from typing import Final

# NOTE: uncomment these to include any queries that this grammar contains:

HIGHLIGHTS_QUERY: Final[str]
INJECTIONS_QUERY: Final[str]
LOCALS_QUERY: Final[str]
TAGS_QUERY: Final[str]

def language() -> object: ...
