/* Automatically @generated by tree-sitter v0.25.7 */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 2739
#define LARGE_STATE_COUNT 161
#define SYMBOL_COUNT 291
#define ALIAS_COUNT 10
#define TOKEN_COUNT 136
#define EXTERNAL_TOKEN_COUNT 10
#define FIELD_COUNT 27
#define MAX_ALIAS_SEQUENCE_LENGTH 10
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 125
#define SUPERTYPE_COUNT 0

enum ts_symbol_identifiers {
  sym_identifier = 1,
  aux_sym_block_token1 = 2,
  anon_sym_RPAREN = 3,
  anon_sym_LPAREN = 4,
  anon_sym_LBRACE = 5,
  anon_sym_RBRACE = 6,
  aux_sym_subroutine_block_token1 = 7,
  aux_sym_if_block_token1 = 8,
  aux_sym_elif_block_token1 = 9,
  aux_sym_while_block_token1 = 10,
  aux_sym_while_block_token2 = 11,
  aux_sym__repeat_block_pre_condition_format_token1 = 12,
  anon_sym_EQ = 13,
  anon_sym_PLUS = 14,
  anon_sym_DASH = 15,
  anon_sym_TILDE = 16,
  anon_sym_BANG = 17,
  anon_sym_PLUS_PLUS = 18,
  anon_sym_DASH_DASH = 19,
  anon_sym_CARET_CARET = 20,
  anon_sym_PIPE_PIPE = 21,
  anon_sym_STAR = 22,
  anon_sym_SLASH = 23,
  anon_sym_PERCENT = 24,
  anon_sym_PIPE = 25,
  anon_sym_CARET = 26,
  anon_sym_EQ_EQ = 27,
  anon_sym_BANG_EQ = 28,
  anon_sym_GT = 29,
  anon_sym_GT_EQ = 30,
  anon_sym_LT_EQ = 31,
  anon_sym_LT = 32,
  anon_sym_LT_LT = 33,
  anon_sym_GT_GT = 34,
  anon_sym_DOT_DOT = 35,
  anon_sym_AMP_AMP = 36,
  anon_sym_AMP = 37,
  anon_sym_COLON = 38,
  aux_sym_macro_definition_token1 = 39,
  aux_sym_parameter_declaration_token1 = 40,
  aux_sym_parameter_declaration_token2 = 41,
  aux_sym_parameter_declaration_token3 = 42,
  aux_sym_parameter_declaration_token4 = 43,
  aux_sym_subroutine_call_expression_token1 = 44,
  aux_sym_macro_definition_command_token1 = 45,
  aux_sym_macro_definition_command_token2 = 46,
  aux_sym_macro_definition_command_token3 = 47,
  aux_sym__host_command_with_default_format_token1 = 48,
  anon_sym_DOT = 49,
  anon_sym_COMMA = 50,
  aux_sym__host_command_argument_list_default_format_token1 = 51,
  aux_sym__hll_non_default_format_var_command_token1 = 52,
  aux_sym__hll_non_default_format_var_command_token2 = 53,
  aux_sym__hll_non_default_format_var_command_token3 = 54,
  aux_sym__hll_non_default_format_var_command_token4 = 55,
  aux_sym__hll_non_default_format_var_command_token5 = 56,
  aux_sym__hll_var_definition_command_token1 = 57,
  aux_sym__var_command_with_non_hll_format_token1 = 58,
  aux_sym__var_command_with_non_hll_format_token2 = 59,
  aux_sym__var_command_with_non_hll_format_token3 = 60,
  aux_sym__hll_var_definition_command_expression_token1 = 61,
  aux_sym__hll_var_definition_command_expression_token2 = 62,
  anon_sym_QMARK = 63,
  anon_sym_BSLASH = 64,
  anon_sym_DASH_GT = 65,
  anon_sym_LBRACK = 66,
  anon_sym_RBRACK = 67,
  anon_sym_sizeof = 68,
  anon_sym_STAR_EQ = 69,
  anon_sym_SLASH_EQ = 70,
  anon_sym_PERCENT_EQ = 71,
  anon_sym_PLUS_EQ = 72,
  anon_sym_DASH_EQ = 73,
  anon_sym_LT_LT_EQ = 74,
  anon_sym_GT_GT_EQ = 75,
  anon_sym_AMP_EQ = 76,
  anon_sym_CARET_EQ = 77,
  anon_sym_PIPE_EQ = 78,
  anon_sym_L_SQUOTE = 79,
  anon_sym_u_SQUOTE = 80,
  anon_sym_U_SQUOTE = 81,
  anon_sym_u8_SQUOTE = 82,
  anon_sym_SQUOTE = 83,
  aux_sym_hll_char_literal_token1 = 84,
  anon_sym_L_DQUOTE = 85,
  anon_sym_u_DQUOTE = 86,
  anon_sym_U_DQUOTE = 87,
  anon_sym_u8_DQUOTE = 88,
  anon_sym_DQUOTE = 89,
  aux_sym_hll_string_literal_token1 = 90,
  anon_sym_const = 91,
  anon_sym_volatile = 92,
  anon_sym_struct = 93,
  anon_sym_union = 94,
  anon_sym_enum = 95,
  anon_sym_signed = 96,
  anon_sym_unsigned = 97,
  anon_sym_long = 98,
  anon_sym_short = 99,
  sym_hll_primitive_type = 100,
  sym_hll_escape_sequence = 101,
  aux_sym__hll_expression_separator_token1 = 102,
  aux_sym__hll_expression_separator_token2 = 103,
  anon_sym_COLON_COLON_COLON = 104,
  anon_sym_COLON_COLON = 105,
  sym_access_class = 106,
  aux_sym_integer_token1 = 107,
  aux_sym_integer_token2 = 108,
  sym_float = 109,
  aux_sym_bitmask_token1 = 110,
  aux_sym_bitmask_token2 = 111,
  aux_sym_string_token1 = 112,
  aux_sym_string_token2 = 113,
  aux_sym_character_token1 = 114,
  aux_sym_character_token2 = 115,
  aux_sym_symbol_token1 = 116,
  aux_sym_symbol_token2 = 117,
  aux_sym_symbol_token3 = 118,
  sym__composed_name = 119,
  sym_time = 120,
  sym_frequency = 121,
  sym_percentage = 122,
  sym_file_handle = 123,
  sym_comment = 124,
  sym__blank_line = 125,
  sym__line_continuation = 126,
  sym__blank = 127,
  sym_label_identifier = 128,
  sym__and_operator_pre_hook = 129,
  sym__decimal_number = 130,
  sym__decimal_number_pre_hook = 131,
  sym_hll_number_literal = 132,
  sym__hll_number_literal_pre_hook = 133,
  sym_path = 134,
  sym__ampersand_char = 135,
  sym_script = 136,
  sym_block = 137,
  sym__statement = 138,
  sym__expression = 139,
  sym__parenthesized_expression = 140,
  sym__path_expression = 141,
  sym_labeled_expression = 142,
  sym_subroutine_block = 143,
  sym_if_block = 144,
  sym_elif_block = 145,
  sym_else_block = 146,
  sym_while_block = 147,
  sym_repeat_block = 148,
  sym__repeat_block_pre_condition_format = 149,
  sym__repeat_block_post_condition_format = 150,
  sym_macro_assignment_expression = 151,
  sym_unary_expression = 152,
  sym__unary_expression = 153,
  sym__prefix_postfix_expression = 154,
  sym_binary_expression = 155,
  sym__binary_expression = 156,
  sym__and_expression = 157,
  sym_assignment_expression = 158,
  sym__label = 159,
  sym_macro_definition = 160,
  sym_parameter_declaration = 161,
  sym_subroutine_call_expression = 162,
  sym_macro = 163,
  sym_recursive_macro_expansion = 164,
  sym_macro_definition_command = 165,
  sym_command_expression = 166,
  sym__host_command = 167,
  sym__host_command_with_default_format = 168,
  sym__other_command = 169,
  sym__command_identifier = 170,
  sym__command_arguments = 171,
  sym__host_command_argument_list_default_format = 172,
  sym__command_argument = 173,
  sym_option_expression = 174,
  sym_format_expression = 175,
  sym_call_expression = 176,
  sym__function_identifier = 177,
  sym_argument_list = 178,
  sym__var_command = 179,
  sym__hll_var_command = 180,
  sym__hll_default_format_var_command = 181,
  sym__hll_non_default_format_var_command = 182,
  sym__non_hll_var_command = 183,
  sym__hll_var_definition_command = 184,
  sym__hll_var_command_with_default_format = 185,
  sym__var_command_with_non_hll_format = 186,
  sym__hll_var_definition_command_expression = 187,
  sym__hll_var_if_command = 188,
  sym__hll_var_command_argument_list_default_format = 189,
  sym__hll_var_command_argument_list_non_default_format = 190,
  sym__hll_var_pattern_command_argument_list_format = 191,
  sym__var_command_argument_list_non_hll_format = 192,
  sym__var_definition_command_argument_list_format = 193,
  sym__hll_var_call_expression_function = 194,
  sym__hll_var_call_expression = 195,
  sym_hll_var_call_expression_argument_list = 196,
  sym_hll_type_pattern = 197,
  sym_trace32_hll_variable = 198,
  sym__hll_expression = 199,
  sym_hll_call_expression = 200,
  sym_hll_field_expression = 201,
  sym_hll_subscript_expression = 202,
  sym_hll_cast_expression = 203,
  sym_hll_sizeof_expression = 204,
  sym_hll_pointer_expression = 205,
  sym_hll_comma_expression = 206,
  sym_hll_update_expression = 207,
  sym_hll_unary_expression = 208,
  sym_hll_binary_expression = 209,
  sym_hll_range_expression = 210,
  sym_hll_assignment_expression = 211,
  sym__hll_assignment_left_expression = 212,
  sym_hll_conditional_expression = 213,
  sym_hll_char_literal = 214,
  sym_hll_string_literal = 215,
  sym_hll_parenthesized_expression = 216,
  sym_hll_type_descriptor = 217,
  sym_hll_type_qualifier = 218,
  sym__hll_declaration_modifiers = 219,
  sym__hll_type_specifier = 220,
  sym_hll_struct_specifier = 221,
  sym_hll_union_specifier = 222,
  sym_hll_enum_specifier = 223,
  sym_hll_sized_type_specifier = 224,
  sym_hll_declaration = 225,
  sym__hll_declaration_specifiers = 226,
  sym__hll_declarator = 227,
  sym_hll_pointer_declarator = 228,
  sym_hll_function_declarator = 229,
  sym_hll_array_declarator = 230,
  sym_hll_parenthesized_declarator = 231,
  sym__hll_pointer_declarator_specifier = 232,
  sym__hll_abstract_declarator = 233,
  sym_hll_abstract_pointer_declarator = 234,
  sym_hll_abstract_function_declarator = 235,
  sym_hll_abstract_array_declarator = 236,
  sym_hll_abstract_parenthesized_declarator = 237,
  sym_hll_parameter_list = 238,
  sym_hll_parameter_declaration = 239,
  sym_hll_argument_list = 240,
  sym__hll_type_identifier = 241,
  sym__hll_field_identifier = 242,
  sym__literal = 243,
  sym__address_expression = 244,
  sym_memory_space = 245,
  sym_address = 246,
  sym__parenthesized_address_expression = 247,
  sym__address_binary_expression = 248,
  sym_integer = 249,
  sym_bitmask = 250,
  sym_string = 251,
  sym_character = 252,
  sym_symbol = 253,
  sym__special_command_operator = 254,
  sym__file_dialog = 255,
  aux_sym_script_repeat1 = 256,
  aux_sym_block_repeat1 = 257,
  aux_sym_block_repeat2 = 258,
  aux_sym__path_expression_repeat1 = 259,
  aux_sym__path_expression_repeat2 = 260,
  aux_sym_subroutine_block_repeat1 = 261,
  aux_sym_if_block_repeat1 = 262,
  aux_sym_if_block_repeat2 = 263,
  aux_sym_if_block_repeat3 = 264,
  aux_sym_macro_definition_repeat1 = 265,
  aux_sym__command_identifier_repeat1 = 266,
  aux_sym__command_arguments_repeat1 = 267,
  aux_sym__command_arguments_repeat2 = 268,
  aux_sym__command_arguments_repeat3 = 269,
  aux_sym__host_command_argument_list_default_format_repeat1 = 270,
  aux_sym__host_command_argument_list_default_format_repeat2 = 271,
  aux_sym_format_expression_repeat1 = 272,
  aux_sym__function_identifier_repeat1 = 273,
  aux_sym_argument_list_repeat1 = 274,
  aux_sym__hll_var_command_argument_list_default_format_repeat1 = 275,
  aux_sym__hll_var_command_argument_list_default_format_repeat2 = 276,
  aux_sym__hll_var_pattern_command_argument_list_format_repeat1 = 277,
  aux_sym__hll_var_pattern_command_argument_list_format_repeat2 = 278,
  aux_sym__var_command_argument_list_non_hll_format_repeat1 = 279,
  aux_sym_hll_type_pattern_repeat1 = 280,
  aux_sym_hll_string_literal_repeat1 = 281,
  aux_sym_hll_type_descriptor_repeat1 = 282,
  aux_sym_hll_type_descriptor_repeat2 = 283,
  aux_sym_hll_sized_type_specifier_repeat1 = 284,
  aux_sym__hll_declaration_specifiers_repeat1 = 285,
  aux_sym__hll_declaration_specifiers_repeat2 = 286,
  aux_sym__hll_pointer_declarator_specifier_repeat1 = 287,
  aux_sym_hll_parameter_list_repeat1 = 288,
  aux_sym_hll_argument_list_repeat1 = 289,
  aux_sym_string_repeat1 = 290,
  anon_alias_sym_IF = 291,
  anon_alias_sym_command = 292,
  anon_alias_sym_function = 293,
  alias_sym_hll_field_identifier = 294,
  alias_sym_hll_type_identifier = 295,
  anon_alias_sym_module = 296,
  anon_alias_sym_name = 297,
  anon_alias_sym_subcommand = 298,
  anon_alias_sym_subfunction = 299,
  anon_alias_sym_type = 300,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_identifier] = "identifier",
  [aux_sym_block_token1] = "block_token1",
  [anon_sym_RPAREN] = ")",
  [anon_sym_LPAREN] = "(",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [aux_sym_subroutine_block_token1] = "identifier",
  [aux_sym_if_block_token1] = "identifier",
  [aux_sym_elif_block_token1] = "identifier",
  [aux_sym_while_block_token1] = "identifier",
  [aux_sym_while_block_token2] = "identifier",
  [aux_sym__repeat_block_pre_condition_format_token1] = "identifier",
  [anon_sym_EQ] = "=",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_TILDE] = "~",
  [anon_sym_BANG] = "!",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_DASH_DASH] = "--",
  [anon_sym_CARET_CARET] = "^^",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_STAR] = "*",
  [anon_sym_SLASH] = "/",
  [anon_sym_PERCENT] = "%",
  [anon_sym_PIPE] = "|",
  [anon_sym_CARET] = "^",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_GT] = ">",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_LT] = "<",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_DOT_DOT] = "..",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_AMP] = "&",
  [anon_sym_COLON] = ":",
  [aux_sym_macro_definition_token1] = "macro_definition_token1",
  [aux_sym_parameter_declaration_token1] = "identifier",
  [aux_sym_parameter_declaration_token2] = "identifier",
  [aux_sym_parameter_declaration_token3] = "identifier",
  [aux_sym_parameter_declaration_token4] = "identifier",
  [aux_sym_subroutine_call_expression_token1] = "identifier",
  [aux_sym_macro_definition_command_token1] = "GLOBAL",
  [aux_sym_macro_definition_command_token2] = "LOCAL",
  [aux_sym_macro_definition_command_token3] = "PRIVATE",
  [aux_sym__host_command_with_default_format_token1] = "OS",
  [anon_sym_DOT] = ".",
  [anon_sym_COMMA] = ",",
  [aux_sym__host_command_argument_list_default_format_token1] = "_host_command_argument_list_default_format_token1",
  [aux_sym__hll_non_default_format_var_command_token1] = "Var",
  [aux_sym__hll_non_default_format_var_command_token2] = "AddSticker",
  [aux_sym__hll_non_default_format_var_command_token3] = "EXPORT",
  [aux_sym__hll_non_default_format_var_command_token4] = "WRITE",
  [aux_sym__hll_non_default_format_var_command_token5] = "PATtern",
  [aux_sym__hll_var_definition_command_token1] = "_hll_var_definition_command_token1",
  [aux_sym__var_command_with_non_hll_format_token1] = "Local",
  [aux_sym__var_command_with_non_hll_format_token2] = "Ref",
  [aux_sym__var_command_with_non_hll_format_token3] = "TREE",
  [aux_sym__hll_var_definition_command_expression_token1] = "NEWLOCAL",
  [aux_sym__hll_var_definition_command_expression_token2] = "NEWGLOBAL",
  [anon_sym_QMARK] = "\?",
  [anon_sym_BSLASH] = "\\",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_sizeof] = "sizeof",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_LT_LT_EQ] = "<<=",
  [anon_sym_GT_GT_EQ] = ">>=",
  [anon_sym_AMP_EQ] = "&=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_L_SQUOTE] = "L'",
  [anon_sym_u_SQUOTE] = "u'",
  [anon_sym_U_SQUOTE] = "U'",
  [anon_sym_u8_SQUOTE] = "u8'",
  [anon_sym_SQUOTE] = "'",
  [aux_sym_hll_char_literal_token1] = "character",
  [anon_sym_L_DQUOTE] = "L\"",
  [anon_sym_u_DQUOTE] = "u\"",
  [anon_sym_U_DQUOTE] = "U\"",
  [anon_sym_u8_DQUOTE] = "u8\"",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym_hll_string_literal_token1] = "hll_string_literal_token1",
  [anon_sym_const] = "const",
  [anon_sym_volatile] = "volatile",
  [anon_sym_struct] = "struct",
  [anon_sym_union] = "union",
  [anon_sym_enum] = "enum",
  [anon_sym_signed] = "signed",
  [anon_sym_unsigned] = "unsigned",
  [anon_sym_long] = "long",
  [anon_sym_short] = "short",
  [sym_hll_primitive_type] = "hll_primitive_type",
  [sym_hll_escape_sequence] = "hll_escape_sequence",
  [aux_sym__hll_expression_separator_token1] = "_hll_expression_separator_token1",
  [aux_sym__hll_expression_separator_token2] = "_hll_expression_separator_token2",
  [anon_sym_COLON_COLON_COLON] = ":::",
  [anon_sym_COLON_COLON] = "::",
  [sym_access_class] = "access_class",
  [aux_sym_integer_token1] = "integer_token1",
  [aux_sym_integer_token2] = "integer_token2",
  [sym_float] = "float",
  [aux_sym_bitmask_token1] = "bitmask_token1",
  [aux_sym_bitmask_token2] = "bitmask_token2",
  [aux_sym_string_token1] = "string_token1",
  [aux_sym_string_token2] = "string_token2",
  [aux_sym_character_token1] = "character_token1",
  [aux_sym_character_token2] = "character_token2",
  [aux_sym_symbol_token1] = "symbol_token1",
  [aux_sym_symbol_token2] = "symbol_token2",
  [aux_sym_symbol_token3] = "symbol_token3",
  [sym__composed_name] = "identifier",
  [sym_time] = "time",
  [sym_frequency] = "frequency",
  [sym_percentage] = "percentage",
  [sym_file_handle] = "file_handle",
  [sym_comment] = "comment",
  [sym__blank_line] = "_blank_line",
  [sym__line_continuation] = "_line_continuation",
  [sym__blank] = "_blank",
  [sym_label_identifier] = "identifier",
  [sym__and_operator_pre_hook] = "_and_operator_pre_hook",
  [sym__decimal_number] = "_decimal_number",
  [sym__decimal_number_pre_hook] = "_decimal_number_pre_hook",
  [sym_hll_number_literal] = "hll_number_literal",
  [sym__hll_number_literal_pre_hook] = "_hll_number_literal_pre_hook",
  [sym_path] = "path",
  [sym__ampersand_char] = "_ampersand_char",
  [sym_script] = "script",
  [sym_block] = "block",
  [sym__statement] = "_statement",
  [sym__expression] = "_expression",
  [sym__parenthesized_expression] = "_parenthesized_expression",
  [sym__path_expression] = "_path_expression",
  [sym_labeled_expression] = "labeled_expression",
  [sym_subroutine_block] = "subroutine_block",
  [sym_if_block] = "if_block",
  [sym_elif_block] = "elif_block",
  [sym_else_block] = "else_block",
  [sym_while_block] = "while_block",
  [sym_repeat_block] = "repeat_block",
  [sym__repeat_block_pre_condition_format] = "_repeat_block_pre_condition_format",
  [sym__repeat_block_post_condition_format] = "_repeat_block_post_condition_format",
  [sym_macro_assignment_expression] = "assignment_expression",
  [sym_unary_expression] = "unary_expression",
  [sym__unary_expression] = "_unary_expression",
  [sym__prefix_postfix_expression] = "_prefix_postfix_expression",
  [sym_binary_expression] = "binary_expression",
  [sym__binary_expression] = "_binary_expression",
  [sym__and_expression] = "_and_expression",
  [sym_assignment_expression] = "assignment_expression",
  [sym__label] = "_label",
  [sym_macro_definition] = "macro_definition",
  [sym_parameter_declaration] = "parameter_declaration",
  [sym_subroutine_call_expression] = "subroutine_call_expression",
  [sym_macro] = "macro",
  [sym_recursive_macro_expansion] = "recursive_macro_expansion",
  [sym_macro_definition_command] = "identifier",
  [sym_command_expression] = "command_expression",
  [sym__host_command] = "_host_command",
  [sym__host_command_with_default_format] = "identifier",
  [sym__other_command] = "_other_command",
  [sym__command_identifier] = "identifier",
  [sym__command_arguments] = "argument_list",
  [sym__host_command_argument_list_default_format] = "argument_list",
  [sym__command_argument] = "_command_argument",
  [sym_option_expression] = "option_expression",
  [sym_format_expression] = "format_expression",
  [sym_call_expression] = "call_expression",
  [sym__function_identifier] = "identifier",
  [sym_argument_list] = "argument_list",
  [sym__var_command] = "_var_command",
  [sym__hll_var_command] = "_hll_var_command",
  [sym__hll_default_format_var_command] = "_hll_default_format_var_command",
  [sym__hll_non_default_format_var_command] = "_hll_non_default_format_var_command",
  [sym__non_hll_var_command] = "_non_hll_var_command",
  [sym__hll_var_definition_command] = "_hll_var_definition_command",
  [sym__hll_var_command_with_default_format] = "identifier",
  [sym__var_command_with_non_hll_format] = "identifier",
  [sym__hll_var_definition_command_expression] = "identifier",
  [sym__hll_var_if_command] = "identifier",
  [sym__hll_var_command_argument_list_default_format] = "argument_list",
  [sym__hll_var_command_argument_list_non_default_format] = "argument_list",
  [sym__hll_var_pattern_command_argument_list_format] = "argument_list",
  [sym__var_command_argument_list_non_hll_format] = "argument_list",
  [sym__var_definition_command_argument_list_format] = "argument_list",
  [sym__hll_var_call_expression_function] = "identifier",
  [sym__hll_var_call_expression] = "_hll_var_call_expression",
  [sym_hll_var_call_expression_argument_list] = "argument_list",
  [sym_hll_type_pattern] = "hll_type_pattern",
  [sym_trace32_hll_variable] = "trace32_hll_variable",
  [sym__hll_expression] = "_hll_expression",
  [sym_hll_call_expression] = "hll_call_expression",
  [sym_hll_field_expression] = "hll_field_expression",
  [sym_hll_subscript_expression] = "hll_subscript_expression",
  [sym_hll_cast_expression] = "hll_cast_expression",
  [sym_hll_sizeof_expression] = "hll_sizeof_expression",
  [sym_hll_pointer_expression] = "hll_pointer_expression",
  [sym_hll_comma_expression] = "hll_comma_expression",
  [sym_hll_update_expression] = "hll_update_expression",
  [sym_hll_unary_expression] = "hll_unary_expression",
  [sym_hll_binary_expression] = "hll_binary_expression",
  [sym_hll_range_expression] = "hll_range_expression",
  [sym_hll_assignment_expression] = "hll_assignment_expression",
  [sym__hll_assignment_left_expression] = "_hll_assignment_left_expression",
  [sym_hll_conditional_expression] = "hll_conditional_expression",
  [sym_hll_char_literal] = "hll_char_literal",
  [sym_hll_string_literal] = "hll_string_literal",
  [sym_hll_parenthesized_expression] = "hll_parenthesized_expression",
  [sym_hll_type_descriptor] = "hll_type_descriptor",
  [sym_hll_type_qualifier] = "hll_type_qualifier",
  [sym__hll_declaration_modifiers] = "_hll_declaration_modifiers",
  [sym__hll_type_specifier] = "_hll_type_specifier",
  [sym_hll_struct_specifier] = "hll_struct_specifier",
  [sym_hll_union_specifier] = "hll_union_specifier",
  [sym_hll_enum_specifier] = "hll_enum_specifier",
  [sym_hll_sized_type_specifier] = "hll_sized_type_specifier",
  [sym_hll_declaration] = "hll_declaration",
  [sym__hll_declaration_specifiers] = "_hll_declaration_specifiers",
  [sym__hll_declarator] = "_hll_declarator",
  [sym_hll_pointer_declarator] = "hll_pointer_declarator",
  [sym_hll_function_declarator] = "hll_function_declarator",
  [sym_hll_array_declarator] = "hll_array_declarator",
  [sym_hll_parenthesized_declarator] = "hll_parenthesized_declarator",
  [sym__hll_pointer_declarator_specifier] = "_hll_pointer_declarator_specifier",
  [sym__hll_abstract_declarator] = "_hll_abstract_declarator",
  [sym_hll_abstract_pointer_declarator] = "hll_abstract_pointer_declarator",
  [sym_hll_abstract_function_declarator] = "hll_abstract_function_declarator",
  [sym_hll_abstract_array_declarator] = "hll_abstract_array_declarator",
  [sym_hll_abstract_parenthesized_declarator] = "hll_abstract_parenthesized_declarator",
  [sym_hll_parameter_list] = "hll_parameter_list",
  [sym_hll_parameter_declaration] = "hll_parameter_declaration",
  [sym_hll_argument_list] = "hll_argument_list",
  [sym__hll_type_identifier] = "_hll_type_identifier",
  [sym__hll_field_identifier] = "_hll_field_identifier",
  [sym__literal] = "_literal",
  [sym__address_expression] = "_address_expression",
  [sym_memory_space] = "memory_space",
  [sym_address] = "address",
  [sym__parenthesized_address_expression] = "_parenthesized_address_expression",
  [sym__address_binary_expression] = "binary_expression",
  [sym_integer] = "integer",
  [sym_bitmask] = "bitmask",
  [sym_string] = "string",
  [sym_character] = "character",
  [sym_symbol] = "symbol",
  [sym__special_command_operator] = "literal",
  [sym__file_dialog] = "_file_dialog",
  [aux_sym_script_repeat1] = "script_repeat1",
  [aux_sym_block_repeat1] = "block_repeat1",
  [aux_sym_block_repeat2] = "block_repeat2",
  [aux_sym__path_expression_repeat1] = "_path_expression_repeat1",
  [aux_sym__path_expression_repeat2] = "_path_expression_repeat2",
  [aux_sym_subroutine_block_repeat1] = "subroutine_block_repeat1",
  [aux_sym_if_block_repeat1] = "if_block_repeat1",
  [aux_sym_if_block_repeat2] = "if_block_repeat2",
  [aux_sym_if_block_repeat3] = "if_block_repeat3",
  [aux_sym_macro_definition_repeat1] = "macro_definition_repeat1",
  [aux_sym__command_identifier_repeat1] = "_command_identifier_repeat1",
  [aux_sym__command_arguments_repeat1] = "_command_arguments_repeat1",
  [aux_sym__command_arguments_repeat2] = "_command_arguments_repeat2",
  [aux_sym__command_arguments_repeat3] = "_command_arguments_repeat3",
  [aux_sym__host_command_argument_list_default_format_repeat1] = "_host_command_argument_list_default_format_repeat1",
  [aux_sym__host_command_argument_list_default_format_repeat2] = "_host_command_argument_list_default_format_repeat2",
  [aux_sym_format_expression_repeat1] = "format_expression_repeat1",
  [aux_sym__function_identifier_repeat1] = "_function_identifier_repeat1",
  [aux_sym_argument_list_repeat1] = "argument_list_repeat1",
  [aux_sym__hll_var_command_argument_list_default_format_repeat1] = "_hll_var_command_argument_list_default_format_repeat1",
  [aux_sym__hll_var_command_argument_list_default_format_repeat2] = "_hll_var_command_argument_list_default_format_repeat2",
  [aux_sym__hll_var_pattern_command_argument_list_format_repeat1] = "_hll_var_pattern_command_argument_list_format_repeat1",
  [aux_sym__hll_var_pattern_command_argument_list_format_repeat2] = "_hll_var_pattern_command_argument_list_format_repeat2",
  [aux_sym__var_command_argument_list_non_hll_format_repeat1] = "_var_command_argument_list_non_hll_format_repeat1",
  [aux_sym_hll_type_pattern_repeat1] = "hll_type_pattern_repeat1",
  [aux_sym_hll_string_literal_repeat1] = "hll_string_literal_repeat1",
  [aux_sym_hll_type_descriptor_repeat1] = "hll_type_descriptor_repeat1",
  [aux_sym_hll_type_descriptor_repeat2] = "hll_type_descriptor_repeat2",
  [aux_sym_hll_sized_type_specifier_repeat1] = "hll_sized_type_specifier_repeat1",
  [aux_sym__hll_declaration_specifiers_repeat1] = "_hll_declaration_specifiers_repeat1",
  [aux_sym__hll_declaration_specifiers_repeat2] = "_hll_declaration_specifiers_repeat2",
  [aux_sym__hll_pointer_declarator_specifier_repeat1] = "_hll_pointer_declarator_specifier_repeat1",
  [aux_sym_hll_parameter_list_repeat1] = "hll_parameter_list_repeat1",
  [aux_sym_hll_argument_list_repeat1] = "hll_argument_list_repeat1",
  [aux_sym_string_repeat1] = "string_repeat1",
  [anon_alias_sym_IF] = "IF",
  [anon_alias_sym_command] = "command",
  [anon_alias_sym_function] = "function",
  [alias_sym_hll_field_identifier] = "hll_field_identifier",
  [alias_sym_hll_type_identifier] = "hll_type_identifier",
  [anon_alias_sym_module] = "module",
  [anon_alias_sym_name] = "name",
  [anon_alias_sym_subcommand] = "subcommand",
  [anon_alias_sym_subfunction] = "subfunction",
  [anon_alias_sym_type] = "type",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_identifier] = sym_identifier,
  [aux_sym_block_token1] = aux_sym_block_token1,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [aux_sym_subroutine_block_token1] = sym_identifier,
  [aux_sym_if_block_token1] = sym_identifier,
  [aux_sym_elif_block_token1] = sym_identifier,
  [aux_sym_while_block_token1] = sym_identifier,
  [aux_sym_while_block_token2] = sym_identifier,
  [aux_sym__repeat_block_pre_condition_format_token1] = sym_identifier,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [anon_sym_CARET_CARET] = anon_sym_CARET_CARET,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_COLON] = anon_sym_COLON,
  [aux_sym_macro_definition_token1] = aux_sym_macro_definition_token1,
  [aux_sym_parameter_declaration_token1] = sym_identifier,
  [aux_sym_parameter_declaration_token2] = sym_identifier,
  [aux_sym_parameter_declaration_token3] = sym_identifier,
  [aux_sym_parameter_declaration_token4] = sym_identifier,
  [aux_sym_subroutine_call_expression_token1] = sym_identifier,
  [aux_sym_macro_definition_command_token1] = aux_sym_macro_definition_command_token1,
  [aux_sym_macro_definition_command_token2] = aux_sym_macro_definition_command_token2,
  [aux_sym_macro_definition_command_token3] = aux_sym_macro_definition_command_token3,
  [aux_sym__host_command_with_default_format_token1] = aux_sym__host_command_with_default_format_token1,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [aux_sym__host_command_argument_list_default_format_token1] = aux_sym__host_command_argument_list_default_format_token1,
  [aux_sym__hll_non_default_format_var_command_token1] = aux_sym__hll_non_default_format_var_command_token1,
  [aux_sym__hll_non_default_format_var_command_token2] = aux_sym__hll_non_default_format_var_command_token2,
  [aux_sym__hll_non_default_format_var_command_token3] = aux_sym__hll_non_default_format_var_command_token3,
  [aux_sym__hll_non_default_format_var_command_token4] = aux_sym__hll_non_default_format_var_command_token4,
  [aux_sym__hll_non_default_format_var_command_token5] = aux_sym__hll_non_default_format_var_command_token5,
  [aux_sym__hll_var_definition_command_token1] = aux_sym__hll_var_definition_command_token1,
  [aux_sym__var_command_with_non_hll_format_token1] = aux_sym__var_command_with_non_hll_format_token1,
  [aux_sym__var_command_with_non_hll_format_token2] = aux_sym__var_command_with_non_hll_format_token2,
  [aux_sym__var_command_with_non_hll_format_token3] = aux_sym__var_command_with_non_hll_format_token3,
  [aux_sym__hll_var_definition_command_expression_token1] = aux_sym__hll_var_definition_command_expression_token1,
  [aux_sym__hll_var_definition_command_expression_token2] = aux_sym__hll_var_definition_command_expression_token2,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_BSLASH] = anon_sym_BSLASH,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_sizeof] = anon_sym_sizeof,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_LT_LT_EQ] = anon_sym_LT_LT_EQ,
  [anon_sym_GT_GT_EQ] = anon_sym_GT_GT_EQ,
  [anon_sym_AMP_EQ] = anon_sym_AMP_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_L_SQUOTE] = anon_sym_L_SQUOTE,
  [anon_sym_u_SQUOTE] = anon_sym_u_SQUOTE,
  [anon_sym_U_SQUOTE] = anon_sym_U_SQUOTE,
  [anon_sym_u8_SQUOTE] = anon_sym_u8_SQUOTE,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [aux_sym_hll_char_literal_token1] = sym_character,
  [anon_sym_L_DQUOTE] = anon_sym_L_DQUOTE,
  [anon_sym_u_DQUOTE] = anon_sym_u_DQUOTE,
  [anon_sym_U_DQUOTE] = anon_sym_U_DQUOTE,
  [anon_sym_u8_DQUOTE] = anon_sym_u8_DQUOTE,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym_hll_string_literal_token1] = aux_sym_hll_string_literal_token1,
  [anon_sym_const] = anon_sym_const,
  [anon_sym_volatile] = anon_sym_volatile,
  [anon_sym_struct] = anon_sym_struct,
  [anon_sym_union] = anon_sym_union,
  [anon_sym_enum] = anon_sym_enum,
  [anon_sym_signed] = anon_sym_signed,
  [anon_sym_unsigned] = anon_sym_unsigned,
  [anon_sym_long] = anon_sym_long,
  [anon_sym_short] = anon_sym_short,
  [sym_hll_primitive_type] = sym_hll_primitive_type,
  [sym_hll_escape_sequence] = sym_hll_escape_sequence,
  [aux_sym__hll_expression_separator_token1] = aux_sym__hll_expression_separator_token1,
  [aux_sym__hll_expression_separator_token2] = aux_sym__hll_expression_separator_token2,
  [anon_sym_COLON_COLON_COLON] = anon_sym_COLON_COLON_COLON,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [sym_access_class] = sym_access_class,
  [aux_sym_integer_token1] = aux_sym_integer_token1,
  [aux_sym_integer_token2] = aux_sym_integer_token2,
  [sym_float] = sym_float,
  [aux_sym_bitmask_token1] = aux_sym_bitmask_token1,
  [aux_sym_bitmask_token2] = aux_sym_bitmask_token2,
  [aux_sym_string_token1] = aux_sym_string_token1,
  [aux_sym_string_token2] = aux_sym_string_token2,
  [aux_sym_character_token1] = aux_sym_character_token1,
  [aux_sym_character_token2] = aux_sym_character_token2,
  [aux_sym_symbol_token1] = aux_sym_symbol_token1,
  [aux_sym_symbol_token2] = aux_sym_symbol_token2,
  [aux_sym_symbol_token3] = aux_sym_symbol_token3,
  [sym__composed_name] = sym_identifier,
  [sym_time] = sym_time,
  [sym_frequency] = sym_frequency,
  [sym_percentage] = sym_percentage,
  [sym_file_handle] = sym_file_handle,
  [sym_comment] = sym_comment,
  [sym__blank_line] = sym__blank_line,
  [sym__line_continuation] = sym__line_continuation,
  [sym__blank] = sym__blank,
  [sym_label_identifier] = sym_identifier,
  [sym__and_operator_pre_hook] = sym__and_operator_pre_hook,
  [sym__decimal_number] = sym__decimal_number,
  [sym__decimal_number_pre_hook] = sym__decimal_number_pre_hook,
  [sym_hll_number_literal] = sym_hll_number_literal,
  [sym__hll_number_literal_pre_hook] = sym__hll_number_literal_pre_hook,
  [sym_path] = sym_path,
  [sym__ampersand_char] = sym__ampersand_char,
  [sym_script] = sym_script,
  [sym_block] = sym_block,
  [sym__statement] = sym__statement,
  [sym__expression] = sym__expression,
  [sym__parenthesized_expression] = sym__parenthesized_expression,
  [sym__path_expression] = sym__path_expression,
  [sym_labeled_expression] = sym_labeled_expression,
  [sym_subroutine_block] = sym_subroutine_block,
  [sym_if_block] = sym_if_block,
  [sym_elif_block] = sym_elif_block,
  [sym_else_block] = sym_else_block,
  [sym_while_block] = sym_while_block,
  [sym_repeat_block] = sym_repeat_block,
  [sym__repeat_block_pre_condition_format] = sym__repeat_block_pre_condition_format,
  [sym__repeat_block_post_condition_format] = sym__repeat_block_post_condition_format,
  [sym_macro_assignment_expression] = sym_assignment_expression,
  [sym_unary_expression] = sym_unary_expression,
  [sym__unary_expression] = sym__unary_expression,
  [sym__prefix_postfix_expression] = sym__prefix_postfix_expression,
  [sym_binary_expression] = sym_binary_expression,
  [sym__binary_expression] = sym__binary_expression,
  [sym__and_expression] = sym__and_expression,
  [sym_assignment_expression] = sym_assignment_expression,
  [sym__label] = sym__label,
  [sym_macro_definition] = sym_macro_definition,
  [sym_parameter_declaration] = sym_parameter_declaration,
  [sym_subroutine_call_expression] = sym_subroutine_call_expression,
  [sym_macro] = sym_macro,
  [sym_recursive_macro_expansion] = sym_recursive_macro_expansion,
  [sym_macro_definition_command] = sym_identifier,
  [sym_command_expression] = sym_command_expression,
  [sym__host_command] = sym__host_command,
  [sym__host_command_with_default_format] = sym_identifier,
  [sym__other_command] = sym__other_command,
  [sym__command_identifier] = sym_identifier,
  [sym__command_arguments] = sym_argument_list,
  [sym__host_command_argument_list_default_format] = sym_argument_list,
  [sym__command_argument] = sym__command_argument,
  [sym_option_expression] = sym_option_expression,
  [sym_format_expression] = sym_format_expression,
  [sym_call_expression] = sym_call_expression,
  [sym__function_identifier] = sym_identifier,
  [sym_argument_list] = sym_argument_list,
  [sym__var_command] = sym__var_command,
  [sym__hll_var_command] = sym__hll_var_command,
  [sym__hll_default_format_var_command] = sym__hll_default_format_var_command,
  [sym__hll_non_default_format_var_command] = sym__hll_non_default_format_var_command,
  [sym__non_hll_var_command] = sym__non_hll_var_command,
  [sym__hll_var_definition_command] = sym__hll_var_definition_command,
  [sym__hll_var_command_with_default_format] = sym_identifier,
  [sym__var_command_with_non_hll_format] = sym_identifier,
  [sym__hll_var_definition_command_expression] = sym_identifier,
  [sym__hll_var_if_command] = sym_identifier,
  [sym__hll_var_command_argument_list_default_format] = sym_argument_list,
  [sym__hll_var_command_argument_list_non_default_format] = sym_argument_list,
  [sym__hll_var_pattern_command_argument_list_format] = sym_argument_list,
  [sym__var_command_argument_list_non_hll_format] = sym_argument_list,
  [sym__var_definition_command_argument_list_format] = sym_argument_list,
  [sym__hll_var_call_expression_function] = sym_identifier,
  [sym__hll_var_call_expression] = sym__hll_var_call_expression,
  [sym_hll_var_call_expression_argument_list] = sym_argument_list,
  [sym_hll_type_pattern] = sym_hll_type_pattern,
  [sym_trace32_hll_variable] = sym_trace32_hll_variable,
  [sym__hll_expression] = sym__hll_expression,
  [sym_hll_call_expression] = sym_hll_call_expression,
  [sym_hll_field_expression] = sym_hll_field_expression,
  [sym_hll_subscript_expression] = sym_hll_subscript_expression,
  [sym_hll_cast_expression] = sym_hll_cast_expression,
  [sym_hll_sizeof_expression] = sym_hll_sizeof_expression,
  [sym_hll_pointer_expression] = sym_hll_pointer_expression,
  [sym_hll_comma_expression] = sym_hll_comma_expression,
  [sym_hll_update_expression] = sym_hll_update_expression,
  [sym_hll_unary_expression] = sym_hll_unary_expression,
  [sym_hll_binary_expression] = sym_hll_binary_expression,
  [sym_hll_range_expression] = sym_hll_range_expression,
  [sym_hll_assignment_expression] = sym_hll_assignment_expression,
  [sym__hll_assignment_left_expression] = sym__hll_assignment_left_expression,
  [sym_hll_conditional_expression] = sym_hll_conditional_expression,
  [sym_hll_char_literal] = sym_hll_char_literal,
  [sym_hll_string_literal] = sym_hll_string_literal,
  [sym_hll_parenthesized_expression] = sym_hll_parenthesized_expression,
  [sym_hll_type_descriptor] = sym_hll_type_descriptor,
  [sym_hll_type_qualifier] = sym_hll_type_qualifier,
  [sym__hll_declaration_modifiers] = sym__hll_declaration_modifiers,
  [sym__hll_type_specifier] = sym__hll_type_specifier,
  [sym_hll_struct_specifier] = sym_hll_struct_specifier,
  [sym_hll_union_specifier] = sym_hll_union_specifier,
  [sym_hll_enum_specifier] = sym_hll_enum_specifier,
  [sym_hll_sized_type_specifier] = sym_hll_sized_type_specifier,
  [sym_hll_declaration] = sym_hll_declaration,
  [sym__hll_declaration_specifiers] = sym__hll_declaration_specifiers,
  [sym__hll_declarator] = sym__hll_declarator,
  [sym_hll_pointer_declarator] = sym_hll_pointer_declarator,
  [sym_hll_function_declarator] = sym_hll_function_declarator,
  [sym_hll_array_declarator] = sym_hll_array_declarator,
  [sym_hll_parenthesized_declarator] = sym_hll_parenthesized_declarator,
  [sym__hll_pointer_declarator_specifier] = sym__hll_pointer_declarator_specifier,
  [sym__hll_abstract_declarator] = sym__hll_abstract_declarator,
  [sym_hll_abstract_pointer_declarator] = sym_hll_abstract_pointer_declarator,
  [sym_hll_abstract_function_declarator] = sym_hll_abstract_function_declarator,
  [sym_hll_abstract_array_declarator] = sym_hll_abstract_array_declarator,
  [sym_hll_abstract_parenthesized_declarator] = sym_hll_abstract_parenthesized_declarator,
  [sym_hll_parameter_list] = sym_hll_parameter_list,
  [sym_hll_parameter_declaration] = sym_hll_parameter_declaration,
  [sym_hll_argument_list] = sym_hll_argument_list,
  [sym__hll_type_identifier] = sym__hll_type_identifier,
  [sym__hll_field_identifier] = sym__hll_field_identifier,
  [sym__literal] = sym__literal,
  [sym__address_expression] = sym__address_expression,
  [sym_memory_space] = sym_memory_space,
  [sym_address] = sym_address,
  [sym__parenthesized_address_expression] = sym__parenthesized_address_expression,
  [sym__address_binary_expression] = sym_binary_expression,
  [sym_integer] = sym_integer,
  [sym_bitmask] = sym_bitmask,
  [sym_string] = sym_string,
  [sym_character] = sym_character,
  [sym_symbol] = sym_symbol,
  [sym__special_command_operator] = sym__special_command_operator,
  [sym__file_dialog] = sym__file_dialog,
  [aux_sym_script_repeat1] = aux_sym_script_repeat1,
  [aux_sym_block_repeat1] = aux_sym_block_repeat1,
  [aux_sym_block_repeat2] = aux_sym_block_repeat2,
  [aux_sym__path_expression_repeat1] = aux_sym__path_expression_repeat1,
  [aux_sym__path_expression_repeat2] = aux_sym__path_expression_repeat2,
  [aux_sym_subroutine_block_repeat1] = aux_sym_subroutine_block_repeat1,
  [aux_sym_if_block_repeat1] = aux_sym_if_block_repeat1,
  [aux_sym_if_block_repeat2] = aux_sym_if_block_repeat2,
  [aux_sym_if_block_repeat3] = aux_sym_if_block_repeat3,
  [aux_sym_macro_definition_repeat1] = aux_sym_macro_definition_repeat1,
  [aux_sym__command_identifier_repeat1] = aux_sym__command_identifier_repeat1,
  [aux_sym__command_arguments_repeat1] = aux_sym__command_arguments_repeat1,
  [aux_sym__command_arguments_repeat2] = aux_sym__command_arguments_repeat2,
  [aux_sym__command_arguments_repeat3] = aux_sym__command_arguments_repeat3,
  [aux_sym__host_command_argument_list_default_format_repeat1] = aux_sym__host_command_argument_list_default_format_repeat1,
  [aux_sym__host_command_argument_list_default_format_repeat2] = aux_sym__host_command_argument_list_default_format_repeat2,
  [aux_sym_format_expression_repeat1] = aux_sym_format_expression_repeat1,
  [aux_sym__function_identifier_repeat1] = aux_sym__function_identifier_repeat1,
  [aux_sym_argument_list_repeat1] = aux_sym_argument_list_repeat1,
  [aux_sym__hll_var_command_argument_list_default_format_repeat1] = aux_sym__hll_var_command_argument_list_default_format_repeat1,
  [aux_sym__hll_var_command_argument_list_default_format_repeat2] = aux_sym__hll_var_command_argument_list_default_format_repeat2,
  [aux_sym__hll_var_pattern_command_argument_list_format_repeat1] = aux_sym__hll_var_pattern_command_argument_list_format_repeat1,
  [aux_sym__hll_var_pattern_command_argument_list_format_repeat2] = aux_sym__hll_var_pattern_command_argument_list_format_repeat2,
  [aux_sym__var_command_argument_list_non_hll_format_repeat1] = aux_sym__var_command_argument_list_non_hll_format_repeat1,
  [aux_sym_hll_type_pattern_repeat1] = aux_sym_hll_type_pattern_repeat1,
  [aux_sym_hll_string_literal_repeat1] = aux_sym_hll_string_literal_repeat1,
  [aux_sym_hll_type_descriptor_repeat1] = aux_sym_hll_type_descriptor_repeat1,
  [aux_sym_hll_type_descriptor_repeat2] = aux_sym_hll_type_descriptor_repeat2,
  [aux_sym_hll_sized_type_specifier_repeat1] = aux_sym_hll_sized_type_specifier_repeat1,
  [aux_sym__hll_declaration_specifiers_repeat1] = aux_sym__hll_declaration_specifiers_repeat1,
  [aux_sym__hll_declaration_specifiers_repeat2] = aux_sym__hll_declaration_specifiers_repeat2,
  [aux_sym__hll_pointer_declarator_specifier_repeat1] = aux_sym__hll_pointer_declarator_specifier_repeat1,
  [aux_sym_hll_parameter_list_repeat1] = aux_sym_hll_parameter_list_repeat1,
  [aux_sym_hll_argument_list_repeat1] = aux_sym_hll_argument_list_repeat1,
  [aux_sym_string_repeat1] = aux_sym_string_repeat1,
  [anon_alias_sym_IF] = anon_alias_sym_IF,
  [anon_alias_sym_command] = anon_alias_sym_command,
  [anon_alias_sym_function] = anon_alias_sym_function,
  [alias_sym_hll_field_identifier] = alias_sym_hll_field_identifier,
  [alias_sym_hll_type_identifier] = alias_sym_hll_type_identifier,
  [anon_alias_sym_module] = anon_alias_sym_module,
  [anon_alias_sym_name] = anon_alias_sym_name,
  [anon_alias_sym_subcommand] = anon_alias_sym_subcommand,
  [anon_alias_sym_subfunction] = anon_alias_sym_subfunction,
  [anon_alias_sym_type] = anon_alias_sym_type,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_block_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_subroutine_block_token1] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_if_block_token1] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_elif_block_token1] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_while_block_token1] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_while_block_token2] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__repeat_block_pre_condition_format_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_macro_definition_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_parameter_declaration_token1] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_parameter_declaration_token2] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_parameter_declaration_token3] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_parameter_declaration_token4] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_subroutine_call_expression_token1] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_macro_definition_command_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_macro_definition_command_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_macro_definition_command_token3] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__host_command_with_default_format_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__host_command_argument_list_default_format_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__hll_non_default_format_var_command_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__hll_non_default_format_var_command_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__hll_non_default_format_var_command_token3] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__hll_non_default_format_var_command_token4] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__hll_non_default_format_var_command_token5] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__hll_var_definition_command_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__var_command_with_non_hll_format_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__var_command_with_non_hll_format_token2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__var_command_with_non_hll_format_token3] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__hll_var_definition_command_expression_token1] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__hll_var_definition_command_expression_token2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BSLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_sizeof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_L_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_U_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u8_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_hll_char_literal_token1] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_L_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_U_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u8_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_hll_string_literal_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_const] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_volatile] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_struct] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_union] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_enum] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_signed] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unsigned] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_long] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_short] = {
    .visible = true,
    .named = false,
  },
  [sym_hll_primitive_type] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__hll_expression_separator_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__hll_expression_separator_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_COLON_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [sym_access_class] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_integer_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_integer_token2] = {
    .visible = false,
    .named = false,
  },
  [sym_float] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_bitmask_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_bitmask_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_character_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_character_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_symbol_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_symbol_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_symbol_token3] = {
    .visible = false,
    .named = false,
  },
  [sym__composed_name] = {
    .visible = true,
    .named = true,
  },
  [sym_time] = {
    .visible = true,
    .named = true,
  },
  [sym_frequency] = {
    .visible = true,
    .named = true,
  },
  [sym_percentage] = {
    .visible = true,
    .named = true,
  },
  [sym_file_handle] = {
    .visible = true,
    .named = true,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym__blank_line] = {
    .visible = false,
    .named = true,
  },
  [sym__line_continuation] = {
    .visible = false,
    .named = true,
  },
  [sym__blank] = {
    .visible = false,
    .named = true,
  },
  [sym_label_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym__and_operator_pre_hook] = {
    .visible = false,
    .named = true,
  },
  [sym__decimal_number] = {
    .visible = false,
    .named = true,
  },
  [sym__decimal_number_pre_hook] = {
    .visible = false,
    .named = true,
  },
  [sym_hll_number_literal] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_number_literal_pre_hook] = {
    .visible = false,
    .named = true,
  },
  [sym_path] = {
    .visible = true,
    .named = true,
  },
  [sym__ampersand_char] = {
    .visible = false,
    .named = true,
  },
  [sym_script] = {
    .visible = true,
    .named = true,
  },
  [sym_block] = {
    .visible = true,
    .named = true,
  },
  [sym__statement] = {
    .visible = false,
    .named = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
  },
  [sym__parenthesized_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__path_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_labeled_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_subroutine_block] = {
    .visible = true,
    .named = true,
  },
  [sym_if_block] = {
    .visible = true,
    .named = true,
  },
  [sym_elif_block] = {
    .visible = true,
    .named = true,
  },
  [sym_else_block] = {
    .visible = true,
    .named = true,
  },
  [sym_while_block] = {
    .visible = true,
    .named = true,
  },
  [sym_repeat_block] = {
    .visible = true,
    .named = true,
  },
  [sym__repeat_block_pre_condition_format] = {
    .visible = false,
    .named = true,
  },
  [sym__repeat_block_post_condition_format] = {
    .visible = false,
    .named = true,
  },
  [sym_macro_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__unary_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__prefix_postfix_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__binary_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__and_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__label] = {
    .visible = false,
    .named = true,
  },
  [sym_macro_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_subroutine_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_macro] = {
    .visible = true,
    .named = true,
  },
  [sym_recursive_macro_expansion] = {
    .visible = true,
    .named = true,
  },
  [sym_macro_definition_command] = {
    .visible = true,
    .named = true,
  },
  [sym_command_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__host_command] = {
    .visible = false,
    .named = true,
  },
  [sym__host_command_with_default_format] = {
    .visible = true,
    .named = true,
  },
  [sym__other_command] = {
    .visible = false,
    .named = true,
  },
  [sym__command_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym__command_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym__host_command_argument_list_default_format] = {
    .visible = true,
    .named = true,
  },
  [sym__command_argument] = {
    .visible = false,
    .named = true,
  },
  [sym_option_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_format_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__function_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym__var_command] = {
    .visible = false,
    .named = true,
  },
  [sym__hll_var_command] = {
    .visible = false,
    .named = true,
  },
  [sym__hll_default_format_var_command] = {
    .visible = false,
    .named = true,
  },
  [sym__hll_non_default_format_var_command] = {
    .visible = false,
    .named = true,
  },
  [sym__non_hll_var_command] = {
    .visible = false,
    .named = true,
  },
  [sym__hll_var_definition_command] = {
    .visible = false,
    .named = true,
  },
  [sym__hll_var_command_with_default_format] = {
    .visible = true,
    .named = true,
  },
  [sym__var_command_with_non_hll_format] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_var_definition_command_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_var_if_command] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_var_command_argument_list_default_format] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_var_command_argument_list_non_default_format] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_var_pattern_command_argument_list_format] = {
    .visible = true,
    .named = true,
  },
  [sym__var_command_argument_list_non_hll_format] = {
    .visible = true,
    .named = true,
  },
  [sym__var_definition_command_argument_list_format] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_var_call_expression_function] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_var_call_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_hll_var_call_expression_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_type_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_trace32_hll_variable] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_hll_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_field_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_subscript_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_cast_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_sizeof_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_pointer_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_comma_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_update_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_range_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_assignment_left_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_hll_conditional_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_char_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_type_descriptor] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_type_qualifier] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_declaration_modifiers] = {
    .visible = false,
    .named = true,
  },
  [sym__hll_type_specifier] = {
    .visible = false,
    .named = true,
  },
  [sym_hll_struct_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_union_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_enum_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_sized_type_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_declaration_specifiers] = {
    .visible = false,
    .named = true,
  },
  [sym__hll_declarator] = {
    .visible = false,
    .named = true,
  },
  [sym_hll_pointer_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_function_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_array_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_parenthesized_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_pointer_declarator_specifier] = {
    .visible = false,
    .named = true,
  },
  [sym__hll_abstract_declarator] = {
    .visible = false,
    .named = true,
  },
  [sym_hll_abstract_pointer_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_abstract_function_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_abstract_array_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_abstract_parenthesized_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_parameter_list] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_parameter_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_hll_argument_list] = {
    .visible = true,
    .named = true,
  },
  [sym__hll_type_identifier] = {
    .visible = false,
    .named = true,
  },
  [sym__hll_field_identifier] = {
    .visible = false,
    .named = true,
  },
  [sym__literal] = {
    .visible = false,
    .named = true,
  },
  [sym__address_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_memory_space] = {
    .visible = true,
    .named = true,
  },
  [sym_address] = {
    .visible = true,
    .named = true,
  },
  [sym__parenthesized_address_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__address_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_integer] = {
    .visible = true,
    .named = true,
  },
  [sym_bitmask] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_character] = {
    .visible = true,
    .named = true,
  },
  [sym_symbol] = {
    .visible = true,
    .named = true,
  },
  [sym__special_command_operator] = {
    .visible = true,
    .named = true,
  },
  [sym__file_dialog] = {
    .visible = false,
    .named = true,
  },
  [aux_sym_script_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_block_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__path_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__path_expression_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_subroutine_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_if_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_if_block_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_if_block_repeat3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_macro_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__command_identifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__command_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__command_arguments_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__command_arguments_repeat3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__host_command_argument_list_default_format_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__host_command_argument_list_default_format_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_format_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__function_identifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_argument_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__hll_var_command_argument_list_default_format_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__hll_var_command_argument_list_default_format_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__hll_var_pattern_command_argument_list_format_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__hll_var_pattern_command_argument_list_format_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__var_command_argument_list_non_hll_format_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_hll_type_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_hll_string_literal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_hll_type_descriptor_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_hll_type_descriptor_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_hll_sized_type_specifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__hll_declaration_specifiers_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__hll_declaration_specifiers_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__hll_pointer_declarator_specifier_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_hll_parameter_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_hll_argument_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [anon_alias_sym_IF] = {
    .visible = true,
    .named = false,
  },
  [anon_alias_sym_command] = {
    .visible = true,
    .named = false,
  },
  [anon_alias_sym_function] = {
    .visible = true,
    .named = false,
  },
  [alias_sym_hll_field_identifier] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_hll_type_identifier] = {
    .visible = true,
    .named = true,
  },
  [anon_alias_sym_module] = {
    .visible = true,
    .named = false,
  },
  [anon_alias_sym_name] = {
    .visible = true,
    .named = false,
  },
  [anon_alias_sym_subcommand] = {
    .visible = true,
    .named = false,
  },
  [anon_alias_sym_subfunction] = {
    .visible = true,
    .named = false,
  },
  [anon_alias_sym_type] = {
    .visible = true,
    .named = false,
  },
};

enum ts_field_identifiers {
  field_access = 1,
  field_alternative = 2,
  field_argument = 3,
  field_arguments = 4,
  field_command = 5,
  field_condition = 6,
  field_consequence = 7,
  field_declarator = 8,
  field_field = 9,
  field_function = 10,
  field_index = 11,
  field_label = 12,
  field_left = 13,
  field_location = 14,
  field_machine = 15,
  field_macro = 16,
  field_name = 17,
  field_operator = 18,
  field_option = 19,
  field_parameters = 20,
  field_right = 21,
  field_segment = 22,
  field_size = 23,
  field_space = 24,
  field_subroutine = 25,
  field_type = 26,
  field_value = 27,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_access] = "access",
  [field_alternative] = "alternative",
  [field_argument] = "argument",
  [field_arguments] = "arguments",
  [field_command] = "command",
  [field_condition] = "condition",
  [field_consequence] = "consequence",
  [field_declarator] = "declarator",
  [field_field] = "field",
  [field_function] = "function",
  [field_index] = "index",
  [field_label] = "label",
  [field_left] = "left",
  [field_location] = "location",
  [field_machine] = "machine",
  [field_macro] = "macro",
  [field_name] = "name",
  [field_operator] = "operator",
  [field_option] = "option",
  [field_parameters] = "parameters",
  [field_right] = "right",
  [field_segment] = "segment",
  [field_size] = "size",
  [field_space] = "space",
  [field_subroutine] = "subroutine",
  [field_type] = "type",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [2] = {.index = 0, .length = 2},
  [3] = {.index = 2, .length = 1},
  [4] = {.index = 3, .length = 2},
  [5] = {.index = 5, .length = 3},
  [6] = {.index = 8, .length = 1},
  [8] = {.index = 9, .length = 1},
  [10] = {.index = 10, .length = 2},
  [11] = {.index = 12, .length = 3},
  [12] = {.index = 15, .length = 2},
  [14] = {.index = 17, .length = 2},
  [15] = {.index = 19, .length = 3},
  [16] = {.index = 22, .length = 1},
  [17] = {.index = 23, .length = 2},
  [18] = {.index = 25, .length = 2},
  [22] = {.index = 27, .length = 2},
  [24] = {.index = 29, .length = 1},
  [25] = {.index = 30, .length = 2},
  [26] = {.index = 32, .length = 2},
  [27] = {.index = 34, .length = 2},
  [28] = {.index = 36, .length = 2},
  [29] = {.index = 38, .length = 1},
  [30] = {.index = 39, .length = 1},
  [31] = {.index = 40, .length = 1},
  [32] = {.index = 41, .length = 1},
  [33] = {.index = 42, .length = 1},
  [34] = {.index = 43, .length = 2},
  [36] = {.index = 45, .length = 2},
  [37] = {.index = 47, .length = 2},
  [38] = {.index = 49, .length = 3},
  [39] = {.index = 52, .length = 1},
  [40] = {.index = 53, .length = 2},
  [41] = {.index = 55, .length = 1},
  [43] = {.index = 56, .length = 2},
  [45] = {.index = 58, .length = 2},
  [47] = {.index = 60, .length = 3},
  [48] = {.index = 63, .length = 2},
  [49] = {.index = 65, .length = 2},
  [50] = {.index = 67, .length = 4},
  [52] = {.index = 71, .length = 3},
  [53] = {.index = 74, .length = 3},
  [54] = {.index = 77, .length = 2},
  [55] = {.index = 79, .length = 1},
  [56] = {.index = 80, .length = 3},
  [57] = {.index = 83, .length = 2},
  [58] = {.index = 85, .length = 1},
  [59] = {.index = 86, .length = 2},
  [60] = {.index = 88, .length = 2},
  [61] = {.index = 90, .length = 2},
  [62] = {.index = 92, .length = 2},
  [63] = {.index = 94, .length = 2},
  [64] = {.index = 96, .length = 2},
  [65] = {.index = 98, .length = 2},
  [66] = {.index = 100, .length = 2},
  [67] = {.index = 102, .length = 1},
  [68] = {.index = 103, .length = 2},
  [69] = {.index = 105, .length = 2},
  [70] = {.index = 107, .length = 2},
  [71] = {.index = 109, .length = 3},
  [72] = {.index = 112, .length = 3},
  [73] = {.index = 115, .length = 2},
  [74] = {.index = 117, .length = 2},
  [75] = {.index = 119, .length = 1},
  [76] = {.index = 120, .length = 1},
  [77] = {.index = 121, .length = 2},
  [78] = {.index = 123, .length = 2},
  [79] = {.index = 125, .length = 1},
  [80] = {.index = 126, .length = 1},
  [81] = {.index = 127, .length = 2},
  [82] = {.index = 129, .length = 2},
  [83] = {.index = 131, .length = 2},
  [84] = {.index = 133, .length = 2},
  [85] = {.index = 135, .length = 1},
  [86] = {.index = 136, .length = 3},
  [87] = {.index = 139, .length = 3},
  [88] = {.index = 142, .length = 3},
  [89] = {.index = 145, .length = 2},
  [90] = {.index = 147, .length = 1},
  [91] = {.index = 148, .length = 3},
  [92] = {.index = 151, .length = 2},
  [93] = {.index = 153, .length = 1},
  [94] = {.index = 154, .length = 2},
  [95] = {.index = 156, .length = 2},
  [96] = {.index = 158, .length = 2},
  [97] = {.index = 160, .length = 2},
  [98] = {.index = 162, .length = 2},
  [99] = {.index = 164, .length = 3},
  [100] = {.index = 167, .length = 2},
  [101] = {.index = 169, .length = 3},
  [102] = {.index = 172, .length = 1},
  [103] = {.index = 173, .length = 2},
  [104] = {.index = 175, .length = 2},
  [105] = {.index = 177, .length = 3},
  [106] = {.index = 180, .length = 2},
  [107] = {.index = 182, .length = 3},
  [108] = {.index = 185, .length = 2},
  [109] = {.index = 187, .length = 3},
  [110] = {.index = 190, .length = 2},
  [111] = {.index = 192, .length = 3},
  [112] = {.index = 195, .length = 2},
  [113] = {.index = 197, .length = 3},
  [114] = {.index = 200, .length = 1},
  [115] = {.index = 201, .length = 3},
  [116] = {.index = 204, .length = 4},
  [117] = {.index = 208, .length = 3},
  [118] = {.index = 211, .length = 3},
  [119] = {.index = 214, .length = 3},
  [120] = {.index = 217, .length = 2},
  [121] = {.index = 219, .length = 3},
  [122] = {.index = 222, .length = 4},
  [123] = {.index = 226, .length = 3},
  [124] = {.index = 229, .length = 3},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_command, 0, .inherited = true},
    {field_condition, 0, .inherited = true},
  [2] =
    {field_command, 0},
  [3] =
    {field_arguments, 0, .inherited = true},
    {field_command, 0, .inherited = true},
  [5] =
    {field_arguments, 0, .inherited = true},
    {field_command, 0, .inherited = true},
    {field_value, 0, .inherited = true},
  [8] =
    {field_label, 0},
  [9] =
    {field_label, 0, .inherited = true},
  [10] =
    {field_argument, 0, .inherited = true},
    {field_operator, 0, .inherited = true},
  [12] =
    {field_left, 0, .inherited = true},
    {field_operator, 0, .inherited = true},
    {field_right, 0, .inherited = true},
  [15] =
    {field_arguments, 0, .inherited = true},
    {field_function, 0, .inherited = true},
  [17] =
    {field_arguments, 1, .inherited = true},
    {field_command, 1, .inherited = true},
  [19] =
    {field_arguments, 1, .inherited = true},
    {field_command, 1, .inherited = true},
    {field_value, 1, .inherited = true},
  [22] =
    {field_macro, 1},
  [23] =
    {field_command, 0},
    {field_macro, 1, .inherited = true},
  [25] =
    {field_macro, 0, .inherited = true},
    {field_macro, 1, .inherited = true},
  [27] =
    {field_arguments, 2},
    {field_command, 0},
  [29] =
    {field_type, 0},
  [30] =
    {field_argument, 1},
    {field_operator, 0},
  [32] =
    {field_access, 0},
    {field_location, 1},
  [34] =
    {field_argument, 0},
    {field_operator, 1},
  [36] =
    {field_arguments, 1},
    {field_function, 0},
  [38] =
    {field_segment, 0},
  [39] =
    {field_machine, 0},
  [40] =
    {field_space, 0},
  [41] =
    {field_location, 1},
  [42] =
    {field_value, 1},
  [43] =
    {field_command, 0},
    {field_condition, 2},
  [45] =
    {field_command, 1},
    {field_macro, 2, .inherited = true},
  [47] =
    {field_command, 0},
    {field_subroutine, 2},
  [49] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [52] =
    {field_option, 1},
  [53] =
    {field_declarator, 1},
    {field_type, 0, .inherited = true},
  [55] =
    {field_type, 1},
  [56] =
    {field_access, 0},
    {field_location, 2},
  [58] =
    {field_declarator, 1},
    {field_type, 0},
  [60] =
    {field_argument, 0},
    {field_field, 2},
    {field_operator, 1},
  [63] =
    {field_command, 1},
    {field_subroutine, 3},
  [65] =
    {field_command, 0},
    {field_macro, 3, .inherited = true},
  [67] =
    {field_arguments, 4},
    {field_command, 0},
    {field_command, 1},
    {field_command, 2},
  [71] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 3},
  [74] =
    {field_left, 0},
    {field_operator, 2},
    {field_right, 3},
  [77] =
    {field_value, 1},
    {field_value, 2, .inherited = true},
  [79] =
    {field_name, 2},
  [80] =
    {field_arguments, 2},
    {field_command, 0},
    {field_value, 4},
  [83] =
    {field_declarator, 0},
    {field_parameters, 1},
  [85] =
    {field_declarator, 1},
  [86] =
    {field_declarator, 2},
    {field_type, 0, .inherited = true},
  [88] =
    {field_machine, 0},
    {field_segment, 2},
  [90] =
    {field_machine, 0},
    {field_space, 2},
  [92] =
    {field_segment, 2},
    {field_space, 0},
  [94] =
    {field_left, 0},
    {field_right, 2},
  [96] =
    {field_type, 1},
    {field_value, 3},
  [98] =
    {field_declarator, 2},
    {field_type, 0},
  [100] =
    {field_declarator, 2},
    {field_type, 1},
  [102] =
    {field_type, 2},
  [103] =
    {field_alternative, 3},
    {field_condition, 0},
  [105] =
    {field_argument, 0},
    {field_index, 2},
  [107] =
    {field_command, 1},
    {field_macro, 4, .inherited = true},
  [109] =
    {field_arguments, 4},
    {field_command, 0},
    {field_subroutine, 2},
  [112] =
    {field_left, 0},
    {field_operator, 2},
    {field_right, 4},
  [115] =
    {field_option, 1},
    {field_value, 3},
  [117] =
    {field_value, 0, .inherited = true},
    {field_value, 1, .inherited = true},
  [119] =
    {field_declarator, 2},
  [120] =
    {field_type, 0, .inherited = true},
  [121] =
    {field_declarator, 0},
    {field_parameters, 2},
  [123] =
    {field_type, 1},
    {field_value, 4},
  [125] =
    {field_size, 1},
  [126] =
    {field_declarator, 0},
  [127] =
    {field_declarator, 3},
    {field_type, 0},
  [129] =
    {field_type, 2},
    {field_value, 4},
  [131] =
    {field_declarator, 3},
    {field_type, 1},
  [133] =
    {field_declarator, 3},
    {field_type, 2},
  [135] =
    {field_type, 3},
  [136] =
    {field_alternative, 4},
    {field_condition, 0},
    {field_consequence, 2},
  [139] =
    {field_command, 0},
    {field_command, 3},
    {field_condition, 5},
  [142] =
    {field_arguments, 5},
    {field_command, 1},
    {field_subroutine, 3},
  [145] =
    {field_declarator, 3},
    {field_size, 1},
  [147] =
    {field_declarator, 3},
  [148] =
    {field_machine, 0},
    {field_segment, 4},
    {field_space, 2},
  [151] =
    {field_size, 1},
    {field_size, 2},
  [153] =
    {field_size, 2},
  [154] =
    {field_declarator, 0},
    {field_size, 2},
  [156] =
    {field_declarator, 0},
    {field_declarator, 1},
  [158] =
    {field_type, 2},
    {field_value, 5},
  [160] =
    {field_declarator, 4},
    {field_type, 1},
  [162] =
    {field_declarator, 4},
    {field_type, 2},
  [164] =
    {field_command, 0},
    {field_command, 4},
    {field_condition, 6},
  [167] =
    {field_declarator, 4},
    {field_size, 1},
  [169] =
    {field_declarator, 4},
    {field_size, 1},
    {field_size, 2},
  [172] =
    {field_declarator, 4},
  [173] =
    {field_declarator, 4},
    {field_size, 2},
  [175] =
    {field_size, 2},
    {field_size, 3},
  [177] =
    {field_declarator, 0},
    {field_size, 2},
    {field_size, 3},
  [180] =
    {field_declarator, 0},
    {field_size, 3},
  [182] =
    {field_declarator, 0},
    {field_declarator, 1},
    {field_size, 3},
  [185] =
    {field_declarator, 5},
    {field_type, 2},
  [187] =
    {field_command, 0},
    {field_command, 5},
    {field_condition, 7},
  [190] =
    {field_declarator, 5},
    {field_size, 1},
  [192] =
    {field_declarator, 5},
    {field_size, 1},
    {field_size, 2},
  [195] =
    {field_declarator, 5},
    {field_size, 2},
  [197] =
    {field_declarator, 5},
    {field_size, 2},
    {field_size, 3},
  [200] =
    {field_declarator, 5},
  [201] =
    {field_declarator, 0},
    {field_size, 3},
    {field_size, 4},
  [204] =
    {field_declarator, 0},
    {field_declarator, 1},
    {field_size, 3},
    {field_size, 4},
  [208] =
    {field_declarator, 0},
    {field_declarator, 1},
    {field_size, 4},
  [211] =
    {field_command, 0},
    {field_command, 6},
    {field_condition, 8},
  [214] =
    {field_declarator, 6},
    {field_size, 1},
    {field_size, 2},
  [217] =
    {field_declarator, 6},
    {field_size, 2},
  [219] =
    {field_declarator, 6},
    {field_size, 2},
    {field_size, 3},
  [222] =
    {field_declarator, 0},
    {field_declarator, 1},
    {field_size, 4},
    {field_size, 5},
  [226] =
    {field_declarator, 7},
    {field_size, 2},
    {field_size, 3},
  [229] =
    {field_command, 0},
    {field_command, 2},
    {field_condition, 4},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = anon_alias_sym_command,
  },
  [7] = {
    [1] = anon_alias_sym_name,
  },
  [9] = {
    [0] = anon_alias_sym_function,
  },
  [13] = {
    [0] = anon_alias_sym_module,
  },
  [19] = {
    [2] = anon_alias_sym_subcommand,
  },
  [20] = {
    [2] = anon_alias_sym_IF,
  },
  [21] = {
    [1] = anon_alias_sym_subcommand,
  },
  [23] = {
    [0] = alias_sym_hll_type_identifier,
  },
  [35] = {
    [2] = anon_alias_sym_name,
  },
  [42] = {
    [2] = anon_alias_sym_subfunction,
  },
  [44] = {
    [1] = anon_alias_sym_subfunction,
  },
  [46] = {
    [0] = alias_sym_hll_field_identifier,
  },
  [50] = {
    [1] = sym_identifier,
  },
  [51] = {
    [0] = anon_alias_sym_type,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym_trace32_hll_variable, 2,
    sym_trace32_hll_variable,
    anon_alias_sym_module,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 2,
  [5] = 3,
  [6] = 2,
  [7] = 3,
  [8] = 2,
  [9] = 9,
  [10] = 10,
  [11] = 11,
  [12] = 12,
  [13] = 13,
  [14] = 13,
  [15] = 13,
  [16] = 13,
  [17] = 13,
  [18] = 13,
  [19] = 9,
  [20] = 12,
  [21] = 11,
  [22] = 10,
  [23] = 10,
  [24] = 11,
  [25] = 9,
  [26] = 12,
  [27] = 10,
  [28] = 11,
  [29] = 9,
  [30] = 12,
  [31] = 31,
  [32] = 32,
  [33] = 33,
  [34] = 33,
  [35] = 32,
  [36] = 36,
  [37] = 11,
  [38] = 36,
  [39] = 39,
  [40] = 40,
  [41] = 41,
  [42] = 42,
  [43] = 43,
  [44] = 39,
  [45] = 43,
  [46] = 9,
  [47] = 43,
  [48] = 39,
  [49] = 43,
  [50] = 39,
  [51] = 36,
  [52] = 42,
  [53] = 43,
  [54] = 39,
  [55] = 43,
  [56] = 39,
  [57] = 43,
  [58] = 39,
  [59] = 43,
  [60] = 39,
  [61] = 36,
  [62] = 42,
  [63] = 43,
  [64] = 39,
  [65] = 43,
  [66] = 39,
  [67] = 36,
  [68] = 42,
  [69] = 36,
  [70] = 42,
  [71] = 36,
  [72] = 42,
  [73] = 42,
  [74] = 36,
  [75] = 42,
  [76] = 36,
  [77] = 42,
  [78] = 36,
  [79] = 42,
  [80] = 10,
  [81] = 12,
  [82] = 9,
  [83] = 11,
  [84] = 12,
  [85] = 10,
  [86] = 12,
  [87] = 10,
  [88] = 11,
  [89] = 9,
  [90] = 9,
  [91] = 10,
  [92] = 11,
  [93] = 12,
  [94] = 94,
  [95] = 95,
  [96] = 96,
  [97] = 97,
  [98] = 98,
  [99] = 99,
  [100] = 100,
  [101] = 101,
  [102] = 9,
  [103] = 10,
  [104] = 11,
  [105] = 12,
  [106] = 96,
  [107] = 97,
  [108] = 94,
  [109] = 109,
  [110] = 110,
  [111] = 111,
  [112] = 98,
  [113] = 110,
  [114] = 95,
  [115] = 115,
  [116] = 96,
  [117] = 97,
  [118] = 98,
  [119] = 119,
  [120] = 120,
  [121] = 121,
  [122] = 122,
  [123] = 99,
  [124] = 101,
  [125] = 125,
  [126] = 126,
  [127] = 127,
  [128] = 128,
  [129] = 129,
  [130] = 94,
  [131] = 109,
  [132] = 115,
  [133] = 122,
  [134] = 128,
  [135] = 126,
  [136] = 95,
  [137] = 127,
  [138] = 99,
  [139] = 96,
  [140] = 97,
  [141] = 110,
  [142] = 98,
  [143] = 120,
  [144] = 129,
  [145] = 129,
  [146] = 99,
  [147] = 120,
  [148] = 94,
  [149] = 129,
  [150] = 120,
  [151] = 94,
  [152] = 129,
  [153] = 120,
  [154] = 120,
  [155] = 99,
  [156] = 95,
  [157] = 99,
  [158] = 94,
  [159] = 129,
  [160] = 100,
  [161] = 161,
  [162] = 9,
  [163] = 10,
  [164] = 11,
  [165] = 12,
  [166] = 161,
  [167] = 161,
  [168] = 168,
  [169] = 169,
  [170] = 170,
  [171] = 171,
  [172] = 172,
  [173] = 173,
  [174] = 174,
  [175] = 175,
  [176] = 176,
  [177] = 177,
  [178] = 169,
  [179] = 179,
  [180] = 180,
  [181] = 181,
  [182] = 182,
  [183] = 183,
  [184] = 170,
  [185] = 185,
  [186] = 179,
  [187] = 180,
  [188] = 181,
  [189] = 182,
  [190] = 183,
  [191] = 185,
  [192] = 171,
  [193] = 179,
  [194] = 180,
  [195] = 168,
  [196] = 196,
  [197] = 181,
  [198] = 182,
  [199] = 183,
  [200] = 174,
  [201] = 201,
  [202] = 202,
  [203] = 203,
  [204] = 204,
  [205] = 203,
  [206] = 203,
  [207] = 207,
  [208] = 208,
  [209] = 204,
  [210] = 172,
  [211] = 207,
  [212] = 212,
  [213] = 185,
  [214] = 208,
  [215] = 174,
  [216] = 175,
  [217] = 204,
  [218] = 196,
  [219] = 185,
  [220] = 208,
  [221] = 175,
  [222] = 176,
  [223] = 177,
  [224] = 169,
  [225] = 168,
  [226] = 196,
  [227] = 170,
  [228] = 171,
  [229] = 168,
  [230] = 196,
  [231] = 179,
  [232] = 180,
  [233] = 168,
  [234] = 196,
  [235] = 181,
  [236] = 182,
  [237] = 183,
  [238] = 172,
  [239] = 173,
  [240] = 176,
  [241] = 173,
  [242] = 177,
  [243] = 169,
  [244] = 196,
  [245] = 170,
  [246] = 171,
  [247] = 168,
  [248] = 212,
  [249] = 172,
  [250] = 173,
  [251] = 208,
  [252] = 175,
  [253] = 179,
  [254] = 180,
  [255] = 181,
  [256] = 182,
  [257] = 183,
  [258] = 176,
  [259] = 259,
  [260] = 177,
  [261] = 207,
  [262] = 204,
  [263] = 203,
  [264] = 212,
  [265] = 174,
  [266] = 266,
  [267] = 267,
  [268] = 268,
  [269] = 269,
  [270] = 270,
  [271] = 271,
  [272] = 272,
  [273] = 273,
  [274] = 267,
  [275] = 266,
  [276] = 276,
  [277] = 277,
  [278] = 278,
  [279] = 279,
  [280] = 272,
  [281] = 267,
  [282] = 269,
  [283] = 270,
  [284] = 271,
  [285] = 285,
  [286] = 273,
  [287] = 269,
  [288] = 266,
  [289] = 276,
  [290] = 277,
  [291] = 278,
  [292] = 272,
  [293] = 267,
  [294] = 269,
  [295] = 270,
  [296] = 271,
  [297] = 276,
  [298] = 277,
  [299] = 276,
  [300] = 271,
  [301] = 276,
  [302] = 302,
  [303] = 303,
  [304] = 279,
  [305] = 270,
  [306] = 306,
  [307] = 306,
  [308] = 285,
  [309] = 278,
  [310] = 310,
  [311] = 302,
  [312] = 272,
  [313] = 267,
  [314] = 273,
  [315] = 266,
  [316] = 277,
  [317] = 278,
  [318] = 271,
  [319] = 276,
  [320] = 273,
  [321] = 269,
  [322] = 270,
  [323] = 273,
  [324] = 266,
  [325] = 277,
  [326] = 278,
  [327] = 272,
  [328] = 267,
  [329] = 269,
  [330] = 270,
  [331] = 273,
  [332] = 266,
  [333] = 277,
  [334] = 278,
  [335] = 272,
  [336] = 271,
  [337] = 337,
  [338] = 338,
  [339] = 338,
  [340] = 337,
  [341] = 338,
  [342] = 337,
  [343] = 338,
  [344] = 337,
  [345] = 337,
  [346] = 338,
  [347] = 338,
  [348] = 337,
  [349] = 349,
  [350] = 350,
  [351] = 351,
  [352] = 349,
  [353] = 353,
  [354] = 351,
  [355] = 350,
  [356] = 353,
  [357] = 353,
  [358] = 353,
  [359] = 353,
  [360] = 353,
  [361] = 361,
  [362] = 361,
  [363] = 363,
  [364] = 363,
  [365] = 361,
  [366] = 361,
  [367] = 361,
  [368] = 363,
  [369] = 363,
  [370] = 363,
  [371] = 363,
  [372] = 361,
  [373] = 373,
  [374] = 373,
  [375] = 373,
  [376] = 373,
  [377] = 373,
  [378] = 373,
  [379] = 379,
  [380] = 380,
  [381] = 381,
  [382] = 382,
  [383] = 380,
  [384] = 384,
  [385] = 379,
  [386] = 386,
  [387] = 387,
  [388] = 388,
  [389] = 389,
  [390] = 390,
  [391] = 391,
  [392] = 392,
  [393] = 393,
  [394] = 394,
  [395] = 395,
  [396] = 396,
  [397] = 397,
  [398] = 398,
  [399] = 394,
  [400] = 400,
  [401] = 401,
  [402] = 402,
  [403] = 403,
  [404] = 403,
  [405] = 405,
  [406] = 406,
  [407] = 407,
  [408] = 406,
  [409] = 409,
  [410] = 410,
  [411] = 411,
  [412] = 412,
  [413] = 413,
  [414] = 414,
  [415] = 415,
  [416] = 416,
  [417] = 417,
  [418] = 418,
  [419] = 419,
  [420] = 420,
  [421] = 421,
  [422] = 422,
  [423] = 423,
  [424] = 424,
  [425] = 425,
  [426] = 426,
  [427] = 427,
  [428] = 428,
  [429] = 429,
  [430] = 424,
  [431] = 411,
  [432] = 432,
  [433] = 433,
  [434] = 434,
  [435] = 435,
  [436] = 436,
  [437] = 437,
  [438] = 438,
  [439] = 439,
  [440] = 440,
  [441] = 441,
  [442] = 442,
  [443] = 443,
  [444] = 436,
  [445] = 445,
  [446] = 446,
  [447] = 447,
  [448] = 448,
  [449] = 449,
  [450] = 450,
  [451] = 451,
  [452] = 452,
  [453] = 453,
  [454] = 454,
  [455] = 455,
  [456] = 456,
  [457] = 457,
  [458] = 458,
  [459] = 436,
  [460] = 442,
  [461] = 461,
  [462] = 462,
  [463] = 463,
  [464] = 464,
  [465] = 465,
  [466] = 466,
  [467] = 467,
  [468] = 468,
  [469] = 469,
  [470] = 470,
  [471] = 468,
  [472] = 468,
  [473] = 473,
  [474] = 474,
  [475] = 436,
  [476] = 476,
  [477] = 477,
  [478] = 445,
  [479] = 447,
  [480] = 446,
  [481] = 481,
  [482] = 482,
  [483] = 483,
  [484] = 468,
  [485] = 485,
  [486] = 486,
  [487] = 487,
  [488] = 488,
  [489] = 489,
  [490] = 490,
  [491] = 491,
  [492] = 492,
  [493] = 488,
  [494] = 491,
  [495] = 490,
  [496] = 489,
  [497] = 488,
  [498] = 489,
  [499] = 499,
  [500] = 491,
  [501] = 501,
  [502] = 488,
  [503] = 503,
  [504] = 501,
  [505] = 505,
  [506] = 488,
  [507] = 490,
  [508] = 508,
  [509] = 491,
  [510] = 490,
  [511] = 511,
  [512] = 491,
  [513] = 489,
  [514] = 490,
  [515] = 515,
  [516] = 516,
  [517] = 517,
  [518] = 518,
  [519] = 501,
  [520] = 505,
  [521] = 521,
  [522] = 501,
  [523] = 505,
  [524] = 524,
  [525] = 488,
  [526] = 505,
  [527] = 527,
  [528] = 505,
  [529] = 489,
  [530] = 491,
  [531] = 489,
  [532] = 491,
  [533] = 490,
  [534] = 489,
  [535] = 490,
  [536] = 488,
  [537] = 489,
  [538] = 488,
  [539] = 491,
  [540] = 491,
  [541] = 489,
  [542] = 489,
  [543] = 490,
  [544] = 490,
  [545] = 490,
  [546] = 488,
  [547] = 505,
  [548] = 501,
  [549] = 491,
  [550] = 501,
  [551] = 488,
  [552] = 552,
  [553] = 553,
  [554] = 554,
  [555] = 555,
  [556] = 556,
  [557] = 557,
  [558] = 558,
  [559] = 559,
  [560] = 560,
  [561] = 561,
  [562] = 562,
  [563] = 552,
  [564] = 564,
  [565] = 565,
  [566] = 552,
  [567] = 552,
  [568] = 568,
  [569] = 569,
  [570] = 552,
  [571] = 552,
  [572] = 572,
  [573] = 573,
  [574] = 552,
  [575] = 575,
  [576] = 552,
  [577] = 552,
  [578] = 578,
  [579] = 558,
  [580] = 552,
  [581] = 581,
  [582] = 582,
  [583] = 583,
  [584] = 558,
  [585] = 558,
  [586] = 586,
  [587] = 587,
  [588] = 588,
  [589] = 587,
  [590] = 590,
  [591] = 591,
  [592] = 592,
  [593] = 593,
  [594] = 594,
  [595] = 595,
  [596] = 596,
  [597] = 597,
  [598] = 598,
  [599] = 599,
  [600] = 600,
  [601] = 601,
  [602] = 602,
  [603] = 603,
  [604] = 588,
  [605] = 605,
  [606] = 590,
  [607] = 591,
  [608] = 592,
  [609] = 609,
  [610] = 593,
  [611] = 588,
  [612] = 612,
  [613] = 613,
  [614] = 614,
  [615] = 594,
  [616] = 595,
  [617] = 596,
  [618] = 597,
  [619] = 598,
  [620] = 599,
  [621] = 600,
  [622] = 601,
  [623] = 602,
  [624] = 603,
  [625] = 588,
  [626] = 626,
  [627] = 587,
  [628] = 628,
  [629] = 590,
  [630] = 591,
  [631] = 592,
  [632] = 632,
  [633] = 633,
  [634] = 593,
  [635] = 635,
  [636] = 594,
  [637] = 595,
  [638] = 596,
  [639] = 597,
  [640] = 598,
  [641] = 599,
  [642] = 600,
  [643] = 601,
  [644] = 602,
  [645] = 603,
  [646] = 588,
  [647] = 587,
  [648] = 590,
  [649] = 591,
  [650] = 592,
  [651] = 593,
  [652] = 594,
  [653] = 595,
  [654] = 596,
  [655] = 597,
  [656] = 598,
  [657] = 599,
  [658] = 600,
  [659] = 601,
  [660] = 602,
  [661] = 603,
  [662] = 588,
  [663] = 663,
  [664] = 664,
  [665] = 665,
  [666] = 666,
  [667] = 667,
  [668] = 668,
  [669] = 669,
  [670] = 670,
  [671] = 671,
  [672] = 587,
  [673] = 673,
  [674] = 590,
  [675] = 591,
  [676] = 676,
  [677] = 592,
  [678] = 678,
  [679] = 679,
  [680] = 593,
  [681] = 681,
  [682] = 682,
  [683] = 594,
  [684] = 595,
  [685] = 596,
  [686] = 597,
  [687] = 598,
  [688] = 599,
  [689] = 600,
  [690] = 601,
  [691] = 602,
  [692] = 603,
  [693] = 587,
  [694] = 588,
  [695] = 587,
  [696] = 590,
  [697] = 591,
  [698] = 592,
  [699] = 593,
  [700] = 594,
  [701] = 595,
  [702] = 596,
  [703] = 597,
  [704] = 598,
  [705] = 599,
  [706] = 600,
  [707] = 601,
  [708] = 602,
  [709] = 603,
  [710] = 588,
  [711] = 587,
  [712] = 590,
  [713] = 591,
  [714] = 592,
  [715] = 593,
  [716] = 499,
  [717] = 594,
  [718] = 595,
  [719] = 596,
  [720] = 597,
  [721] = 598,
  [722] = 599,
  [723] = 600,
  [724] = 601,
  [725] = 602,
  [726] = 603,
  [727] = 517,
  [728] = 588,
  [729] = 587,
  [730] = 590,
  [731] = 591,
  [732] = 592,
  [733] = 593,
  [734] = 734,
  [735] = 605,
  [736] = 609,
  [737] = 734,
  [738] = 738,
  [739] = 738,
  [740] = 594,
  [741] = 595,
  [742] = 590,
  [743] = 596,
  [744] = 597,
  [745] = 598,
  [746] = 599,
  [747] = 591,
  [748] = 600,
  [749] = 594,
  [750] = 595,
  [751] = 596,
  [752] = 597,
  [753] = 598,
  [754] = 599,
  [755] = 600,
  [756] = 601,
  [757] = 602,
  [758] = 603,
  [759] = 601,
  [760] = 605,
  [761] = 609,
  [762] = 734,
  [763] = 738,
  [764] = 592,
  [765] = 602,
  [766] = 603,
  [767] = 605,
  [768] = 609,
  [769] = 734,
  [770] = 738,
  [771] = 588,
  [772] = 605,
  [773] = 609,
  [774] = 738,
  [775] = 605,
  [776] = 609,
  [777] = 738,
  [778] = 605,
  [779] = 609,
  [780] = 738,
  [781] = 605,
  [782] = 609,
  [783] = 738,
  [784] = 587,
  [785] = 605,
  [786] = 609,
  [787] = 738,
  [788] = 605,
  [789] = 609,
  [790] = 738,
  [791] = 791,
  [792] = 590,
  [793] = 591,
  [794] = 592,
  [795] = 593,
  [796] = 791,
  [797] = 593,
  [798] = 791,
  [799] = 799,
  [800] = 594,
  [801] = 595,
  [802] = 596,
  [803] = 597,
  [804] = 598,
  [805] = 599,
  [806] = 600,
  [807] = 601,
  [808] = 602,
  [809] = 603,
  [810] = 518,
  [811] = 811,
  [812] = 812,
  [813] = 813,
  [814] = 814,
  [815] = 553,
  [816] = 816,
  [817] = 817,
  [818] = 818,
  [819] = 819,
  [820] = 820,
  [821] = 821,
  [822] = 822,
  [823] = 823,
  [824] = 583,
  [825] = 825,
  [826] = 826,
  [827] = 827,
  [828] = 828,
  [829] = 829,
  [830] = 559,
  [831] = 831,
  [832] = 832,
  [833] = 678,
  [834] = 517,
  [835] = 681,
  [836] = 682,
  [837] = 669,
  [838] = 626,
  [839] = 666,
  [840] = 628,
  [841] = 841,
  [842] = 842,
  [843] = 518,
  [844] = 799,
  [845] = 632,
  [846] = 633,
  [847] = 635,
  [848] = 612,
  [849] = 613,
  [850] = 499,
  [851] = 614,
  [852] = 670,
  [853] = 671,
  [854] = 667,
  [855] = 673,
  [856] = 818,
  [857] = 819,
  [858] = 676,
  [859] = 663,
  [860] = 664,
  [861] = 665,
  [862] = 668,
  [863] = 679,
  [864] = 799,
  [865] = 583,
  [866] = 665,
  [867] = 612,
  [868] = 613,
  [869] = 666,
  [870] = 614,
  [871] = 667,
  [872] = 831,
  [873] = 668,
  [874] = 559,
  [875] = 553,
  [876] = 669,
  [877] = 832,
  [878] = 811,
  [879] = 626,
  [880] = 820,
  [881] = 818,
  [882] = 819,
  [883] = 628,
  [884] = 814,
  [885] = 670,
  [886] = 671,
  [887] = 829,
  [888] = 664,
  [889] = 673,
  [890] = 818,
  [891] = 819,
  [892] = 632,
  [893] = 633,
  [894] = 821,
  [895] = 676,
  [896] = 822,
  [897] = 635,
  [898] = 823,
  [899] = 678,
  [900] = 825,
  [901] = 826,
  [902] = 827,
  [903] = 813,
  [904] = 816,
  [905] = 817,
  [906] = 679,
  [907] = 681,
  [908] = 682,
  [909] = 518,
  [910] = 499,
  [911] = 517,
  [912] = 828,
  [913] = 818,
  [914] = 819,
  [915] = 663,
  [916] = 812,
  [917] = 829,
  [918] = 812,
  [919] = 668,
  [920] = 517,
  [921] = 669,
  [922] = 813,
  [923] = 681,
  [924] = 682,
  [925] = 816,
  [926] = 817,
  [927] = 583,
  [928] = 818,
  [929] = 559,
  [930] = 819,
  [931] = 553,
  [932] = 811,
  [933] = 673,
  [934] = 799,
  [935] = 664,
  [936] = 823,
  [937] = 663,
  [938] = 818,
  [939] = 819,
  [940] = 670,
  [941] = 499,
  [942] = 818,
  [943] = 820,
  [944] = 819,
  [945] = 665,
  [946] = 626,
  [947] = 671,
  [948] = 814,
  [949] = 825,
  [950] = 821,
  [951] = 822,
  [952] = 676,
  [953] = 628,
  [954] = 612,
  [955] = 613,
  [956] = 831,
  [957] = 614,
  [958] = 632,
  [959] = 832,
  [960] = 633,
  [961] = 635,
  [962] = 826,
  [963] = 678,
  [964] = 679,
  [965] = 666,
  [966] = 667,
  [967] = 828,
  [968] = 827,
  [969] = 518,
  [970] = 559,
  [971] = 583,
  [972] = 823,
  [973] = 831,
  [974] = 818,
  [975] = 827,
  [976] = 826,
  [977] = 818,
  [978] = 813,
  [979] = 814,
  [980] = 820,
  [981] = 821,
  [982] = 816,
  [983] = 822,
  [984] = 832,
  [985] = 812,
  [986] = 811,
  [987] = 553,
  [988] = 817,
  [989] = 828,
  [990] = 819,
  [991] = 819,
  [992] = 825,
  [993] = 829,
  [994] = 994,
  [995] = 995,
  [996] = 996,
  [997] = 997,
  [998] = 998,
  [999] = 999,
  [1000] = 842,
  [1001] = 1001,
  [1002] = 842,
  [1003] = 1003,
  [1004] = 1004,
  [1005] = 1005,
  [1006] = 1006,
  [1007] = 1007,
  [1008] = 1008,
  [1009] = 1009,
  [1010] = 1010,
  [1011] = 1011,
  [1012] = 1012,
  [1013] = 842,
  [1014] = 499,
  [1015] = 613,
  [1016] = 517,
  [1017] = 518,
  [1018] = 612,
  [1019] = 628,
  [1020] = 626,
  [1021] = 632,
  [1022] = 633,
  [1023] = 635,
  [1024] = 1024,
  [1025] = 1025,
  [1026] = 821,
  [1027] = 1027,
  [1028] = 822,
  [1029] = 673,
  [1030] = 406,
  [1031] = 1025,
  [1032] = 1032,
  [1033] = 1033,
  [1034] = 1034,
  [1035] = 679,
  [1036] = 1036,
  [1037] = 1034,
  [1038] = 1025,
  [1039] = 828,
  [1040] = 829,
  [1041] = 1034,
  [1042] = 1042,
  [1043] = 583,
  [1044] = 407,
  [1045] = 676,
  [1046] = 678,
  [1047] = 1034,
  [1048] = 1025,
  [1049] = 820,
  [1050] = 1050,
  [1051] = 1051,
  [1052] = 1052,
  [1053] = 670,
  [1054] = 406,
  [1055] = 1055,
  [1056] = 1056,
  [1057] = 681,
  [1058] = 1058,
  [1059] = 405,
  [1060] = 406,
  [1061] = 614,
  [1062] = 682,
  [1063] = 1063,
  [1064] = 671,
  [1065] = 669,
  [1066] = 663,
  [1067] = 664,
  [1068] = 665,
  [1069] = 666,
  [1070] = 667,
  [1071] = 668,
  [1072] = 559,
  [1073] = 417,
  [1074] = 1074,
  [1075] = 614,
  [1076] = 1074,
  [1077] = 1077,
  [1078] = 1078,
  [1079] = 663,
  [1080] = 1080,
  [1081] = 664,
  [1082] = 665,
  [1083] = 666,
  [1084] = 667,
  [1085] = 1085,
  [1086] = 668,
  [1087] = 1085,
  [1088] = 669,
  [1089] = 1074,
  [1090] = 406,
  [1091] = 626,
  [1092] = 670,
  [1093] = 671,
  [1094] = 1074,
  [1095] = 673,
  [1096] = 423,
  [1097] = 406,
  [1098] = 628,
  [1099] = 414,
  [1100] = 517,
  [1101] = 1080,
  [1102] = 676,
  [1103] = 678,
  [1104] = 679,
  [1105] = 406,
  [1106] = 632,
  [1107] = 681,
  [1108] = 682,
  [1109] = 633,
  [1110] = 635,
  [1111] = 426,
  [1112] = 412,
  [1113] = 419,
  [1114] = 1114,
  [1115] = 424,
  [1116] = 428,
  [1117] = 613,
  [1118] = 411,
  [1119] = 1119,
  [1120] = 1120,
  [1121] = 409,
  [1122] = 1122,
  [1123] = 1123,
  [1124] = 420,
  [1125] = 1125,
  [1126] = 424,
  [1127] = 411,
  [1128] = 432,
  [1129] = 410,
  [1130] = 434,
  [1131] = 441,
  [1132] = 499,
  [1133] = 424,
  [1134] = 411,
  [1135] = 1135,
  [1136] = 407,
  [1137] = 1137,
  [1138] = 1138,
  [1139] = 518,
  [1140] = 1140,
  [1141] = 425,
  [1142] = 415,
  [1143] = 1080,
  [1144] = 429,
  [1145] = 427,
  [1146] = 421,
  [1147] = 418,
  [1148] = 405,
  [1149] = 1085,
  [1150] = 1080,
  [1151] = 612,
  [1152] = 1085,
  [1153] = 1153,
  [1154] = 416,
  [1155] = 451,
  [1156] = 426,
  [1157] = 626,
  [1158] = 1158,
  [1159] = 670,
  [1160] = 1160,
  [1161] = 1161,
  [1162] = 1162,
  [1163] = 411,
  [1164] = 612,
  [1165] = 424,
  [1166] = 613,
  [1167] = 411,
  [1168] = 583,
  [1169] = 671,
  [1170] = 467,
  [1171] = 614,
  [1172] = 441,
  [1173] = 420,
  [1174] = 820,
  [1175] = 673,
  [1176] = 1176,
  [1177] = 434,
  [1178] = 628,
  [1179] = 1179,
  [1180] = 1180,
  [1181] = 1181,
  [1182] = 1182,
  [1183] = 676,
  [1184] = 821,
  [1185] = 678,
  [1186] = 1186,
  [1187] = 424,
  [1188] = 428,
  [1189] = 423,
  [1190] = 679,
  [1191] = 828,
  [1192] = 632,
  [1193] = 633,
  [1194] = 1194,
  [1195] = 421,
  [1196] = 829,
  [1197] = 1197,
  [1198] = 663,
  [1199] = 1180,
  [1200] = 1200,
  [1201] = 664,
  [1202] = 1202,
  [1203] = 1203,
  [1204] = 1186,
  [1205] = 665,
  [1206] = 666,
  [1207] = 1207,
  [1208] = 667,
  [1209] = 668,
  [1210] = 669,
  [1211] = 1211,
  [1212] = 670,
  [1213] = 671,
  [1214] = 681,
  [1215] = 673,
  [1216] = 1216,
  [1217] = 614,
  [1218] = 682,
  [1219] = 663,
  [1220] = 664,
  [1221] = 476,
  [1222] = 665,
  [1223] = 666,
  [1224] = 411,
  [1225] = 667,
  [1226] = 668,
  [1227] = 443,
  [1228] = 635,
  [1229] = 429,
  [1230] = 559,
  [1231] = 465,
  [1232] = 466,
  [1233] = 463,
  [1234] = 432,
  [1235] = 676,
  [1236] = 425,
  [1237] = 410,
  [1238] = 1119,
  [1239] = 1207,
  [1240] = 415,
  [1241] = 1158,
  [1242] = 516,
  [1243] = 439,
  [1244] = 1244,
  [1245] = 678,
  [1246] = 679,
  [1247] = 424,
  [1248] = 822,
  [1249] = 414,
  [1250] = 681,
  [1251] = 682,
  [1252] = 416,
  [1253] = 1253,
  [1254] = 461,
  [1255] = 518,
  [1256] = 499,
  [1257] = 462,
  [1258] = 464,
  [1259] = 1253,
  [1260] = 1253,
  [1261] = 412,
  [1262] = 1262,
  [1263] = 450,
  [1264] = 1264,
  [1265] = 452,
  [1266] = 453,
  [1267] = 454,
  [1268] = 455,
  [1269] = 456,
  [1270] = 457,
  [1271] = 458,
  [1272] = 487,
  [1273] = 1273,
  [1274] = 1274,
  [1275] = 1275,
  [1276] = 517,
  [1277] = 1203,
  [1278] = 427,
  [1279] = 417,
  [1280] = 418,
  [1281] = 419,
  [1282] = 669,
  [1283] = 409,
  [1284] = 1284,
  [1285] = 499,
  [1286] = 668,
  [1287] = 1287,
  [1288] = 439,
  [1289] = 1289,
  [1290] = 1290,
  [1291] = 476,
  [1292] = 568,
  [1293] = 1293,
  [1294] = 465,
  [1295] = 1289,
  [1296] = 467,
  [1297] = 492,
  [1298] = 464,
  [1299] = 516,
  [1300] = 1293,
  [1301] = 487,
  [1302] = 669,
  [1303] = 828,
  [1304] = 461,
  [1305] = 457,
  [1306] = 678,
  [1307] = 1307,
  [1308] = 679,
  [1309] = 583,
  [1310] = 462,
  [1311] = 518,
  [1312] = 829,
  [1313] = 821,
  [1314] = 1314,
  [1315] = 456,
  [1316] = 463,
  [1317] = 1317,
  [1318] = 1318,
  [1319] = 450,
  [1320] = 1293,
  [1321] = 1290,
  [1322] = 559,
  [1323] = 1290,
  [1324] = 670,
  [1325] = 455,
  [1326] = 451,
  [1327] = 614,
  [1328] = 681,
  [1329] = 1158,
  [1330] = 682,
  [1331] = 663,
  [1332] = 1293,
  [1333] = 1289,
  [1334] = 1334,
  [1335] = 503,
  [1336] = 676,
  [1337] = 452,
  [1338] = 466,
  [1339] = 453,
  [1340] = 454,
  [1341] = 1216,
  [1342] = 1290,
  [1343] = 517,
  [1344] = 664,
  [1345] = 665,
  [1346] = 666,
  [1347] = 443,
  [1348] = 820,
  [1349] = 1289,
  [1350] = 671,
  [1351] = 667,
  [1352] = 673,
  [1353] = 822,
  [1354] = 458,
  [1355] = 1355,
  [1356] = 1356,
  [1357] = 1357,
  [1358] = 562,
  [1359] = 628,
  [1360] = 1360,
  [1361] = 676,
  [1362] = 583,
  [1363] = 678,
  [1364] = 679,
  [1365] = 1365,
  [1366] = 1360,
  [1367] = 559,
  [1368] = 1368,
  [1369] = 503,
  [1370] = 1370,
  [1371] = 1371,
  [1372] = 632,
  [1373] = 553,
  [1374] = 518,
  [1375] = 499,
  [1376] = 633,
  [1377] = 681,
  [1378] = 1378,
  [1379] = 682,
  [1380] = 1380,
  [1381] = 1381,
  [1382] = 568,
  [1383] = 1383,
  [1384] = 492,
  [1385] = 635,
  [1386] = 1386,
  [1387] = 1387,
  [1388] = 1388,
  [1389] = 612,
  [1390] = 613,
  [1391] = 614,
  [1392] = 1392,
  [1393] = 1393,
  [1394] = 1394,
  [1395] = 1357,
  [1396] = 1396,
  [1397] = 1307,
  [1398] = 1398,
  [1399] = 1399,
  [1400] = 1356,
  [1401] = 1370,
  [1402] = 1371,
  [1403] = 1403,
  [1404] = 1398,
  [1405] = 1399,
  [1406] = 1406,
  [1407] = 1407,
  [1408] = 1408,
  [1409] = 1365,
  [1410] = 1380,
  [1411] = 555,
  [1412] = 1394,
  [1413] = 450,
  [1414] = 451,
  [1415] = 453,
  [1416] = 454,
  [1417] = 455,
  [1418] = 456,
  [1419] = 457,
  [1420] = 458,
  [1421] = 487,
  [1422] = 461,
  [1423] = 462,
  [1424] = 463,
  [1425] = 464,
  [1426] = 467,
  [1427] = 1370,
  [1428] = 1371,
  [1429] = 517,
  [1430] = 1430,
  [1431] = 1406,
  [1432] = 1407,
  [1433] = 1433,
  [1434] = 1406,
  [1435] = 1407,
  [1436] = 1407,
  [1437] = 1407,
  [1438] = 1407,
  [1439] = 1407,
  [1440] = 1407,
  [1441] = 1407,
  [1442] = 1403,
  [1443] = 1430,
  [1444] = 1433,
  [1445] = 1368,
  [1446] = 663,
  [1447] = 664,
  [1448] = 665,
  [1449] = 666,
  [1450] = 667,
  [1451] = 668,
  [1452] = 669,
  [1453] = 626,
  [1454] = 670,
  [1455] = 671,
  [1456] = 1357,
  [1457] = 1407,
  [1458] = 465,
  [1459] = 466,
  [1460] = 452,
  [1461] = 1403,
  [1462] = 1430,
  [1463] = 1433,
  [1464] = 1368,
  [1465] = 1357,
  [1466] = 673,
  [1467] = 1403,
  [1468] = 1368,
  [1469] = 1403,
  [1470] = 1368,
  [1471] = 560,
  [1472] = 581,
  [1473] = 564,
  [1474] = 565,
  [1475] = 573,
  [1476] = 575,
  [1477] = 1357,
  [1478] = 1430,
  [1479] = 1433,
  [1480] = 1403,
  [1481] = 1368,
  [1482] = 569,
  [1483] = 1408,
  [1484] = 1430,
  [1485] = 1433,
  [1486] = 1357,
  [1487] = 1430,
  [1488] = 1433,
  [1489] = 1388,
  [1490] = 575,
  [1491] = 555,
  [1492] = 1492,
  [1493] = 1493,
  [1494] = 559,
  [1495] = 562,
  [1496] = 569,
  [1497] = 1497,
  [1498] = 1498,
  [1499] = 1499,
  [1500] = 842,
  [1501] = 1501,
  [1502] = 1502,
  [1503] = 1360,
  [1504] = 1504,
  [1505] = 560,
  [1506] = 1506,
  [1507] = 583,
  [1508] = 553,
  [1509] = 1509,
  [1510] = 1493,
  [1511] = 1511,
  [1512] = 1493,
  [1513] = 581,
  [1514] = 564,
  [1515] = 565,
  [1516] = 573,
  [1517] = 1517,
  [1518] = 1518,
  [1519] = 1519,
  [1520] = 1520,
  [1521] = 1521,
  [1522] = 1522,
  [1523] = 1523,
  [1524] = 1524,
  [1525] = 1525,
  [1526] = 1526,
  [1527] = 1527,
  [1528] = 1528,
  [1529] = 1529,
  [1530] = 1530,
  [1531] = 1531,
  [1532] = 1532,
  [1533] = 1533,
  [1534] = 1534,
  [1535] = 1535,
  [1536] = 1533,
  [1537] = 1537,
  [1538] = 1538,
  [1539] = 1528,
  [1540] = 1540,
  [1541] = 1502,
  [1542] = 1542,
  [1543] = 1543,
  [1544] = 1526,
  [1545] = 1545,
  [1546] = 1546,
  [1547] = 1547,
  [1548] = 1548,
  [1549] = 1549,
  [1550] = 1550,
  [1551] = 1551,
  [1552] = 1552,
  [1553] = 1553,
  [1554] = 1554,
  [1555] = 1555,
  [1556] = 1556,
  [1557] = 1557,
  [1558] = 1558,
  [1559] = 1499,
  [1560] = 1560,
  [1561] = 1561,
  [1562] = 1562,
  [1563] = 1563,
  [1564] = 1564,
  [1565] = 1565,
  [1566] = 1566,
  [1567] = 1567,
  [1568] = 1568,
  [1569] = 1569,
  [1570] = 1570,
  [1571] = 1571,
  [1572] = 1572,
  [1573] = 1573,
  [1574] = 1574,
  [1575] = 1575,
  [1576] = 1576,
  [1577] = 1577,
  [1578] = 1578,
  [1579] = 1579,
  [1580] = 1580,
  [1581] = 1581,
  [1582] = 1582,
  [1583] = 1583,
  [1584] = 1584,
  [1585] = 1585,
  [1586] = 1586,
  [1587] = 1587,
  [1588] = 1588,
  [1589] = 1589,
  [1590] = 1590,
  [1591] = 1591,
  [1592] = 1592,
  [1593] = 1593,
  [1594] = 1594,
  [1595] = 1567,
  [1596] = 1596,
  [1597] = 1597,
  [1598] = 1533,
  [1599] = 1597,
  [1600] = 1596,
  [1601] = 842,
  [1602] = 1602,
  [1603] = 1603,
  [1604] = 1602,
  [1605] = 1605,
  [1606] = 1606,
  [1607] = 1607,
  [1608] = 1607,
  [1609] = 1609,
  [1610] = 1610,
  [1611] = 1611,
  [1612] = 1612,
  [1613] = 1160,
  [1614] = 1614,
  [1615] = 1615,
  [1616] = 1616,
  [1617] = 1617,
  [1618] = 1618,
  [1619] = 1619,
  [1620] = 1620,
  [1621] = 1621,
  [1622] = 1622,
  [1623] = 1623,
  [1624] = 1624,
  [1625] = 427,
  [1626] = 417,
  [1627] = 1627,
  [1628] = 429,
  [1629] = 418,
  [1630] = 419,
  [1631] = 426,
  [1632] = 1632,
  [1633] = 1633,
  [1634] = 1634,
  [1635] = 1635,
  [1636] = 414,
  [1637] = 1637,
  [1638] = 1638,
  [1639] = 1639,
  [1640] = 1640,
  [1641] = 1641,
  [1642] = 1642,
  [1643] = 1643,
  [1644] = 1644,
  [1645] = 1645,
  [1646] = 1645,
  [1647] = 1647,
  [1648] = 1645,
  [1649] = 1649,
  [1650] = 1650,
  [1651] = 1651,
  [1652] = 1652,
  [1653] = 1653,
  [1654] = 1654,
  [1655] = 1644,
  [1656] = 1644,
  [1657] = 1657,
  [1658] = 1644,
  [1659] = 1645,
  [1660] = 1660,
  [1661] = 1661,
  [1662] = 1662,
  [1663] = 1663,
  [1664] = 1664,
  [1665] = 1665,
  [1666] = 1666,
  [1667] = 1667,
  [1668] = 1668,
  [1669] = 1669,
  [1670] = 1670,
  [1671] = 1671,
  [1672] = 1672,
  [1673] = 1673,
  [1674] = 1674,
  [1675] = 1675,
  [1676] = 1676,
  [1677] = 1677,
  [1678] = 1678,
  [1679] = 1679,
  [1680] = 1671,
  [1681] = 1681,
  [1682] = 1672,
  [1683] = 1673,
  [1684] = 1674,
  [1685] = 1685,
  [1686] = 1686,
  [1687] = 1681,
  [1688] = 1675,
  [1689] = 427,
  [1690] = 418,
  [1691] = 1681,
  [1692] = 419,
  [1693] = 1693,
  [1694] = 1694,
  [1695] = 426,
  [1696] = 1696,
  [1697] = 1697,
  [1698] = 1698,
  [1699] = 1686,
  [1700] = 1686,
  [1701] = 1681,
  [1702] = 1686,
  [1703] = 1662,
  [1704] = 1663,
  [1705] = 1681,
  [1706] = 1686,
  [1707] = 1664,
  [1708] = 1681,
  [1709] = 1698,
  [1710] = 1681,
  [1711] = 1686,
  [1712] = 1681,
  [1713] = 1686,
  [1714] = 1714,
  [1715] = 842,
  [1716] = 1681,
  [1717] = 1686,
  [1718] = 1694,
  [1719] = 1719,
  [1720] = 1720,
  [1721] = 1721,
  [1722] = 1722,
  [1723] = 1723,
  [1724] = 1681,
  [1725] = 1725,
  [1726] = 1686,
  [1727] = 1676,
  [1728] = 1665,
  [1729] = 1666,
  [1730] = 1667,
  [1731] = 1668,
  [1732] = 1669,
  [1733] = 1733,
  [1734] = 1679,
  [1735] = 1661,
  [1736] = 1721,
  [1737] = 1670,
  [1738] = 1714,
  [1739] = 1719,
  [1740] = 1720,
  [1741] = 1722,
  [1742] = 1693,
  [1743] = 1723,
  [1744] = 1686,
  [1745] = 1745,
  [1746] = 1746,
  [1747] = 1747,
  [1748] = 1748,
  [1749] = 1749,
  [1750] = 1750,
  [1751] = 1751,
  [1752] = 1752,
  [1753] = 1024,
  [1754] = 1754,
  [1755] = 1755,
  [1756] = 1756,
  [1757] = 1757,
  [1758] = 1758,
  [1759] = 1759,
  [1760] = 1760,
  [1761] = 1761,
  [1762] = 1762,
  [1763] = 1763,
  [1764] = 1764,
  [1765] = 1765,
  [1766] = 1766,
  [1767] = 1767,
  [1768] = 1768,
  [1769] = 1769,
  [1770] = 1770,
  [1771] = 1771,
  [1772] = 842,
  [1773] = 1293,
  [1774] = 1290,
  [1775] = 1289,
  [1776] = 1776,
  [1777] = 1777,
  [1778] = 1778,
  [1779] = 1778,
  [1780] = 1780,
  [1781] = 1781,
  [1782] = 1778,
  [1783] = 1778,
  [1784] = 1781,
  [1785] = 1381,
  [1786] = 1383,
  [1787] = 1781,
  [1788] = 1778,
  [1789] = 1781,
  [1790] = 1778,
  [1791] = 1290,
  [1792] = 1760,
  [1793] = 1781,
  [1794] = 1293,
  [1795] = 1795,
  [1796] = 1778,
  [1797] = 1778,
  [1798] = 1781,
  [1799] = 1194,
  [1800] = 1781,
  [1801] = 1778,
  [1802] = 1781,
  [1803] = 1803,
  [1804] = 1778,
  [1805] = 1764,
  [1806] = 1781,
  [1807] = 1761,
  [1808] = 1781,
  [1809] = 1778,
  [1810] = 1769,
  [1811] = 1781,
  [1812] = 1812,
  [1813] = 1289,
  [1814] = 1778,
  [1815] = 1815,
  [1816] = 1816,
  [1817] = 1817,
  [1818] = 1781,
  [1819] = 1819,
  [1820] = 1820,
  [1821] = 1819,
  [1822] = 1822,
  [1823] = 842,
  [1824] = 1497,
  [1825] = 1509,
  [1826] = 1826,
  [1827] = 1827,
  [1828] = 1828,
  [1829] = 1829,
  [1830] = 1830,
  [1831] = 1392,
  [1832] = 1819,
  [1833] = 1833,
  [1834] = 1826,
  [1835] = 1822,
  [1836] = 1836,
  [1837] = 1837,
  [1838] = 1837,
  [1839] = 1839,
  [1840] = 1840,
  [1841] = 1841,
  [1842] = 1842,
  [1843] = 1843,
  [1844] = 1844,
  [1845] = 1827,
  [1846] = 1846,
  [1847] = 1847,
  [1848] = 1848,
  [1849] = 1849,
  [1850] = 1850,
  [1851] = 1851,
  [1852] = 1852,
  [1853] = 1853,
  [1854] = 1854,
  [1855] = 1830,
  [1856] = 1833,
  [1857] = 1857,
  [1858] = 1858,
  [1859] = 1859,
  [1860] = 1828,
  [1861] = 1861,
  [1862] = 1843,
  [1863] = 1863,
  [1864] = 1502,
  [1865] = 1865,
  [1866] = 1863,
  [1867] = 1846,
  [1868] = 1868,
  [1869] = 1857,
  [1870] = 1847,
  [1871] = 1871,
  [1872] = 1872,
  [1873] = 1849,
  [1874] = 1851,
  [1875] = 1396,
  [1876] = 1858,
  [1877] = 1836,
  [1878] = 1826,
  [1879] = 1879,
  [1880] = 1880,
  [1881] = 1881,
  [1882] = 1882,
  [1883] = 1844,
  [1884] = 1859,
  [1885] = 1499,
  [1886] = 1886,
  [1887] = 1887,
  [1888] = 1888,
  [1889] = 1827,
  [1890] = 1861,
  [1891] = 1891,
  [1892] = 1892,
  [1893] = 1893,
  [1894] = 1842,
  [1895] = 1895,
  [1896] = 1896,
  [1897] = 1882,
  [1898] = 1844,
  [1899] = 1819,
  [1900] = 1822,
  [1901] = 1826,
  [1902] = 1827,
  [1903] = 1882,
  [1904] = 1819,
  [1905] = 1882,
  [1906] = 1844,
  [1907] = 1819,
  [1908] = 1822,
  [1909] = 1826,
  [1910] = 1827,
  [1911] = 1822,
  [1912] = 1844,
  [1913] = 1913,
  [1914] = 1914,
  [1915] = 1915,
  [1916] = 1882,
  [1917] = 1844,
  [1918] = 1918,
  [1919] = 1822,
  [1920] = 1826,
  [1921] = 1827,
  [1922] = 1922,
  [1923] = 1923,
  [1924] = 1881,
  [1925] = 1882,
  [1926] = 1926,
  [1927] = 1895,
  [1928] = 1928,
  [1929] = 1929,
  [1930] = 1930,
  [1931] = 1931,
  [1932] = 1932,
  [1933] = 1933,
  [1934] = 1934,
  [1935] = 1935,
  [1936] = 1936,
  [1937] = 1937,
  [1938] = 407,
  [1939] = 1939,
  [1940] = 1940,
  [1941] = 1941,
  [1942] = 1942,
  [1943] = 1929,
  [1944] = 1937,
  [1945] = 1945,
  [1946] = 1946,
  [1947] = 1947,
  [1948] = 1928,
  [1949] = 1949,
  [1950] = 1504,
  [1951] = 1951,
  [1952] = 1952,
  [1953] = 1953,
  [1954] = 1954,
  [1955] = 1955,
  [1956] = 1930,
  [1957] = 1957,
  [1958] = 1958,
  [1959] = 1959,
  [1960] = 1952,
  [1961] = 1953,
  [1962] = 1962,
  [1963] = 1963,
  [1964] = 1504,
  [1965] = 1506,
  [1966] = 1935,
  [1967] = 1931,
  [1968] = 1968,
  [1969] = 1946,
  [1970] = 1970,
  [1971] = 1934,
  [1972] = 1936,
  [1973] = 1973,
  [1974] = 1937,
  [1975] = 1946,
  [1976] = 1936,
  [1977] = 1955,
  [1978] = 1978,
  [1979] = 1979,
  [1980] = 1980,
  [1981] = 1981,
  [1982] = 1982,
  [1983] = 1983,
  [1984] = 1939,
  [1985] = 1968,
  [1986] = 1511,
  [1987] = 1987,
  [1988] = 1928,
  [1989] = 1955,
  [1990] = 1887,
  [1991] = 842,
  [1992] = 1992,
  [1993] = 1928,
  [1994] = 1994,
  [1995] = 1995,
  [1996] = 1996,
  [1997] = 1997,
  [1998] = 1998,
  [1999] = 1999,
  [2000] = 2000,
  [2001] = 2001,
  [2002] = 2002,
  [2003] = 2003,
  [2004] = 1871,
  [2005] = 2005,
  [2006] = 1947,
  [2007] = 2007,
  [2008] = 2008,
  [2009] = 2009,
  [2010] = 2010,
  [2011] = 2011,
  [2012] = 2012,
  [2013] = 2007,
  [2014] = 1962,
  [2015] = 2015,
  [2016] = 1936,
  [2017] = 1946,
  [2018] = 1963,
  [2019] = 2019,
  [2020] = 1955,
  [2021] = 2021,
  [2022] = 1934,
  [2023] = 2023,
  [2024] = 1951,
  [2025] = 2023,
  [2026] = 2003,
  [2027] = 1936,
  [2028] = 1946,
  [2029] = 1942,
  [2030] = 1511,
  [2031] = 1498,
  [2032] = 1501,
  [2033] = 2033,
  [2034] = 1498,
  [2035] = 2035,
  [2036] = 405,
  [2037] = 1936,
  [2038] = 1946,
  [2039] = 842,
  [2040] = 1968,
  [2041] = 2041,
  [2042] = 2042,
  [2043] = 2043,
  [2044] = 2044,
  [2045] = 2045,
  [2046] = 2046,
  [2047] = 2047,
  [2048] = 2048,
  [2049] = 2049,
  [2050] = 2050,
  [2051] = 2051,
  [2052] = 2052,
  [2053] = 2053,
  [2054] = 553,
  [2055] = 2052,
  [2056] = 2056,
  [2057] = 2057,
  [2058] = 1568,
  [2059] = 2059,
  [2060] = 1570,
  [2061] = 2061,
  [2062] = 1575,
  [2063] = 1576,
  [2064] = 2033,
  [2065] = 2046,
  [2066] = 2066,
  [2067] = 2067,
  [2068] = 2052,
  [2069] = 1584,
  [2070] = 2070,
  [2071] = 1585,
  [2072] = 2072,
  [2073] = 2073,
  [2074] = 2035,
  [2075] = 2075,
  [2076] = 2046,
  [2077] = 2052,
  [2078] = 1565,
  [2079] = 1569,
  [2080] = 2046,
  [2081] = 2081,
  [2082] = 2082,
  [2083] = 2052,
  [2084] = 2084,
  [2085] = 1567,
  [2086] = 2086,
  [2087] = 2087,
  [2088] = 2046,
  [2089] = 2089,
  [2090] = 2052,
  [2091] = 1556,
  [2092] = 2092,
  [2093] = 553,
  [2094] = 1577,
  [2095] = 2095,
  [2096] = 2096,
  [2097] = 2097,
  [2098] = 2098,
  [2099] = 1543,
  [2100] = 2100,
  [2101] = 2101,
  [2102] = 2102,
  [2103] = 1546,
  [2104] = 2104,
  [2105] = 2105,
  [2106] = 2106,
  [2107] = 1580,
  [2108] = 1582,
  [2109] = 1590,
  [2110] = 1593,
  [2111] = 1517,
  [2112] = 1594,
  [2113] = 1550,
  [2114] = 2114,
  [2115] = 2115,
  [2116] = 1941,
  [2117] = 2059,
  [2118] = 2118,
  [2119] = 2119,
  [2120] = 1525,
  [2121] = 1551,
  [2122] = 1530,
  [2123] = 2052,
  [2124] = 2008,
  [2125] = 1552,
  [2126] = 1554,
  [2127] = 2127,
  [2128] = 2128,
  [2129] = 1555,
  [2130] = 1557,
  [2131] = 1932,
  [2132] = 1581,
  [2133] = 2133,
  [2134] = 1583,
  [2135] = 1586,
  [2136] = 2136,
  [2137] = 2137,
  [2138] = 1545,
  [2139] = 1591,
  [2140] = 1531,
  [2141] = 1532,
  [2142] = 1560,
  [2143] = 1563,
  [2144] = 2144,
  [2145] = 2046,
  [2146] = 1571,
  [2147] = 2118,
  [2148] = 1562,
  [2149] = 1578,
  [2150] = 1520,
  [2151] = 1523,
  [2152] = 1524,
  [2153] = 1527,
  [2154] = 2119,
  [2155] = 1518,
  [2156] = 1522,
  [2157] = 1542,
  [2158] = 1592,
  [2159] = 1529,
  [2160] = 2105,
  [2161] = 1534,
  [2162] = 2162,
  [2163] = 2075,
  [2164] = 1537,
  [2165] = 1538,
  [2166] = 1540,
  [2167] = 2167,
  [2168] = 2168,
  [2169] = 1528,
  [2170] = 2170,
  [2171] = 2171,
  [2172] = 2101,
  [2173] = 2173,
  [2174] = 2052,
  [2175] = 2095,
  [2176] = 2021,
  [2177] = 2170,
  [2178] = 2178,
  [2179] = 2179,
  [2180] = 1547,
  [2181] = 1548,
  [2182] = 1553,
  [2183] = 2127,
  [2184] = 1561,
  [2185] = 2185,
  [2186] = 1564,
  [2187] = 1566,
  [2188] = 1574,
  [2189] = 1587,
  [2190] = 1588,
  [2191] = 2191,
  [2192] = 2192,
  [2193] = 2193,
  [2194] = 1589,
  [2195] = 2046,
  [2196] = 2196,
  [2197] = 2197,
  [2198] = 2198,
  [2199] = 2199,
  [2200] = 1549,
  [2201] = 2201,
  [2202] = 1521,
  [2203] = 1572,
  [2204] = 1535,
  [2205] = 2205,
  [2206] = 2019,
  [2207] = 1526,
  [2208] = 2102,
  [2209] = 2209,
  [2210] = 2162,
  [2211] = 2173,
  [2212] = 2212,
  [2213] = 1573,
  [2214] = 2053,
  [2215] = 2041,
  [2216] = 2216,
  [2217] = 1519,
  [2218] = 2046,
  [2219] = 1558,
  [2220] = 2220,
  [2221] = 2221,
  [2222] = 2222,
  [2223] = 2223,
  [2224] = 2224,
  [2225] = 2225,
  [2226] = 2226,
  [2227] = 2227,
  [2228] = 2228,
  [2229] = 2229,
  [2230] = 2230,
  [2231] = 2231,
  [2232] = 2232,
  [2233] = 2232,
  [2234] = 2234,
  [2235] = 2235,
  [2236] = 2236,
  [2237] = 2237,
  [2238] = 2234,
  [2239] = 2239,
  [2240] = 2240,
  [2241] = 2241,
  [2242] = 2242,
  [2243] = 2243,
  [2244] = 2226,
  [2245] = 2245,
  [2246] = 2246,
  [2247] = 2247,
  [2248] = 2248,
  [2249] = 2222,
  [2250] = 2225,
  [2251] = 2251,
  [2252] = 2252,
  [2253] = 2253,
  [2254] = 2254,
  [2255] = 2255,
  [2256] = 2220,
  [2257] = 2252,
  [2258] = 2258,
  [2259] = 2229,
  [2260] = 2260,
  [2261] = 2224,
  [2262] = 2234,
  [2263] = 2236,
  [2264] = 2237,
  [2265] = 2243,
  [2266] = 2226,
  [2267] = 2245,
  [2268] = 2222,
  [2269] = 2253,
  [2270] = 2254,
  [2271] = 2271,
  [2272] = 2272,
  [2273] = 2242,
  [2274] = 2245,
  [2275] = 2275,
  [2276] = 2229,
  [2277] = 2234,
  [2278] = 2236,
  [2279] = 2237,
  [2280] = 2243,
  [2281] = 2226,
  [2282] = 2245,
  [2283] = 2222,
  [2284] = 2253,
  [2285] = 2254,
  [2286] = 2232,
  [2287] = 2234,
  [2288] = 2288,
  [2289] = 2289,
  [2290] = 2229,
  [2291] = 2252,
  [2292] = 2243,
  [2293] = 2234,
  [2294] = 2236,
  [2295] = 2237,
  [2296] = 2247,
  [2297] = 2243,
  [2298] = 2226,
  [2299] = 2245,
  [2300] = 2222,
  [2301] = 2254,
  [2302] = 2302,
  [2303] = 2235,
  [2304] = 2229,
  [2305] = 2305,
  [2306] = 2306,
  [2307] = 2234,
  [2308] = 2236,
  [2309] = 2237,
  [2310] = 2243,
  [2311] = 2226,
  [2312] = 2245,
  [2313] = 2222,
  [2314] = 2254,
  [2315] = 2235,
  [2316] = 2316,
  [2317] = 2229,
  [2318] = 2236,
  [2319] = 2234,
  [2320] = 2236,
  [2321] = 2237,
  [2322] = 2243,
  [2323] = 2226,
  [2324] = 2245,
  [2325] = 2222,
  [2326] = 2254,
  [2327] = 2236,
  [2328] = 1602,
  [2329] = 2237,
  [2330] = 2229,
  [2331] = 2331,
  [2332] = 2234,
  [2333] = 2236,
  [2334] = 2237,
  [2335] = 2243,
  [2336] = 2226,
  [2337] = 2245,
  [2338] = 2222,
  [2339] = 2254,
  [2340] = 2340,
  [2341] = 2228,
  [2342] = 2254,
  [2343] = 2246,
  [2344] = 2305,
  [2345] = 2345,
  [2346] = 2227,
  [2347] = 2258,
  [2348] = 2348,
  [2349] = 2254,
  [2350] = 2229,
  [2351] = 2351,
  [2352] = 2255,
  [2353] = 2353,
  [2354] = 2354,
  [2355] = 2237,
  [2356] = 2356,
  [2357] = 2357,
  [2358] = 2358,
  [2359] = 2359,
  [2360] = 2358,
  [2361] = 2353,
  [2362] = 2362,
  [2363] = 2228,
  [2364] = 2364,
  [2365] = 2365,
  [2366] = 2229,
  [2367] = 2205,
  [2368] = 2368,
  [2369] = 2369,
  [2370] = 2253,
  [2371] = 2220,
  [2372] = 2372,
  [2373] = 2373,
  [2374] = 2374,
  [2375] = 2288,
  [2376] = 2376,
  [2377] = 2254,
  [2378] = 2378,
  [2379] = 2272,
  [2380] = 2357,
  [2381] = 2381,
  [2382] = 2382,
  [2383] = 2383,
  [2384] = 2384,
  [2385] = 2385,
  [2386] = 2305,
  [2387] = 2222,
  [2388] = 2351,
  [2389] = 2220,
  [2390] = 2225,
  [2391] = 2391,
  [2392] = 2381,
  [2393] = 2305,
  [2394] = 2242,
  [2395] = 2395,
  [2396] = 2243,
  [2397] = 2224,
  [2398] = 2398,
  [2399] = 2399,
  [2400] = 2400,
  [2401] = 2401,
  [2402] = 2402,
  [2403] = 2226,
  [2404] = 2348,
  [2405] = 2228,
  [2406] = 2229,
  [2407] = 2407,
  [2408] = 2245,
  [2409] = 2316,
  [2410] = 2399,
  [2411] = 2231,
  [2412] = 2247,
  [2413] = 2401,
  [2414] = 2414,
  [2415] = 2415,
  [2416] = 2365,
  [2417] = 2417,
  [2418] = 2260,
  [2419] = 2419,
  [2420] = 2420,
  [2421] = 2354,
  [2422] = 2232,
  [2423] = 2220,
  [2424] = 2234,
  [2425] = 2425,
  [2426] = 2235,
  [2427] = 2236,
  [2428] = 2237,
  [2429] = 2429,
  [2430] = 2430,
  [2431] = 2417,
  [2432] = 2242,
  [2433] = 2243,
  [2434] = 2226,
  [2435] = 2383,
  [2436] = 2436,
  [2437] = 2245,
  [2438] = 2359,
  [2439] = 2414,
  [2440] = 2407,
  [2441] = 2247,
  [2442] = 2442,
  [2443] = 2222,
  [2444] = 2444,
  [2445] = 2225,
  [2446] = 2446,
  [2447] = 2253,
  [2448] = 2252,
  [2449] = 2305,
  [2450] = 2419,
  [2451] = 2220,
  [2452] = 2253,
  [2453] = 2398,
  [2454] = 2384,
  [2455] = 2455,
  [2456] = 2254,
  [2457] = 2457,
  [2458] = 2458,
  [2459] = 2459,
  [2460] = 2305,
  [2461] = 2220,
  [2462] = 2254,
  [2463] = 2463,
  [2464] = 2464,
  [2465] = 2220,
  [2466] = 2398,
  [2467] = 2398,
  [2468] = 2398,
  [2469] = 2224,
  [2470] = 2470,
  [2471] = 2220,
  [2472] = 2472,
  [2473] = 2220,
  [2474] = 2220,
  [2475] = 2220,
  [2476] = 2398,
  [2477] = 2220,
  [2478] = 2253,
  [2479] = 2479,
  [2480] = 2480,
  [2481] = 2481,
  [2482] = 2482,
  [2483] = 2483,
  [2484] = 2484,
  [2485] = 2485,
  [2486] = 2486,
  [2487] = 2487,
  [2488] = 2488,
  [2489] = 2489,
  [2490] = 2490,
  [2491] = 405,
  [2492] = 2486,
  [2493] = 2493,
  [2494] = 2494,
  [2495] = 2495,
  [2496] = 2484,
  [2497] = 2497,
  [2498] = 517,
  [2499] = 2482,
  [2500] = 2500,
  [2501] = 2482,
  [2502] = 2486,
  [2503] = 2495,
  [2504] = 2504,
  [2505] = 2505,
  [2506] = 2495,
  [2507] = 2484,
  [2508] = 2508,
  [2509] = 517,
  [2510] = 2495,
  [2511] = 2511,
  [2512] = 2495,
  [2513] = 2495,
  [2514] = 2514,
  [2515] = 407,
  [2516] = 2508,
  [2517] = 2517,
  [2518] = 2518,
  [2519] = 2484,
  [2520] = 2520,
  [2521] = 2521,
  [2522] = 2482,
  [2523] = 2479,
  [2524] = 2524,
  [2525] = 407,
  [2526] = 2521,
  [2527] = 2511,
  [2528] = 2528,
  [2529] = 2529,
  [2530] = 2528,
  [2531] = 2485,
  [2532] = 2532,
  [2533] = 2524,
  [2534] = 2486,
  [2535] = 2535,
  [2536] = 2536,
  [2537] = 499,
  [2538] = 2482,
  [2539] = 2539,
  [2540] = 2482,
  [2541] = 2541,
  [2542] = 2542,
  [2543] = 2260,
  [2544] = 2544,
  [2545] = 2495,
  [2546] = 2546,
  [2547] = 2547,
  [2548] = 2514,
  [2549] = 2549,
  [2550] = 2504,
  [2551] = 2517,
  [2552] = 499,
  [2553] = 2546,
  [2554] = 2483,
  [2555] = 2494,
  [2556] = 518,
  [2557] = 2494,
  [2558] = 2558,
  [2559] = 2495,
  [2560] = 2560,
  [2561] = 2487,
  [2562] = 2524,
  [2563] = 2490,
  [2564] = 2549,
  [2565] = 2504,
  [2566] = 2517,
  [2567] = 2546,
  [2568] = 2487,
  [2569] = 2549,
  [2570] = 2504,
  [2571] = 2517,
  [2572] = 2546,
  [2573] = 2487,
  [2574] = 2549,
  [2575] = 2546,
  [2576] = 2549,
  [2577] = 2546,
  [2578] = 2482,
  [2579] = 405,
  [2580] = 2580,
  [2581] = 2581,
  [2582] = 2582,
  [2583] = 2583,
  [2584] = 2584,
  [2585] = 2585,
  [2586] = 2482,
  [2587] = 2490,
  [2588] = 2482,
  [2589] = 2483,
  [2590] = 2485,
  [2591] = 2591,
  [2592] = 2482,
  [2593] = 2504,
  [2594] = 2517,
  [2595] = 2487,
  [2596] = 2549,
  [2597] = 2482,
  [2598] = 2504,
  [2599] = 2517,
  [2600] = 2487,
  [2601] = 2520,
  [2602] = 575,
  [2603] = 2603,
  [2604] = 2604,
  [2605] = 2605,
  [2606] = 2606,
  [2607] = 2607,
  [2608] = 2608,
  [2609] = 2609,
  [2610] = 2610,
  [2611] = 2611,
  [2612] = 2612,
  [2613] = 2610,
  [2614] = 2611,
  [2615] = 2615,
  [2616] = 2616,
  [2617] = 2617,
  [2618] = 2618,
  [2619] = 2619,
  [2620] = 2620,
  [2621] = 821,
  [2622] = 2622,
  [2623] = 2623,
  [2624] = 2610,
  [2625] = 2610,
  [2626] = 2610,
  [2627] = 2611,
  [2628] = 2611,
  [2629] = 2629,
  [2630] = 405,
  [2631] = 409,
  [2632] = 2632,
  [2633] = 2633,
  [2634] = 2634,
  [2635] = 2635,
  [2636] = 2634,
  [2637] = 2611,
  [2638] = 2638,
  [2639] = 2610,
  [2640] = 2611,
  [2641] = 2606,
  [2642] = 2642,
  [2643] = 2643,
  [2644] = 2644,
  [2645] = 2645,
  [2646] = 2633,
  [2647] = 2606,
  [2648] = 2648,
  [2649] = 2644,
  [2650] = 2650,
  [2651] = 2603,
  [2652] = 2652,
  [2653] = 412,
  [2654] = 2633,
  [2655] = 2623,
  [2656] = 2656,
  [2657] = 2629,
  [2658] = 2658,
  [2659] = 2615,
  [2660] = 2660,
  [2661] = 2645,
  [2662] = 2603,
  [2663] = 2663,
  [2664] = 2648,
  [2665] = 2665,
  [2666] = 2603,
  [2667] = 2667,
  [2668] = 2668,
  [2669] = 2623,
  [2670] = 2670,
  [2671] = 2617,
  [2672] = 2658,
  [2673] = 2673,
  [2674] = 2623,
  [2675] = 2633,
  [2676] = 2616,
  [2677] = 2618,
  [2678] = 2665,
  [2679] = 2617,
  [2680] = 407,
  [2681] = 2673,
  [2682] = 2606,
  [2683] = 2683,
  [2684] = 2667,
  [2685] = 2668,
  [2686] = 2686,
  [2687] = 2665,
  [2688] = 2688,
  [2689] = 2633,
  [2690] = 2604,
  [2691] = 2603,
  [2692] = 2606,
  [2693] = 828,
  [2694] = 2694,
  [2695] = 2683,
  [2696] = 2696,
  [2697] = 2697,
  [2698] = 2606,
  [2699] = 2699,
  [2700] = 2700,
  [2701] = 2656,
  [2702] = 2632,
  [2703] = 2660,
  [2704] = 2660,
  [2705] = 2619,
  [2706] = 2665,
  [2707] = 2603,
  [2708] = 2656,
  [2709] = 2633,
  [2710] = 583,
  [2711] = 2617,
  [2712] = 2712,
  [2713] = 2617,
  [2714] = 2610,
  [2715] = 2611,
  [2716] = 2716,
  [2717] = 2717,
  [2718] = 2718,
  [2719] = 2617,
  [2720] = 2720,
  [2721] = 2603,
  [2722] = 2635,
  [2723] = 2617,
  [2724] = 559,
  [2725] = 2645,
  [2726] = 2726,
  [2727] = 2606,
  [2728] = 829,
  [2729] = 2633,
  [2730] = 2730,
  [2731] = 2731,
  [2732] = 2663,
  [2733] = 2733,
  [2734] = 2603,
  [2735] = 2735,
  [2736] = 2622,
  [2737] = 2673,
  [2738] = 2608,
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(143);
      ADVANCE_MAP(
        '\r', 238,
        '!', 162,
        '"', 247,
        '%', 172,
        '&', 196,
        '\'', 237,
        '(', 148,
        ')', 147,
        '*', 168,
        '+', 154,
        ',', 207,
        '-', 157,
        '.', 205,
        '/', 170,
        ':', 199,
        ';', 209,
        '<', 187,
        '=', 152,
        '>', 182,
        '?', 217,
        '[', 221,
        '\\', 218,
        ']', 222,
        '^', 177,
        '{', 149,
        '|', 174,
        '}', 150,
        '~', 161,
        '\t', 242,
        ' ', 242,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(212);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(213);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n') ADVANCE(208);
      END_STATE();
    case 1:
      if (lookahead == '\n') ADVANCE(344);
      if (lookahead == '\\') ADVANCE(5);
      if (lookahead != 0) ADVANCE(1);
      END_STATE();
    case 2:
      if (lookahead == '\n') ADVANCE(347);
      END_STATE();
    case 3:
      if (lookahead == '\n') ADVANCE(347);
      if (lookahead == '\r') ADVANCE(2);
      END_STATE();
    case 4:
      if (lookahead == '\n') ADVANCE(347);
      if (lookahead == '\r') ADVANCE(2);
      if (lookahead == '`') ADVANCE(138);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(307);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(60);
      END_STATE();
    case 5:
      if (lookahead == '\n') ADVANCE(345);
      if (lookahead == '\r') ADVANCE(6);
      if (lookahead == '\\') ADVANCE(5);
      if (lookahead != 0) ADVANCE(1);
      END_STATE();
    case 6:
      if (lookahead == '\n') ADVANCE(345);
      if (lookahead == '\\') ADVANCE(5);
      if (lookahead != 0) ADVANCE(1);
      END_STATE();
    case 7:
      if (lookahead == '\n') ADVANCE(346);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        '\n', 346,
        '\r', 7,
        '!', 163,
        '"', 247,
        '#', 116,
        '%', 172,
        '&', 196,
        '\'', 237,
        '(', 148,
        '*', 168,
        '+', 155,
        ',', 207,
        '-', 158,
        '.', 49,
        '/', 170,
        '0', 42,
        ':', 200,
        ';', 1,
        '<', 188,
        '=', 153,
        '>', 183,
        '\\', 219,
        '^', 179,
        '_', 330,
        '`', 135,
        '{', 149,
        '|', 176,
        '~', 161,
        '\t', 350,
        ' ', 350,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(43);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(329);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        '\n', 346,
        '\r', 7,
        '!', 163,
        '"', 247,
        '#', 116,
        '%', 172,
        '&', 197,
        '\'', 237,
        '(', 148,
        ')', 147,
        '*', 168,
        '+', 155,
        ',', 207,
        '-', 160,
        '.', 206,
        '/', 170,
        '0', 42,
        ':', 54,
        ';', 1,
        '<', 188,
        '=', 153,
        '>', 183,
        '?', 217,
        'B', 327,
        '[', 221,
        '\\', 219,
        ']', 222,
        '^', 179,
        '_', 330,
        '`', 135,
        '{', 149,
        '|', 176,
        '~', 161,
        '\t', 350,
        ' ', 350,
        'V', 324,
        'v', 324,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(43);
      if (('A' <= lookahead && lookahead <= 'z')) ADVANCE(329);
      END_STATE();
    case 10:
      ADVANCE_MAP(
        '\n', 346,
        '\r', 7,
        '!', 163,
        '"', 247,
        '%', 172,
        '&', 197,
        '\'', 237,
        '(', 148,
        '*', 168,
        '+', 155,
        ',', 256,
        '-', 160,
        '.', 205,
        '/', 170,
        ';', 257,
        '<', 188,
        '=', 63,
        '>', 183,
        '?', 217,
        'L', 317,
        'U', 318,
        '[', 221,
        '\\', 219,
        '^', 177,
        '`', 135,
        'u', 319,
        '|', 176,
        '~', 161,
        '\t', 348,
        ' ', 348,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(77);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 11:
      ADVANCE_MAP(
        '\n', 346,
        '\r', 7,
        '!', 163,
        '"', 247,
        '%', 173,
        '&', 198,
        '\'', 237,
        '(', 148,
        '*', 169,
        '+', 156,
        ',', 256,
        '-', 159,
        '.', 205,
        '/', 171,
        ';', 257,
        '<', 189,
        '=', 153,
        '>', 184,
        '?', 217,
        'L', 317,
        'U', 318,
        '[', 221,
        '\\', 219,
        '^', 178,
        '`', 135,
        'u', 319,
        '|', 175,
        '~', 161,
        '\t', 348,
        ' ', 348,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(77);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 12:
      ADVANCE_MAP(
        '\n', 346,
        '\r', 7,
        '!', 163,
        '"', 247,
        '%', 173,
        '&', 198,
        '\'', 237,
        '(', 148,
        '*', 169,
        '+', 156,
        '-', 159,
        '.', 205,
        '/', 171,
        ';', 1,
        '<', 189,
        '=', 153,
        '>', 184,
        '?', 217,
        'L', 317,
        'U', 318,
        '[', 221,
        '\\', 219,
        '^', 178,
        '`', 135,
        'u', 319,
        '|', 175,
        '~', 161,
        '\t', 350,
        ' ', 350,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(77);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 13:
      ADVANCE_MAP(
        '\n', 346,
        '\r', 7,
        '!', 62,
        '"', 247,
        '%', 173,
        '&', 198,
        '\'', 237,
        '(', 148,
        '*', 169,
        '+', 156,
        ',', 256,
        '-', 159,
        '.', 205,
        '/', 171,
        '0', 102,
        ';', 257,
        '<', 189,
        '=', 153,
        '>', 184,
        '?', 217,
        'L', 36,
        'U', 37,
        '[', 221,
        '\\', 4,
        '^', 178,
        'u', 38,
        '|', 175,
        '\t', 348,
        ' ', 348,
      );
      END_STATE();
    case 14:
      ADVANCE_MAP(
        '\n', 346,
        '\r', 7,
        '!', 62,
        '%', 172,
        '&', 197,
        '(', 148,
        '*', 168,
        '+', 155,
        ',', 207,
        '-', 158,
        '.', 206,
        '/', 170,
        ':', 201,
        ';', 1,
        '<', 188,
        '=', 153,
        '>', 183,
        'B', 328,
        '\\', 4,
        '^', 179,
        '|', 176,
        '\t', 350,
        ' ', 350,
        'V', 325,
        'v', 325,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 15:
      ADVANCE_MAP(
        '\n', 346,
        '\r', 7,
        '!', 62,
        '%', 172,
        '&', 197,
        '(', 148,
        '*', 168,
        '+', 155,
        ',', 256,
        '-', 160,
        '.', 205,
        '/', 170,
        ';', 257,
        '<', 188,
        '=', 63,
        '>', 183,
        '?', 217,
        '[', 221,
        '\\', 4,
        '^', 177,
        '|', 176,
        '\t', 348,
        ' ', 348,
      );
      END_STATE();
    case 16:
      ADVANCE_MAP(
        '\n', 346,
        '\r', 7,
        '!', 62,
        '%', 173,
        '&', 198,
        '(', 148,
        ')', 147,
        '*', 169,
        '+', 156,
        '-', 159,
        '.', 205,
        '/', 171,
        ';', 1,
        '<', 189,
        '=', 153,
        '>', 184,
        '?', 217,
        '[', 221,
        '\\', 4,
        '^', 178,
        '|', 175,
        '\t', 350,
        ' ', 350,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 17:
      ADVANCE_MAP(
        '\n', 346,
        '\r', 7,
        '(', 148,
        '/', 50,
        ';', 1,
        '=', 214,
        '\\', 3,
        '\t', 349,
        ' ', 349,
      );
      END_STATE();
    case 18:
      ADVANCE_MAP(
        '\n', 346,
        '\r', 7,
        '(', 145,
        '*', 168,
        '/', 50,
        ';', 1,
        '?', 217,
        '\\', 3,
        '\t', 350,
        ' ', 350,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 19:
      if (lookahead == '\n') ADVANCE(346);
      if (lookahead == '\r') ADVANCE(7);
      if (lookahead == ',') ADVANCE(256);
      if (lookahead == '/') ADVANCE(50);
      if (lookahead == ';') ADVANCE(257);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(19);
      END_STATE();
    case 20:
      if (lookahead == '\n') ADVANCE(346);
      if (lookahead == '\r') ADVANCE(7);
      if (lookahead == '/') ADVANCE(50);
      if (lookahead == ';') ADVANCE(1);
      if (lookahead == '=') ADVANCE(214);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(20);
      END_STATE();
    case 21:
      if (lookahead == '\n') ADVANCE(346);
      if (lookahead == '\r') ADVANCE(7);
      if (lookahead == '/') ADVANCE(50);
      if (lookahead == ';') ADVANCE(1);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(21);
      END_STATE();
    case 22:
      if (lookahead == '\n') ADVANCE(346);
      if (lookahead == '\r') ADVANCE(7);
      if (lookahead == '/') ADVANCE(211);
      if (lookahead == ';') ADVANCE(209);
      if (lookahead == '\\') ADVANCE(210);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(350);
      if (lookahead != 0) ADVANCE(208);
      END_STATE();
    case 23:
      if (lookahead == '\n') ADVANCE(253);
      if (lookahead == '\r') ADVANCE(253);
      if (lookahead == 'U') ADVANCE(130);
      if (lookahead == 'u') ADVANCE(126);
      if (lookahead == 'x') ADVANCE(124);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (lookahead != 0) ADVANCE(253);
      END_STATE();
    case 24:
      ADVANCE_MAP(
        '!', 162,
        '"', 247,
        '%', 172,
        '&', 196,
        '\'', 237,
        '(', 148,
        ')', 147,
        '*', 168,
        '+', 155,
        '-', 158,
        '/', 170,
        ':', 199,
        ';', 1,
        'L', 317,
        'U', 318,
        '\\', 219,
        ']', 222,
        '`', 135,
        'u', 319,
        '~', 161,
        '\t', 352,
        ' ', 352,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(77);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 25:
      ADVANCE_MAP(
        '!', 163,
        '"', 247,
        '#', 116,
        '%', 172,
        '&', 196,
        '\'', 237,
        '(', 148,
        ')', 147,
        '*', 168,
        '+', 155,
        ',', 207,
        '-', 158,
        '.', 206,
        '/', 170,
        '0', 42,
        ':', 200,
        ';', 1,
        '<', 188,
        '=', 153,
        '>', 183,
        '[', 221,
        '\\', 219,
        ']', 222,
        '^', 179,
        '_', 330,
        '`', 135,
        '{', 149,
        '|', 176,
        '}', 150,
        '~', 161,
        '\t', 352,
        ' ', 352,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(43);
      if (('A' <= lookahead && lookahead <= 'z')) ADVANCE(329);
      END_STATE();
    case 26:
      ADVANCE_MAP(
        '!', 163,
        '"', 247,
        '%', 172,
        '&', 197,
        '\'', 237,
        '(', 148,
        ')', 147,
        '*', 168,
        '+', 155,
        ',', 207,
        '-', 160,
        '.', 206,
        '/', 170,
        ':', 199,
        ';', 1,
        '<', 188,
        '=', 63,
        '>', 183,
        '?', 217,
        'L', 317,
        'U', 318,
        '[', 221,
        '\\', 219,
        ']', 222,
        '^', 177,
        '`', 135,
        'u', 319,
        '|', 176,
        '~', 161,
        '\t', 352,
        ' ', 352,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(77);
      if (('A' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 27:
      ADVANCE_MAP(
        '!', 163,
        '"', 247,
        '%', 172,
        '&', 197,
        '\'', 237,
        '(', 148,
        '*', 168,
        '+', 155,
        ',', 256,
        '-', 160,
        '.', 205,
        '/', 170,
        ';', 257,
        '<', 188,
        '=', 63,
        '>', 183,
        '?', 217,
        'L', 317,
        'U', 318,
        '[', 221,
        '\\', 219,
        '^', 177,
        '`', 135,
        'u', 319,
        '|', 176,
        '~', 161,
        '\t', 351,
        ' ', 351,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(77);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 28:
      ADVANCE_MAP(
        '!', 163,
        '"', 247,
        '%', 173,
        '&', 198,
        '\'', 237,
        '(', 148,
        ')', 147,
        '*', 169,
        '+', 156,
        ',', 207,
        '-', 159,
        '.', 206,
        '/', 171,
        ':', 199,
        ';', 1,
        '<', 189,
        '=', 153,
        '>', 184,
        '?', 217,
        'L', 317,
        'U', 318,
        '[', 221,
        '\\', 219,
        ']', 222,
        '^', 178,
        '`', 135,
        'u', 319,
        '|', 175,
        '~', 161,
        '\t', 352,
        ' ', 352,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(77);
      if (('A' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 29:
      ADVANCE_MAP(
        '!', 163,
        '"', 247,
        '%', 173,
        '&', 198,
        '\'', 237,
        '(', 148,
        '*', 169,
        '+', 156,
        ',', 256,
        '-', 159,
        '.', 205,
        '/', 171,
        ';', 257,
        '<', 189,
        '=', 153,
        '>', 184,
        '?', 217,
        'L', 317,
        'U', 318,
        '[', 221,
        '\\', 219,
        '^', 178,
        '`', 135,
        'u', 319,
        '|', 175,
        '~', 161,
        '\t', 351,
        ' ', 351,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(77);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 30:
      ADVANCE_MAP(
        '!', 62,
        '"', 247,
        '%', 173,
        '&', 198,
        '(', 148,
        ')', 147,
        '*', 169,
        '+', 156,
        ',', 207,
        '-', 159,
        '.', 206,
        '/', 171,
        ':', 199,
        ';', 1,
        '<', 189,
        '=', 153,
        '>', 184,
        '?', 217,
        '[', 221,
        '\\', 4,
        ']', 222,
        '^', 178,
        '{', 149,
        '|', 175,
        '\t', 352,
        ' ', 352,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 31:
      ADVANCE_MAP(
        '!', 62,
        '%', 172,
        '&', 197,
        '(', 148,
        ')', 147,
        '*', 168,
        '+', 155,
        ',', 207,
        '-', 160,
        '.', 206,
        '/', 170,
        ':', 199,
        ';', 1,
        '<', 188,
        '=', 153,
        '>', 183,
        '?', 217,
        '[', 221,
        '\\', 4,
        ']', 222,
        '^', 179,
        '|', 176,
        '}', 150,
        '\t', 352,
        ' ', 352,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= '_') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 32:
      ADVANCE_MAP(
        '!', 62,
        '%', 172,
        '&', 197,
        '(', 148,
        '*', 168,
        '+', 155,
        ',', 256,
        '-', 160,
        '.', 205,
        '/', 170,
        ';', 257,
        '<', 188,
        '=', 63,
        '>', 183,
        '?', 217,
        '[', 221,
        '\\', 4,
        '^', 177,
        '|', 176,
        '\t', 351,
        ' ', 351,
      );
      END_STATE();
    case 33:
      ADVANCE_MAP(
        '!', 62,
        '%', 173,
        '&', 198,
        '(', 148,
        '*', 169,
        '+', 156,
        ',', 256,
        '-', 159,
        '.', 205,
        '/', 171,
        ';', 257,
        '<', 189,
        '=', 153,
        '>', 184,
        '?', 217,
        '[', 221,
        '\\', 4,
        '^', 178,
        '|', 175,
        '\t', 351,
        ' ', 351,
        'L', 215,
        'l', 215,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 34:
      if (lookahead == '"') ADVANCE(247);
      if (lookahead == '&') ADVANCE(196);
      if (lookahead == '/') ADVANCE(211);
      if (lookahead == ';') ADVANCE(209);
      if (lookahead == '\\') ADVANCE(210);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(352);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(208);
      END_STATE();
    case 35:
      if (lookahead == '"') ADVANCE(247);
      if (lookahead == '/') ADVANCE(251);
      if (lookahead == ';') ADVANCE(249);
      if (lookahead == '\\') ADVANCE(23);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(250);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n') ADVANCE(252);
      END_STATE();
    case 36:
      if (lookahead == '"') ADVANCE(243);
      if (lookahead == '\'') ADVANCE(233);
      END_STATE();
    case 37:
      if (lookahead == '"') ADVANCE(245);
      if (lookahead == '\'') ADVANCE(235);
      END_STATE();
    case 38:
      if (lookahead == '"') ADVANCE(244);
      if (lookahead == '\'') ADVANCE(234);
      if (lookahead == '8') ADVANCE(39);
      END_STATE();
    case 39:
      if (lookahead == '"') ADVANCE(246);
      if (lookahead == '\'') ADVANCE(236);
      END_STATE();
    case 40:
      if (lookahead == '"') ADVANCE(248);
      if (lookahead == '&') ADVANCE(196);
      if (lookahead == '/') ADVANCE(278);
      if (lookahead == ';') ADVANCE(275);
      if (lookahead == '\\') ADVANCE(274);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(277);
      if (lookahead != 0) ADVANCE(279);
      END_STATE();
    case 41:
      ADVANCE_MAP(
        '%', 172,
        '(', 148,
        ')', 147,
        '*', 168,
        ',', 207,
        '/', 50,
        ';', 1,
        '?', 217,
        '[', 221,
        '\\', 218,
        '\t', 352,
        ' ', 352,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 42:
      ADVANCE_MAP(
        '%', 315,
        '.', 44,
        'X', 337,
        '\\', 117,
        'm', 331,
        's', 312,
        'x', 334,
        'y', 335,
        'H', 336,
        'h', 336,
        'n', 332,
        'u', 332,
        'G', 333,
        'K', 333,
        'M', 333,
        'g', 333,
        'k', 333,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(43);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 43:
      ADVANCE_MAP(
        '%', 315,
        '.', 44,
        '\\', 117,
        'm', 331,
        's', 312,
        'H', 336,
        'h', 336,
        'n', 332,
        'u', 332,
        'G', 333,
        'K', 333,
        'M', 333,
        'g', 333,
        'k', 333,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(43);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 44:
      ADVANCE_MAP(
        '%', 315,
        'm', 100,
        's', 311,
        'H', 114,
        'h', 114,
        'n', 101,
        'u', 101,
        'G', 107,
        'K', 107,
        'M', 107,
        'g', 107,
        'k', 107,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      END_STATE();
    case 45:
      ADVANCE_MAP(
        '%', 173,
        '&', 64,
        '(', 145,
        ')', 147,
        '*', 65,
        '+', 66,
        ',', 207,
        '-', 67,
        '.', 49,
        '/', 171,
        ';', 1,
        '<', 61,
        '=', 152,
        '>', 72,
        '\\', 3,
        ']', 222,
        '^', 68,
        '|', 69,
        '\t', 352,
        ' ', 352,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 46:
      if (lookahead == '%') ADVANCE(111);
      if (lookahead == '&') ADVANCE(196);
      if (lookahead == '/') ADVANCE(170);
      if (lookahead == ';') ADVANCE(1);
      if (lookahead == '\\') ADVANCE(3);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(352);
      END_STATE();
    case 47:
      ADVANCE_MAP(
        '&', 197,
        '(', 145,
        ')', 147,
        '/', 50,
        ':', 54,
        ';', 1,
        '=', 152,
        'B', 328,
        '\\', 4,
        '\t', 352,
        ' ', 352,
        'V', 325,
        'v', 325,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 48:
      if (lookahead == ',') ADVANCE(256);
      if (lookahead == '/') ADVANCE(50);
      if (lookahead == ';') ADVANCE(257);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(48);
      END_STATE();
    case 49:
      if (lookahead == '.') ADVANCE(194);
      END_STATE();
    case 50:
      if (lookahead == '/') ADVANCE(1);
      END_STATE();
    case 51:
      if (lookahead == '/') ADVANCE(50);
      if (lookahead == ';') ADVANCE(1);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(51);
      END_STATE();
    case 52:
      if (lookahead == '/') ADVANCE(286);
      if (lookahead == ';') ADVANCE(284);
      if (lookahead == '\\') ADVANCE(285);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(287);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n') ADVANCE(283);
      END_STATE();
    case 53:
      if (lookahead == '/') ADVANCE(241);
      if (lookahead == ';') ADVANCE(239);
      if (lookahead == '\\') ADVANCE(240);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(242);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '\'') ADVANCE(238);
      END_STATE();
    case 54:
      if (lookahead == ':') ADVANCE(73);
      END_STATE();
    case 55:
      if (lookahead == ':') ADVANCE(202);
      END_STATE();
    case 56:
      if (lookahead == ':') ADVANCE(131);
      END_STATE();
    case 57:
      if (lookahead == ':') ADVANCE(134);
      END_STATE();
    case 58:
      if (lookahead == ':') ADVANCE(55);
      END_STATE();
    case 59:
      if (lookahead == ':') ADVANCE(133);
      END_STATE();
    case 60:
      if (lookahead == ':') ADVANCE(59);
      if (lookahead == '\\') ADVANCE(98);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(60);
      END_STATE();
    case 61:
      if (lookahead == '<') ADVANCE(70);
      END_STATE();
    case 62:
      if (lookahead == '=') ADVANCE(181);
      END_STATE();
    case 63:
      if (lookahead == '=') ADVANCE(180);
      END_STATE();
    case 64:
      if (lookahead == '=') ADVANCE(230);
      END_STATE();
    case 65:
      if (lookahead == '=') ADVANCE(223);
      END_STATE();
    case 66:
      if (lookahead == '=') ADVANCE(226);
      END_STATE();
    case 67:
      if (lookahead == '=') ADVANCE(227);
      END_STATE();
    case 68:
      if (lookahead == '=') ADVANCE(231);
      END_STATE();
    case 69:
      if (lookahead == '=') ADVANCE(232);
      END_STATE();
    case 70:
      if (lookahead == '=') ADVANCE(228);
      END_STATE();
    case 71:
      if (lookahead == '=') ADVANCE(229);
      END_STATE();
    case 72:
      if (lookahead == '>') ADVANCE(71);
      END_STATE();
    case 73:
      if (lookahead == 'B') ADVANCE(58);
      END_STATE();
    case 74:
      if (lookahead == '\\') ADVANCE(99);
      if (lookahead == '`') ADVANCE(139);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      END_STATE();
    case 75:
      if (lookahead == '\\') ADVANCE(305);
      END_STATE();
    case 76:
      if (lookahead == '\\') ADVANCE(305);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(76);
      END_STATE();
    case 77:
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(77);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 78:
      if (lookahead == '\\') ADVANCE(90);
      END_STATE();
    case 79:
      if (lookahead == '\\') ADVANCE(90);
      if (lookahead == '`') ADVANCE(137);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(80);
      END_STATE();
    case 80:
      if (lookahead == '\\') ADVANCE(90);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(80);
      END_STATE();
    case 81:
      if (lookahead == '\\') ADVANCE(79);
      if (lookahead == '`') ADVANCE(136);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(76);
      END_STATE();
    case 82:
      if (lookahead == '\\') ADVANCE(98);
      END_STATE();
    case 83:
      if (lookahead == '\\') ADVANCE(98);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(83);
      END_STATE();
    case 84:
      if (lookahead == '\\') ADVANCE(119);
      if (lookahead == '`') ADVANCE(135);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(291);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(290);
      END_STATE();
    case 85:
      if (lookahead == '\\') ADVANCE(119);
      if (lookahead == '`') ADVANCE(135);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(301);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(295);
      END_STATE();
    case 86:
      if (lookahead == '\\') ADVANCE(119);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(302);
      END_STATE();
    case 87:
      if (lookahead == '\\') ADVANCE(120);
      if (lookahead == '`') ADVANCE(138);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(308);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(83);
      END_STATE();
    case 88:
      if (lookahead == '`') ADVANCE(135);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(291);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(290);
      END_STATE();
    case 89:
      if (lookahead == '`') ADVANCE(135);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(301);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(295);
      END_STATE();
    case 90:
      if (lookahead == '`') ADVANCE(135);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(290);
      END_STATE();
    case 91:
      if (lookahead == '`') ADVANCE(294);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(91);
      END_STATE();
    case 92:
      if (lookahead == '`') ADVANCE(75);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(92);
      END_STATE();
    case 93:
      if (lookahead == '`') ADVANCE(293);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(93);
      END_STATE();
    case 94:
      if (lookahead == '`') ADVANCE(78);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(94);
      END_STATE();
    case 95:
      if (lookahead == '`') ADVANCE(296);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(95);
      END_STATE();
    case 96:
      if (lookahead == '`') ADVANCE(82);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(96);
      END_STATE();
    case 97:
      if (lookahead == '`') ADVANCE(299);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(97);
      END_STATE();
    case 98:
      if (lookahead == '`') ADVANCE(138);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(308);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(83);
      END_STATE();
    case 99:
      if (lookahead == '`') ADVANCE(140);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(297);
      END_STATE();
    case 100:
      ADVANCE_MAP(
        'm', 100,
        's', 311,
        'H', 114,
        'h', 114,
        'n', 101,
        'u', 101,
        'G', 107,
        'K', 107,
        'M', 107,
        'g', 107,
        'k', 107,
      );
      END_STATE();
    case 101:
      if (lookahead == 's') ADVANCE(311);
      if (lookahead == 'm' ||
          lookahead == 'n' ||
          lookahead == 'u') ADVANCE(101);
      END_STATE();
    case 102:
      if (lookahead == 'y') ADVANCE(121);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(122);
      END_STATE();
    case 103:
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(118);
      END_STATE();
    case 104:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(151);
      END_STATE();
    case 105:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(204);
      END_STATE();
    case 106:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(108);
      END_STATE();
    case 107:
      ADVANCE_MAP(
        'H', 114,
        'h', 114,
        'G', 107,
        'K', 107,
        'M', 107,
        'g', 107,
        'k', 107,
        'm', 107,
      );
      END_STATE();
    case 108:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(110);
      END_STATE();
    case 109:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(112);
      END_STATE();
    case 110:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(104);
      END_STATE();
    case 111:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(109);
      END_STATE();
    case 112:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(105);
      END_STATE();
    case 113:
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(106);
      END_STATE();
    case 114:
      if (lookahead == 'Z' ||
          lookahead == 'z') ADVANCE(313);
      END_STATE();
    case 115:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 116:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(316);
      END_STATE();
    case 117:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(292);
      END_STATE();
    case 118:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(269);
      END_STATE();
    case 119:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(303);
      END_STATE();
    case 120:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(309);
      END_STATE();
    case 121:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(264);
      END_STATE();
    case 122:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(267);
      END_STATE();
    case 123:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(253);
      END_STATE();
    case 124:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(123);
      END_STATE();
    case 125:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(124);
      END_STATE();
    case 126:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(125);
      END_STATE();
    case 127:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(126);
      END_STATE();
    case 128:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(127);
      END_STATE();
    case 129:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(128);
      END_STATE();
    case 130:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(129);
      END_STATE();
    case 131:
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(288);
      END_STATE();
    case 132:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(310);
      END_STATE();
    case 133:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(60);
      END_STATE();
    case 134:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(290);
      END_STATE();
    case 135:
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '`') ADVANCE(91);
      END_STATE();
    case 136:
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '`') ADVANCE(92);
      END_STATE();
    case 137:
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '`') ADVANCE(94);
      END_STATE();
    case 138:
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '`') ADVANCE(96);
      END_STATE();
    case 139:
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '`') ADVANCE(93);
      END_STATE();
    case 140:
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '`') ADVANCE(95);
      END_STATE();
    case 141:
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '`') ADVANCE(97);
      END_STATE();
    case 142:
      if (eof) ADVANCE(143);
      ADVANCE_MAP(
        '\n', 346,
        '\r', 7,
        '"', 247,
        '%', 172,
        '&', 197,
        '\'', 237,
        '(', 145,
        ')', 147,
        ',', 207,
        '.', 205,
        '/', 170,
        ':', 54,
        ';', 1,
        '=', 152,
        'B', 328,
        '\\', 3,
        ']', 222,
        '}', 150,
        '\t', 350,
        ' ', 350,
        'V', 325,
        'v', 325,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(aux_sym_block_token1);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(aux_sym_block_token1);
      if (lookahead == '&') ADVANCE(146);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(aux_sym_block_token1);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(144);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 148:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 149:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(aux_sym_while_block_token2);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(180);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(164);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(164);
      if (lookahead == '=') ADVANCE(226);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(165);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(165);
      if (lookahead == '=') ADVANCE(227);
      if (lookahead == '>') ADVANCE(220);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(165);
      if (lookahead == '>') ADVANCE(220);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(181);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(anon_sym_CARET_CARET);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '=') ADVANCE(223);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '/') ADVANCE(1);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '/') ADVANCE(1);
      if (lookahead == '=') ADVANCE(224);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(225);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(232);
      if (lookahead == '|') ADVANCE(167);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(167);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(231);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '^') ADVANCE(166);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(185);
      if (lookahead == '>') ADVANCE(192);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(185);
      if (lookahead == '>') ADVANCE(193);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(190);
      if (lookahead == '=') ADVANCE(186);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(191);
      if (lookahead == '=') ADVANCE(186);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '=') ADVANCE(228);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '=') ADVANCE(229);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(195);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(195);
      if (lookahead == '=') ADVANCE(230);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(259);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(260);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(aux_sym_macro_definition_token1);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(aux_sym_macro_definition_token1);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(288);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(aux_sym_parameter_declaration_token4);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(194);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(aux_sym__host_command_argument_list_default_format_token1);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(aux_sym__host_command_argument_list_default_format_token1);
      if (lookahead == '\n') ADVANCE(344);
      if (lookahead == '\\') ADVANCE(5);
      if (lookahead != 0) ADVANCE(1);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(aux_sym__host_command_argument_list_default_format_token1);
      if (lookahead == '\n') ADVANCE(347);
      if (lookahead == '\r') ADVANCE(2);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(aux_sym__host_command_argument_list_default_format_token1);
      if (lookahead == '/') ADVANCE(1);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(aux_sym__host_command_argument_list_default_format_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(115);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(aux_sym__host_command_argument_list_default_format_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(aux_sym__hll_var_definition_command_token1);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(214);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(aux_sym__var_command_with_non_hll_format_token1);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(340);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(aux_sym__var_command_with_non_hll_format_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      if (lookahead == '\n') ADVANCE(347);
      if (lookahead == '\r') ADVANCE(2);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_BSLASH);
      if (lookahead == '\n') ADVANCE(347);
      if (lookahead == '\r') ADVANCE(2);
      if (lookahead == '\\') ADVANCE(74);
      if (lookahead == '`') ADVANCE(141);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(80);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(anon_sym_LT_LT_EQ);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(anon_sym_GT_GT_EQ);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(anon_sym_AMP_EQ);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(anon_sym_L_SQUOTE);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(anon_sym_u_SQUOTE);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(anon_sym_U_SQUOTE);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(anon_sym_u8_SQUOTE);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(aux_sym_hll_char_literal_token1);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(aux_sym_hll_char_literal_token1);
      if (lookahead == '\n') ADVANCE(344);
      if (lookahead == '\\') ADVANCE(5);
      if (lookahead != 0) ADVANCE(1);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(aux_sym_hll_char_literal_token1);
      if (lookahead == '\n') ADVANCE(253);
      if (lookahead == '\r') ADVANCE(253);
      if (lookahead == 'U') ADVANCE(130);
      if (lookahead == 'u') ADVANCE(126);
      if (lookahead == 'x') ADVANCE(124);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(255);
      if (lookahead != 0) ADVANCE(253);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(aux_sym_hll_char_literal_token1);
      if (lookahead == '/') ADVANCE(1);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(aux_sym_hll_char_literal_token1);
      if (lookahead == '/') ADVANCE(50);
      if (lookahead == ';') ADVANCE(1);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(51);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(anon_sym_L_DQUOTE);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(anon_sym_u_DQUOTE);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(anon_sym_U_DQUOTE);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(anon_sym_u8_DQUOTE);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      if (lookahead == '"') ADVANCE(280);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(aux_sym_hll_string_literal_token1);
      if (lookahead == '\n') ADVANCE(344);
      if (lookahead == '"') ADVANCE(1);
      if (lookahead == '\\') ADVANCE(5);
      if (lookahead != 0) ADVANCE(249);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(aux_sym_hll_string_literal_token1);
      if (lookahead == '/') ADVANCE(251);
      if (lookahead == ';') ADVANCE(249);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(250);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(252);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(aux_sym_hll_string_literal_token1);
      if (lookahead == '/') ADVANCE(249);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(252);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(aux_sym_hll_string_literal_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(252);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(sym_hll_escape_sequence);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(sym_hll_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(253);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(sym_hll_escape_sequence);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(254);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(aux_sym__hll_expression_separator_token1);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(256);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(aux_sym__hll_expression_separator_token2);
      if (lookahead == '\n') ADVANCE(344);
      if (lookahead == '\\') ADVANCE(5);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(257);
      if (lookahead != 0) ADVANCE(1);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(anon_sym_COLON_COLON_COLON);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      if (lookahead == ':') ADVANCE(258);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      if (lookahead == ':') ADVANCE(258);
      if (lookahead == 'B') ADVANCE(58);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(sym_access_class);
      if (lookahead == ':') ADVANCE(203);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(sym_access_class);
      if (lookahead == ':') ADVANCE(131);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(aux_sym_integer_token1);
      if (lookahead == '\\') ADVANCE(117);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(270);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(aux_sym_integer_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(264);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(aux_sym_integer_token2);
      if (lookahead == '\\') ADVANCE(117);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(271);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(265);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(aux_sym_integer_token2);
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(266);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(aux_sym_integer_token2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(267);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(sym_float);
      ADVANCE_MAP(
        '%', 315,
        'e', 103,
        'm', 100,
        's', 311,
        'H', 114,
        'h', 114,
        'n', 101,
        'u', 101,
        'G', 107,
        'K', 107,
        'M', 107,
        'g', 107,
        'k', 107,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      END_STATE();
    case 269:
      ACCEPT_TOKEN(sym_float);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(269);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(aux_sym_bitmask_token1);
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == 'X' ||
          lookahead == 'x') ADVANCE(270);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(aux_sym_bitmask_token2);
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == 'X' ||
          ('a' <= lookahead && lookahead <= 'f') ||
          lookahead == 'x') ADVANCE(271);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(aux_sym_string_token1);
      if (lookahead == '\n') ADVANCE(275);
      if (lookahead == '\r') ADVANCE(273);
      if (lookahead == '\\') ADVANCE(272);
      if (lookahead == '"' ||
          lookahead == '&') ADVANCE(1);
      if (lookahead != 0) ADVANCE(275);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(aux_sym_string_token1);
      if (lookahead == '\n') ADVANCE(275);
      if (lookahead == '\\') ADVANCE(272);
      if (lookahead == '"' ||
          lookahead == '&') ADVANCE(1);
      if (lookahead != 0) ADVANCE(275);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(aux_sym_string_token1);
      if (lookahead == '\n') ADVANCE(279);
      if (lookahead == '\r') ADVANCE(276);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '&') ADVANCE(279);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(aux_sym_string_token1);
      if (lookahead == '\n') ADVANCE(279);
      if (lookahead == '\\') ADVANCE(272);
      if (lookahead == '"' ||
          lookahead == '&') ADVANCE(1);
      if (lookahead != 0) ADVANCE(275);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(aux_sym_string_token1);
      if (lookahead == '\n') ADVANCE(279);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '&') ADVANCE(279);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(aux_sym_string_token1);
      if (lookahead == '/') ADVANCE(278);
      if (lookahead == ';') ADVANCE(275);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(277);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '&') ADVANCE(279);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(aux_sym_string_token1);
      if (lookahead == '/') ADVANCE(275);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '&') ADVANCE(279);
      END_STATE();
    case 279:
      ACCEPT_TOKEN(aux_sym_string_token1);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '&') ADVANCE(279);
      END_STATE();
    case 280:
      ACCEPT_TOKEN(aux_sym_string_token2);
      END_STATE();
    case 281:
      ACCEPT_TOKEN(aux_sym_character_token1);
      END_STATE();
    case 282:
      ACCEPT_TOKEN(aux_sym_character_token1);
      if (lookahead == '\n') ADVANCE(347);
      END_STATE();
    case 283:
      ACCEPT_TOKEN(aux_sym_character_token2);
      END_STATE();
    case 284:
      ACCEPT_TOKEN(aux_sym_character_token2);
      if (lookahead == '\n') ADVANCE(344);
      if (lookahead == '\\') ADVANCE(5);
      if (lookahead != 0) ADVANCE(1);
      END_STATE();
    case 285:
      ACCEPT_TOKEN(aux_sym_character_token2);
      if (lookahead == '\n') ADVANCE(347);
      if (lookahead == '\r') ADVANCE(282);
      if (lookahead != 0 &&
          lookahead != '\'') ADVANCE(281);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(aux_sym_character_token2);
      if (lookahead == '/') ADVANCE(1);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(aux_sym_character_token2);
      if (lookahead == '/') ADVANCE(50);
      if (lookahead == ';') ADVANCE(1);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(51);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == ':') ADVANCE(56);
      if (lookahead == '\\') ADVANCE(89);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(288);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == ':') ADVANCE(57);
      if (lookahead == '\\') ADVANCE(306);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(289);
      END_STATE();
    case 290:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == ':') ADVANCE(57);
      if (lookahead == '\\') ADVANCE(89);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(290);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == ':') ADVANCE(57);
      if (lookahead == '\\') ADVANCE(85);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(291);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(290);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == '\\') ADVANCE(86);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(292);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == '\\') ADVANCE(306);
      END_STATE();
    case 294:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == '\\') ADVANCE(89);
      END_STATE();
    case 295:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == '\\') ADVANCE(89);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(295);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == '\\') ADVANCE(304);
      END_STATE();
    case 297:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == '\\') ADVANCE(304);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(297);
      END_STATE();
    case 298:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == '\\') ADVANCE(84);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(298);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(300);
      END_STATE();
    case 299:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == '\\') ADVANCE(88);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == '\\') ADVANCE(88);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(300);
      END_STATE();
    case 301:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == '\\') ADVANCE(85);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(301);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(295);
      END_STATE();
    case 302:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (lookahead == '\\') ADVANCE(119);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(302);
      END_STATE();
    case 303:
      ACCEPT_TOKEN(aux_sym_symbol_token1);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(303);
      END_STATE();
    case 304:
      ACCEPT_TOKEN(aux_sym_symbol_token2);
      if (lookahead == '\\') ADVANCE(81);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(292);
      END_STATE();
    case 305:
      ACCEPT_TOKEN(aux_sym_symbol_token2);
      if (lookahead == '\\') ADVANCE(90);
      if (lookahead == '`') ADVANCE(137);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(80);
      END_STATE();
    case 306:
      ACCEPT_TOKEN(aux_sym_symbol_token2);
      if (lookahead == '\\') ADVANCE(90);
      if (lookahead == '`') ADVANCE(141);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(298);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(300);
      END_STATE();
    case 307:
      ACCEPT_TOKEN(aux_sym_symbol_token3);
      if (lookahead == ':') ADVANCE(59);
      if (lookahead == '\\') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(307);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(60);
      END_STATE();
    case 308:
      ACCEPT_TOKEN(aux_sym_symbol_token3);
      if (lookahead == '\\') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(308);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(83);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(aux_sym_symbol_token3);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(309);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(sym__composed_name);
      if (lookahead == '-') ADVANCE(132);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(310);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(sym_time);
      END_STATE();
    case 312:
      ACCEPT_TOKEN(sym_time);
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 313:
      ACCEPT_TOKEN(sym_frequency);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(sym_frequency);
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 315:
      ACCEPT_TOKEN(sym_percentage);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(sym_file_handle);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(316);
      END_STATE();
    case 317:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(243);
      if (lookahead == '\'') ADVANCE(233);
      if (lookahead == ':') ADVANCE(56);
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 318:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(245);
      if (lookahead == '\'') ADVANCE(235);
      if (lookahead == ':') ADVANCE(56);
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 319:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(244);
      if (lookahead == '\'') ADVANCE(234);
      if (lookahead == '8') ADVANCE(320);
      if (lookahead == ':') ADVANCE(56);
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 320:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(246);
      if (lookahead == '\'') ADVANCE(236);
      if (lookahead == ':') ADVANCE(56);
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '-') ADVANCE(132);
      if (lookahead == '.') ADVANCE(113);
      if (lookahead == ':') ADVANCE(262);
      if (lookahead == '\\') ADVANCE(117);
      if (lookahead == '_') ADVANCE(330);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '-') ADVANCE(132);
      if (lookahead == ':') ADVANCE(262);
      if (lookahead == '\\') ADVANCE(117);
      if (lookahead == '_') ADVANCE(330);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(321);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '-') ADVANCE(132);
      if (lookahead == ':') ADVANCE(262);
      if (lookahead == '\\') ADVANCE(117);
      if (lookahead == '_') ADVANCE(330);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      END_STATE();
    case 324:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '.') ADVANCE(113);
      if (lookahead == ':') ADVANCE(262);
      if (lookahead == '\\') ADVANCE(117);
      if (lookahead == '_') ADVANCE(330);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(322);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      END_STATE();
    case 325:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '.') ADVANCE(113);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(342);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '.') ADVANCE(113);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 327:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == ':') ADVANCE(261);
      if (lookahead == '\\') ADVANCE(117);
      if (lookahead == '_') ADVANCE(330);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == ':') ADVANCE(55);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 329:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == ':') ADVANCE(262);
      if (lookahead == '\\') ADVANCE(117);
      if (lookahead == '_') ADVANCE(330);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(323);
      END_STATE();
    case 330:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == ':') ADVANCE(56);
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(330);
      END_STATE();
    case 331:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '\\', 117,
        'm', 331,
        's', 312,
        'H', 336,
        'h', 336,
        'n', 332,
        'u', 332,
        'G', 333,
        'K', 333,
        'M', 333,
        'g', 333,
        'k', 333,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 332:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(117);
      if (lookahead == 's') ADVANCE(312);
      if (lookahead == 'm' ||
          lookahead == 'n' ||
          lookahead == 'u') ADVANCE(332);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '\\', 117,
        'H', 336,
        'h', 336,
        'G', 333,
        'K', 333,
        'M', 333,
        'g', 333,
        'k', 333,
        'm', 333,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(117);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(271);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(265);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 335:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(117);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(270);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(263);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 336:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(117);
      if (lookahead == 'Z' ||
          lookahead == 'z') ADVANCE(314);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Y') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'y')) ADVANCE(338);
      END_STATE();
    case 337:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(266);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 338:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(338);
      END_STATE();
    case 339:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(341);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 340:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(339);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(216);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 342:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(326);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 343:
      ACCEPT_TOKEN(sym_identifier);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(343);
      END_STATE();
    case 344:
      ACCEPT_TOKEN(sym_comment);
      END_STATE();
    case 345:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\n') ADVANCE(344);
      if (lookahead == '\\') ADVANCE(5);
      if (lookahead != 0) ADVANCE(1);
      END_STATE();
    case 346:
      ACCEPT_TOKEN(sym__blank_line);
      END_STATE();
    case 347:
      ACCEPT_TOKEN(sym__line_continuation);
      END_STATE();
    case 348:
      ACCEPT_TOKEN(sym__blank);
      if (lookahead == '\n') ADVANCE(346);
      if (lookahead == '\r') ADVANCE(7);
      if (lookahead == ',') ADVANCE(256);
      if (lookahead == '/') ADVANCE(50);
      if (lookahead == ';') ADVANCE(257);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(19);
      END_STATE();
    case 349:
      ACCEPT_TOKEN(sym__blank);
      if (lookahead == '\n') ADVANCE(346);
      if (lookahead == '\r') ADVANCE(7);
      if (lookahead == '/') ADVANCE(50);
      if (lookahead == ';') ADVANCE(1);
      if (lookahead == '=') ADVANCE(214);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(20);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(sym__blank);
      if (lookahead == '\n') ADVANCE(346);
      if (lookahead == '\r') ADVANCE(7);
      if (lookahead == '/') ADVANCE(50);
      if (lookahead == ';') ADVANCE(1);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(21);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(sym__blank);
      if (lookahead == ',') ADVANCE(256);
      if (lookahead == '/') ADVANCE(50);
      if (lookahead == ';') ADVANCE(257);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(48);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(sym__blank);
      if (lookahead == '/') ADVANCE(50);
      if (lookahead == ';') ADVANCE(1);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(51);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      ADVANCE_MAP(
        'E', 1,
        'I', 2,
        'L', 3,
        'S', 4,
        'V', 5,
        'b', 6,
        'c', 7,
        'd', 8,
        'e', 9,
        'f', 10,
        'i', 11,
        'l', 12,
        's', 13,
        'u', 14,
        'v', 15,
        'A', 16,
        'a', 16,
        'G', 17,
        'g', 17,
        'N', 18,
        'n', 18,
        'O', 19,
        'o', 19,
        'P', 20,
        'p', 20,
        'R', 21,
        'r', 21,
        'T', 22,
        't', 22,
        'W', 23,
        'w', 23,
      );
      END_STATE();
    case 1:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(24);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(25);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(26);
      END_STATE();
    case 2:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(27);
      END_STATE();
    case 3:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(28);
      END_STATE();
    case 4:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(29);
      END_STATE();
    case 5:
      ACCEPT_TOKEN(aux_sym__hll_non_default_format_var_command_token1);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(30);
      END_STATE();
    case 6:
      if (lookahead == 'o') ADVANCE(31);
      END_STATE();
    case 7:
      if (lookahead == 'h') ADVANCE(32);
      if (lookahead == 'o') ADVANCE(33);
      END_STATE();
    case 8:
      if (lookahead == 'o') ADVANCE(34);
      END_STATE();
    case 9:
      if (lookahead == 'N') ADVANCE(25);
      if (lookahead == 'n') ADVANCE(35);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(24);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(26);
      END_STATE();
    case 10:
      if (lookahead == 'l') ADVANCE(36);
      END_STATE();
    case 11:
      if (lookahead == 'n') ADVANCE(37);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(27);
      END_STATE();
    case 12:
      if (lookahead == 'O') ADVANCE(28);
      if (lookahead == 'o') ADVANCE(38);
      END_STATE();
    case 13:
      if (lookahead == 'h') ADVANCE(39);
      if (lookahead == 'i') ADVANCE(40);
      if (lookahead == 's') ADVANCE(41);
      if (lookahead == 't') ADVANCE(42);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(29);
      END_STATE();
    case 14:
      if (lookahead == 'i') ADVANCE(43);
      if (lookahead == 'n') ADVANCE(44);
      END_STATE();
    case 15:
      ACCEPT_TOKEN(aux_sym__hll_non_default_format_var_command_token1);
      if (lookahead == 'o') ADVANCE(45);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(30);
      END_STATE();
    case 16:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(46);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(47);
      END_STATE();
    case 17:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(48);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(49);
      END_STATE();
    case 18:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(50);
      END_STATE();
    case 19:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(51);
      END_STATE();
    case 20:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(52);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(53);
      END_STATE();
    case 21:
      ACCEPT_TOKEN(aux_sym__var_command_with_non_hll_format_token2);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(54);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(55);
      END_STATE();
    case 22:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(56);
      END_STATE();
    case 23:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(57);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(58);
      END_STATE();
    case 24:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(59);
      END_STATE();
    case 25:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(60);
      END_STATE();
    case 26:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(61);
      END_STATE();
    case 27:
      ACCEPT_TOKEN(aux_sym_if_block_token1);
      END_STATE();
    case 28:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(62);
      END_STATE();
    case 29:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(63);
      END_STATE();
    case 30:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(64);
      END_STATE();
    case 31:
      if (lookahead == 'o') ADVANCE(65);
      END_STATE();
    case 32:
      if (lookahead == 'a') ADVANCE(66);
      END_STATE();
    case 33:
      if (lookahead == 'n') ADVANCE(67);
      END_STATE();
    case 34:
      if (lookahead == 'u') ADVANCE(68);
      END_STATE();
    case 35:
      if (lookahead == 'u') ADVANCE(69);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(60);
      END_STATE();
    case 36:
      if (lookahead == 'o') ADVANCE(70);
      END_STATE();
    case 37:
      if (lookahead == 't') ADVANCE(71);
      END_STATE();
    case 38:
      if (lookahead == 'n') ADVANCE(72);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(62);
      END_STATE();
    case 39:
      if (lookahead == 'o') ADVANCE(73);
      END_STATE();
    case 40:
      if (lookahead == 'g') ADVANCE(74);
      if (lookahead == 'z') ADVANCE(75);
      END_STATE();
    case 41:
      if (lookahead == 'i') ADVANCE(76);
      END_STATE();
    case 42:
      if (lookahead == 'r') ADVANCE(77);
      END_STATE();
    case 43:
      if (lookahead == 'n') ADVANCE(78);
      END_STATE();
    case 44:
      if (lookahead == 'i') ADVANCE(79);
      if (lookahead == 's') ADVANCE(80);
      END_STATE();
    case 45:
      if (lookahead == 'i') ADVANCE(81);
      if (lookahead == 'l') ADVANCE(82);
      END_STATE();
    case 46:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(83);
      END_STATE();
    case 47:
      ACCEPT_TOKEN(aux_sym__hll_non_default_format_var_command_token2);
      END_STATE();
    case 48:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(84);
      END_STATE();
    case 49:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(85);
      END_STATE();
    case 50:
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(86);
      END_STATE();
    case 51:
      ACCEPT_TOKEN(aux_sym__host_command_with_default_format_token1);
      END_STATE();
    case 52:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(87);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(88);
      END_STATE();
    case 53:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(89);
      END_STATE();
    case 54:
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(90);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(91);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(92);
      END_STATE();
    case 55:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(93);
      END_STATE();
    case 56:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(94);
      END_STATE();
    case 57:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(95);
      END_STATE();
    case 58:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(96);
      END_STATE();
    case 59:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(97);
      END_STATE();
    case 60:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(98);
      END_STATE();
    case 61:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(99);
      END_STATE();
    case 62:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(100);
      END_STATE();
    case 63:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(101);
      END_STATE();
    case 64:
      ACCEPT_TOKEN(aux_sym__hll_non_default_format_var_command_token1);
      END_STATE();
    case 65:
      if (lookahead == 'l') ADVANCE(102);
      END_STATE();
    case 66:
      if (lookahead == 'r') ADVANCE(103);
      END_STATE();
    case 67:
      if (lookahead == 's') ADVANCE(104);
      END_STATE();
    case 68:
      if (lookahead == 'b') ADVANCE(105);
      END_STATE();
    case 69:
      if (lookahead == 'm') ADVANCE(106);
      END_STATE();
    case 70:
      if (lookahead == 'a') ADVANCE(107);
      END_STATE();
    case 71:
      ACCEPT_TOKEN(sym_hll_primitive_type);
      if (lookahead == '1') ADVANCE(108);
      if (lookahead == '3') ADVANCE(109);
      if (lookahead == '6') ADVANCE(110);
      if (lookahead == '8') ADVANCE(111);
      if (lookahead == 'p') ADVANCE(112);
      END_STATE();
    case 72:
      if (lookahead == 'g') ADVANCE(113);
      END_STATE();
    case 73:
      if (lookahead == 'r') ADVANCE(114);
      END_STATE();
    case 74:
      if (lookahead == 'n') ADVANCE(115);
      END_STATE();
    case 75:
      if (lookahead == 'e') ADVANCE(116);
      END_STATE();
    case 76:
      if (lookahead == 'z') ADVANCE(117);
      END_STATE();
    case 77:
      if (lookahead == 'u') ADVANCE(118);
      END_STATE();
    case 78:
      if (lookahead == 't') ADVANCE(119);
      END_STATE();
    case 79:
      if (lookahead == 'o') ADVANCE(120);
      END_STATE();
    case 80:
      if (lookahead == 'i') ADVANCE(121);
      END_STATE();
    case 81:
      if (lookahead == 'd') ADVANCE(102);
      END_STATE();
    case 82:
      if (lookahead == 'a') ADVANCE(122);
      END_STATE();
    case 83:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(123);
      END_STATE();
    case 84:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(124);
      END_STATE();
    case 85:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(125);
      END_STATE();
    case 86:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(126);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(127);
      END_STATE();
    case 87:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(128);
      END_STATE();
    case 88:
      ACCEPT_TOKEN(aux_sym__hll_non_default_format_var_command_token5);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(129);
      END_STATE();
    case 89:
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(130);
      END_STATE();
    case 90:
      ACCEPT_TOKEN(aux_sym__var_command_with_non_hll_format_token2);
      END_STATE();
    case 91:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(131);
      END_STATE();
    case 92:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(132);
      END_STATE();
    case 93:
      ACCEPT_TOKEN(aux_sym__repeat_block_pre_condition_format_token1);
      END_STATE();
    case 94:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(133);
      END_STATE();
    case 95:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(134);
      END_STATE();
    case 96:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(135);
      END_STATE();
    case 97:
      ACCEPT_TOKEN(aux_sym_elif_block_token1);
      END_STATE();
    case 98:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(136);
      END_STATE();
    case 99:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(137);
      END_STATE();
    case 100:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(138);
      END_STATE();
    case 101:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(139);
      END_STATE();
    case 102:
      ACCEPT_TOKEN(sym_hll_primitive_type);
      END_STATE();
    case 103:
      ACCEPT_TOKEN(sym_hll_primitive_type);
      if (lookahead == '1') ADVANCE(140);
      if (lookahead == '3') ADVANCE(141);
      if (lookahead == '6') ADVANCE(142);
      if (lookahead == '8') ADVANCE(143);
      if (lookahead == 'p') ADVANCE(144);
      END_STATE();
    case 104:
      if (lookahead == 't') ADVANCE(145);
      END_STATE();
    case 105:
      if (lookahead == 'l') ADVANCE(146);
      END_STATE();
    case 106:
      ACCEPT_TOKEN(anon_sym_enum);
      END_STATE();
    case 107:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 108:
      if (lookahead == '6') ADVANCE(147);
      END_STATE();
    case 109:
      if (lookahead == '2') ADVANCE(148);
      END_STATE();
    case 110:
      if (lookahead == '4') ADVANCE(149);
      END_STATE();
    case 111:
      if (lookahead == '_') ADVANCE(150);
      END_STATE();
    case 112:
      if (lookahead == 't') ADVANCE(151);
      END_STATE();
    case 113:
      ACCEPT_TOKEN(anon_sym_long);
      END_STATE();
    case 114:
      if (lookahead == 't') ADVANCE(152);
      END_STATE();
    case 115:
      if (lookahead == 'e') ADVANCE(153);
      END_STATE();
    case 116:
      if (lookahead == '_') ADVANCE(154);
      if (lookahead == 'o') ADVANCE(155);
      END_STATE();
    case 117:
      if (lookahead == 'e') ADVANCE(156);
      END_STATE();
    case 118:
      if (lookahead == 'c') ADVANCE(157);
      END_STATE();
    case 119:
      if (lookahead == '1') ADVANCE(158);
      if (lookahead == '3') ADVANCE(159);
      if (lookahead == '6') ADVANCE(160);
      if (lookahead == '8') ADVANCE(161);
      if (lookahead == 'p') ADVANCE(162);
      END_STATE();
    case 120:
      if (lookahead == 'n') ADVANCE(163);
      END_STATE();
    case 121:
      if (lookahead == 'g') ADVANCE(164);
      END_STATE();
    case 122:
      if (lookahead == 't') ADVANCE(165);
      END_STATE();
    case 123:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(166);
      END_STATE();
    case 124:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(167);
      END_STATE();
    case 125:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(168);
      END_STATE();
    case 126:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(169);
      END_STATE();
    case 127:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(170);
      END_STATE();
    case 128:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(171);
      END_STATE();
    case 129:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(172);
      END_STATE();
    case 130:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(173);
      END_STATE();
    case 131:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(174);
      END_STATE();
    case 132:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(175);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(aux_sym__var_command_with_non_hll_format_token3);
      END_STATE();
    case 134:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(176);
      END_STATE();
    case 135:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(177);
      END_STATE();
    case 136:
      ACCEPT_TOKEN(aux_sym_parameter_declaration_token3);
      END_STATE();
    case 137:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(178);
      END_STATE();
    case 138:
      ACCEPT_TOKEN(aux_sym_macro_definition_command_token2);
      END_STATE();
    case 139:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(179);
      END_STATE();
    case 140:
      if (lookahead == '6') ADVANCE(180);
      END_STATE();
    case 141:
      if (lookahead == '2') ADVANCE(181);
      END_STATE();
    case 142:
      if (lookahead == '4') ADVANCE(182);
      END_STATE();
    case 143:
      if (lookahead == '_') ADVANCE(183);
      END_STATE();
    case 144:
      if (lookahead == 't') ADVANCE(184);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(anon_sym_const);
      END_STATE();
    case 146:
      if (lookahead == 'e') ADVANCE(102);
      END_STATE();
    case 147:
      if (lookahead == '_') ADVANCE(185);
      END_STATE();
    case 148:
      if (lookahead == '_') ADVANCE(186);
      END_STATE();
    case 149:
      if (lookahead == '_') ADVANCE(187);
      END_STATE();
    case 150:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 151:
      if (lookahead == 'r') ADVANCE(188);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(anon_sym_short);
      END_STATE();
    case 153:
      if (lookahead == 'd') ADVANCE(189);
      END_STATE();
    case 154:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 155:
      if (lookahead == 'f') ADVANCE(190);
      END_STATE();
    case 156:
      if (lookahead == '_') ADVANCE(191);
      END_STATE();
    case 157:
      if (lookahead == 't') ADVANCE(192);
      END_STATE();
    case 158:
      if (lookahead == '6') ADVANCE(193);
      END_STATE();
    case 159:
      if (lookahead == '2') ADVANCE(194);
      END_STATE();
    case 160:
      if (lookahead == '4') ADVANCE(195);
      END_STATE();
    case 161:
      if (lookahead == '_') ADVANCE(196);
      END_STATE();
    case 162:
      if (lookahead == 't') ADVANCE(197);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_union);
      END_STATE();
    case 164:
      if (lookahead == 'n') ADVANCE(198);
      END_STATE();
    case 165:
      if (lookahead == 'i') ADVANCE(199);
      END_STATE();
    case 166:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(200);
      END_STATE();
    case 167:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(201);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(aux_sym_subroutine_call_expression_token1);
      END_STATE();
    case 169:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(202);
      END_STATE();
    case 170:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(203);
      END_STATE();
    case 171:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(204);
      END_STATE();
    case 172:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(205);
      END_STATE();
    case 173:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(206);
      END_STATE();
    case 174:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(93);
      END_STATE();
    case 175:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(207);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(aux_sym_while_block_token1);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(aux_sym__hll_non_default_format_var_command_token4);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(aux_sym__hll_non_default_format_var_command_token3);
      END_STATE();
    case 179:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(208);
      END_STATE();
    case 180:
      if (lookahead == '_') ADVANCE(209);
      END_STATE();
    case 181:
      if (lookahead == '_') ADVANCE(210);
      END_STATE();
    case 182:
      if (lookahead == '_') ADVANCE(211);
      END_STATE();
    case 183:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 184:
      if (lookahead == 'r') ADVANCE(212);
      END_STATE();
    case 185:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 186:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 187:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 188:
      if (lookahead == '_') ADVANCE(213);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_signed);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_sizeof);
      END_STATE();
    case 191:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_struct);
      END_STATE();
    case 193:
      if (lookahead == '_') ADVANCE(214);
      END_STATE();
    case 194:
      if (lookahead == '_') ADVANCE(215);
      END_STATE();
    case 195:
      if (lookahead == '_') ADVANCE(216);
      END_STATE();
    case 196:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 197:
      if (lookahead == 'r') ADVANCE(217);
      END_STATE();
    case 198:
      if (lookahead == 'e') ADVANCE(218);
      END_STATE();
    case 199:
      if (lookahead == 'l') ADVANCE(219);
      END_STATE();
    case 200:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(220);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(aux_sym_macro_definition_command_token1);
      END_STATE();
    case 202:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(221);
      END_STATE();
    case 203:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(222);
      END_STATE();
    case 204:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(223);
      END_STATE();
    case 205:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(224);
      END_STATE();
    case 206:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(225);
      END_STATE();
    case 207:
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(226);
      END_STATE();
    case 208:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(227);
      END_STATE();
    case 209:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 210:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 211:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 212:
      if (lookahead == '_') ADVANCE(228);
      END_STATE();
    case 213:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 214:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 215:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 216:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 217:
      if (lookahead == '_') ADVANCE(229);
      END_STATE();
    case 218:
      if (lookahead == 'd') ADVANCE(230);
      END_STATE();
    case 219:
      if (lookahead == 'e') ADVANCE(231);
      END_STATE();
    case 220:
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(232);
      END_STATE();
    case 221:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(233);
      END_STATE();
    case 222:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(234);
      END_STATE();
    case 223:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(235);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(aux_sym__hll_non_default_format_var_command_token5);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(aux_sym_macro_definition_command_token3);
      END_STATE();
    case 226:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(236);
      END_STATE();
    case 227:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(237);
      END_STATE();
    case 228:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 229:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(anon_sym_unsigned);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(anon_sym_volatile);
      END_STATE();
    case 232:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(238);
      END_STATE();
    case 233:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(239);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(aux_sym__hll_var_definition_command_expression_token1);
      END_STATE();
    case 235:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(240);
      END_STATE();
    case 236:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(241);
      END_STATE();
    case 237:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(242);
      END_STATE();
    case 238:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(47);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(aux_sym__hll_var_definition_command_expression_token2);
      END_STATE();
    case 240:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(243);
      END_STATE();
    case 241:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(244);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(aux_sym_subroutine_block_token1);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(aux_sym_parameter_declaration_token1);
      END_STATE();
    case 244:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(245);
      END_STATE();
    case 245:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(246);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(aux_sym_parameter_declaration_token2);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 142, .external_lex_state = 2},
  [2] = {.lex_state = 9, .external_lex_state = 3},
  [3] = {.lex_state = 9, .external_lex_state = 3},
  [4] = {.lex_state = 25, .external_lex_state = 4},
  [5] = {.lex_state = 25, .external_lex_state = 4},
  [6] = {.lex_state = 8, .external_lex_state = 4},
  [7] = {.lex_state = 8, .external_lex_state = 4},
  [8] = {.lex_state = 25, .external_lex_state = 4},
  [9] = {.lex_state = 28, .external_lex_state = 5},
  [10] = {.lex_state = 28, .external_lex_state = 5},
  [11] = {.lex_state = 28, .external_lex_state = 5},
  [12] = {.lex_state = 28, .external_lex_state = 5},
  [13] = {.lex_state = 14, .external_lex_state = 6},
  [14] = {.lex_state = 14, .external_lex_state = 6},
  [15] = {.lex_state = 14, .external_lex_state = 6},
  [16] = {.lex_state = 14, .external_lex_state = 6},
  [17] = {.lex_state = 14, .external_lex_state = 6},
  [18] = {.lex_state = 14, .external_lex_state = 6},
  [19] = {.lex_state = 11, .external_lex_state = 5},
  [20] = {.lex_state = 11, .external_lex_state = 5},
  [21] = {.lex_state = 11, .external_lex_state = 5},
  [22] = {.lex_state = 11, .external_lex_state = 5},
  [23] = {.lex_state = 29, .external_lex_state = 5},
  [24] = {.lex_state = 29, .external_lex_state = 5},
  [25] = {.lex_state = 29, .external_lex_state = 5},
  [26] = {.lex_state = 29, .external_lex_state = 5},
  [27] = {.lex_state = 12, .external_lex_state = 5},
  [28] = {.lex_state = 12, .external_lex_state = 5},
  [29] = {.lex_state = 12, .external_lex_state = 5},
  [30] = {.lex_state = 12, .external_lex_state = 5},
  [31] = {.lex_state = 8, .external_lex_state = 7},
  [32] = {.lex_state = 25, .external_lex_state = 7},
  [33] = {.lex_state = 25, .external_lex_state = 7},
  [34] = {.lex_state = 25, .external_lex_state = 7},
  [35] = {.lex_state = 25, .external_lex_state = 7},
  [36] = {.lex_state = 24, .external_lex_state = 8},
  [37] = {.lex_state = 26, .external_lex_state = 5},
  [38] = {.lex_state = 24, .external_lex_state = 8},
  [39] = {.lex_state = 24, .external_lex_state = 8},
  [40] = {.lex_state = 25, .external_lex_state = 7},
  [41] = {.lex_state = 25, .external_lex_state = 7},
  [42] = {.lex_state = 24, .external_lex_state = 8},
  [43] = {.lex_state = 24, .external_lex_state = 8},
  [44] = {.lex_state = 24, .external_lex_state = 8},
  [45] = {.lex_state = 24, .external_lex_state = 8},
  [46] = {.lex_state = 26, .external_lex_state = 5},
  [47] = {.lex_state = 24, .external_lex_state = 8},
  [48] = {.lex_state = 24, .external_lex_state = 8},
  [49] = {.lex_state = 24, .external_lex_state = 8},
  [50] = {.lex_state = 24, .external_lex_state = 8},
  [51] = {.lex_state = 24, .external_lex_state = 8},
  [52] = {.lex_state = 24, .external_lex_state = 8},
  [53] = {.lex_state = 24, .external_lex_state = 8},
  [54] = {.lex_state = 24, .external_lex_state = 8},
  [55] = {.lex_state = 24, .external_lex_state = 8},
  [56] = {.lex_state = 24, .external_lex_state = 8},
  [57] = {.lex_state = 24, .external_lex_state = 8},
  [58] = {.lex_state = 24, .external_lex_state = 8},
  [59] = {.lex_state = 24, .external_lex_state = 8},
  [60] = {.lex_state = 24, .external_lex_state = 8},
  [61] = {.lex_state = 24, .external_lex_state = 8},
  [62] = {.lex_state = 24, .external_lex_state = 8},
  [63] = {.lex_state = 24, .external_lex_state = 8},
  [64] = {.lex_state = 24, .external_lex_state = 8},
  [65] = {.lex_state = 24, .external_lex_state = 8},
  [66] = {.lex_state = 24, .external_lex_state = 8},
  [67] = {.lex_state = 24, .external_lex_state = 8},
  [68] = {.lex_state = 24, .external_lex_state = 8},
  [69] = {.lex_state = 24, .external_lex_state = 8},
  [70] = {.lex_state = 24, .external_lex_state = 8},
  [71] = {.lex_state = 24, .external_lex_state = 8},
  [72] = {.lex_state = 24, .external_lex_state = 8},
  [73] = {.lex_state = 24, .external_lex_state = 8},
  [74] = {.lex_state = 24, .external_lex_state = 8},
  [75] = {.lex_state = 24, .external_lex_state = 8},
  [76] = {.lex_state = 24, .external_lex_state = 8},
  [77] = {.lex_state = 24, .external_lex_state = 8},
  [78] = {.lex_state = 24, .external_lex_state = 8},
  [79] = {.lex_state = 24, .external_lex_state = 8},
  [80] = {.lex_state = 26, .external_lex_state = 5},
  [81] = {.lex_state = 26, .external_lex_state = 5},
  [82] = {.lex_state = 10, .external_lex_state = 5},
  [83] = {.lex_state = 10, .external_lex_state = 5},
  [84] = {.lex_state = 10, .external_lex_state = 5},
  [85] = {.lex_state = 10, .external_lex_state = 5},
  [86] = {.lex_state = 26, .external_lex_state = 5},
  [87] = {.lex_state = 26, .external_lex_state = 5},
  [88] = {.lex_state = 26, .external_lex_state = 5},
  [89] = {.lex_state = 26, .external_lex_state = 5},
  [90] = {.lex_state = 27, .external_lex_state = 5},
  [91] = {.lex_state = 27, .external_lex_state = 5},
  [92] = {.lex_state = 27, .external_lex_state = 5},
  [93] = {.lex_state = 27, .external_lex_state = 5},
  [94] = {.lex_state = 25, .external_lex_state = 7},
  [95] = {.lex_state = 25, .external_lex_state = 7},
  [96] = {.lex_state = 25, .external_lex_state = 7},
  [97] = {.lex_state = 25, .external_lex_state = 7},
  [98] = {.lex_state = 25, .external_lex_state = 7},
  [99] = {.lex_state = 25, .external_lex_state = 7},
  [100] = {.lex_state = 25, .external_lex_state = 7},
  [101] = {.lex_state = 25, .external_lex_state = 7},
  [102] = {.lex_state = 10, .external_lex_state = 5},
  [103] = {.lex_state = 10, .external_lex_state = 5},
  [104] = {.lex_state = 10, .external_lex_state = 5},
  [105] = {.lex_state = 10, .external_lex_state = 5},
  [106] = {.lex_state = 25, .external_lex_state = 7},
  [107] = {.lex_state = 25, .external_lex_state = 7},
  [108] = {.lex_state = 25, .external_lex_state = 7},
  [109] = {.lex_state = 25, .external_lex_state = 7},
  [110] = {.lex_state = 25, .external_lex_state = 7},
  [111] = {.lex_state = 25, .external_lex_state = 7},
  [112] = {.lex_state = 25, .external_lex_state = 7},
  [113] = {.lex_state = 25, .external_lex_state = 7},
  [114] = {.lex_state = 25, .external_lex_state = 7},
  [115] = {.lex_state = 25, .external_lex_state = 7},
  [116] = {.lex_state = 25, .external_lex_state = 7},
  [117] = {.lex_state = 25, .external_lex_state = 7},
  [118] = {.lex_state = 25, .external_lex_state = 7},
  [119] = {.lex_state = 25, .external_lex_state = 7},
  [120] = {.lex_state = 25, .external_lex_state = 7},
  [121] = {.lex_state = 25, .external_lex_state = 7},
  [122] = {.lex_state = 25, .external_lex_state = 7},
  [123] = {.lex_state = 25, .external_lex_state = 7},
  [124] = {.lex_state = 25, .external_lex_state = 7},
  [125] = {.lex_state = 25, .external_lex_state = 7},
  [126] = {.lex_state = 25, .external_lex_state = 7},
  [127] = {.lex_state = 25, .external_lex_state = 7},
  [128] = {.lex_state = 25, .external_lex_state = 7},
  [129] = {.lex_state = 25, .external_lex_state = 7},
  [130] = {.lex_state = 25, .external_lex_state = 7},
  [131] = {.lex_state = 25, .external_lex_state = 7},
  [132] = {.lex_state = 25, .external_lex_state = 7},
  [133] = {.lex_state = 25, .external_lex_state = 7},
  [134] = {.lex_state = 25, .external_lex_state = 7},
  [135] = {.lex_state = 25, .external_lex_state = 7},
  [136] = {.lex_state = 25, .external_lex_state = 7},
  [137] = {.lex_state = 25, .external_lex_state = 7},
  [138] = {.lex_state = 25, .external_lex_state = 7},
  [139] = {.lex_state = 25, .external_lex_state = 7},
  [140] = {.lex_state = 25, .external_lex_state = 7},
  [141] = {.lex_state = 25, .external_lex_state = 7},
  [142] = {.lex_state = 25, .external_lex_state = 7},
  [143] = {.lex_state = 25, .external_lex_state = 7},
  [144] = {.lex_state = 25, .external_lex_state = 7},
  [145] = {.lex_state = 25, .external_lex_state = 7},
  [146] = {.lex_state = 25, .external_lex_state = 7},
  [147] = {.lex_state = 25, .external_lex_state = 7},
  [148] = {.lex_state = 25, .external_lex_state = 7},
  [149] = {.lex_state = 25, .external_lex_state = 7},
  [150] = {.lex_state = 25, .external_lex_state = 7},
  [151] = {.lex_state = 25, .external_lex_state = 7},
  [152] = {.lex_state = 25, .external_lex_state = 7},
  [153] = {.lex_state = 25, .external_lex_state = 7},
  [154] = {.lex_state = 25, .external_lex_state = 7},
  [155] = {.lex_state = 25, .external_lex_state = 7},
  [156] = {.lex_state = 25, .external_lex_state = 7},
  [157] = {.lex_state = 25, .external_lex_state = 7},
  [158] = {.lex_state = 25, .external_lex_state = 7},
  [159] = {.lex_state = 25, .external_lex_state = 7},
  [160] = {.lex_state = 25, .external_lex_state = 7},
  [161] = {.lex_state = 25, .external_lex_state = 7},
  [162] = {.lex_state = 10, .external_lex_state = 5},
  [163] = {.lex_state = 10, .external_lex_state = 5},
  [164] = {.lex_state = 10, .external_lex_state = 5},
  [165] = {.lex_state = 10, .external_lex_state = 5},
  [166] = {.lex_state = 25, .external_lex_state = 7},
  [167] = {.lex_state = 25, .external_lex_state = 7},
  [168] = {.lex_state = 25, .external_lex_state = 7},
  [169] = {.lex_state = 25, .external_lex_state = 7},
  [170] = {.lex_state = 25, .external_lex_state = 7},
  [171] = {.lex_state = 25, .external_lex_state = 7},
  [172] = {.lex_state = 25, .external_lex_state = 7},
  [173] = {.lex_state = 25, .external_lex_state = 7},
  [174] = {.lex_state = 25, .external_lex_state = 7},
  [175] = {.lex_state = 25, .external_lex_state = 7},
  [176] = {.lex_state = 25, .external_lex_state = 7},
  [177] = {.lex_state = 25, .external_lex_state = 7},
  [178] = {.lex_state = 25, .external_lex_state = 7},
  [179] = {.lex_state = 25, .external_lex_state = 7},
  [180] = {.lex_state = 25, .external_lex_state = 7},
  [181] = {.lex_state = 25, .external_lex_state = 7},
  [182] = {.lex_state = 25, .external_lex_state = 7},
  [183] = {.lex_state = 25, .external_lex_state = 7},
  [184] = {.lex_state = 25, .external_lex_state = 7},
  [185] = {.lex_state = 25, .external_lex_state = 7},
  [186] = {.lex_state = 25, .external_lex_state = 7},
  [187] = {.lex_state = 25, .external_lex_state = 7},
  [188] = {.lex_state = 25, .external_lex_state = 7},
  [189] = {.lex_state = 25, .external_lex_state = 7},
  [190] = {.lex_state = 25, .external_lex_state = 7},
  [191] = {.lex_state = 25, .external_lex_state = 7},
  [192] = {.lex_state = 25, .external_lex_state = 7},
  [193] = {.lex_state = 25, .external_lex_state = 7},
  [194] = {.lex_state = 25, .external_lex_state = 7},
  [195] = {.lex_state = 25, .external_lex_state = 7},
  [196] = {.lex_state = 25, .external_lex_state = 7},
  [197] = {.lex_state = 25, .external_lex_state = 7},
  [198] = {.lex_state = 25, .external_lex_state = 7},
  [199] = {.lex_state = 25, .external_lex_state = 7},
  [200] = {.lex_state = 25, .external_lex_state = 7},
  [201] = {.lex_state = 25, .external_lex_state = 7},
  [202] = {.lex_state = 25, .external_lex_state = 7},
  [203] = {.lex_state = 25, .external_lex_state = 7},
  [204] = {.lex_state = 25, .external_lex_state = 7},
  [205] = {.lex_state = 25, .external_lex_state = 7},
  [206] = {.lex_state = 25, .external_lex_state = 7},
  [207] = {.lex_state = 25, .external_lex_state = 7},
  [208] = {.lex_state = 25, .external_lex_state = 7},
  [209] = {.lex_state = 25, .external_lex_state = 7},
  [210] = {.lex_state = 25, .external_lex_state = 7},
  [211] = {.lex_state = 25, .external_lex_state = 7},
  [212] = {.lex_state = 25, .external_lex_state = 7},
  [213] = {.lex_state = 25, .external_lex_state = 7},
  [214] = {.lex_state = 25, .external_lex_state = 7},
  [215] = {.lex_state = 25, .external_lex_state = 7},
  [216] = {.lex_state = 25, .external_lex_state = 7},
  [217] = {.lex_state = 25, .external_lex_state = 7},
  [218] = {.lex_state = 25, .external_lex_state = 7},
  [219] = {.lex_state = 25, .external_lex_state = 7},
  [220] = {.lex_state = 25, .external_lex_state = 7},
  [221] = {.lex_state = 25, .external_lex_state = 7},
  [222] = {.lex_state = 25, .external_lex_state = 7},
  [223] = {.lex_state = 25, .external_lex_state = 7},
  [224] = {.lex_state = 25, .external_lex_state = 7},
  [225] = {.lex_state = 25, .external_lex_state = 7},
  [226] = {.lex_state = 25, .external_lex_state = 7},
  [227] = {.lex_state = 25, .external_lex_state = 7},
  [228] = {.lex_state = 25, .external_lex_state = 7},
  [229] = {.lex_state = 25, .external_lex_state = 7},
  [230] = {.lex_state = 25, .external_lex_state = 7},
  [231] = {.lex_state = 25, .external_lex_state = 7},
  [232] = {.lex_state = 25, .external_lex_state = 7},
  [233] = {.lex_state = 25, .external_lex_state = 7},
  [234] = {.lex_state = 25, .external_lex_state = 7},
  [235] = {.lex_state = 25, .external_lex_state = 7},
  [236] = {.lex_state = 25, .external_lex_state = 7},
  [237] = {.lex_state = 25, .external_lex_state = 7},
  [238] = {.lex_state = 25, .external_lex_state = 7},
  [239] = {.lex_state = 25, .external_lex_state = 7},
  [240] = {.lex_state = 25, .external_lex_state = 7},
  [241] = {.lex_state = 25, .external_lex_state = 7},
  [242] = {.lex_state = 25, .external_lex_state = 7},
  [243] = {.lex_state = 25, .external_lex_state = 7},
  [244] = {.lex_state = 25, .external_lex_state = 7},
  [245] = {.lex_state = 25, .external_lex_state = 7},
  [246] = {.lex_state = 25, .external_lex_state = 7},
  [247] = {.lex_state = 25, .external_lex_state = 7},
  [248] = {.lex_state = 25, .external_lex_state = 7},
  [249] = {.lex_state = 25, .external_lex_state = 7},
  [250] = {.lex_state = 25, .external_lex_state = 7},
  [251] = {.lex_state = 25, .external_lex_state = 7},
  [252] = {.lex_state = 25, .external_lex_state = 7},
  [253] = {.lex_state = 25, .external_lex_state = 7},
  [254] = {.lex_state = 25, .external_lex_state = 7},
  [255] = {.lex_state = 25, .external_lex_state = 7},
  [256] = {.lex_state = 25, .external_lex_state = 7},
  [257] = {.lex_state = 25, .external_lex_state = 7},
  [258] = {.lex_state = 25, .external_lex_state = 7},
  [259] = {.lex_state = 25, .external_lex_state = 7},
  [260] = {.lex_state = 25, .external_lex_state = 7},
  [261] = {.lex_state = 25, .external_lex_state = 7},
  [262] = {.lex_state = 25, .external_lex_state = 7},
  [263] = {.lex_state = 25, .external_lex_state = 7},
  [264] = {.lex_state = 25, .external_lex_state = 7},
  [265] = {.lex_state = 25, .external_lex_state = 7},
  [266] = {.lex_state = 142, .external_lex_state = 2},
  [267] = {.lex_state = 142, .external_lex_state = 2},
  [268] = {.lex_state = 142, .external_lex_state = 2},
  [269] = {.lex_state = 142, .external_lex_state = 2},
  [270] = {.lex_state = 142, .external_lex_state = 2},
  [271] = {.lex_state = 142, .external_lex_state = 2},
  [272] = {.lex_state = 142, .external_lex_state = 2},
  [273] = {.lex_state = 142, .external_lex_state = 2},
  [274] = {.lex_state = 142, .external_lex_state = 2},
  [275] = {.lex_state = 142, .external_lex_state = 2},
  [276] = {.lex_state = 142, .external_lex_state = 2},
  [277] = {.lex_state = 142, .external_lex_state = 2},
  [278] = {.lex_state = 142, .external_lex_state = 2},
  [279] = {.lex_state = 142, .external_lex_state = 2},
  [280] = {.lex_state = 142, .external_lex_state = 2},
  [281] = {.lex_state = 142, .external_lex_state = 2},
  [282] = {.lex_state = 142, .external_lex_state = 2},
  [283] = {.lex_state = 142, .external_lex_state = 2},
  [284] = {.lex_state = 142, .external_lex_state = 2},
  [285] = {.lex_state = 142, .external_lex_state = 2},
  [286] = {.lex_state = 142, .external_lex_state = 2},
  [287] = {.lex_state = 142, .external_lex_state = 2},
  [288] = {.lex_state = 142, .external_lex_state = 2},
  [289] = {.lex_state = 142, .external_lex_state = 2},
  [290] = {.lex_state = 142, .external_lex_state = 2},
  [291] = {.lex_state = 142, .external_lex_state = 2},
  [292] = {.lex_state = 142, .external_lex_state = 2},
  [293] = {.lex_state = 142, .external_lex_state = 2},
  [294] = {.lex_state = 142, .external_lex_state = 2},
  [295] = {.lex_state = 142, .external_lex_state = 2},
  [296] = {.lex_state = 142, .external_lex_state = 2},
  [297] = {.lex_state = 142, .external_lex_state = 2},
  [298] = {.lex_state = 142, .external_lex_state = 2},
  [299] = {.lex_state = 142, .external_lex_state = 2},
  [300] = {.lex_state = 142, .external_lex_state = 2},
  [301] = {.lex_state = 142, .external_lex_state = 2},
  [302] = {.lex_state = 142, .external_lex_state = 2},
  [303] = {.lex_state = 142, .external_lex_state = 2},
  [304] = {.lex_state = 142, .external_lex_state = 2},
  [305] = {.lex_state = 142, .external_lex_state = 2},
  [306] = {.lex_state = 142, .external_lex_state = 2},
  [307] = {.lex_state = 142, .external_lex_state = 2},
  [308] = {.lex_state = 142, .external_lex_state = 2},
  [309] = {.lex_state = 142, .external_lex_state = 2},
  [310] = {.lex_state = 142, .external_lex_state = 2},
  [311] = {.lex_state = 142, .external_lex_state = 2},
  [312] = {.lex_state = 142, .external_lex_state = 2},
  [313] = {.lex_state = 142, .external_lex_state = 2},
  [314] = {.lex_state = 142, .external_lex_state = 2},
  [315] = {.lex_state = 142, .external_lex_state = 2},
  [316] = {.lex_state = 142, .external_lex_state = 2},
  [317] = {.lex_state = 142, .external_lex_state = 2},
  [318] = {.lex_state = 142, .external_lex_state = 2},
  [319] = {.lex_state = 142, .external_lex_state = 2},
  [320] = {.lex_state = 142, .external_lex_state = 2},
  [321] = {.lex_state = 142, .external_lex_state = 2},
  [322] = {.lex_state = 142, .external_lex_state = 2},
  [323] = {.lex_state = 142, .external_lex_state = 2},
  [324] = {.lex_state = 142, .external_lex_state = 2},
  [325] = {.lex_state = 142, .external_lex_state = 2},
  [326] = {.lex_state = 142, .external_lex_state = 2},
  [327] = {.lex_state = 142, .external_lex_state = 2},
  [328] = {.lex_state = 142, .external_lex_state = 2},
  [329] = {.lex_state = 142, .external_lex_state = 2},
  [330] = {.lex_state = 142, .external_lex_state = 2},
  [331] = {.lex_state = 142, .external_lex_state = 2},
  [332] = {.lex_state = 142, .external_lex_state = 2},
  [333] = {.lex_state = 142, .external_lex_state = 2},
  [334] = {.lex_state = 142, .external_lex_state = 2},
  [335] = {.lex_state = 142, .external_lex_state = 2},
  [336] = {.lex_state = 142, .external_lex_state = 2},
  [337] = {.lex_state = 142, .external_lex_state = 2},
  [338] = {.lex_state = 142, .external_lex_state = 2},
  [339] = {.lex_state = 142, .external_lex_state = 2},
  [340] = {.lex_state = 142, .external_lex_state = 2},
  [341] = {.lex_state = 142, .external_lex_state = 2},
  [342] = {.lex_state = 142, .external_lex_state = 2},
  [343] = {.lex_state = 142, .external_lex_state = 2},
  [344] = {.lex_state = 142, .external_lex_state = 2},
  [345] = {.lex_state = 142, .external_lex_state = 2},
  [346] = {.lex_state = 142, .external_lex_state = 2},
  [347] = {.lex_state = 142, .external_lex_state = 2},
  [348] = {.lex_state = 142, .external_lex_state = 2},
  [349] = {.lex_state = 47, .external_lex_state = 2},
  [350] = {.lex_state = 47, .external_lex_state = 2},
  [351] = {.lex_state = 47, .external_lex_state = 2},
  [352] = {.lex_state = 47, .external_lex_state = 2},
  [353] = {.lex_state = 142, .external_lex_state = 2},
  [354] = {.lex_state = 47, .external_lex_state = 2},
  [355] = {.lex_state = 47, .external_lex_state = 2},
  [356] = {.lex_state = 142, .external_lex_state = 2},
  [357] = {.lex_state = 142, .external_lex_state = 2},
  [358] = {.lex_state = 142, .external_lex_state = 2},
  [359] = {.lex_state = 142, .external_lex_state = 2},
  [360] = {.lex_state = 142, .external_lex_state = 2},
  [361] = {.lex_state = 47, .external_lex_state = 2},
  [362] = {.lex_state = 47, .external_lex_state = 2},
  [363] = {.lex_state = 47, .external_lex_state = 2},
  [364] = {.lex_state = 47, .external_lex_state = 2},
  [365] = {.lex_state = 47, .external_lex_state = 2},
  [366] = {.lex_state = 47, .external_lex_state = 2},
  [367] = {.lex_state = 47, .external_lex_state = 2},
  [368] = {.lex_state = 47, .external_lex_state = 2},
  [369] = {.lex_state = 47, .external_lex_state = 2},
  [370] = {.lex_state = 47, .external_lex_state = 2},
  [371] = {.lex_state = 47, .external_lex_state = 2},
  [372] = {.lex_state = 47, .external_lex_state = 2},
  [373] = {.lex_state = 47, .external_lex_state = 2},
  [374] = {.lex_state = 47, .external_lex_state = 2},
  [375] = {.lex_state = 47, .external_lex_state = 2},
  [376] = {.lex_state = 47, .external_lex_state = 2},
  [377] = {.lex_state = 47, .external_lex_state = 2},
  [378] = {.lex_state = 47, .external_lex_state = 2},
  [379] = {.lex_state = 24, .external_lex_state = 8},
  [380] = {.lex_state = 24, .external_lex_state = 8},
  [381] = {.lex_state = 24, .external_lex_state = 8},
  [382] = {.lex_state = 24, .external_lex_state = 8},
  [383] = {.lex_state = 24, .external_lex_state = 8},
  [384] = {.lex_state = 24, .external_lex_state = 8},
  [385] = {.lex_state = 24, .external_lex_state = 8},
  [386] = {.lex_state = 24, .external_lex_state = 8},
  [387] = {.lex_state = 24, .external_lex_state = 8},
  [388] = {.lex_state = 24, .external_lex_state = 8},
  [389] = {.lex_state = 24, .external_lex_state = 8},
  [390] = {.lex_state = 24, .external_lex_state = 8},
  [391] = {.lex_state = 24, .external_lex_state = 8},
  [392] = {.lex_state = 24, .external_lex_state = 8},
  [393] = {.lex_state = 24, .external_lex_state = 8},
  [394] = {.lex_state = 24, .external_lex_state = 8},
  [395] = {.lex_state = 24, .external_lex_state = 8},
  [396] = {.lex_state = 24, .external_lex_state = 8},
  [397] = {.lex_state = 24, .external_lex_state = 8},
  [398] = {.lex_state = 24, .external_lex_state = 8},
  [399] = {.lex_state = 24, .external_lex_state = 8},
  [400] = {.lex_state = 24, .external_lex_state = 8},
  [401] = {.lex_state = 24, .external_lex_state = 8},
  [402] = {.lex_state = 24, .external_lex_state = 8},
  [403] = {.lex_state = 24, .external_lex_state = 8},
  [404] = {.lex_state = 24, .external_lex_state = 8},
  [405] = {.lex_state = 14, .external_lex_state = 9},
  [406] = {.lex_state = 14, .external_lex_state = 9},
  [407] = {.lex_state = 14, .external_lex_state = 9},
  [408] = {.lex_state = 14, .external_lex_state = 9},
  [409] = {.lex_state = 14, .external_lex_state = 6},
  [410] = {.lex_state = 14, .external_lex_state = 6},
  [411] = {.lex_state = 14, .external_lex_state = 6},
  [412] = {.lex_state = 14, .external_lex_state = 6},
  [413] = {.lex_state = 24, .external_lex_state = 8},
  [414] = {.lex_state = 14, .external_lex_state = 6},
  [415] = {.lex_state = 14, .external_lex_state = 6},
  [416] = {.lex_state = 14, .external_lex_state = 6},
  [417] = {.lex_state = 14, .external_lex_state = 6},
  [418] = {.lex_state = 14, .external_lex_state = 6},
  [419] = {.lex_state = 14, .external_lex_state = 6},
  [420] = {.lex_state = 14, .external_lex_state = 6},
  [421] = {.lex_state = 14, .external_lex_state = 6},
  [422] = {.lex_state = 24, .external_lex_state = 8},
  [423] = {.lex_state = 14, .external_lex_state = 6},
  [424] = {.lex_state = 14, .external_lex_state = 6},
  [425] = {.lex_state = 14, .external_lex_state = 6},
  [426] = {.lex_state = 14, .external_lex_state = 6},
  [427] = {.lex_state = 14, .external_lex_state = 6},
  [428] = {.lex_state = 14, .external_lex_state = 6},
  [429] = {.lex_state = 14, .external_lex_state = 6},
  [430] = {.lex_state = 14, .external_lex_state = 6},
  [431] = {.lex_state = 14, .external_lex_state = 6},
  [432] = {.lex_state = 14, .external_lex_state = 6},
  [433] = {.lex_state = 24, .external_lex_state = 8},
  [434] = {.lex_state = 14, .external_lex_state = 6},
  [435] = {.lex_state = 24, .external_lex_state = 8},
  [436] = {.lex_state = 24, .external_lex_state = 8},
  [437] = {.lex_state = 24, .external_lex_state = 8},
  [438] = {.lex_state = 24, .external_lex_state = 8},
  [439] = {.lex_state = 14, .external_lex_state = 9},
  [440] = {.lex_state = 24, .external_lex_state = 8},
  [441] = {.lex_state = 14, .external_lex_state = 6},
  [442] = {.lex_state = 24, .external_lex_state = 8},
  [443] = {.lex_state = 14, .external_lex_state = 9},
  [444] = {.lex_state = 24, .external_lex_state = 8},
  [445] = {.lex_state = 24, .external_lex_state = 8},
  [446] = {.lex_state = 24, .external_lex_state = 8},
  [447] = {.lex_state = 24, .external_lex_state = 8},
  [448] = {.lex_state = 24, .external_lex_state = 8},
  [449] = {.lex_state = 24, .external_lex_state = 8},
  [450] = {.lex_state = 14, .external_lex_state = 6},
  [451] = {.lex_state = 14, .external_lex_state = 6},
  [452] = {.lex_state = 14, .external_lex_state = 6},
  [453] = {.lex_state = 14, .external_lex_state = 6},
  [454] = {.lex_state = 14, .external_lex_state = 6},
  [455] = {.lex_state = 14, .external_lex_state = 6},
  [456] = {.lex_state = 14, .external_lex_state = 6},
  [457] = {.lex_state = 14, .external_lex_state = 6},
  [458] = {.lex_state = 14, .external_lex_state = 6},
  [459] = {.lex_state = 24, .external_lex_state = 8},
  [460] = {.lex_state = 24, .external_lex_state = 8},
  [461] = {.lex_state = 14, .external_lex_state = 6},
  [462] = {.lex_state = 14, .external_lex_state = 6},
  [463] = {.lex_state = 14, .external_lex_state = 6},
  [464] = {.lex_state = 14, .external_lex_state = 6},
  [465] = {.lex_state = 14, .external_lex_state = 6},
  [466] = {.lex_state = 14, .external_lex_state = 6},
  [467] = {.lex_state = 14, .external_lex_state = 6},
  [468] = {.lex_state = 24, .external_lex_state = 8},
  [469] = {.lex_state = 24, .external_lex_state = 8},
  [470] = {.lex_state = 24, .external_lex_state = 8},
  [471] = {.lex_state = 24, .external_lex_state = 8},
  [472] = {.lex_state = 24, .external_lex_state = 8},
  [473] = {.lex_state = 24, .external_lex_state = 8},
  [474] = {.lex_state = 24, .external_lex_state = 8},
  [475] = {.lex_state = 24, .external_lex_state = 8},
  [476] = {.lex_state = 14, .external_lex_state = 6},
  [477] = {.lex_state = 24, .external_lex_state = 8},
  [478] = {.lex_state = 24, .external_lex_state = 8},
  [479] = {.lex_state = 24, .external_lex_state = 8},
  [480] = {.lex_state = 24, .external_lex_state = 8},
  [481] = {.lex_state = 24, .external_lex_state = 8},
  [482] = {.lex_state = 24, .external_lex_state = 8},
  [483] = {.lex_state = 24, .external_lex_state = 8},
  [484] = {.lex_state = 24, .external_lex_state = 8},
  [485] = {.lex_state = 24, .external_lex_state = 8},
  [486] = {.lex_state = 24, .external_lex_state = 8},
  [487] = {.lex_state = 14, .external_lex_state = 6},
  [488] = {.lex_state = 24, .external_lex_state = 8},
  [489] = {.lex_state = 24, .external_lex_state = 8},
  [490] = {.lex_state = 24, .external_lex_state = 8},
  [491] = {.lex_state = 24, .external_lex_state = 8},
  [492] = {.lex_state = 14, .external_lex_state = 9},
  [493] = {.lex_state = 24, .external_lex_state = 8},
  [494] = {.lex_state = 24, .external_lex_state = 8},
  [495] = {.lex_state = 24, .external_lex_state = 8},
  [496] = {.lex_state = 24, .external_lex_state = 8},
  [497] = {.lex_state = 24, .external_lex_state = 8},
  [498] = {.lex_state = 24, .external_lex_state = 8},
  [499] = {.lex_state = 14, .external_lex_state = 6},
  [500] = {.lex_state = 24, .external_lex_state = 8},
  [501] = {.lex_state = 24, .external_lex_state = 8},
  [502] = {.lex_state = 24, .external_lex_state = 8},
  [503] = {.lex_state = 14, .external_lex_state = 9},
  [504] = {.lex_state = 24, .external_lex_state = 8},
  [505] = {.lex_state = 24, .external_lex_state = 8},
  [506] = {.lex_state = 24, .external_lex_state = 8},
  [507] = {.lex_state = 24, .external_lex_state = 8},
  [508] = {.lex_state = 24, .external_lex_state = 8},
  [509] = {.lex_state = 24, .external_lex_state = 8},
  [510] = {.lex_state = 24, .external_lex_state = 8},
  [511] = {.lex_state = 24, .external_lex_state = 8},
  [512] = {.lex_state = 24, .external_lex_state = 8},
  [513] = {.lex_state = 24, .external_lex_state = 8},
  [514] = {.lex_state = 24, .external_lex_state = 8},
  [515] = {.lex_state = 24, .external_lex_state = 8},
  [516] = {.lex_state = 14, .external_lex_state = 6},
  [517] = {.lex_state = 14, .external_lex_state = 6},
  [518] = {.lex_state = 14, .external_lex_state = 6},
  [519] = {.lex_state = 24, .external_lex_state = 8},
  [520] = {.lex_state = 24, .external_lex_state = 8},
  [521] = {.lex_state = 24, .external_lex_state = 8},
  [522] = {.lex_state = 24, .external_lex_state = 8},
  [523] = {.lex_state = 24, .external_lex_state = 8},
  [524] = {.lex_state = 24, .external_lex_state = 8},
  [525] = {.lex_state = 24, .external_lex_state = 8},
  [526] = {.lex_state = 24, .external_lex_state = 8},
  [527] = {.lex_state = 24, .external_lex_state = 8},
  [528] = {.lex_state = 24, .external_lex_state = 8},
  [529] = {.lex_state = 24, .external_lex_state = 8},
  [530] = {.lex_state = 24, .external_lex_state = 8},
  [531] = {.lex_state = 24, .external_lex_state = 8},
  [532] = {.lex_state = 24, .external_lex_state = 8},
  [533] = {.lex_state = 24, .external_lex_state = 8},
  [534] = {.lex_state = 24, .external_lex_state = 8},
  [535] = {.lex_state = 24, .external_lex_state = 8},
  [536] = {.lex_state = 24, .external_lex_state = 8},
  [537] = {.lex_state = 24, .external_lex_state = 8},
  [538] = {.lex_state = 24, .external_lex_state = 8},
  [539] = {.lex_state = 24, .external_lex_state = 8},
  [540] = {.lex_state = 24, .external_lex_state = 8},
  [541] = {.lex_state = 24, .external_lex_state = 8},
  [542] = {.lex_state = 24, .external_lex_state = 8},
  [543] = {.lex_state = 24, .external_lex_state = 8},
  [544] = {.lex_state = 24, .external_lex_state = 8},
  [545] = {.lex_state = 24, .external_lex_state = 8},
  [546] = {.lex_state = 24, .external_lex_state = 8},
  [547] = {.lex_state = 24, .external_lex_state = 8},
  [548] = {.lex_state = 24, .external_lex_state = 8},
  [549] = {.lex_state = 24, .external_lex_state = 8},
  [550] = {.lex_state = 24, .external_lex_state = 8},
  [551] = {.lex_state = 24, .external_lex_state = 8},
  [552] = {.lex_state = 24, .external_lex_state = 8},
  [553] = {.lex_state = 14, .external_lex_state = 6},
  [554] = {.lex_state = 24, .external_lex_state = 8},
  [555] = {.lex_state = 14, .external_lex_state = 6},
  [556] = {.lex_state = 24, .external_lex_state = 8},
  [557] = {.lex_state = 24, .external_lex_state = 8},
  [558] = {.lex_state = 24, .external_lex_state = 8},
  [559] = {.lex_state = 14, .external_lex_state = 6},
  [560] = {.lex_state = 14, .external_lex_state = 6},
  [561] = {.lex_state = 24, .external_lex_state = 8},
  [562] = {.lex_state = 14, .external_lex_state = 6},
  [563] = {.lex_state = 24, .external_lex_state = 8},
  [564] = {.lex_state = 14, .external_lex_state = 6},
  [565] = {.lex_state = 14, .external_lex_state = 6},
  [566] = {.lex_state = 24, .external_lex_state = 8},
  [567] = {.lex_state = 24, .external_lex_state = 8},
  [568] = {.lex_state = 14, .external_lex_state = 6},
  [569] = {.lex_state = 14, .external_lex_state = 6},
  [570] = {.lex_state = 24, .external_lex_state = 8},
  [571] = {.lex_state = 24, .external_lex_state = 8},
  [572] = {.lex_state = 24, .external_lex_state = 8},
  [573] = {.lex_state = 14, .external_lex_state = 6},
  [574] = {.lex_state = 24, .external_lex_state = 8},
  [575] = {.lex_state = 14, .external_lex_state = 6},
  [576] = {.lex_state = 24, .external_lex_state = 8},
  [577] = {.lex_state = 24, .external_lex_state = 8},
  [578] = {.lex_state = 24, .external_lex_state = 8},
  [579] = {.lex_state = 24, .external_lex_state = 8},
  [580] = {.lex_state = 24, .external_lex_state = 8},
  [581] = {.lex_state = 14, .external_lex_state = 6},
  [582] = {.lex_state = 24, .external_lex_state = 8},
  [583] = {.lex_state = 14, .external_lex_state = 6},
  [584] = {.lex_state = 24, .external_lex_state = 8},
  [585] = {.lex_state = 24, .external_lex_state = 8},
  [586] = {.lex_state = 24, .external_lex_state = 8},
  [587] = {.lex_state = 24, .external_lex_state = 8},
  [588] = {.lex_state = 24, .external_lex_state = 8},
  [589] = {.lex_state = 24, .external_lex_state = 8},
  [590] = {.lex_state = 24, .external_lex_state = 8},
  [591] = {.lex_state = 24, .external_lex_state = 8},
  [592] = {.lex_state = 24, .external_lex_state = 8},
  [593] = {.lex_state = 24, .external_lex_state = 8},
  [594] = {.lex_state = 24, .external_lex_state = 8},
  [595] = {.lex_state = 24, .external_lex_state = 8},
  [596] = {.lex_state = 24, .external_lex_state = 8},
  [597] = {.lex_state = 24, .external_lex_state = 8},
  [598] = {.lex_state = 24, .external_lex_state = 8},
  [599] = {.lex_state = 24, .external_lex_state = 8},
  [600] = {.lex_state = 24, .external_lex_state = 8},
  [601] = {.lex_state = 24, .external_lex_state = 8},
  [602] = {.lex_state = 24, .external_lex_state = 8},
  [603] = {.lex_state = 24, .external_lex_state = 8},
  [604] = {.lex_state = 24, .external_lex_state = 8},
  [605] = {.lex_state = 24, .external_lex_state = 8},
  [606] = {.lex_state = 24, .external_lex_state = 8},
  [607] = {.lex_state = 24, .external_lex_state = 8},
  [608] = {.lex_state = 24, .external_lex_state = 8},
  [609] = {.lex_state = 24, .external_lex_state = 8},
  [610] = {.lex_state = 24, .external_lex_state = 8},
  [611] = {.lex_state = 24, .external_lex_state = 8},
  [612] = {.lex_state = 28, .external_lex_state = 10},
  [613] = {.lex_state = 28, .external_lex_state = 10},
  [614] = {.lex_state = 28, .external_lex_state = 10},
  [615] = {.lex_state = 24, .external_lex_state = 8},
  [616] = {.lex_state = 24, .external_lex_state = 8},
  [617] = {.lex_state = 24, .external_lex_state = 8},
  [618] = {.lex_state = 24, .external_lex_state = 8},
  [619] = {.lex_state = 24, .external_lex_state = 8},
  [620] = {.lex_state = 24, .external_lex_state = 8},
  [621] = {.lex_state = 24, .external_lex_state = 8},
  [622] = {.lex_state = 24, .external_lex_state = 8},
  [623] = {.lex_state = 24, .external_lex_state = 8},
  [624] = {.lex_state = 24, .external_lex_state = 8},
  [625] = {.lex_state = 24, .external_lex_state = 8},
  [626] = {.lex_state = 28, .external_lex_state = 10},
  [627] = {.lex_state = 24, .external_lex_state = 8},
  [628] = {.lex_state = 28, .external_lex_state = 10},
  [629] = {.lex_state = 24, .external_lex_state = 8},
  [630] = {.lex_state = 24, .external_lex_state = 8},
  [631] = {.lex_state = 24, .external_lex_state = 8},
  [632] = {.lex_state = 28, .external_lex_state = 10},
  [633] = {.lex_state = 28, .external_lex_state = 10},
  [634] = {.lex_state = 24, .external_lex_state = 8},
  [635] = {.lex_state = 28, .external_lex_state = 10},
  [636] = {.lex_state = 24, .external_lex_state = 8},
  [637] = {.lex_state = 24, .external_lex_state = 8},
  [638] = {.lex_state = 24, .external_lex_state = 8},
  [639] = {.lex_state = 24, .external_lex_state = 8},
  [640] = {.lex_state = 24, .external_lex_state = 8},
  [641] = {.lex_state = 24, .external_lex_state = 8},
  [642] = {.lex_state = 24, .external_lex_state = 8},
  [643] = {.lex_state = 24, .external_lex_state = 8},
  [644] = {.lex_state = 24, .external_lex_state = 8},
  [645] = {.lex_state = 24, .external_lex_state = 8},
  [646] = {.lex_state = 24, .external_lex_state = 8},
  [647] = {.lex_state = 24, .external_lex_state = 8},
  [648] = {.lex_state = 24, .external_lex_state = 8},
  [649] = {.lex_state = 24, .external_lex_state = 8},
  [650] = {.lex_state = 24, .external_lex_state = 8},
  [651] = {.lex_state = 24, .external_lex_state = 8},
  [652] = {.lex_state = 24, .external_lex_state = 8},
  [653] = {.lex_state = 24, .external_lex_state = 8},
  [654] = {.lex_state = 24, .external_lex_state = 8},
  [655] = {.lex_state = 24, .external_lex_state = 8},
  [656] = {.lex_state = 24, .external_lex_state = 8},
  [657] = {.lex_state = 24, .external_lex_state = 8},
  [658] = {.lex_state = 24, .external_lex_state = 8},
  [659] = {.lex_state = 24, .external_lex_state = 8},
  [660] = {.lex_state = 24, .external_lex_state = 8},
  [661] = {.lex_state = 24, .external_lex_state = 8},
  [662] = {.lex_state = 24, .external_lex_state = 8},
  [663] = {.lex_state = 28, .external_lex_state = 10},
  [664] = {.lex_state = 28, .external_lex_state = 10},
  [665] = {.lex_state = 28, .external_lex_state = 10},
  [666] = {.lex_state = 28, .external_lex_state = 10},
  [667] = {.lex_state = 28, .external_lex_state = 10},
  [668] = {.lex_state = 28, .external_lex_state = 10},
  [669] = {.lex_state = 28, .external_lex_state = 10},
  [670] = {.lex_state = 28, .external_lex_state = 10},
  [671] = {.lex_state = 28, .external_lex_state = 10},
  [672] = {.lex_state = 24, .external_lex_state = 8},
  [673] = {.lex_state = 28, .external_lex_state = 10},
  [674] = {.lex_state = 24, .external_lex_state = 8},
  [675] = {.lex_state = 24, .external_lex_state = 8},
  [676] = {.lex_state = 28, .external_lex_state = 10},
  [677] = {.lex_state = 24, .external_lex_state = 8},
  [678] = {.lex_state = 28, .external_lex_state = 10},
  [679] = {.lex_state = 28, .external_lex_state = 10},
  [680] = {.lex_state = 24, .external_lex_state = 8},
  [681] = {.lex_state = 28, .external_lex_state = 10},
  [682] = {.lex_state = 28, .external_lex_state = 10},
  [683] = {.lex_state = 24, .external_lex_state = 8},
  [684] = {.lex_state = 24, .external_lex_state = 8},
  [685] = {.lex_state = 24, .external_lex_state = 8},
  [686] = {.lex_state = 24, .external_lex_state = 8},
  [687] = {.lex_state = 24, .external_lex_state = 8},
  [688] = {.lex_state = 24, .external_lex_state = 8},
  [689] = {.lex_state = 24, .external_lex_state = 8},
  [690] = {.lex_state = 24, .external_lex_state = 8},
  [691] = {.lex_state = 24, .external_lex_state = 8},
  [692] = {.lex_state = 24, .external_lex_state = 8},
  [693] = {.lex_state = 24, .external_lex_state = 8},
  [694] = {.lex_state = 24, .external_lex_state = 8},
  [695] = {.lex_state = 24, .external_lex_state = 8},
  [696] = {.lex_state = 24, .external_lex_state = 8},
  [697] = {.lex_state = 24, .external_lex_state = 8},
  [698] = {.lex_state = 24, .external_lex_state = 8},
  [699] = {.lex_state = 24, .external_lex_state = 8},
  [700] = {.lex_state = 24, .external_lex_state = 8},
  [701] = {.lex_state = 24, .external_lex_state = 8},
  [702] = {.lex_state = 24, .external_lex_state = 8},
  [703] = {.lex_state = 24, .external_lex_state = 8},
  [704] = {.lex_state = 24, .external_lex_state = 8},
  [705] = {.lex_state = 24, .external_lex_state = 8},
  [706] = {.lex_state = 24, .external_lex_state = 8},
  [707] = {.lex_state = 24, .external_lex_state = 8},
  [708] = {.lex_state = 24, .external_lex_state = 8},
  [709] = {.lex_state = 24, .external_lex_state = 8},
  [710] = {.lex_state = 24, .external_lex_state = 8},
  [711] = {.lex_state = 24, .external_lex_state = 8},
  [712] = {.lex_state = 24, .external_lex_state = 8},
  [713] = {.lex_state = 24, .external_lex_state = 8},
  [714] = {.lex_state = 24, .external_lex_state = 8},
  [715] = {.lex_state = 24, .external_lex_state = 8},
  [716] = {.lex_state = 30, .external_lex_state = 10},
  [717] = {.lex_state = 24, .external_lex_state = 8},
  [718] = {.lex_state = 24, .external_lex_state = 8},
  [719] = {.lex_state = 24, .external_lex_state = 8},
  [720] = {.lex_state = 24, .external_lex_state = 8},
  [721] = {.lex_state = 24, .external_lex_state = 8},
  [722] = {.lex_state = 24, .external_lex_state = 8},
  [723] = {.lex_state = 24, .external_lex_state = 8},
  [724] = {.lex_state = 24, .external_lex_state = 8},
  [725] = {.lex_state = 24, .external_lex_state = 8},
  [726] = {.lex_state = 24, .external_lex_state = 8},
  [727] = {.lex_state = 30, .external_lex_state = 10},
  [728] = {.lex_state = 24, .external_lex_state = 8},
  [729] = {.lex_state = 24, .external_lex_state = 8},
  [730] = {.lex_state = 24, .external_lex_state = 8},
  [731] = {.lex_state = 24, .external_lex_state = 8},
  [732] = {.lex_state = 24, .external_lex_state = 8},
  [733] = {.lex_state = 24, .external_lex_state = 8},
  [734] = {.lex_state = 24, .external_lex_state = 8},
  [735] = {.lex_state = 24, .external_lex_state = 8},
  [736] = {.lex_state = 24, .external_lex_state = 8},
  [737] = {.lex_state = 24, .external_lex_state = 8},
  [738] = {.lex_state = 24, .external_lex_state = 8},
  [739] = {.lex_state = 24, .external_lex_state = 8},
  [740] = {.lex_state = 24, .external_lex_state = 8},
  [741] = {.lex_state = 24, .external_lex_state = 8},
  [742] = {.lex_state = 24, .external_lex_state = 8},
  [743] = {.lex_state = 24, .external_lex_state = 8},
  [744] = {.lex_state = 24, .external_lex_state = 8},
  [745] = {.lex_state = 24, .external_lex_state = 8},
  [746] = {.lex_state = 24, .external_lex_state = 8},
  [747] = {.lex_state = 24, .external_lex_state = 8},
  [748] = {.lex_state = 24, .external_lex_state = 8},
  [749] = {.lex_state = 24, .external_lex_state = 8},
  [750] = {.lex_state = 24, .external_lex_state = 8},
  [751] = {.lex_state = 24, .external_lex_state = 8},
  [752] = {.lex_state = 24, .external_lex_state = 8},
  [753] = {.lex_state = 24, .external_lex_state = 8},
  [754] = {.lex_state = 24, .external_lex_state = 8},
  [755] = {.lex_state = 24, .external_lex_state = 8},
  [756] = {.lex_state = 24, .external_lex_state = 8},
  [757] = {.lex_state = 24, .external_lex_state = 8},
  [758] = {.lex_state = 24, .external_lex_state = 8},
  [759] = {.lex_state = 24, .external_lex_state = 8},
  [760] = {.lex_state = 24, .external_lex_state = 8},
  [761] = {.lex_state = 24, .external_lex_state = 8},
  [762] = {.lex_state = 24, .external_lex_state = 8},
  [763] = {.lex_state = 24, .external_lex_state = 8},
  [764] = {.lex_state = 24, .external_lex_state = 8},
  [765] = {.lex_state = 24, .external_lex_state = 8},
  [766] = {.lex_state = 24, .external_lex_state = 8},
  [767] = {.lex_state = 24, .external_lex_state = 8},
  [768] = {.lex_state = 24, .external_lex_state = 8},
  [769] = {.lex_state = 24, .external_lex_state = 8},
  [770] = {.lex_state = 24, .external_lex_state = 8},
  [771] = {.lex_state = 24, .external_lex_state = 8},
  [772] = {.lex_state = 24, .external_lex_state = 8},
  [773] = {.lex_state = 24, .external_lex_state = 8},
  [774] = {.lex_state = 24, .external_lex_state = 8},
  [775] = {.lex_state = 24, .external_lex_state = 8},
  [776] = {.lex_state = 24, .external_lex_state = 8},
  [777] = {.lex_state = 24, .external_lex_state = 8},
  [778] = {.lex_state = 24, .external_lex_state = 8},
  [779] = {.lex_state = 24, .external_lex_state = 8},
  [780] = {.lex_state = 24, .external_lex_state = 8},
  [781] = {.lex_state = 24, .external_lex_state = 8},
  [782] = {.lex_state = 24, .external_lex_state = 8},
  [783] = {.lex_state = 24, .external_lex_state = 8},
  [784] = {.lex_state = 24, .external_lex_state = 8},
  [785] = {.lex_state = 24, .external_lex_state = 8},
  [786] = {.lex_state = 24, .external_lex_state = 8},
  [787] = {.lex_state = 24, .external_lex_state = 8},
  [788] = {.lex_state = 24, .external_lex_state = 8},
  [789] = {.lex_state = 24, .external_lex_state = 8},
  [790] = {.lex_state = 24, .external_lex_state = 8},
  [791] = {.lex_state = 24, .external_lex_state = 8},
  [792] = {.lex_state = 24, .external_lex_state = 8},
  [793] = {.lex_state = 24, .external_lex_state = 8},
  [794] = {.lex_state = 24, .external_lex_state = 8},
  [795] = {.lex_state = 24, .external_lex_state = 8},
  [796] = {.lex_state = 24, .external_lex_state = 8},
  [797] = {.lex_state = 24, .external_lex_state = 8},
  [798] = {.lex_state = 24, .external_lex_state = 8},
  [799] = {.lex_state = 28, .external_lex_state = 10},
  [800] = {.lex_state = 24, .external_lex_state = 8},
  [801] = {.lex_state = 24, .external_lex_state = 8},
  [802] = {.lex_state = 24, .external_lex_state = 8},
  [803] = {.lex_state = 24, .external_lex_state = 8},
  [804] = {.lex_state = 24, .external_lex_state = 8},
  [805] = {.lex_state = 24, .external_lex_state = 8},
  [806] = {.lex_state = 24, .external_lex_state = 8},
  [807] = {.lex_state = 24, .external_lex_state = 8},
  [808] = {.lex_state = 24, .external_lex_state = 8},
  [809] = {.lex_state = 24, .external_lex_state = 8},
  [810] = {.lex_state = 30, .external_lex_state = 10},
  [811] = {.lex_state = 28, .external_lex_state = 10},
  [812] = {.lex_state = 28, .external_lex_state = 10},
  [813] = {.lex_state = 28, .external_lex_state = 10},
  [814] = {.lex_state = 28, .external_lex_state = 10},
  [815] = {.lex_state = 28, .external_lex_state = 10},
  [816] = {.lex_state = 28, .external_lex_state = 10},
  [817] = {.lex_state = 28, .external_lex_state = 10},
  [818] = {.lex_state = 28, .external_lex_state = 10},
  [819] = {.lex_state = 28, .external_lex_state = 10},
  [820] = {.lex_state = 28, .external_lex_state = 10},
  [821] = {.lex_state = 28, .external_lex_state = 10},
  [822] = {.lex_state = 28, .external_lex_state = 10},
  [823] = {.lex_state = 28, .external_lex_state = 10},
  [824] = {.lex_state = 28, .external_lex_state = 10},
  [825] = {.lex_state = 28, .external_lex_state = 10},
  [826] = {.lex_state = 28, .external_lex_state = 10},
  [827] = {.lex_state = 28, .external_lex_state = 10},
  [828] = {.lex_state = 28, .external_lex_state = 10},
  [829] = {.lex_state = 28, .external_lex_state = 10},
  [830] = {.lex_state = 28, .external_lex_state = 10},
  [831] = {.lex_state = 28, .external_lex_state = 10},
  [832] = {.lex_state = 28, .external_lex_state = 10},
  [833] = {.lex_state = 11, .external_lex_state = 10},
  [834] = {.lex_state = 13, .external_lex_state = 10},
  [835] = {.lex_state = 11, .external_lex_state = 10},
  [836] = {.lex_state = 11, .external_lex_state = 10},
  [837] = {.lex_state = 11, .external_lex_state = 10},
  [838] = {.lex_state = 11, .external_lex_state = 10},
  [839] = {.lex_state = 11, .external_lex_state = 10},
  [840] = {.lex_state = 11, .external_lex_state = 10},
  [841] = {.lex_state = 30, .external_lex_state = 10},
  [842] = {.lex_state = 24, .external_lex_state = 8},
  [843] = {.lex_state = 13, .external_lex_state = 10},
  [844] = {.lex_state = 11, .external_lex_state = 10},
  [845] = {.lex_state = 11, .external_lex_state = 10},
  [846] = {.lex_state = 11, .external_lex_state = 10},
  [847] = {.lex_state = 11, .external_lex_state = 10},
  [848] = {.lex_state = 11, .external_lex_state = 10},
  [849] = {.lex_state = 11, .external_lex_state = 10},
  [850] = {.lex_state = 13, .external_lex_state = 10},
  [851] = {.lex_state = 11, .external_lex_state = 10},
  [852] = {.lex_state = 11, .external_lex_state = 10},
  [853] = {.lex_state = 11, .external_lex_state = 10},
  [854] = {.lex_state = 11, .external_lex_state = 10},
  [855] = {.lex_state = 11, .external_lex_state = 10},
  [856] = {.lex_state = 28, .external_lex_state = 10},
  [857] = {.lex_state = 28, .external_lex_state = 10},
  [858] = {.lex_state = 11, .external_lex_state = 10},
  [859] = {.lex_state = 11, .external_lex_state = 10},
  [860] = {.lex_state = 11, .external_lex_state = 10},
  [861] = {.lex_state = 11, .external_lex_state = 10},
  [862] = {.lex_state = 11, .external_lex_state = 10},
  [863] = {.lex_state = 11, .external_lex_state = 10},
  [864] = {.lex_state = 29, .external_lex_state = 10},
  [865] = {.lex_state = 11, .external_lex_state = 10},
  [866] = {.lex_state = 29, .external_lex_state = 10},
  [867] = {.lex_state = 29, .external_lex_state = 10},
  [868] = {.lex_state = 29, .external_lex_state = 10},
  [869] = {.lex_state = 29, .external_lex_state = 10},
  [870] = {.lex_state = 29, .external_lex_state = 10},
  [871] = {.lex_state = 29, .external_lex_state = 10},
  [872] = {.lex_state = 11, .external_lex_state = 10},
  [873] = {.lex_state = 29, .external_lex_state = 10},
  [874] = {.lex_state = 11, .external_lex_state = 10},
  [875] = {.lex_state = 11, .external_lex_state = 10},
  [876] = {.lex_state = 29, .external_lex_state = 10},
  [877] = {.lex_state = 11, .external_lex_state = 10},
  [878] = {.lex_state = 11, .external_lex_state = 10},
  [879] = {.lex_state = 29, .external_lex_state = 10},
  [880] = {.lex_state = 11, .external_lex_state = 10},
  [881] = {.lex_state = 11, .external_lex_state = 10},
  [882] = {.lex_state = 11, .external_lex_state = 10},
  [883] = {.lex_state = 29, .external_lex_state = 10},
  [884] = {.lex_state = 11, .external_lex_state = 10},
  [885] = {.lex_state = 29, .external_lex_state = 10},
  [886] = {.lex_state = 29, .external_lex_state = 10},
  [887] = {.lex_state = 11, .external_lex_state = 10},
  [888] = {.lex_state = 29, .external_lex_state = 10},
  [889] = {.lex_state = 29, .external_lex_state = 10},
  [890] = {.lex_state = 11, .external_lex_state = 10},
  [891] = {.lex_state = 11, .external_lex_state = 10},
  [892] = {.lex_state = 29, .external_lex_state = 10},
  [893] = {.lex_state = 29, .external_lex_state = 10},
  [894] = {.lex_state = 11, .external_lex_state = 10},
  [895] = {.lex_state = 29, .external_lex_state = 10},
  [896] = {.lex_state = 11, .external_lex_state = 10},
  [897] = {.lex_state = 29, .external_lex_state = 10},
  [898] = {.lex_state = 11, .external_lex_state = 10},
  [899] = {.lex_state = 29, .external_lex_state = 10},
  [900] = {.lex_state = 11, .external_lex_state = 10},
  [901] = {.lex_state = 11, .external_lex_state = 10},
  [902] = {.lex_state = 11, .external_lex_state = 10},
  [903] = {.lex_state = 11, .external_lex_state = 10},
  [904] = {.lex_state = 11, .external_lex_state = 10},
  [905] = {.lex_state = 11, .external_lex_state = 10},
  [906] = {.lex_state = 29, .external_lex_state = 10},
  [907] = {.lex_state = 29, .external_lex_state = 10},
  [908] = {.lex_state = 29, .external_lex_state = 10},
  [909] = {.lex_state = 33, .external_lex_state = 10},
  [910] = {.lex_state = 33, .external_lex_state = 10},
  [911] = {.lex_state = 33, .external_lex_state = 10},
  [912] = {.lex_state = 11, .external_lex_state = 10},
  [913] = {.lex_state = 28, .external_lex_state = 10},
  [914] = {.lex_state = 28, .external_lex_state = 10},
  [915] = {.lex_state = 29, .external_lex_state = 10},
  [916] = {.lex_state = 11, .external_lex_state = 10},
  [917] = {.lex_state = 29, .external_lex_state = 10},
  [918] = {.lex_state = 29, .external_lex_state = 10},
  [919] = {.lex_state = 12, .external_lex_state = 10},
  [920] = {.lex_state = 16, .external_lex_state = 10},
  [921] = {.lex_state = 12, .external_lex_state = 10},
  [922] = {.lex_state = 29, .external_lex_state = 10},
  [923] = {.lex_state = 12, .external_lex_state = 10},
  [924] = {.lex_state = 12, .external_lex_state = 10},
  [925] = {.lex_state = 29, .external_lex_state = 10},
  [926] = {.lex_state = 29, .external_lex_state = 10},
  [927] = {.lex_state = 29, .external_lex_state = 10},
  [928] = {.lex_state = 29, .external_lex_state = 10},
  [929] = {.lex_state = 29, .external_lex_state = 10},
  [930] = {.lex_state = 29, .external_lex_state = 10},
  [931] = {.lex_state = 29, .external_lex_state = 10},
  [932] = {.lex_state = 29, .external_lex_state = 10},
  [933] = {.lex_state = 12, .external_lex_state = 10},
  [934] = {.lex_state = 12, .external_lex_state = 10},
  [935] = {.lex_state = 12, .external_lex_state = 10},
  [936] = {.lex_state = 29, .external_lex_state = 10},
  [937] = {.lex_state = 12, .external_lex_state = 10},
  [938] = {.lex_state = 29, .external_lex_state = 10},
  [939] = {.lex_state = 29, .external_lex_state = 10},
  [940] = {.lex_state = 12, .external_lex_state = 10},
  [941] = {.lex_state = 16, .external_lex_state = 10},
  [942] = {.lex_state = 29, .external_lex_state = 10},
  [943] = {.lex_state = 29, .external_lex_state = 10},
  [944] = {.lex_state = 29, .external_lex_state = 10},
  [945] = {.lex_state = 12, .external_lex_state = 10},
  [946] = {.lex_state = 12, .external_lex_state = 10},
  [947] = {.lex_state = 12, .external_lex_state = 10},
  [948] = {.lex_state = 29, .external_lex_state = 10},
  [949] = {.lex_state = 29, .external_lex_state = 10},
  [950] = {.lex_state = 29, .external_lex_state = 10},
  [951] = {.lex_state = 29, .external_lex_state = 10},
  [952] = {.lex_state = 12, .external_lex_state = 10},
  [953] = {.lex_state = 12, .external_lex_state = 10},
  [954] = {.lex_state = 12, .external_lex_state = 10},
  [955] = {.lex_state = 12, .external_lex_state = 10},
  [956] = {.lex_state = 29, .external_lex_state = 10},
  [957] = {.lex_state = 12, .external_lex_state = 10},
  [958] = {.lex_state = 12, .external_lex_state = 10},
  [959] = {.lex_state = 29, .external_lex_state = 10},
  [960] = {.lex_state = 12, .external_lex_state = 10},
  [961] = {.lex_state = 12, .external_lex_state = 10},
  [962] = {.lex_state = 29, .external_lex_state = 10},
  [963] = {.lex_state = 12, .external_lex_state = 10},
  [964] = {.lex_state = 12, .external_lex_state = 10},
  [965] = {.lex_state = 12, .external_lex_state = 10},
  [966] = {.lex_state = 12, .external_lex_state = 10},
  [967] = {.lex_state = 29, .external_lex_state = 10},
  [968] = {.lex_state = 29, .external_lex_state = 10},
  [969] = {.lex_state = 16, .external_lex_state = 10},
  [970] = {.lex_state = 12, .external_lex_state = 10},
  [971] = {.lex_state = 12, .external_lex_state = 10},
  [972] = {.lex_state = 12, .external_lex_state = 10},
  [973] = {.lex_state = 12, .external_lex_state = 10},
  [974] = {.lex_state = 12, .external_lex_state = 10},
  [975] = {.lex_state = 12, .external_lex_state = 10},
  [976] = {.lex_state = 12, .external_lex_state = 10},
  [977] = {.lex_state = 12, .external_lex_state = 10},
  [978] = {.lex_state = 12, .external_lex_state = 10},
  [979] = {.lex_state = 12, .external_lex_state = 10},
  [980] = {.lex_state = 12, .external_lex_state = 10},
  [981] = {.lex_state = 12, .external_lex_state = 10},
  [982] = {.lex_state = 12, .external_lex_state = 10},
  [983] = {.lex_state = 12, .external_lex_state = 10},
  [984] = {.lex_state = 12, .external_lex_state = 10},
  [985] = {.lex_state = 12, .external_lex_state = 10},
  [986] = {.lex_state = 12, .external_lex_state = 10},
  [987] = {.lex_state = 12, .external_lex_state = 10},
  [988] = {.lex_state = 12, .external_lex_state = 10},
  [989] = {.lex_state = 12, .external_lex_state = 10},
  [990] = {.lex_state = 12, .external_lex_state = 10},
  [991] = {.lex_state = 12, .external_lex_state = 10},
  [992] = {.lex_state = 12, .external_lex_state = 10},
  [993] = {.lex_state = 12, .external_lex_state = 10},
  [994] = {.lex_state = 25, .external_lex_state = 7},
  [995] = {.lex_state = 25, .external_lex_state = 7},
  [996] = {.lex_state = 25, .external_lex_state = 7},
  [997] = {.lex_state = 25, .external_lex_state = 7},
  [998] = {.lex_state = 25, .external_lex_state = 7},
  [999] = {.lex_state = 25, .external_lex_state = 7},
  [1000] = {.lex_state = 8, .external_lex_state = 7},
  [1001] = {.lex_state = 8, .external_lex_state = 7},
  [1002] = {.lex_state = 25, .external_lex_state = 7},
  [1003] = {.lex_state = 8, .external_lex_state = 7},
  [1004] = {.lex_state = 8, .external_lex_state = 7},
  [1005] = {.lex_state = 8, .external_lex_state = 7},
  [1006] = {.lex_state = 25, .external_lex_state = 7},
  [1007] = {.lex_state = 25, .external_lex_state = 7},
  [1008] = {.lex_state = 25, .external_lex_state = 7},
  [1009] = {.lex_state = 25, .external_lex_state = 7},
  [1010] = {.lex_state = 25, .external_lex_state = 7},
  [1011] = {.lex_state = 25, .external_lex_state = 7},
  [1012] = {.lex_state = 25, .external_lex_state = 7},
  [1013] = {.lex_state = 45},
  [1014] = {.lex_state = 31, .external_lex_state = 10},
  [1015] = {.lex_state = 26, .external_lex_state = 10},
  [1016] = {.lex_state = 31, .external_lex_state = 10},
  [1017] = {.lex_state = 31, .external_lex_state = 10},
  [1018] = {.lex_state = 26, .external_lex_state = 10},
  [1019] = {.lex_state = 26, .external_lex_state = 10},
  [1020] = {.lex_state = 26, .external_lex_state = 10},
  [1021] = {.lex_state = 26, .external_lex_state = 10},
  [1022] = {.lex_state = 26, .external_lex_state = 10},
  [1023] = {.lex_state = 26, .external_lex_state = 10},
  [1024] = {.lex_state = 24, .external_lex_state = 8},
  [1025] = {.lex_state = 26, .external_lex_state = 10},
  [1026] = {.lex_state = 26, .external_lex_state = 10},
  [1027] = {.lex_state = 10, .external_lex_state = 10},
  [1028] = {.lex_state = 26, .external_lex_state = 10},
  [1029] = {.lex_state = 26, .external_lex_state = 10},
  [1030] = {.lex_state = 25, .external_lex_state = 11},
  [1031] = {.lex_state = 26, .external_lex_state = 10},
  [1032] = {.lex_state = 24, .external_lex_state = 8},
  [1033] = {.lex_state = 10, .external_lex_state = 10},
  [1034] = {.lex_state = 26, .external_lex_state = 10},
  [1035] = {.lex_state = 26, .external_lex_state = 10},
  [1036] = {.lex_state = 24, .external_lex_state = 8},
  [1037] = {.lex_state = 26, .external_lex_state = 10},
  [1038] = {.lex_state = 26, .external_lex_state = 10},
  [1039] = {.lex_state = 26, .external_lex_state = 10},
  [1040] = {.lex_state = 26, .external_lex_state = 10},
  [1041] = {.lex_state = 26, .external_lex_state = 10},
  [1042] = {.lex_state = 24, .external_lex_state = 8},
  [1043] = {.lex_state = 26, .external_lex_state = 10},
  [1044] = {.lex_state = 25, .external_lex_state = 11},
  [1045] = {.lex_state = 26, .external_lex_state = 10},
  [1046] = {.lex_state = 26, .external_lex_state = 10},
  [1047] = {.lex_state = 26, .external_lex_state = 10},
  [1048] = {.lex_state = 26, .external_lex_state = 10},
  [1049] = {.lex_state = 26, .external_lex_state = 10},
  [1050] = {.lex_state = 10, .external_lex_state = 10},
  [1051] = {.lex_state = 24, .external_lex_state = 8},
  [1052] = {.lex_state = 10, .external_lex_state = 10},
  [1053] = {.lex_state = 26, .external_lex_state = 10},
  [1054] = {.lex_state = 25, .external_lex_state = 11},
  [1055] = {.lex_state = 10, .external_lex_state = 10},
  [1056] = {.lex_state = 10, .external_lex_state = 10},
  [1057] = {.lex_state = 26, .external_lex_state = 10},
  [1058] = {.lex_state = 10, .external_lex_state = 10},
  [1059] = {.lex_state = 25, .external_lex_state = 11},
  [1060] = {.lex_state = 25, .external_lex_state = 11},
  [1061] = {.lex_state = 26, .external_lex_state = 10},
  [1062] = {.lex_state = 26, .external_lex_state = 10},
  [1063] = {.lex_state = 10, .external_lex_state = 10},
  [1064] = {.lex_state = 26, .external_lex_state = 10},
  [1065] = {.lex_state = 26, .external_lex_state = 10},
  [1066] = {.lex_state = 26, .external_lex_state = 10},
  [1067] = {.lex_state = 26, .external_lex_state = 10},
  [1068] = {.lex_state = 26, .external_lex_state = 10},
  [1069] = {.lex_state = 26, .external_lex_state = 10},
  [1070] = {.lex_state = 26, .external_lex_state = 10},
  [1071] = {.lex_state = 26, .external_lex_state = 10},
  [1072] = {.lex_state = 26, .external_lex_state = 10},
  [1073] = {.lex_state = 25, .external_lex_state = 12},
  [1074] = {.lex_state = 26, .external_lex_state = 10},
  [1075] = {.lex_state = 10, .external_lex_state = 10},
  [1076] = {.lex_state = 26, .external_lex_state = 10},
  [1077] = {.lex_state = 25, .external_lex_state = 7},
  [1078] = {.lex_state = 25, .external_lex_state = 7},
  [1079] = {.lex_state = 10, .external_lex_state = 10},
  [1080] = {.lex_state = 26, .external_lex_state = 10},
  [1081] = {.lex_state = 10, .external_lex_state = 10},
  [1082] = {.lex_state = 10, .external_lex_state = 10},
  [1083] = {.lex_state = 10, .external_lex_state = 10},
  [1084] = {.lex_state = 10, .external_lex_state = 10},
  [1085] = {.lex_state = 26, .external_lex_state = 10},
  [1086] = {.lex_state = 10, .external_lex_state = 10},
  [1087] = {.lex_state = 26, .external_lex_state = 10},
  [1088] = {.lex_state = 10, .external_lex_state = 10},
  [1089] = {.lex_state = 26, .external_lex_state = 10},
  [1090] = {.lex_state = 8, .external_lex_state = 11},
  [1091] = {.lex_state = 10, .external_lex_state = 10},
  [1092] = {.lex_state = 10, .external_lex_state = 10},
  [1093] = {.lex_state = 10, .external_lex_state = 10},
  [1094] = {.lex_state = 26, .external_lex_state = 10},
  [1095] = {.lex_state = 10, .external_lex_state = 10},
  [1096] = {.lex_state = 25, .external_lex_state = 12},
  [1097] = {.lex_state = 8, .external_lex_state = 11},
  [1098] = {.lex_state = 10, .external_lex_state = 10},
  [1099] = {.lex_state = 25, .external_lex_state = 12},
  [1100] = {.lex_state = 15, .external_lex_state = 10},
  [1101] = {.lex_state = 26, .external_lex_state = 10},
  [1102] = {.lex_state = 10, .external_lex_state = 10},
  [1103] = {.lex_state = 10, .external_lex_state = 10},
  [1104] = {.lex_state = 10, .external_lex_state = 10},
  [1105] = {.lex_state = 25, .external_lex_state = 11},
  [1106] = {.lex_state = 10, .external_lex_state = 10},
  [1107] = {.lex_state = 10, .external_lex_state = 10},
  [1108] = {.lex_state = 10, .external_lex_state = 10},
  [1109] = {.lex_state = 10, .external_lex_state = 10},
  [1110] = {.lex_state = 10, .external_lex_state = 10},
  [1111] = {.lex_state = 25, .external_lex_state = 12},
  [1112] = {.lex_state = 25, .external_lex_state = 12},
  [1113] = {.lex_state = 25, .external_lex_state = 12},
  [1114] = {.lex_state = 25, .external_lex_state = 7},
  [1115] = {.lex_state = 25, .external_lex_state = 12},
  [1116] = {.lex_state = 25, .external_lex_state = 12},
  [1117] = {.lex_state = 10, .external_lex_state = 10},
  [1118] = {.lex_state = 25, .external_lex_state = 12},
  [1119] = {.lex_state = 25, .external_lex_state = 13},
  [1120] = {.lex_state = 24, .external_lex_state = 8},
  [1121] = {.lex_state = 25, .external_lex_state = 12},
  [1122] = {.lex_state = 26, .external_lex_state = 10},
  [1123] = {.lex_state = 24, .external_lex_state = 8},
  [1124] = {.lex_state = 25, .external_lex_state = 12},
  [1125] = {.lex_state = 27, .external_lex_state = 10},
  [1126] = {.lex_state = 25, .external_lex_state = 12},
  [1127] = {.lex_state = 25, .external_lex_state = 12},
  [1128] = {.lex_state = 25, .external_lex_state = 12},
  [1129] = {.lex_state = 25, .external_lex_state = 12},
  [1130] = {.lex_state = 25, .external_lex_state = 12},
  [1131] = {.lex_state = 25, .external_lex_state = 13},
  [1132] = {.lex_state = 15, .external_lex_state = 10},
  [1133] = {.lex_state = 25, .external_lex_state = 12},
  [1134] = {.lex_state = 25, .external_lex_state = 12},
  [1135] = {.lex_state = 24, .external_lex_state = 8},
  [1136] = {.lex_state = 8, .external_lex_state = 11},
  [1137] = {.lex_state = 24, .external_lex_state = 8},
  [1138] = {.lex_state = 26, .external_lex_state = 10},
  [1139] = {.lex_state = 15, .external_lex_state = 10},
  [1140] = {.lex_state = 25, .external_lex_state = 7},
  [1141] = {.lex_state = 25, .external_lex_state = 12},
  [1142] = {.lex_state = 25, .external_lex_state = 12},
  [1143] = {.lex_state = 26, .external_lex_state = 10},
  [1144] = {.lex_state = 25, .external_lex_state = 12},
  [1145] = {.lex_state = 25, .external_lex_state = 12},
  [1146] = {.lex_state = 25, .external_lex_state = 12},
  [1147] = {.lex_state = 25, .external_lex_state = 12},
  [1148] = {.lex_state = 8, .external_lex_state = 11},
  [1149] = {.lex_state = 26, .external_lex_state = 10},
  [1150] = {.lex_state = 26, .external_lex_state = 10},
  [1151] = {.lex_state = 10, .external_lex_state = 10},
  [1152] = {.lex_state = 26, .external_lex_state = 10},
  [1153] = {.lex_state = 27, .external_lex_state = 10},
  [1154] = {.lex_state = 25, .external_lex_state = 12},
  [1155] = {.lex_state = 25, .external_lex_state = 12},
  [1156] = {.lex_state = 8, .external_lex_state = 12},
  [1157] = {.lex_state = 27, .external_lex_state = 10},
  [1158] = {.lex_state = 8, .external_lex_state = 11},
  [1159] = {.lex_state = 26, .external_lex_state = 10},
  [1160] = {.lex_state = 24, .external_lex_state = 8},
  [1161] = {.lex_state = 26, .external_lex_state = 10},
  [1162] = {.lex_state = 26, .external_lex_state = 10},
  [1163] = {.lex_state = 8, .external_lex_state = 12},
  [1164] = {.lex_state = 27, .external_lex_state = 10},
  [1165] = {.lex_state = 25, .external_lex_state = 12},
  [1166] = {.lex_state = 27, .external_lex_state = 10},
  [1167] = {.lex_state = 25, .external_lex_state = 12},
  [1168] = {.lex_state = 10, .external_lex_state = 10},
  [1169] = {.lex_state = 26, .external_lex_state = 10},
  [1170] = {.lex_state = 25, .external_lex_state = 12},
  [1171] = {.lex_state = 26, .external_lex_state = 10},
  [1172] = {.lex_state = 8, .external_lex_state = 13},
  [1173] = {.lex_state = 8, .external_lex_state = 12},
  [1174] = {.lex_state = 10, .external_lex_state = 10},
  [1175] = {.lex_state = 26, .external_lex_state = 10},
  [1176] = {.lex_state = 26, .external_lex_state = 10},
  [1177] = {.lex_state = 8, .external_lex_state = 12},
  [1178] = {.lex_state = 27, .external_lex_state = 10},
  [1179] = {.lex_state = 26, .external_lex_state = 10},
  [1180] = {.lex_state = 26, .external_lex_state = 10},
  [1181] = {.lex_state = 26, .external_lex_state = 10},
  [1182] = {.lex_state = 26, .external_lex_state = 10},
  [1183] = {.lex_state = 26, .external_lex_state = 10},
  [1184] = {.lex_state = 10, .external_lex_state = 10},
  [1185] = {.lex_state = 26, .external_lex_state = 10},
  [1186] = {.lex_state = 26, .external_lex_state = 10},
  [1187] = {.lex_state = 8, .external_lex_state = 12},
  [1188] = {.lex_state = 8, .external_lex_state = 12},
  [1189] = {.lex_state = 8, .external_lex_state = 12},
  [1190] = {.lex_state = 26, .external_lex_state = 10},
  [1191] = {.lex_state = 10, .external_lex_state = 10},
  [1192] = {.lex_state = 27, .external_lex_state = 10},
  [1193] = {.lex_state = 27, .external_lex_state = 10},
  [1194] = {.lex_state = 24, .external_lex_state = 8},
  [1195] = {.lex_state = 8, .external_lex_state = 12},
  [1196] = {.lex_state = 10, .external_lex_state = 10},
  [1197] = {.lex_state = 26, .external_lex_state = 10},
  [1198] = {.lex_state = 27, .external_lex_state = 10},
  [1199] = {.lex_state = 26, .external_lex_state = 10},
  [1200] = {.lex_state = 26, .external_lex_state = 10},
  [1201] = {.lex_state = 27, .external_lex_state = 10},
  [1202] = {.lex_state = 26, .external_lex_state = 10},
  [1203] = {.lex_state = 26, .external_lex_state = 10},
  [1204] = {.lex_state = 26, .external_lex_state = 10},
  [1205] = {.lex_state = 27, .external_lex_state = 10},
  [1206] = {.lex_state = 27, .external_lex_state = 10},
  [1207] = {.lex_state = 26, .external_lex_state = 10},
  [1208] = {.lex_state = 27, .external_lex_state = 10},
  [1209] = {.lex_state = 27, .external_lex_state = 10},
  [1210] = {.lex_state = 27, .external_lex_state = 10},
  [1211] = {.lex_state = 26, .external_lex_state = 10},
  [1212] = {.lex_state = 27, .external_lex_state = 10},
  [1213] = {.lex_state = 27, .external_lex_state = 10},
  [1214] = {.lex_state = 26, .external_lex_state = 10},
  [1215] = {.lex_state = 27, .external_lex_state = 10},
  [1216] = {.lex_state = 25, .external_lex_state = 13},
  [1217] = {.lex_state = 27, .external_lex_state = 10},
  [1218] = {.lex_state = 26, .external_lex_state = 10},
  [1219] = {.lex_state = 26, .external_lex_state = 10},
  [1220] = {.lex_state = 26, .external_lex_state = 10},
  [1221] = {.lex_state = 25, .external_lex_state = 12},
  [1222] = {.lex_state = 26, .external_lex_state = 10},
  [1223] = {.lex_state = 26, .external_lex_state = 10},
  [1224] = {.lex_state = 8, .external_lex_state = 12},
  [1225] = {.lex_state = 26, .external_lex_state = 10},
  [1226] = {.lex_state = 26, .external_lex_state = 10},
  [1227] = {.lex_state = 25, .external_lex_state = 11},
  [1228] = {.lex_state = 27, .external_lex_state = 10},
  [1229] = {.lex_state = 8, .external_lex_state = 12},
  [1230] = {.lex_state = 10, .external_lex_state = 10},
  [1231] = {.lex_state = 25, .external_lex_state = 12},
  [1232] = {.lex_state = 25, .external_lex_state = 12},
  [1233] = {.lex_state = 25, .external_lex_state = 12},
  [1234] = {.lex_state = 8, .external_lex_state = 12},
  [1235] = {.lex_state = 27, .external_lex_state = 10},
  [1236] = {.lex_state = 8, .external_lex_state = 12},
  [1237] = {.lex_state = 8, .external_lex_state = 12},
  [1238] = {.lex_state = 8, .external_lex_state = 13},
  [1239] = {.lex_state = 26, .external_lex_state = 10},
  [1240] = {.lex_state = 8, .external_lex_state = 12},
  [1241] = {.lex_state = 8, .external_lex_state = 11},
  [1242] = {.lex_state = 25, .external_lex_state = 13},
  [1243] = {.lex_state = 25, .external_lex_state = 11},
  [1244] = {.lex_state = 26, .external_lex_state = 10},
  [1245] = {.lex_state = 27, .external_lex_state = 10},
  [1246] = {.lex_state = 27, .external_lex_state = 10},
  [1247] = {.lex_state = 8, .external_lex_state = 12},
  [1248] = {.lex_state = 10, .external_lex_state = 10},
  [1249] = {.lex_state = 8, .external_lex_state = 12},
  [1250] = {.lex_state = 27, .external_lex_state = 10},
  [1251] = {.lex_state = 27, .external_lex_state = 10},
  [1252] = {.lex_state = 8, .external_lex_state = 12},
  [1253] = {.lex_state = 25, .external_lex_state = 12},
  [1254] = {.lex_state = 25, .external_lex_state = 12},
  [1255] = {.lex_state = 32, .external_lex_state = 10},
  [1256] = {.lex_state = 32, .external_lex_state = 10},
  [1257] = {.lex_state = 25, .external_lex_state = 12},
  [1258] = {.lex_state = 25, .external_lex_state = 12},
  [1259] = {.lex_state = 25, .external_lex_state = 12},
  [1260] = {.lex_state = 25, .external_lex_state = 12},
  [1261] = {.lex_state = 8, .external_lex_state = 12},
  [1262] = {.lex_state = 26, .external_lex_state = 10},
  [1263] = {.lex_state = 25, .external_lex_state = 12},
  [1264] = {.lex_state = 26, .external_lex_state = 10},
  [1265] = {.lex_state = 25, .external_lex_state = 12},
  [1266] = {.lex_state = 25, .external_lex_state = 12},
  [1267] = {.lex_state = 25, .external_lex_state = 12},
  [1268] = {.lex_state = 25, .external_lex_state = 12},
  [1269] = {.lex_state = 25, .external_lex_state = 12},
  [1270] = {.lex_state = 25, .external_lex_state = 12},
  [1271] = {.lex_state = 25, .external_lex_state = 12},
  [1272] = {.lex_state = 25, .external_lex_state = 12},
  [1273] = {.lex_state = 26, .external_lex_state = 10},
  [1274] = {.lex_state = 26, .external_lex_state = 10},
  [1275] = {.lex_state = 26, .external_lex_state = 10},
  [1276] = {.lex_state = 32, .external_lex_state = 10},
  [1277] = {.lex_state = 26, .external_lex_state = 10},
  [1278] = {.lex_state = 8, .external_lex_state = 12},
  [1279] = {.lex_state = 8, .external_lex_state = 12},
  [1280] = {.lex_state = 8, .external_lex_state = 12},
  [1281] = {.lex_state = 8, .external_lex_state = 12},
  [1282] = {.lex_state = 26, .external_lex_state = 10},
  [1283] = {.lex_state = 8, .external_lex_state = 12},
  [1284] = {.lex_state = 26, .external_lex_state = 10},
  [1285] = {.lex_state = 31, .external_lex_state = 12},
  [1286] = {.lex_state = 10, .external_lex_state = 10},
  [1287] = {.lex_state = 24, .external_lex_state = 8},
  [1288] = {.lex_state = 9, .external_lex_state = 11},
  [1289] = {.lex_state = 142, .external_lex_state = 2},
  [1290] = {.lex_state = 142, .external_lex_state = 2},
  [1291] = {.lex_state = 9, .external_lex_state = 12},
  [1292] = {.lex_state = 25, .external_lex_state = 13},
  [1293] = {.lex_state = 142, .external_lex_state = 2},
  [1294] = {.lex_state = 9, .external_lex_state = 12},
  [1295] = {.lex_state = 142, .external_lex_state = 2},
  [1296] = {.lex_state = 9, .external_lex_state = 12},
  [1297] = {.lex_state = 25, .external_lex_state = 11},
  [1298] = {.lex_state = 9, .external_lex_state = 12},
  [1299] = {.lex_state = 8, .external_lex_state = 13},
  [1300] = {.lex_state = 142, .external_lex_state = 2},
  [1301] = {.lex_state = 9, .external_lex_state = 12},
  [1302] = {.lex_state = 10, .external_lex_state = 10},
  [1303] = {.lex_state = 27, .external_lex_state = 10},
  [1304] = {.lex_state = 9, .external_lex_state = 12},
  [1305] = {.lex_state = 9, .external_lex_state = 12},
  [1306] = {.lex_state = 10, .external_lex_state = 10},
  [1307] = {.lex_state = 9, .external_lex_state = 12},
  [1308] = {.lex_state = 10, .external_lex_state = 10},
  [1309] = {.lex_state = 27, .external_lex_state = 10},
  [1310] = {.lex_state = 9, .external_lex_state = 12},
  [1311] = {.lex_state = 31, .external_lex_state = 12},
  [1312] = {.lex_state = 27, .external_lex_state = 10},
  [1313] = {.lex_state = 27, .external_lex_state = 10},
  [1314] = {.lex_state = 24, .external_lex_state = 8},
  [1315] = {.lex_state = 9, .external_lex_state = 12},
  [1316] = {.lex_state = 9, .external_lex_state = 12},
  [1317] = {.lex_state = 25, .external_lex_state = 12},
  [1318] = {.lex_state = 25, .external_lex_state = 12},
  [1319] = {.lex_state = 9, .external_lex_state = 12},
  [1320] = {.lex_state = 142, .external_lex_state = 2},
  [1321] = {.lex_state = 142, .external_lex_state = 2},
  [1322] = {.lex_state = 27, .external_lex_state = 10},
  [1323] = {.lex_state = 142, .external_lex_state = 2},
  [1324] = {.lex_state = 10, .external_lex_state = 10},
  [1325] = {.lex_state = 9, .external_lex_state = 12},
  [1326] = {.lex_state = 9, .external_lex_state = 12},
  [1327] = {.lex_state = 10, .external_lex_state = 10},
  [1328] = {.lex_state = 10, .external_lex_state = 10},
  [1329] = {.lex_state = 25, .external_lex_state = 11},
  [1330] = {.lex_state = 10, .external_lex_state = 10},
  [1331] = {.lex_state = 10, .external_lex_state = 10},
  [1332] = {.lex_state = 142, .external_lex_state = 2},
  [1333] = {.lex_state = 142, .external_lex_state = 2},
  [1334] = {.lex_state = 25, .external_lex_state = 12},
  [1335] = {.lex_state = 25, .external_lex_state = 11},
  [1336] = {.lex_state = 10, .external_lex_state = 10},
  [1337] = {.lex_state = 9, .external_lex_state = 12},
  [1338] = {.lex_state = 9, .external_lex_state = 12},
  [1339] = {.lex_state = 9, .external_lex_state = 12},
  [1340] = {.lex_state = 9, .external_lex_state = 12},
  [1341] = {.lex_state = 8, .external_lex_state = 13},
  [1342] = {.lex_state = 142, .external_lex_state = 2},
  [1343] = {.lex_state = 31, .external_lex_state = 12},
  [1344] = {.lex_state = 10, .external_lex_state = 10},
  [1345] = {.lex_state = 10, .external_lex_state = 10},
  [1346] = {.lex_state = 10, .external_lex_state = 10},
  [1347] = {.lex_state = 9, .external_lex_state = 11},
  [1348] = {.lex_state = 27, .external_lex_state = 10},
  [1349] = {.lex_state = 142, .external_lex_state = 2},
  [1350] = {.lex_state = 10, .external_lex_state = 10},
  [1351] = {.lex_state = 10, .external_lex_state = 10},
  [1352] = {.lex_state = 10, .external_lex_state = 10},
  [1353] = {.lex_state = 27, .external_lex_state = 10},
  [1354] = {.lex_state = 9, .external_lex_state = 12},
  [1355] = {.lex_state = 10, .external_lex_state = 10},
  [1356] = {.lex_state = 9, .external_lex_state = 12},
  [1357] = {.lex_state = 9, .external_lex_state = 12},
  [1358] = {.lex_state = 25, .external_lex_state = 12},
  [1359] = {.lex_state = 9, .external_lex_state = 10},
  [1360] = {.lex_state = 9, .external_lex_state = 12},
  [1361] = {.lex_state = 9, .external_lex_state = 10},
  [1362] = {.lex_state = 25, .external_lex_state = 12},
  [1363] = {.lex_state = 9, .external_lex_state = 10},
  [1364] = {.lex_state = 9, .external_lex_state = 10},
  [1365] = {.lex_state = 9, .external_lex_state = 12},
  [1366] = {.lex_state = 9, .external_lex_state = 12},
  [1367] = {.lex_state = 25, .external_lex_state = 12},
  [1368] = {.lex_state = 9, .external_lex_state = 10},
  [1369] = {.lex_state = 9, .external_lex_state = 11},
  [1370] = {.lex_state = 25, .external_lex_state = 12},
  [1371] = {.lex_state = 25, .external_lex_state = 12},
  [1372] = {.lex_state = 9, .external_lex_state = 10},
  [1373] = {.lex_state = 25, .external_lex_state = 12},
  [1374] = {.lex_state = 14, .external_lex_state = 12},
  [1375] = {.lex_state = 14, .external_lex_state = 12},
  [1376] = {.lex_state = 9, .external_lex_state = 10},
  [1377] = {.lex_state = 9, .external_lex_state = 10},
  [1378] = {.lex_state = 9, .external_lex_state = 10},
  [1379] = {.lex_state = 9, .external_lex_state = 10},
  [1380] = {.lex_state = 9, .external_lex_state = 12},
  [1381] = {.lex_state = 142, .external_lex_state = 2},
  [1382] = {.lex_state = 8, .external_lex_state = 13},
  [1383] = {.lex_state = 142, .external_lex_state = 2},
  [1384] = {.lex_state = 9, .external_lex_state = 11},
  [1385] = {.lex_state = 9, .external_lex_state = 10},
  [1386] = {.lex_state = 9, .external_lex_state = 10},
  [1387] = {.lex_state = 9, .external_lex_state = 10},
  [1388] = {.lex_state = 9, .external_lex_state = 12},
  [1389] = {.lex_state = 9, .external_lex_state = 10},
  [1390] = {.lex_state = 9, .external_lex_state = 10},
  [1391] = {.lex_state = 9, .external_lex_state = 10},
  [1392] = {.lex_state = 142, .external_lex_state = 2},
  [1393] = {.lex_state = 9, .external_lex_state = 10},
  [1394] = {.lex_state = 9, .external_lex_state = 12},
  [1395] = {.lex_state = 9, .external_lex_state = 12},
  [1396] = {.lex_state = 142, .external_lex_state = 2},
  [1397] = {.lex_state = 25, .external_lex_state = 12},
  [1398] = {.lex_state = 9, .external_lex_state = 12},
  [1399] = {.lex_state = 9, .external_lex_state = 12},
  [1400] = {.lex_state = 9, .external_lex_state = 12},
  [1401] = {.lex_state = 25, .external_lex_state = 12},
  [1402] = {.lex_state = 25, .external_lex_state = 12},
  [1403] = {.lex_state = 9, .external_lex_state = 12},
  [1404] = {.lex_state = 9, .external_lex_state = 12},
  [1405] = {.lex_state = 9, .external_lex_state = 12},
  [1406] = {.lex_state = 9, .external_lex_state = 10},
  [1407] = {.lex_state = 26, .external_lex_state = 10},
  [1408] = {.lex_state = 9, .external_lex_state = 12},
  [1409] = {.lex_state = 9, .external_lex_state = 12},
  [1410] = {.lex_state = 9, .external_lex_state = 12},
  [1411] = {.lex_state = 25, .external_lex_state = 12},
  [1412] = {.lex_state = 9, .external_lex_state = 12},
  [1413] = {.lex_state = 25, .external_lex_state = 12},
  [1414] = {.lex_state = 25, .external_lex_state = 12},
  [1415] = {.lex_state = 25, .external_lex_state = 12},
  [1416] = {.lex_state = 25, .external_lex_state = 12},
  [1417] = {.lex_state = 25, .external_lex_state = 12},
  [1418] = {.lex_state = 25, .external_lex_state = 12},
  [1419] = {.lex_state = 25, .external_lex_state = 12},
  [1420] = {.lex_state = 25, .external_lex_state = 12},
  [1421] = {.lex_state = 25, .external_lex_state = 12},
  [1422] = {.lex_state = 25, .external_lex_state = 12},
  [1423] = {.lex_state = 25, .external_lex_state = 12},
  [1424] = {.lex_state = 25, .external_lex_state = 12},
  [1425] = {.lex_state = 25, .external_lex_state = 12},
  [1426] = {.lex_state = 25, .external_lex_state = 12},
  [1427] = {.lex_state = 25, .external_lex_state = 12},
  [1428] = {.lex_state = 25, .external_lex_state = 12},
  [1429] = {.lex_state = 14, .external_lex_state = 12},
  [1430] = {.lex_state = 9, .external_lex_state = 12},
  [1431] = {.lex_state = 9, .external_lex_state = 10},
  [1432] = {.lex_state = 26, .external_lex_state = 10},
  [1433] = {.lex_state = 9, .external_lex_state = 10},
  [1434] = {.lex_state = 9, .external_lex_state = 10},
  [1435] = {.lex_state = 26, .external_lex_state = 10},
  [1436] = {.lex_state = 26, .external_lex_state = 10},
  [1437] = {.lex_state = 26, .external_lex_state = 10},
  [1438] = {.lex_state = 26, .external_lex_state = 10},
  [1439] = {.lex_state = 26, .external_lex_state = 10},
  [1440] = {.lex_state = 26, .external_lex_state = 10},
  [1441] = {.lex_state = 26, .external_lex_state = 10},
  [1442] = {.lex_state = 9, .external_lex_state = 12},
  [1443] = {.lex_state = 9, .external_lex_state = 12},
  [1444] = {.lex_state = 9, .external_lex_state = 10},
  [1445] = {.lex_state = 9, .external_lex_state = 10},
  [1446] = {.lex_state = 9, .external_lex_state = 10},
  [1447] = {.lex_state = 9, .external_lex_state = 10},
  [1448] = {.lex_state = 9, .external_lex_state = 10},
  [1449] = {.lex_state = 9, .external_lex_state = 10},
  [1450] = {.lex_state = 9, .external_lex_state = 10},
  [1451] = {.lex_state = 9, .external_lex_state = 10},
  [1452] = {.lex_state = 9, .external_lex_state = 10},
  [1453] = {.lex_state = 9, .external_lex_state = 10},
  [1454] = {.lex_state = 9, .external_lex_state = 10},
  [1455] = {.lex_state = 9, .external_lex_state = 10},
  [1456] = {.lex_state = 9, .external_lex_state = 12},
  [1457] = {.lex_state = 26, .external_lex_state = 10},
  [1458] = {.lex_state = 25, .external_lex_state = 12},
  [1459] = {.lex_state = 25, .external_lex_state = 12},
  [1460] = {.lex_state = 25, .external_lex_state = 12},
  [1461] = {.lex_state = 9, .external_lex_state = 12},
  [1462] = {.lex_state = 9, .external_lex_state = 12},
  [1463] = {.lex_state = 9, .external_lex_state = 10},
  [1464] = {.lex_state = 9, .external_lex_state = 10},
  [1465] = {.lex_state = 9, .external_lex_state = 12},
  [1466] = {.lex_state = 9, .external_lex_state = 10},
  [1467] = {.lex_state = 9, .external_lex_state = 12},
  [1468] = {.lex_state = 9, .external_lex_state = 10},
  [1469] = {.lex_state = 9, .external_lex_state = 12},
  [1470] = {.lex_state = 9, .external_lex_state = 10},
  [1471] = {.lex_state = 25, .external_lex_state = 12},
  [1472] = {.lex_state = 25, .external_lex_state = 12},
  [1473] = {.lex_state = 25, .external_lex_state = 12},
  [1474] = {.lex_state = 25, .external_lex_state = 12},
  [1475] = {.lex_state = 25, .external_lex_state = 12},
  [1476] = {.lex_state = 25, .external_lex_state = 12},
  [1477] = {.lex_state = 9, .external_lex_state = 12},
  [1478] = {.lex_state = 9, .external_lex_state = 12},
  [1479] = {.lex_state = 9, .external_lex_state = 10},
  [1480] = {.lex_state = 9, .external_lex_state = 12},
  [1481] = {.lex_state = 9, .external_lex_state = 10},
  [1482] = {.lex_state = 25, .external_lex_state = 12},
  [1483] = {.lex_state = 9, .external_lex_state = 12},
  [1484] = {.lex_state = 9, .external_lex_state = 12},
  [1485] = {.lex_state = 9, .external_lex_state = 10},
  [1486] = {.lex_state = 9, .external_lex_state = 12},
  [1487] = {.lex_state = 9, .external_lex_state = 12},
  [1488] = {.lex_state = 9, .external_lex_state = 10},
  [1489] = {.lex_state = 9, .external_lex_state = 12},
  [1490] = {.lex_state = 9, .external_lex_state = 12},
  [1491] = {.lex_state = 9, .external_lex_state = 12},
  [1492] = {.lex_state = 30},
  [1493] = {.lex_state = 25, .external_lex_state = 12},
  [1494] = {.lex_state = 9, .external_lex_state = 12},
  [1495] = {.lex_state = 9, .external_lex_state = 12},
  [1496] = {.lex_state = 9, .external_lex_state = 12},
  [1497] = {.lex_state = 142, .external_lex_state = 2},
  [1498] = {.lex_state = 142, .external_lex_state = 2},
  [1499] = {.lex_state = 142, .external_lex_state = 2},
  [1500] = {.lex_state = 142, .external_lex_state = 2},
  [1501] = {.lex_state = 142, .external_lex_state = 2},
  [1502] = {.lex_state = 142, .external_lex_state = 2},
  [1503] = {.lex_state = 25, .external_lex_state = 12},
  [1504] = {.lex_state = 142, .external_lex_state = 2},
  [1505] = {.lex_state = 9, .external_lex_state = 12},
  [1506] = {.lex_state = 142, .external_lex_state = 2},
  [1507] = {.lex_state = 9, .external_lex_state = 12},
  [1508] = {.lex_state = 9, .external_lex_state = 12},
  [1509] = {.lex_state = 142, .external_lex_state = 2},
  [1510] = {.lex_state = 25, .external_lex_state = 12},
  [1511] = {.lex_state = 142, .external_lex_state = 2},
  [1512] = {.lex_state = 25, .external_lex_state = 12},
  [1513] = {.lex_state = 9, .external_lex_state = 12},
  [1514] = {.lex_state = 9, .external_lex_state = 12},
  [1515] = {.lex_state = 9, .external_lex_state = 12},
  [1516] = {.lex_state = 9, .external_lex_state = 12},
  [1517] = {.lex_state = 142, .external_lex_state = 2},
  [1518] = {.lex_state = 142, .external_lex_state = 2},
  [1519] = {.lex_state = 142, .external_lex_state = 2},
  [1520] = {.lex_state = 142, .external_lex_state = 2},
  [1521] = {.lex_state = 142, .external_lex_state = 2},
  [1522] = {.lex_state = 142, .external_lex_state = 2},
  [1523] = {.lex_state = 142, .external_lex_state = 2},
  [1524] = {.lex_state = 142, .external_lex_state = 2},
  [1525] = {.lex_state = 142, .external_lex_state = 2},
  [1526] = {.lex_state = 142, .external_lex_state = 2},
  [1527] = {.lex_state = 142, .external_lex_state = 2},
  [1528] = {.lex_state = 142, .external_lex_state = 2},
  [1529] = {.lex_state = 142, .external_lex_state = 2},
  [1530] = {.lex_state = 142, .external_lex_state = 2},
  [1531] = {.lex_state = 142, .external_lex_state = 2},
  [1532] = {.lex_state = 142, .external_lex_state = 2},
  [1533] = {.lex_state = 9, .external_lex_state = 12},
  [1534] = {.lex_state = 142, .external_lex_state = 2},
  [1535] = {.lex_state = 142, .external_lex_state = 2},
  [1536] = {.lex_state = 9, .external_lex_state = 12},
  [1537] = {.lex_state = 142, .external_lex_state = 2},
  [1538] = {.lex_state = 142, .external_lex_state = 2},
  [1539] = {.lex_state = 142, .external_lex_state = 2},
  [1540] = {.lex_state = 142, .external_lex_state = 2},
  [1541] = {.lex_state = 142, .external_lex_state = 2},
  [1542] = {.lex_state = 142, .external_lex_state = 2},
  [1543] = {.lex_state = 142, .external_lex_state = 2},
  [1544] = {.lex_state = 142, .external_lex_state = 2},
  [1545] = {.lex_state = 142, .external_lex_state = 2},
  [1546] = {.lex_state = 142, .external_lex_state = 2},
  [1547] = {.lex_state = 142, .external_lex_state = 2},
  [1548] = {.lex_state = 142, .external_lex_state = 2},
  [1549] = {.lex_state = 142, .external_lex_state = 2},
  [1550] = {.lex_state = 142, .external_lex_state = 2},
  [1551] = {.lex_state = 142, .external_lex_state = 2},
  [1552] = {.lex_state = 142, .external_lex_state = 2},
  [1553] = {.lex_state = 142, .external_lex_state = 2},
  [1554] = {.lex_state = 142, .external_lex_state = 2},
  [1555] = {.lex_state = 142, .external_lex_state = 2},
  [1556] = {.lex_state = 142, .external_lex_state = 2},
  [1557] = {.lex_state = 142, .external_lex_state = 2},
  [1558] = {.lex_state = 142, .external_lex_state = 2},
  [1559] = {.lex_state = 142, .external_lex_state = 2},
  [1560] = {.lex_state = 142, .external_lex_state = 2},
  [1561] = {.lex_state = 142, .external_lex_state = 2},
  [1562] = {.lex_state = 142, .external_lex_state = 2},
  [1563] = {.lex_state = 142, .external_lex_state = 2},
  [1564] = {.lex_state = 142, .external_lex_state = 2},
  [1565] = {.lex_state = 142, .external_lex_state = 2},
  [1566] = {.lex_state = 142, .external_lex_state = 2},
  [1567] = {.lex_state = 142, .external_lex_state = 2},
  [1568] = {.lex_state = 142, .external_lex_state = 2},
  [1569] = {.lex_state = 142, .external_lex_state = 2},
  [1570] = {.lex_state = 142, .external_lex_state = 2},
  [1571] = {.lex_state = 142, .external_lex_state = 2},
  [1572] = {.lex_state = 142, .external_lex_state = 2},
  [1573] = {.lex_state = 142, .external_lex_state = 2},
  [1574] = {.lex_state = 142, .external_lex_state = 2},
  [1575] = {.lex_state = 142, .external_lex_state = 2},
  [1576] = {.lex_state = 142, .external_lex_state = 2},
  [1577] = {.lex_state = 142, .external_lex_state = 2},
  [1578] = {.lex_state = 142, .external_lex_state = 2},
  [1579] = {.lex_state = 24, .external_lex_state = 8},
  [1580] = {.lex_state = 142, .external_lex_state = 2},
  [1581] = {.lex_state = 142, .external_lex_state = 2},
  [1582] = {.lex_state = 142, .external_lex_state = 2},
  [1583] = {.lex_state = 142, .external_lex_state = 2},
  [1584] = {.lex_state = 142, .external_lex_state = 2},
  [1585] = {.lex_state = 142, .external_lex_state = 2},
  [1586] = {.lex_state = 142, .external_lex_state = 2},
  [1587] = {.lex_state = 142, .external_lex_state = 2},
  [1588] = {.lex_state = 142, .external_lex_state = 2},
  [1589] = {.lex_state = 142, .external_lex_state = 2},
  [1590] = {.lex_state = 142, .external_lex_state = 2},
  [1591] = {.lex_state = 142, .external_lex_state = 2},
  [1592] = {.lex_state = 142, .external_lex_state = 2},
  [1593] = {.lex_state = 142, .external_lex_state = 2},
  [1594] = {.lex_state = 142, .external_lex_state = 2},
  [1595] = {.lex_state = 142, .external_lex_state = 2},
  [1596] = {.lex_state = 30},
  [1597] = {.lex_state = 30},
  [1598] = {.lex_state = 25, .external_lex_state = 12},
  [1599] = {.lex_state = 30},
  [1600] = {.lex_state = 30},
  [1601] = {.lex_state = 47, .external_lex_state = 2},
  [1602] = {.lex_state = 142, .external_lex_state = 2},
  [1603] = {.lex_state = 30},
  [1604] = {.lex_state = 142, .external_lex_state = 2},
  [1605] = {.lex_state = 30},
  [1606] = {.lex_state = 142, .external_lex_state = 2},
  [1607] = {.lex_state = 30},
  [1608] = {.lex_state = 30},
  [1609] = {.lex_state = 41},
  [1610] = {.lex_state = 41},
  [1611] = {.lex_state = 41},
  [1612] = {.lex_state = 41},
  [1613] = {.lex_state = 41},
  [1614] = {.lex_state = 30},
  [1615] = {.lex_state = 41},
  [1616] = {.lex_state = 13, .external_lex_state = 14},
  [1617] = {.lex_state = 31},
  [1618] = {.lex_state = 31},
  [1619] = {.lex_state = 31},
  [1620] = {.lex_state = 31},
  [1621] = {.lex_state = 31},
  [1622] = {.lex_state = 31},
  [1623] = {.lex_state = 31},
  [1624] = {.lex_state = 31},
  [1625] = {.lex_state = 25},
  [1626] = {.lex_state = 25},
  [1627] = {.lex_state = 41},
  [1628] = {.lex_state = 25},
  [1629] = {.lex_state = 25},
  [1630] = {.lex_state = 25},
  [1631] = {.lex_state = 25},
  [1632] = {.lex_state = 41},
  [1633] = {.lex_state = 41},
  [1634] = {.lex_state = 41},
  [1635] = {.lex_state = 41},
  [1636] = {.lex_state = 25},
  [1637] = {.lex_state = 41},
  [1638] = {.lex_state = 41},
  [1639] = {.lex_state = 41},
  [1640] = {.lex_state = 41},
  [1641] = {.lex_state = 41},
  [1642] = {.lex_state = 41},
  [1643] = {.lex_state = 31},
  [1644] = {.lex_state = 25},
  [1645] = {.lex_state = 25},
  [1646] = {.lex_state = 25},
  [1647] = {.lex_state = 31},
  [1648] = {.lex_state = 25},
  [1649] = {.lex_state = 30},
  [1650] = {.lex_state = 31},
  [1651] = {.lex_state = 30},
  [1652] = {.lex_state = 31},
  [1653] = {.lex_state = 31},
  [1654] = {.lex_state = 31},
  [1655] = {.lex_state = 25},
  [1656] = {.lex_state = 25},
  [1657] = {.lex_state = 31},
  [1658] = {.lex_state = 25},
  [1659] = {.lex_state = 25},
  [1660] = {.lex_state = 31},
  [1661] = {.lex_state = 41},
  [1662] = {.lex_state = 41},
  [1663] = {.lex_state = 41},
  [1664] = {.lex_state = 41},
  [1665] = {.lex_state = 41},
  [1666] = {.lex_state = 41},
  [1667] = {.lex_state = 41},
  [1668] = {.lex_state = 41},
  [1669] = {.lex_state = 41},
  [1670] = {.lex_state = 41},
  [1671] = {.lex_state = 41},
  [1672] = {.lex_state = 41},
  [1673] = {.lex_state = 41},
  [1674] = {.lex_state = 41},
  [1675] = {.lex_state = 41},
  [1676] = {.lex_state = 41},
  [1677] = {.lex_state = 41},
  [1678] = {.lex_state = 25},
  [1679] = {.lex_state = 41},
  [1680] = {.lex_state = 41},
  [1681] = {.lex_state = 28},
  [1682] = {.lex_state = 41},
  [1683] = {.lex_state = 41},
  [1684] = {.lex_state = 41},
  [1685] = {.lex_state = 41},
  [1686] = {.lex_state = 28},
  [1687] = {.lex_state = 28},
  [1688] = {.lex_state = 41},
  [1689] = {.lex_state = 25},
  [1690] = {.lex_state = 25},
  [1691] = {.lex_state = 28},
  [1692] = {.lex_state = 25},
  [1693] = {.lex_state = 41},
  [1694] = {.lex_state = 41},
  [1695] = {.lex_state = 25},
  [1696] = {.lex_state = 25},
  [1697] = {.lex_state = 25},
  [1698] = {.lex_state = 41},
  [1699] = {.lex_state = 28},
  [1700] = {.lex_state = 28},
  [1701] = {.lex_state = 28},
  [1702] = {.lex_state = 28},
  [1703] = {.lex_state = 41},
  [1704] = {.lex_state = 41},
  [1705] = {.lex_state = 28},
  [1706] = {.lex_state = 28},
  [1707] = {.lex_state = 41},
  [1708] = {.lex_state = 28},
  [1709] = {.lex_state = 41},
  [1710] = {.lex_state = 28},
  [1711] = {.lex_state = 28},
  [1712] = {.lex_state = 28},
  [1713] = {.lex_state = 28},
  [1714] = {.lex_state = 41},
  [1715] = {.lex_state = 41},
  [1716] = {.lex_state = 28},
  [1717] = {.lex_state = 28},
  [1718] = {.lex_state = 41},
  [1719] = {.lex_state = 41},
  [1720] = {.lex_state = 41},
  [1721] = {.lex_state = 41},
  [1722] = {.lex_state = 41},
  [1723] = {.lex_state = 41},
  [1724] = {.lex_state = 28},
  [1725] = {.lex_state = 25},
  [1726] = {.lex_state = 28},
  [1727] = {.lex_state = 41},
  [1728] = {.lex_state = 41},
  [1729] = {.lex_state = 41},
  [1730] = {.lex_state = 41},
  [1731] = {.lex_state = 41},
  [1732] = {.lex_state = 41},
  [1733] = {.lex_state = 30},
  [1734] = {.lex_state = 41},
  [1735] = {.lex_state = 41},
  [1736] = {.lex_state = 41},
  [1737] = {.lex_state = 41},
  [1738] = {.lex_state = 41},
  [1739] = {.lex_state = 41},
  [1740] = {.lex_state = 41},
  [1741] = {.lex_state = 41},
  [1742] = {.lex_state = 41},
  [1743] = {.lex_state = 41},
  [1744] = {.lex_state = 28},
  [1745] = {.lex_state = 25},
  [1746] = {.lex_state = 25},
  [1747] = {.lex_state = 25, .external_lex_state = 15},
  [1748] = {.lex_state = 31},
  [1749] = {.lex_state = 41},
  [1750] = {.lex_state = 25},
  [1751] = {.lex_state = 33},
  [1752] = {.lex_state = 25},
  [1753] = {.lex_state = 31},
  [1754] = {.lex_state = 41},
  [1755] = {.lex_state = 41},
  [1756] = {.lex_state = 33},
  [1757] = {.lex_state = 41},
  [1758] = {.lex_state = 41},
  [1759] = {.lex_state = 41},
  [1760] = {.lex_state = 41},
  [1761] = {.lex_state = 41},
  [1762] = {.lex_state = 41},
  [1763] = {.lex_state = 31},
  [1764] = {.lex_state = 41},
  [1765] = {.lex_state = 41},
  [1766] = {.lex_state = 41},
  [1767] = {.lex_state = 34, .external_lex_state = 16},
  [1768] = {.lex_state = 41},
  [1769] = {.lex_state = 41},
  [1770] = {.lex_state = 41},
  [1771] = {.lex_state = 41},
  [1772] = {.lex_state = 25, .external_lex_state = 15},
  [1773] = {.lex_state = 16},
  [1774] = {.lex_state = 16},
  [1775] = {.lex_state = 16},
  [1776] = {.lex_state = 25},
  [1777] = {.lex_state = 34, .external_lex_state = 16},
  [1778] = {.lex_state = 40, .external_lex_state = 17},
  [1779] = {.lex_state = 40, .external_lex_state = 17},
  [1780] = {.lex_state = 25},
  [1781] = {.lex_state = 40, .external_lex_state = 17},
  [1782] = {.lex_state = 40, .external_lex_state = 17},
  [1783] = {.lex_state = 40, .external_lex_state = 17},
  [1784] = {.lex_state = 40, .external_lex_state = 17},
  [1785] = {.lex_state = 18},
  [1786] = {.lex_state = 18},
  [1787] = {.lex_state = 40, .external_lex_state = 17},
  [1788] = {.lex_state = 40, .external_lex_state = 17},
  [1789] = {.lex_state = 40, .external_lex_state = 17},
  [1790] = {.lex_state = 40, .external_lex_state = 17},
  [1791] = {.lex_state = 16},
  [1792] = {.lex_state = 41},
  [1793] = {.lex_state = 40, .external_lex_state = 17},
  [1794] = {.lex_state = 16},
  [1795] = {.lex_state = 25},
  [1796] = {.lex_state = 40, .external_lex_state = 17},
  [1797] = {.lex_state = 40, .external_lex_state = 17},
  [1798] = {.lex_state = 40, .external_lex_state = 17},
  [1799] = {.lex_state = 31},
  [1800] = {.lex_state = 40, .external_lex_state = 17},
  [1801] = {.lex_state = 40, .external_lex_state = 17},
  [1802] = {.lex_state = 40, .external_lex_state = 17},
  [1803] = {.lex_state = 30},
  [1804] = {.lex_state = 40, .external_lex_state = 17},
  [1805] = {.lex_state = 41},
  [1806] = {.lex_state = 40, .external_lex_state = 17},
  [1807] = {.lex_state = 41},
  [1808] = {.lex_state = 40, .external_lex_state = 17},
  [1809] = {.lex_state = 40, .external_lex_state = 17},
  [1810] = {.lex_state = 41},
  [1811] = {.lex_state = 40, .external_lex_state = 17},
  [1812] = {.lex_state = 30},
  [1813] = {.lex_state = 16},
  [1814] = {.lex_state = 40, .external_lex_state = 17},
  [1815] = {.lex_state = 25},
  [1816] = {.lex_state = 40, .external_lex_state = 17},
  [1817] = {.lex_state = 41},
  [1818] = {.lex_state = 40, .external_lex_state = 17},
  [1819] = {.lex_state = 16},
  [1820] = {.lex_state = 31},
  [1821] = {.lex_state = 16},
  [1822] = {.lex_state = 16},
  [1823] = {.lex_state = 18},
  [1824] = {.lex_state = 18},
  [1825] = {.lex_state = 18},
  [1826] = {.lex_state = 16},
  [1827] = {.lex_state = 16},
  [1828] = {.lex_state = 17},
  [1829] = {.lex_state = 25},
  [1830] = {.lex_state = 17},
  [1831] = {.lex_state = 16},
  [1832] = {.lex_state = 16},
  [1833] = {.lex_state = 17},
  [1834] = {.lex_state = 16},
  [1835] = {.lex_state = 16},
  [1836] = {.lex_state = 25},
  [1837] = {.lex_state = 25},
  [1838] = {.lex_state = 17},
  [1839] = {.lex_state = 25},
  [1840] = {.lex_state = 25},
  [1841] = {.lex_state = 25},
  [1842] = {.lex_state = 17},
  [1843] = {.lex_state = 17},
  [1844] = {.lex_state = 16},
  [1845] = {.lex_state = 16},
  [1846] = {.lex_state = 25},
  [1847] = {.lex_state = 25},
  [1848] = {.lex_state = 17},
  [1849] = {.lex_state = 25},
  [1850] = {.lex_state = 25},
  [1851] = {.lex_state = 25},
  [1852] = {.lex_state = 25},
  [1853] = {.lex_state = 25},
  [1854] = {.lex_state = 17},
  [1855] = {.lex_state = 25},
  [1856] = {.lex_state = 25},
  [1857] = {.lex_state = 25},
  [1858] = {.lex_state = 25},
  [1859] = {.lex_state = 25},
  [1860] = {.lex_state = 25},
  [1861] = {.lex_state = 25},
  [1862] = {.lex_state = 25},
  [1863] = {.lex_state = 25},
  [1864] = {.lex_state = 16},
  [1865] = {.lex_state = 31},
  [1866] = {.lex_state = 17},
  [1867] = {.lex_state = 17},
  [1868] = {.lex_state = 22},
  [1869] = {.lex_state = 17},
  [1870] = {.lex_state = 17},
  [1871] = {.lex_state = 18},
  [1872] = {.lex_state = 30},
  [1873] = {.lex_state = 17},
  [1874] = {.lex_state = 17},
  [1875] = {.lex_state = 16},
  [1876] = {.lex_state = 17},
  [1877] = {.lex_state = 17},
  [1878] = {.lex_state = 16},
  [1879] = {.lex_state = 25},
  [1880] = {.lex_state = 25},
  [1881] = {.lex_state = 25},
  [1882] = {.lex_state = 16},
  [1883] = {.lex_state = 16},
  [1884] = {.lex_state = 17},
  [1885] = {.lex_state = 16},
  [1886] = {.lex_state = 31},
  [1887] = {.lex_state = 18},
  [1888] = {.lex_state = 30},
  [1889] = {.lex_state = 16},
  [1890] = {.lex_state = 17},
  [1891] = {.lex_state = 25},
  [1892] = {.lex_state = 25},
  [1893] = {.lex_state = 30},
  [1894] = {.lex_state = 25},
  [1895] = {.lex_state = 25},
  [1896] = {.lex_state = 31},
  [1897] = {.lex_state = 16},
  [1898] = {.lex_state = 16},
  [1899] = {.lex_state = 16},
  [1900] = {.lex_state = 16},
  [1901] = {.lex_state = 16},
  [1902] = {.lex_state = 16},
  [1903] = {.lex_state = 16},
  [1904] = {.lex_state = 16},
  [1905] = {.lex_state = 16},
  [1906] = {.lex_state = 16},
  [1907] = {.lex_state = 16},
  [1908] = {.lex_state = 16},
  [1909] = {.lex_state = 16},
  [1910] = {.lex_state = 16},
  [1911] = {.lex_state = 16},
  [1912] = {.lex_state = 16},
  [1913] = {.lex_state = 25},
  [1914] = {.lex_state = 25},
  [1915] = {.lex_state = 25},
  [1916] = {.lex_state = 16},
  [1917] = {.lex_state = 16},
  [1918] = {.lex_state = 31},
  [1919] = {.lex_state = 16},
  [1920] = {.lex_state = 16},
  [1921] = {.lex_state = 16},
  [1922] = {.lex_state = 25},
  [1923] = {.lex_state = 25},
  [1924] = {.lex_state = 17},
  [1925] = {.lex_state = 16},
  [1926] = {.lex_state = 25},
  [1927] = {.lex_state = 17},
  [1928] = {.lex_state = 25},
  [1929] = {.lex_state = 25},
  [1930] = {.lex_state = 25},
  [1931] = {.lex_state = 47},
  [1932] = {.lex_state = 142},
  [1933] = {.lex_state = 25, .external_lex_state = 16},
  [1934] = {.lex_state = 47},
  [1935] = {.lex_state = 47},
  [1936] = {.lex_state = 47},
  [1937] = {.lex_state = 25},
  [1938] = {.lex_state = 40, .external_lex_state = 17},
  [1939] = {.lex_state = 47},
  [1940] = {.lex_state = 142},
  [1941] = {.lex_state = 142},
  [1942] = {.lex_state = 47},
  [1943] = {.lex_state = 25},
  [1944] = {.lex_state = 25},
  [1945] = {.lex_state = 25},
  [1946] = {.lex_state = 47},
  [1947] = {.lex_state = 47},
  [1948] = {.lex_state = 25},
  [1949] = {.lex_state = 25},
  [1950] = {.lex_state = 16},
  [1951] = {.lex_state = 47},
  [1952] = {.lex_state = 25},
  [1953] = {.lex_state = 25},
  [1954] = {.lex_state = 25},
  [1955] = {.lex_state = 25},
  [1956] = {.lex_state = 25},
  [1957] = {.lex_state = 25},
  [1958] = {.lex_state = 25},
  [1959] = {.lex_state = 25},
  [1960] = {.lex_state = 25},
  [1961] = {.lex_state = 25},
  [1962] = {.lex_state = 47},
  [1963] = {.lex_state = 47},
  [1964] = {.lex_state = 16},
  [1965] = {.lex_state = 16},
  [1966] = {.lex_state = 47},
  [1967] = {.lex_state = 47},
  [1968] = {.lex_state = 47},
  [1969] = {.lex_state = 47},
  [1970] = {.lex_state = 25},
  [1971] = {.lex_state = 47},
  [1972] = {.lex_state = 47},
  [1973] = {.lex_state = 16},
  [1974] = {.lex_state = 25},
  [1975] = {.lex_state = 47},
  [1976] = {.lex_state = 47},
  [1977] = {.lex_state = 25},
  [1978] = {.lex_state = 25},
  [1979] = {.lex_state = 25},
  [1980] = {.lex_state = 25},
  [1981] = {.lex_state = 25},
  [1982] = {.lex_state = 25},
  [1983] = {.lex_state = 25},
  [1984] = {.lex_state = 47},
  [1985] = {.lex_state = 47},
  [1986] = {.lex_state = 16},
  [1987] = {.lex_state = 25},
  [1988] = {.lex_state = 25},
  [1989] = {.lex_state = 25},
  [1990] = {.lex_state = 31},
  [1991] = {.lex_state = 46, .external_lex_state = 16},
  [1992] = {.lex_state = 25},
  [1993] = {.lex_state = 25},
  [1994] = {.lex_state = 142},
  [1995] = {.lex_state = 25},
  [1996] = {.lex_state = 25},
  [1997] = {.lex_state = 25},
  [1998] = {.lex_state = 25},
  [1999] = {.lex_state = 25},
  [2000] = {.lex_state = 25},
  [2001] = {.lex_state = 25},
  [2002] = {.lex_state = 25},
  [2003] = {.lex_state = 46, .external_lex_state = 16},
  [2004] = {.lex_state = 31},
  [2005] = {.lex_state = 25},
  [2006] = {.lex_state = 47},
  [2007] = {.lex_state = 30},
  [2008] = {.lex_state = 25},
  [2009] = {.lex_state = 25},
  [2010] = {.lex_state = 25},
  [2011] = {.lex_state = 25},
  [2012] = {.lex_state = 25},
  [2013] = {.lex_state = 30},
  [2014] = {.lex_state = 47},
  [2015] = {.lex_state = 25},
  [2016] = {.lex_state = 47},
  [2017] = {.lex_state = 47},
  [2018] = {.lex_state = 47},
  [2019] = {.lex_state = 25},
  [2020] = {.lex_state = 25},
  [2021] = {.lex_state = 25},
  [2022] = {.lex_state = 47},
  [2023] = {.lex_state = 46, .external_lex_state = 16},
  [2024] = {.lex_state = 47},
  [2025] = {.lex_state = 46, .external_lex_state = 16},
  [2026] = {.lex_state = 46, .external_lex_state = 16},
  [2027] = {.lex_state = 47},
  [2028] = {.lex_state = 47},
  [2029] = {.lex_state = 47},
  [2030] = {.lex_state = 16},
  [2031] = {.lex_state = 16},
  [2032] = {.lex_state = 16},
  [2033] = {.lex_state = 18},
  [2034] = {.lex_state = 16},
  [2035] = {.lex_state = 142},
  [2036] = {.lex_state = 40, .external_lex_state = 17},
  [2037] = {.lex_state = 47},
  [2038] = {.lex_state = 47},
  [2039] = {.lex_state = 34, .external_lex_state = 16},
  [2040] = {.lex_state = 47},
  [2041] = {.lex_state = 25},
  [2042] = {.lex_state = 25},
  [2043] = {.lex_state = 25},
  [2044] = {.lex_state = 25},
  [2045] = {.lex_state = 25},
  [2046] = {.lex_state = 35},
  [2047] = {.lex_state = 25},
  [2048] = {.lex_state = 25},
  [2049] = {.lex_state = 142},
  [2050] = {.lex_state = 25},
  [2051] = {.lex_state = 30},
  [2052] = {.lex_state = 35},
  [2053] = {.lex_state = 142},
  [2054] = {.lex_state = 25},
  [2055] = {.lex_state = 35},
  [2056] = {.lex_state = 30},
  [2057] = {.lex_state = 30},
  [2058] = {.lex_state = 16},
  [2059] = {.lex_state = 17},
  [2060] = {.lex_state = 16},
  [2061] = {.lex_state = 142},
  [2062] = {.lex_state = 16},
  [2063] = {.lex_state = 16},
  [2064] = {.lex_state = 31},
  [2065] = {.lex_state = 35},
  [2066] = {.lex_state = 142},
  [2067] = {.lex_state = 142},
  [2068] = {.lex_state = 35},
  [2069] = {.lex_state = 16},
  [2070] = {.lex_state = 142},
  [2071] = {.lex_state = 16},
  [2072] = {.lex_state = 30},
  [2073] = {.lex_state = 25},
  [2074] = {.lex_state = 25},
  [2075] = {.lex_state = 25},
  [2076] = {.lex_state = 35},
  [2077] = {.lex_state = 35},
  [2078] = {.lex_state = 16},
  [2079] = {.lex_state = 16},
  [2080] = {.lex_state = 35},
  [2081] = {.lex_state = 30},
  [2082] = {.lex_state = 142},
  [2083] = {.lex_state = 35},
  [2084] = {.lex_state = 30},
  [2085] = {.lex_state = 16},
  [2086] = {.lex_state = 25},
  [2087] = {.lex_state = 25},
  [2088] = {.lex_state = 35},
  [2089] = {.lex_state = 25},
  [2090] = {.lex_state = 35},
  [2091] = {.lex_state = 16},
  [2092] = {.lex_state = 25},
  [2093] = {.lex_state = 17},
  [2094] = {.lex_state = 16},
  [2095] = {.lex_state = 142},
  [2096] = {.lex_state = 142},
  [2097] = {.lex_state = 30},
  [2098] = {.lex_state = 142},
  [2099] = {.lex_state = 16},
  [2100] = {.lex_state = 142},
  [2101] = {.lex_state = 25},
  [2102] = {.lex_state = 142},
  [2103] = {.lex_state = 16},
  [2104] = {.lex_state = 35},
  [2105] = {.lex_state = 142},
  [2106] = {.lex_state = 142},
  [2107] = {.lex_state = 16},
  [2108] = {.lex_state = 16},
  [2109] = {.lex_state = 16},
  [2110] = {.lex_state = 16},
  [2111] = {.lex_state = 16},
  [2112] = {.lex_state = 16},
  [2113] = {.lex_state = 16},
  [2114] = {.lex_state = 142},
  [2115] = {.lex_state = 142},
  [2116] = {.lex_state = 25},
  [2117] = {.lex_state = 25},
  [2118] = {.lex_state = 17},
  [2119] = {.lex_state = 25},
  [2120] = {.lex_state = 16},
  [2121] = {.lex_state = 16},
  [2122] = {.lex_state = 16},
  [2123] = {.lex_state = 35},
  [2124] = {.lex_state = 17},
  [2125] = {.lex_state = 16},
  [2126] = {.lex_state = 16},
  [2127] = {.lex_state = 142},
  [2128] = {.lex_state = 142},
  [2129] = {.lex_state = 16},
  [2130] = {.lex_state = 16},
  [2131] = {.lex_state = 25},
  [2132] = {.lex_state = 16},
  [2133] = {.lex_state = 142},
  [2134] = {.lex_state = 16},
  [2135] = {.lex_state = 16},
  [2136] = {.lex_state = 22},
  [2137] = {.lex_state = 25},
  [2138] = {.lex_state = 16},
  [2139] = {.lex_state = 16},
  [2140] = {.lex_state = 16},
  [2141] = {.lex_state = 16},
  [2142] = {.lex_state = 16},
  [2143] = {.lex_state = 16},
  [2144] = {.lex_state = 25, .external_lex_state = 16},
  [2145] = {.lex_state = 35},
  [2146] = {.lex_state = 16},
  [2147] = {.lex_state = 25},
  [2148] = {.lex_state = 16},
  [2149] = {.lex_state = 16},
  [2150] = {.lex_state = 16},
  [2151] = {.lex_state = 16},
  [2152] = {.lex_state = 16},
  [2153] = {.lex_state = 16},
  [2154] = {.lex_state = 17},
  [2155] = {.lex_state = 16},
  [2156] = {.lex_state = 16},
  [2157] = {.lex_state = 16},
  [2158] = {.lex_state = 16},
  [2159] = {.lex_state = 16},
  [2160] = {.lex_state = 142},
  [2161] = {.lex_state = 16},
  [2162] = {.lex_state = 142},
  [2163] = {.lex_state = 17},
  [2164] = {.lex_state = 16},
  [2165] = {.lex_state = 16},
  [2166] = {.lex_state = 16},
  [2167] = {.lex_state = 142},
  [2168] = {.lex_state = 142},
  [2169] = {.lex_state = 16},
  [2170] = {.lex_state = 142},
  [2171] = {.lex_state = 25},
  [2172] = {.lex_state = 25},
  [2173] = {.lex_state = 142},
  [2174] = {.lex_state = 35},
  [2175] = {.lex_state = 142},
  [2176] = {.lex_state = 17},
  [2177] = {.lex_state = 142},
  [2178] = {.lex_state = 30},
  [2179] = {.lex_state = 25},
  [2180] = {.lex_state = 16},
  [2181] = {.lex_state = 16},
  [2182] = {.lex_state = 16},
  [2183] = {.lex_state = 142},
  [2184] = {.lex_state = 16},
  [2185] = {.lex_state = 16},
  [2186] = {.lex_state = 16},
  [2187] = {.lex_state = 16},
  [2188] = {.lex_state = 16},
  [2189] = {.lex_state = 16},
  [2190] = {.lex_state = 16},
  [2191] = {.lex_state = 142},
  [2192] = {.lex_state = 22},
  [2193] = {.lex_state = 142},
  [2194] = {.lex_state = 16},
  [2195] = {.lex_state = 35},
  [2196] = {.lex_state = 142},
  [2197] = {.lex_state = 142},
  [2198] = {.lex_state = 25, .external_lex_state = 16},
  [2199] = {.lex_state = 142},
  [2200] = {.lex_state = 16},
  [2201] = {.lex_state = 142},
  [2202] = {.lex_state = 16},
  [2203] = {.lex_state = 16},
  [2204] = {.lex_state = 16},
  [2205] = {.lex_state = 142},
  [2206] = {.lex_state = 17},
  [2207] = {.lex_state = 16},
  [2208] = {.lex_state = 142},
  [2209] = {.lex_state = 142},
  [2210] = {.lex_state = 142},
  [2211] = {.lex_state = 142},
  [2212] = {.lex_state = 142},
  [2213] = {.lex_state = 16},
  [2214] = {.lex_state = 142},
  [2215] = {.lex_state = 17},
  [2216] = {.lex_state = 142},
  [2217] = {.lex_state = 16},
  [2218] = {.lex_state = 35},
  [2219] = {.lex_state = 16},
  [2220] = {.lex_state = 142},
  [2221] = {.lex_state = 25},
  [2222] = {.lex_state = 25},
  [2223] = {.lex_state = 25},
  [2224] = {.lex_state = 25},
  [2225] = {.lex_state = 142},
  [2226] = {.lex_state = 25},
  [2227] = {.lex_state = 25},
  [2228] = {.lex_state = 25},
  [2229] = {.lex_state = 25},
  [2230] = {.lex_state = 25},
  [2231] = {.lex_state = 25},
  [2232] = {.lex_state = 25},
  [2233] = {.lex_state = 25},
  [2234] = {.lex_state = 25},
  [2235] = {.lex_state = 25},
  [2236] = {.lex_state = 25},
  [2237] = {.lex_state = 25},
  [2238] = {.lex_state = 25},
  [2239] = {.lex_state = 25},
  [2240] = {.lex_state = 142},
  [2241] = {.lex_state = 142},
  [2242] = {.lex_state = 25},
  [2243] = {.lex_state = 25},
  [2244] = {.lex_state = 25},
  [2245] = {.lex_state = 25},
  [2246] = {.lex_state = 142},
  [2247] = {.lex_state = 142},
  [2248] = {.lex_state = 142},
  [2249] = {.lex_state = 25},
  [2250] = {.lex_state = 142},
  [2251] = {.lex_state = 142},
  [2252] = {.lex_state = 142},
  [2253] = {.lex_state = 30},
  [2254] = {.lex_state = 30},
  [2255] = {.lex_state = 25},
  [2256] = {.lex_state = 142},
  [2257] = {.lex_state = 142},
  [2258] = {.lex_state = 25},
  [2259] = {.lex_state = 25},
  [2260] = {.lex_state = 142},
  [2261] = {.lex_state = 25},
  [2262] = {.lex_state = 25},
  [2263] = {.lex_state = 25},
  [2264] = {.lex_state = 25},
  [2265] = {.lex_state = 25},
  [2266] = {.lex_state = 25},
  [2267] = {.lex_state = 25},
  [2268] = {.lex_state = 25},
  [2269] = {.lex_state = 30},
  [2270] = {.lex_state = 30},
  [2271] = {.lex_state = 16},
  [2272] = {.lex_state = 25},
  [2273] = {.lex_state = 25},
  [2274] = {.lex_state = 25},
  [2275] = {.lex_state = 142},
  [2276] = {.lex_state = 25},
  [2277] = {.lex_state = 25},
  [2278] = {.lex_state = 25},
  [2279] = {.lex_state = 25},
  [2280] = {.lex_state = 25},
  [2281] = {.lex_state = 25},
  [2282] = {.lex_state = 25},
  [2283] = {.lex_state = 25},
  [2284] = {.lex_state = 30},
  [2285] = {.lex_state = 30},
  [2286] = {.lex_state = 25},
  [2287] = {.lex_state = 25},
  [2288] = {.lex_state = 25},
  [2289] = {.lex_state = 142},
  [2290] = {.lex_state = 25},
  [2291] = {.lex_state = 142},
  [2292] = {.lex_state = 25},
  [2293] = {.lex_state = 25},
  [2294] = {.lex_state = 25},
  [2295] = {.lex_state = 25},
  [2296] = {.lex_state = 142},
  [2297] = {.lex_state = 25},
  [2298] = {.lex_state = 25},
  [2299] = {.lex_state = 25},
  [2300] = {.lex_state = 25},
  [2301] = {.lex_state = 30},
  [2302] = {.lex_state = 25},
  [2303] = {.lex_state = 25},
  [2304] = {.lex_state = 25},
  [2305] = {.lex_state = 142},
  [2306] = {.lex_state = 25},
  [2307] = {.lex_state = 25},
  [2308] = {.lex_state = 25},
  [2309] = {.lex_state = 25},
  [2310] = {.lex_state = 25},
  [2311] = {.lex_state = 25},
  [2312] = {.lex_state = 25},
  [2313] = {.lex_state = 25},
  [2314] = {.lex_state = 30},
  [2315] = {.lex_state = 25},
  [2316] = {.lex_state = 25},
  [2317] = {.lex_state = 25},
  [2318] = {.lex_state = 25},
  [2319] = {.lex_state = 25},
  [2320] = {.lex_state = 25},
  [2321] = {.lex_state = 25},
  [2322] = {.lex_state = 25},
  [2323] = {.lex_state = 25},
  [2324] = {.lex_state = 25},
  [2325] = {.lex_state = 25},
  [2326] = {.lex_state = 30},
  [2327] = {.lex_state = 25},
  [2328] = {.lex_state = 16},
  [2329] = {.lex_state = 25},
  [2330] = {.lex_state = 25},
  [2331] = {.lex_state = 25},
  [2332] = {.lex_state = 25},
  [2333] = {.lex_state = 25},
  [2334] = {.lex_state = 25},
  [2335] = {.lex_state = 25},
  [2336] = {.lex_state = 25},
  [2337] = {.lex_state = 25},
  [2338] = {.lex_state = 25},
  [2339] = {.lex_state = 30},
  [2340] = {.lex_state = 25},
  [2341] = {.lex_state = 25},
  [2342] = {.lex_state = 30},
  [2343] = {.lex_state = 142},
  [2344] = {.lex_state = 142},
  [2345] = {.lex_state = 25},
  [2346] = {.lex_state = 25},
  [2347] = {.lex_state = 25},
  [2348] = {.lex_state = 25},
  [2349] = {.lex_state = 30},
  [2350] = {.lex_state = 25},
  [2351] = {.lex_state = 25},
  [2352] = {.lex_state = 25},
  [2353] = {.lex_state = 25},
  [2354] = {.lex_state = 30},
  [2355] = {.lex_state = 25},
  [2356] = {.lex_state = 142},
  [2357] = {.lex_state = 25},
  [2358] = {.lex_state = 30},
  [2359] = {.lex_state = 25},
  [2360] = {.lex_state = 30},
  [2361] = {.lex_state = 25},
  [2362] = {.lex_state = 142},
  [2363] = {.lex_state = 25},
  [2364] = {.lex_state = 25},
  [2365] = {.lex_state = 25},
  [2366] = {.lex_state = 25},
  [2367] = {.lex_state = 25},
  [2368] = {.lex_state = 142},
  [2369] = {.lex_state = 25},
  [2370] = {.lex_state = 30},
  [2371] = {.lex_state = 142},
  [2372] = {.lex_state = 25},
  [2373] = {.lex_state = 9},
  [2374] = {.lex_state = 25},
  [2375] = {.lex_state = 25},
  [2376] = {.lex_state = 142},
  [2377] = {.lex_state = 30},
  [2378] = {.lex_state = 25},
  [2379] = {.lex_state = 25},
  [2380] = {.lex_state = 25},
  [2381] = {.lex_state = 25},
  [2382] = {.lex_state = 25},
  [2383] = {.lex_state = 16},
  [2384] = {.lex_state = 142},
  [2385] = {.lex_state = 142},
  [2386] = {.lex_state = 142},
  [2387] = {.lex_state = 25},
  [2388] = {.lex_state = 25},
  [2389] = {.lex_state = 142},
  [2390] = {.lex_state = 142},
  [2391] = {.lex_state = 142},
  [2392] = {.lex_state = 25},
  [2393] = {.lex_state = 142},
  [2394] = {.lex_state = 25},
  [2395] = {.lex_state = 25},
  [2396] = {.lex_state = 25},
  [2397] = {.lex_state = 25},
  [2398] = {.lex_state = 30},
  [2399] = {.lex_state = 30},
  [2400] = {.lex_state = 9},
  [2401] = {.lex_state = 47},
  [2402] = {.lex_state = 47},
  [2403] = {.lex_state = 25},
  [2404] = {.lex_state = 25},
  [2405] = {.lex_state = 25},
  [2406] = {.lex_state = 25},
  [2407] = {.lex_state = 16},
  [2408] = {.lex_state = 25},
  [2409] = {.lex_state = 25},
  [2410] = {.lex_state = 30},
  [2411] = {.lex_state = 25},
  [2412] = {.lex_state = 142},
  [2413] = {.lex_state = 47},
  [2414] = {.lex_state = 16},
  [2415] = {.lex_state = 142},
  [2416] = {.lex_state = 25},
  [2417] = {.lex_state = 142},
  [2418] = {.lex_state = 142},
  [2419] = {.lex_state = 142},
  [2420] = {.lex_state = 142},
  [2421] = {.lex_state = 30},
  [2422] = {.lex_state = 25},
  [2423] = {.lex_state = 142},
  [2424] = {.lex_state = 25},
  [2425] = {.lex_state = 142},
  [2426] = {.lex_state = 25},
  [2427] = {.lex_state = 25},
  [2428] = {.lex_state = 25},
  [2429] = {.lex_state = 142},
  [2430] = {.lex_state = 142},
  [2431] = {.lex_state = 142},
  [2432] = {.lex_state = 25},
  [2433] = {.lex_state = 25},
  [2434] = {.lex_state = 25},
  [2435] = {.lex_state = 16},
  [2436] = {.lex_state = 142},
  [2437] = {.lex_state = 25},
  [2438] = {.lex_state = 25},
  [2439] = {.lex_state = 16},
  [2440] = {.lex_state = 16},
  [2441] = {.lex_state = 142},
  [2442] = {.lex_state = 31},
  [2443] = {.lex_state = 25},
  [2444] = {.lex_state = 142},
  [2445] = {.lex_state = 142},
  [2446] = {.lex_state = 25},
  [2447] = {.lex_state = 30},
  [2448] = {.lex_state = 142},
  [2449] = {.lex_state = 142},
  [2450] = {.lex_state = 142},
  [2451] = {.lex_state = 142},
  [2452] = {.lex_state = 30},
  [2453] = {.lex_state = 30},
  [2454] = {.lex_state = 142},
  [2455] = {.lex_state = 25},
  [2456] = {.lex_state = 30},
  [2457] = {.lex_state = 142},
  [2458] = {.lex_state = 142},
  [2459] = {.lex_state = 142},
  [2460] = {.lex_state = 142},
  [2461] = {.lex_state = 142},
  [2462] = {.lex_state = 30},
  [2463] = {.lex_state = 142},
  [2464] = {.lex_state = 142},
  [2465] = {.lex_state = 142},
  [2466] = {.lex_state = 30},
  [2467] = {.lex_state = 30},
  [2468] = {.lex_state = 30},
  [2469] = {.lex_state = 25},
  [2470] = {.lex_state = 25},
  [2471] = {.lex_state = 142},
  [2472] = {.lex_state = 25},
  [2473] = {.lex_state = 142},
  [2474] = {.lex_state = 142},
  [2475] = {.lex_state = 142},
  [2476] = {.lex_state = 30},
  [2477] = {.lex_state = 142},
  [2478] = {.lex_state = 30},
  [2479] = {.lex_state = 25},
  [2480] = {.lex_state = 25},
  [2481] = {.lex_state = 25},
  [2482] = {.lex_state = 142},
  [2483] = {.lex_state = 9},
  [2484] = {.lex_state = 30},
  [2485] = {.lex_state = 9},
  [2486] = {.lex_state = 142, .external_lex_state = 10},
  [2487] = {.lex_state = 25},
  [2488] = {.lex_state = 25},
  [2489] = {.lex_state = 25},
  [2490] = {.lex_state = 30},
  [2491] = {.lex_state = 142},
  [2492] = {.lex_state = 142, .external_lex_state = 10},
  [2493] = {.lex_state = 25},
  [2494] = {.lex_state = 52},
  [2495] = {.lex_state = 53},
  [2496] = {.lex_state = 30},
  [2497] = {.lex_state = 142},
  [2498] = {.lex_state = 47},
  [2499] = {.lex_state = 142},
  [2500] = {.lex_state = 17},
  [2501] = {.lex_state = 142},
  [2502] = {.lex_state = 142, .external_lex_state = 10},
  [2503] = {.lex_state = 53},
  [2504] = {.lex_state = 25},
  [2505] = {.lex_state = 9},
  [2506] = {.lex_state = 53},
  [2507] = {.lex_state = 30},
  [2508] = {.lex_state = 25},
  [2509] = {.lex_state = 142},
  [2510] = {.lex_state = 53},
  [2511] = {.lex_state = 25},
  [2512] = {.lex_state = 53},
  [2513] = {.lex_state = 53},
  [2514] = {.lex_state = 25},
  [2515] = {.lex_state = 25},
  [2516] = {.lex_state = 25},
  [2517] = {.lex_state = 25},
  [2518] = {.lex_state = 142},
  [2519] = {.lex_state = 30},
  [2520] = {.lex_state = 25},
  [2521] = {.lex_state = 25},
  [2522] = {.lex_state = 142},
  [2523] = {.lex_state = 25},
  [2524] = {.lex_state = 25, .external_lex_state = 16},
  [2525] = {.lex_state = 142},
  [2526] = {.lex_state = 25},
  [2527] = {.lex_state = 25},
  [2528] = {.lex_state = 25},
  [2529] = {.lex_state = 142},
  [2530] = {.lex_state = 25},
  [2531] = {.lex_state = 9},
  [2532] = {.lex_state = 17},
  [2533] = {.lex_state = 25, .external_lex_state = 16},
  [2534] = {.lex_state = 142, .external_lex_state = 10},
  [2535] = {.lex_state = 142},
  [2536] = {.lex_state = 25},
  [2537] = {.lex_state = 47},
  [2538] = {.lex_state = 142},
  [2539] = {.lex_state = 25},
  [2540] = {.lex_state = 142},
  [2541] = {.lex_state = 25},
  [2542] = {.lex_state = 142},
  [2543] = {.lex_state = 25},
  [2544] = {.lex_state = 25},
  [2545] = {.lex_state = 53},
  [2546] = {.lex_state = 25},
  [2547] = {.lex_state = 142},
  [2548] = {.lex_state = 25},
  [2549] = {.lex_state = 25},
  [2550] = {.lex_state = 25},
  [2551] = {.lex_state = 25},
  [2552] = {.lex_state = 142},
  [2553] = {.lex_state = 25},
  [2554] = {.lex_state = 9},
  [2555] = {.lex_state = 52},
  [2556] = {.lex_state = 47},
  [2557] = {.lex_state = 52},
  [2558] = {.lex_state = 142},
  [2559] = {.lex_state = 53},
  [2560] = {.lex_state = 25},
  [2561] = {.lex_state = 25},
  [2562] = {.lex_state = 25, .external_lex_state = 16},
  [2563] = {.lex_state = 30},
  [2564] = {.lex_state = 25},
  [2565] = {.lex_state = 25},
  [2566] = {.lex_state = 25},
  [2567] = {.lex_state = 25},
  [2568] = {.lex_state = 25},
  [2569] = {.lex_state = 25},
  [2570] = {.lex_state = 25},
  [2571] = {.lex_state = 25},
  [2572] = {.lex_state = 25},
  [2573] = {.lex_state = 25},
  [2574] = {.lex_state = 25},
  [2575] = {.lex_state = 25},
  [2576] = {.lex_state = 25},
  [2577] = {.lex_state = 25},
  [2578] = {.lex_state = 142},
  [2579] = {.lex_state = 25},
  [2580] = {.lex_state = 25},
  [2581] = {.lex_state = 25},
  [2582] = {.lex_state = 25},
  [2583] = {.lex_state = 142},
  [2584] = {.lex_state = 25},
  [2585] = {.lex_state = 25},
  [2586] = {.lex_state = 142},
  [2587] = {.lex_state = 30},
  [2588] = {.lex_state = 142},
  [2589] = {.lex_state = 9},
  [2590] = {.lex_state = 9},
  [2591] = {.lex_state = 25},
  [2592] = {.lex_state = 142},
  [2593] = {.lex_state = 25},
  [2594] = {.lex_state = 25},
  [2595] = {.lex_state = 25},
  [2596] = {.lex_state = 25},
  [2597] = {.lex_state = 142},
  [2598] = {.lex_state = 25},
  [2599] = {.lex_state = 25},
  [2600] = {.lex_state = 25},
  [2601] = {.lex_state = 25},
  [2602] = {.lex_state = 25},
  [2603] = {.lex_state = 142},
  [2604] = {.lex_state = 142},
  [2605] = {.lex_state = 142},
  [2606] = {.lex_state = 142},
  [2607] = {.lex_state = 25},
  [2608] = {.lex_state = 30},
  [2609] = {.lex_state = 142},
  [2610] = {.lex_state = 30},
  [2611] = {.lex_state = 30},
  [2612] = {.lex_state = 142},
  [2613] = {.lex_state = 30},
  [2614] = {.lex_state = 30},
  [2615] = {.lex_state = 142},
  [2616] = {.lex_state = 142},
  [2617] = {.lex_state = 0, .external_lex_state = 18},
  [2618] = {.lex_state = 142},
  [2619] = {.lex_state = 142},
  [2620] = {.lex_state = 142},
  [2621] = {.lex_state = 142},
  [2622] = {.lex_state = 142},
  [2623] = {.lex_state = 142},
  [2624] = {.lex_state = 30},
  [2625] = {.lex_state = 30},
  [2626] = {.lex_state = 30},
  [2627] = {.lex_state = 30},
  [2628] = {.lex_state = 30},
  [2629] = {.lex_state = 142},
  [2630] = {.lex_state = 0, .external_lex_state = 15},
  [2631] = {.lex_state = 142},
  [2632] = {.lex_state = 142},
  [2633] = {.lex_state = 142},
  [2634] = {.lex_state = 142},
  [2635] = {.lex_state = 30},
  [2636] = {.lex_state = 142},
  [2637] = {.lex_state = 30},
  [2638] = {.lex_state = 142},
  [2639] = {.lex_state = 30},
  [2640] = {.lex_state = 30},
  [2641] = {.lex_state = 142},
  [2642] = {.lex_state = 30},
  [2643] = {.lex_state = 142},
  [2644] = {.lex_state = 142},
  [2645] = {.lex_state = 30},
  [2646] = {.lex_state = 142},
  [2647] = {.lex_state = 142},
  [2648] = {.lex_state = 142},
  [2649] = {.lex_state = 142},
  [2650] = {.lex_state = 142},
  [2651] = {.lex_state = 142},
  [2652] = {.lex_state = 142},
  [2653] = {.lex_state = 142},
  [2654] = {.lex_state = 142},
  [2655] = {.lex_state = 142},
  [2656] = {.lex_state = 0, .external_lex_state = 15},
  [2657] = {.lex_state = 142},
  [2658] = {.lex_state = 142},
  [2659] = {.lex_state = 142},
  [2660] = {.lex_state = 142},
  [2661] = {.lex_state = 30},
  [2662] = {.lex_state = 142},
  [2663] = {.lex_state = 30},
  [2664] = {.lex_state = 142},
  [2665] = {.lex_state = 0, .external_lex_state = 19},
  [2666] = {.lex_state = 142},
  [2667] = {.lex_state = 142},
  [2668] = {.lex_state = 142},
  [2669] = {.lex_state = 142},
  [2670] = {.lex_state = 142},
  [2671] = {.lex_state = 0, .external_lex_state = 18},
  [2672] = {.lex_state = 142},
  [2673] = {.lex_state = 30},
  [2674] = {.lex_state = 142},
  [2675] = {.lex_state = 142},
  [2676] = {.lex_state = 142},
  [2677] = {.lex_state = 142},
  [2678] = {.lex_state = 0, .external_lex_state = 19},
  [2679] = {.lex_state = 0, .external_lex_state = 18},
  [2680] = {.lex_state = 0, .external_lex_state = 15},
  [2681] = {.lex_state = 30},
  [2682] = {.lex_state = 142},
  [2683] = {.lex_state = 142},
  [2684] = {.lex_state = 142},
  [2685] = {.lex_state = 142},
  [2686] = {.lex_state = 142},
  [2687] = {.lex_state = 0, .external_lex_state = 19},
  [2688] = {.lex_state = 25},
  [2689] = {.lex_state = 142},
  [2690] = {.lex_state = 142},
  [2691] = {.lex_state = 142},
  [2692] = {.lex_state = 142},
  [2693] = {.lex_state = 142},
  [2694] = {.lex_state = 25},
  [2695] = {.lex_state = 142},
  [2696] = {.lex_state = 142},
  [2697] = {.lex_state = 142},
  [2698] = {.lex_state = 142},
  [2699] = {.lex_state = 25},
  [2700] = {.lex_state = 30},
  [2701] = {.lex_state = 0, .external_lex_state = 15},
  [2702] = {.lex_state = 142},
  [2703] = {.lex_state = 142},
  [2704] = {.lex_state = 142},
  [2705] = {.lex_state = 142},
  [2706] = {.lex_state = 0, .external_lex_state = 19},
  [2707] = {.lex_state = 142},
  [2708] = {.lex_state = 0, .external_lex_state = 15},
  [2709] = {.lex_state = 142},
  [2710] = {.lex_state = 25},
  [2711] = {.lex_state = 0, .external_lex_state = 18},
  [2712] = {.lex_state = 142},
  [2713] = {.lex_state = 0, .external_lex_state = 18},
  [2714] = {.lex_state = 30},
  [2715] = {.lex_state = 30},
  [2716] = {.lex_state = 142},
  [2717] = {.lex_state = 30},
  [2718] = {.lex_state = 142},
  [2719] = {.lex_state = 0, .external_lex_state = 18},
  [2720] = {.lex_state = 0},
  [2721] = {.lex_state = 142},
  [2722] = {.lex_state = 30},
  [2723] = {.lex_state = 0, .external_lex_state = 18},
  [2724] = {.lex_state = 25},
  [2725] = {.lex_state = 30},
  [2726] = {.lex_state = 142},
  [2727] = {.lex_state = 142},
  [2728] = {.lex_state = 142},
  [2729] = {.lex_state = 142},
  [2730] = {.lex_state = 25},
  [2731] = {.lex_state = 28},
  [2732] = {.lex_state = 30},
  [2733] = {.lex_state = 9},
  [2734] = {.lex_state = 142},
  [2735] = {.lex_state = 30},
  [2736] = {.lex_state = 142},
  [2737] = {.lex_state = 30},
  [2738] = {.lex_state = 30},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_identifier] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [aux_sym_subroutine_block_token1] = ACTIONS(1),
    [aux_sym_if_block_token1] = ACTIONS(1),
    [aux_sym_elif_block_token1] = ACTIONS(1),
    [aux_sym_while_block_token1] = ACTIONS(1),
    [aux_sym__repeat_block_pre_condition_format_token1] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [aux_sym_parameter_declaration_token1] = ACTIONS(1),
    [aux_sym_parameter_declaration_token2] = ACTIONS(1),
    [aux_sym_parameter_declaration_token3] = ACTIONS(1),
    [aux_sym_subroutine_call_expression_token1] = ACTIONS(1),
    [aux_sym_macro_definition_command_token1] = ACTIONS(1),
    [aux_sym_macro_definition_command_token2] = ACTIONS(1),
    [aux_sym_macro_definition_command_token3] = ACTIONS(1),
    [aux_sym__host_command_with_default_format_token1] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [aux_sym__host_command_argument_list_default_format_token1] = ACTIONS(1),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(1),
    [aux_sym__hll_non_default_format_var_command_token2] = ACTIONS(1),
    [aux_sym__hll_non_default_format_var_command_token3] = ACTIONS(1),
    [aux_sym__hll_non_default_format_var_command_token4] = ACTIONS(1),
    [aux_sym__hll_non_default_format_var_command_token5] = ACTIONS(1),
    [aux_sym__var_command_with_non_hll_format_token2] = ACTIONS(1),
    [aux_sym__var_command_with_non_hll_format_token3] = ACTIONS(1),
    [aux_sym__hll_var_definition_command_expression_token1] = ACTIONS(1),
    [aux_sym__hll_var_definition_command_expression_token2] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_BSLASH] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_sizeof] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [aux_sym_hll_char_literal_token1] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_const] = ACTIONS(1),
    [anon_sym_volatile] = ACTIONS(1),
    [anon_sym_struct] = ACTIONS(1),
    [anon_sym_union] = ACTIONS(1),
    [anon_sym_enum] = ACTIONS(1),
    [anon_sym_signed] = ACTIONS(1),
    [anon_sym_unsigned] = ACTIONS(1),
    [anon_sym_long] = ACTIONS(1),
    [anon_sym_short] = ACTIONS(1),
    [sym_hll_primitive_type] = ACTIONS(1),
    [aux_sym_character_token2] = ACTIONS(1),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(1),
    [sym_label_identifier] = ACTIONS(1),
    [sym__and_operator_pre_hook] = ACTIONS(1),
    [sym__decimal_number] = ACTIONS(1),
    [sym__decimal_number_pre_hook] = ACTIONS(1),
    [sym_hll_number_literal] = ACTIONS(1),
    [sym__hll_number_literal_pre_hook] = ACTIONS(1),
    [sym_path] = ACTIONS(1),
    [sym__ampersand_char] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_script] = STATE(2720),
    [sym_block] = STATE(303),
    [sym__statement] = STATE(303),
    [sym_labeled_expression] = STATE(303),
    [sym_subroutine_block] = STATE(303),
    [sym_if_block] = STATE(303),
    [sym_while_block] = STATE(303),
    [sym_repeat_block] = STATE(303),
    [sym__repeat_block_pre_condition_format] = STATE(1531),
    [sym__repeat_block_post_condition_format] = STATE(1532),
    [sym_macro_assignment_expression] = STATE(303),
    [sym__label] = STATE(359),
    [sym_macro_definition] = STATE(303),
    [sym_parameter_declaration] = STATE(303),
    [sym_subroutine_call_expression] = STATE(303),
    [sym_macro] = STATE(2316),
    [sym_recursive_macro_expansion] = STATE(1396),
    [sym_macro_definition_command] = STATE(2404),
    [sym_command_expression] = STATE(303),
    [sym__host_command] = STATE(2629),
    [sym__host_command_with_default_format] = STATE(2493),
    [sym__other_command] = STATE(2672),
    [sym__command_identifier] = STATE(2241),
    [sym__var_command] = STATE(2618),
    [sym__hll_var_command] = STATE(2650),
    [sym__hll_default_format_var_command] = STATE(2670),
    [sym__hll_non_default_format_var_command] = STATE(2697),
    [sym__non_hll_var_command] = STATE(2718),
    [sym__hll_var_definition_command] = STATE(2605),
    [sym__hll_var_command_with_default_format] = STATE(2425),
    [sym__var_command_with_non_hll_format] = STATE(2429),
    [sym__hll_var_definition_command_expression] = STATE(2544),
    [sym__hll_var_if_command] = STATE(2546),
    [aux_sym_script_repeat1] = STATE(303),
    [aux_sym_block_repeat1] = STATE(361),
    [ts_builtin_sym_end] = ACTIONS(5),
    [sym_identifier] = ACTIONS(7),
    [aux_sym_block_token1] = ACTIONS(9),
    [aux_sym_subroutine_block_token1] = ACTIONS(11),
    [aux_sym_if_block_token1] = ACTIONS(13),
    [aux_sym_while_block_token1] = ACTIONS(15),
    [aux_sym_while_block_token2] = ACTIONS(17),
    [aux_sym__repeat_block_pre_condition_format_token1] = ACTIONS(19),
    [anon_sym_AMP_AMP] = ACTIONS(21),
    [anon_sym_AMP] = ACTIONS(23),
    [aux_sym_macro_definition_token1] = ACTIONS(25),
    [aux_sym_parameter_declaration_token1] = ACTIONS(27),
    [aux_sym_parameter_declaration_token2] = ACTIONS(27),
    [aux_sym_parameter_declaration_token3] = ACTIONS(29),
    [aux_sym_subroutine_call_expression_token1] = ACTIONS(31),
    [aux_sym_macro_definition_command_token1] = ACTIONS(33),
    [aux_sym_macro_definition_command_token2] = ACTIONS(33),
    [aux_sym_macro_definition_command_token3] = ACTIONS(33),
    [aux_sym__host_command_with_default_format_token1] = ACTIONS(35),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(37),
    [sym_comment] = ACTIONS(3),
    [sym__blank_line] = ACTIONS(39),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(41),
    [sym_label_identifier] = ACTIONS(43),
  },
  [STATE(2)] = {
    [sym__expression] = STATE(452),
    [sym__parenthesized_expression] = STATE(452),
    [sym__path_expression] = STATE(452),
    [sym_unary_expression] = STATE(452),
    [sym__unary_expression] = STATE(581),
    [sym__prefix_postfix_expression] = STATE(564),
    [sym_binary_expression] = STATE(452),
    [sym__binary_expression] = STATE(565),
    [sym__and_expression] = STATE(573),
    [sym_assignment_expression] = STATE(452),
    [sym_macro] = STATE(406),
    [sym_call_expression] = STATE(424),
    [sym__function_identifier] = STATE(2589),
    [sym__hll_var_call_expression_function] = STATE(2590),
    [sym__hll_var_call_expression] = STATE(428),
    [sym_trace32_hll_variable] = STATE(575),
    [sym__literal] = STATE(452),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(211),
    [sym_address] = STATE(452),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(411),
    [sym_bitmask] = STATE(452),
    [sym_string] = STATE(452),
    [sym_character] = STATE(452),
    [sym_symbol] = STATE(452),
    [sym__special_command_operator] = STATE(452),
    [sym__file_dialog] = STATE(569),
    [aux_sym__path_expression_repeat1] = STATE(443),
    [aux_sym__path_expression_repeat2] = STATE(452),
    [sym_identifier] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [aux_sym_subroutine_block_token1] = ACTIONS(51),
    [aux_sym_if_block_token1] = ACTIONS(51),
    [aux_sym_while_block_token1] = ACTIONS(51),
    [aux_sym_while_block_token2] = ACTIONS(53),
    [aux_sym__repeat_block_pre_condition_format_token1] = ACTIONS(51),
    [anon_sym_EQ] = ACTIONS(51),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(57),
    [anon_sym_BANG] = ACTIONS(55),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_CARET_CARET] = ACTIONS(53),
    [anon_sym_PIPE_PIPE] = ACTIONS(53),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_SLASH] = ACTIONS(51),
    [anon_sym_PERCENT] = ACTIONS(53),
    [anon_sym_PIPE] = ACTIONS(51),
    [anon_sym_CARET] = ACTIONS(51),
    [anon_sym_EQ_EQ] = ACTIONS(53),
    [anon_sym_BANG_EQ] = ACTIONS(53),
    [anon_sym_GT] = ACTIONS(51),
    [anon_sym_GT_EQ] = ACTIONS(53),
    [anon_sym_LT_EQ] = ACTIONS(53),
    [anon_sym_LT] = ACTIONS(51),
    [anon_sym_LT_LT] = ACTIONS(53),
    [anon_sym_GT_GT] = ACTIONS(53),
    [anon_sym_DOT_DOT] = ACTIONS(53),
    [anon_sym_AMP_AMP] = ACTIONS(53),
    [anon_sym_AMP] = ACTIONS(63),
    [aux_sym_macro_definition_token1] = ACTIONS(51),
    [aux_sym_parameter_declaration_token1] = ACTIONS(51),
    [aux_sym_parameter_declaration_token2] = ACTIONS(51),
    [aux_sym_parameter_declaration_token3] = ACTIONS(51),
    [aux_sym_subroutine_call_expression_token1] = ACTIONS(51),
    [aux_sym_macro_definition_command_token1] = ACTIONS(51),
    [aux_sym_macro_definition_command_token2] = ACTIONS(51),
    [aux_sym_macro_definition_command_token3] = ACTIONS(51),
    [aux_sym__host_command_with_default_format_token1] = ACTIONS(51),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(67),
    [anon_sym_SQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_access_class] = ACTIONS(73),
    [aux_sym_integer_token1] = ACTIONS(75),
    [aux_sym_integer_token2] = ACTIONS(75),
    [sym_float] = ACTIONS(77),
    [aux_sym_bitmask_token1] = ACTIONS(79),
    [aux_sym_bitmask_token2] = ACTIONS(79),
    [aux_sym_symbol_token1] = ACTIONS(81),
    [aux_sym_symbol_token2] = ACTIONS(83),
    [sym__composed_name] = ACTIONS(85),
    [sym_time] = ACTIONS(77),
    [sym_frequency] = ACTIONS(77),
    [sym_percentage] = ACTIONS(85),
    [sym_file_handle] = ACTIONS(85),
    [sym_comment] = ACTIONS(53),
    [sym__blank_line] = ACTIONS(53),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(51),
    [sym_label_identifier] = ACTIONS(53),
    [sym__and_operator_pre_hook] = ACTIONS(53),
    [sym__decimal_number_pre_hook] = ACTIONS(87),
    [sym_path] = ACTIONS(85),
  },
  [STATE(3)] = {
    [sym__expression] = STATE(1512),
    [sym__parenthesized_expression] = STATE(1512),
    [sym__path_expression] = STATE(1512),
    [sym_unary_expression] = STATE(1512),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1512),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1512),
    [sym_macro] = STATE(408),
    [sym_call_expression] = STATE(430),
    [sym__function_identifier] = STATE(2589),
    [sym__hll_var_call_expression_function] = STATE(2590),
    [sym__hll_var_call_expression] = STATE(428),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1512),
    [sym__address_expression] = STATE(420),
    [sym_memory_space] = STATE(212),
    [sym_address] = STATE(1512),
    [sym__parenthesized_address_expression] = STATE(420),
    [sym__address_binary_expression] = STATE(414),
    [sym_integer] = STATE(431),
    [sym_bitmask] = STATE(1512),
    [sym_string] = STATE(1512),
    [sym_character] = STATE(1512),
    [sym_symbol] = STATE(1512),
    [sym__special_command_operator] = STATE(1512),
    [sym__file_dialog] = STATE(1482),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(89),
    [anon_sym_LPAREN] = ACTIONS(92),
    [anon_sym_LBRACE] = ACTIONS(95),
    [aux_sym_subroutine_block_token1] = ACTIONS(97),
    [aux_sym_if_block_token1] = ACTIONS(97),
    [aux_sym_while_block_token1] = ACTIONS(97),
    [aux_sym_while_block_token2] = ACTIONS(99),
    [aux_sym__repeat_block_pre_condition_format_token1] = ACTIONS(97),
    [anon_sym_EQ] = ACTIONS(97),
    [anon_sym_PLUS] = ACTIONS(101),
    [anon_sym_DASH] = ACTIONS(101),
    [anon_sym_TILDE] = ACTIONS(104),
    [anon_sym_BANG] = ACTIONS(106),
    [anon_sym_PLUS_PLUS] = ACTIONS(108),
    [anon_sym_DASH_DASH] = ACTIONS(108),
    [anon_sym_CARET_CARET] = ACTIONS(99),
    [anon_sym_PIPE_PIPE] = ACTIONS(99),
    [anon_sym_STAR] = ACTIONS(111),
    [anon_sym_SLASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(99),
    [anon_sym_PIPE] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(99),
    [anon_sym_BANG_EQ] = ACTIONS(99),
    [anon_sym_GT] = ACTIONS(97),
    [anon_sym_GT_EQ] = ACTIONS(99),
    [anon_sym_LT_EQ] = ACTIONS(99),
    [anon_sym_LT] = ACTIONS(97),
    [anon_sym_LT_LT] = ACTIONS(99),
    [anon_sym_GT_GT] = ACTIONS(99),
    [anon_sym_DOT_DOT] = ACTIONS(99),
    [anon_sym_AMP_AMP] = ACTIONS(99),
    [anon_sym_AMP] = ACTIONS(114),
    [aux_sym_macro_definition_token1] = ACTIONS(97),
    [aux_sym_parameter_declaration_token1] = ACTIONS(97),
    [aux_sym_parameter_declaration_token2] = ACTIONS(97),
    [aux_sym_parameter_declaration_token3] = ACTIONS(97),
    [aux_sym_subroutine_call_expression_token1] = ACTIONS(97),
    [aux_sym_macro_definition_command_token1] = ACTIONS(97),
    [aux_sym_macro_definition_command_token2] = ACTIONS(97),
    [aux_sym_macro_definition_command_token3] = ACTIONS(97),
    [aux_sym__host_command_with_default_format_token1] = ACTIONS(97),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(117),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(75),
    [aux_sym_integer_token2] = ACTIONS(75),
    [sym_float] = ACTIONS(128),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(136),
    [sym_time] = ACTIONS(128),
    [sym_frequency] = ACTIONS(128),
    [sym_percentage] = ACTIONS(136),
    [sym_file_handle] = ACTIONS(136),
    [sym_comment] = ACTIONS(99),
    [sym__blank_line] = ACTIONS(99),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(97),
    [sym_label_identifier] = ACTIONS(99),
    [sym__and_operator_pre_hook] = ACTIONS(99),
    [sym__decimal_number_pre_hook] = ACTIONS(87),
    [sym_path] = ACTIONS(138),
  },
  [STATE(4)] = {
    [sym__expression] = STATE(1265),
    [sym__parenthesized_expression] = STATE(1265),
    [sym__path_expression] = STATE(1265),
    [sym_unary_expression] = STATE(1265),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1265),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1265),
    [sym_macro] = STATE(1030),
    [sym_call_expression] = STATE(1126),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1265),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1265),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1127),
    [sym_bitmask] = STATE(1265),
    [sym_string] = STATE(1265),
    [sym_character] = STATE(1265),
    [sym_symbol] = STATE(1265),
    [sym__special_command_operator] = STATE(1265),
    [sym__file_dialog] = STATE(1482),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_RPAREN] = ACTIONS(53),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_RBRACE] = ACTIONS(53),
    [anon_sym_EQ] = ACTIONS(51),
    [anon_sym_PLUS] = ACTIONS(146),
    [anon_sym_DASH] = ACTIONS(146),
    [anon_sym_TILDE] = ACTIONS(148),
    [anon_sym_BANG] = ACTIONS(146),
    [anon_sym_PLUS_PLUS] = ACTIONS(150),
    [anon_sym_DASH_DASH] = ACTIONS(150),
    [anon_sym_CARET_CARET] = ACTIONS(53),
    [anon_sym_PIPE_PIPE] = ACTIONS(53),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_SLASH] = ACTIONS(51),
    [anon_sym_PERCENT] = ACTIONS(53),
    [anon_sym_PIPE] = ACTIONS(51),
    [anon_sym_CARET] = ACTIONS(51),
    [anon_sym_EQ_EQ] = ACTIONS(53),
    [anon_sym_BANG_EQ] = ACTIONS(53),
    [anon_sym_GT] = ACTIONS(51),
    [anon_sym_GT_EQ] = ACTIONS(53),
    [anon_sym_LT_EQ] = ACTIONS(53),
    [anon_sym_LT] = ACTIONS(51),
    [anon_sym_LT_LT] = ACTIONS(53),
    [anon_sym_GT_GT] = ACTIONS(53),
    [anon_sym_DOT_DOT] = ACTIONS(53),
    [anon_sym_AMP] = ACTIONS(154),
    [anon_sym_COMMA] = ACTIONS(53),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(158),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(160),
    [sym_time] = ACTIONS(158),
    [sym_frequency] = ACTIONS(158),
    [sym_percentage] = ACTIONS(160),
    [sym_file_handle] = ACTIONS(160),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(51),
    [sym__and_operator_pre_hook] = ACTIONS(53),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(5)] = {
    [sym__expression] = STATE(1493),
    [sym__parenthesized_expression] = STATE(1493),
    [sym__path_expression] = STATE(1493),
    [sym_unary_expression] = STATE(1493),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1493),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1493),
    [sym_macro] = STATE(1060),
    [sym_call_expression] = STATE(1133),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1493),
    [sym__address_expression] = STATE(1124),
    [sym_memory_space] = STATE(264),
    [sym_address] = STATE(1493),
    [sym__parenthesized_address_expression] = STATE(1124),
    [sym__address_binary_expression] = STATE(1099),
    [sym_integer] = STATE(1134),
    [sym_bitmask] = STATE(1493),
    [sym_string] = STATE(1493),
    [sym_character] = STATE(1493),
    [sym_symbol] = STATE(1493),
    [sym__special_command_operator] = STATE(1493),
    [sym__file_dialog] = STATE(1482),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_RPAREN] = ACTIONS(99),
    [anon_sym_LPAREN] = ACTIONS(164),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_RBRACE] = ACTIONS(99),
    [anon_sym_EQ] = ACTIONS(97),
    [anon_sym_PLUS] = ACTIONS(101),
    [anon_sym_DASH] = ACTIONS(101),
    [anon_sym_TILDE] = ACTIONS(104),
    [anon_sym_BANG] = ACTIONS(106),
    [anon_sym_PLUS_PLUS] = ACTIONS(108),
    [anon_sym_DASH_DASH] = ACTIONS(108),
    [anon_sym_CARET_CARET] = ACTIONS(99),
    [anon_sym_PIPE_PIPE] = ACTIONS(99),
    [anon_sym_STAR] = ACTIONS(111),
    [anon_sym_SLASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(99),
    [anon_sym_PIPE] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(99),
    [anon_sym_BANG_EQ] = ACTIONS(99),
    [anon_sym_GT] = ACTIONS(97),
    [anon_sym_GT_EQ] = ACTIONS(99),
    [anon_sym_LT_EQ] = ACTIONS(99),
    [anon_sym_LT] = ACTIONS(97),
    [anon_sym_LT_LT] = ACTIONS(99),
    [anon_sym_GT_GT] = ACTIONS(99),
    [anon_sym_DOT_DOT] = ACTIONS(99),
    [anon_sym_AMP] = ACTIONS(154),
    [anon_sym_COMMA] = ACTIONS(99),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(169),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(171),
    [sym_time] = ACTIONS(169),
    [sym_frequency] = ACTIONS(169),
    [sym_percentage] = ACTIONS(171),
    [sym_file_handle] = ACTIONS(171),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(97),
    [sym__and_operator_pre_hook] = ACTIONS(99),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(6)] = {
    [sym__expression] = STATE(1337),
    [sym__parenthesized_expression] = STATE(1337),
    [sym__path_expression] = STATE(1337),
    [sym_unary_expression] = STATE(1337),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1337),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1337),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1337),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1337),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1337),
    [sym_string] = STATE(1337),
    [sym_character] = STATE(1337),
    [sym_symbol] = STATE(1337),
    [sym__special_command_operator] = STATE(1337),
    [sym__file_dialog] = STATE(1496),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_EQ] = ACTIONS(51),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(179),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_CARET_CARET] = ACTIONS(53),
    [anon_sym_PIPE_PIPE] = ACTIONS(53),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_SLASH] = ACTIONS(51),
    [anon_sym_PERCENT] = ACTIONS(53),
    [anon_sym_PIPE] = ACTIONS(51),
    [anon_sym_CARET] = ACTIONS(51),
    [anon_sym_EQ_EQ] = ACTIONS(53),
    [anon_sym_BANG_EQ] = ACTIONS(53),
    [anon_sym_GT] = ACTIONS(51),
    [anon_sym_GT_EQ] = ACTIONS(53),
    [anon_sym_LT_EQ] = ACTIONS(53),
    [anon_sym_LT] = ACTIONS(51),
    [anon_sym_LT_LT] = ACTIONS(53),
    [anon_sym_GT_GT] = ACTIONS(53),
    [anon_sym_DOT_DOT] = ACTIONS(53),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(53),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(199),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(207),
    [sym_time] = ACTIONS(199),
    [sym_frequency] = ACTIONS(199),
    [sym_percentage] = ACTIONS(207),
    [sym_file_handle] = ACTIONS(207),
    [sym_comment] = ACTIONS(53),
    [sym__blank_line] = ACTIONS(53),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(51),
    [sym__and_operator_pre_hook] = ACTIONS(53),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(7)] = {
    [sym__expression] = STATE(1510),
    [sym__parenthesized_expression] = STATE(1510),
    [sym__path_expression] = STATE(1510),
    [sym_unary_expression] = STATE(1510),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1510),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1510),
    [sym_macro] = STATE(1090),
    [sym_call_expression] = STATE(1247),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1510),
    [sym__address_expression] = STATE(1173),
    [sym_memory_space] = STATE(248),
    [sym_address] = STATE(1510),
    [sym__parenthesized_address_expression] = STATE(1173),
    [sym__address_binary_expression] = STATE(1249),
    [sym_integer] = STATE(1163),
    [sym_bitmask] = STATE(1510),
    [sym_string] = STATE(1510),
    [sym_character] = STATE(1510),
    [sym_symbol] = STATE(1510),
    [sym__special_command_operator] = STATE(1510),
    [sym__file_dialog] = STATE(1482),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(213),
    [anon_sym_LBRACE] = ACTIONS(216),
    [anon_sym_EQ] = ACTIONS(97),
    [anon_sym_PLUS] = ACTIONS(101),
    [anon_sym_DASH] = ACTIONS(101),
    [anon_sym_TILDE] = ACTIONS(104),
    [anon_sym_BANG] = ACTIONS(106),
    [anon_sym_PLUS_PLUS] = ACTIONS(108),
    [anon_sym_DASH_DASH] = ACTIONS(108),
    [anon_sym_CARET_CARET] = ACTIONS(99),
    [anon_sym_PIPE_PIPE] = ACTIONS(99),
    [anon_sym_STAR] = ACTIONS(111),
    [anon_sym_SLASH] = ACTIONS(97),
    [anon_sym_PERCENT] = ACTIONS(99),
    [anon_sym_PIPE] = ACTIONS(97),
    [anon_sym_CARET] = ACTIONS(97),
    [anon_sym_EQ_EQ] = ACTIONS(99),
    [anon_sym_BANG_EQ] = ACTIONS(99),
    [anon_sym_GT] = ACTIONS(97),
    [anon_sym_GT_EQ] = ACTIONS(99),
    [anon_sym_LT_EQ] = ACTIONS(99),
    [anon_sym_LT] = ACTIONS(97),
    [anon_sym_LT_LT] = ACTIONS(99),
    [anon_sym_GT_GT] = ACTIONS(99),
    [anon_sym_DOT_DOT] = ACTIONS(99),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(99),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(218),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(220),
    [sym_time] = ACTIONS(218),
    [sym_frequency] = ACTIONS(218),
    [sym_percentage] = ACTIONS(220),
    [sym_file_handle] = ACTIONS(220),
    [sym_comment] = ACTIONS(99),
    [sym__blank_line] = ACTIONS(99),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(97),
    [sym__and_operator_pre_hook] = ACTIONS(99),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(138),
  },
  [STATE(8)] = {
    [sym__expression] = STATE(1460),
    [sym__parenthesized_expression] = STATE(1460),
    [sym__path_expression] = STATE(1460),
    [sym_unary_expression] = STATE(1460),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1460),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1460),
    [sym_macro] = STATE(1054),
    [sym_call_expression] = STATE(1115),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1460),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1460),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1118),
    [sym_bitmask] = STATE(1460),
    [sym_string] = STATE(1460),
    [sym_character] = STATE(1460),
    [sym_symbol] = STATE(1460),
    [sym__special_command_operator] = STATE(1460),
    [sym__file_dialog] = STATE(1482),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_EQ] = ACTIONS(51),
    [anon_sym_PLUS] = ACTIONS(106),
    [anon_sym_DASH] = ACTIONS(106),
    [anon_sym_TILDE] = ACTIONS(104),
    [anon_sym_BANG] = ACTIONS(106),
    [anon_sym_PLUS_PLUS] = ACTIONS(222),
    [anon_sym_DASH_DASH] = ACTIONS(222),
    [anon_sym_CARET_CARET] = ACTIONS(53),
    [anon_sym_PIPE_PIPE] = ACTIONS(53),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_SLASH] = ACTIONS(51),
    [anon_sym_PERCENT] = ACTIONS(53),
    [anon_sym_PIPE] = ACTIONS(51),
    [anon_sym_CARET] = ACTIONS(51),
    [anon_sym_EQ_EQ] = ACTIONS(53),
    [anon_sym_BANG_EQ] = ACTIONS(53),
    [anon_sym_GT] = ACTIONS(51),
    [anon_sym_GT_EQ] = ACTIONS(53),
    [anon_sym_LT_EQ] = ACTIONS(53),
    [anon_sym_LT] = ACTIONS(51),
    [anon_sym_LT_LT] = ACTIONS(53),
    [anon_sym_GT_GT] = ACTIONS(53),
    [anon_sym_DOT_DOT] = ACTIONS(53),
    [anon_sym_AMP] = ACTIONS(154),
    [anon_sym_COMMA] = ACTIONS(53),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(224),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(226),
    [sym_time] = ACTIONS(224),
    [sym_frequency] = ACTIONS(224),
    [sym_percentage] = ACTIONS(226),
    [sym_file_handle] = ACTIONS(226),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(51),
    [sym__and_operator_pre_hook] = ACTIONS(53),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(9)] = {
    [sym_trace32_hll_variable] = STATE(819),
    [sym__hll_expression] = STATE(628),
    [sym_hll_call_expression] = STATE(818),
    [sym_hll_field_expression] = STATE(818),
    [sym_hll_subscript_expression] = STATE(818),
    [sym_hll_cast_expression] = STATE(628),
    [sym_hll_sizeof_expression] = STATE(628),
    [sym_hll_pointer_expression] = STATE(818),
    [sym_hll_update_expression] = STATE(628),
    [sym_hll_unary_expression] = STATE(628),
    [sym_hll_binary_expression] = STATE(628),
    [sym_hll_assignment_expression] = STATE(628),
    [sym__hll_assignment_left_expression] = STATE(1708),
    [sym_hll_conditional_expression] = STATE(628),
    [sym_hll_char_literal] = STATE(628),
    [sym_hll_string_literal] = STATE(628),
    [sym_hll_parenthesized_expression] = STATE(818),
    [sym_symbol] = STATE(628),
    [sym_identifier] = ACTIONS(228),
    [anon_sym_RPAREN] = ACTIONS(230),
    [anon_sym_LPAREN] = ACTIONS(232),
    [anon_sym_EQ] = ACTIONS(234),
    [anon_sym_PLUS] = ACTIONS(236),
    [anon_sym_DASH] = ACTIONS(236),
    [anon_sym_TILDE] = ACTIONS(238),
    [anon_sym_BANG] = ACTIONS(236),
    [anon_sym_PLUS_PLUS] = ACTIONS(240),
    [anon_sym_DASH_DASH] = ACTIONS(240),
    [anon_sym_PIPE_PIPE] = ACTIONS(230),
    [anon_sym_STAR] = ACTIONS(242),
    [anon_sym_SLASH] = ACTIONS(234),
    [anon_sym_PERCENT] = ACTIONS(234),
    [anon_sym_PIPE] = ACTIONS(234),
    [anon_sym_CARET] = ACTIONS(234),
    [anon_sym_EQ_EQ] = ACTIONS(230),
    [anon_sym_BANG_EQ] = ACTIONS(230),
    [anon_sym_GT] = ACTIONS(234),
    [anon_sym_GT_EQ] = ACTIONS(230),
    [anon_sym_LT_EQ] = ACTIONS(230),
    [anon_sym_LT] = ACTIONS(234),
    [anon_sym_LT_LT] = ACTIONS(234),
    [anon_sym_GT_GT] = ACTIONS(234),
    [anon_sym_DOT_DOT] = ACTIONS(230),
    [anon_sym_AMP_AMP] = ACTIONS(230),
    [anon_sym_AMP] = ACTIONS(242),
    [anon_sym_COLON] = ACTIONS(230),
    [anon_sym_DOT] = ACTIONS(234),
    [anon_sym_COMMA] = ACTIONS(230),
    [anon_sym_QMARK] = ACTIONS(230),
    [anon_sym_BSLASH] = ACTIONS(244),
    [anon_sym_DASH_GT] = ACTIONS(230),
    [anon_sym_LBRACK] = ACTIONS(230),
    [anon_sym_RBRACK] = ACTIONS(230),
    [anon_sym_sizeof] = ACTIONS(246),
    [anon_sym_STAR_EQ] = ACTIONS(230),
    [anon_sym_SLASH_EQ] = ACTIONS(230),
    [anon_sym_PERCENT_EQ] = ACTIONS(230),
    [anon_sym_PLUS_EQ] = ACTIONS(230),
    [anon_sym_DASH_EQ] = ACTIONS(230),
    [anon_sym_LT_LT_EQ] = ACTIONS(230),
    [anon_sym_GT_GT_EQ] = ACTIONS(230),
    [anon_sym_AMP_EQ] = ACTIONS(230),
    [anon_sym_CARET_EQ] = ACTIONS(230),
    [anon_sym_PIPE_EQ] = ACTIONS(230),
    [anon_sym_L_SQUOTE] = ACTIONS(248),
    [anon_sym_u_SQUOTE] = ACTIONS(248),
    [anon_sym_U_SQUOTE] = ACTIONS(248),
    [anon_sym_u8_SQUOTE] = ACTIONS(248),
    [anon_sym_SQUOTE] = ACTIONS(248),
    [anon_sym_L_DQUOTE] = ACTIONS(250),
    [anon_sym_u_DQUOTE] = ACTIONS(250),
    [anon_sym_U_DQUOTE] = ACTIONS(250),
    [anon_sym_u8_DQUOTE] = ACTIONS(250),
    [anon_sym_DQUOTE] = ACTIONS(250),
    [aux_sym_symbol_token1] = ACTIONS(252),
    [aux_sym_symbol_token2] = ACTIONS(254),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(234),
    [sym__hll_number_literal_pre_hook] = ACTIONS(256),
  },
  [STATE(10)] = {
    [sym_trace32_hll_variable] = STATE(819),
    [sym__hll_expression] = STATE(632),
    [sym_hll_call_expression] = STATE(818),
    [sym_hll_field_expression] = STATE(818),
    [sym_hll_subscript_expression] = STATE(818),
    [sym_hll_cast_expression] = STATE(632),
    [sym_hll_sizeof_expression] = STATE(632),
    [sym_hll_pointer_expression] = STATE(818),
    [sym_hll_update_expression] = STATE(632),
    [sym_hll_unary_expression] = STATE(632),
    [sym_hll_binary_expression] = STATE(632),
    [sym_hll_assignment_expression] = STATE(632),
    [sym__hll_assignment_left_expression] = STATE(1708),
    [sym_hll_conditional_expression] = STATE(632),
    [sym_hll_char_literal] = STATE(632),
    [sym_hll_string_literal] = STATE(632),
    [sym_hll_parenthesized_expression] = STATE(818),
    [sym_symbol] = STATE(632),
    [sym_identifier] = ACTIONS(228),
    [anon_sym_RPAREN] = ACTIONS(258),
    [anon_sym_LPAREN] = ACTIONS(232),
    [anon_sym_EQ] = ACTIONS(260),
    [anon_sym_PLUS] = ACTIONS(236),
    [anon_sym_DASH] = ACTIONS(236),
    [anon_sym_TILDE] = ACTIONS(238),
    [anon_sym_BANG] = ACTIONS(236),
    [anon_sym_PLUS_PLUS] = ACTIONS(240),
    [anon_sym_DASH_DASH] = ACTIONS(240),
    [anon_sym_PIPE_PIPE] = ACTIONS(258),
    [anon_sym_STAR] = ACTIONS(242),
    [anon_sym_SLASH] = ACTIONS(260),
    [anon_sym_PERCENT] = ACTIONS(260),
    [anon_sym_PIPE] = ACTIONS(260),
    [anon_sym_CARET] = ACTIONS(260),
    [anon_sym_EQ_EQ] = ACTIONS(258),
    [anon_sym_BANG_EQ] = ACTIONS(258),
    [anon_sym_GT] = ACTIONS(260),
    [anon_sym_GT_EQ] = ACTIONS(258),
    [anon_sym_LT_EQ] = ACTIONS(258),
    [anon_sym_LT] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(260),
    [anon_sym_GT_GT] = ACTIONS(260),
    [anon_sym_DOT_DOT] = ACTIONS(258),
    [anon_sym_AMP_AMP] = ACTIONS(258),
    [anon_sym_AMP] = ACTIONS(242),
    [anon_sym_COLON] = ACTIONS(258),
    [anon_sym_DOT] = ACTIONS(260),
    [anon_sym_COMMA] = ACTIONS(258),
    [anon_sym_QMARK] = ACTIONS(258),
    [anon_sym_BSLASH] = ACTIONS(244),
    [anon_sym_DASH_GT] = ACTIONS(258),
    [anon_sym_LBRACK] = ACTIONS(258),
    [anon_sym_RBRACK] = ACTIONS(258),
    [anon_sym_sizeof] = ACTIONS(246),
    [anon_sym_STAR_EQ] = ACTIONS(258),
    [anon_sym_SLASH_EQ] = ACTIONS(258),
    [anon_sym_PERCENT_EQ] = ACTIONS(258),
    [anon_sym_PLUS_EQ] = ACTIONS(258),
    [anon_sym_DASH_EQ] = ACTIONS(258),
    [anon_sym_LT_LT_EQ] = ACTIONS(258),
    [anon_sym_GT_GT_EQ] = ACTIONS(258),
    [anon_sym_AMP_EQ] = ACTIONS(258),
    [anon_sym_CARET_EQ] = ACTIONS(258),
    [anon_sym_PIPE_EQ] = ACTIONS(258),
    [anon_sym_L_SQUOTE] = ACTIONS(248),
    [anon_sym_u_SQUOTE] = ACTIONS(248),
    [anon_sym_U_SQUOTE] = ACTIONS(248),
    [anon_sym_u8_SQUOTE] = ACTIONS(248),
    [anon_sym_SQUOTE] = ACTIONS(248),
    [anon_sym_L_DQUOTE] = ACTIONS(250),
    [anon_sym_u_DQUOTE] = ACTIONS(250),
    [anon_sym_U_DQUOTE] = ACTIONS(250),
    [anon_sym_u8_DQUOTE] = ACTIONS(250),
    [anon_sym_DQUOTE] = ACTIONS(250),
    [aux_sym_symbol_token1] = ACTIONS(252),
    [aux_sym_symbol_token2] = ACTIONS(254),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(260),
    [sym__hll_number_literal_pre_hook] = ACTIONS(256),
  },
  [STATE(11)] = {
    [sym_trace32_hll_variable] = STATE(819),
    [sym__hll_expression] = STATE(633),
    [sym_hll_call_expression] = STATE(818),
    [sym_hll_field_expression] = STATE(818),
    [sym_hll_subscript_expression] = STATE(818),
    [sym_hll_cast_expression] = STATE(633),
    [sym_hll_sizeof_expression] = STATE(633),
    [sym_hll_pointer_expression] = STATE(818),
    [sym_hll_update_expression] = STATE(633),
    [sym_hll_unary_expression] = STATE(633),
    [sym_hll_binary_expression] = STATE(633),
    [sym_hll_assignment_expression] = STATE(633),
    [sym__hll_assignment_left_expression] = STATE(1708),
    [sym_hll_conditional_expression] = STATE(633),
    [sym_hll_char_literal] = STATE(633),
    [sym_hll_string_literal] = STATE(633),
    [sym_hll_parenthesized_expression] = STATE(818),
    [sym_symbol] = STATE(633),
    [sym_identifier] = ACTIONS(228),
    [anon_sym_RPAREN] = ACTIONS(262),
    [anon_sym_LPAREN] = ACTIONS(232),
    [anon_sym_EQ] = ACTIONS(264),
    [anon_sym_PLUS] = ACTIONS(236),
    [anon_sym_DASH] = ACTIONS(236),
    [anon_sym_TILDE] = ACTIONS(238),
    [anon_sym_BANG] = ACTIONS(236),
    [anon_sym_PLUS_PLUS] = ACTIONS(240),
    [anon_sym_DASH_DASH] = ACTIONS(240),
    [anon_sym_PIPE_PIPE] = ACTIONS(262),
    [anon_sym_STAR] = ACTIONS(242),
    [anon_sym_SLASH] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(264),
    [anon_sym_PIPE] = ACTIONS(264),
    [anon_sym_CARET] = ACTIONS(264),
    [anon_sym_EQ_EQ] = ACTIONS(262),
    [anon_sym_BANG_EQ] = ACTIONS(262),
    [anon_sym_GT] = ACTIONS(264),
    [anon_sym_GT_EQ] = ACTIONS(262),
    [anon_sym_LT_EQ] = ACTIONS(262),
    [anon_sym_LT] = ACTIONS(264),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_GT_GT] = ACTIONS(264),
    [anon_sym_DOT_DOT] = ACTIONS(262),
    [anon_sym_AMP_AMP] = ACTIONS(262),
    [anon_sym_AMP] = ACTIONS(242),
    [anon_sym_COLON] = ACTIONS(262),
    [anon_sym_DOT] = ACTIONS(264),
    [anon_sym_COMMA] = ACTIONS(262),
    [anon_sym_QMARK] = ACTIONS(262),
    [anon_sym_BSLASH] = ACTIONS(244),
    [anon_sym_DASH_GT] = ACTIONS(262),
    [anon_sym_LBRACK] = ACTIONS(262),
    [anon_sym_RBRACK] = ACTIONS(262),
    [anon_sym_sizeof] = ACTIONS(246),
    [anon_sym_STAR_EQ] = ACTIONS(262),
    [anon_sym_SLASH_EQ] = ACTIONS(262),
    [anon_sym_PERCENT_EQ] = ACTIONS(262),
    [anon_sym_PLUS_EQ] = ACTIONS(262),
    [anon_sym_DASH_EQ] = ACTIONS(262),
    [anon_sym_LT_LT_EQ] = ACTIONS(262),
    [anon_sym_GT_GT_EQ] = ACTIONS(262),
    [anon_sym_AMP_EQ] = ACTIONS(262),
    [anon_sym_CARET_EQ] = ACTIONS(262),
    [anon_sym_PIPE_EQ] = ACTIONS(262),
    [anon_sym_L_SQUOTE] = ACTIONS(248),
    [anon_sym_u_SQUOTE] = ACTIONS(248),
    [anon_sym_U_SQUOTE] = ACTIONS(248),
    [anon_sym_u8_SQUOTE] = ACTIONS(248),
    [anon_sym_SQUOTE] = ACTIONS(248),
    [anon_sym_L_DQUOTE] = ACTIONS(250),
    [anon_sym_u_DQUOTE] = ACTIONS(250),
    [anon_sym_U_DQUOTE] = ACTIONS(250),
    [anon_sym_u8_DQUOTE] = ACTIONS(250),
    [anon_sym_DQUOTE] = ACTIONS(250),
    [aux_sym_symbol_token1] = ACTIONS(252),
    [aux_sym_symbol_token2] = ACTIONS(254),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(264),
    [sym__hll_number_literal_pre_hook] = ACTIONS(256),
  },
  [STATE(12)] = {
    [sym_trace32_hll_variable] = STATE(819),
    [sym__hll_expression] = STATE(635),
    [sym_hll_call_expression] = STATE(818),
    [sym_hll_field_expression] = STATE(818),
    [sym_hll_subscript_expression] = STATE(818),
    [sym_hll_cast_expression] = STATE(635),
    [sym_hll_sizeof_expression] = STATE(635),
    [sym_hll_pointer_expression] = STATE(818),
    [sym_hll_update_expression] = STATE(635),
    [sym_hll_unary_expression] = STATE(635),
    [sym_hll_binary_expression] = STATE(635),
    [sym_hll_assignment_expression] = STATE(635),
    [sym__hll_assignment_left_expression] = STATE(1708),
    [sym_hll_conditional_expression] = STATE(635),
    [sym_hll_char_literal] = STATE(635),
    [sym_hll_string_literal] = STATE(635),
    [sym_hll_parenthesized_expression] = STATE(818),
    [sym_symbol] = STATE(635),
    [sym_identifier] = ACTIONS(228),
    [anon_sym_RPAREN] = ACTIONS(266),
    [anon_sym_LPAREN] = ACTIONS(232),
    [anon_sym_EQ] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(236),
    [anon_sym_DASH] = ACTIONS(236),
    [anon_sym_TILDE] = ACTIONS(238),
    [anon_sym_BANG] = ACTIONS(236),
    [anon_sym_PLUS_PLUS] = ACTIONS(240),
    [anon_sym_DASH_DASH] = ACTIONS(240),
    [anon_sym_PIPE_PIPE] = ACTIONS(266),
    [anon_sym_STAR] = ACTIONS(242),
    [anon_sym_SLASH] = ACTIONS(268),
    [anon_sym_PERCENT] = ACTIONS(268),
    [anon_sym_PIPE] = ACTIONS(268),
    [anon_sym_CARET] = ACTIONS(268),
    [anon_sym_EQ_EQ] = ACTIONS(266),
    [anon_sym_BANG_EQ] = ACTIONS(266),
    [anon_sym_GT] = ACTIONS(268),
    [anon_sym_GT_EQ] = ACTIONS(266),
    [anon_sym_LT_EQ] = ACTIONS(266),
    [anon_sym_LT] = ACTIONS(268),
    [anon_sym_LT_LT] = ACTIONS(268),
    [anon_sym_GT_GT] = ACTIONS(268),
    [anon_sym_DOT_DOT] = ACTIONS(266),
    [anon_sym_AMP_AMP] = ACTIONS(266),
    [anon_sym_AMP] = ACTIONS(242),
    [anon_sym_COLON] = ACTIONS(266),
    [anon_sym_DOT] = ACTIONS(268),
    [anon_sym_COMMA] = ACTIONS(266),
    [anon_sym_QMARK] = ACTIONS(266),
    [anon_sym_BSLASH] = ACTIONS(244),
    [anon_sym_DASH_GT] = ACTIONS(266),
    [anon_sym_LBRACK] = ACTIONS(266),
    [anon_sym_RBRACK] = ACTIONS(266),
    [anon_sym_sizeof] = ACTIONS(246),
    [anon_sym_STAR_EQ] = ACTIONS(266),
    [anon_sym_SLASH_EQ] = ACTIONS(266),
    [anon_sym_PERCENT_EQ] = ACTIONS(266),
    [anon_sym_PLUS_EQ] = ACTIONS(266),
    [anon_sym_DASH_EQ] = ACTIONS(266),
    [anon_sym_LT_LT_EQ] = ACTIONS(266),
    [anon_sym_GT_GT_EQ] = ACTIONS(266),
    [anon_sym_AMP_EQ] = ACTIONS(266),
    [anon_sym_CARET_EQ] = ACTIONS(266),
    [anon_sym_PIPE_EQ] = ACTIONS(266),
    [anon_sym_L_SQUOTE] = ACTIONS(248),
    [anon_sym_u_SQUOTE] = ACTIONS(248),
    [anon_sym_U_SQUOTE] = ACTIONS(248),
    [anon_sym_u8_SQUOTE] = ACTIONS(248),
    [anon_sym_SQUOTE] = ACTIONS(248),
    [anon_sym_L_DQUOTE] = ACTIONS(250),
    [anon_sym_u_DQUOTE] = ACTIONS(250),
    [anon_sym_U_DQUOTE] = ACTIONS(250),
    [anon_sym_u8_DQUOTE] = ACTIONS(250),
    [anon_sym_DQUOTE] = ACTIONS(250),
    [aux_sym_symbol_token1] = ACTIONS(252),
    [aux_sym_symbol_token2] = ACTIONS(254),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(268),
    [sym__hll_number_literal_pre_hook] = ACTIONS(256),
  },
  [STATE(13)] = {
    [sym__statement] = STATE(1543),
    [sym_labeled_expression] = STATE(1543),
    [sym_subroutine_block] = STATE(1543),
    [sym_if_block] = STATE(1543),
    [sym_while_block] = STATE(1543),
    [sym_repeat_block] = STATE(1543),
    [sym__repeat_block_pre_condition_format] = STATE(1531),
    [sym__repeat_block_post_condition_format] = STATE(1532),
    [sym_macro_assignment_expression] = STATE(1543),
    [sym__label] = STATE(359),
    [sym_macro_definition] = STATE(1543),
    [sym_parameter_declaration] = STATE(1543),
    [sym_subroutine_call_expression] = STATE(1543),
    [sym_macro] = STATE(2316),
    [sym_recursive_macro_expansion] = STATE(1396),
    [sym_macro_definition_command] = STATE(2404),
    [sym_command_expression] = STATE(1543),
    [sym__host_command] = STATE(2629),
    [sym__host_command_with_default_format] = STATE(2493),
    [sym__other_command] = STATE(2672),
    [sym__command_identifier] = STATE(2241),
    [sym_argument_list] = STATE(432),
    [sym__var_command] = STATE(2618),
    [sym__hll_var_command] = STATE(2650),
    [sym__hll_default_format_var_command] = STATE(2670),
    [sym__hll_non_default_format_var_command] = STATE(2697),
    [sym__non_hll_var_command] = STATE(2718),
    [sym__hll_var_definition_command] = STATE(2605),
    [sym__hll_var_command_with_default_format] = STATE(2425),
    [sym__var_command_with_non_hll_format] = STATE(2429),
    [sym__hll_var_definition_command_expression] = STATE(2544),
    [sym__hll_var_if_command] = STATE(2546),
    [aux_sym_block_repeat1] = STATE(368),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(270),
    [aux_sym_subroutine_block_token1] = ACTIONS(11),
    [aux_sym_if_block_token1] = ACTIONS(13),
    [aux_sym_while_block_token1] = ACTIONS(15),
    [aux_sym_while_block_token2] = ACTIONS(17),
    [aux_sym__repeat_block_pre_condition_format_token1] = ACTIONS(19),
    [anon_sym_EQ] = ACTIONS(272),
    [anon_sym_PLUS] = ACTIONS(274),
    [anon_sym_DASH] = ACTIONS(274),
    [anon_sym_PLUS_PLUS] = ACTIONS(276),
    [anon_sym_DASH_DASH] = ACTIONS(276),
    [anon_sym_CARET_CARET] = ACTIONS(278),
    [anon_sym_PIPE_PIPE] = ACTIONS(280),
    [anon_sym_STAR] = ACTIONS(282),
    [anon_sym_SLASH] = ACTIONS(284),
    [anon_sym_PERCENT] = ACTIONS(282),
    [anon_sym_PIPE] = ACTIONS(286),
    [anon_sym_CARET] = ACTIONS(288),
    [anon_sym_EQ_EQ] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(292),
    [anon_sym_GT_EQ] = ACTIONS(290),
    [anon_sym_LT_EQ] = ACTIONS(290),
    [anon_sym_LT] = ACTIONS(292),
    [anon_sym_LT_LT] = ACTIONS(294),
    [anon_sym_GT_GT] = ACTIONS(294),
    [anon_sym_DOT_DOT] = ACTIONS(296),
    [anon_sym_AMP_AMP] = ACTIONS(21),
    [anon_sym_AMP] = ACTIONS(23),
    [aux_sym_macro_definition_token1] = ACTIONS(25),
    [aux_sym_parameter_declaration_token1] = ACTIONS(27),
    [aux_sym_parameter_declaration_token2] = ACTIONS(27),
    [aux_sym_parameter_declaration_token3] = ACTIONS(29),
    [aux_sym_subroutine_call_expression_token1] = ACTIONS(31),
    [aux_sym_macro_definition_command_token1] = ACTIONS(33),
    [aux_sym_macro_definition_command_token2] = ACTIONS(33),
    [aux_sym_macro_definition_command_token3] = ACTIONS(33),
    [aux_sym__host_command_with_default_format_token1] = ACTIONS(35),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(37),
    [sym_comment] = ACTIONS(298),
    [sym__blank_line] = ACTIONS(298),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(300),
    [sym_label_identifier] = ACTIONS(43),
    [sym__and_operator_pre_hook] = ACTIONS(302),
  },
  [STATE(14)] = {
    [sym__statement] = STATE(1543),
    [sym_labeled_expression] = STATE(1543),
    [sym_subroutine_block] = STATE(1543),
    [sym_if_block] = STATE(1543),
    [sym_while_block] = STATE(1543),
    [sym_repeat_block] = STATE(1543),
    [sym__repeat_block_pre_condition_format] = STATE(1531),
    [sym__repeat_block_post_condition_format] = STATE(1532),
    [sym_macro_assignment_expression] = STATE(1543),
    [sym__label] = STATE(357),
    [sym_macro_definition] = STATE(1543),
    [sym_parameter_declaration] = STATE(1543),
    [sym_subroutine_call_expression] = STATE(1543),
    [sym_macro] = STATE(2316),
    [sym_recursive_macro_expansion] = STATE(1396),
    [sym_macro_definition_command] = STATE(2404),
    [sym_command_expression] = STATE(1543),
    [sym__host_command] = STATE(2629),
    [sym__host_command_with_default_format] = STATE(2493),
    [sym__other_command] = STATE(2672),
    [sym__command_identifier] = STATE(2241),
    [sym_argument_list] = STATE(432),
    [sym__var_command] = STATE(2618),
    [sym__hll_var_command] = STATE(2650),
    [sym__hll_default_format_var_command] = STATE(2670),
    [sym__hll_non_default_format_var_command] = STATE(2697),
    [sym__non_hll_var_command] = STATE(2718),
    [sym__hll_var_definition_command] = STATE(2605),
    [sym__hll_var_command_with_default_format] = STATE(2425),
    [sym__var_command_with_non_hll_format] = STATE(2429),
    [sym__hll_var_definition_command_expression] = STATE(2544),
    [sym__hll_var_if_command] = STATE(2567),
    [aux_sym_block_repeat1] = STATE(370),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(270),
    [aux_sym_subroutine_block_token1] = ACTIONS(11),
    [aux_sym_if_block_token1] = ACTIONS(304),
    [aux_sym_while_block_token1] = ACTIONS(306),
    [aux_sym_while_block_token2] = ACTIONS(308),
    [aux_sym__repeat_block_pre_condition_format_token1] = ACTIONS(310),
    [anon_sym_EQ] = ACTIONS(272),
    [anon_sym_PLUS] = ACTIONS(274),
    [anon_sym_DASH] = ACTIONS(274),
    [anon_sym_PLUS_PLUS] = ACTIONS(276),
    [anon_sym_DASH_DASH] = ACTIONS(276),
    [anon_sym_CARET_CARET] = ACTIONS(278),
    [anon_sym_PIPE_PIPE] = ACTIONS(280),
    [anon_sym_STAR] = ACTIONS(282),
    [anon_sym_SLASH] = ACTIONS(284),
    [anon_sym_PERCENT] = ACTIONS(282),
    [anon_sym_PIPE] = ACTIONS(286),
    [anon_sym_CARET] = ACTIONS(288),
    [anon_sym_EQ_EQ] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(292),
    [anon_sym_GT_EQ] = ACTIONS(290),
    [anon_sym_LT_EQ] = ACTIONS(290),
    [anon_sym_LT] = ACTIONS(292),
    [anon_sym_LT_LT] = ACTIONS(294),
    [anon_sym_GT_GT] = ACTIONS(294),
    [anon_sym_DOT_DOT] = ACTIONS(296),
    [anon_sym_AMP_AMP] = ACTIONS(21),
    [anon_sym_AMP] = ACTIONS(23),
    [aux_sym_macro_definition_token1] = ACTIONS(25),
    [aux_sym_parameter_declaration_token1] = ACTIONS(27),
    [aux_sym_parameter_declaration_token2] = ACTIONS(27),
    [aux_sym_parameter_declaration_token3] = ACTIONS(29),
    [aux_sym_subroutine_call_expression_token1] = ACTIONS(31),
    [aux_sym_macro_definition_command_token1] = ACTIONS(33),
    [aux_sym_macro_definition_command_token2] = ACTIONS(33),
    [aux_sym_macro_definition_command_token3] = ACTIONS(33),
    [aux_sym__host_command_with_default_format_token1] = ACTIONS(35),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(37),
    [sym_comment] = ACTIONS(312),
    [sym__blank_line] = ACTIONS(312),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(314),
    [sym_label_identifier] = ACTIONS(43),
    [sym__and_operator_pre_hook] = ACTIONS(302),
  },
  [STATE(15)] = {
    [sym__statement] = STATE(1543),
    [sym_labeled_expression] = STATE(1543),
    [sym_subroutine_block] = STATE(1543),
    [sym_if_block] = STATE(1543),
    [sym_while_block] = STATE(1543),
    [sym_repeat_block] = STATE(1543),
    [sym__repeat_block_pre_condition_format] = STATE(1531),
    [sym__repeat_block_post_condition_format] = STATE(1532),
    [sym_macro_assignment_expression] = STATE(1543),
    [sym__label] = STATE(358),
    [sym_macro_definition] = STATE(1543),
    [sym_parameter_declaration] = STATE(1543),
    [sym_subroutine_call_expression] = STATE(1543),
    [sym_macro] = STATE(2316),
    [sym_recursive_macro_expansion] = STATE(1396),
    [sym_macro_definition_command] = STATE(2404),
    [sym_command_expression] = STATE(1543),
    [sym__host_command] = STATE(2629),
    [sym__host_command_with_default_format] = STATE(2493),
    [sym__other_command] = STATE(2672),
    [sym__command_identifier] = STATE(2241),
    [sym_argument_list] = STATE(432),
    [sym__var_command] = STATE(2618),
    [sym__hll_var_command] = STATE(2650),
    [sym__hll_default_format_var_command] = STATE(2670),
    [sym__hll_non_default_format_var_command] = STATE(2697),
    [sym__non_hll_var_command] = STATE(2718),
    [sym__hll_var_definition_command] = STATE(2605),
    [sym__hll_var_command_with_default_format] = STATE(2425),
    [sym__var_command_with_non_hll_format] = STATE(2429),
    [sym__hll_var_definition_command_expression] = STATE(2544),
    [sym__hll_var_if_command] = STATE(2572),
    [aux_sym_block_repeat1] = STATE(371),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(270),
    [aux_sym_subroutine_block_token1] = ACTIONS(11),
    [aux_sym_if_block_token1] = ACTIONS(316),
    [aux_sym_while_block_token1] = ACTIONS(318),
    [aux_sym_while_block_token2] = ACTIONS(320),
    [aux_sym__repeat_block_pre_condition_format_token1] = ACTIONS(322),
    [anon_sym_EQ] = ACTIONS(272),
    [anon_sym_PLUS] = ACTIONS(274),
    [anon_sym_DASH] = ACTIONS(274),
    [anon_sym_PLUS_PLUS] = ACTIONS(276),
    [anon_sym_DASH_DASH] = ACTIONS(276),
    [anon_sym_CARET_CARET] = ACTIONS(278),
    [anon_sym_PIPE_PIPE] = ACTIONS(280),
    [anon_sym_STAR] = ACTIONS(282),
    [anon_sym_SLASH] = ACTIONS(284),
    [anon_sym_PERCENT] = ACTIONS(282),
    [anon_sym_PIPE] = ACTIONS(286),
    [anon_sym_CARET] = ACTIONS(288),
    [anon_sym_EQ_EQ] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(292),
    [anon_sym_GT_EQ] = ACTIONS(290),
    [anon_sym_LT_EQ] = ACTIONS(290),
    [anon_sym_LT] = ACTIONS(292),
    [anon_sym_LT_LT] = ACTIONS(294),
    [anon_sym_GT_GT] = ACTIONS(294),
    [anon_sym_DOT_DOT] = ACTIONS(296),
    [anon_sym_AMP_AMP] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(23),
    [aux_sym_macro_definition_token1] = ACTIONS(25),
    [aux_sym_parameter_declaration_token1] = ACTIONS(27),
    [aux_sym_parameter_declaration_token2] = ACTIONS(27),
    [aux_sym_parameter_declaration_token3] = ACTIONS(29),
    [aux_sym_subroutine_call_expression_token1] = ACTIONS(31),
    [aux_sym_macro_definition_command_token1] = ACTIONS(33),
    [aux_sym_macro_definition_command_token2] = ACTIONS(33),
    [aux_sym_macro_definition_command_token3] = ACTIONS(33),
    [aux_sym__host_command_with_default_format_token1] = ACTIONS(35),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(37),
    [sym_comment] = ACTIONS(326),
    [sym__blank_line] = ACTIONS(326),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(328),
    [sym_label_identifier] = ACTIONS(43),
    [sym__and_operator_pre_hook] = ACTIONS(302),
  },
  [STATE(16)] = {
    [sym__statement] = STATE(1543),
    [sym_labeled_expression] = STATE(1543),
    [sym_subroutine_block] = STATE(1543),
    [sym_if_block] = STATE(1543),
    [sym_while_block] = STATE(1543),
    [sym_repeat_block] = STATE(1543),
    [sym__repeat_block_pre_condition_format] = STATE(1531),
    [sym__repeat_block_post_condition_format] = STATE(1532),
    [sym_macro_assignment_expression] = STATE(1543),
    [sym__label] = STATE(360),
    [sym_macro_definition] = STATE(1543),
    [sym_parameter_declaration] = STATE(1543),
    [sym_subroutine_call_expression] = STATE(1543),
    [sym_macro] = STATE(2316),
    [sym_recursive_macro_expansion] = STATE(1396),
    [sym_macro_definition_command] = STATE(2404),
    [sym_command_expression] = STATE(1543),
    [sym__host_command] = STATE(2629),
    [sym__host_command_with_default_format] = STATE(2493),
    [sym__other_command] = STATE(2672),
    [sym__command_identifier] = STATE(2241),
    [sym_argument_list] = STATE(432),
    [sym__var_command] = STATE(2618),
    [sym__hll_var_command] = STATE(2650),
    [sym__hll_default_format_var_command] = STATE(2670),
    [sym__hll_non_default_format_var_command] = STATE(2697),
    [sym__non_hll_var_command] = STATE(2718),
    [sym__hll_var_definition_command] = STATE(2605),
    [sym__hll_var_command_with_default_format] = STATE(2425),
    [sym__var_command_with_non_hll_format] = STATE(2429),
    [sym__hll_var_definition_command_expression] = STATE(2544),
    [sym__hll_var_if_command] = STATE(2553),
    [aux_sym_block_repeat1] = STATE(364),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(270),
    [aux_sym_subroutine_block_token1] = ACTIONS(11),
    [aux_sym_if_block_token1] = ACTIONS(330),
    [aux_sym_while_block_token1] = ACTIONS(332),
    [aux_sym_while_block_token2] = ACTIONS(334),
    [aux_sym__repeat_block_pre_condition_format_token1] = ACTIONS(336),
    [anon_sym_EQ] = ACTIONS(272),
    [anon_sym_PLUS] = ACTIONS(274),
    [anon_sym_DASH] = ACTIONS(274),
    [anon_sym_PLUS_PLUS] = ACTIONS(276),
    [anon_sym_DASH_DASH] = ACTIONS(276),
    [anon_sym_CARET_CARET] = ACTIONS(278),
    [anon_sym_PIPE_PIPE] = ACTIONS(280),
    [anon_sym_STAR] = ACTIONS(282),
    [anon_sym_SLASH] = ACTIONS(284),
    [anon_sym_PERCENT] = ACTIONS(282),
    [anon_sym_PIPE] = ACTIONS(286),
    [anon_sym_CARET] = ACTIONS(288),
    [anon_sym_EQ_EQ] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(292),
    [anon_sym_GT_EQ] = ACTIONS(290),
    [anon_sym_LT_EQ] = ACTIONS(290),
    [anon_sym_LT] = ACTIONS(292),
    [anon_sym_LT_LT] = ACTIONS(294),
    [anon_sym_GT_GT] = ACTIONS(294),
    [anon_sym_DOT_DOT] = ACTIONS(296),
    [anon_sym_AMP_AMP] = ACTIONS(324),
    [anon_sym_AMP] = ACTIONS(23),
    [aux_sym_macro_definition_token1] = ACTIONS(25),
    [aux_sym_parameter_declaration_token1] = ACTIONS(27),
    [aux_sym_parameter_declaration_token2] = ACTIONS(27),
    [aux_sym_parameter_declaration_token3] = ACTIONS(29),
    [aux_sym_subroutine_call_expression_token1] = ACTIONS(31),
    [aux_sym_macro_definition_command_token1] = ACTIONS(33),
    [aux_sym_macro_definition_command_token2] = ACTIONS(33),
    [aux_sym_macro_definition_command_token3] = ACTIONS(33),
    [aux_sym__host_command_with_default_format_token1] = ACTIONS(35),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(37),
    [sym_comment] = ACTIONS(338),
    [sym__blank_line] = ACTIONS(338),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(340),
    [sym_label_identifier] = ACTIONS(43),
    [sym__and_operator_pre_hook] = ACTIONS(302),
  },
  [STATE(17)] = {
    [sym__statement] = STATE(2099),
    [sym_labeled_expression] = STATE(2099),
    [sym_subroutine_block] = STATE(2099),
    [sym_if_block] = STATE(2099),
    [sym_while_block] = STATE(2099),
    [sym_repeat_block] = STATE(2099),
    [sym__repeat_block_pre_condition_format] = STATE(2140),
    [sym__repeat_block_post_condition_format] = STATE(2141),
    [sym_macro_assignment_expression] = STATE(2099),
    [sym__label] = STATE(353),
    [sym_macro_definition] = STATE(2099),
    [sym_parameter_declaration] = STATE(2099),
    [sym_subroutine_call_expression] = STATE(2099),
    [sym_macro] = STATE(2409),
    [sym_recursive_macro_expansion] = STATE(1875),
    [sym_macro_definition_command] = STATE(2348),
    [sym_command_expression] = STATE(2099),
    [sym__host_command] = STATE(2657),
    [sym__host_command_with_default_format] = STATE(2493),
    [sym__other_command] = STATE(2658),
    [sym__command_identifier] = STATE(2241),
    [sym_argument_list] = STATE(432),
    [sym__var_command] = STATE(2677),
    [sym__hll_var_command] = STATE(2650),
    [sym__hll_default_format_var_command] = STATE(2670),
    [sym__hll_non_default_format_var_command] = STATE(2697),
    [sym__non_hll_var_command] = STATE(2718),
    [sym__hll_var_definition_command] = STATE(2605),
    [sym__hll_var_command_with_default_format] = STATE(2425),
    [sym__var_command_with_non_hll_format] = STATE(2429),
    [sym__hll_var_definition_command_expression] = STATE(2544),
    [sym__hll_var_if_command] = STATE(2575),
    [aux_sym_block_repeat1] = STATE(363),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(270),
    [aux_sym_subroutine_block_token1] = ACTIONS(342),
    [aux_sym_if_block_token1] = ACTIONS(344),
    [aux_sym_while_block_token1] = ACTIONS(346),
    [aux_sym_while_block_token2] = ACTIONS(348),
    [aux_sym__repeat_block_pre_condition_format_token1] = ACTIONS(350),
    [anon_sym_EQ] = ACTIONS(272),
    [anon_sym_PLUS] = ACTIONS(274),
    [anon_sym_DASH] = ACTIONS(274),
    [anon_sym_PLUS_PLUS] = ACTIONS(276),
    [anon_sym_DASH_DASH] = ACTIONS(276),
    [anon_sym_CARET_CARET] = ACTIONS(278),
    [anon_sym_PIPE_PIPE] = ACTIONS(280),
    [anon_sym_STAR] = ACTIONS(282),
    [anon_sym_SLASH] = ACTIONS(284),
    [anon_sym_PERCENT] = ACTIONS(282),
    [anon_sym_PIPE] = ACTIONS(286),
    [anon_sym_CARET] = ACTIONS(288),
    [anon_sym_EQ_EQ] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(292),
    [anon_sym_GT_EQ] = ACTIONS(290),
    [anon_sym_LT_EQ] = ACTIONS(290),
    [anon_sym_LT] = ACTIONS(292),
    [anon_sym_LT_LT] = ACTIONS(294),
    [anon_sym_GT_GT] = ACTIONS(294),
    [anon_sym_DOT_DOT] = ACTIONS(296),
    [anon_sym_AMP_AMP] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(23),
    [aux_sym_macro_definition_token1] = ACTIONS(354),
    [aux_sym_parameter_declaration_token1] = ACTIONS(356),
    [aux_sym_parameter_declaration_token2] = ACTIONS(356),
    [aux_sym_parameter_declaration_token3] = ACTIONS(358),
    [aux_sym_subroutine_call_expression_token1] = ACTIONS(360),
    [aux_sym_macro_definition_command_token1] = ACTIONS(33),
    [aux_sym_macro_definition_command_token2] = ACTIONS(33),
    [aux_sym_macro_definition_command_token3] = ACTIONS(33),
    [aux_sym__host_command_with_default_format_token1] = ACTIONS(35),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(37),
    [sym_comment] = ACTIONS(362),
    [sym__blank_line] = ACTIONS(362),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(364),
    [sym_label_identifier] = ACTIONS(43),
    [sym__and_operator_pre_hook] = ACTIONS(302),
  },
  [STATE(18)] = {
    [sym__statement] = STATE(2099),
    [sym_labeled_expression] = STATE(2099),
    [sym_subroutine_block] = STATE(2099),
    [sym_if_block] = STATE(2099),
    [sym_while_block] = STATE(2099),
    [sym_repeat_block] = STATE(2099),
    [sym__repeat_block_pre_condition_format] = STATE(2140),
    [sym__repeat_block_post_condition_format] = STATE(2141),
    [sym_macro_assignment_expression] = STATE(2099),
    [sym__label] = STATE(356),
    [sym_macro_definition] = STATE(2099),
    [sym_parameter_declaration] = STATE(2099),
    [sym_subroutine_call_expression] = STATE(2099),
    [sym_macro] = STATE(2409),
    [sym_recursive_macro_expansion] = STATE(1875),
    [sym_macro_definition_command] = STATE(2348),
    [sym_command_expression] = STATE(2099),
    [sym__host_command] = STATE(2657),
    [sym__host_command_with_default_format] = STATE(2493),
    [sym__other_command] = STATE(2658),
    [sym__command_identifier] = STATE(2241),
    [sym_argument_list] = STATE(432),
    [sym__var_command] = STATE(2677),
    [sym__hll_var_command] = STATE(2650),
    [sym__hll_default_format_var_command] = STATE(2670),
    [sym__hll_non_default_format_var_command] = STATE(2697),
    [sym__non_hll_var_command] = STATE(2718),
    [sym__hll_var_definition_command] = STATE(2605),
    [sym__hll_var_command_with_default_format] = STATE(2425),
    [sym__var_command_with_non_hll_format] = STATE(2429),
    [sym__hll_var_definition_command_expression] = STATE(2544),
    [sym__hll_var_if_command] = STATE(2577),
    [aux_sym_block_repeat1] = STATE(369),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_LPAREN] = ACTIONS(270),
    [aux_sym_subroutine_block_token1] = ACTIONS(342),
    [aux_sym_if_block_token1] = ACTIONS(366),
    [aux_sym_while_block_token1] = ACTIONS(368),
    [aux_sym_while_block_token2] = ACTIONS(370),
    [aux_sym__repeat_block_pre_condition_format_token1] = ACTIONS(372),
    [anon_sym_EQ] = ACTIONS(272),
    [anon_sym_PLUS] = ACTIONS(274),
    [anon_sym_DASH] = ACTIONS(274),
    [anon_sym_PLUS_PLUS] = ACTIONS(276),
    [anon_sym_DASH_DASH] = ACTIONS(276),
    [anon_sym_CARET_CARET] = ACTIONS(278),
    [anon_sym_PIPE_PIPE] = ACTIONS(280),
    [anon_sym_STAR] = ACTIONS(282),
    [anon_sym_SLASH] = ACTIONS(284),
    [anon_sym_PERCENT] = ACTIONS(282),
    [anon_sym_PIPE] = ACTIONS(286),
    [anon_sym_CARET] = ACTIONS(288),
    [anon_sym_EQ_EQ] = ACTIONS(290),
    [anon_sym_BANG_EQ] = ACTIONS(290),
    [anon_sym_GT] = ACTIONS(292),
    [anon_sym_GT_EQ] = ACTIONS(290),
    [anon_sym_LT_EQ] = ACTIONS(290),
    [anon_sym_LT] = ACTIONS(292),
    [anon_sym_LT_LT] = ACTIONS(294),
    [anon_sym_GT_GT] = ACTIONS(294),
    [anon_sym_DOT_DOT] = ACTIONS(296),
    [anon_sym_AMP_AMP] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(23),
    [aux_sym_macro_definition_token1] = ACTIONS(354),
    [aux_sym_parameter_declaration_token1] = ACTIONS(356),
    [aux_sym_parameter_declaration_token2] = ACTIONS(356),
    [aux_sym_parameter_declaration_token3] = ACTIONS(358),
    [aux_sym_subroutine_call_expression_token1] = ACTIONS(360),
    [aux_sym_macro_definition_command_token1] = ACTIONS(33),
    [aux_sym_macro_definition_command_token2] = ACTIONS(33),
    [aux_sym_macro_definition_command_token3] = ACTIONS(33),
    [aux_sym__host_command_with_default_format_token1] = ACTIONS(35),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(37),
    [sym_comment] = ACTIONS(374),
    [sym__blank_line] = ACTIONS(374),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(376),
    [sym_label_identifier] = ACTIONS(43),
    [sym__and_operator_pre_hook] = ACTIONS(302),
  },
  [STATE(19)] = {
    [sym_trace32_hll_variable] = STATE(882),
    [sym__hll_expression] = STATE(840),
    [sym_hll_call_expression] = STATE(881),
    [sym_hll_field_expression] = STATE(881),
    [sym_hll_subscript_expression] = STATE(881),
    [sym_hll_cast_expression] = STATE(840),
    [sym_hll_sizeof_expression] = STATE(840),
    [sym_hll_pointer_expression] = STATE(881),
    [sym_hll_update_expression] = STATE(840),
    [sym_hll_unary_expression] = STATE(840),
    [sym_hll_binary_expression] = STATE(840),
    [sym_hll_assignment_expression] = STATE(840),
    [sym__hll_assignment_left_expression] = STATE(1701),
    [sym_hll_conditional_expression] = STATE(840),
    [sym_hll_char_literal] = STATE(840),
    [sym_hll_string_literal] = STATE(840),
    [sym_hll_parenthesized_expression] = STATE(881),
    [sym_symbol] = STATE(840),
    [sym_identifier] = ACTIONS(378),
    [anon_sym_LPAREN] = ACTIONS(380),
    [anon_sym_EQ] = ACTIONS(234),
    [anon_sym_PLUS] = ACTIONS(382),
    [anon_sym_DASH] = ACTIONS(382),
    [anon_sym_TILDE] = ACTIONS(384),
    [anon_sym_BANG] = ACTIONS(382),
    [anon_sym_PLUS_PLUS] = ACTIONS(386),
    [anon_sym_DASH_DASH] = ACTIONS(386),
    [anon_sym_PIPE_PIPE] = ACTIONS(230),
    [anon_sym_STAR] = ACTIONS(388),
    [anon_sym_SLASH] = ACTIONS(234),
    [anon_sym_PERCENT] = ACTIONS(234),
    [anon_sym_PIPE] = ACTIONS(234),
    [anon_sym_CARET] = ACTIONS(234),
    [anon_sym_EQ_EQ] = ACTIONS(230),
    [anon_sym_BANG_EQ] = ACTIONS(230),
    [anon_sym_GT] = ACTIONS(234),
    [anon_sym_GT_EQ] = ACTIONS(230),
    [anon_sym_LT_EQ] = ACTIONS(230),
    [anon_sym_LT] = ACTIONS(234),
    [anon_sym_LT_LT] = ACTIONS(234),
    [anon_sym_GT_GT] = ACTIONS(234),
    [anon_sym_AMP_AMP] = ACTIONS(230),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT] = ACTIONS(230),
    [anon_sym_QMARK] = ACTIONS(230),
    [anon_sym_BSLASH] = ACTIONS(390),
    [anon_sym_DASH_GT] = ACTIONS(230),
    [anon_sym_LBRACK] = ACTIONS(230),
    [anon_sym_sizeof] = ACTIONS(392),
    [anon_sym_STAR_EQ] = ACTIONS(230),
    [anon_sym_SLASH_EQ] = ACTIONS(230),
    [anon_sym_PERCENT_EQ] = ACTIONS(230),
    [anon_sym_PLUS_EQ] = ACTIONS(230),
    [anon_sym_DASH_EQ] = ACTIONS(230),
    [anon_sym_LT_LT_EQ] = ACTIONS(230),
    [anon_sym_GT_GT_EQ] = ACTIONS(230),
    [anon_sym_AMP_EQ] = ACTIONS(230),
    [anon_sym_CARET_EQ] = ACTIONS(230),
    [anon_sym_PIPE_EQ] = ACTIONS(230),
    [anon_sym_L_SQUOTE] = ACTIONS(394),
    [anon_sym_u_SQUOTE] = ACTIONS(394),
    [anon_sym_U_SQUOTE] = ACTIONS(394),
    [anon_sym_u8_SQUOTE] = ACTIONS(394),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_L_DQUOTE] = ACTIONS(396),
    [anon_sym_u_DQUOTE] = ACTIONS(396),
    [anon_sym_U_DQUOTE] = ACTIONS(396),
    [anon_sym_u8_DQUOTE] = ACTIONS(396),
    [anon_sym_DQUOTE] = ACTIONS(396),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(230),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(234),
    [aux_sym_symbol_token1] = ACTIONS(398),
    [aux_sym_symbol_token2] = ACTIONS(400),
    [sym_comment] = ACTIONS(230),
    [sym__blank_line] = ACTIONS(230),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(234),
    [sym__hll_number_literal_pre_hook] = ACTIONS(402),
  },
  [STATE(20)] = {
    [sym_trace32_hll_variable] = STATE(882),
    [sym__hll_expression] = STATE(847),
    [sym_hll_call_expression] = STATE(881),
    [sym_hll_field_expression] = STATE(881),
    [sym_hll_subscript_expression] = STATE(881),
    [sym_hll_cast_expression] = STATE(847),
    [sym_hll_sizeof_expression] = STATE(847),
    [sym_hll_pointer_expression] = STATE(881),
    [sym_hll_update_expression] = STATE(847),
    [sym_hll_unary_expression] = STATE(847),
    [sym_hll_binary_expression] = STATE(847),
    [sym_hll_assignment_expression] = STATE(847),
    [sym__hll_assignment_left_expression] = STATE(1701),
    [sym_hll_conditional_expression] = STATE(847),
    [sym_hll_char_literal] = STATE(847),
    [sym_hll_string_literal] = STATE(847),
    [sym_hll_parenthesized_expression] = STATE(881),
    [sym_symbol] = STATE(847),
    [sym_identifier] = ACTIONS(378),
    [anon_sym_LPAREN] = ACTIONS(380),
    [anon_sym_EQ] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(382),
    [anon_sym_DASH] = ACTIONS(382),
    [anon_sym_TILDE] = ACTIONS(384),
    [anon_sym_BANG] = ACTIONS(382),
    [anon_sym_PLUS_PLUS] = ACTIONS(386),
    [anon_sym_DASH_DASH] = ACTIONS(386),
    [anon_sym_PIPE_PIPE] = ACTIONS(266),
    [anon_sym_STAR] = ACTIONS(388),
    [anon_sym_SLASH] = ACTIONS(268),
    [anon_sym_PERCENT] = ACTIONS(268),
    [anon_sym_PIPE] = ACTIONS(268),
    [anon_sym_CARET] = ACTIONS(268),
    [anon_sym_EQ_EQ] = ACTIONS(266),
    [anon_sym_BANG_EQ] = ACTIONS(266),
    [anon_sym_GT] = ACTIONS(268),
    [anon_sym_GT_EQ] = ACTIONS(266),
    [anon_sym_LT_EQ] = ACTIONS(266),
    [anon_sym_LT] = ACTIONS(268),
    [anon_sym_LT_LT] = ACTIONS(268),
    [anon_sym_GT_GT] = ACTIONS(268),
    [anon_sym_AMP_AMP] = ACTIONS(266),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT] = ACTIONS(266),
    [anon_sym_QMARK] = ACTIONS(266),
    [anon_sym_BSLASH] = ACTIONS(390),
    [anon_sym_DASH_GT] = ACTIONS(266),
    [anon_sym_LBRACK] = ACTIONS(266),
    [anon_sym_sizeof] = ACTIONS(392),
    [anon_sym_STAR_EQ] = ACTIONS(266),
    [anon_sym_SLASH_EQ] = ACTIONS(266),
    [anon_sym_PERCENT_EQ] = ACTIONS(266),
    [anon_sym_PLUS_EQ] = ACTIONS(266),
    [anon_sym_DASH_EQ] = ACTIONS(266),
    [anon_sym_LT_LT_EQ] = ACTIONS(266),
    [anon_sym_GT_GT_EQ] = ACTIONS(266),
    [anon_sym_AMP_EQ] = ACTIONS(266),
    [anon_sym_CARET_EQ] = ACTIONS(266),
    [anon_sym_PIPE_EQ] = ACTIONS(266),
    [anon_sym_L_SQUOTE] = ACTIONS(394),
    [anon_sym_u_SQUOTE] = ACTIONS(394),
    [anon_sym_U_SQUOTE] = ACTIONS(394),
    [anon_sym_u8_SQUOTE] = ACTIONS(394),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_L_DQUOTE] = ACTIONS(396),
    [anon_sym_u_DQUOTE] = ACTIONS(396),
    [anon_sym_U_DQUOTE] = ACTIONS(396),
    [anon_sym_u8_DQUOTE] = ACTIONS(396),
    [anon_sym_DQUOTE] = ACTIONS(396),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(266),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(268),
    [aux_sym_symbol_token1] = ACTIONS(398),
    [aux_sym_symbol_token2] = ACTIONS(400),
    [sym_comment] = ACTIONS(266),
    [sym__blank_line] = ACTIONS(266),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(268),
    [sym__hll_number_literal_pre_hook] = ACTIONS(402),
  },
  [STATE(21)] = {
    [sym_trace32_hll_variable] = STATE(882),
    [sym__hll_expression] = STATE(846),
    [sym_hll_call_expression] = STATE(881),
    [sym_hll_field_expression] = STATE(881),
    [sym_hll_subscript_expression] = STATE(881),
    [sym_hll_cast_expression] = STATE(846),
    [sym_hll_sizeof_expression] = STATE(846),
    [sym_hll_pointer_expression] = STATE(881),
    [sym_hll_update_expression] = STATE(846),
    [sym_hll_unary_expression] = STATE(846),
    [sym_hll_binary_expression] = STATE(846),
    [sym_hll_assignment_expression] = STATE(846),
    [sym__hll_assignment_left_expression] = STATE(1701),
    [sym_hll_conditional_expression] = STATE(846),
    [sym_hll_char_literal] = STATE(846),
    [sym_hll_string_literal] = STATE(846),
    [sym_hll_parenthesized_expression] = STATE(881),
    [sym_symbol] = STATE(846),
    [sym_identifier] = ACTIONS(378),
    [anon_sym_LPAREN] = ACTIONS(380),
    [anon_sym_EQ] = ACTIONS(264),
    [anon_sym_PLUS] = ACTIONS(382),
    [anon_sym_DASH] = ACTIONS(382),
    [anon_sym_TILDE] = ACTIONS(384),
    [anon_sym_BANG] = ACTIONS(382),
    [anon_sym_PLUS_PLUS] = ACTIONS(386),
    [anon_sym_DASH_DASH] = ACTIONS(386),
    [anon_sym_PIPE_PIPE] = ACTIONS(262),
    [anon_sym_STAR] = ACTIONS(388),
    [anon_sym_SLASH] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(264),
    [anon_sym_PIPE] = ACTIONS(264),
    [anon_sym_CARET] = ACTIONS(264),
    [anon_sym_EQ_EQ] = ACTIONS(262),
    [anon_sym_BANG_EQ] = ACTIONS(262),
    [anon_sym_GT] = ACTIONS(264),
    [anon_sym_GT_EQ] = ACTIONS(262),
    [anon_sym_LT_EQ] = ACTIONS(262),
    [anon_sym_LT] = ACTIONS(264),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_GT_GT] = ACTIONS(264),
    [anon_sym_AMP_AMP] = ACTIONS(262),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT] = ACTIONS(262),
    [anon_sym_QMARK] = ACTIONS(262),
    [anon_sym_BSLASH] = ACTIONS(390),
    [anon_sym_DASH_GT] = ACTIONS(262),
    [anon_sym_LBRACK] = ACTIONS(262),
    [anon_sym_sizeof] = ACTIONS(392),
    [anon_sym_STAR_EQ] = ACTIONS(262),
    [anon_sym_SLASH_EQ] = ACTIONS(262),
    [anon_sym_PERCENT_EQ] = ACTIONS(262),
    [anon_sym_PLUS_EQ] = ACTIONS(262),
    [anon_sym_DASH_EQ] = ACTIONS(262),
    [anon_sym_LT_LT_EQ] = ACTIONS(262),
    [anon_sym_GT_GT_EQ] = ACTIONS(262),
    [anon_sym_AMP_EQ] = ACTIONS(262),
    [anon_sym_CARET_EQ] = ACTIONS(262),
    [anon_sym_PIPE_EQ] = ACTIONS(262),
    [anon_sym_L_SQUOTE] = ACTIONS(394),
    [anon_sym_u_SQUOTE] = ACTIONS(394),
    [anon_sym_U_SQUOTE] = ACTIONS(394),
    [anon_sym_u8_SQUOTE] = ACTIONS(394),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_L_DQUOTE] = ACTIONS(396),
    [anon_sym_u_DQUOTE] = ACTIONS(396),
    [anon_sym_U_DQUOTE] = ACTIONS(396),
    [anon_sym_u8_DQUOTE] = ACTIONS(396),
    [anon_sym_DQUOTE] = ACTIONS(396),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(262),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(264),
    [aux_sym_symbol_token1] = ACTIONS(398),
    [aux_sym_symbol_token2] = ACTIONS(400),
    [sym_comment] = ACTIONS(262),
    [sym__blank_line] = ACTIONS(262),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(264),
    [sym__hll_number_literal_pre_hook] = ACTIONS(402),
  },
  [STATE(22)] = {
    [sym_trace32_hll_variable] = STATE(882),
    [sym__hll_expression] = STATE(845),
    [sym_hll_call_expression] = STATE(881),
    [sym_hll_field_expression] = STATE(881),
    [sym_hll_subscript_expression] = STATE(881),
    [sym_hll_cast_expression] = STATE(845),
    [sym_hll_sizeof_expression] = STATE(845),
    [sym_hll_pointer_expression] = STATE(881),
    [sym_hll_update_expression] = STATE(845),
    [sym_hll_unary_expression] = STATE(845),
    [sym_hll_binary_expression] = STATE(845),
    [sym_hll_assignment_expression] = STATE(845),
    [sym__hll_assignment_left_expression] = STATE(1701),
    [sym_hll_conditional_expression] = STATE(845),
    [sym_hll_char_literal] = STATE(845),
    [sym_hll_string_literal] = STATE(845),
    [sym_hll_parenthesized_expression] = STATE(881),
    [sym_symbol] = STATE(845),
    [sym_identifier] = ACTIONS(378),
    [anon_sym_LPAREN] = ACTIONS(380),
    [anon_sym_EQ] = ACTIONS(260),
    [anon_sym_PLUS] = ACTIONS(382),
    [anon_sym_DASH] = ACTIONS(382),
    [anon_sym_TILDE] = ACTIONS(384),
    [anon_sym_BANG] = ACTIONS(382),
    [anon_sym_PLUS_PLUS] = ACTIONS(386),
    [anon_sym_DASH_DASH] = ACTIONS(386),
    [anon_sym_PIPE_PIPE] = ACTIONS(258),
    [anon_sym_STAR] = ACTIONS(388),
    [anon_sym_SLASH] = ACTIONS(260),
    [anon_sym_PERCENT] = ACTIONS(260),
    [anon_sym_PIPE] = ACTIONS(260),
    [anon_sym_CARET] = ACTIONS(260),
    [anon_sym_EQ_EQ] = ACTIONS(258),
    [anon_sym_BANG_EQ] = ACTIONS(258),
    [anon_sym_GT] = ACTIONS(260),
    [anon_sym_GT_EQ] = ACTIONS(258),
    [anon_sym_LT_EQ] = ACTIONS(258),
    [anon_sym_LT] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(260),
    [anon_sym_GT_GT] = ACTIONS(260),
    [anon_sym_AMP_AMP] = ACTIONS(258),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT] = ACTIONS(258),
    [anon_sym_QMARK] = ACTIONS(258),
    [anon_sym_BSLASH] = ACTIONS(390),
    [anon_sym_DASH_GT] = ACTIONS(258),
    [anon_sym_LBRACK] = ACTIONS(258),
    [anon_sym_sizeof] = ACTIONS(392),
    [anon_sym_STAR_EQ] = ACTIONS(258),
    [anon_sym_SLASH_EQ] = ACTIONS(258),
    [anon_sym_PERCENT_EQ] = ACTIONS(258),
    [anon_sym_PLUS_EQ] = ACTIONS(258),
    [anon_sym_DASH_EQ] = ACTIONS(258),
    [anon_sym_LT_LT_EQ] = ACTIONS(258),
    [anon_sym_GT_GT_EQ] = ACTIONS(258),
    [anon_sym_AMP_EQ] = ACTIONS(258),
    [anon_sym_CARET_EQ] = ACTIONS(258),
    [anon_sym_PIPE_EQ] = ACTIONS(258),
    [anon_sym_L_SQUOTE] = ACTIONS(394),
    [anon_sym_u_SQUOTE] = ACTIONS(394),
    [anon_sym_U_SQUOTE] = ACTIONS(394),
    [anon_sym_u8_SQUOTE] = ACTIONS(394),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_L_DQUOTE] = ACTIONS(396),
    [anon_sym_u_DQUOTE] = ACTIONS(396),
    [anon_sym_U_DQUOTE] = ACTIONS(396),
    [anon_sym_u8_DQUOTE] = ACTIONS(396),
    [anon_sym_DQUOTE] = ACTIONS(396),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(258),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(260),
    [aux_sym_symbol_token1] = ACTIONS(398),
    [aux_sym_symbol_token2] = ACTIONS(400),
    [sym_comment] = ACTIONS(258),
    [sym__blank_line] = ACTIONS(258),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(260),
    [sym__hll_number_literal_pre_hook] = ACTIONS(402),
  },
  [STATE(23)] = {
    [sym_trace32_hll_variable] = STATE(930),
    [sym__hll_expression] = STATE(892),
    [sym_hll_call_expression] = STATE(928),
    [sym_hll_field_expression] = STATE(928),
    [sym_hll_subscript_expression] = STATE(928),
    [sym_hll_cast_expression] = STATE(892),
    [sym_hll_sizeof_expression] = STATE(892),
    [sym_hll_pointer_expression] = STATE(928),
    [sym_hll_update_expression] = STATE(892),
    [sym_hll_unary_expression] = STATE(892),
    [sym_hll_binary_expression] = STATE(892),
    [sym_hll_assignment_expression] = STATE(892),
    [sym__hll_assignment_left_expression] = STATE(1716),
    [sym_hll_conditional_expression] = STATE(892),
    [sym_hll_char_literal] = STATE(892),
    [sym_hll_string_literal] = STATE(892),
    [sym_hll_parenthesized_expression] = STATE(928),
    [sym_symbol] = STATE(892),
    [sym_identifier] = ACTIONS(404),
    [anon_sym_LPAREN] = ACTIONS(406),
    [anon_sym_EQ] = ACTIONS(260),
    [anon_sym_PLUS] = ACTIONS(408),
    [anon_sym_DASH] = ACTIONS(408),
    [anon_sym_TILDE] = ACTIONS(410),
    [anon_sym_BANG] = ACTIONS(408),
    [anon_sym_PLUS_PLUS] = ACTIONS(412),
    [anon_sym_DASH_DASH] = ACTIONS(412),
    [anon_sym_PIPE_PIPE] = ACTIONS(258),
    [anon_sym_STAR] = ACTIONS(414),
    [anon_sym_SLASH] = ACTIONS(260),
    [anon_sym_PERCENT] = ACTIONS(260),
    [anon_sym_PIPE] = ACTIONS(260),
    [anon_sym_CARET] = ACTIONS(260),
    [anon_sym_EQ_EQ] = ACTIONS(258),
    [anon_sym_BANG_EQ] = ACTIONS(258),
    [anon_sym_GT] = ACTIONS(260),
    [anon_sym_GT_EQ] = ACTIONS(258),
    [anon_sym_LT_EQ] = ACTIONS(258),
    [anon_sym_LT] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(260),
    [anon_sym_GT_GT] = ACTIONS(260),
    [anon_sym_AMP_AMP] = ACTIONS(258),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(258),
    [anon_sym_QMARK] = ACTIONS(258),
    [anon_sym_BSLASH] = ACTIONS(416),
    [anon_sym_DASH_GT] = ACTIONS(258),
    [anon_sym_LBRACK] = ACTIONS(258),
    [anon_sym_sizeof] = ACTIONS(418),
    [anon_sym_STAR_EQ] = ACTIONS(258),
    [anon_sym_SLASH_EQ] = ACTIONS(258),
    [anon_sym_PERCENT_EQ] = ACTIONS(258),
    [anon_sym_PLUS_EQ] = ACTIONS(258),
    [anon_sym_DASH_EQ] = ACTIONS(258),
    [anon_sym_LT_LT_EQ] = ACTIONS(258),
    [anon_sym_GT_GT_EQ] = ACTIONS(258),
    [anon_sym_AMP_EQ] = ACTIONS(258),
    [anon_sym_CARET_EQ] = ACTIONS(258),
    [anon_sym_PIPE_EQ] = ACTIONS(258),
    [anon_sym_L_SQUOTE] = ACTIONS(420),
    [anon_sym_u_SQUOTE] = ACTIONS(420),
    [anon_sym_U_SQUOTE] = ACTIONS(420),
    [anon_sym_u8_SQUOTE] = ACTIONS(420),
    [anon_sym_SQUOTE] = ACTIONS(420),
    [anon_sym_L_DQUOTE] = ACTIONS(422),
    [anon_sym_u_DQUOTE] = ACTIONS(422),
    [anon_sym_U_DQUOTE] = ACTIONS(422),
    [anon_sym_u8_DQUOTE] = ACTIONS(422),
    [anon_sym_DQUOTE] = ACTIONS(422),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(258),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(260),
    [aux_sym_symbol_token1] = ACTIONS(424),
    [aux_sym_symbol_token2] = ACTIONS(426),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(260),
    [sym__hll_number_literal_pre_hook] = ACTIONS(428),
  },
  [STATE(24)] = {
    [sym_trace32_hll_variable] = STATE(930),
    [sym__hll_expression] = STATE(893),
    [sym_hll_call_expression] = STATE(928),
    [sym_hll_field_expression] = STATE(928),
    [sym_hll_subscript_expression] = STATE(928),
    [sym_hll_cast_expression] = STATE(893),
    [sym_hll_sizeof_expression] = STATE(893),
    [sym_hll_pointer_expression] = STATE(928),
    [sym_hll_update_expression] = STATE(893),
    [sym_hll_unary_expression] = STATE(893),
    [sym_hll_binary_expression] = STATE(893),
    [sym_hll_assignment_expression] = STATE(893),
    [sym__hll_assignment_left_expression] = STATE(1716),
    [sym_hll_conditional_expression] = STATE(893),
    [sym_hll_char_literal] = STATE(893),
    [sym_hll_string_literal] = STATE(893),
    [sym_hll_parenthesized_expression] = STATE(928),
    [sym_symbol] = STATE(893),
    [sym_identifier] = ACTIONS(404),
    [anon_sym_LPAREN] = ACTIONS(406),
    [anon_sym_EQ] = ACTIONS(264),
    [anon_sym_PLUS] = ACTIONS(408),
    [anon_sym_DASH] = ACTIONS(408),
    [anon_sym_TILDE] = ACTIONS(410),
    [anon_sym_BANG] = ACTIONS(408),
    [anon_sym_PLUS_PLUS] = ACTIONS(412),
    [anon_sym_DASH_DASH] = ACTIONS(412),
    [anon_sym_PIPE_PIPE] = ACTIONS(262),
    [anon_sym_STAR] = ACTIONS(414),
    [anon_sym_SLASH] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(264),
    [anon_sym_PIPE] = ACTIONS(264),
    [anon_sym_CARET] = ACTIONS(264),
    [anon_sym_EQ_EQ] = ACTIONS(262),
    [anon_sym_BANG_EQ] = ACTIONS(262),
    [anon_sym_GT] = ACTIONS(264),
    [anon_sym_GT_EQ] = ACTIONS(262),
    [anon_sym_LT_EQ] = ACTIONS(262),
    [anon_sym_LT] = ACTIONS(264),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_GT_GT] = ACTIONS(264),
    [anon_sym_AMP_AMP] = ACTIONS(262),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(262),
    [anon_sym_QMARK] = ACTIONS(262),
    [anon_sym_BSLASH] = ACTIONS(416),
    [anon_sym_DASH_GT] = ACTIONS(262),
    [anon_sym_LBRACK] = ACTIONS(262),
    [anon_sym_sizeof] = ACTIONS(418),
    [anon_sym_STAR_EQ] = ACTIONS(262),
    [anon_sym_SLASH_EQ] = ACTIONS(262),
    [anon_sym_PERCENT_EQ] = ACTIONS(262),
    [anon_sym_PLUS_EQ] = ACTIONS(262),
    [anon_sym_DASH_EQ] = ACTIONS(262),
    [anon_sym_LT_LT_EQ] = ACTIONS(262),
    [anon_sym_GT_GT_EQ] = ACTIONS(262),
    [anon_sym_AMP_EQ] = ACTIONS(262),
    [anon_sym_CARET_EQ] = ACTIONS(262),
    [anon_sym_PIPE_EQ] = ACTIONS(262),
    [anon_sym_L_SQUOTE] = ACTIONS(420),
    [anon_sym_u_SQUOTE] = ACTIONS(420),
    [anon_sym_U_SQUOTE] = ACTIONS(420),
    [anon_sym_u8_SQUOTE] = ACTIONS(420),
    [anon_sym_SQUOTE] = ACTIONS(420),
    [anon_sym_L_DQUOTE] = ACTIONS(422),
    [anon_sym_u_DQUOTE] = ACTIONS(422),
    [anon_sym_U_DQUOTE] = ACTIONS(422),
    [anon_sym_u8_DQUOTE] = ACTIONS(422),
    [anon_sym_DQUOTE] = ACTIONS(422),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(262),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(264),
    [aux_sym_symbol_token1] = ACTIONS(424),
    [aux_sym_symbol_token2] = ACTIONS(426),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(264),
    [sym__hll_number_literal_pre_hook] = ACTIONS(428),
  },
  [STATE(25)] = {
    [sym_trace32_hll_variable] = STATE(930),
    [sym__hll_expression] = STATE(883),
    [sym_hll_call_expression] = STATE(928),
    [sym_hll_field_expression] = STATE(928),
    [sym_hll_subscript_expression] = STATE(928),
    [sym_hll_cast_expression] = STATE(883),
    [sym_hll_sizeof_expression] = STATE(883),
    [sym_hll_pointer_expression] = STATE(928),
    [sym_hll_update_expression] = STATE(883),
    [sym_hll_unary_expression] = STATE(883),
    [sym_hll_binary_expression] = STATE(883),
    [sym_hll_assignment_expression] = STATE(883),
    [sym__hll_assignment_left_expression] = STATE(1716),
    [sym_hll_conditional_expression] = STATE(883),
    [sym_hll_char_literal] = STATE(883),
    [sym_hll_string_literal] = STATE(883),
    [sym_hll_parenthesized_expression] = STATE(928),
    [sym_symbol] = STATE(883),
    [sym_identifier] = ACTIONS(404),
    [anon_sym_LPAREN] = ACTIONS(406),
    [anon_sym_EQ] = ACTIONS(234),
    [anon_sym_PLUS] = ACTIONS(408),
    [anon_sym_DASH] = ACTIONS(408),
    [anon_sym_TILDE] = ACTIONS(410),
    [anon_sym_BANG] = ACTIONS(408),
    [anon_sym_PLUS_PLUS] = ACTIONS(412),
    [anon_sym_DASH_DASH] = ACTIONS(412),
    [anon_sym_PIPE_PIPE] = ACTIONS(230),
    [anon_sym_STAR] = ACTIONS(414),
    [anon_sym_SLASH] = ACTIONS(234),
    [anon_sym_PERCENT] = ACTIONS(234),
    [anon_sym_PIPE] = ACTIONS(234),
    [anon_sym_CARET] = ACTIONS(234),
    [anon_sym_EQ_EQ] = ACTIONS(230),
    [anon_sym_BANG_EQ] = ACTIONS(230),
    [anon_sym_GT] = ACTIONS(234),
    [anon_sym_GT_EQ] = ACTIONS(230),
    [anon_sym_LT_EQ] = ACTIONS(230),
    [anon_sym_LT] = ACTIONS(234),
    [anon_sym_LT_LT] = ACTIONS(234),
    [anon_sym_GT_GT] = ACTIONS(234),
    [anon_sym_AMP_AMP] = ACTIONS(230),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(230),
    [anon_sym_QMARK] = ACTIONS(230),
    [anon_sym_BSLASH] = ACTIONS(416),
    [anon_sym_DASH_GT] = ACTIONS(230),
    [anon_sym_LBRACK] = ACTIONS(230),
    [anon_sym_sizeof] = ACTIONS(418),
    [anon_sym_STAR_EQ] = ACTIONS(230),
    [anon_sym_SLASH_EQ] = ACTIONS(230),
    [anon_sym_PERCENT_EQ] = ACTIONS(230),
    [anon_sym_PLUS_EQ] = ACTIONS(230),
    [anon_sym_DASH_EQ] = ACTIONS(230),
    [anon_sym_LT_LT_EQ] = ACTIONS(230),
    [anon_sym_GT_GT_EQ] = ACTIONS(230),
    [anon_sym_AMP_EQ] = ACTIONS(230),
    [anon_sym_CARET_EQ] = ACTIONS(230),
    [anon_sym_PIPE_EQ] = ACTIONS(230),
    [anon_sym_L_SQUOTE] = ACTIONS(420),
    [anon_sym_u_SQUOTE] = ACTIONS(420),
    [anon_sym_U_SQUOTE] = ACTIONS(420),
    [anon_sym_u8_SQUOTE] = ACTIONS(420),
    [anon_sym_SQUOTE] = ACTIONS(420),
    [anon_sym_L_DQUOTE] = ACTIONS(422),
    [anon_sym_u_DQUOTE] = ACTIONS(422),
    [anon_sym_U_DQUOTE] = ACTIONS(422),
    [anon_sym_u8_DQUOTE] = ACTIONS(422),
    [anon_sym_DQUOTE] = ACTIONS(422),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(230),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(234),
    [aux_sym_symbol_token1] = ACTIONS(424),
    [aux_sym_symbol_token2] = ACTIONS(426),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(234),
    [sym__hll_number_literal_pre_hook] = ACTIONS(428),
  },
  [STATE(26)] = {
    [sym_trace32_hll_variable] = STATE(930),
    [sym__hll_expression] = STATE(897),
    [sym_hll_call_expression] = STATE(928),
    [sym_hll_field_expression] = STATE(928),
    [sym_hll_subscript_expression] = STATE(928),
    [sym_hll_cast_expression] = STATE(897),
    [sym_hll_sizeof_expression] = STATE(897),
    [sym_hll_pointer_expression] = STATE(928),
    [sym_hll_update_expression] = STATE(897),
    [sym_hll_unary_expression] = STATE(897),
    [sym_hll_binary_expression] = STATE(897),
    [sym_hll_assignment_expression] = STATE(897),
    [sym__hll_assignment_left_expression] = STATE(1716),
    [sym_hll_conditional_expression] = STATE(897),
    [sym_hll_char_literal] = STATE(897),
    [sym_hll_string_literal] = STATE(897),
    [sym_hll_parenthesized_expression] = STATE(928),
    [sym_symbol] = STATE(897),
    [sym_identifier] = ACTIONS(404),
    [anon_sym_LPAREN] = ACTIONS(406),
    [anon_sym_EQ] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(408),
    [anon_sym_DASH] = ACTIONS(408),
    [anon_sym_TILDE] = ACTIONS(410),
    [anon_sym_BANG] = ACTIONS(408),
    [anon_sym_PLUS_PLUS] = ACTIONS(412),
    [anon_sym_DASH_DASH] = ACTIONS(412),
    [anon_sym_PIPE_PIPE] = ACTIONS(266),
    [anon_sym_STAR] = ACTIONS(414),
    [anon_sym_SLASH] = ACTIONS(268),
    [anon_sym_PERCENT] = ACTIONS(268),
    [anon_sym_PIPE] = ACTIONS(268),
    [anon_sym_CARET] = ACTIONS(268),
    [anon_sym_EQ_EQ] = ACTIONS(266),
    [anon_sym_BANG_EQ] = ACTIONS(266),
    [anon_sym_GT] = ACTIONS(268),
    [anon_sym_GT_EQ] = ACTIONS(266),
    [anon_sym_LT_EQ] = ACTIONS(266),
    [anon_sym_LT] = ACTIONS(268),
    [anon_sym_LT_LT] = ACTIONS(268),
    [anon_sym_GT_GT] = ACTIONS(268),
    [anon_sym_AMP_AMP] = ACTIONS(266),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(266),
    [anon_sym_QMARK] = ACTIONS(266),
    [anon_sym_BSLASH] = ACTIONS(416),
    [anon_sym_DASH_GT] = ACTIONS(266),
    [anon_sym_LBRACK] = ACTIONS(266),
    [anon_sym_sizeof] = ACTIONS(418),
    [anon_sym_STAR_EQ] = ACTIONS(266),
    [anon_sym_SLASH_EQ] = ACTIONS(266),
    [anon_sym_PERCENT_EQ] = ACTIONS(266),
    [anon_sym_PLUS_EQ] = ACTIONS(266),
    [anon_sym_DASH_EQ] = ACTIONS(266),
    [anon_sym_LT_LT_EQ] = ACTIONS(266),
    [anon_sym_GT_GT_EQ] = ACTIONS(266),
    [anon_sym_AMP_EQ] = ACTIONS(266),
    [anon_sym_CARET_EQ] = ACTIONS(266),
    [anon_sym_PIPE_EQ] = ACTIONS(266),
    [anon_sym_L_SQUOTE] = ACTIONS(420),
    [anon_sym_u_SQUOTE] = ACTIONS(420),
    [anon_sym_U_SQUOTE] = ACTIONS(420),
    [anon_sym_u8_SQUOTE] = ACTIONS(420),
    [anon_sym_SQUOTE] = ACTIONS(420),
    [anon_sym_L_DQUOTE] = ACTIONS(422),
    [anon_sym_u_DQUOTE] = ACTIONS(422),
    [anon_sym_U_DQUOTE] = ACTIONS(422),
    [anon_sym_u8_DQUOTE] = ACTIONS(422),
    [anon_sym_DQUOTE] = ACTIONS(422),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(266),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(268),
    [aux_sym_symbol_token1] = ACTIONS(424),
    [aux_sym_symbol_token2] = ACTIONS(426),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(268),
    [sym__hll_number_literal_pre_hook] = ACTIONS(428),
  },
  [STATE(27)] = {
    [sym_trace32_hll_variable] = STATE(991),
    [sym__hll_expression] = STATE(958),
    [sym_hll_call_expression] = STATE(977),
    [sym_hll_field_expression] = STATE(977),
    [sym_hll_subscript_expression] = STATE(977),
    [sym_hll_cast_expression] = STATE(958),
    [sym_hll_sizeof_expression] = STATE(958),
    [sym_hll_pointer_expression] = STATE(977),
    [sym_hll_update_expression] = STATE(958),
    [sym_hll_unary_expression] = STATE(958),
    [sym_hll_binary_expression] = STATE(958),
    [sym_hll_assignment_expression] = STATE(958),
    [sym__hll_assignment_left_expression] = STATE(1691),
    [sym_hll_conditional_expression] = STATE(958),
    [sym_hll_char_literal] = STATE(958),
    [sym_hll_string_literal] = STATE(958),
    [sym_hll_parenthesized_expression] = STATE(977),
    [sym_symbol] = STATE(958),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_LPAREN] = ACTIONS(432),
    [anon_sym_EQ] = ACTIONS(260),
    [anon_sym_PLUS] = ACTIONS(434),
    [anon_sym_DASH] = ACTIONS(434),
    [anon_sym_TILDE] = ACTIONS(436),
    [anon_sym_BANG] = ACTIONS(434),
    [anon_sym_PLUS_PLUS] = ACTIONS(438),
    [anon_sym_DASH_DASH] = ACTIONS(438),
    [anon_sym_PIPE_PIPE] = ACTIONS(258),
    [anon_sym_STAR] = ACTIONS(440),
    [anon_sym_SLASH] = ACTIONS(260),
    [anon_sym_PERCENT] = ACTIONS(260),
    [anon_sym_PIPE] = ACTIONS(260),
    [anon_sym_CARET] = ACTIONS(260),
    [anon_sym_EQ_EQ] = ACTIONS(258),
    [anon_sym_BANG_EQ] = ACTIONS(258),
    [anon_sym_GT] = ACTIONS(260),
    [anon_sym_GT_EQ] = ACTIONS(258),
    [anon_sym_LT_EQ] = ACTIONS(258),
    [anon_sym_LT] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(260),
    [anon_sym_GT_GT] = ACTIONS(260),
    [anon_sym_AMP_AMP] = ACTIONS(258),
    [anon_sym_AMP] = ACTIONS(440),
    [anon_sym_DOT] = ACTIONS(258),
    [anon_sym_QMARK] = ACTIONS(258),
    [anon_sym_BSLASH] = ACTIONS(442),
    [anon_sym_DASH_GT] = ACTIONS(258),
    [anon_sym_LBRACK] = ACTIONS(258),
    [anon_sym_sizeof] = ACTIONS(444),
    [anon_sym_STAR_EQ] = ACTIONS(258),
    [anon_sym_SLASH_EQ] = ACTIONS(258),
    [anon_sym_PERCENT_EQ] = ACTIONS(258),
    [anon_sym_PLUS_EQ] = ACTIONS(258),
    [anon_sym_DASH_EQ] = ACTIONS(258),
    [anon_sym_LT_LT_EQ] = ACTIONS(258),
    [anon_sym_GT_GT_EQ] = ACTIONS(258),
    [anon_sym_AMP_EQ] = ACTIONS(258),
    [anon_sym_CARET_EQ] = ACTIONS(258),
    [anon_sym_PIPE_EQ] = ACTIONS(258),
    [anon_sym_L_SQUOTE] = ACTIONS(446),
    [anon_sym_u_SQUOTE] = ACTIONS(446),
    [anon_sym_U_SQUOTE] = ACTIONS(446),
    [anon_sym_u8_SQUOTE] = ACTIONS(446),
    [anon_sym_SQUOTE] = ACTIONS(446),
    [anon_sym_L_DQUOTE] = ACTIONS(448),
    [anon_sym_u_DQUOTE] = ACTIONS(448),
    [anon_sym_U_DQUOTE] = ACTIONS(448),
    [anon_sym_u8_DQUOTE] = ACTIONS(448),
    [anon_sym_DQUOTE] = ACTIONS(448),
    [aux_sym_symbol_token1] = ACTIONS(450),
    [aux_sym_symbol_token2] = ACTIONS(452),
    [sym_comment] = ACTIONS(258),
    [sym__blank_line] = ACTIONS(258),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(260),
    [sym__hll_number_literal_pre_hook] = ACTIONS(454),
  },
  [STATE(28)] = {
    [sym_trace32_hll_variable] = STATE(991),
    [sym__hll_expression] = STATE(960),
    [sym_hll_call_expression] = STATE(977),
    [sym_hll_field_expression] = STATE(977),
    [sym_hll_subscript_expression] = STATE(977),
    [sym_hll_cast_expression] = STATE(960),
    [sym_hll_sizeof_expression] = STATE(960),
    [sym_hll_pointer_expression] = STATE(977),
    [sym_hll_update_expression] = STATE(960),
    [sym_hll_unary_expression] = STATE(960),
    [sym_hll_binary_expression] = STATE(960),
    [sym_hll_assignment_expression] = STATE(960),
    [sym__hll_assignment_left_expression] = STATE(1691),
    [sym_hll_conditional_expression] = STATE(960),
    [sym_hll_char_literal] = STATE(960),
    [sym_hll_string_literal] = STATE(960),
    [sym_hll_parenthesized_expression] = STATE(977),
    [sym_symbol] = STATE(960),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_LPAREN] = ACTIONS(432),
    [anon_sym_EQ] = ACTIONS(264),
    [anon_sym_PLUS] = ACTIONS(434),
    [anon_sym_DASH] = ACTIONS(434),
    [anon_sym_TILDE] = ACTIONS(436),
    [anon_sym_BANG] = ACTIONS(434),
    [anon_sym_PLUS_PLUS] = ACTIONS(438),
    [anon_sym_DASH_DASH] = ACTIONS(438),
    [anon_sym_PIPE_PIPE] = ACTIONS(262),
    [anon_sym_STAR] = ACTIONS(440),
    [anon_sym_SLASH] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(264),
    [anon_sym_PIPE] = ACTIONS(264),
    [anon_sym_CARET] = ACTIONS(264),
    [anon_sym_EQ_EQ] = ACTIONS(262),
    [anon_sym_BANG_EQ] = ACTIONS(262),
    [anon_sym_GT] = ACTIONS(264),
    [anon_sym_GT_EQ] = ACTIONS(262),
    [anon_sym_LT_EQ] = ACTIONS(262),
    [anon_sym_LT] = ACTIONS(264),
    [anon_sym_LT_LT] = ACTIONS(264),
    [anon_sym_GT_GT] = ACTIONS(264),
    [anon_sym_AMP_AMP] = ACTIONS(262),
    [anon_sym_AMP] = ACTIONS(440),
    [anon_sym_DOT] = ACTIONS(262),
    [anon_sym_QMARK] = ACTIONS(262),
    [anon_sym_BSLASH] = ACTIONS(442),
    [anon_sym_DASH_GT] = ACTIONS(262),
    [anon_sym_LBRACK] = ACTIONS(262),
    [anon_sym_sizeof] = ACTIONS(444),
    [anon_sym_STAR_EQ] = ACTIONS(262),
    [anon_sym_SLASH_EQ] = ACTIONS(262),
    [anon_sym_PERCENT_EQ] = ACTIONS(262),
    [anon_sym_PLUS_EQ] = ACTIONS(262),
    [anon_sym_DASH_EQ] = ACTIONS(262),
    [anon_sym_LT_LT_EQ] = ACTIONS(262),
    [anon_sym_GT_GT_EQ] = ACTIONS(262),
    [anon_sym_AMP_EQ] = ACTIONS(262),
    [anon_sym_CARET_EQ] = ACTIONS(262),
    [anon_sym_PIPE_EQ] = ACTIONS(262),
    [anon_sym_L_SQUOTE] = ACTIONS(446),
    [anon_sym_u_SQUOTE] = ACTIONS(446),
    [anon_sym_U_SQUOTE] = ACTIONS(446),
    [anon_sym_u8_SQUOTE] = ACTIONS(446),
    [anon_sym_SQUOTE] = ACTIONS(446),
    [anon_sym_L_DQUOTE] = ACTIONS(448),
    [anon_sym_u_DQUOTE] = ACTIONS(448),
    [anon_sym_U_DQUOTE] = ACTIONS(448),
    [anon_sym_u8_DQUOTE] = ACTIONS(448),
    [anon_sym_DQUOTE] = ACTIONS(448),
    [aux_sym_symbol_token1] = ACTIONS(450),
    [aux_sym_symbol_token2] = ACTIONS(452),
    [sym_comment] = ACTIONS(262),
    [sym__blank_line] = ACTIONS(262),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(264),
    [sym__hll_number_literal_pre_hook] = ACTIONS(454),
  },
  [STATE(29)] = {
    [sym_trace32_hll_variable] = STATE(991),
    [sym__hll_expression] = STATE(953),
    [sym_hll_call_expression] = STATE(977),
    [sym_hll_field_expression] = STATE(977),
    [sym_hll_subscript_expression] = STATE(977),
    [sym_hll_cast_expression] = STATE(953),
    [sym_hll_sizeof_expression] = STATE(953),
    [sym_hll_pointer_expression] = STATE(977),
    [sym_hll_update_expression] = STATE(953),
    [sym_hll_unary_expression] = STATE(953),
    [sym_hll_binary_expression] = STATE(953),
    [sym_hll_assignment_expression] = STATE(953),
    [sym__hll_assignment_left_expression] = STATE(1691),
    [sym_hll_conditional_expression] = STATE(953),
    [sym_hll_char_literal] = STATE(953),
    [sym_hll_string_literal] = STATE(953),
    [sym_hll_parenthesized_expression] = STATE(977),
    [sym_symbol] = STATE(953),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_LPAREN] = ACTIONS(432),
    [anon_sym_EQ] = ACTIONS(234),
    [anon_sym_PLUS] = ACTIONS(434),
    [anon_sym_DASH] = ACTIONS(434),
    [anon_sym_TILDE] = ACTIONS(436),
    [anon_sym_BANG] = ACTIONS(434),
    [anon_sym_PLUS_PLUS] = ACTIONS(438),
    [anon_sym_DASH_DASH] = ACTIONS(438),
    [anon_sym_PIPE_PIPE] = ACTIONS(230),
    [anon_sym_STAR] = ACTIONS(440),
    [anon_sym_SLASH] = ACTIONS(234),
    [anon_sym_PERCENT] = ACTIONS(234),
    [anon_sym_PIPE] = ACTIONS(234),
    [anon_sym_CARET] = ACTIONS(234),
    [anon_sym_EQ_EQ] = ACTIONS(230),
    [anon_sym_BANG_EQ] = ACTIONS(230),
    [anon_sym_GT] = ACTIONS(234),
    [anon_sym_GT_EQ] = ACTIONS(230),
    [anon_sym_LT_EQ] = ACTIONS(230),
    [anon_sym_LT] = ACTIONS(234),
    [anon_sym_LT_LT] = ACTIONS(234),
    [anon_sym_GT_GT] = ACTIONS(234),
    [anon_sym_AMP_AMP] = ACTIONS(230),
    [anon_sym_AMP] = ACTIONS(440),
    [anon_sym_DOT] = ACTIONS(230),
    [anon_sym_QMARK] = ACTIONS(230),
    [anon_sym_BSLASH] = ACTIONS(442),
    [anon_sym_DASH_GT] = ACTIONS(230),
    [anon_sym_LBRACK] = ACTIONS(230),
    [anon_sym_sizeof] = ACTIONS(444),
    [anon_sym_STAR_EQ] = ACTIONS(230),
    [anon_sym_SLASH_EQ] = ACTIONS(230),
    [anon_sym_PERCENT_EQ] = ACTIONS(230),
    [anon_sym_PLUS_EQ] = ACTIONS(230),
    [anon_sym_DASH_EQ] = ACTIONS(230),
    [anon_sym_LT_LT_EQ] = ACTIONS(230),
    [anon_sym_GT_GT_EQ] = ACTIONS(230),
    [anon_sym_AMP_EQ] = ACTIONS(230),
    [anon_sym_CARET_EQ] = ACTIONS(230),
    [anon_sym_PIPE_EQ] = ACTIONS(230),
    [anon_sym_L_SQUOTE] = ACTIONS(446),
    [anon_sym_u_SQUOTE] = ACTIONS(446),
    [anon_sym_U_SQUOTE] = ACTIONS(446),
    [anon_sym_u8_SQUOTE] = ACTIONS(446),
    [anon_sym_SQUOTE] = ACTIONS(446),
    [anon_sym_L_DQUOTE] = ACTIONS(448),
    [anon_sym_u_DQUOTE] = ACTIONS(448),
    [anon_sym_U_DQUOTE] = ACTIONS(448),
    [anon_sym_u8_DQUOTE] = ACTIONS(448),
    [anon_sym_DQUOTE] = ACTIONS(448),
    [aux_sym_symbol_token1] = ACTIONS(450),
    [aux_sym_symbol_token2] = ACTIONS(452),
    [sym_comment] = ACTIONS(230),
    [sym__blank_line] = ACTIONS(230),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(234),
    [sym__hll_number_literal_pre_hook] = ACTIONS(454),
  },
  [STATE(30)] = {
    [sym_trace32_hll_variable] = STATE(991),
    [sym__hll_expression] = STATE(961),
    [sym_hll_call_expression] = STATE(977),
    [sym_hll_field_expression] = STATE(977),
    [sym_hll_subscript_expression] = STATE(977),
    [sym_hll_cast_expression] = STATE(961),
    [sym_hll_sizeof_expression] = STATE(961),
    [sym_hll_pointer_expression] = STATE(977),
    [sym_hll_update_expression] = STATE(961),
    [sym_hll_unary_expression] = STATE(961),
    [sym_hll_binary_expression] = STATE(961),
    [sym_hll_assignment_expression] = STATE(961),
    [sym__hll_assignment_left_expression] = STATE(1691),
    [sym_hll_conditional_expression] = STATE(961),
    [sym_hll_char_literal] = STATE(961),
    [sym_hll_string_literal] = STATE(961),
    [sym_hll_parenthesized_expression] = STATE(977),
    [sym_symbol] = STATE(961),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_LPAREN] = ACTIONS(432),
    [anon_sym_EQ] = ACTIONS(268),
    [anon_sym_PLUS] = ACTIONS(434),
    [anon_sym_DASH] = ACTIONS(434),
    [anon_sym_TILDE] = ACTIONS(436),
    [anon_sym_BANG] = ACTIONS(434),
    [anon_sym_PLUS_PLUS] = ACTIONS(438),
    [anon_sym_DASH_DASH] = ACTIONS(438),
    [anon_sym_PIPE_PIPE] = ACTIONS(266),
    [anon_sym_STAR] = ACTIONS(440),
    [anon_sym_SLASH] = ACTIONS(268),
    [anon_sym_PERCENT] = ACTIONS(268),
    [anon_sym_PIPE] = ACTIONS(268),
    [anon_sym_CARET] = ACTIONS(268),
    [anon_sym_EQ_EQ] = ACTIONS(266),
    [anon_sym_BANG_EQ] = ACTIONS(266),
    [anon_sym_GT] = ACTIONS(268),
    [anon_sym_GT_EQ] = ACTIONS(266),
    [anon_sym_LT_EQ] = ACTIONS(266),
    [anon_sym_LT] = ACTIONS(268),
    [anon_sym_LT_LT] = ACTIONS(268),
    [anon_sym_GT_GT] = ACTIONS(268),
    [anon_sym_AMP_AMP] = ACTIONS(266),
    [anon_sym_AMP] = ACTIONS(440),
    [anon_sym_DOT] = ACTIONS(266),
    [anon_sym_QMARK] = ACTIONS(266),
    [anon_sym_BSLASH] = ACTIONS(442),
    [anon_sym_DASH_GT] = ACTIONS(266),
    [anon_sym_LBRACK] = ACTIONS(266),
    [anon_sym_sizeof] = ACTIONS(444),
    [anon_sym_STAR_EQ] = ACTIONS(266),
    [anon_sym_SLASH_EQ] = ACTIONS(266),
    [anon_sym_PERCENT_EQ] = ACTIONS(266),
    [anon_sym_PLUS_EQ] = ACTIONS(266),
    [anon_sym_DASH_EQ] = ACTIONS(266),
    [anon_sym_LT_LT_EQ] = ACTIONS(266),
    [anon_sym_GT_GT_EQ] = ACTIONS(266),
    [anon_sym_AMP_EQ] = ACTIONS(266),
    [anon_sym_CARET_EQ] = ACTIONS(266),
    [anon_sym_PIPE_EQ] = ACTIONS(266),
    [anon_sym_L_SQUOTE] = ACTIONS(446),
    [anon_sym_u_SQUOTE] = ACTIONS(446),
    [anon_sym_U_SQUOTE] = ACTIONS(446),
    [anon_sym_u8_SQUOTE] = ACTIONS(446),
    [anon_sym_SQUOTE] = ACTIONS(446),
    [anon_sym_L_DQUOTE] = ACTIONS(448),
    [anon_sym_u_DQUOTE] = ACTIONS(448),
    [anon_sym_U_DQUOTE] = ACTIONS(448),
    [anon_sym_u8_DQUOTE] = ACTIONS(448),
    [anon_sym_DQUOTE] = ACTIONS(448),
    [aux_sym_symbol_token1] = ACTIONS(450),
    [aux_sym_symbol_token2] = ACTIONS(452),
    [sym_comment] = ACTIONS(266),
    [sym__blank_line] = ACTIONS(266),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(268),
    [sym__hll_number_literal_pre_hook] = ACTIONS(454),
  },
  [STATE(31)] = {
    [sym__expression] = STATE(1307),
    [sym__parenthesized_expression] = STATE(1307),
    [sym__path_expression] = STATE(1307),
    [sym_unary_expression] = STATE(1307),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1307),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1307),
    [sym_macro] = STATE(1097),
    [sym__command_arguments] = STATE(2686),
    [sym__command_argument] = STATE(2128),
    [sym_option_expression] = STATE(2167),
    [sym_format_expression] = STATE(2128),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1307),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1307),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1307),
    [sym_string] = STATE(1307),
    [sym_character] = STATE(1307),
    [sym_symbol] = STATE(1307),
    [sym__special_command_operator] = STATE(1307),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1000),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [aux_sym__command_arguments_repeat1] = STATE(41),
    [aux_sym__command_arguments_repeat2] = STATE(2240),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_SLASH] = ACTIONS(456),
    [anon_sym_PERCENT] = ACTIONS(458),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(460),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(462),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(464),
    [sym_time] = ACTIONS(462),
    [sym_frequency] = ACTIONS(462),
    [sym_percentage] = ACTIONS(464),
    [sym_file_handle] = ACTIONS(464),
    [sym_comment] = ACTIONS(466),
    [sym__blank_line] = ACTIONS(466),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(468),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(32)] = {
    [sym__expression] = STATE(1307),
    [sym__parenthesized_expression] = STATE(1307),
    [sym__path_expression] = STATE(1307),
    [sym_unary_expression] = STATE(1307),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1307),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1307),
    [sym_macro] = STATE(1097),
    [sym__command_arguments] = STATE(2736),
    [sym__command_argument] = STATE(2128),
    [sym_option_expression] = STATE(2167),
    [sym_format_expression] = STATE(2128),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1307),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1307),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1307),
    [sym_string] = STATE(1307),
    [sym_character] = STATE(1307),
    [sym_symbol] = STATE(1307),
    [sym__special_command_operator] = STATE(1307),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [aux_sym__command_arguments_repeat1] = STATE(41),
    [aux_sym__command_arguments_repeat2] = STATE(2240),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_SLASH] = ACTIONS(456),
    [anon_sym_PERCENT] = ACTIONS(458),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(460),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(462),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(464),
    [sym_time] = ACTIONS(462),
    [sym_frequency] = ACTIONS(462),
    [sym_percentage] = ACTIONS(464),
    [sym_file_handle] = ACTIONS(464),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(33)] = {
    [sym__expression] = STATE(1307),
    [sym__parenthesized_expression] = STATE(1307),
    [sym__path_expression] = STATE(1307),
    [sym_unary_expression] = STATE(1307),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1307),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1307),
    [sym_macro] = STATE(1097),
    [sym__command_arguments] = STATE(2604),
    [sym__command_argument] = STATE(2128),
    [sym_option_expression] = STATE(2167),
    [sym_format_expression] = STATE(2128),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1307),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1307),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1307),
    [sym_string] = STATE(1307),
    [sym_character] = STATE(1307),
    [sym_symbol] = STATE(1307),
    [sym__special_command_operator] = STATE(1307),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [aux_sym__command_arguments_repeat1] = STATE(41),
    [aux_sym__command_arguments_repeat2] = STATE(2240),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_SLASH] = ACTIONS(456),
    [anon_sym_PERCENT] = ACTIONS(458),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(460),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(462),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(464),
    [sym_time] = ACTIONS(462),
    [sym_frequency] = ACTIONS(462),
    [sym_percentage] = ACTIONS(464),
    [sym_file_handle] = ACTIONS(464),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(34)] = {
    [sym__expression] = STATE(1307),
    [sym__parenthesized_expression] = STATE(1307),
    [sym__path_expression] = STATE(1307),
    [sym_unary_expression] = STATE(1307),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1307),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1307),
    [sym_macro] = STATE(1097),
    [sym__command_arguments] = STATE(2690),
    [sym__command_argument] = STATE(2128),
    [sym_option_expression] = STATE(2167),
    [sym_format_expression] = STATE(2128),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1307),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1307),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1307),
    [sym_string] = STATE(1307),
    [sym_character] = STATE(1307),
    [sym_symbol] = STATE(1307),
    [sym__special_command_operator] = STATE(1307),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [aux_sym__command_arguments_repeat1] = STATE(41),
    [aux_sym__command_arguments_repeat2] = STATE(2240),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_SLASH] = ACTIONS(456),
    [anon_sym_PERCENT] = ACTIONS(458),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(460),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(462),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(464),
    [sym_time] = ACTIONS(462),
    [sym_frequency] = ACTIONS(462),
    [sym_percentage] = ACTIONS(464),
    [sym_file_handle] = ACTIONS(464),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(35)] = {
    [sym__expression] = STATE(1307),
    [sym__parenthesized_expression] = STATE(1307),
    [sym__path_expression] = STATE(1307),
    [sym_unary_expression] = STATE(1307),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1307),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1307),
    [sym_macro] = STATE(1097),
    [sym__command_arguments] = STATE(2622),
    [sym__command_argument] = STATE(2128),
    [sym_option_expression] = STATE(2167),
    [sym_format_expression] = STATE(2128),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1307),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1307),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1307),
    [sym_string] = STATE(1307),
    [sym_character] = STATE(1307),
    [sym_symbol] = STATE(1307),
    [sym__special_command_operator] = STATE(1307),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [aux_sym__command_arguments_repeat1] = STATE(41),
    [aux_sym__command_arguments_repeat2] = STATE(2240),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_SLASH] = ACTIONS(456),
    [anon_sym_PERCENT] = ACTIONS(458),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(460),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(462),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(464),
    [sym_time] = ACTIONS(462),
    [sym_frequency] = ACTIONS(462),
    [sym_percentage] = ACTIONS(464),
    [sym_file_handle] = ACTIONS(464),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(36)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1143),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1143),
    [sym_hll_sizeof_expression] = STATE(1143),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2405),
    [sym_hll_update_expression] = STATE(1143),
    [sym_hll_unary_expression] = STATE(1143),
    [sym_hll_binary_expression] = STATE(1143),
    [sym_hll_assignment_expression] = STATE(1143),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1143),
    [sym_hll_char_literal] = STATE(1143),
    [sym_hll_string_literal] = STATE(1143),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2290),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1143),
    [aux_sym_block_repeat1] = STATE(73),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(508),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(37)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1022),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1022),
    [sym_hll_sizeof_expression] = STATE(1022),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_update_expression] = STATE(1022),
    [sym_hll_unary_expression] = STATE(1022),
    [sym_hll_binary_expression] = STATE(1022),
    [sym_hll_assignment_expression] = STATE(1022),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1022),
    [sym_hll_char_literal] = STATE(1022),
    [sym_hll_string_literal] = STATE(1022),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_symbol] = STATE(1022),
    [sym_identifier] = ACTIONS(512),
    [anon_sym_RPAREN] = ACTIONS(262),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(476),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_PIPE_PIPE] = ACTIONS(262),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_SLASH] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(262),
    [anon_sym_PIPE] = ACTIONS(264),
    [anon_sym_CARET] = ACTIONS(262),
    [anon_sym_EQ_EQ] = ACTIONS(262),
    [anon_sym_BANG_EQ] = ACTIONS(262),
    [anon_sym_GT] = ACTIONS(264),
    [anon_sym_GT_EQ] = ACTIONS(262),
    [anon_sym_LT_EQ] = ACTIONS(262),
    [anon_sym_LT] = ACTIONS(264),
    [anon_sym_LT_LT] = ACTIONS(262),
    [anon_sym_GT_GT] = ACTIONS(262),
    [anon_sym_DOT_DOT] = ACTIONS(262),
    [anon_sym_AMP_AMP] = ACTIONS(262),
    [anon_sym_AMP] = ACTIONS(242),
    [anon_sym_DOT] = ACTIONS(264),
    [anon_sym_COMMA] = ACTIONS(262),
    [anon_sym_QMARK] = ACTIONS(262),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_DASH_GT] = ACTIONS(262),
    [anon_sym_LBRACK] = ACTIONS(262),
    [anon_sym_RBRACK] = ACTIONS(262),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(264),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(38)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1101),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1101),
    [sym_hll_sizeof_expression] = STATE(1101),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2341),
    [sym_hll_update_expression] = STATE(1101),
    [sym_hll_unary_expression] = STATE(1101),
    [sym_hll_binary_expression] = STATE(1101),
    [sym_hll_assignment_expression] = STATE(1101),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1101),
    [sym_hll_char_literal] = STATE(1101),
    [sym_hll_string_literal] = STATE(1101),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2350),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1101),
    [aux_sym_block_repeat1] = STATE(42),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(514),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(39)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1149),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1149),
    [sym_hll_sizeof_expression] = STATE(1149),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2315),
    [sym_hll_update_expression] = STATE(1149),
    [sym_hll_unary_expression] = STATE(1149),
    [sym_hll_binary_expression] = STATE(1149),
    [sym_hll_assignment_expression] = STATE(1149),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1149),
    [sym_hll_char_literal] = STATE(1149),
    [sym_hll_string_literal] = STATE(1149),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2274),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1149),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(40)] = {
    [sym__expression] = STATE(1397),
    [sym__parenthesized_expression] = STATE(1397),
    [sym__path_expression] = STATE(1397),
    [sym_unary_expression] = STATE(1397),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1397),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1397),
    [sym_macro] = STATE(1054),
    [sym__command_argument] = STATE(2455),
    [sym_format_expression] = STATE(2455),
    [sym_call_expression] = STATE(1115),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1397),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1397),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1118),
    [sym_bitmask] = STATE(1397),
    [sym_string] = STATE(1397),
    [sym_character] = STATE(1397),
    [sym_symbol] = STATE(1397),
    [sym__special_command_operator] = STATE(1397),
    [sym__file_dialog] = STATE(1482),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [aux_sym__command_arguments_repeat1] = STATE(40),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_LPAREN] = ACTIONS(523),
    [anon_sym_LBRACE] = ACTIONS(526),
    [anon_sym_PLUS] = ACTIONS(529),
    [anon_sym_DASH] = ACTIONS(529),
    [anon_sym_TILDE] = ACTIONS(532),
    [anon_sym_BANG] = ACTIONS(532),
    [anon_sym_PLUS_PLUS] = ACTIONS(535),
    [anon_sym_DASH_DASH] = ACTIONS(535),
    [anon_sym_STAR] = ACTIONS(538),
    [anon_sym_PERCENT] = ACTIONS(541),
    [anon_sym_AMP] = ACTIONS(544),
    [anon_sym_COMMA] = ACTIONS(547),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(550),
    [anon_sym_BSLASH] = ACTIONS(553),
    [anon_sym_SQUOTE] = ACTIONS(556),
    [anon_sym_DQUOTE] = ACTIONS(559),
    [sym_access_class] = ACTIONS(562),
    [aux_sym_integer_token1] = ACTIONS(565),
    [aux_sym_integer_token2] = ACTIONS(565),
    [sym_float] = ACTIONS(568),
    [aux_sym_bitmask_token1] = ACTIONS(571),
    [aux_sym_bitmask_token2] = ACTIONS(571),
    [aux_sym_symbol_token1] = ACTIONS(574),
    [aux_sym_symbol_token2] = ACTIONS(577),
    [sym__composed_name] = ACTIONS(580),
    [sym_time] = ACTIONS(568),
    [sym_frequency] = ACTIONS(568),
    [sym_percentage] = ACTIONS(580),
    [sym_file_handle] = ACTIONS(580),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__decimal_number_pre_hook] = ACTIONS(583),
    [sym_path] = ACTIONS(586),
  },
  [STATE(41)] = {
    [sym__expression] = STATE(1307),
    [sym__parenthesized_expression] = STATE(1307),
    [sym__path_expression] = STATE(1307),
    [sym_unary_expression] = STATE(1307),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1307),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1307),
    [sym_macro] = STATE(1097),
    [sym__command_argument] = STATE(2061),
    [sym_format_expression] = STATE(2061),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1307),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1307),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1307),
    [sym_string] = STATE(1307),
    [sym_character] = STATE(1307),
    [sym_symbol] = STATE(1307),
    [sym__special_command_operator] = STATE(1307),
    [sym__file_dialog] = STATE(1496),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [aux_sym__command_arguments_repeat1] = STATE(40),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_PERCENT] = ACTIONS(458),
    [anon_sym_AMP] = ACTIONS(187),
    [anon_sym_COMMA] = ACTIONS(589),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(462),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(464),
    [sym_time] = ACTIONS(462),
    [sym_frequency] = ACTIONS(462),
    [sym_percentage] = ACTIONS(464),
    [sym_file_handle] = ACTIONS(464),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(42)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1149),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1149),
    [sym_hll_sizeof_expression] = STATE(1149),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2315),
    [sym_hll_update_expression] = STATE(1149),
    [sym_hll_unary_expression] = STATE(1149),
    [sym_hll_binary_expression] = STATE(1149),
    [sym_hll_assignment_expression] = STATE(1149),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1149),
    [sym_hll_char_literal] = STATE(1149),
    [sym_hll_string_literal] = STATE(1149),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2318),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1149),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(43)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1150),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1150),
    [sym_hll_sizeof_expression] = STATE(1150),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2363),
    [sym_hll_update_expression] = STATE(1150),
    [sym_hll_unary_expression] = STATE(1150),
    [sym_hll_binary_expression] = STATE(1150),
    [sym_hll_assignment_expression] = STATE(1150),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1150),
    [sym_hll_char_literal] = STATE(1150),
    [sym_hll_string_literal] = STATE(1150),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2329),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1150),
    [aux_sym_block_repeat1] = STATE(44),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(591),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(44)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1087),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1087),
    [sym_hll_sizeof_expression] = STATE(1087),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2303),
    [sym_hll_update_expression] = STATE(1087),
    [sym_hll_unary_expression] = STATE(1087),
    [sym_hll_binary_expression] = STATE(1087),
    [sym_hll_assignment_expression] = STATE(1087),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1087),
    [sym_hll_char_literal] = STATE(1087),
    [sym_hll_string_literal] = STATE(1087),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2408),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1087),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(45)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1101),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1101),
    [sym_hll_sizeof_expression] = STATE(1101),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2341),
    [sym_hll_update_expression] = STATE(1101),
    [sym_hll_unary_expression] = STATE(1101),
    [sym_hll_binary_expression] = STATE(1101),
    [sym_hll_assignment_expression] = STATE(1101),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1101),
    [sym_hll_char_literal] = STATE(1101),
    [sym_hll_string_literal] = STATE(1101),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2355),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1101),
    [aux_sym_block_repeat1] = STATE(39),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(593),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(46)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1019),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1019),
    [sym_hll_sizeof_expression] = STATE(1019),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_update_expression] = STATE(1019),
    [sym_hll_unary_expression] = STATE(1019),
    [sym_hll_binary_expression] = STATE(1019),
    [sym_hll_assignment_expression] = STATE(1019),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1019),
    [sym_hll_char_literal] = STATE(1019),
    [sym_hll_string_literal] = STATE(1019),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_symbol] = STATE(1019),
    [sym_identifier] = ACTIONS(512),
    [anon_sym_RPAREN] = ACTIONS(230),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(476),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_PIPE_PIPE] = ACTIONS(230),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_SLASH] = ACTIONS(234),
    [anon_sym_PERCENT] = ACTIONS(230),
    [anon_sym_PIPE] = ACTIONS(234),
    [anon_sym_CARET] = ACTIONS(230),
    [anon_sym_EQ_EQ] = ACTIONS(230),
    [anon_sym_BANG_EQ] = ACTIONS(230),
    [anon_sym_GT] = ACTIONS(234),
    [anon_sym_GT_EQ] = ACTIONS(230),
    [anon_sym_LT_EQ] = ACTIONS(230),
    [anon_sym_LT] = ACTIONS(234),
    [anon_sym_LT_LT] = ACTIONS(230),
    [anon_sym_GT_GT] = ACTIONS(230),
    [anon_sym_DOT_DOT] = ACTIONS(230),
    [anon_sym_AMP_AMP] = ACTIONS(230),
    [anon_sym_AMP] = ACTIONS(242),
    [anon_sym_DOT] = ACTIONS(234),
    [anon_sym_COMMA] = ACTIONS(230),
    [anon_sym_QMARK] = ACTIONS(230),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_DASH_GT] = ACTIONS(230),
    [anon_sym_LBRACK] = ACTIONS(230),
    [anon_sym_RBRACK] = ACTIONS(230),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(234),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(47)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1143),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1143),
    [sym_hll_sizeof_expression] = STATE(1143),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2405),
    [sym_hll_update_expression] = STATE(1143),
    [sym_hll_unary_expression] = STATE(1143),
    [sym_hll_binary_expression] = STATE(1143),
    [sym_hll_assignment_expression] = STATE(1143),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1143),
    [sym_hll_char_literal] = STATE(1143),
    [sym_hll_string_literal] = STATE(1143),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2428),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1143),
    [aux_sym_block_repeat1] = STATE(48),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(595),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(48)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1152),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1152),
    [sym_hll_sizeof_expression] = STATE(1152),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2426),
    [sym_hll_update_expression] = STATE(1152),
    [sym_hll_unary_expression] = STATE(1152),
    [sym_hll_binary_expression] = STATE(1152),
    [sym_hll_assignment_expression] = STATE(1152),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1152),
    [sym_hll_char_literal] = STATE(1152),
    [sym_hll_string_literal] = STATE(1152),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2437),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1152),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(49)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1080),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1080),
    [sym_hll_sizeof_expression] = STATE(1080),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2228),
    [sym_hll_update_expression] = STATE(1080),
    [sym_hll_unary_expression] = STATE(1080),
    [sym_hll_binary_expression] = STATE(1080),
    [sym_hll_assignment_expression] = STATE(1080),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1080),
    [sym_hll_char_literal] = STATE(1080),
    [sym_hll_string_literal] = STATE(1080),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2309),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1080),
    [aux_sym_block_repeat1] = STATE(50),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(597),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(50)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1085),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1085),
    [sym_hll_sizeof_expression] = STATE(1085),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2235),
    [sym_hll_update_expression] = STATE(1085),
    [sym_hll_unary_expression] = STATE(1085),
    [sym_hll_binary_expression] = STATE(1085),
    [sym_hll_assignment_expression] = STATE(1085),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1085),
    [sym_hll_char_literal] = STATE(1085),
    [sym_hll_string_literal] = STATE(1085),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2312),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1085),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(51)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1150),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1150),
    [sym_hll_sizeof_expression] = STATE(1150),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2363),
    [sym_hll_update_expression] = STATE(1150),
    [sym_hll_unary_expression] = STATE(1150),
    [sym_hll_binary_expression] = STATE(1150),
    [sym_hll_assignment_expression] = STATE(1150),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1150),
    [sym_hll_char_literal] = STATE(1150),
    [sym_hll_string_literal] = STATE(1150),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2366),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1150),
    [aux_sym_block_repeat1] = STATE(52),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(599),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(52)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1087),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1087),
    [sym_hll_sizeof_expression] = STATE(1087),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2303),
    [sym_hll_update_expression] = STATE(1087),
    [sym_hll_unary_expression] = STATE(1087),
    [sym_hll_binary_expression] = STATE(1087),
    [sym_hll_assignment_expression] = STATE(1087),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1087),
    [sym_hll_char_literal] = STATE(1087),
    [sym_hll_string_literal] = STATE(1087),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2327),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1087),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(53)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1101),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1101),
    [sym_hll_sizeof_expression] = STATE(1101),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2341),
    [sym_hll_update_expression] = STATE(1101),
    [sym_hll_unary_expression] = STATE(1101),
    [sym_hll_binary_expression] = STATE(1101),
    [sym_hll_assignment_expression] = STATE(1101),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1101),
    [sym_hll_char_literal] = STATE(1101),
    [sym_hll_string_literal] = STATE(1101),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2237),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1101),
    [aux_sym_block_repeat1] = STATE(54),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(601),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(54)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1149),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1149),
    [sym_hll_sizeof_expression] = STATE(1149),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2315),
    [sym_hll_update_expression] = STATE(1149),
    [sym_hll_unary_expression] = STATE(1149),
    [sym_hll_binary_expression] = STATE(1149),
    [sym_hll_assignment_expression] = STATE(1149),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1149),
    [sym_hll_char_literal] = STATE(1149),
    [sym_hll_string_literal] = STATE(1149),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2245),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1149),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(55)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1150),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1150),
    [sym_hll_sizeof_expression] = STATE(1150),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2363),
    [sym_hll_update_expression] = STATE(1150),
    [sym_hll_unary_expression] = STATE(1150),
    [sym_hll_binary_expression] = STATE(1150),
    [sym_hll_assignment_expression] = STATE(1150),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1150),
    [sym_hll_char_literal] = STATE(1150),
    [sym_hll_string_literal] = STATE(1150),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2264),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1150),
    [aux_sym_block_repeat1] = STATE(56),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(603),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(56)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1087),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1087),
    [sym_hll_sizeof_expression] = STATE(1087),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2303),
    [sym_hll_update_expression] = STATE(1087),
    [sym_hll_unary_expression] = STATE(1087),
    [sym_hll_binary_expression] = STATE(1087),
    [sym_hll_assignment_expression] = STATE(1087),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1087),
    [sym_hll_char_literal] = STATE(1087),
    [sym_hll_string_literal] = STATE(1087),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2267),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1087),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(57)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1143),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1143),
    [sym_hll_sizeof_expression] = STATE(1143),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2405),
    [sym_hll_update_expression] = STATE(1143),
    [sym_hll_unary_expression] = STATE(1143),
    [sym_hll_binary_expression] = STATE(1143),
    [sym_hll_assignment_expression] = STATE(1143),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1143),
    [sym_hll_char_literal] = STATE(1143),
    [sym_hll_string_literal] = STATE(1143),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2279),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1143),
    [aux_sym_block_repeat1] = STATE(58),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(605),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(58)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1152),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1152),
    [sym_hll_sizeof_expression] = STATE(1152),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2426),
    [sym_hll_update_expression] = STATE(1152),
    [sym_hll_unary_expression] = STATE(1152),
    [sym_hll_binary_expression] = STATE(1152),
    [sym_hll_assignment_expression] = STATE(1152),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1152),
    [sym_hll_char_literal] = STATE(1152),
    [sym_hll_string_literal] = STATE(1152),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2282),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1152),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(59)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1080),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1080),
    [sym_hll_sizeof_expression] = STATE(1080),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2228),
    [sym_hll_update_expression] = STATE(1080),
    [sym_hll_unary_expression] = STATE(1080),
    [sym_hll_binary_expression] = STATE(1080),
    [sym_hll_assignment_expression] = STATE(1080),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1080),
    [sym_hll_char_literal] = STATE(1080),
    [sym_hll_string_literal] = STATE(1080),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2321),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1080),
    [aux_sym_block_repeat1] = STATE(60),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(607),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(60)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1085),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1085),
    [sym_hll_sizeof_expression] = STATE(1085),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2235),
    [sym_hll_update_expression] = STATE(1085),
    [sym_hll_unary_expression] = STATE(1085),
    [sym_hll_binary_expression] = STATE(1085),
    [sym_hll_assignment_expression] = STATE(1085),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1085),
    [sym_hll_char_literal] = STATE(1085),
    [sym_hll_string_literal] = STATE(1085),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2324),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1085),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(61)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1143),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1143),
    [sym_hll_sizeof_expression] = STATE(1143),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2405),
    [sym_hll_update_expression] = STATE(1143),
    [sym_hll_unary_expression] = STATE(1143),
    [sym_hll_binary_expression] = STATE(1143),
    [sym_hll_assignment_expression] = STATE(1143),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1143),
    [sym_hll_char_literal] = STATE(1143),
    [sym_hll_string_literal] = STATE(1143),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2406),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1143),
    [aux_sym_block_repeat1] = STATE(62),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(609),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(62)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1152),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1152),
    [sym_hll_sizeof_expression] = STATE(1152),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2426),
    [sym_hll_update_expression] = STATE(1152),
    [sym_hll_unary_expression] = STATE(1152),
    [sym_hll_binary_expression] = STATE(1152),
    [sym_hll_assignment_expression] = STATE(1152),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1152),
    [sym_hll_char_literal] = STATE(1152),
    [sym_hll_string_literal] = STATE(1152),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2427),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1152),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(63)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1143),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1143),
    [sym_hll_sizeof_expression] = STATE(1143),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2405),
    [sym_hll_update_expression] = STATE(1143),
    [sym_hll_unary_expression] = STATE(1143),
    [sym_hll_binary_expression] = STATE(1143),
    [sym_hll_assignment_expression] = STATE(1143),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1143),
    [sym_hll_char_literal] = STATE(1143),
    [sym_hll_string_literal] = STATE(1143),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2295),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1143),
    [aux_sym_block_repeat1] = STATE(64),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(611),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(64)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1152),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1152),
    [sym_hll_sizeof_expression] = STATE(1152),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2426),
    [sym_hll_update_expression] = STATE(1152),
    [sym_hll_unary_expression] = STATE(1152),
    [sym_hll_binary_expression] = STATE(1152),
    [sym_hll_assignment_expression] = STATE(1152),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1152),
    [sym_hll_char_literal] = STATE(1152),
    [sym_hll_string_literal] = STATE(1152),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2299),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1152),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(65)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1080),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1080),
    [sym_hll_sizeof_expression] = STATE(1080),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2228),
    [sym_hll_update_expression] = STATE(1080),
    [sym_hll_unary_expression] = STATE(1080),
    [sym_hll_binary_expression] = STATE(1080),
    [sym_hll_assignment_expression] = STATE(1080),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1080),
    [sym_hll_char_literal] = STATE(1080),
    [sym_hll_string_literal] = STATE(1080),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2334),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1080),
    [aux_sym_block_repeat1] = STATE(66),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(613),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(66)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1085),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1085),
    [sym_hll_sizeof_expression] = STATE(1085),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2235),
    [sym_hll_update_expression] = STATE(1085),
    [sym_hll_unary_expression] = STATE(1085),
    [sym_hll_binary_expression] = STATE(1085),
    [sym_hll_assignment_expression] = STATE(1085),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1085),
    [sym_hll_char_literal] = STATE(1085),
    [sym_hll_string_literal] = STATE(1085),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2337),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1085),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(67)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1101),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1101),
    [sym_hll_sizeof_expression] = STATE(1101),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2341),
    [sym_hll_update_expression] = STATE(1101),
    [sym_hll_unary_expression] = STATE(1101),
    [sym_hll_binary_expression] = STATE(1101),
    [sym_hll_assignment_expression] = STATE(1101),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1101),
    [sym_hll_char_literal] = STATE(1101),
    [sym_hll_string_literal] = STATE(1101),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2229),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1101),
    [aux_sym_block_repeat1] = STATE(68),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(615),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(68)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1149),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1149),
    [sym_hll_sizeof_expression] = STATE(1149),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2315),
    [sym_hll_update_expression] = STATE(1149),
    [sym_hll_unary_expression] = STATE(1149),
    [sym_hll_binary_expression] = STATE(1149),
    [sym_hll_assignment_expression] = STATE(1149),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1149),
    [sym_hll_char_literal] = STATE(1149),
    [sym_hll_string_literal] = STATE(1149),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2236),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1149),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(69)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1150),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1150),
    [sym_hll_sizeof_expression] = STATE(1150),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2363),
    [sym_hll_update_expression] = STATE(1150),
    [sym_hll_unary_expression] = STATE(1150),
    [sym_hll_binary_expression] = STATE(1150),
    [sym_hll_assignment_expression] = STATE(1150),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1150),
    [sym_hll_char_literal] = STATE(1150),
    [sym_hll_string_literal] = STATE(1150),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2259),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1150),
    [aux_sym_block_repeat1] = STATE(70),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(617),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(70)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1087),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1087),
    [sym_hll_sizeof_expression] = STATE(1087),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2303),
    [sym_hll_update_expression] = STATE(1087),
    [sym_hll_unary_expression] = STATE(1087),
    [sym_hll_binary_expression] = STATE(1087),
    [sym_hll_assignment_expression] = STATE(1087),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1087),
    [sym_hll_char_literal] = STATE(1087),
    [sym_hll_string_literal] = STATE(1087),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2263),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1087),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(71)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1143),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1143),
    [sym_hll_sizeof_expression] = STATE(1143),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2405),
    [sym_hll_update_expression] = STATE(1143),
    [sym_hll_unary_expression] = STATE(1143),
    [sym_hll_binary_expression] = STATE(1143),
    [sym_hll_assignment_expression] = STATE(1143),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1143),
    [sym_hll_char_literal] = STATE(1143),
    [sym_hll_string_literal] = STATE(1143),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2276),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1143),
    [aux_sym_block_repeat1] = STATE(72),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(619),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(72)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1152),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1152),
    [sym_hll_sizeof_expression] = STATE(1152),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2426),
    [sym_hll_update_expression] = STATE(1152),
    [sym_hll_unary_expression] = STATE(1152),
    [sym_hll_binary_expression] = STATE(1152),
    [sym_hll_assignment_expression] = STATE(1152),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1152),
    [sym_hll_char_literal] = STATE(1152),
    [sym_hll_string_literal] = STATE(1152),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2278),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1152),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(73)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1152),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1152),
    [sym_hll_sizeof_expression] = STATE(1152),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2426),
    [sym_hll_update_expression] = STATE(1152),
    [sym_hll_unary_expression] = STATE(1152),
    [sym_hll_binary_expression] = STATE(1152),
    [sym_hll_assignment_expression] = STATE(1152),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1152),
    [sym_hll_char_literal] = STATE(1152),
    [sym_hll_string_literal] = STATE(1152),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2294),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1152),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(74)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1080),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1080),
    [sym_hll_sizeof_expression] = STATE(1080),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2228),
    [sym_hll_update_expression] = STATE(1080),
    [sym_hll_unary_expression] = STATE(1080),
    [sym_hll_binary_expression] = STATE(1080),
    [sym_hll_assignment_expression] = STATE(1080),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1080),
    [sym_hll_char_literal] = STATE(1080),
    [sym_hll_string_literal] = STATE(1080),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2304),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1080),
    [aux_sym_block_repeat1] = STATE(75),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(621),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(75)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1085),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1085),
    [sym_hll_sizeof_expression] = STATE(1085),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2235),
    [sym_hll_update_expression] = STATE(1085),
    [sym_hll_unary_expression] = STATE(1085),
    [sym_hll_binary_expression] = STATE(1085),
    [sym_hll_assignment_expression] = STATE(1085),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1085),
    [sym_hll_char_literal] = STATE(1085),
    [sym_hll_string_literal] = STATE(1085),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2308),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1085),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(76)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1080),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1080),
    [sym_hll_sizeof_expression] = STATE(1080),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2228),
    [sym_hll_update_expression] = STATE(1080),
    [sym_hll_unary_expression] = STATE(1080),
    [sym_hll_binary_expression] = STATE(1080),
    [sym_hll_assignment_expression] = STATE(1080),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1080),
    [sym_hll_char_literal] = STATE(1080),
    [sym_hll_string_literal] = STATE(1080),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2317),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1080),
    [aux_sym_block_repeat1] = STATE(77),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(623),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(77)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1085),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1085),
    [sym_hll_sizeof_expression] = STATE(1085),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2235),
    [sym_hll_update_expression] = STATE(1085),
    [sym_hll_unary_expression] = STATE(1085),
    [sym_hll_binary_expression] = STATE(1085),
    [sym_hll_assignment_expression] = STATE(1085),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1085),
    [sym_hll_char_literal] = STATE(1085),
    [sym_hll_string_literal] = STATE(1085),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2320),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1085),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(78)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1080),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1080),
    [sym_hll_sizeof_expression] = STATE(1080),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2228),
    [sym_hll_update_expression] = STATE(1080),
    [sym_hll_unary_expression] = STATE(1080),
    [sym_hll_binary_expression] = STATE(1080),
    [sym_hll_assignment_expression] = STATE(1080),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1080),
    [sym_hll_char_literal] = STATE(1080),
    [sym_hll_string_literal] = STATE(1080),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2330),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1622),
    [sym_hll_struct_specifier] = STATE(1622),
    [sym_hll_union_specifier] = STATE(1622),
    [sym_hll_enum_specifier] = STATE(1622),
    [sym_hll_sized_type_specifier] = STATE(1622),
    [sym__hll_type_identifier] = STATE(1622),
    [sym_symbol] = STATE(1080),
    [aux_sym_block_repeat1] = STATE(79),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(502),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(625),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(79)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1085),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1085),
    [sym_hll_sizeof_expression] = STATE(1085),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_comma_expression] = STATE(2235),
    [sym_hll_update_expression] = STATE(1085),
    [sym_hll_unary_expression] = STATE(1085),
    [sym_hll_binary_expression] = STATE(1085),
    [sym_hll_assignment_expression] = STATE(1085),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1085),
    [sym_hll_char_literal] = STATE(1085),
    [sym_hll_string_literal] = STATE(1085),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_hll_type_descriptor] = STATE(2333),
    [sym_hll_type_qualifier] = STATE(1605),
    [sym__hll_type_specifier] = STATE(1621),
    [sym_hll_struct_specifier] = STATE(1621),
    [sym_hll_union_specifier] = STATE(1621),
    [sym_hll_enum_specifier] = STATE(1621),
    [sym_hll_sized_type_specifier] = STATE(1621),
    [sym__hll_type_identifier] = STATE(1621),
    [sym_symbol] = STATE(1085),
    [aux_sym_block_repeat1] = STATE(582),
    [aux_sym_hll_type_descriptor_repeat1] = STATE(1605),
    [aux_sym_hll_sized_type_specifier_repeat1] = STATE(1812),
    [sym_identifier] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(492),
    [anon_sym_volatile] = ACTIONS(492),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(496),
    [anon_sym_enum] = ACTIONS(498),
    [anon_sym_signed] = ACTIONS(500),
    [anon_sym_unsigned] = ACTIONS(500),
    [anon_sym_long] = ACTIONS(500),
    [anon_sym_short] = ACTIONS(500),
    [sym_hll_primitive_type] = ACTIONS(516),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(518),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(80)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1021),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1021),
    [sym_hll_sizeof_expression] = STATE(1021),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_update_expression] = STATE(1021),
    [sym_hll_unary_expression] = STATE(1021),
    [sym_hll_binary_expression] = STATE(1021),
    [sym_hll_assignment_expression] = STATE(1021),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1021),
    [sym_hll_char_literal] = STATE(1021),
    [sym_hll_string_literal] = STATE(1021),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_symbol] = STATE(1021),
    [sym_identifier] = ACTIONS(512),
    [anon_sym_RPAREN] = ACTIONS(258),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(476),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_PIPE_PIPE] = ACTIONS(258),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_SLASH] = ACTIONS(260),
    [anon_sym_PERCENT] = ACTIONS(258),
    [anon_sym_PIPE] = ACTIONS(260),
    [anon_sym_CARET] = ACTIONS(258),
    [anon_sym_EQ_EQ] = ACTIONS(258),
    [anon_sym_BANG_EQ] = ACTIONS(258),
    [anon_sym_GT] = ACTIONS(260),
    [anon_sym_GT_EQ] = ACTIONS(258),
    [anon_sym_LT_EQ] = ACTIONS(258),
    [anon_sym_LT] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(258),
    [anon_sym_GT_GT] = ACTIONS(258),
    [anon_sym_DOT_DOT] = ACTIONS(258),
    [anon_sym_AMP_AMP] = ACTIONS(258),
    [anon_sym_AMP] = ACTIONS(242),
    [anon_sym_DOT] = ACTIONS(260),
    [anon_sym_COMMA] = ACTIONS(258),
    [anon_sym_QMARK] = ACTIONS(258),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_DASH_GT] = ACTIONS(258),
    [anon_sym_LBRACK] = ACTIONS(258),
    [anon_sym_RBRACK] = ACTIONS(258),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(260),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(81)] = {
    [sym_trace32_hll_variable] = STATE(857),
    [sym__hll_expression] = STATE(1023),
    [sym_hll_call_expression] = STATE(856),
    [sym_hll_field_expression] = STATE(856),
    [sym_hll_subscript_expression] = STATE(856),
    [sym_hll_cast_expression] = STATE(1023),
    [sym_hll_sizeof_expression] = STATE(1023),
    [sym_hll_pointer_expression] = STATE(856),
    [sym_hll_update_expression] = STATE(1023),
    [sym_hll_unary_expression] = STATE(1023),
    [sym_hll_binary_expression] = STATE(1023),
    [sym_hll_assignment_expression] = STATE(1023),
    [sym__hll_assignment_left_expression] = STATE(1681),
    [sym_hll_conditional_expression] = STATE(1023),
    [sym_hll_char_literal] = STATE(1023),
    [sym_hll_string_literal] = STATE(1023),
    [sym_hll_parenthesized_expression] = STATE(856),
    [sym_symbol] = STATE(1023),
    [sym_identifier] = ACTIONS(512),
    [anon_sym_RPAREN] = ACTIONS(266),
    [anon_sym_LPAREN] = ACTIONS(474),
    [anon_sym_PLUS] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(476),
    [anon_sym_TILDE] = ACTIONS(478),
    [anon_sym_BANG] = ACTIONS(476),
    [anon_sym_PLUS_PLUS] = ACTIONS(480),
    [anon_sym_DASH_DASH] = ACTIONS(480),
    [anon_sym_PIPE_PIPE] = ACTIONS(266),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_SLASH] = ACTIONS(268),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_PIPE] = ACTIONS(268),
    [anon_sym_CARET] = ACTIONS(266),
    [anon_sym_EQ_EQ] = ACTIONS(266),
    [anon_sym_BANG_EQ] = ACTIONS(266),
    [anon_sym_GT] = ACTIONS(268),
    [anon_sym_GT_EQ] = ACTIONS(266),
    [anon_sym_LT_EQ] = ACTIONS(266),
    [anon_sym_LT] = ACTIONS(268),
    [anon_sym_LT_LT] = ACTIONS(266),
    [anon_sym_GT_GT] = ACTIONS(266),
    [anon_sym_DOT_DOT] = ACTIONS(266),
    [anon_sym_AMP_AMP] = ACTIONS(266),
    [anon_sym_AMP] = ACTIONS(242),
    [anon_sym_DOT] = ACTIONS(268),
    [anon_sym_COMMA] = ACTIONS(266),
    [anon_sym_QMARK] = ACTIONS(266),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_DASH_GT] = ACTIONS(266),
    [anon_sym_LBRACK] = ACTIONS(266),
    [anon_sym_RBRACK] = ACTIONS(266),
    [anon_sym_sizeof] = ACTIONS(486),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(268),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(82)] = {
    [sym_trace32_hll_variable] = STATE(891),
    [sym__hll_expression] = STATE(1098),
    [sym_hll_call_expression] = STATE(890),
    [sym_hll_field_expression] = STATE(890),
    [sym_hll_subscript_expression] = STATE(890),
    [sym_hll_cast_expression] = STATE(1098),
    [sym_hll_sizeof_expression] = STATE(1098),
    [sym_hll_pointer_expression] = STATE(890),
    [sym_hll_update_expression] = STATE(1098),
    [sym_hll_unary_expression] = STATE(1098),
    [sym_hll_binary_expression] = STATE(1098),
    [sym_hll_assignment_expression] = STATE(1098),
    [sym__hll_assignment_left_expression] = STATE(1724),
    [sym_hll_conditional_expression] = STATE(1098),
    [sym_hll_char_literal] = STATE(1098),
    [sym_hll_string_literal] = STATE(1098),
    [sym_hll_parenthesized_expression] = STATE(890),
    [sym_symbol] = STATE(1098),
    [sym_identifier] = ACTIONS(627),
    [anon_sym_LPAREN] = ACTIONS(629),
    [anon_sym_PLUS] = ACTIONS(631),
    [anon_sym_DASH] = ACTIONS(631),
    [anon_sym_TILDE] = ACTIONS(633),
    [anon_sym_BANG] = ACTIONS(631),
    [anon_sym_PLUS_PLUS] = ACTIONS(635),
    [anon_sym_DASH_DASH] = ACTIONS(635),
    [anon_sym_PIPE_PIPE] = ACTIONS(230),
    [anon_sym_STAR] = ACTIONS(637),
    [anon_sym_SLASH] = ACTIONS(234),
    [anon_sym_PERCENT] = ACTIONS(230),
    [anon_sym_PIPE] = ACTIONS(234),
    [anon_sym_CARET] = ACTIONS(230),
    [anon_sym_EQ_EQ] = ACTIONS(230),
    [anon_sym_BANG_EQ] = ACTIONS(230),
    [anon_sym_GT] = ACTIONS(234),
    [anon_sym_GT_EQ] = ACTIONS(230),
    [anon_sym_LT_EQ] = ACTIONS(230),
    [anon_sym_LT] = ACTIONS(234),
    [anon_sym_LT_LT] = ACTIONS(230),
    [anon_sym_GT_GT] = ACTIONS(230),
    [anon_sym_AMP_AMP] = ACTIONS(230),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT] = ACTIONS(230),
    [anon_sym_QMARK] = ACTIONS(230),
    [anon_sym_BSLASH] = ACTIONS(639),
    [anon_sym_DASH_GT] = ACTIONS(230),
    [anon_sym_LBRACK] = ACTIONS(230),
    [anon_sym_sizeof] = ACTIONS(641),
    [anon_sym_L_SQUOTE] = ACTIONS(643),
    [anon_sym_u_SQUOTE] = ACTIONS(643),
    [anon_sym_U_SQUOTE] = ACTIONS(643),
    [anon_sym_u8_SQUOTE] = ACTIONS(643),
    [anon_sym_SQUOTE] = ACTIONS(643),
    [anon_sym_L_DQUOTE] = ACTIONS(645),
    [anon_sym_u_DQUOTE] = ACTIONS(645),
    [anon_sym_U_DQUOTE] = ACTIONS(645),
    [anon_sym_u8_DQUOTE] = ACTIONS(645),
    [anon_sym_DQUOTE] = ACTIONS(645),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(230),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(234),
    [aux_sym_symbol_token1] = ACTIONS(647),
    [aux_sym_symbol_token2] = ACTIONS(649),
    [sym_comment] = ACTIONS(230),
    [sym__blank_line] = ACTIONS(230),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(234),
    [sym__hll_number_literal_pre_hook] = ACTIONS(651),
  },
  [STATE(83)] = {
    [sym_trace32_hll_variable] = STATE(891),
    [sym__hll_expression] = STATE(1109),
    [sym_hll_call_expression] = STATE(890),
    [sym_hll_field_expression] = STATE(890),
    [sym_hll_subscript_expression] = STATE(890),
    [sym_hll_cast_expression] = STATE(1109),
    [sym_hll_sizeof_expression] = STATE(1109),
    [sym_hll_pointer_expression] = STATE(890),
    [sym_hll_update_expression] = STATE(1109),
    [sym_hll_unary_expression] = STATE(1109),
    [sym_hll_binary_expression] = STATE(1109),
    [sym_hll_assignment_expression] = STATE(1109),
    [sym__hll_assignment_left_expression] = STATE(1724),
    [sym_hll_conditional_expression] = STATE(1109),
    [sym_hll_char_literal] = STATE(1109),
    [sym_hll_string_literal] = STATE(1109),
    [sym_hll_parenthesized_expression] = STATE(890),
    [sym_symbol] = STATE(1109),
    [sym_identifier] = ACTIONS(627),
    [anon_sym_LPAREN] = ACTIONS(629),
    [anon_sym_PLUS] = ACTIONS(631),
    [anon_sym_DASH] = ACTIONS(631),
    [anon_sym_TILDE] = ACTIONS(633),
    [anon_sym_BANG] = ACTIONS(631),
    [anon_sym_PLUS_PLUS] = ACTIONS(635),
    [anon_sym_DASH_DASH] = ACTIONS(635),
    [anon_sym_PIPE_PIPE] = ACTIONS(262),
    [anon_sym_STAR] = ACTIONS(637),
    [anon_sym_SLASH] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(262),
    [anon_sym_PIPE] = ACTIONS(264),
    [anon_sym_CARET] = ACTIONS(262),
    [anon_sym_EQ_EQ] = ACTIONS(262),
    [anon_sym_BANG_EQ] = ACTIONS(262),
    [anon_sym_GT] = ACTIONS(264),
    [anon_sym_GT_EQ] = ACTIONS(262),
    [anon_sym_LT_EQ] = ACTIONS(262),
    [anon_sym_LT] = ACTIONS(264),
    [anon_sym_LT_LT] = ACTIONS(262),
    [anon_sym_GT_GT] = ACTIONS(262),
    [anon_sym_AMP_AMP] = ACTIONS(262),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT] = ACTIONS(262),
    [anon_sym_QMARK] = ACTIONS(262),
    [anon_sym_BSLASH] = ACTIONS(639),
    [anon_sym_DASH_GT] = ACTIONS(262),
    [anon_sym_LBRACK] = ACTIONS(262),
    [anon_sym_sizeof] = ACTIONS(641),
    [anon_sym_L_SQUOTE] = ACTIONS(643),
    [anon_sym_u_SQUOTE] = ACTIONS(643),
    [anon_sym_U_SQUOTE] = ACTIONS(643),
    [anon_sym_u8_SQUOTE] = ACTIONS(643),
    [anon_sym_SQUOTE] = ACTIONS(643),
    [anon_sym_L_DQUOTE] = ACTIONS(645),
    [anon_sym_u_DQUOTE] = ACTIONS(645),
    [anon_sym_U_DQUOTE] = ACTIONS(645),
    [anon_sym_u8_DQUOTE] = ACTIONS(645),
    [anon_sym_DQUOTE] = ACTIONS(645),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(262),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(264),
    [aux_sym_symbol_token1] = ACTIONS(647),
    [aux_sym_symbol_token2] = ACTIONS(649),
    [sym_comment] = ACTIONS(262),
    [sym__blank_line] = ACTIONS(262),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(264),
    [sym__hll_number_literal_pre_hook] = ACTIONS(651),
  },
  [STATE(84)] = {
    [sym_trace32_hll_variable] = STATE(891),
    [sym__hll_expression] = STATE(1110),
    [sym_hll_call_expression] = STATE(890),
    [sym_hll_field_expression] = STATE(890),
    [sym_hll_subscript_expression] = STATE(890),
    [sym_hll_cast_expression] = STATE(1110),
    [sym_hll_sizeof_expression] = STATE(1110),
    [sym_hll_pointer_expression] = STATE(890),
    [sym_hll_update_expression] = STATE(1110),
    [sym_hll_unary_expression] = STATE(1110),
    [sym_hll_binary_expression] = STATE(1110),
    [sym_hll_assignment_expression] = STATE(1110),
    [sym__hll_assignment_left_expression] = STATE(1724),
    [sym_hll_conditional_expression] = STATE(1110),
    [sym_hll_char_literal] = STATE(1110),
    [sym_hll_string_literal] = STATE(1110),
    [sym_hll_parenthesized_expression] = STATE(890),
    [sym_symbol] = STATE(1110),
    [sym_identifier] = ACTIONS(627),
    [anon_sym_LPAREN] = ACTIONS(629),
    [anon_sym_PLUS] = ACTIONS(631),
    [anon_sym_DASH] = ACTIONS(631),
    [anon_sym_TILDE] = ACTIONS(633),
    [anon_sym_BANG] = ACTIONS(631),
    [anon_sym_PLUS_PLUS] = ACTIONS(635),
    [anon_sym_DASH_DASH] = ACTIONS(635),
    [anon_sym_PIPE_PIPE] = ACTIONS(266),
    [anon_sym_STAR] = ACTIONS(637),
    [anon_sym_SLASH] = ACTIONS(268),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_PIPE] = ACTIONS(268),
    [anon_sym_CARET] = ACTIONS(266),
    [anon_sym_EQ_EQ] = ACTIONS(266),
    [anon_sym_BANG_EQ] = ACTIONS(266),
    [anon_sym_GT] = ACTIONS(268),
    [anon_sym_GT_EQ] = ACTIONS(266),
    [anon_sym_LT_EQ] = ACTIONS(266),
    [anon_sym_LT] = ACTIONS(268),
    [anon_sym_LT_LT] = ACTIONS(266),
    [anon_sym_GT_GT] = ACTIONS(266),
    [anon_sym_AMP_AMP] = ACTIONS(266),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT] = ACTIONS(266),
    [anon_sym_QMARK] = ACTIONS(266),
    [anon_sym_BSLASH] = ACTIONS(639),
    [anon_sym_DASH_GT] = ACTIONS(266),
    [anon_sym_LBRACK] = ACTIONS(266),
    [anon_sym_sizeof] = ACTIONS(641),
    [anon_sym_L_SQUOTE] = ACTIONS(643),
    [anon_sym_u_SQUOTE] = ACTIONS(643),
    [anon_sym_U_SQUOTE] = ACTIONS(643),
    [anon_sym_u8_SQUOTE] = ACTIONS(643),
    [anon_sym_SQUOTE] = ACTIONS(643),
    [anon_sym_L_DQUOTE] = ACTIONS(645),
    [anon_sym_u_DQUOTE] = ACTIONS(645),
    [anon_sym_U_DQUOTE] = ACTIONS(645),
    [anon_sym_u8_DQUOTE] = ACTIONS(645),
    [anon_sym_DQUOTE] = ACTIONS(645),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(266),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(268),
    [aux_sym_symbol_token1] = ACTIONS(647),
    [aux_sym_symbol_token2] = ACTIONS(649),
    [sym_comment] = ACTIONS(266),
    [sym__blank_line] = ACTIONS(266),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(268),
    [sym__hll_number_literal_pre_hook] = ACTIONS(651),
  },
  [STATE(85)] = {
    [sym_trace32_hll_variable] = STATE(891),
    [sym__hll_expression] = STATE(1106),
    [sym_hll_call_expression] = STATE(890),
    [sym_hll_field_expression] = STATE(890),
    [sym_hll_subscript_expression] = STATE(890),
    [sym_hll_cast_expression] = STATE(1106),
    [sym_hll_sizeof_expression] = STATE(1106),
    [sym_hll_pointer_expression] = STATE(890),
    [sym_hll_update_expression] = STATE(1106),
    [sym_hll_unary_expression] = STATE(1106),
    [sym_hll_binary_expression] = STATE(1106),
    [sym_hll_assignment_expression] = STATE(1106),
    [sym__hll_assignment_left_expression] = STATE(1724),
    [sym_hll_conditional_expression] = STATE(1106),
    [sym_hll_char_literal] = STATE(1106),
    [sym_hll_string_literal] = STATE(1106),
    [sym_hll_parenthesized_expression] = STATE(890),
    [sym_symbol] = STATE(1106),
    [sym_identifier] = ACTIONS(627),
    [anon_sym_LPAREN] = ACTIONS(629),
    [anon_sym_PLUS] = ACTIONS(631),
    [anon_sym_DASH] = ACTIONS(631),
    [anon_sym_TILDE] = ACTIONS(633),
    [anon_sym_BANG] = ACTIONS(631),
    [anon_sym_PLUS_PLUS] = ACTIONS(635),
    [anon_sym_DASH_DASH] = ACTIONS(635),
    [anon_sym_PIPE_PIPE] = ACTIONS(258),
    [anon_sym_STAR] = ACTIONS(637),
    [anon_sym_SLASH] = ACTIONS(260),
    [anon_sym_PERCENT] = ACTIONS(258),
    [anon_sym_PIPE] = ACTIONS(260),
    [anon_sym_CARET] = ACTIONS(258),
    [anon_sym_EQ_EQ] = ACTIONS(258),
    [anon_sym_BANG_EQ] = ACTIONS(258),
    [anon_sym_GT] = ACTIONS(260),
    [anon_sym_GT_EQ] = ACTIONS(258),
    [anon_sym_LT_EQ] = ACTIONS(258),
    [anon_sym_LT] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(258),
    [anon_sym_GT_GT] = ACTIONS(258),
    [anon_sym_AMP_AMP] = ACTIONS(258),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT] = ACTIONS(258),
    [anon_sym_QMARK] = ACTIONS(258),
    [anon_sym_BSLASH] = ACTIONS(639),
    [anon_sym_DASH_GT] = ACTIONS(258),
    [anon_sym_LBRACK] = ACTIONS(258),
    [anon_sym_sizeof] = ACTIONS(641),
    [anon_sym_L_SQUOTE] = ACTIONS(643),
    [anon_sym_u_SQUOTE] = ACTIONS(643),
    [anon_sym_U_SQUOTE] = ACTIONS(643),
    [anon_sym_u8_SQUOTE] = ACTIONS(643),
    [anon_sym_SQUOTE] = ACTIONS(643),
    [anon_sym_L_DQUOTE] = ACTIONS(645),
    [anon_sym_u_DQUOTE] = ACTIONS(645),
    [anon_sym_U_DQUOTE] = ACTIONS(645),
    [anon_sym_u8_DQUOTE] = ACTIONS(645),
    [anon_sym_DQUOTE] = ACTIONS(645),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(258),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(260),
    [aux_sym_symbol_token1] = ACTIONS(647),
    [aux_sym_symbol_token2] = ACTIONS(649),
    [sym_comment] = ACTIONS(258),
    [sym__blank_line] = ACTIONS(258),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(260),
    [sym__hll_number_literal_pre_hook] = ACTIONS(651),
  },
  [STATE(86)] = {
    [sym_trace32_hll_variable] = STATE(914),
    [sym__hll_expression] = STATE(1023),
    [sym_hll_call_expression] = STATE(913),
    [sym_hll_field_expression] = STATE(913),
    [sym_hll_subscript_expression] = STATE(913),
    [sym_hll_cast_expression] = STATE(1023),
    [sym_hll_sizeof_expression] = STATE(1023),
    [sym_hll_pointer_expression] = STATE(913),
    [sym_hll_update_expression] = STATE(1023),
    [sym_hll_unary_expression] = STATE(1023),
    [sym_hll_binary_expression] = STATE(1023),
    [sym_hll_assignment_expression] = STATE(1023),
    [sym__hll_assignment_left_expression] = STATE(1705),
    [sym_hll_conditional_expression] = STATE(1023),
    [sym_hll_char_literal] = STATE(1023),
    [sym_hll_string_literal] = STATE(1023),
    [sym_hll_parenthesized_expression] = STATE(913),
    [sym_symbol] = STATE(1023),
    [sym_identifier] = ACTIONS(653),
    [anon_sym_RPAREN] = ACTIONS(266),
    [anon_sym_LPAREN] = ACTIONS(655),
    [anon_sym_PLUS] = ACTIONS(657),
    [anon_sym_DASH] = ACTIONS(657),
    [anon_sym_TILDE] = ACTIONS(659),
    [anon_sym_BANG] = ACTIONS(657),
    [anon_sym_PLUS_PLUS] = ACTIONS(661),
    [anon_sym_DASH_DASH] = ACTIONS(661),
    [anon_sym_PIPE_PIPE] = ACTIONS(266),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_SLASH] = ACTIONS(268),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_PIPE] = ACTIONS(268),
    [anon_sym_CARET] = ACTIONS(266),
    [anon_sym_EQ_EQ] = ACTIONS(266),
    [anon_sym_BANG_EQ] = ACTIONS(266),
    [anon_sym_GT] = ACTIONS(268),
    [anon_sym_GT_EQ] = ACTIONS(266),
    [anon_sym_LT_EQ] = ACTIONS(266),
    [anon_sym_LT] = ACTIONS(268),
    [anon_sym_LT_LT] = ACTIONS(266),
    [anon_sym_GT_GT] = ACTIONS(266),
    [anon_sym_AMP_AMP] = ACTIONS(266),
    [anon_sym_AMP] = ACTIONS(242),
    [anon_sym_COLON] = ACTIONS(266),
    [anon_sym_DOT] = ACTIONS(266),
    [anon_sym_QMARK] = ACTIONS(266),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_DASH_GT] = ACTIONS(266),
    [anon_sym_LBRACK] = ACTIONS(266),
    [anon_sym_RBRACK] = ACTIONS(266),
    [anon_sym_sizeof] = ACTIONS(663),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(87)] = {
    [sym_trace32_hll_variable] = STATE(914),
    [sym__hll_expression] = STATE(1021),
    [sym_hll_call_expression] = STATE(913),
    [sym_hll_field_expression] = STATE(913),
    [sym_hll_subscript_expression] = STATE(913),
    [sym_hll_cast_expression] = STATE(1021),
    [sym_hll_sizeof_expression] = STATE(1021),
    [sym_hll_pointer_expression] = STATE(913),
    [sym_hll_update_expression] = STATE(1021),
    [sym_hll_unary_expression] = STATE(1021),
    [sym_hll_binary_expression] = STATE(1021),
    [sym_hll_assignment_expression] = STATE(1021),
    [sym__hll_assignment_left_expression] = STATE(1705),
    [sym_hll_conditional_expression] = STATE(1021),
    [sym_hll_char_literal] = STATE(1021),
    [sym_hll_string_literal] = STATE(1021),
    [sym_hll_parenthesized_expression] = STATE(913),
    [sym_symbol] = STATE(1021),
    [sym_identifier] = ACTIONS(653),
    [anon_sym_RPAREN] = ACTIONS(258),
    [anon_sym_LPAREN] = ACTIONS(655),
    [anon_sym_PLUS] = ACTIONS(657),
    [anon_sym_DASH] = ACTIONS(657),
    [anon_sym_TILDE] = ACTIONS(659),
    [anon_sym_BANG] = ACTIONS(657),
    [anon_sym_PLUS_PLUS] = ACTIONS(661),
    [anon_sym_DASH_DASH] = ACTIONS(661),
    [anon_sym_PIPE_PIPE] = ACTIONS(258),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_SLASH] = ACTIONS(260),
    [anon_sym_PERCENT] = ACTIONS(258),
    [anon_sym_PIPE] = ACTIONS(260),
    [anon_sym_CARET] = ACTIONS(258),
    [anon_sym_EQ_EQ] = ACTIONS(258),
    [anon_sym_BANG_EQ] = ACTIONS(258),
    [anon_sym_GT] = ACTIONS(260),
    [anon_sym_GT_EQ] = ACTIONS(258),
    [anon_sym_LT_EQ] = ACTIONS(258),
    [anon_sym_LT] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(258),
    [anon_sym_GT_GT] = ACTIONS(258),
    [anon_sym_AMP_AMP] = ACTIONS(258),
    [anon_sym_AMP] = ACTIONS(242),
    [anon_sym_COLON] = ACTIONS(258),
    [anon_sym_DOT] = ACTIONS(258),
    [anon_sym_QMARK] = ACTIONS(258),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_DASH_GT] = ACTIONS(258),
    [anon_sym_LBRACK] = ACTIONS(258),
    [anon_sym_RBRACK] = ACTIONS(258),
    [anon_sym_sizeof] = ACTIONS(663),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(88)] = {
    [sym_trace32_hll_variable] = STATE(914),
    [sym__hll_expression] = STATE(1022),
    [sym_hll_call_expression] = STATE(913),
    [sym_hll_field_expression] = STATE(913),
    [sym_hll_subscript_expression] = STATE(913),
    [sym_hll_cast_expression] = STATE(1022),
    [sym_hll_sizeof_expression] = STATE(1022),
    [sym_hll_pointer_expression] = STATE(913),
    [sym_hll_update_expression] = STATE(1022),
    [sym_hll_unary_expression] = STATE(1022),
    [sym_hll_binary_expression] = STATE(1022),
    [sym_hll_assignment_expression] = STATE(1022),
    [sym__hll_assignment_left_expression] = STATE(1705),
    [sym_hll_conditional_expression] = STATE(1022),
    [sym_hll_char_literal] = STATE(1022),
    [sym_hll_string_literal] = STATE(1022),
    [sym_hll_parenthesized_expression] = STATE(913),
    [sym_symbol] = STATE(1022),
    [sym_identifier] = ACTIONS(653),
    [anon_sym_RPAREN] = ACTIONS(262),
    [anon_sym_LPAREN] = ACTIONS(655),
    [anon_sym_PLUS] = ACTIONS(657),
    [anon_sym_DASH] = ACTIONS(657),
    [anon_sym_TILDE] = ACTIONS(659),
    [anon_sym_BANG] = ACTIONS(657),
    [anon_sym_PLUS_PLUS] = ACTIONS(661),
    [anon_sym_DASH_DASH] = ACTIONS(661),
    [anon_sym_PIPE_PIPE] = ACTIONS(262),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_SLASH] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(262),
    [anon_sym_PIPE] = ACTIONS(264),
    [anon_sym_CARET] = ACTIONS(262),
    [anon_sym_EQ_EQ] = ACTIONS(262),
    [anon_sym_BANG_EQ] = ACTIONS(262),
    [anon_sym_GT] = ACTIONS(264),
    [anon_sym_GT_EQ] = ACTIONS(262),
    [anon_sym_LT_EQ] = ACTIONS(262),
    [anon_sym_LT] = ACTIONS(264),
    [anon_sym_LT_LT] = ACTIONS(262),
    [anon_sym_GT_GT] = ACTIONS(262),
    [anon_sym_AMP_AMP] = ACTIONS(262),
    [anon_sym_AMP] = ACTIONS(242),
    [anon_sym_COLON] = ACTIONS(262),
    [anon_sym_DOT] = ACTIONS(262),
    [anon_sym_QMARK] = ACTIONS(262),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_DASH_GT] = ACTIONS(262),
    [anon_sym_LBRACK] = ACTIONS(262),
    [anon_sym_RBRACK] = ACTIONS(262),
    [anon_sym_sizeof] = ACTIONS(663),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(89)] = {
    [sym_trace32_hll_variable] = STATE(914),
    [sym__hll_expression] = STATE(1019),
    [sym_hll_call_expression] = STATE(913),
    [sym_hll_field_expression] = STATE(913),
    [sym_hll_subscript_expression] = STATE(913),
    [sym_hll_cast_expression] = STATE(1019),
    [sym_hll_sizeof_expression] = STATE(1019),
    [sym_hll_pointer_expression] = STATE(913),
    [sym_hll_update_expression] = STATE(1019),
    [sym_hll_unary_expression] = STATE(1019),
    [sym_hll_binary_expression] = STATE(1019),
    [sym_hll_assignment_expression] = STATE(1019),
    [sym__hll_assignment_left_expression] = STATE(1705),
    [sym_hll_conditional_expression] = STATE(1019),
    [sym_hll_char_literal] = STATE(1019),
    [sym_hll_string_literal] = STATE(1019),
    [sym_hll_parenthesized_expression] = STATE(913),
    [sym_symbol] = STATE(1019),
    [sym_identifier] = ACTIONS(653),
    [anon_sym_RPAREN] = ACTIONS(230),
    [anon_sym_LPAREN] = ACTIONS(655),
    [anon_sym_PLUS] = ACTIONS(657),
    [anon_sym_DASH] = ACTIONS(657),
    [anon_sym_TILDE] = ACTIONS(659),
    [anon_sym_BANG] = ACTIONS(657),
    [anon_sym_PLUS_PLUS] = ACTIONS(661),
    [anon_sym_DASH_DASH] = ACTIONS(661),
    [anon_sym_PIPE_PIPE] = ACTIONS(230),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_SLASH] = ACTIONS(234),
    [anon_sym_PERCENT] = ACTIONS(230),
    [anon_sym_PIPE] = ACTIONS(234),
    [anon_sym_CARET] = ACTIONS(230),
    [anon_sym_EQ_EQ] = ACTIONS(230),
    [anon_sym_BANG_EQ] = ACTIONS(230),
    [anon_sym_GT] = ACTIONS(234),
    [anon_sym_GT_EQ] = ACTIONS(230),
    [anon_sym_LT_EQ] = ACTIONS(230),
    [anon_sym_LT] = ACTIONS(234),
    [anon_sym_LT_LT] = ACTIONS(230),
    [anon_sym_GT_GT] = ACTIONS(230),
    [anon_sym_AMP_AMP] = ACTIONS(230),
    [anon_sym_AMP] = ACTIONS(242),
    [anon_sym_COLON] = ACTIONS(230),
    [anon_sym_DOT] = ACTIONS(230),
    [anon_sym_QMARK] = ACTIONS(230),
    [anon_sym_BSLASH] = ACTIONS(484),
    [anon_sym_DASH_GT] = ACTIONS(230),
    [anon_sym_LBRACK] = ACTIONS(230),
    [anon_sym_RBRACK] = ACTIONS(230),
    [anon_sym_sizeof] = ACTIONS(663),
    [anon_sym_L_SQUOTE] = ACTIONS(488),
    [anon_sym_u_SQUOTE] = ACTIONS(488),
    [anon_sym_U_SQUOTE] = ACTIONS(488),
    [anon_sym_u8_SQUOTE] = ACTIONS(488),
    [anon_sym_SQUOTE] = ACTIONS(488),
    [anon_sym_L_DQUOTE] = ACTIONS(490),
    [anon_sym_u_DQUOTE] = ACTIONS(490),
    [anon_sym_U_DQUOTE] = ACTIONS(490),
    [anon_sym_u8_DQUOTE] = ACTIONS(490),
    [anon_sym_DQUOTE] = ACTIONS(490),
    [aux_sym_symbol_token1] = ACTIONS(504),
    [aux_sym_symbol_token2] = ACTIONS(506),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__hll_number_literal_pre_hook] = ACTIONS(510),
  },
  [STATE(90)] = {
    [sym_trace32_hll_variable] = STATE(939),
    [sym__hll_expression] = STATE(1178),
    [sym_hll_call_expression] = STATE(938),
    [sym_hll_field_expression] = STATE(938),
    [sym_hll_subscript_expression] = STATE(938),
    [sym_hll_cast_expression] = STATE(1178),
    [sym_hll_sizeof_expression] = STATE(1178),
    [sym_hll_pointer_expression] = STATE(938),
    [sym_hll_update_expression] = STATE(1178),
    [sym_hll_unary_expression] = STATE(1178),
    [sym_hll_binary_expression] = STATE(1178),
    [sym_hll_assignment_expression] = STATE(1178),
    [sym__hll_assignment_left_expression] = STATE(1712),
    [sym_hll_conditional_expression] = STATE(1178),
    [sym_hll_char_literal] = STATE(1178),
    [sym_hll_string_literal] = STATE(1178),
    [sym_hll_parenthesized_expression] = STATE(938),
    [sym_symbol] = STATE(1178),
    [sym_identifier] = ACTIONS(665),
    [anon_sym_LPAREN] = ACTIONS(667),
    [anon_sym_PLUS] = ACTIONS(669),
    [anon_sym_DASH] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(671),
    [anon_sym_BANG] = ACTIONS(669),
    [anon_sym_PLUS_PLUS] = ACTIONS(673),
    [anon_sym_DASH_DASH] = ACTIONS(673),
    [anon_sym_PIPE_PIPE] = ACTIONS(230),
    [anon_sym_STAR] = ACTIONS(675),
    [anon_sym_SLASH] = ACTIONS(234),
    [anon_sym_PERCENT] = ACTIONS(230),
    [anon_sym_PIPE] = ACTIONS(234),
    [anon_sym_CARET] = ACTIONS(230),
    [anon_sym_EQ_EQ] = ACTIONS(230),
    [anon_sym_BANG_EQ] = ACTIONS(230),
    [anon_sym_GT] = ACTIONS(234),
    [anon_sym_GT_EQ] = ACTIONS(230),
    [anon_sym_LT_EQ] = ACTIONS(230),
    [anon_sym_LT] = ACTIONS(234),
    [anon_sym_LT_LT] = ACTIONS(230),
    [anon_sym_GT_GT] = ACTIONS(230),
    [anon_sym_AMP_AMP] = ACTIONS(230),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(230),
    [anon_sym_QMARK] = ACTIONS(230),
    [anon_sym_BSLASH] = ACTIONS(677),
    [anon_sym_DASH_GT] = ACTIONS(230),
    [anon_sym_LBRACK] = ACTIONS(230),
    [anon_sym_sizeof] = ACTIONS(679),
    [anon_sym_L_SQUOTE] = ACTIONS(681),
    [anon_sym_u_SQUOTE] = ACTIONS(681),
    [anon_sym_U_SQUOTE] = ACTIONS(681),
    [anon_sym_u8_SQUOTE] = ACTIONS(681),
    [anon_sym_SQUOTE] = ACTIONS(681),
    [anon_sym_L_DQUOTE] = ACTIONS(683),
    [anon_sym_u_DQUOTE] = ACTIONS(683),
    [anon_sym_U_DQUOTE] = ACTIONS(683),
    [anon_sym_u8_DQUOTE] = ACTIONS(683),
    [anon_sym_DQUOTE] = ACTIONS(683),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(230),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(234),
    [aux_sym_symbol_token1] = ACTIONS(685),
    [aux_sym_symbol_token2] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(234),
    [sym__hll_number_literal_pre_hook] = ACTIONS(689),
  },
  [STATE(91)] = {
    [sym_trace32_hll_variable] = STATE(939),
    [sym__hll_expression] = STATE(1192),
    [sym_hll_call_expression] = STATE(938),
    [sym_hll_field_expression] = STATE(938),
    [sym_hll_subscript_expression] = STATE(938),
    [sym_hll_cast_expression] = STATE(1192),
    [sym_hll_sizeof_expression] = STATE(1192),
    [sym_hll_pointer_expression] = STATE(938),
    [sym_hll_update_expression] = STATE(1192),
    [sym_hll_unary_expression] = STATE(1192),
    [sym_hll_binary_expression] = STATE(1192),
    [sym_hll_assignment_expression] = STATE(1192),
    [sym__hll_assignment_left_expression] = STATE(1712),
    [sym_hll_conditional_expression] = STATE(1192),
    [sym_hll_char_literal] = STATE(1192),
    [sym_hll_string_literal] = STATE(1192),
    [sym_hll_parenthesized_expression] = STATE(938),
    [sym_symbol] = STATE(1192),
    [sym_identifier] = ACTIONS(665),
    [anon_sym_LPAREN] = ACTIONS(667),
    [anon_sym_PLUS] = ACTIONS(669),
    [anon_sym_DASH] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(671),
    [anon_sym_BANG] = ACTIONS(669),
    [anon_sym_PLUS_PLUS] = ACTIONS(673),
    [anon_sym_DASH_DASH] = ACTIONS(673),
    [anon_sym_PIPE_PIPE] = ACTIONS(258),
    [anon_sym_STAR] = ACTIONS(675),
    [anon_sym_SLASH] = ACTIONS(260),
    [anon_sym_PERCENT] = ACTIONS(258),
    [anon_sym_PIPE] = ACTIONS(260),
    [anon_sym_CARET] = ACTIONS(258),
    [anon_sym_EQ_EQ] = ACTIONS(258),
    [anon_sym_BANG_EQ] = ACTIONS(258),
    [anon_sym_GT] = ACTIONS(260),
    [anon_sym_GT_EQ] = ACTIONS(258),
    [anon_sym_LT_EQ] = ACTIONS(258),
    [anon_sym_LT] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(258),
    [anon_sym_GT_GT] = ACTIONS(258),
    [anon_sym_AMP_AMP] = ACTIONS(258),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(258),
    [anon_sym_QMARK] = ACTIONS(258),
    [anon_sym_BSLASH] = ACTIONS(677),
    [anon_sym_DASH_GT] = ACTIONS(258),
    [anon_sym_LBRACK] = ACTIONS(258),
    [anon_sym_sizeof] = ACTIONS(679),
    [anon_sym_L_SQUOTE] = ACTIONS(681),
    [anon_sym_u_SQUOTE] = ACTIONS(681),
    [anon_sym_U_SQUOTE] = ACTIONS(681),
    [anon_sym_u8_SQUOTE] = ACTIONS(681),
    [anon_sym_SQUOTE] = ACTIONS(681),
    [anon_sym_L_DQUOTE] = ACTIONS(683),
    [anon_sym_u_DQUOTE] = ACTIONS(683),
    [anon_sym_U_DQUOTE] = ACTIONS(683),
    [anon_sym_u8_DQUOTE] = ACTIONS(683),
    [anon_sym_DQUOTE] = ACTIONS(683),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(258),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(260),
    [aux_sym_symbol_token1] = ACTIONS(685),
    [aux_sym_symbol_token2] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(260),
    [sym__hll_number_literal_pre_hook] = ACTIONS(689),
  },
  [STATE(92)] = {
    [sym_trace32_hll_variable] = STATE(939),
    [sym__hll_expression] = STATE(1193),
    [sym_hll_call_expression] = STATE(938),
    [sym_hll_field_expression] = STATE(938),
    [sym_hll_subscript_expression] = STATE(938),
    [sym_hll_cast_expression] = STATE(1193),
    [sym_hll_sizeof_expression] = STATE(1193),
    [sym_hll_pointer_expression] = STATE(938),
    [sym_hll_update_expression] = STATE(1193),
    [sym_hll_unary_expression] = STATE(1193),
    [sym_hll_binary_expression] = STATE(1193),
    [sym_hll_assignment_expression] = STATE(1193),
    [sym__hll_assignment_left_expression] = STATE(1712),
    [sym_hll_conditional_expression] = STATE(1193),
    [sym_hll_char_literal] = STATE(1193),
    [sym_hll_string_literal] = STATE(1193),
    [sym_hll_parenthesized_expression] = STATE(938),
    [sym_symbol] = STATE(1193),
    [sym_identifier] = ACTIONS(665),
    [anon_sym_LPAREN] = ACTIONS(667),
    [anon_sym_PLUS] = ACTIONS(669),
    [anon_sym_DASH] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(671),
    [anon_sym_BANG] = ACTIONS(669),
    [anon_sym_PLUS_PLUS] = ACTIONS(673),
    [anon_sym_DASH_DASH] = ACTIONS(673),
    [anon_sym_PIPE_PIPE] = ACTIONS(262),
    [anon_sym_STAR] = ACTIONS(675),
    [anon_sym_SLASH] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(262),
    [anon_sym_PIPE] = ACTIONS(264),
    [anon_sym_CARET] = ACTIONS(262),
    [anon_sym_EQ_EQ] = ACTIONS(262),
    [anon_sym_BANG_EQ] = ACTIONS(262),
    [anon_sym_GT] = ACTIONS(264),
    [anon_sym_GT_EQ] = ACTIONS(262),
    [anon_sym_LT_EQ] = ACTIONS(262),
    [anon_sym_LT] = ACTIONS(264),
    [anon_sym_LT_LT] = ACTIONS(262),
    [anon_sym_GT_GT] = ACTIONS(262),
    [anon_sym_AMP_AMP] = ACTIONS(262),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(262),
    [anon_sym_QMARK] = ACTIONS(262),
    [anon_sym_BSLASH] = ACTIONS(677),
    [anon_sym_DASH_GT] = ACTIONS(262),
    [anon_sym_LBRACK] = ACTIONS(262),
    [anon_sym_sizeof] = ACTIONS(679),
    [anon_sym_L_SQUOTE] = ACTIONS(681),
    [anon_sym_u_SQUOTE] = ACTIONS(681),
    [anon_sym_U_SQUOTE] = ACTIONS(681),
    [anon_sym_u8_SQUOTE] = ACTIONS(681),
    [anon_sym_SQUOTE] = ACTIONS(681),
    [anon_sym_L_DQUOTE] = ACTIONS(683),
    [anon_sym_u_DQUOTE] = ACTIONS(683),
    [anon_sym_U_DQUOTE] = ACTIONS(683),
    [anon_sym_u8_DQUOTE] = ACTIONS(683),
    [anon_sym_DQUOTE] = ACTIONS(683),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(262),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(264),
    [aux_sym_symbol_token1] = ACTIONS(685),
    [aux_sym_symbol_token2] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(264),
    [sym__hll_number_literal_pre_hook] = ACTIONS(689),
  },
  [STATE(93)] = {
    [sym_trace32_hll_variable] = STATE(939),
    [sym__hll_expression] = STATE(1228),
    [sym_hll_call_expression] = STATE(938),
    [sym_hll_field_expression] = STATE(938),
    [sym_hll_subscript_expression] = STATE(938),
    [sym_hll_cast_expression] = STATE(1228),
    [sym_hll_sizeof_expression] = STATE(1228),
    [sym_hll_pointer_expression] = STATE(938),
    [sym_hll_update_expression] = STATE(1228),
    [sym_hll_unary_expression] = STATE(1228),
    [sym_hll_binary_expression] = STATE(1228),
    [sym_hll_assignment_expression] = STATE(1228),
    [sym__hll_assignment_left_expression] = STATE(1712),
    [sym_hll_conditional_expression] = STATE(1228),
    [sym_hll_char_literal] = STATE(1228),
    [sym_hll_string_literal] = STATE(1228),
    [sym_hll_parenthesized_expression] = STATE(938),
    [sym_symbol] = STATE(1228),
    [sym_identifier] = ACTIONS(665),
    [anon_sym_LPAREN] = ACTIONS(667),
    [anon_sym_PLUS] = ACTIONS(669),
    [anon_sym_DASH] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(671),
    [anon_sym_BANG] = ACTIONS(669),
    [anon_sym_PLUS_PLUS] = ACTIONS(673),
    [anon_sym_DASH_DASH] = ACTIONS(673),
    [anon_sym_PIPE_PIPE] = ACTIONS(266),
    [anon_sym_STAR] = ACTIONS(675),
    [anon_sym_SLASH] = ACTIONS(268),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_PIPE] = ACTIONS(268),
    [anon_sym_CARET] = ACTIONS(266),
    [anon_sym_EQ_EQ] = ACTIONS(266),
    [anon_sym_BANG_EQ] = ACTIONS(266),
    [anon_sym_GT] = ACTIONS(268),
    [anon_sym_GT_EQ] = ACTIONS(266),
    [anon_sym_LT_EQ] = ACTIONS(266),
    [anon_sym_LT] = ACTIONS(268),
    [anon_sym_LT_LT] = ACTIONS(266),
    [anon_sym_GT_GT] = ACTIONS(266),
    [anon_sym_AMP_AMP] = ACTIONS(266),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(266),
    [anon_sym_QMARK] = ACTIONS(266),
    [anon_sym_BSLASH] = ACTIONS(677),
    [anon_sym_DASH_GT] = ACTIONS(266),
    [anon_sym_LBRACK] = ACTIONS(266),
    [anon_sym_sizeof] = ACTIONS(679),
    [anon_sym_L_SQUOTE] = ACTIONS(681),
    [anon_sym_u_SQUOTE] = ACTIONS(681),
    [anon_sym_U_SQUOTE] = ACTIONS(681),
    [anon_sym_u8_SQUOTE] = ACTIONS(681),
    [anon_sym_SQUOTE] = ACTIONS(681),
    [anon_sym_L_DQUOTE] = ACTIONS(683),
    [anon_sym_u_DQUOTE] = ACTIONS(683),
    [anon_sym_U_DQUOTE] = ACTIONS(683),
    [anon_sym_u8_DQUOTE] = ACTIONS(683),
    [anon_sym_DQUOTE] = ACTIONS(683),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(266),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(268),
    [aux_sym_symbol_token1] = ACTIONS(685),
    [aux_sym_symbol_token2] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(268),
    [sym__hll_number_literal_pre_hook] = ACTIONS(689),
  },
  [STATE(94)] = {
    [sym__expression] = STATE(1443),
    [sym__parenthesized_expression] = STATE(1443),
    [sym__path_expression] = STATE(1443),
    [sym_unary_expression] = STATE(1443),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1443),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1443),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1443),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1443),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1443),
    [sym_string] = STATE(1443),
    [sym_character] = STATE(1443),
    [sym_symbol] = STATE(1443),
    [sym__special_command_operator] = STATE(1443),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(691),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(693),
    [sym_time] = ACTIONS(691),
    [sym_frequency] = ACTIONS(691),
    [sym_percentage] = ACTIONS(693),
    [sym_file_handle] = ACTIONS(693),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(95)] = {
    [sym__expression] = STATE(1263),
    [sym__parenthesized_expression] = STATE(1263),
    [sym__path_expression] = STATE(1263),
    [sym_unary_expression] = STATE(1263),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1263),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1263),
    [sym_macro] = STATE(1030),
    [sym_call_expression] = STATE(1126),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1263),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1263),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1127),
    [sym_bitmask] = STATE(1263),
    [sym_string] = STATE(1263),
    [sym_character] = STATE(1263),
    [sym_symbol] = STATE(1263),
    [sym__special_command_operator] = STATE(1263),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(96),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(146),
    [anon_sym_DASH] = ACTIONS(146),
    [anon_sym_TILDE] = ACTIONS(148),
    [anon_sym_BANG] = ACTIONS(148),
    [anon_sym_PLUS_PLUS] = ACTIONS(150),
    [anon_sym_DASH_DASH] = ACTIONS(150),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(695),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(697),
    [sym_time] = ACTIONS(695),
    [sym_frequency] = ACTIONS(695),
    [sym_percentage] = ACTIONS(697),
    [sym_file_handle] = ACTIONS(697),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(699),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(96)] = {
    [sym__expression] = STATE(1233),
    [sym__parenthesized_expression] = STATE(1233),
    [sym__path_expression] = STATE(1233),
    [sym_unary_expression] = STATE(1233),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1233),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1233),
    [sym_macro] = STATE(1030),
    [sym_call_expression] = STATE(1126),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1233),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1233),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1127),
    [sym_bitmask] = STATE(1233),
    [sym_string] = STATE(1233),
    [sym_character] = STATE(1233),
    [sym_symbol] = STATE(1233),
    [sym__special_command_operator] = STATE(1233),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(146),
    [anon_sym_DASH] = ACTIONS(146),
    [anon_sym_TILDE] = ACTIONS(148),
    [anon_sym_BANG] = ACTIONS(148),
    [anon_sym_PLUS_PLUS] = ACTIONS(150),
    [anon_sym_DASH_DASH] = ACTIONS(150),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(701),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(703),
    [sym_time] = ACTIONS(701),
    [sym_frequency] = ACTIONS(701),
    [sym_percentage] = ACTIONS(703),
    [sym_file_handle] = ACTIONS(703),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(97)] = {
    [sym__expression] = STATE(1258),
    [sym__parenthesized_expression] = STATE(1258),
    [sym__path_expression] = STATE(1258),
    [sym_unary_expression] = STATE(1258),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1258),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1258),
    [sym_macro] = STATE(1030),
    [sym_call_expression] = STATE(1126),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1258),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1258),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1127),
    [sym_bitmask] = STATE(1258),
    [sym_string] = STATE(1258),
    [sym_character] = STATE(1258),
    [sym_symbol] = STATE(1258),
    [sym__special_command_operator] = STATE(1258),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(98),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(146),
    [anon_sym_DASH] = ACTIONS(146),
    [anon_sym_TILDE] = ACTIONS(148),
    [anon_sym_BANG] = ACTIONS(148),
    [anon_sym_PLUS_PLUS] = ACTIONS(150),
    [anon_sym_DASH_DASH] = ACTIONS(150),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(705),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(707),
    [sym_time] = ACTIONS(705),
    [sym_frequency] = ACTIONS(705),
    [sym_percentage] = ACTIONS(707),
    [sym_file_handle] = ACTIONS(707),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(709),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(98)] = {
    [sym__expression] = STATE(1170),
    [sym__parenthesized_expression] = STATE(1170),
    [sym__path_expression] = STATE(1170),
    [sym_unary_expression] = STATE(1170),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1170),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1170),
    [sym_macro] = STATE(1030),
    [sym_call_expression] = STATE(1126),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1170),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1170),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1127),
    [sym_bitmask] = STATE(1170),
    [sym_string] = STATE(1170),
    [sym_character] = STATE(1170),
    [sym_symbol] = STATE(1170),
    [sym__special_command_operator] = STATE(1170),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(146),
    [anon_sym_DASH] = ACTIONS(146),
    [anon_sym_TILDE] = ACTIONS(148),
    [anon_sym_BANG] = ACTIONS(148),
    [anon_sym_PLUS_PLUS] = ACTIONS(150),
    [anon_sym_DASH_DASH] = ACTIONS(150),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(711),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(713),
    [sym_time] = ACTIONS(711),
    [sym_frequency] = ACTIONS(711),
    [sym_percentage] = ACTIONS(713),
    [sym_file_handle] = ACTIONS(713),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(99)] = {
    [sym__expression] = STATE(13),
    [sym__parenthesized_expression] = STATE(13),
    [sym__path_expression] = STATE(13),
    [sym_unary_expression] = STATE(13),
    [sym__unary_expression] = STATE(581),
    [sym__prefix_postfix_expression] = STATE(564),
    [sym_binary_expression] = STATE(13),
    [sym__binary_expression] = STATE(565),
    [sym__and_expression] = STATE(573),
    [sym_assignment_expression] = STATE(13),
    [sym_macro] = STATE(406),
    [sym_call_expression] = STATE(424),
    [sym__function_identifier] = STATE(2589),
    [sym__hll_var_call_expression_function] = STATE(2590),
    [sym__hll_var_call_expression] = STATE(428),
    [sym_trace32_hll_variable] = STATE(575),
    [sym__literal] = STATE(13),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(211),
    [sym_address] = STATE(13),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(411),
    [sym_bitmask] = STATE(13),
    [sym_string] = STATE(13),
    [sym_character] = STATE(13),
    [sym_symbol] = STATE(13),
    [sym__special_command_operator] = STATE(13),
    [sym__file_dialog] = STATE(569),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(443),
    [aux_sym__path_expression_repeat2] = STATE(13),
    [sym_identifier] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(57),
    [anon_sym_BANG] = ACTIONS(57),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(715),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(67),
    [anon_sym_SQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_access_class] = ACTIONS(73),
    [aux_sym_integer_token1] = ACTIONS(75),
    [aux_sym_integer_token2] = ACTIONS(75),
    [sym_float] = ACTIONS(717),
    [aux_sym_bitmask_token1] = ACTIONS(79),
    [aux_sym_bitmask_token2] = ACTIONS(79),
    [aux_sym_symbol_token1] = ACTIONS(81),
    [aux_sym_symbol_token2] = ACTIONS(83),
    [sym__composed_name] = ACTIONS(719),
    [sym_time] = ACTIONS(717),
    [sym_frequency] = ACTIONS(717),
    [sym_percentage] = ACTIONS(719),
    [sym_file_handle] = ACTIONS(719),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(87),
    [sym_path] = ACTIONS(719),
  },
  [STATE(100)] = {
    [sym__expression] = STATE(1483),
    [sym__parenthesized_expression] = STATE(1483),
    [sym__path_expression] = STATE(1483),
    [sym_unary_expression] = STATE(1483),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1483),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1483),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1483),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1483),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1483),
    [sym_string] = STATE(1483),
    [sym_character] = STATE(1483),
    [sym_symbol] = STATE(1483),
    [sym__special_command_operator] = STATE(1483),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(721),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(723),
    [sym_time] = ACTIONS(721),
    [sym_frequency] = ACTIONS(721),
    [sym_percentage] = ACTIONS(723),
    [sym_file_handle] = ACTIONS(723),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(101)] = {
    [sym__expression] = STATE(1356),
    [sym__parenthesized_expression] = STATE(1356),
    [sym__path_expression] = STATE(1356),
    [sym_unary_expression] = STATE(1356),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1356),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1356),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1356),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1356),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1356),
    [sym_string] = STATE(1356),
    [sym_character] = STATE(1356),
    [sym_symbol] = STATE(1356),
    [sym__special_command_operator] = STATE(1356),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(135),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(725),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(727),
    [sym_time] = ACTIONS(725),
    [sym_frequency] = ACTIONS(725),
    [sym_percentage] = ACTIONS(727),
    [sym_file_handle] = ACTIONS(727),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(729),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(102)] = {
    [sym_trace32_hll_variable] = STATE(944),
    [sym__hll_expression] = STATE(1178),
    [sym_hll_call_expression] = STATE(942),
    [sym_hll_field_expression] = STATE(942),
    [sym_hll_subscript_expression] = STATE(942),
    [sym_hll_cast_expression] = STATE(1178),
    [sym_hll_sizeof_expression] = STATE(1178),
    [sym_hll_pointer_expression] = STATE(942),
    [sym_hll_update_expression] = STATE(1178),
    [sym_hll_unary_expression] = STATE(1178),
    [sym_hll_binary_expression] = STATE(1178),
    [sym_hll_assignment_expression] = STATE(1178),
    [sym__hll_assignment_left_expression] = STATE(1710),
    [sym_hll_conditional_expression] = STATE(1178),
    [sym_hll_char_literal] = STATE(1178),
    [sym_hll_string_literal] = STATE(1178),
    [sym_hll_parenthesized_expression] = STATE(942),
    [sym_symbol] = STATE(1178),
    [sym_identifier] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_PLUS] = ACTIONS(735),
    [anon_sym_DASH] = ACTIONS(735),
    [anon_sym_TILDE] = ACTIONS(737),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_PLUS_PLUS] = ACTIONS(739),
    [anon_sym_DASH_DASH] = ACTIONS(739),
    [anon_sym_PIPE_PIPE] = ACTIONS(230),
    [anon_sym_STAR] = ACTIONS(675),
    [anon_sym_SLASH] = ACTIONS(234),
    [anon_sym_PERCENT] = ACTIONS(230),
    [anon_sym_PIPE] = ACTIONS(234),
    [anon_sym_CARET] = ACTIONS(230),
    [anon_sym_EQ_EQ] = ACTIONS(230),
    [anon_sym_BANG_EQ] = ACTIONS(230),
    [anon_sym_GT] = ACTIONS(234),
    [anon_sym_GT_EQ] = ACTIONS(230),
    [anon_sym_LT_EQ] = ACTIONS(230),
    [anon_sym_LT] = ACTIONS(234),
    [anon_sym_LT_LT] = ACTIONS(230),
    [anon_sym_GT_GT] = ACTIONS(230),
    [anon_sym_AMP_AMP] = ACTIONS(230),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(230),
    [anon_sym_QMARK] = ACTIONS(230),
    [anon_sym_BSLASH] = ACTIONS(677),
    [anon_sym_DASH_GT] = ACTIONS(230),
    [anon_sym_LBRACK] = ACTIONS(230),
    [anon_sym_sizeof] = ACTIONS(741),
    [anon_sym_L_SQUOTE] = ACTIONS(681),
    [anon_sym_u_SQUOTE] = ACTIONS(681),
    [anon_sym_U_SQUOTE] = ACTIONS(681),
    [anon_sym_u8_SQUOTE] = ACTIONS(681),
    [anon_sym_SQUOTE] = ACTIONS(681),
    [anon_sym_L_DQUOTE] = ACTIONS(683),
    [anon_sym_u_DQUOTE] = ACTIONS(683),
    [anon_sym_U_DQUOTE] = ACTIONS(683),
    [anon_sym_u8_DQUOTE] = ACTIONS(683),
    [anon_sym_DQUOTE] = ACTIONS(683),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(230),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(234),
    [aux_sym_symbol_token1] = ACTIONS(685),
    [aux_sym_symbol_token2] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__hll_number_literal_pre_hook] = ACTIONS(689),
  },
  [STATE(103)] = {
    [sym_trace32_hll_variable] = STATE(944),
    [sym__hll_expression] = STATE(1192),
    [sym_hll_call_expression] = STATE(942),
    [sym_hll_field_expression] = STATE(942),
    [sym_hll_subscript_expression] = STATE(942),
    [sym_hll_cast_expression] = STATE(1192),
    [sym_hll_sizeof_expression] = STATE(1192),
    [sym_hll_pointer_expression] = STATE(942),
    [sym_hll_update_expression] = STATE(1192),
    [sym_hll_unary_expression] = STATE(1192),
    [sym_hll_binary_expression] = STATE(1192),
    [sym_hll_assignment_expression] = STATE(1192),
    [sym__hll_assignment_left_expression] = STATE(1710),
    [sym_hll_conditional_expression] = STATE(1192),
    [sym_hll_char_literal] = STATE(1192),
    [sym_hll_string_literal] = STATE(1192),
    [sym_hll_parenthesized_expression] = STATE(942),
    [sym_symbol] = STATE(1192),
    [sym_identifier] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_PLUS] = ACTIONS(735),
    [anon_sym_DASH] = ACTIONS(735),
    [anon_sym_TILDE] = ACTIONS(737),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_PLUS_PLUS] = ACTIONS(739),
    [anon_sym_DASH_DASH] = ACTIONS(739),
    [anon_sym_PIPE_PIPE] = ACTIONS(258),
    [anon_sym_STAR] = ACTIONS(675),
    [anon_sym_SLASH] = ACTIONS(260),
    [anon_sym_PERCENT] = ACTIONS(258),
    [anon_sym_PIPE] = ACTIONS(260),
    [anon_sym_CARET] = ACTIONS(258),
    [anon_sym_EQ_EQ] = ACTIONS(258),
    [anon_sym_BANG_EQ] = ACTIONS(258),
    [anon_sym_GT] = ACTIONS(260),
    [anon_sym_GT_EQ] = ACTIONS(258),
    [anon_sym_LT_EQ] = ACTIONS(258),
    [anon_sym_LT] = ACTIONS(260),
    [anon_sym_LT_LT] = ACTIONS(258),
    [anon_sym_GT_GT] = ACTIONS(258),
    [anon_sym_AMP_AMP] = ACTIONS(258),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(258),
    [anon_sym_QMARK] = ACTIONS(258),
    [anon_sym_BSLASH] = ACTIONS(677),
    [anon_sym_DASH_GT] = ACTIONS(258),
    [anon_sym_LBRACK] = ACTIONS(258),
    [anon_sym_sizeof] = ACTIONS(741),
    [anon_sym_L_SQUOTE] = ACTIONS(681),
    [anon_sym_u_SQUOTE] = ACTIONS(681),
    [anon_sym_U_SQUOTE] = ACTIONS(681),
    [anon_sym_u8_SQUOTE] = ACTIONS(681),
    [anon_sym_SQUOTE] = ACTIONS(681),
    [anon_sym_L_DQUOTE] = ACTIONS(683),
    [anon_sym_u_DQUOTE] = ACTIONS(683),
    [anon_sym_U_DQUOTE] = ACTIONS(683),
    [anon_sym_u8_DQUOTE] = ACTIONS(683),
    [anon_sym_DQUOTE] = ACTIONS(683),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(258),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(260),
    [aux_sym_symbol_token1] = ACTIONS(685),
    [aux_sym_symbol_token2] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__hll_number_literal_pre_hook] = ACTIONS(689),
  },
  [STATE(104)] = {
    [sym_trace32_hll_variable] = STATE(944),
    [sym__hll_expression] = STATE(1193),
    [sym_hll_call_expression] = STATE(942),
    [sym_hll_field_expression] = STATE(942),
    [sym_hll_subscript_expression] = STATE(942),
    [sym_hll_cast_expression] = STATE(1193),
    [sym_hll_sizeof_expression] = STATE(1193),
    [sym_hll_pointer_expression] = STATE(942),
    [sym_hll_update_expression] = STATE(1193),
    [sym_hll_unary_expression] = STATE(1193),
    [sym_hll_binary_expression] = STATE(1193),
    [sym_hll_assignment_expression] = STATE(1193),
    [sym__hll_assignment_left_expression] = STATE(1710),
    [sym_hll_conditional_expression] = STATE(1193),
    [sym_hll_char_literal] = STATE(1193),
    [sym_hll_string_literal] = STATE(1193),
    [sym_hll_parenthesized_expression] = STATE(942),
    [sym_symbol] = STATE(1193),
    [sym_identifier] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_PLUS] = ACTIONS(735),
    [anon_sym_DASH] = ACTIONS(735),
    [anon_sym_TILDE] = ACTIONS(737),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_PLUS_PLUS] = ACTIONS(739),
    [anon_sym_DASH_DASH] = ACTIONS(739),
    [anon_sym_PIPE_PIPE] = ACTIONS(262),
    [anon_sym_STAR] = ACTIONS(675),
    [anon_sym_SLASH] = ACTIONS(264),
    [anon_sym_PERCENT] = ACTIONS(262),
    [anon_sym_PIPE] = ACTIONS(264),
    [anon_sym_CARET] = ACTIONS(262),
    [anon_sym_EQ_EQ] = ACTIONS(262),
    [anon_sym_BANG_EQ] = ACTIONS(262),
    [anon_sym_GT] = ACTIONS(264),
    [anon_sym_GT_EQ] = ACTIONS(262),
    [anon_sym_LT_EQ] = ACTIONS(262),
    [anon_sym_LT] = ACTIONS(264),
    [anon_sym_LT_LT] = ACTIONS(262),
    [anon_sym_GT_GT] = ACTIONS(262),
    [anon_sym_AMP_AMP] = ACTIONS(262),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(262),
    [anon_sym_QMARK] = ACTIONS(262),
    [anon_sym_BSLASH] = ACTIONS(677),
    [anon_sym_DASH_GT] = ACTIONS(262),
    [anon_sym_LBRACK] = ACTIONS(262),
    [anon_sym_sizeof] = ACTIONS(741),
    [anon_sym_L_SQUOTE] = ACTIONS(681),
    [anon_sym_u_SQUOTE] = ACTIONS(681),
    [anon_sym_U_SQUOTE] = ACTIONS(681),
    [anon_sym_u8_SQUOTE] = ACTIONS(681),
    [anon_sym_SQUOTE] = ACTIONS(681),
    [anon_sym_L_DQUOTE] = ACTIONS(683),
    [anon_sym_u_DQUOTE] = ACTIONS(683),
    [anon_sym_U_DQUOTE] = ACTIONS(683),
    [anon_sym_u8_DQUOTE] = ACTIONS(683),
    [anon_sym_DQUOTE] = ACTIONS(683),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(262),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(264),
    [aux_sym_symbol_token1] = ACTIONS(685),
    [aux_sym_symbol_token2] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__hll_number_literal_pre_hook] = ACTIONS(689),
  },
  [STATE(105)] = {
    [sym_trace32_hll_variable] = STATE(944),
    [sym__hll_expression] = STATE(1228),
    [sym_hll_call_expression] = STATE(942),
    [sym_hll_field_expression] = STATE(942),
    [sym_hll_subscript_expression] = STATE(942),
    [sym_hll_cast_expression] = STATE(1228),
    [sym_hll_sizeof_expression] = STATE(1228),
    [sym_hll_pointer_expression] = STATE(942),
    [sym_hll_update_expression] = STATE(1228),
    [sym_hll_unary_expression] = STATE(1228),
    [sym_hll_binary_expression] = STATE(1228),
    [sym_hll_assignment_expression] = STATE(1228),
    [sym__hll_assignment_left_expression] = STATE(1710),
    [sym_hll_conditional_expression] = STATE(1228),
    [sym_hll_char_literal] = STATE(1228),
    [sym_hll_string_literal] = STATE(1228),
    [sym_hll_parenthesized_expression] = STATE(942),
    [sym_symbol] = STATE(1228),
    [sym_identifier] = ACTIONS(731),
    [anon_sym_LPAREN] = ACTIONS(733),
    [anon_sym_PLUS] = ACTIONS(735),
    [anon_sym_DASH] = ACTIONS(735),
    [anon_sym_TILDE] = ACTIONS(737),
    [anon_sym_BANG] = ACTIONS(735),
    [anon_sym_PLUS_PLUS] = ACTIONS(739),
    [anon_sym_DASH_DASH] = ACTIONS(739),
    [anon_sym_PIPE_PIPE] = ACTIONS(266),
    [anon_sym_STAR] = ACTIONS(675),
    [anon_sym_SLASH] = ACTIONS(268),
    [anon_sym_PERCENT] = ACTIONS(266),
    [anon_sym_PIPE] = ACTIONS(268),
    [anon_sym_CARET] = ACTIONS(266),
    [anon_sym_EQ_EQ] = ACTIONS(266),
    [anon_sym_BANG_EQ] = ACTIONS(266),
    [anon_sym_GT] = ACTIONS(268),
    [anon_sym_GT_EQ] = ACTIONS(266),
    [anon_sym_LT_EQ] = ACTIONS(266),
    [anon_sym_LT] = ACTIONS(268),
    [anon_sym_LT_LT] = ACTIONS(266),
    [anon_sym_GT_GT] = ACTIONS(266),
    [anon_sym_AMP_AMP] = ACTIONS(266),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(266),
    [anon_sym_QMARK] = ACTIONS(266),
    [anon_sym_BSLASH] = ACTIONS(677),
    [anon_sym_DASH_GT] = ACTIONS(266),
    [anon_sym_LBRACK] = ACTIONS(266),
    [anon_sym_sizeof] = ACTIONS(741),
    [anon_sym_L_SQUOTE] = ACTIONS(681),
    [anon_sym_u_SQUOTE] = ACTIONS(681),
    [anon_sym_U_SQUOTE] = ACTIONS(681),
    [anon_sym_u8_SQUOTE] = ACTIONS(681),
    [anon_sym_SQUOTE] = ACTIONS(681),
    [anon_sym_L_DQUOTE] = ACTIONS(683),
    [anon_sym_u_DQUOTE] = ACTIONS(683),
    [anon_sym_U_DQUOTE] = ACTIONS(683),
    [anon_sym_u8_DQUOTE] = ACTIONS(683),
    [anon_sym_DQUOTE] = ACTIONS(683),
    [aux_sym__hll_expression_separator_token1] = ACTIONS(266),
    [aux_sym__hll_expression_separator_token2] = ACTIONS(268),
    [aux_sym_symbol_token1] = ACTIONS(685),
    [aux_sym_symbol_token2] = ACTIONS(687),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__hll_number_literal_pre_hook] = ACTIONS(689),
  },
  [STATE(106)] = {
    [sym__expression] = STATE(1316),
    [sym__parenthesized_expression] = STATE(1316),
    [sym__path_expression] = STATE(1316),
    [sym_unary_expression] = STATE(1316),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1316),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1316),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1316),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1316),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1316),
    [sym_string] = STATE(1316),
    [sym_character] = STATE(1316),
    [sym_symbol] = STATE(1316),
    [sym__special_command_operator] = STATE(1316),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(743),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(745),
    [sym_time] = ACTIONS(743),
    [sym_frequency] = ACTIONS(743),
    [sym_percentage] = ACTIONS(745),
    [sym_file_handle] = ACTIONS(745),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(107)] = {
    [sym__expression] = STATE(1298),
    [sym__parenthesized_expression] = STATE(1298),
    [sym__path_expression] = STATE(1298),
    [sym_unary_expression] = STATE(1298),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1298),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1298),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1298),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1298),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1298),
    [sym_string] = STATE(1298),
    [sym_character] = STATE(1298),
    [sym_symbol] = STATE(1298),
    [sym__special_command_operator] = STATE(1298),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(112),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(747),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(749),
    [sym_time] = ACTIONS(747),
    [sym_frequency] = ACTIONS(747),
    [sym_percentage] = ACTIONS(749),
    [sym_file_handle] = ACTIONS(749),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(751),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(108)] = {
    [sym__expression] = STATE(1430),
    [sym__parenthesized_expression] = STATE(1430),
    [sym__path_expression] = STATE(1430),
    [sym_unary_expression] = STATE(1430),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1430),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1430),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1430),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1430),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1430),
    [sym_string] = STATE(1430),
    [sym_character] = STATE(1430),
    [sym_symbol] = STATE(1430),
    [sym__special_command_operator] = STATE(1430),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(753),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(755),
    [sym_time] = ACTIONS(753),
    [sym_frequency] = ACTIONS(753),
    [sym_percentage] = ACTIONS(755),
    [sym_file_handle] = ACTIONS(755),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(109)] = {
    [sym__expression] = STATE(1365),
    [sym__parenthesized_expression] = STATE(1365),
    [sym__path_expression] = STATE(1365),
    [sym_unary_expression] = STATE(1365),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1365),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1365),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1365),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1365),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1365),
    [sym_string] = STATE(1365),
    [sym_character] = STATE(1365),
    [sym_symbol] = STATE(1365),
    [sym__special_command_operator] = STATE(1365),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(757),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(759),
    [sym_time] = ACTIONS(757),
    [sym_frequency] = ACTIONS(757),
    [sym_percentage] = ACTIONS(759),
    [sym_file_handle] = ACTIONS(759),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(110)] = {
    [sym__expression] = STATE(1493),
    [sym__parenthesized_expression] = STATE(1493),
    [sym__path_expression] = STATE(1493),
    [sym_unary_expression] = STATE(1493),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1493),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1493),
    [sym_macro] = STATE(1241),
    [sym_call_expression] = STATE(1115),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1536),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1536),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1536),
    [sym_string] = STATE(1536),
    [sym_character] = STATE(1536),
    [sym_symbol] = STATE(1536),
    [sym__special_command_operator] = STATE(1536),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(761),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(106),
    [anon_sym_DASH] = ACTIONS(106),
    [anon_sym_TILDE] = ACTIONS(104),
    [anon_sym_BANG] = ACTIONS(104),
    [anon_sym_PLUS_PLUS] = ACTIONS(222),
    [anon_sym_DASH_DASH] = ACTIONS(222),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(763),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(765),
    [sym_time] = ACTIONS(763),
    [sym_frequency] = ACTIONS(763),
    [sym_percentage] = ACTIONS(765),
    [sym_file_handle] = ACTIONS(765),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(138),
  },
  [STATE(111)] = {
    [sym__expression] = STATE(1334),
    [sym__parenthesized_expression] = STATE(1334),
    [sym__path_expression] = STATE(1334),
    [sym_unary_expression] = STATE(1334),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1334),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1334),
    [sym_macro] = STATE(1030),
    [sym_call_expression] = STATE(1126),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1334),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1334),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1127),
    [sym_bitmask] = STATE(1334),
    [sym_string] = STATE(1334),
    [sym_character] = STATE(1334),
    [sym_symbol] = STATE(1334),
    [sym__special_command_operator] = STATE(1334),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(119),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(146),
    [anon_sym_DASH] = ACTIONS(146),
    [anon_sym_TILDE] = ACTIONS(148),
    [anon_sym_BANG] = ACTIONS(148),
    [anon_sym_PLUS_PLUS] = ACTIONS(150),
    [anon_sym_DASH_DASH] = ACTIONS(150),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(767),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(769),
    [sym_time] = ACTIONS(767),
    [sym_frequency] = ACTIONS(767),
    [sym_percentage] = ACTIONS(769),
    [sym_file_handle] = ACTIONS(769),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(771),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(112)] = {
    [sym__expression] = STATE(1296),
    [sym__parenthesized_expression] = STATE(1296),
    [sym__path_expression] = STATE(1296),
    [sym_unary_expression] = STATE(1296),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1296),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1296),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1296),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1296),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1296),
    [sym_string] = STATE(1296),
    [sym_character] = STATE(1296),
    [sym_symbol] = STATE(1296),
    [sym__special_command_operator] = STATE(1296),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(773),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(775),
    [sym_time] = ACTIONS(773),
    [sym_frequency] = ACTIONS(773),
    [sym_percentage] = ACTIONS(775),
    [sym_file_handle] = ACTIONS(775),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(113)] = {
    [sym__expression] = STATE(1493),
    [sym__parenthesized_expression] = STATE(1493),
    [sym__path_expression] = STATE(1493),
    [sym_unary_expression] = STATE(1493),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1493),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1493),
    [sym_macro] = STATE(1158),
    [sym_call_expression] = STATE(1115),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1533),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1533),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1533),
    [sym_string] = STATE(1533),
    [sym_character] = STATE(1533),
    [sym_symbol] = STATE(1533),
    [sym__special_command_operator] = STATE(1533),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(777),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(106),
    [anon_sym_DASH] = ACTIONS(106),
    [anon_sym_TILDE] = ACTIONS(104),
    [anon_sym_BANG] = ACTIONS(104),
    [anon_sym_PLUS_PLUS] = ACTIONS(222),
    [anon_sym_DASH_DASH] = ACTIONS(222),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(779),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(781),
    [sym_time] = ACTIONS(779),
    [sym_frequency] = ACTIONS(779),
    [sym_percentage] = ACTIONS(781),
    [sym_file_handle] = ACTIONS(781),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(138),
  },
  [STATE(114)] = {
    [sym__expression] = STATE(1413),
    [sym__parenthesized_expression] = STATE(1413),
    [sym__path_expression] = STATE(1413),
    [sym_unary_expression] = STATE(1413),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1413),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1413),
    [sym_macro] = STATE(1054),
    [sym_call_expression] = STATE(1115),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1413),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1413),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1118),
    [sym_bitmask] = STATE(1413),
    [sym_string] = STATE(1413),
    [sym_character] = STATE(1413),
    [sym_symbol] = STATE(1413),
    [sym__special_command_operator] = STATE(1413),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(116),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(106),
    [anon_sym_DASH] = ACTIONS(106),
    [anon_sym_TILDE] = ACTIONS(104),
    [anon_sym_BANG] = ACTIONS(104),
    [anon_sym_PLUS_PLUS] = ACTIONS(222),
    [anon_sym_DASH_DASH] = ACTIONS(222),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(783),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(785),
    [sym_time] = ACTIONS(783),
    [sym_frequency] = ACTIONS(783),
    [sym_percentage] = ACTIONS(785),
    [sym_file_handle] = ACTIONS(785),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(787),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(115)] = {
    [sym__expression] = STATE(1380),
    [sym__parenthesized_expression] = STATE(1380),
    [sym__path_expression] = STATE(1380),
    [sym_unary_expression] = STATE(1380),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1380),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1380),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1380),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1380),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1380),
    [sym_string] = STATE(1380),
    [sym_character] = STATE(1380),
    [sym_symbol] = STATE(1380),
    [sym__special_command_operator] = STATE(1380),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(789),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(791),
    [sym_time] = ACTIONS(789),
    [sym_frequency] = ACTIONS(789),
    [sym_percentage] = ACTIONS(791),
    [sym_file_handle] = ACTIONS(791),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(116)] = {
    [sym__expression] = STATE(1424),
    [sym__parenthesized_expression] = STATE(1424),
    [sym__path_expression] = STATE(1424),
    [sym_unary_expression] = STATE(1424),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1424),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1424),
    [sym_macro] = STATE(1054),
    [sym_call_expression] = STATE(1115),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1424),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1424),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1118),
    [sym_bitmask] = STATE(1424),
    [sym_string] = STATE(1424),
    [sym_character] = STATE(1424),
    [sym_symbol] = STATE(1424),
    [sym__special_command_operator] = STATE(1424),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(106),
    [anon_sym_DASH] = ACTIONS(106),
    [anon_sym_TILDE] = ACTIONS(104),
    [anon_sym_BANG] = ACTIONS(104),
    [anon_sym_PLUS_PLUS] = ACTIONS(222),
    [anon_sym_DASH_DASH] = ACTIONS(222),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(793),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(795),
    [sym_time] = ACTIONS(793),
    [sym_frequency] = ACTIONS(793),
    [sym_percentage] = ACTIONS(795),
    [sym_file_handle] = ACTIONS(795),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(117)] = {
    [sym__expression] = STATE(1425),
    [sym__parenthesized_expression] = STATE(1425),
    [sym__path_expression] = STATE(1425),
    [sym_unary_expression] = STATE(1425),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1425),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1425),
    [sym_macro] = STATE(1054),
    [sym_call_expression] = STATE(1115),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1425),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1425),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1118),
    [sym_bitmask] = STATE(1425),
    [sym_string] = STATE(1425),
    [sym_character] = STATE(1425),
    [sym_symbol] = STATE(1425),
    [sym__special_command_operator] = STATE(1425),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(118),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(106),
    [anon_sym_DASH] = ACTIONS(106),
    [anon_sym_TILDE] = ACTIONS(104),
    [anon_sym_BANG] = ACTIONS(104),
    [anon_sym_PLUS_PLUS] = ACTIONS(222),
    [anon_sym_DASH_DASH] = ACTIONS(222),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(797),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(799),
    [sym_time] = ACTIONS(797),
    [sym_frequency] = ACTIONS(797),
    [sym_percentage] = ACTIONS(799),
    [sym_file_handle] = ACTIONS(799),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(801),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(118)] = {
    [sym__expression] = STATE(1426),
    [sym__parenthesized_expression] = STATE(1426),
    [sym__path_expression] = STATE(1426),
    [sym_unary_expression] = STATE(1426),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1426),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1426),
    [sym_macro] = STATE(1054),
    [sym_call_expression] = STATE(1115),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1426),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1426),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1118),
    [sym_bitmask] = STATE(1426),
    [sym_string] = STATE(1426),
    [sym_character] = STATE(1426),
    [sym_symbol] = STATE(1426),
    [sym__special_command_operator] = STATE(1426),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(106),
    [anon_sym_DASH] = ACTIONS(106),
    [anon_sym_TILDE] = ACTIONS(104),
    [anon_sym_BANG] = ACTIONS(104),
    [anon_sym_PLUS_PLUS] = ACTIONS(222),
    [anon_sym_DASH_DASH] = ACTIONS(222),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(803),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(805),
    [sym_time] = ACTIONS(803),
    [sym_frequency] = ACTIONS(803),
    [sym_percentage] = ACTIONS(805),
    [sym_file_handle] = ACTIONS(805),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(119)] = {
    [sym__expression] = STATE(1317),
    [sym__parenthesized_expression] = STATE(1317),
    [sym__path_expression] = STATE(1317),
    [sym_unary_expression] = STATE(1317),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1317),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1317),
    [sym_macro] = STATE(1030),
    [sym_call_expression] = STATE(1126),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1317),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1317),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1127),
    [sym_bitmask] = STATE(1317),
    [sym_string] = STATE(1317),
    [sym_character] = STATE(1317),
    [sym_symbol] = STATE(1317),
    [sym__special_command_operator] = STATE(1317),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(146),
    [anon_sym_DASH] = ACTIONS(146),
    [anon_sym_TILDE] = ACTIONS(148),
    [anon_sym_BANG] = ACTIONS(148),
    [anon_sym_PLUS_PLUS] = ACTIONS(150),
    [anon_sym_DASH_DASH] = ACTIONS(150),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(807),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(809),
    [sym_time] = ACTIONS(807),
    [sym_frequency] = ACTIONS(807),
    [sym_percentage] = ACTIONS(809),
    [sym_file_handle] = ACTIONS(809),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(120)] = {
    [sym__expression] = STATE(1403),
    [sym__parenthesized_expression] = STATE(1403),
    [sym__path_expression] = STATE(1403),
    [sym_unary_expression] = STATE(1403),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1403),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1403),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1403),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1403),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1403),
    [sym_string] = STATE(1403),
    [sym_character] = STATE(1403),
    [sym_symbol] = STATE(1403),
    [sym__special_command_operator] = STATE(1403),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(811),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(813),
    [sym_time] = ACTIONS(811),
    [sym_frequency] = ACTIONS(811),
    [sym_percentage] = ACTIONS(813),
    [sym_file_handle] = ACTIONS(813),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(121)] = {
    [sym__expression] = STATE(1317),
    [sym__parenthesized_expression] = STATE(1317),
    [sym__path_expression] = STATE(1317),
    [sym_unary_expression] = STATE(1317),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1317),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1317),
    [sym_macro] = STATE(1030),
    [sym_call_expression] = STATE(1126),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1317),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1317),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1127),
    [sym_bitmask] = STATE(1317),
    [sym_string] = STATE(1317),
    [sym_character] = STATE(1317),
    [sym_symbol] = STATE(1317),
    [sym__special_command_operator] = STATE(1317),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(125),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(146),
    [anon_sym_DASH] = ACTIONS(146),
    [anon_sym_TILDE] = ACTIONS(148),
    [anon_sym_BANG] = ACTIONS(148),
    [anon_sym_PLUS_PLUS] = ACTIONS(150),
    [anon_sym_DASH_DASH] = ACTIONS(150),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(807),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(809),
    [sym_time] = ACTIONS(807),
    [sym_frequency] = ACTIONS(807),
    [sym_percentage] = ACTIONS(809),
    [sym_file_handle] = ACTIONS(809),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(815),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(122)] = {
    [sym__expression] = STATE(1388),
    [sym__parenthesized_expression] = STATE(1388),
    [sym__path_expression] = STATE(1388),
    [sym_unary_expression] = STATE(1388),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1388),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1388),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1388),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1388),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1388),
    [sym_string] = STATE(1388),
    [sym_character] = STATE(1388),
    [sym_symbol] = STATE(1388),
    [sym__special_command_operator] = STATE(1388),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(817),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(819),
    [sym_time] = ACTIONS(817),
    [sym_frequency] = ACTIONS(817),
    [sym_percentage] = ACTIONS(819),
    [sym_file_handle] = ACTIONS(819),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(123)] = {
    [sym__expression] = STATE(16),
    [sym__parenthesized_expression] = STATE(16),
    [sym__path_expression] = STATE(16),
    [sym_unary_expression] = STATE(16),
    [sym__unary_expression] = STATE(581),
    [sym__prefix_postfix_expression] = STATE(564),
    [sym_binary_expression] = STATE(16),
    [sym__binary_expression] = STATE(565),
    [sym__and_expression] = STATE(573),
    [sym_assignment_expression] = STATE(16),
    [sym_macro] = STATE(406),
    [sym_call_expression] = STATE(424),
    [sym__function_identifier] = STATE(2589),
    [sym__hll_var_call_expression_function] = STATE(2590),
    [sym__hll_var_call_expression] = STATE(428),
    [sym_trace32_hll_variable] = STATE(575),
    [sym__literal] = STATE(16),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(211),
    [sym_address] = STATE(16),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(411),
    [sym_bitmask] = STATE(16),
    [sym_string] = STATE(16),
    [sym_character] = STATE(16),
    [sym_symbol] = STATE(16),
    [sym__special_command_operator] = STATE(16),
    [sym__file_dialog] = STATE(569),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(443),
    [aux_sym__path_expression_repeat2] = STATE(16),
    [sym_identifier] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(57),
    [anon_sym_BANG] = ACTIONS(57),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(715),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(67),
    [anon_sym_SQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_access_class] = ACTIONS(73),
    [aux_sym_integer_token1] = ACTIONS(75),
    [aux_sym_integer_token2] = ACTIONS(75),
    [sym_float] = ACTIONS(821),
    [aux_sym_bitmask_token1] = ACTIONS(79),
    [aux_sym_bitmask_token2] = ACTIONS(79),
    [aux_sym_symbol_token1] = ACTIONS(81),
    [aux_sym_symbol_token2] = ACTIONS(83),
    [sym__composed_name] = ACTIONS(823),
    [sym_time] = ACTIONS(821),
    [sym_frequency] = ACTIONS(821),
    [sym_percentage] = ACTIONS(823),
    [sym_file_handle] = ACTIONS(823),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(87),
    [sym_path] = ACTIONS(823),
  },
  [STATE(124)] = {
    [sym__expression] = STATE(1400),
    [sym__parenthesized_expression] = STATE(1400),
    [sym__path_expression] = STATE(1400),
    [sym_unary_expression] = STATE(1400),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1400),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1400),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1400),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1400),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1400),
    [sym_string] = STATE(1400),
    [sym_character] = STATE(1400),
    [sym_symbol] = STATE(1400),
    [sym__special_command_operator] = STATE(1400),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(126),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(825),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(827),
    [sym_time] = ACTIONS(825),
    [sym_frequency] = ACTIONS(825),
    [sym_percentage] = ACTIONS(827),
    [sym_file_handle] = ACTIONS(827),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(829),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(125)] = {
    [sym__expression] = STATE(1318),
    [sym__parenthesized_expression] = STATE(1318),
    [sym__path_expression] = STATE(1318),
    [sym_unary_expression] = STATE(1318),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1318),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1318),
    [sym_macro] = STATE(1030),
    [sym_call_expression] = STATE(1126),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1318),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1318),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1127),
    [sym_bitmask] = STATE(1318),
    [sym_string] = STATE(1318),
    [sym_character] = STATE(1318),
    [sym_symbol] = STATE(1318),
    [sym__special_command_operator] = STATE(1318),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(140),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(146),
    [anon_sym_DASH] = ACTIONS(146),
    [anon_sym_TILDE] = ACTIONS(148),
    [anon_sym_BANG] = ACTIONS(148),
    [anon_sym_PLUS_PLUS] = ACTIONS(150),
    [anon_sym_DASH_DASH] = ACTIONS(150),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(831),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(833),
    [sym_time] = ACTIONS(831),
    [sym_frequency] = ACTIONS(831),
    [sym_percentage] = ACTIONS(833),
    [sym_file_handle] = ACTIONS(833),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(126)] = {
    [sym__expression] = STATE(1404),
    [sym__parenthesized_expression] = STATE(1404),
    [sym__path_expression] = STATE(1404),
    [sym_unary_expression] = STATE(1404),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1404),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1404),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1404),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1404),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1404),
    [sym_string] = STATE(1404),
    [sym_character] = STATE(1404),
    [sym_symbol] = STATE(1404),
    [sym__special_command_operator] = STATE(1404),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(835),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(837),
    [sym_time] = ACTIONS(835),
    [sym_frequency] = ACTIONS(835),
    [sym_percentage] = ACTIONS(837),
    [sym_file_handle] = ACTIONS(837),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(127)] = {
    [sym__expression] = STATE(1405),
    [sym__parenthesized_expression] = STATE(1405),
    [sym__path_expression] = STATE(1405),
    [sym_unary_expression] = STATE(1405),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1405),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1405),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1405),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1405),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1405),
    [sym_string] = STATE(1405),
    [sym_character] = STATE(1405),
    [sym_symbol] = STATE(1405),
    [sym__special_command_operator] = STATE(1405),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(160),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(839),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(841),
    [sym_time] = ACTIONS(839),
    [sym_frequency] = ACTIONS(839),
    [sym_percentage] = ACTIONS(841),
    [sym_file_handle] = ACTIONS(841),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(843),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(128)] = {
    [sym__expression] = STATE(1394),
    [sym__parenthesized_expression] = STATE(1394),
    [sym__path_expression] = STATE(1394),
    [sym_unary_expression] = STATE(1394),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1394),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1394),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1394),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1394),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1394),
    [sym_string] = STATE(1394),
    [sym_character] = STATE(1394),
    [sym_symbol] = STATE(1394),
    [sym__special_command_operator] = STATE(1394),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(845),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(847),
    [sym_time] = ACTIONS(845),
    [sym_frequency] = ACTIONS(845),
    [sym_percentage] = ACTIONS(847),
    [sym_file_handle] = ACTIONS(847),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(129)] = {
    [sym__expression] = STATE(1395),
    [sym__parenthesized_expression] = STATE(1395),
    [sym__path_expression] = STATE(1395),
    [sym_unary_expression] = STATE(1395),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1395),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1395),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1395),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1395),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1395),
    [sym_string] = STATE(1395),
    [sym_character] = STATE(1395),
    [sym_symbol] = STATE(1395),
    [sym__special_command_operator] = STATE(1395),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(849),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(851),
    [sym_time] = ACTIONS(849),
    [sym_frequency] = ACTIONS(849),
    [sym_percentage] = ACTIONS(851),
    [sym_file_handle] = ACTIONS(851),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(130)] = {
    [sym__expression] = STATE(1484),
    [sym__parenthesized_expression] = STATE(1484),
    [sym__path_expression] = STATE(1484),
    [sym_unary_expression] = STATE(1484),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1484),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1484),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1484),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1484),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1484),
    [sym_string] = STATE(1484),
    [sym_character] = STATE(1484),
    [sym_symbol] = STATE(1484),
    [sym__special_command_operator] = STATE(1484),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(853),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(855),
    [sym_time] = ACTIONS(853),
    [sym_frequency] = ACTIONS(853),
    [sym_percentage] = ACTIONS(855),
    [sym_file_handle] = ACTIONS(855),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(131)] = {
    [sym__expression] = STATE(1409),
    [sym__parenthesized_expression] = STATE(1409),
    [sym__path_expression] = STATE(1409),
    [sym_unary_expression] = STATE(1409),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1409),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1409),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1409),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1409),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1409),
    [sym_string] = STATE(1409),
    [sym_character] = STATE(1409),
    [sym_symbol] = STATE(1409),
    [sym__special_command_operator] = STATE(1409),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(857),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(859),
    [sym_time] = ACTIONS(857),
    [sym_frequency] = ACTIONS(857),
    [sym_percentage] = ACTIONS(859),
    [sym_file_handle] = ACTIONS(859),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(132)] = {
    [sym__expression] = STATE(1410),
    [sym__parenthesized_expression] = STATE(1410),
    [sym__path_expression] = STATE(1410),
    [sym_unary_expression] = STATE(1410),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1410),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1410),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1410),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1410),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1410),
    [sym_string] = STATE(1410),
    [sym_character] = STATE(1410),
    [sym_symbol] = STATE(1410),
    [sym__special_command_operator] = STATE(1410),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(861),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(863),
    [sym_time] = ACTIONS(861),
    [sym_frequency] = ACTIONS(861),
    [sym_percentage] = ACTIONS(863),
    [sym_file_handle] = ACTIONS(863),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(133)] = {
    [sym__expression] = STATE(1489),
    [sym__parenthesized_expression] = STATE(1489),
    [sym__path_expression] = STATE(1489),
    [sym_unary_expression] = STATE(1489),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1489),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1489),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1489),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1489),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1489),
    [sym_string] = STATE(1489),
    [sym_character] = STATE(1489),
    [sym_symbol] = STATE(1489),
    [sym__special_command_operator] = STATE(1489),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(865),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(867),
    [sym_time] = ACTIONS(865),
    [sym_frequency] = ACTIONS(865),
    [sym_percentage] = ACTIONS(867),
    [sym_file_handle] = ACTIONS(867),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(134)] = {
    [sym__expression] = STATE(1412),
    [sym__parenthesized_expression] = STATE(1412),
    [sym__path_expression] = STATE(1412),
    [sym_unary_expression] = STATE(1412),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1412),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1412),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1412),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1412),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1412),
    [sym_string] = STATE(1412),
    [sym_character] = STATE(1412),
    [sym_symbol] = STATE(1412),
    [sym__special_command_operator] = STATE(1412),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(869),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(871),
    [sym_time] = ACTIONS(869),
    [sym_frequency] = ACTIONS(869),
    [sym_percentage] = ACTIONS(871),
    [sym_file_handle] = ACTIONS(871),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(135)] = {
    [sym__expression] = STATE(1398),
    [sym__parenthesized_expression] = STATE(1398),
    [sym__path_expression] = STATE(1398),
    [sym_unary_expression] = STATE(1398),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1398),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1398),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1398),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1398),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1398),
    [sym_string] = STATE(1398),
    [sym_character] = STATE(1398),
    [sym_symbol] = STATE(1398),
    [sym__special_command_operator] = STATE(1398),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(873),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(875),
    [sym_time] = ACTIONS(873),
    [sym_frequency] = ACTIONS(873),
    [sym_percentage] = ACTIONS(875),
    [sym_file_handle] = ACTIONS(875),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(136)] = {
    [sym__expression] = STATE(450),
    [sym__parenthesized_expression] = STATE(450),
    [sym__path_expression] = STATE(450),
    [sym_unary_expression] = STATE(450),
    [sym__unary_expression] = STATE(581),
    [sym__prefix_postfix_expression] = STATE(564),
    [sym_binary_expression] = STATE(450),
    [sym__binary_expression] = STATE(565),
    [sym__and_expression] = STATE(573),
    [sym_assignment_expression] = STATE(450),
    [sym_macro] = STATE(406),
    [sym_call_expression] = STATE(424),
    [sym__function_identifier] = STATE(2589),
    [sym__hll_var_call_expression_function] = STATE(2590),
    [sym__hll_var_call_expression] = STATE(428),
    [sym_trace32_hll_variable] = STATE(575),
    [sym__literal] = STATE(450),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(211),
    [sym_address] = STATE(450),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(411),
    [sym_bitmask] = STATE(450),
    [sym_string] = STATE(450),
    [sym_character] = STATE(450),
    [sym_symbol] = STATE(450),
    [sym__special_command_operator] = STATE(450),
    [sym__file_dialog] = STATE(569),
    [aux_sym_block_repeat1] = STATE(139),
    [aux_sym__path_expression_repeat1] = STATE(443),
    [aux_sym__path_expression_repeat2] = STATE(450),
    [sym_identifier] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(57),
    [anon_sym_BANG] = ACTIONS(57),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(715),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(67),
    [anon_sym_SQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_access_class] = ACTIONS(73),
    [aux_sym_integer_token1] = ACTIONS(75),
    [aux_sym_integer_token2] = ACTIONS(75),
    [sym_float] = ACTIONS(877),
    [aux_sym_bitmask_token1] = ACTIONS(79),
    [aux_sym_bitmask_token2] = ACTIONS(79),
    [aux_sym_symbol_token1] = ACTIONS(81),
    [aux_sym_symbol_token2] = ACTIONS(83),
    [sym__composed_name] = ACTIONS(879),
    [sym_time] = ACTIONS(877),
    [sym_frequency] = ACTIONS(877),
    [sym_percentage] = ACTIONS(879),
    [sym_file_handle] = ACTIONS(879),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(881),
    [sym__decimal_number_pre_hook] = ACTIONS(87),
    [sym_path] = ACTIONS(879),
  },
  [STATE(137)] = {
    [sym__expression] = STATE(1399),
    [sym__parenthesized_expression] = STATE(1399),
    [sym__path_expression] = STATE(1399),
    [sym_unary_expression] = STATE(1399),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1399),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1399),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1399),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1399),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1399),
    [sym_string] = STATE(1399),
    [sym_character] = STATE(1399),
    [sym_symbol] = STATE(1399),
    [sym__special_command_operator] = STATE(1399),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(100),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(883),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(885),
    [sym_time] = ACTIONS(883),
    [sym_frequency] = ACTIONS(883),
    [sym_percentage] = ACTIONS(885),
    [sym_file_handle] = ACTIONS(885),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(887),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(138)] = {
    [sym__expression] = STATE(17),
    [sym__parenthesized_expression] = STATE(17),
    [sym__path_expression] = STATE(17),
    [sym_unary_expression] = STATE(17),
    [sym__unary_expression] = STATE(581),
    [sym__prefix_postfix_expression] = STATE(564),
    [sym_binary_expression] = STATE(17),
    [sym__binary_expression] = STATE(565),
    [sym__and_expression] = STATE(573),
    [sym_assignment_expression] = STATE(17),
    [sym_macro] = STATE(406),
    [sym_call_expression] = STATE(424),
    [sym__function_identifier] = STATE(2589),
    [sym__hll_var_call_expression_function] = STATE(2590),
    [sym__hll_var_call_expression] = STATE(428),
    [sym_trace32_hll_variable] = STATE(575),
    [sym__literal] = STATE(17),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(211),
    [sym_address] = STATE(17),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(411),
    [sym_bitmask] = STATE(17),
    [sym_string] = STATE(17),
    [sym_character] = STATE(17),
    [sym_symbol] = STATE(17),
    [sym__special_command_operator] = STATE(17),
    [sym__file_dialog] = STATE(569),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(443),
    [aux_sym__path_expression_repeat2] = STATE(17),
    [sym_identifier] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(57),
    [anon_sym_BANG] = ACTIONS(57),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(715),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(67),
    [anon_sym_SQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_access_class] = ACTIONS(73),
    [aux_sym_integer_token1] = ACTIONS(75),
    [aux_sym_integer_token2] = ACTIONS(75),
    [sym_float] = ACTIONS(889),
    [aux_sym_bitmask_token1] = ACTIONS(79),
    [aux_sym_bitmask_token2] = ACTIONS(79),
    [aux_sym_symbol_token1] = ACTIONS(81),
    [aux_sym_symbol_token2] = ACTIONS(83),
    [sym__composed_name] = ACTIONS(891),
    [sym_time] = ACTIONS(889),
    [sym_frequency] = ACTIONS(889),
    [sym_percentage] = ACTIONS(891),
    [sym_file_handle] = ACTIONS(891),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(87),
    [sym_path] = ACTIONS(891),
  },
  [STATE(139)] = {
    [sym__expression] = STATE(463),
    [sym__parenthesized_expression] = STATE(463),
    [sym__path_expression] = STATE(463),
    [sym_unary_expression] = STATE(463),
    [sym__unary_expression] = STATE(581),
    [sym__prefix_postfix_expression] = STATE(564),
    [sym_binary_expression] = STATE(463),
    [sym__binary_expression] = STATE(565),
    [sym__and_expression] = STATE(573),
    [sym_assignment_expression] = STATE(463),
    [sym_macro] = STATE(406),
    [sym_call_expression] = STATE(424),
    [sym__function_identifier] = STATE(2589),
    [sym__hll_var_call_expression_function] = STATE(2590),
    [sym__hll_var_call_expression] = STATE(428),
    [sym_trace32_hll_variable] = STATE(575),
    [sym__literal] = STATE(463),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(211),
    [sym_address] = STATE(463),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(411),
    [sym_bitmask] = STATE(463),
    [sym_string] = STATE(463),
    [sym_character] = STATE(463),
    [sym_symbol] = STATE(463),
    [sym__special_command_operator] = STATE(463),
    [sym__file_dialog] = STATE(569),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(443),
    [aux_sym__path_expression_repeat2] = STATE(463),
    [sym_identifier] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(57),
    [anon_sym_BANG] = ACTIONS(57),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(715),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(67),
    [anon_sym_SQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_access_class] = ACTIONS(73),
    [aux_sym_integer_token1] = ACTIONS(75),
    [aux_sym_integer_token2] = ACTIONS(75),
    [sym_float] = ACTIONS(893),
    [aux_sym_bitmask_token1] = ACTIONS(79),
    [aux_sym_bitmask_token2] = ACTIONS(79),
    [aux_sym_symbol_token1] = ACTIONS(81),
    [aux_sym_symbol_token2] = ACTIONS(83),
    [sym__composed_name] = ACTIONS(895),
    [sym_time] = ACTIONS(893),
    [sym_frequency] = ACTIONS(893),
    [sym_percentage] = ACTIONS(895),
    [sym_file_handle] = ACTIONS(895),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(87),
    [sym_path] = ACTIONS(895),
  },
  [STATE(140)] = {
    [sym__expression] = STATE(464),
    [sym__parenthesized_expression] = STATE(464),
    [sym__path_expression] = STATE(464),
    [sym_unary_expression] = STATE(464),
    [sym__unary_expression] = STATE(581),
    [sym__prefix_postfix_expression] = STATE(564),
    [sym_binary_expression] = STATE(464),
    [sym__binary_expression] = STATE(565),
    [sym__and_expression] = STATE(573),
    [sym_assignment_expression] = STATE(464),
    [sym_macro] = STATE(406),
    [sym_call_expression] = STATE(424),
    [sym__function_identifier] = STATE(2589),
    [sym__hll_var_call_expression_function] = STATE(2590),
    [sym__hll_var_call_expression] = STATE(428),
    [sym_trace32_hll_variable] = STATE(575),
    [sym__literal] = STATE(464),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(211),
    [sym_address] = STATE(464),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(411),
    [sym_bitmask] = STATE(464),
    [sym_string] = STATE(464),
    [sym_character] = STATE(464),
    [sym_symbol] = STATE(464),
    [sym__special_command_operator] = STATE(464),
    [sym__file_dialog] = STATE(569),
    [aux_sym_block_repeat1] = STATE(142),
    [aux_sym__path_expression_repeat1] = STATE(443),
    [aux_sym__path_expression_repeat2] = STATE(464),
    [sym_identifier] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(57),
    [anon_sym_BANG] = ACTIONS(57),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(715),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(67),
    [anon_sym_SQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_access_class] = ACTIONS(73),
    [aux_sym_integer_token1] = ACTIONS(75),
    [aux_sym_integer_token2] = ACTIONS(75),
    [sym_float] = ACTIONS(897),
    [aux_sym_bitmask_token1] = ACTIONS(79),
    [aux_sym_bitmask_token2] = ACTIONS(79),
    [aux_sym_symbol_token1] = ACTIONS(81),
    [aux_sym_symbol_token2] = ACTIONS(83),
    [sym__composed_name] = ACTIONS(899),
    [sym_time] = ACTIONS(897),
    [sym_frequency] = ACTIONS(897),
    [sym_percentage] = ACTIONS(899),
    [sym_file_handle] = ACTIONS(899),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(901),
    [sym__decimal_number_pre_hook] = ACTIONS(87),
    [sym_path] = ACTIONS(899),
  },
  [STATE(141)] = {
    [sym__expression] = STATE(1493),
    [sym__parenthesized_expression] = STATE(1493),
    [sym__path_expression] = STATE(1493),
    [sym_unary_expression] = STATE(1493),
    [sym__unary_expression] = STATE(1472),
    [sym__prefix_postfix_expression] = STATE(1473),
    [sym_binary_expression] = STATE(1493),
    [sym__binary_expression] = STATE(1474),
    [sym__and_expression] = STATE(1475),
    [sym_assignment_expression] = STATE(1493),
    [sym_macro] = STATE(1329),
    [sym_call_expression] = STATE(1115),
    [sym__function_identifier] = STATE(2483),
    [sym__hll_var_call_expression_function] = STATE(2485),
    [sym__hll_var_call_expression] = STATE(1116),
    [sym_trace32_hll_variable] = STATE(1476),
    [sym__literal] = STATE(1598),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(261),
    [sym_address] = STATE(1598),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1118),
    [sym_bitmask] = STATE(1598),
    [sym_string] = STATE(1598),
    [sym_character] = STATE(1598),
    [sym_symbol] = STATE(1598),
    [sym__special_command_operator] = STATE(1598),
    [sym__file_dialog] = STATE(1482),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1227),
    [aux_sym__path_expression_repeat2] = STATE(1119),
    [sym_identifier] = ACTIONS(903),
    [anon_sym_LPAREN] = ACTIONS(142),
    [anon_sym_LBRACE] = ACTIONS(144),
    [anon_sym_PLUS] = ACTIONS(106),
    [anon_sym_DASH] = ACTIONS(106),
    [anon_sym_TILDE] = ACTIONS(104),
    [anon_sym_BANG] = ACTIONS(104),
    [anon_sym_PLUS_PLUS] = ACTIONS(222),
    [anon_sym_DASH_DASH] = ACTIONS(222),
    [anon_sym_STAR] = ACTIONS(152),
    [anon_sym_AMP] = ACTIONS(154),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(120),
    [anon_sym_SQUOTE] = ACTIONS(122),
    [anon_sym_DQUOTE] = ACTIONS(124),
    [sym_access_class] = ACTIONS(126),
    [aux_sym_integer_token1] = ACTIONS(156),
    [aux_sym_integer_token2] = ACTIONS(156),
    [sym_float] = ACTIONS(905),
    [aux_sym_bitmask_token1] = ACTIONS(130),
    [aux_sym_bitmask_token2] = ACTIONS(130),
    [aux_sym_symbol_token1] = ACTIONS(132),
    [aux_sym_symbol_token2] = ACTIONS(134),
    [sym__composed_name] = ACTIONS(907),
    [sym_time] = ACTIONS(905),
    [sym_frequency] = ACTIONS(905),
    [sym_percentage] = ACTIONS(907),
    [sym_file_handle] = ACTIONS(907),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(162),
    [sym_path] = ACTIONS(138),
  },
  [STATE(142)] = {
    [sym__expression] = STATE(467),
    [sym__parenthesized_expression] = STATE(467),
    [sym__path_expression] = STATE(467),
    [sym_unary_expression] = STATE(467),
    [sym__unary_expression] = STATE(581),
    [sym__prefix_postfix_expression] = STATE(564),
    [sym_binary_expression] = STATE(467),
    [sym__binary_expression] = STATE(565),
    [sym__and_expression] = STATE(573),
    [sym_assignment_expression] = STATE(467),
    [sym_macro] = STATE(406),
    [sym_call_expression] = STATE(424),
    [sym__function_identifier] = STATE(2589),
    [sym__hll_var_call_expression_function] = STATE(2590),
    [sym__hll_var_call_expression] = STATE(428),
    [sym_trace32_hll_variable] = STATE(575),
    [sym__literal] = STATE(467),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(211),
    [sym_address] = STATE(467),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(411),
    [sym_bitmask] = STATE(467),
    [sym_string] = STATE(467),
    [sym_character] = STATE(467),
    [sym_symbol] = STATE(467),
    [sym__special_command_operator] = STATE(467),
    [sym__file_dialog] = STATE(569),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(443),
    [aux_sym__path_expression_repeat2] = STATE(467),
    [sym_identifier] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(57),
    [anon_sym_BANG] = ACTIONS(57),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(715),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(67),
    [anon_sym_SQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_access_class] = ACTIONS(73),
    [aux_sym_integer_token1] = ACTIONS(75),
    [aux_sym_integer_token2] = ACTIONS(75),
    [sym_float] = ACTIONS(909),
    [aux_sym_bitmask_token1] = ACTIONS(79),
    [aux_sym_bitmask_token2] = ACTIONS(79),
    [aux_sym_symbol_token1] = ACTIONS(81),
    [aux_sym_symbol_token2] = ACTIONS(83),
    [sym__composed_name] = ACTIONS(911),
    [sym_time] = ACTIONS(909),
    [sym_frequency] = ACTIONS(909),
    [sym_percentage] = ACTIONS(911),
    [sym_file_handle] = ACTIONS(911),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(87),
    [sym_path] = ACTIONS(911),
  },
  [STATE(143)] = {
    [sym__expression] = STATE(1442),
    [sym__parenthesized_expression] = STATE(1442),
    [sym__path_expression] = STATE(1442),
    [sym_unary_expression] = STATE(1442),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1442),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1442),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1442),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1442),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1442),
    [sym_string] = STATE(1442),
    [sym_character] = STATE(1442),
    [sym_symbol] = STATE(1442),
    [sym__special_command_operator] = STATE(1442),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(913),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(915),
    [sym_time] = ACTIONS(913),
    [sym_frequency] = ACTIONS(913),
    [sym_percentage] = ACTIONS(915),
    [sym_file_handle] = ACTIONS(915),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(144)] = {
    [sym__expression] = STATE(1486),
    [sym__parenthesized_expression] = STATE(1486),
    [sym__path_expression] = STATE(1486),
    [sym_unary_expression] = STATE(1486),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1486),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1486),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1486),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1486),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1486),
    [sym_string] = STATE(1486),
    [sym_character] = STATE(1486),
    [sym_symbol] = STATE(1486),
    [sym__special_command_operator] = STATE(1486),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(917),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(919),
    [sym_time] = ACTIONS(917),
    [sym_frequency] = ACTIONS(917),
    [sym_percentage] = ACTIONS(919),
    [sym_file_handle] = ACTIONS(919),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(145)] = {
    [sym__expression] = STATE(1456),
    [sym__parenthesized_expression] = STATE(1456),
    [sym__path_expression] = STATE(1456),
    [sym_unary_expression] = STATE(1456),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1456),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1456),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1456),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1456),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1456),
    [sym_string] = STATE(1456),
    [sym_character] = STATE(1456),
    [sym_symbol] = STATE(1456),
    [sym__special_command_operator] = STATE(1456),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(921),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(923),
    [sym_time] = ACTIONS(921),
    [sym_frequency] = ACTIONS(921),
    [sym_percentage] = ACTIONS(923),
    [sym_file_handle] = ACTIONS(923),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(146)] = {
    [sym__expression] = STATE(18),
    [sym__parenthesized_expression] = STATE(18),
    [sym__path_expression] = STATE(18),
    [sym_unary_expression] = STATE(18),
    [sym__unary_expression] = STATE(581),
    [sym__prefix_postfix_expression] = STATE(564),
    [sym_binary_expression] = STATE(18),
    [sym__binary_expression] = STATE(565),
    [sym__and_expression] = STATE(573),
    [sym_assignment_expression] = STATE(18),
    [sym_macro] = STATE(406),
    [sym_call_expression] = STATE(424),
    [sym__function_identifier] = STATE(2589),
    [sym__hll_var_call_expression_function] = STATE(2590),
    [sym__hll_var_call_expression] = STATE(428),
    [sym_trace32_hll_variable] = STATE(575),
    [sym__literal] = STATE(18),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(211),
    [sym_address] = STATE(18),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(411),
    [sym_bitmask] = STATE(18),
    [sym_string] = STATE(18),
    [sym_character] = STATE(18),
    [sym_symbol] = STATE(18),
    [sym__special_command_operator] = STATE(18),
    [sym__file_dialog] = STATE(569),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(443),
    [aux_sym__path_expression_repeat2] = STATE(18),
    [sym_identifier] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(57),
    [anon_sym_BANG] = ACTIONS(57),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(715),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(67),
    [anon_sym_SQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_access_class] = ACTIONS(73),
    [aux_sym_integer_token1] = ACTIONS(75),
    [aux_sym_integer_token2] = ACTIONS(75),
    [sym_float] = ACTIONS(925),
    [aux_sym_bitmask_token1] = ACTIONS(79),
    [aux_sym_bitmask_token2] = ACTIONS(79),
    [aux_sym_symbol_token1] = ACTIONS(81),
    [aux_sym_symbol_token2] = ACTIONS(83),
    [sym__composed_name] = ACTIONS(927),
    [sym_time] = ACTIONS(925),
    [sym_frequency] = ACTIONS(925),
    [sym_percentage] = ACTIONS(927),
    [sym_file_handle] = ACTIONS(927),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(87),
    [sym_path] = ACTIONS(927),
  },
  [STATE(147)] = {
    [sym__expression] = STATE(1461),
    [sym__parenthesized_expression] = STATE(1461),
    [sym__path_expression] = STATE(1461),
    [sym_unary_expression] = STATE(1461),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1461),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1461),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1461),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1461),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1461),
    [sym_string] = STATE(1461),
    [sym_character] = STATE(1461),
    [sym_symbol] = STATE(1461),
    [sym__special_command_operator] = STATE(1461),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(929),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(931),
    [sym_time] = ACTIONS(929),
    [sym_frequency] = ACTIONS(929),
    [sym_percentage] = ACTIONS(931),
    [sym_file_handle] = ACTIONS(931),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(148)] = {
    [sym__expression] = STATE(1462),
    [sym__parenthesized_expression] = STATE(1462),
    [sym__path_expression] = STATE(1462),
    [sym_unary_expression] = STATE(1462),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1462),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1462),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1462),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1462),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1462),
    [sym_string] = STATE(1462),
    [sym_character] = STATE(1462),
    [sym_symbol] = STATE(1462),
    [sym__special_command_operator] = STATE(1462),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(933),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(935),
    [sym_time] = ACTIONS(933),
    [sym_frequency] = ACTIONS(933),
    [sym_percentage] = ACTIONS(935),
    [sym_file_handle] = ACTIONS(935),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(149)] = {
    [sym__expression] = STATE(1465),
    [sym__parenthesized_expression] = STATE(1465),
    [sym__path_expression] = STATE(1465),
    [sym_unary_expression] = STATE(1465),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1465),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1465),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1465),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1465),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1465),
    [sym_string] = STATE(1465),
    [sym_character] = STATE(1465),
    [sym_symbol] = STATE(1465),
    [sym__special_command_operator] = STATE(1465),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(937),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(939),
    [sym_time] = ACTIONS(937),
    [sym_frequency] = ACTIONS(937),
    [sym_percentage] = ACTIONS(939),
    [sym_file_handle] = ACTIONS(939),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(150)] = {
    [sym__expression] = STATE(1467),
    [sym__parenthesized_expression] = STATE(1467),
    [sym__path_expression] = STATE(1467),
    [sym_unary_expression] = STATE(1467),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1467),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1467),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1467),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1467),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1467),
    [sym_string] = STATE(1467),
    [sym_character] = STATE(1467),
    [sym_symbol] = STATE(1467),
    [sym__special_command_operator] = STATE(1467),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(941),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(943),
    [sym_time] = ACTIONS(941),
    [sym_frequency] = ACTIONS(941),
    [sym_percentage] = ACTIONS(943),
    [sym_file_handle] = ACTIONS(943),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(151)] = {
    [sym__expression] = STATE(1478),
    [sym__parenthesized_expression] = STATE(1478),
    [sym__path_expression] = STATE(1478),
    [sym_unary_expression] = STATE(1478),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1478),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1478),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1478),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1478),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1478),
    [sym_string] = STATE(1478),
    [sym_character] = STATE(1478),
    [sym_symbol] = STATE(1478),
    [sym__special_command_operator] = STATE(1478),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(945),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(947),
    [sym_time] = ACTIONS(945),
    [sym_frequency] = ACTIONS(945),
    [sym_percentage] = ACTIONS(947),
    [sym_file_handle] = ACTIONS(947),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(152)] = {
    [sym__expression] = STATE(1477),
    [sym__parenthesized_expression] = STATE(1477),
    [sym__path_expression] = STATE(1477),
    [sym_unary_expression] = STATE(1477),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1477),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1477),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1477),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1477),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1477),
    [sym_string] = STATE(1477),
    [sym_character] = STATE(1477),
    [sym_symbol] = STATE(1477),
    [sym__special_command_operator] = STATE(1477),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(949),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(951),
    [sym_time] = ACTIONS(949),
    [sym_frequency] = ACTIONS(949),
    [sym_percentage] = ACTIONS(951),
    [sym_file_handle] = ACTIONS(951),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(153)] = {
    [sym__expression] = STATE(1469),
    [sym__parenthesized_expression] = STATE(1469),
    [sym__path_expression] = STATE(1469),
    [sym_unary_expression] = STATE(1469),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1469),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1469),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1469),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1469),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1469),
    [sym_string] = STATE(1469),
    [sym_character] = STATE(1469),
    [sym_symbol] = STATE(1469),
    [sym__special_command_operator] = STATE(1469),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(953),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(955),
    [sym_time] = ACTIONS(953),
    [sym_frequency] = ACTIONS(953),
    [sym_percentage] = ACTIONS(955),
    [sym_file_handle] = ACTIONS(955),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(154)] = {
    [sym__expression] = STATE(1480),
    [sym__parenthesized_expression] = STATE(1480),
    [sym__path_expression] = STATE(1480),
    [sym_unary_expression] = STATE(1480),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1480),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1480),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1480),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1480),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1480),
    [sym_string] = STATE(1480),
    [sym_character] = STATE(1480),
    [sym_symbol] = STATE(1480),
    [sym__special_command_operator] = STATE(1480),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(957),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(959),
    [sym_time] = ACTIONS(957),
    [sym_frequency] = ACTIONS(957),
    [sym_percentage] = ACTIONS(959),
    [sym_file_handle] = ACTIONS(959),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(155)] = {
    [sym__expression] = STATE(14),
    [sym__parenthesized_expression] = STATE(14),
    [sym__path_expression] = STATE(14),
    [sym_unary_expression] = STATE(14),
    [sym__unary_expression] = STATE(581),
    [sym__prefix_postfix_expression] = STATE(564),
    [sym_binary_expression] = STATE(14),
    [sym__binary_expression] = STATE(565),
    [sym__and_expression] = STATE(573),
    [sym_assignment_expression] = STATE(14),
    [sym_macro] = STATE(406),
    [sym_call_expression] = STATE(424),
    [sym__function_identifier] = STATE(2589),
    [sym__hll_var_call_expression_function] = STATE(2590),
    [sym__hll_var_call_expression] = STATE(428),
    [sym_trace32_hll_variable] = STATE(575),
    [sym__literal] = STATE(14),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(211),
    [sym_address] = STATE(14),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(411),
    [sym_bitmask] = STATE(14),
    [sym_string] = STATE(14),
    [sym_character] = STATE(14),
    [sym_symbol] = STATE(14),
    [sym__special_command_operator] = STATE(14),
    [sym__file_dialog] = STATE(569),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(443),
    [aux_sym__path_expression_repeat2] = STATE(14),
    [sym_identifier] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(57),
    [anon_sym_BANG] = ACTIONS(57),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(715),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(67),
    [anon_sym_SQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_access_class] = ACTIONS(73),
    [aux_sym_integer_token1] = ACTIONS(75),
    [aux_sym_integer_token2] = ACTIONS(75),
    [sym_float] = ACTIONS(961),
    [aux_sym_bitmask_token1] = ACTIONS(79),
    [aux_sym_bitmask_token2] = ACTIONS(79),
    [aux_sym_symbol_token1] = ACTIONS(81),
    [aux_sym_symbol_token2] = ACTIONS(83),
    [sym__composed_name] = ACTIONS(963),
    [sym_time] = ACTIONS(961),
    [sym_frequency] = ACTIONS(961),
    [sym_percentage] = ACTIONS(963),
    [sym_file_handle] = ACTIONS(963),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(87),
    [sym_path] = ACTIONS(963),
  },
  [STATE(156)] = {
    [sym__expression] = STATE(1319),
    [sym__parenthesized_expression] = STATE(1319),
    [sym__path_expression] = STATE(1319),
    [sym_unary_expression] = STATE(1319),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1319),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1319),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1319),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1319),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1319),
    [sym_string] = STATE(1319),
    [sym_character] = STATE(1319),
    [sym_symbol] = STATE(1319),
    [sym__special_command_operator] = STATE(1319),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(106),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(965),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(967),
    [sym_time] = ACTIONS(965),
    [sym_frequency] = ACTIONS(965),
    [sym_percentage] = ACTIONS(967),
    [sym_file_handle] = ACTIONS(967),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(969),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(157)] = {
    [sym__expression] = STATE(15),
    [sym__parenthesized_expression] = STATE(15),
    [sym__path_expression] = STATE(15),
    [sym_unary_expression] = STATE(15),
    [sym__unary_expression] = STATE(581),
    [sym__prefix_postfix_expression] = STATE(564),
    [sym_binary_expression] = STATE(15),
    [sym__binary_expression] = STATE(565),
    [sym__and_expression] = STATE(573),
    [sym_assignment_expression] = STATE(15),
    [sym_macro] = STATE(406),
    [sym_call_expression] = STATE(424),
    [sym__function_identifier] = STATE(2589),
    [sym__hll_var_call_expression_function] = STATE(2590),
    [sym__hll_var_call_expression] = STATE(428),
    [sym_trace32_hll_variable] = STATE(575),
    [sym__literal] = STATE(15),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(211),
    [sym_address] = STATE(15),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(411),
    [sym_bitmask] = STATE(15),
    [sym_string] = STATE(15),
    [sym_character] = STATE(15),
    [sym_symbol] = STATE(15),
    [sym__special_command_operator] = STATE(15),
    [sym__file_dialog] = STATE(569),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(443),
    [aux_sym__path_expression_repeat2] = STATE(15),
    [sym_identifier] = ACTIONS(45),
    [anon_sym_LPAREN] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_PLUS] = ACTIONS(55),
    [anon_sym_DASH] = ACTIONS(55),
    [anon_sym_TILDE] = ACTIONS(57),
    [anon_sym_BANG] = ACTIONS(57),
    [anon_sym_PLUS_PLUS] = ACTIONS(59),
    [anon_sym_DASH_DASH] = ACTIONS(59),
    [anon_sym_STAR] = ACTIONS(61),
    [anon_sym_AMP] = ACTIONS(715),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(67),
    [anon_sym_SQUOTE] = ACTIONS(69),
    [anon_sym_DQUOTE] = ACTIONS(71),
    [sym_access_class] = ACTIONS(73),
    [aux_sym_integer_token1] = ACTIONS(75),
    [aux_sym_integer_token2] = ACTIONS(75),
    [sym_float] = ACTIONS(971),
    [aux_sym_bitmask_token1] = ACTIONS(79),
    [aux_sym_bitmask_token2] = ACTIONS(79),
    [aux_sym_symbol_token1] = ACTIONS(81),
    [aux_sym_symbol_token2] = ACTIONS(83),
    [sym__composed_name] = ACTIONS(973),
    [sym_time] = ACTIONS(971),
    [sym_frequency] = ACTIONS(971),
    [sym_percentage] = ACTIONS(973),
    [sym_file_handle] = ACTIONS(973),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(87),
    [sym_path] = ACTIONS(973),
  },
  [STATE(158)] = {
    [sym__expression] = STATE(1487),
    [sym__parenthesized_expression] = STATE(1487),
    [sym__path_expression] = STATE(1487),
    [sym_unary_expression] = STATE(1487),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1487),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1487),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1487),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1487),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1487),
    [sym_string] = STATE(1487),
    [sym_character] = STATE(1487),
    [sym_symbol] = STATE(1487),
    [sym__special_command_operator] = STATE(1487),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(975),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(977),
    [sym_time] = ACTIONS(975),
    [sym_frequency] = ACTIONS(975),
    [sym_percentage] = ACTIONS(977),
    [sym_file_handle] = ACTIONS(977),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(159)] = {
    [sym__expression] = STATE(1357),
    [sym__parenthesized_expression] = STATE(1357),
    [sym__path_expression] = STATE(1357),
    [sym_unary_expression] = STATE(1357),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1357),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1357),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1357),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1357),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1357),
    [sym_string] = STATE(1357),
    [sym_character] = STATE(1357),
    [sym_symbol] = STATE(1357),
    [sym__special_command_operator] = STATE(1357),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(979),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(981),
    [sym_time] = ACTIONS(979),
    [sym_frequency] = ACTIONS(979),
    [sym_percentage] = ACTIONS(981),
    [sym_file_handle] = ACTIONS(981),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
  [STATE(160)] = {
    [sym__expression] = STATE(1408),
    [sym__parenthesized_expression] = STATE(1408),
    [sym__path_expression] = STATE(1408),
    [sym_unary_expression] = STATE(1408),
    [sym__unary_expression] = STATE(1513),
    [sym__prefix_postfix_expression] = STATE(1514),
    [sym_binary_expression] = STATE(1408),
    [sym__binary_expression] = STATE(1515),
    [sym__and_expression] = STATE(1516),
    [sym_assignment_expression] = STATE(1408),
    [sym_macro] = STATE(1097),
    [sym_call_expression] = STATE(1187),
    [sym__function_identifier] = STATE(2554),
    [sym__hll_var_call_expression_function] = STATE(2531),
    [sym__hll_var_call_expression] = STATE(1188),
    [sym_trace32_hll_variable] = STATE(1490),
    [sym__literal] = STATE(1408),
    [sym__address_expression] = STATE(1678),
    [sym_memory_space] = STATE(207),
    [sym_address] = STATE(1408),
    [sym__parenthesized_address_expression] = STATE(1678),
    [sym__address_binary_expression] = STATE(1636),
    [sym_integer] = STATE(1224),
    [sym_bitmask] = STATE(1408),
    [sym_string] = STATE(1408),
    [sym_character] = STATE(1408),
    [sym_symbol] = STATE(1408),
    [sym__special_command_operator] = STATE(1408),
    [sym__file_dialog] = STATE(1496),
    [aux_sym_block_repeat1] = STATE(1002),
    [aux_sym__path_expression_repeat1] = STATE(1347),
    [aux_sym__path_expression_repeat2] = STATE(1238),
    [sym_identifier] = ACTIONS(173),
    [anon_sym_LPAREN] = ACTIONS(175),
    [anon_sym_LBRACE] = ACTIONS(177),
    [anon_sym_PLUS] = ACTIONS(179),
    [anon_sym_DASH] = ACTIONS(179),
    [anon_sym_TILDE] = ACTIONS(181),
    [anon_sym_BANG] = ACTIONS(181),
    [anon_sym_PLUS_PLUS] = ACTIONS(183),
    [anon_sym_DASH_DASH] = ACTIONS(183),
    [anon_sym_STAR] = ACTIONS(185),
    [anon_sym_AMP] = ACTIONS(187),
    [aux_sym__hll_non_default_format_var_command_token1] = ACTIONS(65),
    [anon_sym_BSLASH] = ACTIONS(189),
    [anon_sym_SQUOTE] = ACTIONS(191),
    [anon_sym_DQUOTE] = ACTIONS(193),
    [sym_access_class] = ACTIONS(195),
    [aux_sym_integer_token1] = ACTIONS(197),
    [aux_sym_integer_token2] = ACTIONS(197),
    [sym_float] = ACTIONS(983),
    [aux_sym_bitmask_token1] = ACTIONS(201),
    [aux_sym_bitmask_token2] = ACTIONS(201),
    [aux_sym_symbol_token1] = ACTIONS(203),
    [aux_sym_symbol_token2] = ACTIONS(205),
    [sym__composed_name] = ACTIONS(985),
    [sym_time] = ACTIONS(983),
    [sym_frequency] = ACTIONS(983),
    [sym_percentage] = ACTIONS(985),
    [sym_file_handle] = ACTIONS(985),
    [sym_comment] = ACTIONS(3),
    [sym__line_continuation] = ACTIONS(3),
    [sym__blank] = ACTIONS(470),
    [sym__decimal_number_pre_hook] = ACTIONS(209),
    [sym_path] = ACTIONS(211),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 41,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(987), 1,
      anon_sym_RPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(989), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(991), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1253), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [147] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(440), 1,
      anon_sym_AMP,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(999), 1,
      anon_sym_TILDE,
    ACTIONS(1003), 1,
      anon_sym_STAR,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(997), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
    ACTIONS(234), 4,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1359), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
    ACTIONS(230), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      sym_comment,
      sym__blank_line,
  [254] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(440), 1,
      anon_sym_AMP,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(999), 1,
      anon_sym_TILDE,
    ACTIONS(1003), 1,
      anon_sym_STAR,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(997), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
    ACTIONS(260), 4,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1372), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
    ACTIONS(258), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      sym_comment,
      sym__blank_line,
  [361] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(440), 1,
      anon_sym_AMP,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(999), 1,
      anon_sym_TILDE,
    ACTIONS(1003), 1,
      anon_sym_STAR,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(997), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
    ACTIONS(264), 4,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1376), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
    ACTIONS(262), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      sym_comment,
      sym__blank_line,
  [468] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(440), 1,
      anon_sym_AMP,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(999), 1,
      anon_sym_TILDE,
    ACTIONS(1003), 1,
      anon_sym_STAR,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(997), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
    ACTIONS(268), 4,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1385), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
    ACTIONS(266), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      sym_comment,
      sym__blank_line,
  [575] = 41,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(1009), 1,
      anon_sym_RPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1011), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1013), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1260), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [722] = 41,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(1015), 1,
      anon_sym_RPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1017), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1019), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1259), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [869] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1645), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1021), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1023), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1427), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [1013] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1025), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1027), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1420), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [1157] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1029), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1031), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1421), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [1301] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(224), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(226), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1460), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [1445] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1033), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1035), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1422), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [1589] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1037), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1039), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1423), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [1733] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1041), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1043), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1340), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [1877] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1045), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1047), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1325), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [2021] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1049), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1051), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1315), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [2165] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1053), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1055), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1305), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [2309] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1057), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1059), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1354), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [2453] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(95), 1,
      anon_sym_LBRACE,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(715), 1,
      anon_sym_AMP,
    ACTIONS(1061), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(408), 1,
      sym_macro,
    STATE(414), 1,
      sym__address_binary_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(430), 1,
      sym_call_expression,
    STATE(431), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(427), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(128), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(136), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1512), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [2597] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(95), 1,
      anon_sym_LBRACE,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(715), 1,
      anon_sym_AMP,
    ACTIONS(1061), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(408), 1,
      sym_macro,
    STATE(414), 1,
      sym__address_binary_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(430), 1,
      sym_call_expression,
    STATE(431), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(417), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(128), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(136), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1512), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [2741] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(95), 1,
      anon_sym_LBRACE,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(715), 1,
      anon_sym_AMP,
    ACTIONS(1061), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(408), 1,
      sym_macro,
    STATE(414), 1,
      sym__address_binary_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(430), 1,
      sym_call_expression,
    STATE(431), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(418), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(128), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(136), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1512), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [2885] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(95), 1,
      anon_sym_LBRACE,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(715), 1,
      anon_sym_AMP,
    ACTIONS(1061), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(408), 1,
      sym_macro,
    STATE(414), 1,
      sym__address_binary_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(430), 1,
      sym_call_expression,
    STATE(431), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(419), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(128), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(136), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1512), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [3029] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(95), 1,
      anon_sym_LBRACE,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(715), 1,
      anon_sym_AMP,
    ACTIONS(1061), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(408), 1,
      sym_macro,
    STATE(414), 1,
      sym__address_binary_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(430), 1,
      sym_call_expression,
    STATE(431), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(426), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(128), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(136), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1512), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [3173] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1063), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1065), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1301), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [3317] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1067), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1069), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1339), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [3461] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(167), 1,
      anon_sym_LBRACE,
    ACTIONS(1071), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1060), 1,
      sym_macro,
    STATE(1099), 1,
      sym__address_binary_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1133), 1,
      sym_call_expression,
    STATE(1134), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1145), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [3605] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(167), 1,
      anon_sym_LBRACE,
    ACTIONS(1071), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1060), 1,
      sym_macro,
    STATE(1099), 1,
      sym__address_binary_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1133), 1,
      sym_call_expression,
    STATE(1134), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1073), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [3749] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(167), 1,
      anon_sym_LBRACE,
    ACTIONS(1071), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1060), 1,
      sym_macro,
    STATE(1099), 1,
      sym__address_binary_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1133), 1,
      sym_call_expression,
    STATE(1134), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1147), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [3893] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(167), 1,
      anon_sym_LBRACE,
    ACTIONS(1071), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1060), 1,
      sym_macro,
    STATE(1099), 1,
      sym__address_binary_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1133), 1,
      sym_call_expression,
    STATE(1134), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1113), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [4037] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(167), 1,
      anon_sym_LBRACE,
    ACTIONS(1071), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1060), 1,
      sym_macro,
    STATE(1099), 1,
      sym__address_binary_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1133), 1,
      sym_call_expression,
    STATE(1134), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1111), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [4181] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1073), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1075), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1415), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [4325] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(199), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(207), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1337), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [4469] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1689), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [4613] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1626), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [4757] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1645), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1077), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1079), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1401), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [4901] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1658), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1081), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1083), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1402), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [5045] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1690), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [5189] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1692), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [5333] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1695), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [5477] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1085), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1087), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1416), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [5621] = 40,
    ACTIONS(1089), 1,
      sym_identifier,
    ACTIONS(1092), 1,
      anon_sym_LPAREN,
    ACTIONS(1095), 1,
      anon_sym_LBRACE,
    ACTIONS(1107), 1,
      anon_sym_STAR,
    ACTIONS(1110), 1,
      anon_sym_AMP,
    ACTIONS(1113), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(1116), 1,
      anon_sym_BSLASH,
    ACTIONS(1119), 1,
      anon_sym_SQUOTE,
    ACTIONS(1122), 1,
      anon_sym_DQUOTE,
    ACTIONS(1125), 1,
      sym_access_class,
    ACTIONS(1137), 1,
      aux_sym_symbol_token1,
    ACTIONS(1140), 1,
      aux_sym_symbol_token2,
    ACTIONS(1146), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(1149), 1,
      sym_path,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1098), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1101), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1104), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1128), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(1134), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1696), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1131), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1143), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [5765] = 40,
    ACTIONS(1152), 1,
      sym_identifier,
    ACTIONS(1155), 1,
      anon_sym_LPAREN,
    ACTIONS(1158), 1,
      anon_sym_LBRACE,
    ACTIONS(1170), 1,
      anon_sym_STAR,
    ACTIONS(1173), 1,
      anon_sym_AMP,
    ACTIONS(1176), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(1179), 1,
      anon_sym_BSLASH,
    ACTIONS(1182), 1,
      anon_sym_SQUOTE,
    ACTIONS(1185), 1,
      anon_sym_DQUOTE,
    ACTIONS(1188), 1,
      sym_access_class,
    ACTIONS(1200), 1,
      aux_sym_symbol_token1,
    ACTIONS(1203), 1,
      aux_sym_symbol_token2,
    ACTIONS(1209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(1212), 1,
      sym_path,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1161), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1164), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1167), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1191), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(1197), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1697), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1194), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1206), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [5909] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1215), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1217), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1338), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [6053] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1219), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1221), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(465), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [6193] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1223), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1225), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1232), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [6337] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1227), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1229), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(466), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [6477] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(216), 1,
      anon_sym_LBRACE,
    ACTIONS(1231), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1090), 1,
      sym_macro,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1163), 1,
      sym_integer,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1247), 1,
      sym_call_expression,
    STATE(1249), 1,
      sym__address_binary_expression,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1177), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(218), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(220), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1510), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [6621] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1233), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1235), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1155), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [6765] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1237), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1239), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1294), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [6909] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1241), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1243), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1304), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [7053] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(95), 1,
      anon_sym_LBRACE,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(715), 1,
      anon_sym_AMP,
    ACTIONS(1061), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(408), 1,
      sym_macro,
    STATE(414), 1,
      sym__address_binary_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(430), 1,
      sym_call_expression,
    STATE(431), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(434), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(128), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(136), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1512), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [7197] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(95), 1,
      anon_sym_LBRACE,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(715), 1,
      anon_sym_AMP,
    ACTIONS(1061), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(408), 1,
      sym_macro,
    STATE(414), 1,
      sym__address_binary_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(430), 1,
      sym_call_expression,
    STATE(431), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(425), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(128), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(136), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1512), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [7341] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1245), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1247), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1266), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [7485] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1249), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1251), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(451), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [7625] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1253), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1255), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1267), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [7769] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1257), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1259), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1268), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [7913] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1261), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1263), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1231), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [8057] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1658), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1265), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1267), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1428), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [8201] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1269), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1271), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(453), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [8341] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1273), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1275), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1414), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [8485] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1277), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1279), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(455), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [8625] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1281), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1283), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(456), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [8765] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1285), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1287), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(457), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [8905] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1289), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1291), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(458), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [9045] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1646), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1021), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1023), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1427), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [9189] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1656), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1265), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1267), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1428), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [9333] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1293), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1295), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(487), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [9473] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(77), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(85), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(452), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [9613] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1648), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1021), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1023), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1427), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [9757] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1655), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1265), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1267), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1428), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [9901] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(216), 1,
      anon_sym_LBRACE,
    ACTIONS(1231), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1090), 1,
      sym_macro,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1163), 1,
      sym_integer,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1247), 1,
      sym_call_expression,
    STATE(1249), 1,
      sym__address_binary_expression,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1278), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(218), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(220), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1510), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [10045] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(216), 1,
      anon_sym_LBRACE,
    ACTIONS(1231), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1090), 1,
      sym_macro,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1163), 1,
      sym_integer,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1247), 1,
      sym_call_expression,
    STATE(1249), 1,
      sym__address_binary_expression,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1279), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(218), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(220), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1510), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [10189] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1659), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1021), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1023), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1427), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [10333] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1644), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1265), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1267), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1428), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [10477] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(216), 1,
      anon_sym_LBRACE,
    ACTIONS(1231), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1090), 1,
      sym_macro,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1163), 1,
      sym_integer,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1247), 1,
      sym_call_expression,
    STATE(1249), 1,
      sym__address_binary_expression,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1280), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(218), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(220), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1510), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [10621] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(216), 1,
      anon_sym_LBRACE,
    ACTIONS(1231), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1090), 1,
      sym_macro,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1163), 1,
      sym_integer,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1247), 1,
      sym_call_expression,
    STATE(1249), 1,
      sym__address_binary_expression,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1281), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(218), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(220), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1510), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [10765] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(216), 1,
      anon_sym_LBRACE,
    ACTIONS(1231), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1090), 1,
      sym_macro,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1163), 1,
      sym_integer,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1247), 1,
      sym_call_expression,
    STATE(1249), 1,
      sym__address_binary_expression,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1156), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(218), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(220), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1510), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [10909] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1297), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1299), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(461), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [11049] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1301), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1303), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(462), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [11189] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1305), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1307), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1269), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [11333] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1309), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1311), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1310), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [11477] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1313), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1315), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1270), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [11621] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1317), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1319), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1271), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [11765] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1658), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1321), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1323), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1371), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [11909] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1325), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1327), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1272), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [12053] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(158), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(160), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1265), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [12197] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1645), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1329), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1331), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1370), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [12341] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(216), 1,
      anon_sym_LBRACE,
    ACTIONS(1231), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1090), 1,
      sym_macro,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1163), 1,
      sym_integer,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1247), 1,
      sym_call_expression,
    STATE(1249), 1,
      sym__address_binary_expression,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1236), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(218), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(220), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1510), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [12485] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1333), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1335), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1254), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [12629] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1030), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1126), 1,
      sym_call_expression,
    STATE(1127), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(146), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(148), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(150), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1337), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1339), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1257), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [12773] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(173), 1,
      sym_identifier,
    ACTIONS(175), 1,
      anon_sym_LPAREN,
    ACTIONS(177), 1,
      anon_sym_LBRACE,
    ACTIONS(185), 1,
      anon_sym_STAR,
    ACTIONS(187), 1,
      anon_sym_AMP,
    ACTIONS(189), 1,
      anon_sym_BSLASH,
    ACTIONS(191), 1,
      anon_sym_SQUOTE,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(195), 1,
      sym_access_class,
    ACTIONS(203), 1,
      aux_sym_symbol_token1,
    ACTIONS(205), 1,
      aux_sym_symbol_token2,
    ACTIONS(209), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(211), 1,
      sym_path,
    STATE(207), 1,
      sym_memory_space,
    STATE(1097), 1,
      sym_macro,
    STATE(1187), 1,
      sym_call_expression,
    STATE(1188), 1,
      sym__hll_var_call_expression,
    STATE(1224), 1,
      sym_integer,
    STATE(1238), 1,
      aux_sym__path_expression_repeat2,
    STATE(1347), 1,
      aux_sym__path_expression_repeat1,
    STATE(1490), 1,
      sym_trace32_hll_variable,
    STATE(1496), 1,
      sym__file_dialog,
    STATE(1513), 1,
      sym__unary_expression,
    STATE(1514), 1,
      sym__prefix_postfix_expression,
    STATE(1515), 1,
      sym__binary_expression,
    STATE(1516), 1,
      sym__and_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2531), 1,
      sym__hll_var_call_expression_function,
    STATE(2554), 1,
      sym__function_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(179), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(181), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(183), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(197), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(201), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1341), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1343), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1326), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [12917] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1345), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1347), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1417), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [13061] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1105), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1165), 1,
      sym_call_expression,
    STATE(1167), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1625), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [13205] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1105), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1165), 1,
      sym_call_expression,
    STATE(1167), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1626), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [13349] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1105), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1165), 1,
      sym_call_expression,
    STATE(1167), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1629), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [13493] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1105), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1165), 1,
      sym_call_expression,
    STATE(1167), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1630), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [13637] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1105), 1,
      sym_macro,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1165), 1,
      sym_call_expression,
    STATE(1167), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1631), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [13781] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1349), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1351), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1418), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [13925] = 40,
    ACTIONS(1353), 1,
      sym_identifier,
    ACTIONS(1357), 1,
      anon_sym_LPAREN,
    ACTIONS(1361), 1,
      anon_sym_LBRACE,
    ACTIONS(1377), 1,
      anon_sym_STAR,
    ACTIONS(1381), 1,
      anon_sym_AMP,
    ACTIONS(1385), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(1389), 1,
      anon_sym_BSLASH,
    ACTIONS(1393), 1,
      anon_sym_SQUOTE,
    ACTIONS(1397), 1,
      anon_sym_DQUOTE,
    ACTIONS(1401), 1,
      sym_access_class,
    ACTIONS(1417), 1,
      aux_sym_symbol_token1,
    ACTIONS(1421), 1,
      aux_sym_symbol_token2,
    ACTIONS(1429), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(1433), 1,
      sym_path,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1365), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1369), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1373), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1405), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(1413), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1725), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1409), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1425), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [14069] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1437), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1439), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1419), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [14213] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(167), 1,
      anon_sym_LBRACE,
    ACTIONS(1071), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1060), 1,
      sym_macro,
    STATE(1099), 1,
      sym__address_binary_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1133), 1,
      sym_call_expression,
    STATE(1134), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1130), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [14357] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1441), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1443), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1458), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [14501] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(142), 1,
      anon_sym_LPAREN,
    ACTIONS(144), 1,
      anon_sym_LBRACE,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    STATE(261), 1,
      sym_memory_space,
    STATE(1054), 1,
      sym_macro,
    STATE(1115), 1,
      sym_call_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1118), 1,
      sym_integer,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1445), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1447), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1459), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [14645] = 40,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(120), 1,
      anon_sym_BSLASH,
    ACTIONS(122), 1,
      anon_sym_SQUOTE,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(126), 1,
      sym_access_class,
    ACTIONS(132), 1,
      aux_sym_symbol_token1,
    ACTIONS(134), 1,
      aux_sym_symbol_token2,
    ACTIONS(138), 1,
      sym_path,
    ACTIONS(140), 1,
      sym_identifier,
    ACTIONS(152), 1,
      anon_sym_STAR,
    ACTIONS(154), 1,
      anon_sym_AMP,
    ACTIONS(162), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(167), 1,
      anon_sym_LBRACE,
    ACTIONS(1071), 1,
      anon_sym_LPAREN,
    STATE(261), 1,
      sym_memory_space,
    STATE(1060), 1,
      sym_macro,
    STATE(1099), 1,
      sym__address_binary_expression,
    STATE(1116), 1,
      sym__hll_var_call_expression,
    STATE(1119), 1,
      aux_sym__path_expression_repeat2,
    STATE(1133), 1,
      sym_call_expression,
    STATE(1134), 1,
      sym_integer,
    STATE(1227), 1,
      aux_sym__path_expression_repeat1,
    STATE(1472), 1,
      sym__unary_expression,
    STATE(1473), 1,
      sym__prefix_postfix_expression,
    STATE(1474), 1,
      sym__binary_expression,
    STATE(1475), 1,
      sym__and_expression,
    STATE(1476), 1,
      sym_trace32_hll_variable,
    STATE(1482), 1,
      sym__file_dialog,
    STATE(2483), 1,
      sym__function_identifier,
    STATE(2485), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(104), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(130), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    ACTIONS(156), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(222), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    STATE(1141), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(169), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(171), 3,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(1493), 13,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
  [14789] = 38,
    ACTIONS(45), 1,
      sym_identifier,
    ACTIONS(47), 1,
      anon_sym_LPAREN,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(61), 1,
      anon_sym_STAR,
    ACTIONS(65), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(67), 1,
      anon_sym_BSLASH,
    ACTIONS(69), 1,
      anon_sym_SQUOTE,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(73), 1,
      sym_access_class,
    ACTIONS(81), 1,
      aux_sym_symbol_token1,
    ACTIONS(83), 1,
      aux_sym_symbol_token2,
    ACTIONS(87), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(211), 1,
      sym_memory_space,
    STATE(406), 1,
      sym_macro,
    STATE(411), 1,
      sym_integer,
    STATE(424), 1,
      sym_call_expression,
    STATE(428), 1,
      sym__hll_var_call_expression,
    STATE(443), 1,
      aux_sym__path_expression_repeat1,
    STATE(564), 1,
      sym__prefix_postfix_expression,
    STATE(565), 1,
      sym__binary_expression,
    STATE(569), 1,
      sym__file_dialog,
    STATE(573), 1,
      sym__and_expression,
    STATE(575), 1,
      sym_trace32_hll_variable,
    STATE(581), 1,
      sym__unary_expression,
    STATE(1636), 1,
      sym__address_binary_expression,
    STATE(2589), 1,
      sym__function_identifier,
    STATE(2590), 1,
      sym__hll_var_call_expression_function,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(55), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(57), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(59), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(75), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    ACTIONS(79), 2,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
    STATE(1678), 2,
      sym__address_expression,
      sym__parenthesized_address_expression,
    ACTIONS(1449), 3,
      sym_float,
      sym_time,
      sym_frequency,
    ACTIONS(1451), 4,
      sym_path,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
    STATE(454), 14,
      sym__expression,
      sym__parenthesized_expression,
      sym__path_expression,
      sym_unary_expression,
      sym_binary_expression,
      sym_assignment_expression,
      sym__literal,
      sym_address,
      sym_bitmask,
      sym_string,
      sym_character,
      sym_symbol,
      sym__special_command_operator,
      aux_sym__path_expression_repeat2,
  [14929] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(41), 1,
      sym__blank,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1453), 1,
      sym_comment,
    ACTIONS(1455), 1,
      sym__blank_line,
    STATE(309), 1,
      aux_sym_if_block_repeat1,
    STATE(359), 1,
      sym__label,
    STATE(361), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2029), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1568), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [15079] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1457), 1,
      sym_comment,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1461), 1,
      sym__blank,
    STATE(360), 1,
      sym__label,
    STATE(372), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2018), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1538), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [15229] = 43,
    ACTIONS(1463), 1,
      ts_builtin_sym_end,
    ACTIONS(1465), 1,
      sym_identifier,
    ACTIONS(1468), 1,
      aux_sym_block_token1,
    ACTIONS(1471), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(1474), 1,
      aux_sym_if_block_token1,
    ACTIONS(1477), 1,
      aux_sym_while_block_token1,
    ACTIONS(1480), 1,
      aux_sym_while_block_token2,
    ACTIONS(1483), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1486), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1489), 1,
      anon_sym_AMP,
    ACTIONS(1492), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(1498), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(1501), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1507), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(1510), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(1513), 1,
      sym__blank_line,
    ACTIONS(1516), 1,
      sym__blank,
    ACTIONS(1519), 1,
      sym_label_identifier,
    STATE(359), 1,
      sym__label,
    STATE(361), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1495), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(1504), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(268), 13,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
      aux_sym_script_repeat1,
  [15375] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1461), 1,
      sym__blank,
    ACTIONS(1522), 1,
      sym_comment,
    ACTIONS(1524), 1,
      sym__blank_line,
    STATE(270), 1,
      aux_sym_if_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(372), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1985), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1498), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [15525] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1461), 1,
      sym__blank,
    ACTIONS(1526), 1,
      sym_comment,
    STATE(360), 1,
      sym__label,
    STATE(372), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2022), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1504), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [15675] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1528), 1,
      sym_comment,
    ACTIONS(1530), 1,
      sym__blank_line,
    ACTIONS(1532), 1,
      sym__blank,
    STATE(276), 1,
      aux_sym_if_block_repeat1,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2027), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1321), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [15825] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1534), 1,
      sym_comment,
    ACTIONS(1536), 1,
      sym__blank_line,
    ACTIONS(1538), 1,
      sym__blank,
    STATE(274), 1,
      aux_sym_if_block_repeat1,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2014), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1537), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [15975] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1461), 1,
      sym__blank,
    ACTIONS(1540), 1,
      sym_comment,
    ACTIONS(1542), 1,
      sym__blank_line,
    STATE(277), 1,
      aux_sym_if_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(372), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1984), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1519), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [16125] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1457), 1,
      sym_comment,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1538), 1,
      sym__blank,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2018), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1538), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [16275] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1453), 1,
      sym_comment,
    ACTIONS(1461), 1,
      sym__blank,
    ACTIONS(1544), 1,
      sym__blank_line,
    STATE(278), 1,
      aux_sym_if_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(372), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2029), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1568), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [16425] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1532), 1,
      sym__blank,
    ACTIONS(1546), 1,
      sym_comment,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2028), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1332), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [16575] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1461), 1,
      sym__blank,
    ACTIONS(1548), 1,
      sym_comment,
    STATE(360), 1,
      sym__label,
    STATE(372), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2006), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1577), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [16725] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1461), 1,
      sym__blank,
    ACTIONS(1550), 1,
      sym_comment,
    STATE(360), 1,
      sym__label,
    STATE(372), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2024), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1580), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [16875] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1552), 1,
      anon_sym_RPAREN,
    ACTIONS(1554), 1,
      sym__blank_line,
    ACTIONS(1556), 1,
      sym__blank,
    STATE(351), 1,
      aux_sym_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(310), 13,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
      aux_sym_block_repeat2,
  [17021] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1560), 1,
      sym_comment,
    ACTIONS(1562), 1,
      sym__blank_line,
    ACTIONS(1564), 1,
      sym__blank,
    STATE(281), 1,
      aux_sym_if_block_repeat1,
    STATE(353), 1,
      sym__label,
    STATE(362), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1962), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2164), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [17171] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1564), 1,
      sym__blank,
    ACTIONS(1566), 1,
      sym_comment,
    STATE(353), 1,
      sym__label,
    STATE(362), 1,
      aux_sym_block_repeat1,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1963), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2165), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [17321] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1568), 1,
      sym_comment,
    ACTIONS(1570), 1,
      sym__blank_line,
    ACTIONS(1572), 1,
      sym__blank,
    STATE(283), 1,
      aux_sym_if_block_repeat1,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(2040), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2031), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [17471] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1572), 1,
      sym__blank,
    ACTIONS(1574), 1,
      sym_comment,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1934), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1964), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [17621] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1538), 1,
      sym__blank,
    ACTIONS(1576), 1,
      sym_comment,
    ACTIONS(1578), 1,
      sym__blank_line,
    STATE(289), 1,
      aux_sym_if_block_repeat1,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1936), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1342), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [17771] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1554), 1,
      sym__blank_line,
    ACTIONS(1580), 1,
      anon_sym_RPAREN,
    ACTIONS(1582), 1,
      sym__blank,
    STATE(350), 1,
      aux_sym_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(310), 13,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
      aux_sym_block_repeat2,
  [17917] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1564), 1,
      sym__blank,
    ACTIONS(1584), 1,
      sym_comment,
    ACTIONS(1586), 1,
      sym__blank_line,
    STATE(290), 1,
      aux_sym_if_block_repeat1,
    STATE(353), 1,
      sym__label,
    STATE(362), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1939), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2217), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [18067] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(41), 1,
      sym__blank,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1522), 1,
      sym_comment,
    ACTIONS(1588), 1,
      sym__blank_line,
    STATE(305), 1,
      aux_sym_if_block_repeat1,
    STATE(359), 1,
      sym__label,
    STATE(361), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1985), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1498), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [18217] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1564), 1,
      sym__blank,
    ACTIONS(1590), 1,
      sym_comment,
    ACTIONS(1592), 1,
      sym__blank_line,
    STATE(291), 1,
      aux_sym_if_block_repeat1,
    STATE(353), 1,
      sym__label,
    STATE(362), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1942), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2058), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [18367] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1538), 1,
      sym__blank,
    ACTIONS(1594), 1,
      sym_comment,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(1946), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1320), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [18517] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1564), 1,
      sym__blank,
    ACTIONS(1596), 1,
      sym_comment,
    STATE(353), 1,
      sym__label,
    STATE(362), 1,
      aux_sym_block_repeat1,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1947), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2094), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [18667] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1564), 1,
      sym__blank,
    ACTIONS(1598), 1,
      sym_comment,
    STATE(353), 1,
      sym__label,
    STATE(362), 1,
      aux_sym_block_repeat1,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1951), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2107), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [18817] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1532), 1,
      sym__blank,
    ACTIONS(1534), 1,
      sym_comment,
    ACTIONS(1600), 1,
      sym__blank_line,
    STATE(293), 1,
      aux_sym_if_block_repeat1,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2014), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1537), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [18967] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1457), 1,
      sym_comment,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1532), 1,
      sym__blank,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2018), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1538), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [19117] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1572), 1,
      sym__blank,
    ACTIONS(1602), 1,
      sym_comment,
    ACTIONS(1604), 1,
      sym__blank_line,
    STATE(295), 1,
      aux_sym_if_block_repeat1,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1968), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2034), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [19267] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1572), 1,
      sym__blank,
    ACTIONS(1606), 1,
      sym_comment,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1971), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1950), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [19417] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1532), 1,
      sym__blank,
    ACTIONS(1608), 1,
      sym_comment,
    ACTIONS(1610), 1,
      sym__blank_line,
    STATE(299), 1,
      aux_sym_if_block_repeat1,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2016), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1290), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [19567] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1538), 1,
      sym__blank,
    ACTIONS(1612), 1,
      sym_comment,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(1969), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1300), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [19717] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(41), 1,
      sym__blank,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1548), 1,
      sym_comment,
    STATE(359), 1,
      sym__label,
    STATE(361), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2006), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1577), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [19867] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1532), 1,
      sym__blank,
    ACTIONS(1614), 1,
      sym_comment,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2017), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1293), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [20017] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1564), 1,
      sym__blank,
    ACTIONS(1616), 1,
      sym_comment,
    ACTIONS(1618), 1,
      sym__blank_line,
    STATE(301), 1,
      aux_sym_if_block_repeat1,
    STATE(353), 1,
      sym__label,
    STATE(362), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1972), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1791), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [20167] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1564), 1,
      sym__blank,
    ACTIONS(1620), 1,
      sym_comment,
    STATE(353), 1,
      sym__label,
    STATE(362), 1,
      aux_sym_block_repeat1,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1975), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1794), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [20317] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1622), 1,
      anon_sym_RPAREN,
    ACTIONS(1624), 1,
      sym__blank_line,
    ACTIONS(1626), 1,
      sym__blank,
    STATE(349), 1,
      aux_sym_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(304), 13,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
      aux_sym_block_repeat2,
  [20463] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(41), 1,
      sym__blank,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1628), 1,
      ts_builtin_sym_end,
    ACTIONS(1630), 1,
      sym__blank_line,
    STATE(359), 1,
      sym__label,
    STATE(361), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(268), 13,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
      aux_sym_script_repeat1,
  [20609] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1554), 1,
      sym__blank_line,
    ACTIONS(1632), 1,
      anon_sym_RPAREN,
    ACTIONS(1634), 1,
      sym__blank,
    STATE(354), 1,
      aux_sym_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(310), 13,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
      aux_sym_block_repeat2,
  [20755] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(41), 1,
      sym__blank,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1526), 1,
      sym_comment,
    STATE(359), 1,
      sym__label,
    STATE(361), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2022), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1504), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [20905] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1632), 1,
      anon_sym_RPAREN,
    ACTIONS(1634), 1,
      sym__blank,
    ACTIONS(1636), 1,
      sym__blank_line,
    STATE(354), 1,
      aux_sym_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(308), 13,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
      aux_sym_block_repeat2,
  [21051] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1552), 1,
      anon_sym_RPAREN,
    ACTIONS(1556), 1,
      sym__blank,
    ACTIONS(1638), 1,
      sym__blank_line,
    STATE(351), 1,
      aux_sym_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(285), 13,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
      aux_sym_block_repeat2,
  [21197] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1554), 1,
      sym__blank_line,
    ACTIONS(1640), 1,
      anon_sym_RPAREN,
    ACTIONS(1642), 1,
      sym__blank,
    STATE(355), 1,
      aux_sym_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(310), 13,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
      aux_sym_block_repeat2,
  [21343] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(41), 1,
      sym__blank,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1550), 1,
      sym_comment,
    STATE(359), 1,
      sym__label,
    STATE(361), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2024), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1580), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [21493] = 43,
    ACTIONS(1644), 1,
      sym_identifier,
    ACTIONS(1647), 1,
      aux_sym_block_token1,
    ACTIONS(1650), 1,
      anon_sym_RPAREN,
    ACTIONS(1652), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(1655), 1,
      aux_sym_if_block_token1,
    ACTIONS(1658), 1,
      aux_sym_while_block_token1,
    ACTIONS(1661), 1,
      aux_sym_while_block_token2,
    ACTIONS(1664), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1667), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1670), 1,
      anon_sym_AMP,
    ACTIONS(1673), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(1679), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(1682), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1688), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(1691), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(1694), 1,
      sym__blank_line,
    ACTIONS(1697), 1,
      sym__blank,
    ACTIONS(1700), 1,
      sym_label_identifier,
    STATE(360), 1,
      sym__label,
    STATE(372), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1676), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(1685), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(310), 13,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
      aux_sym_block_repeat2,
  [21639] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1703), 1,
      anon_sym_RPAREN,
    ACTIONS(1705), 1,
      sym__blank_line,
    ACTIONS(1707), 1,
      sym__blank,
    STATE(352), 1,
      aux_sym_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(279), 13,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
      aux_sym_block_repeat2,
  [21785] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1560), 1,
      sym_comment,
    ACTIONS(1572), 1,
      sym__blank,
    ACTIONS(1709), 1,
      sym__blank_line,
    STATE(313), 1,
      aux_sym_if_block_repeat1,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1962), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2164), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [21935] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1566), 1,
      sym_comment,
    ACTIONS(1572), 1,
      sym__blank,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1963), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2165), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [22085] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1572), 1,
      sym__blank,
    ACTIONS(1584), 1,
      sym_comment,
    ACTIONS(1711), 1,
      sym__blank_line,
    STATE(316), 1,
      aux_sym_if_block_repeat1,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1939), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2217), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [22235] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1572), 1,
      sym__blank,
    ACTIONS(1590), 1,
      sym_comment,
    ACTIONS(1713), 1,
      sym__blank_line,
    STATE(317), 1,
      aux_sym_if_block_repeat1,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1942), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2058), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [22385] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1572), 1,
      sym__blank,
    ACTIONS(1596), 1,
      sym_comment,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1947), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2094), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [22535] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1572), 1,
      sym__blank,
    ACTIONS(1598), 1,
      sym_comment,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(1951), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2107), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [22685] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1572), 1,
      sym__blank,
    ACTIONS(1715), 1,
      sym_comment,
    ACTIONS(1717), 1,
      sym__blank_line,
    STATE(319), 1,
      aux_sym_if_block_repeat1,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(2037), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1774), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [22835] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1572), 1,
      sym__blank,
    ACTIONS(1719), 1,
      sym_comment,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(2038), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1773), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [22985] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(41), 1,
      sym__blank,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1540), 1,
      sym_comment,
    ACTIONS(1721), 1,
      sym__blank_line,
    STATE(298), 1,
      aux_sym_if_block_repeat1,
    STATE(359), 1,
      sym__label,
    STATE(361), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1984), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1519), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [23135] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1522), 1,
      sym_comment,
    ACTIONS(1538), 1,
      sym__blank,
    ACTIONS(1723), 1,
      sym__blank_line,
    STATE(322), 1,
      aux_sym_if_block_repeat1,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1985), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1498), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [23285] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1526), 1,
      sym_comment,
    ACTIONS(1538), 1,
      sym__blank,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2022), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1504), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [23435] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1538), 1,
      sym__blank,
    ACTIONS(1540), 1,
      sym_comment,
    ACTIONS(1725), 1,
      sym__blank_line,
    STATE(325), 1,
      aux_sym_if_block_repeat1,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1984), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1519), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [23585] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1453), 1,
      sym_comment,
    ACTIONS(1538), 1,
      sym__blank,
    ACTIONS(1727), 1,
      sym__blank_line,
    STATE(326), 1,
      aux_sym_if_block_repeat1,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2029), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1568), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [23735] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1538), 1,
      sym__blank,
    ACTIONS(1548), 1,
      sym_comment,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2006), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1577), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [23885] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1538), 1,
      sym__blank,
    ACTIONS(1550), 1,
      sym_comment,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2024), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1580), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [24035] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(41), 1,
      sym__blank,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1534), 1,
      sym_comment,
    ACTIONS(1729), 1,
      sym__blank_line,
    STATE(328), 1,
      aux_sym_if_block_repeat1,
    STATE(359), 1,
      sym__label,
    STATE(361), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2014), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1537), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [24185] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(41), 1,
      sym__blank,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1457), 1,
      sym_comment,
    ACTIONS(1459), 1,
      sym__blank_line,
    STATE(359), 1,
      sym__label,
    STATE(361), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2018), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1538), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [24335] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1522), 1,
      sym_comment,
    ACTIONS(1532), 1,
      sym__blank,
    ACTIONS(1731), 1,
      sym__blank_line,
    STATE(330), 1,
      aux_sym_if_block_repeat1,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1985), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1498), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [24485] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1526), 1,
      sym_comment,
    ACTIONS(1532), 1,
      sym__blank,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2022), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1504), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [24635] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1532), 1,
      sym__blank,
    ACTIONS(1540), 1,
      sym_comment,
    ACTIONS(1733), 1,
      sym__blank_line,
    STATE(333), 1,
      aux_sym_if_block_repeat1,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1984), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1519), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [24785] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1453), 1,
      sym_comment,
    ACTIONS(1532), 1,
      sym__blank,
    ACTIONS(1735), 1,
      sym__blank_line,
    STATE(334), 1,
      aux_sym_if_block_repeat1,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2029), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1568), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [24935] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1532), 1,
      sym__blank,
    ACTIONS(1548), 1,
      sym_comment,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2006), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1577), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [25085] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1459), 1,
      sym__blank_line,
    ACTIONS(1532), 1,
      sym__blank,
    ACTIONS(1550), 1,
      sym_comment,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    STATE(2024), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1580), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [25235] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1461), 1,
      sym__blank,
    ACTIONS(1534), 1,
      sym_comment,
    ACTIONS(1737), 1,
      sym__blank_line,
    STATE(267), 1,
      aux_sym_if_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(372), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2014), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1537), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [25385] = 45,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1538), 1,
      sym__blank,
    ACTIONS(1739), 1,
      sym_comment,
    ACTIONS(1741), 1,
      sym__blank_line,
    STATE(297), 1,
      aux_sym_if_block_repeat1,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1976), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1323), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [25535] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(41), 1,
      sym__blank,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1743), 1,
      sym__blank_line,
    STATE(339), 1,
      aux_sym_if_block_repeat1,
    STATE(359), 1,
      sym__label,
    STATE(361), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1560), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [25680] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1564), 1,
      sym__blank,
    ACTIONS(1745), 1,
      sym__blank_line,
    STATE(353), 1,
      sym__label,
    STATE(362), 1,
      aux_sym_block_repeat1,
    STATE(1602), 1,
      aux_sym_if_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2155), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [25825] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(41), 1,
      sym__blank,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1745), 1,
      sym__blank_line,
    STATE(359), 1,
      sym__label,
    STATE(361), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1602), 1,
      aux_sym_if_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1518), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [25970] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1572), 1,
      sym__blank,
    ACTIONS(1747), 1,
      sym__blank_line,
    STATE(341), 1,
      aux_sym_if_block_repeat1,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2142), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [26115] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1572), 1,
      sym__blank,
    ACTIONS(1745), 1,
      sym__blank_line,
    STATE(356), 1,
      sym__label,
    STATE(366), 1,
      aux_sym_block_repeat1,
    STATE(1602), 1,
      aux_sym_if_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2155), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [26260] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1461), 1,
      sym__blank,
    ACTIONS(1749), 1,
      sym__blank_line,
    STATE(343), 1,
      aux_sym_if_block_repeat1,
    STATE(360), 1,
      sym__label,
    STATE(372), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1560), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [26405] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1461), 1,
      sym__blank,
    ACTIONS(1745), 1,
      sym__blank_line,
    STATE(360), 1,
      sym__label,
    STATE(372), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1602), 1,
      aux_sym_if_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1518), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [26550] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1538), 1,
      sym__blank,
    ACTIONS(1751), 1,
      sym__blank_line,
    STATE(347), 1,
      aux_sym_if_block_repeat1,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1560), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [26695] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1532), 1,
      sym__blank,
    ACTIONS(1753), 1,
      sym__blank_line,
    STATE(346), 1,
      aux_sym_if_block_repeat1,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1560), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [26840] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1532), 1,
      sym__blank,
    ACTIONS(1745), 1,
      sym__blank_line,
    STATE(358), 1,
      sym__label,
    STATE(365), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1602), 1,
      aux_sym_if_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1518), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [26985] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1538), 1,
      sym__blank,
    ACTIONS(1745), 1,
      sym__blank_line,
    STATE(357), 1,
      sym__label,
    STATE(367), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1602), 1,
      aux_sym_if_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1518), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [27130] = 43,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(1564), 1,
      sym__blank,
    ACTIONS(1755), 1,
      sym__blank_line,
    STATE(338), 1,
      aux_sym_if_block_repeat1,
    STATE(353), 1,
      sym__label,
    STATE(362), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2142), 12,
      sym_block,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [27275] = 42,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1632), 1,
      anon_sym_RPAREN,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(360), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [27415] = 42,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(1759), 1,
      sym__blank,
    ACTIONS(1761), 1,
      anon_sym_RPAREN,
    STATE(360), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [27555] = 42,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1580), 1,
      anon_sym_RPAREN,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(360), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [27695] = 42,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1552), 1,
      anon_sym_RPAREN,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(360), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [27835] = 42,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1763), 1,
      sym_comment,
    ACTIONS(1765), 1,
      sym__blank_line,
    ACTIONS(1767), 1,
      sym__blank,
    STATE(353), 1,
      sym__label,
    STATE(374), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2148), 11,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [27975] = 42,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1640), 1,
      anon_sym_RPAREN,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(360), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [28115] = 42,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(1759), 1,
      sym__blank,
    ACTIONS(1769), 1,
      anon_sym_RPAREN,
    STATE(360), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [28255] = 42,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1763), 1,
      sym_comment,
    ACTIONS(1765), 1,
      sym__blank_line,
    ACTIONS(1771), 1,
      sym__blank,
    STATE(356), 1,
      sym__label,
    STATE(377), 1,
      aux_sym_block_repeat1,
    STATE(1875), 1,
      sym_recursive_macro_expansion,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2148), 11,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [28395] = 42,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1773), 1,
      sym_comment,
    ACTIONS(1775), 1,
      sym__blank_line,
    ACTIONS(1777), 1,
      sym__blank,
    STATE(357), 1,
      sym__label,
    STATE(376), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1562), 11,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [28535] = 42,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1773), 1,
      sym_comment,
    ACTIONS(1775), 1,
      sym__blank_line,
    ACTIONS(1779), 1,
      sym__blank,
    STATE(358), 1,
      sym__label,
    STATE(378), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1562), 11,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [28675] = 42,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1773), 1,
      sym_comment,
    ACTIONS(1775), 1,
      sym__blank_line,
    ACTIONS(1781), 1,
      sym__blank,
    STATE(359), 1,
      sym__label,
    STATE(375), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1562), 11,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [28815] = 42,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1773), 1,
      sym_comment,
    ACTIONS(1775), 1,
      sym__blank_line,
    ACTIONS(1783), 1,
      sym__blank,
    STATE(360), 1,
      sym__label,
    STATE(373), 1,
      aux_sym_block_repeat1,
    STATE(1396), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1562), 11,
      sym__statement,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [28955] = 41,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(359), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [29092] = 41,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1759), 1,
      sym__blank,
    ACTIONS(1785), 1,
      aux_sym_block_token1,
    STATE(353), 1,
      sym__label,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(1831), 1,
      sym_recursive_macro_expansion,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2158), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [29229] = 41,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1759), 1,
      sym__blank,
    ACTIONS(1787), 1,
      anon_sym_EQ,
    STATE(353), 1,
      sym__label,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(1831), 1,
      sym_recursive_macro_expansion,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2158), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [29366] = 41,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1759), 1,
      sym__blank,
    ACTIONS(1787), 1,
      anon_sym_EQ,
    STATE(360), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [29503] = 41,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(358), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [29640] = 41,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1759), 1,
      sym__blank,
    ACTIONS(1785), 1,
      aux_sym_block_token1,
    STATE(356), 1,
      sym__label,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(1831), 1,
      sym_recursive_macro_expansion,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2158), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [29777] = 41,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(357), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [29914] = 41,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1759), 1,
      sym__blank,
    ACTIONS(1787), 1,
      anon_sym_EQ,
    STATE(359), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [30051] = 41,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1759), 1,
      sym__blank,
    ACTIONS(1787), 1,
      anon_sym_EQ,
    STATE(356), 1,
      sym__label,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(1831), 1,
      sym_recursive_macro_expansion,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2158), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [30188] = 41,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1759), 1,
      sym__blank,
    ACTIONS(1787), 1,
      anon_sym_EQ,
    STATE(357), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [30325] = 41,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1759), 1,
      sym__blank,
    ACTIONS(1787), 1,
      anon_sym_EQ,
    STATE(358), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [30462] = 41,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(360), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [30599] = 40,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(330), 1,
      aux_sym_if_block_token1,
    ACTIONS(332), 1,
      aux_sym_while_block_token1,
    ACTIONS(334), 1,
      aux_sym_while_block_token2,
    ACTIONS(336), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(360), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2553), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [30733] = 40,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(344), 1,
      aux_sym_if_block_token1,
    ACTIONS(346), 1,
      aux_sym_while_block_token1,
    ACTIONS(348), 1,
      aux_sym_while_block_token2,
    ACTIONS(350), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(353), 1,
      sym__label,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(1831), 1,
      sym_recursive_macro_expansion,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2575), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2158), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [30867] = 40,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(13), 1,
      aux_sym_if_block_token1,
    ACTIONS(15), 1,
      aux_sym_while_block_token1,
    ACTIONS(17), 1,
      aux_sym_while_block_token2,
    ACTIONS(19), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(359), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2546), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [31001] = 40,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(21), 1,
      anon_sym_AMP_AMP,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(304), 1,
      aux_sym_if_block_token1,
    ACTIONS(306), 1,
      aux_sym_while_block_token1,
    ACTIONS(308), 1,
      aux_sym_while_block_token2,
    ACTIONS(310), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(357), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2567), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [31135] = 40,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(342), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(352), 1,
      anon_sym_AMP_AMP,
    ACTIONS(354), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(358), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(360), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(366), 1,
      aux_sym_if_block_token1,
    ACTIONS(368), 1,
      aux_sym_while_block_token1,
    ACTIONS(370), 1,
      aux_sym_while_block_token2,
    ACTIONS(372), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(356), 1,
      sym__label,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(1831), 1,
      sym_recursive_macro_expansion,
    STATE(2140), 1,
      sym__repeat_block_pre_condition_format,
    STATE(2141), 1,
      sym__repeat_block_post_condition_format,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2348), 1,
      sym_macro_definition_command,
    STATE(2409), 1,
      sym_macro,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2577), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2657), 1,
      sym__host_command,
    STATE(2658), 1,
      sym__other_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2677), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(356), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(2158), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [31269] = 40,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(11), 1,
      aux_sym_subroutine_block_token1,
    ACTIONS(23), 1,
      anon_sym_AMP,
    ACTIONS(25), 1,
      aux_sym_macro_definition_token1,
    ACTIONS(29), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(31), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(37), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(43), 1,
      sym_label_identifier,
    ACTIONS(316), 1,
      aux_sym_if_block_token1,
    ACTIONS(318), 1,
      aux_sym_while_block_token1,
    ACTIONS(320), 1,
      aux_sym_while_block_token2,
    ACTIONS(322), 1,
      aux_sym__repeat_block_pre_condition_format_token1,
    ACTIONS(324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1759), 1,
      sym__blank,
    STATE(358), 1,
      sym__label,
    STATE(1392), 1,
      sym_recursive_macro_expansion,
    STATE(1531), 1,
      sym__repeat_block_pre_condition_format,
    STATE(1532), 1,
      sym__repeat_block_post_condition_format,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2316), 1,
      sym_macro,
    STATE(2404), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2572), 1,
      sym__hll_var_if_command,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2618), 1,
      sym__var_command,
    STATE(2629), 1,
      sym__host_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2672), 1,
      sym__other_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(27), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
    STATE(1592), 10,
      sym_labeled_expression,
      sym_subroutine_block,
      sym_if_block,
      sym_while_block,
      sym_repeat_block,
      sym_macro_assignment_expression,
      sym_macro_definition,
      sym_parameter_declaration,
      sym_subroutine_call_expression,
      sym_command_expression,
  [31403] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1789), 1,
      anon_sym_STAR,
    ACTIONS(1791), 1,
      anon_sym_RBRACK,
    ACTIONS(1795), 1,
      sym__blank,
    STATE(403), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1024), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1204), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [31504] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1797), 1,
      anon_sym_STAR,
    ACTIONS(1799), 1,
      anon_sym_RBRACK,
    ACTIONS(1801), 1,
      sym__blank,
    STATE(394), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(385), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1180), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [31605] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1803), 1,
      anon_sym_STAR,
    ACTIONS(1805), 1,
      anon_sym_RBRACK,
    ACTIONS(1807), 1,
      sym__blank,
    STATE(397), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(387), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1161), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [31706] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1809), 1,
      anon_sym_STAR,
    ACTIONS(1811), 1,
      anon_sym_RBRACK,
    ACTIONS(1813), 1,
      sym__blank,
    STATE(393), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(391), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1244), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [31807] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1815), 1,
      anon_sym_STAR,
    ACTIONS(1817), 1,
      anon_sym_RBRACK,
    ACTIONS(1819), 1,
      sym__blank,
    STATE(399), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(379), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1199), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [31908] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1821), 1,
      anon_sym_STAR,
    ACTIONS(1823), 1,
      anon_sym_RBRACK,
    ACTIONS(1825), 1,
      sym__blank,
    STATE(398), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1024), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1274), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [32009] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1827), 1,
      anon_sym_STAR,
    ACTIONS(1829), 1,
      anon_sym_RBRACK,
    ACTIONS(1831), 1,
      sym__blank,
    STATE(404), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1024), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1186), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [32110] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1833), 1,
      anon_sym_STAR,
    ACTIONS(1835), 1,
      anon_sym_RBRACK,
    ACTIONS(1837), 1,
      sym__blank,
    STATE(396), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(388), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1275), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [32211] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1839), 1,
      anon_sym_STAR,
    ACTIONS(1841), 1,
      anon_sym_RBRACK,
    ACTIONS(1843), 1,
      sym__blank,
    STATE(400), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1024), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1202), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [32312] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1845), 1,
      anon_sym_STAR,
    ACTIONS(1847), 1,
      anon_sym_RBRACK,
    ACTIONS(1849), 1,
      sym__blank,
    STATE(392), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1024), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1284), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [32413] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1851), 1,
      anon_sym_STAR,
    ACTIONS(1853), 1,
      anon_sym_RBRACK,
    ACTIONS(1855), 1,
      sym__blank,
    STATE(401), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(384), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1211), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [32514] = 25,
    ACTIONS(458), 1,
      anon_sym_PERCENT,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(413), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat2,
    STATE(561), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat1,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    STATE(2385), 1,
      sym_format_expression,
    STATE(2712), 1,
      sym__hll_var_command_argument_list_default_format,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1056), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [32617] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1859), 1,
      anon_sym_STAR,
    ACTIONS(1861), 1,
      anon_sym_RBRACK,
    ACTIONS(1863), 1,
      sym__blank,
    STATE(395), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1024), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1181), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [32718] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(1865), 1,
      anon_sym_STAR,
    ACTIONS(1867), 1,
      anon_sym_RBRACK,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1197), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [32818] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(1861), 1,
      anon_sym_RBRACK,
    ACTIONS(1869), 1,
      anon_sym_STAR,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1176), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [32918] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1829), 1,
      anon_sym_RBRACK,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(1871), 1,
      anon_sym_STAR,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1277), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [33018] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(1873), 1,
      anon_sym_STAR,
    ACTIONS(1875), 1,
      anon_sym_RBRACK,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1182), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [33118] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1847), 1,
      anon_sym_RBRACK,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(1877), 1,
      anon_sym_STAR,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1179), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [33218] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1841), 1,
      anon_sym_RBRACK,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(1879), 1,
      anon_sym_STAR,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1200), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [33318] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(1881), 1,
      anon_sym_STAR,
    ACTIONS(1883), 1,
      anon_sym_RBRACK,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1162), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [33418] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1791), 1,
      anon_sym_RBRACK,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(1885), 1,
      anon_sym_STAR,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1203), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [33518] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(1887), 1,
      anon_sym_STAR,
    ACTIONS(1889), 1,
      anon_sym_RBRACK,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1262), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [33618] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1823), 1,
      anon_sym_RBRACK,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(1891), 1,
      anon_sym_STAR,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1273), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [33718] = 24,
    ACTIONS(458), 1,
      anon_sym_PERCENT,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(433), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat2,
    STATE(556), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat1,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    STATE(2420), 1,
      sym_format_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1058), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [33818] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(1893), 1,
      anon_sym_STAR,
    ACTIONS(1895), 1,
      anon_sym_RBRACK,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1207), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [33918] = 24,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(482), 1,
      anon_sym_AMP,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(1897), 1,
      anon_sym_STAR,
    ACTIONS(1899), 1,
      anon_sym_RBRACK,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1793), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1239), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [34018] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1903), 23,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1901), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [34075] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1919), 1,
      sym_path,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 16,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(1905), 19,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [34142] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1923), 23,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1921), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [34199] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1925), 1,
      sym_path,
    ACTIONS(1917), 7,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1909), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1912), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
    ACTIONS(1915), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
  [34262] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1929), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1927), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [34318] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1933), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1931), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [34374] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1939), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1942), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1937), 16,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(1935), 19,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [34438] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1947), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1945), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [34494] = 22,
    ACTIONS(458), 1,
      anon_sym_PERCENT,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(422), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat2,
    STATE(578), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat1,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    STATE(2430), 1,
      sym_format_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1055), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [34588] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1951), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1949), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [34644] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1955), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1953), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [34700] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1959), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1957), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [34756] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1963), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1961), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [34812] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1965), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1963), 19,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1961), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [34870] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1971), 1,
      anon_sym_SLASH,
    ACTIONS(1973), 1,
      anon_sym_CARET,
    ACTIONS(1967), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1969), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(1965), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1963), 17,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1961), 22,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [34936] = 12,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1971), 1,
      anon_sym_SLASH,
    ACTIONS(1973), 1,
      anon_sym_CARET,
    ACTIONS(1979), 1,
      anon_sym_PIPE,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1967), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1969), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(1965), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1977), 16,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(1975), 19,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [35010] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1989), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1987), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [35066] = 22,
    ACTIONS(1991), 1,
      sym_identifier,
    ACTIONS(1994), 1,
      anon_sym_LPAREN,
    ACTIONS(2009), 1,
      anon_sym_PERCENT,
    ACTIONS(2012), 1,
      anon_sym_BSLASH,
    ACTIONS(2015), 1,
      anon_sym_sizeof,
    ACTIONS(2024), 1,
      aux_sym_symbol_token1,
    ACTIONS(2027), 1,
      aux_sym_symbol_token2,
    ACTIONS(2030), 1,
      sym__hll_number_literal_pre_hook,
    STATE(422), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat2,
    STATE(554), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat1,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    STATE(2584), 1,
      sym_format_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2000), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2003), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2006), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2018), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(2021), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1125), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [35160] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2035), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(2033), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [35216] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 16,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(1905), 19,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [35280] = 14,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1971), 1,
      anon_sym_SLASH,
    ACTIONS(1973), 1,
      anon_sym_CARET,
    ACTIONS(1979), 1,
      anon_sym_PIPE,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1967), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1969), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(1965), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(2044), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2042), 6,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2039), 10,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2037), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [35358] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1971), 1,
      anon_sym_SLASH,
    ACTIONS(1967), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1969), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(1965), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1963), 17,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1961), 23,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [35422] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1971), 1,
      anon_sym_SLASH,
    ACTIONS(1969), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(1965), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1963), 17,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1961), 25,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [35484] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2049), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(2047), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [35540] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2053), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(2051), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [35596] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1917), 7,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1909), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1912), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
    ACTIONS(1915), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
  [35656] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1917), 7,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1939), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1942), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
    ACTIONS(1915), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
  [35716] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2057), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(2055), 26,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_COLON,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_COLON_COLON,
      sym_identifier,
      sym__blank,
  [35772] = 22,
    ACTIONS(458), 1,
      anon_sym_PERCENT,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(422), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat2,
    STATE(557), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat1,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    STATE(2356), 1,
      sym_format_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1027), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [35866] = 12,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1971), 1,
      anon_sym_SLASH,
    ACTIONS(1973), 1,
      anon_sym_CARET,
    ACTIONS(1979), 1,
      anon_sym_PIPE,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(1967), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1969), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(1965), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(2061), 16,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2059), 19,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [35940] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2063), 1,
      anon_sym_RBRACK,
    ACTIONS(2065), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2331), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [36031] = 21,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(2067), 1,
      anon_sym_RPAREN,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1031), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [36122] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2069), 1,
      anon_sym_RBRACK,
    ACTIONS(2071), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2472), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [36213] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2073), 1,
      anon_sym_RBRACK,
    ACTIONS(2075), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2221), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [36304] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2081), 1,
      sym_path,
    STATE(439), 1,
      aux_sym__path_expression_repeat1,
    ACTIONS(2079), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2077), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [36363] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2084), 1,
      anon_sym_RBRACK,
    ACTIONS(2086), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2369), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [36454] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2092), 1,
      anon_sym_AMP,
    STATE(441), 1,
      aux_sym__path_expression_repeat2,
    STATE(2656), 1,
      sym_macro,
    ACTIONS(2090), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2088), 23,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [36515] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2095), 1,
      anon_sym_RBRACK,
    ACTIONS(2097), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2375), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [36606] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2103), 1,
      sym_path,
    STATE(439), 1,
      aux_sym__path_expression_repeat1,
    ACTIONS(2101), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2099), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [36665] = 21,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(2105), 1,
      anon_sym_RPAREN,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1048), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [36756] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2107), 1,
      anon_sym_RBRACK,
    ACTIONS(2109), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2379), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [36847] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2111), 1,
      anon_sym_RBRACK,
    ACTIONS(2113), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2380), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [36938] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2115), 1,
      anon_sym_RBRACK,
    ACTIONS(2117), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2381), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [37029] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2119), 1,
      anon_sym_RBRACK,
    ACTIONS(2121), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2239), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [37120] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2123), 1,
      anon_sym_RBRACK,
    ACTIONS(2125), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2378), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [37211] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(272), 1,
      anon_sym_EQ,
    ACTIONS(278), 1,
      anon_sym_CARET_CARET,
    ACTIONS(280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(284), 1,
      anon_sym_SLASH,
    ACTIONS(286), 1,
      anon_sym_PIPE,
    ACTIONS(288), 1,
      anon_sym_CARET,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    ACTIONS(302), 1,
      sym__and_operator_pre_hook,
    ACTIONS(2131), 1,
      sym__blank,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(274), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(282), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(292), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(290), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2129), 6,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2127), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [37302] = 11,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(284), 1,
      anon_sym_SLASH,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(282), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2135), 13,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2133), 23,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [37373] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(2135), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2133), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [37432] = 17,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(284), 1,
      anon_sym_SLASH,
    ACTIONS(286), 1,
      anon_sym_PIPE,
    ACTIONS(288), 1,
      anon_sym_CARET,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    ACTIONS(302), 1,
      sym__and_operator_pre_hook,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(274), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(282), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(292), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(290), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2135), 8,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2133), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [37515] = 18,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(278), 1,
      anon_sym_CARET_CARET,
    ACTIONS(284), 1,
      anon_sym_SLASH,
    ACTIONS(286), 1,
      anon_sym_PIPE,
    ACTIONS(288), 1,
      anon_sym_CARET,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    ACTIONS(302), 1,
      sym__and_operator_pre_hook,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(274), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(282), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(292), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(290), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2135), 7,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2133), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [37600] = 9,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2135), 15,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2133), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [37667] = 16,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(284), 1,
      anon_sym_SLASH,
    ACTIONS(288), 1,
      anon_sym_CARET,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2137), 1,
      sym__and_operator_pre_hook,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(274), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(282), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(292), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(290), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2135), 8,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2133), 18,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [37748] = 15,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(284), 1,
      anon_sym_SLASH,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2137), 1,
      sym__and_operator_pre_hook,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(274), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(282), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(292), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(290), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2135), 8,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2133), 19,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [37827] = 12,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(284), 1,
      anon_sym_SLASH,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(274), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(282), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2135), 13,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2133), 21,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [37900] = 21,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(2140), 1,
      anon_sym_RPAREN,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1025), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [37991] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2142), 1,
      anon_sym_RBRACK,
    ACTIONS(2144), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2288), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [38082] = 17,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(284), 1,
      anon_sym_SLASH,
    ACTIONS(286), 1,
      anon_sym_PIPE,
    ACTIONS(288), 1,
      anon_sym_CARET,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    ACTIONS(302), 1,
      sym__and_operator_pre_hook,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(274), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(282), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(292), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(290), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2148), 8,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2146), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [38165] = 14,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(284), 1,
      anon_sym_SLASH,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(274), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(282), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(292), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(290), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2148), 9,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2146), 19,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [38242] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(272), 1,
      anon_sym_EQ,
    ACTIONS(278), 1,
      anon_sym_CARET_CARET,
    ACTIONS(280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(284), 1,
      anon_sym_SLASH,
    ACTIONS(286), 1,
      anon_sym_PIPE,
    ACTIONS(288), 1,
      anon_sym_CARET,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    ACTIONS(302), 1,
      sym__and_operator_pre_hook,
    ACTIONS(2131), 1,
      sym__blank,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(274), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(282), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(292), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(290), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2152), 6,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2150), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [38333] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(272), 1,
      anon_sym_EQ,
    ACTIONS(278), 1,
      anon_sym_CARET_CARET,
    ACTIONS(280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(284), 1,
      anon_sym_SLASH,
    ACTIONS(286), 1,
      anon_sym_PIPE,
    ACTIONS(288), 1,
      anon_sym_CARET,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    ACTIONS(302), 1,
      sym__and_operator_pre_hook,
    ACTIONS(2131), 1,
      sym__blank,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(274), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(282), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(292), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(290), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2156), 6,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2154), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [38424] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2160), 17,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2158), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [38489] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2164), 18,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2162), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [38552] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(272), 1,
      anon_sym_EQ,
    ACTIONS(278), 1,
      anon_sym_CARET_CARET,
    ACTIONS(280), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(284), 1,
      anon_sym_SLASH,
    ACTIONS(286), 1,
      anon_sym_PIPE,
    ACTIONS(288), 1,
      anon_sym_CARET,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    ACTIONS(302), 1,
      sym__and_operator_pre_hook,
    ACTIONS(2131), 1,
      sym__blank,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(274), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(282), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(292), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(294), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(290), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2168), 6,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2166), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [38643] = 21,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(2170), 1,
      anon_sym_RPAREN,
    ACTIONS(2172), 1,
      sym__blank,
    STATE(444), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1047), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [38734] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2174), 1,
      anon_sym_RBRACK,
    ACTIONS(2176), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2223), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [38825] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2178), 1,
      anon_sym_RBRACK,
    ACTIONS(2180), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2302), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [38916] = 21,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(2182), 1,
      anon_sym_RPAREN,
    ACTIONS(2184), 1,
      sym__blank,
    STATE(436), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1034), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [39007] = 21,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(2186), 1,
      anon_sym_RPAREN,
    ACTIONS(2188), 1,
      sym__blank,
    STATE(475), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1037), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [39098] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2190), 1,
      anon_sym_RBRACK,
    ACTIONS(2192), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2230), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [39189] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2194), 1,
      anon_sym_RBRACK,
    ACTIONS(2196), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2382), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [39280] = 21,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(2198), 1,
      anon_sym_RPAREN,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1038), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [39371] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2200), 1,
      anon_sym_LPAREN,
    ACTIONS(2203), 1,
      anon_sym_DOT,
    STATE(2373), 1,
      aux_sym__function_identifier_repeat1,
    ACTIONS(1907), 20,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(1905), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [39432] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2205), 1,
      anon_sym_RBRACK,
    ACTIONS(2207), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2306), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [39523] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2209), 1,
      anon_sym_RBRACK,
    ACTIONS(2211), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2272), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [39614] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2213), 1,
      anon_sym_RBRACK,
    ACTIONS(2215), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2392), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [39705] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2217), 1,
      anon_sym_RBRACK,
    ACTIONS(2219), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2357), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [39796] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2221), 1,
      anon_sym_RBRACK,
    ACTIONS(2223), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2470), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [39887] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2225), 1,
      anon_sym_RBRACK,
    ACTIONS(2227), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2340), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [39978] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2229), 1,
      anon_sym_RBRACK,
    ACTIONS(2231), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2345), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [40069] = 21,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(2233), 1,
      anon_sym_RPAREN,
    ACTIONS(2235), 1,
      sym__blank,
    STATE(459), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1041), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [40160] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2237), 1,
      anon_sym_RBRACK,
    ACTIONS(2239), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2364), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [40251] = 21,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2241), 1,
      anon_sym_RBRACK,
    ACTIONS(2243), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    STATE(2372), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [40342] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(296), 1,
      anon_sym_DOT_DOT,
    STATE(432), 1,
      sym_argument_list,
    STATE(2224), 1,
      aux_sym_block_repeat1,
    ACTIONS(276), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2135), 17,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      aux_sym_while_block_token2,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2133), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [40407] = 20,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(924), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [40495] = 20,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1103), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [40583] = 20,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2245), 1,
      sym__blank,
    STATE(497), 1,
      aux_sym_block_repeat1,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1104), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [40671] = 20,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2247), 1,
      sym__blank,
    STATE(529), 1,
      aux_sym_block_repeat1,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(673), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [40759] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1919), 1,
      sym_path,
    ACTIONS(2251), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2249), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [40815] = 20,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1062), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [40903] = 20,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2253), 1,
      sym__blank,
    STATE(489), 1,
      aux_sym_block_repeat1,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1095), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [40991] = 20,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2255), 1,
      sym__blank,
    STATE(502), 1,
      aux_sym_block_repeat1,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1190), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [41079] = 20,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1046), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [41167] = 20,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1108), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [41255] = 20,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(963), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [41343] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2259), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      aux_sym_symbol_token3,
      sym_comment,
      sym__blank_line,
    ACTIONS(2257), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [41397] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(2261), 1,
      sym__blank,
    STATE(542), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1466), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [41485] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1433), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [41573] = 20,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1218), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [41661] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2079), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2077), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [41715] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1444), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [41803] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1445), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [41891] = 20,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(682), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [41979] = 20,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2263), 1,
      sym__blank,
    STATE(488), 1,
      aux_sym_block_repeat1,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(964), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [42067] = 20,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1138), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [42155] = 20,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2265), 1,
      sym__blank,
    STATE(537), 1,
      aux_sym_block_repeat1,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1175), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [42243] = 20,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(2267), 1,
      sym__blank,
    STATE(493), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1035), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [42331] = 20,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1386), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [42419] = 20,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    ACTIONS(2269), 1,
      sym__blank,
    STATE(513), 1,
      aux_sym_block_repeat1,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1352), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [42507] = 20,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1306), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [42595] = 20,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    ACTIONS(2271), 1,
      sym__blank,
    STATE(551), 1,
      aux_sym_block_repeat1,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1308), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [42683] = 20,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2273), 1,
      sym__blank,
    STATE(527), 1,
      aux_sym_block_repeat1,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1387), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [42771] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    STATE(503), 1,
      sym_macro,
    ACTIONS(2251), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2249), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [42827] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2277), 22,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      aux_sym_symbol_token3,
      sym_comment,
      sym__blank_line,
    ACTIONS(2275), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [42881] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2283), 1,
      aux_sym_symbol_token3,
    ACTIONS(2281), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2279), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [42937] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1463), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [43025] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1464), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [43113] = 20,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(2285), 1,
      sym__blank,
    STATE(508), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1122), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [43201] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1479), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [43289] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1468), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [43377] = 20,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2287), 1,
      sym__blank,
    STATE(511), 1,
      aux_sym_block_repeat1,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1378), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [43465] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1379), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [43553] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1470), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [43641] = 20,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1393), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [43729] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1481), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [43817] = 20,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(678), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [43905] = 20,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2289), 1,
      sym__blank,
    STATE(531), 1,
      aux_sym_block_repeat1,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(855), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [43993] = 20,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(833), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [44081] = 20,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2291), 1,
      sym__blank,
    STATE(534), 1,
      aux_sym_block_repeat1,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1215), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [44169] = 20,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2293), 1,
      sym__blank,
    STATE(538), 1,
      aux_sym_block_repeat1,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(863), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [44257] = 20,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1245), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [44345] = 20,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2295), 1,
      sym__blank,
    STATE(536), 1,
      aux_sym_block_repeat1,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1246), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [44433] = 20,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1251), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [44521] = 20,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1185), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [44609] = 20,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(836), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [44697] = 20,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2297), 1,
      sym__blank,
    STATE(541), 1,
      aux_sym_block_repeat1,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(889), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [44785] = 20,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2299), 1,
      sym__blank,
    STATE(498), 1,
      aux_sym_block_repeat1,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(933), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [44873] = 20,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(899), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [44961] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1363), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [45049] = 20,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2301), 1,
      sym__blank,
    STATE(546), 1,
      aux_sym_block_repeat1,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(906), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [45137] = 20,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2303), 1,
      sym__blank,
    STATE(506), 1,
      aux_sym_block_repeat1,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(679), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [45225] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(2305), 1,
      sym__blank,
    STATE(525), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1364), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [45313] = 20,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(908), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [45401] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1368), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [45489] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1488), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [45577] = 20,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(2307), 1,
      sym__blank,
    STATE(496), 1,
      aux_sym_block_repeat1,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1029), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [45665] = 20,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1485), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [45753] = 20,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1330), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [45841] = 19,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2309), 1,
      anon_sym_COLON,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1438), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [45926] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2313), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2311), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [45979] = 19,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(586), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat1,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1153), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [46064] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2317), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2315), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [46117] = 19,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(586), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat1,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1063), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [46202] = 19,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(586), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat1,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1050), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [46287] = 19,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    STATE(2669), 1,
      sym_hll_range_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1089), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [46372] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2321), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2319), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [46425] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2325), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2323), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [46478] = 19,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(586), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat1,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1052), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [46563] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2329), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2327), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [46616] = 19,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2331), 1,
      anon_sym_COLON,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1432), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [46701] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2335), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2333), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [46754] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2339), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2337), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [46807] = 19,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2341), 1,
      anon_sym_COLON,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1407), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [46892] = 19,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2343), 1,
      anon_sym_COLON,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1457), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [46977] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2090), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2088), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [47030] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2347), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2345), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [47083] = 19,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2349), 1,
      anon_sym_COLON,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1439), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [47168] = 19,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2351), 1,
      anon_sym_COLON,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1441), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [47253] = 19,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    STATE(2480), 1,
      sym_hll_comma_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1264), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [47338] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2339), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2337), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [47391] = 19,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2353), 1,
      anon_sym_COLON,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1437), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [47476] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2357), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2355), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [47529] = 19,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2359), 1,
      anon_sym_COLON,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1436), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [47614] = 19,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2361), 1,
      anon_sym_COLON,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1435), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [47699] = 19,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(586), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat1,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1033), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [47784] = 19,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    STATE(2623), 1,
      sym_hll_range_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1074), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [47869] = 19,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2363), 1,
      anon_sym_COLON,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1440), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [47954] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2335), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2333), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [48007] = 13,
    ACTIONS(2365), 1,
      sym_identifier,
    ACTIONS(2372), 1,
      anon_sym_struct,
    ACTIONS(2375), 1,
      anon_sym_union,
    ACTIONS(2378), 1,
      anon_sym_enum,
    ACTIONS(2384), 1,
      sym_hll_primitive_type,
    ACTIONS(2387), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(1812), 1,
      aux_sym_hll_sized_type_specifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2381), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1620), 6,
      sym__hll_type_specifier,
      sym_hll_struct_specifier,
      sym_hll_union_specifier,
      sym_hll_enum_specifier,
      sym_hll_sized_type_specifier,
      sym__hll_type_identifier,
    ACTIONS(2370), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      anon_sym_const,
      anon_sym_volatile,
      aux_sym_symbol_token2,
    ACTIONS(2368), 19,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [48080] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2392), 21,
      sym_label_identifier,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      aux_sym_while_block_token2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(2390), 24,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [48133] = 19,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    STATE(2674), 1,
      sym_hll_range_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1076), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [48218] = 19,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    STATE(2655), 1,
      sym_hll_range_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1094), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [48303] = 19,
    ACTIONS(2394), 1,
      sym_identifier,
    ACTIONS(2397), 1,
      anon_sym_LPAREN,
    ACTIONS(2412), 1,
      anon_sym_BSLASH,
    ACTIONS(2415), 1,
      anon_sym_sizeof,
    ACTIONS(2424), 1,
      aux_sym_symbol_token1,
    ACTIONS(2427), 1,
      aux_sym_symbol_token2,
    ACTIONS(2430), 1,
      sym__hll_number_literal_pre_hook,
    STATE(586), 1,
      aux_sym__hll_var_command_argument_list_default_format_repeat1,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2400), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2403), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(2406), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2409), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(2418), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(2421), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1355), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [48388] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(952), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [48470] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1019), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [48552] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1045), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [48634] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1021), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [48716] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1022), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [48798] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1057), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [48880] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1023), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [48962] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(937), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [49044] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(935), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [49126] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(945), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [49208] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(965), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [49290] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(966), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [49372] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(919), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [49454] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(921), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [49536] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(946), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [49618] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(940), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [49700] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(947), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [49782] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(953), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [49864] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1389), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [49946] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(958), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [50028] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(960), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [50110] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(923), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [50192] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1390), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [50274] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(961), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [50356] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1359), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [50438] = 8,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2437), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2433), 24,
      anon_sym_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [50500] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2447), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2445), 22,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [50564] = 10,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2453), 12,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2451), 22,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [50630] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(859), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [50712] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(860), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [50794] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(861), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [50876] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(839), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [50958] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(854), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [51040] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(862), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [51122] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(837), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [51204] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(838), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [51286] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(852), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [51368] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(853), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [51450] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(840), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [51532] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2459), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 22,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [51596] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(858), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [51678] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2463), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2461), 22,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [51742] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(845), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [51824] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(846), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [51906] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(835), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [51988] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2467), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2465), 22,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [52052] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2471), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2469), 22,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [52116] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(847), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [52198] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2475), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2473), 22,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [52262] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1219), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [52344] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1220), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [52426] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1222), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [52508] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1223), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [52590] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1225), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [52672] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1226), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [52754] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1282), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [52836] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1020), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [52918] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1159), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53000] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1169), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53082] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1019), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53164] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1183), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53246] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1021), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53328] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1022), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53410] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1214), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53492] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1023), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53574] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(663), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53656] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(664), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53738] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(665), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53820] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(666), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53902] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(667), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [53984] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(668), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [54066] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(669), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [54148] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(626), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [54230] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(670), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [54312] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(671), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [54394] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(628), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [54476] = 11,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 22,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [54544] = 19,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2481), 1,
      anon_sym_PIPE,
    ACTIONS(2483), 1,
      anon_sym_CARET,
    ACTIONS(2491), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2493), 1,
      anon_sym_AMP,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2485), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2487), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2489), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2457), 17,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_DOT_DOT,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [54628] = 10,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 12,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 22,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [54694] = 17,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2483), 1,
      anon_sym_CARET,
    ACTIONS(2493), 1,
      anon_sym_AMP,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2485), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2487), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2489), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2459), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      sym__blank,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2457), 18,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [54774] = 16,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2493), 1,
      anon_sym_AMP,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2485), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2487), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2489), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      sym__blank,
    ACTIONS(2457), 18,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [54852] = 14,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2487), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2489), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 5,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 20,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [54926] = 12,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 22,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [54996] = 18,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2481), 1,
      anon_sym_PIPE,
    ACTIONS(2483), 1,
      anon_sym_CARET,
    ACTIONS(2493), 1,
      anon_sym_AMP,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2485), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2487), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2489), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2457), 18,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [55078] = 15,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2485), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2487), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2489), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 5,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 18,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [55154] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(676), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [55236] = 21,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2481), 1,
      anon_sym_PIPE,
    ACTIONS(2483), 1,
      anon_sym_CARET,
    ACTIONS(2491), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2493), 1,
      anon_sym_AMP,
    ACTIONS(2499), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2501), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2485), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2487), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2489), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2497), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2495), 15,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [55324] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(632), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [55406] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(633), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [55488] = 21,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2481), 1,
      anon_sym_PIPE,
    ACTIONS(2483), 1,
      anon_sym_CARET,
    ACTIONS(2491), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2493), 1,
      anon_sym_AMP,
    ACTIONS(2499), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2501), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2485), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2487), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2489), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2505), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2503), 15,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [55576] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(681), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [55658] = 21,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2481), 1,
      anon_sym_PIPE,
    ACTIONS(2483), 1,
      anon_sym_CARET,
    ACTIONS(2491), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2493), 1,
      anon_sym_AMP,
    ACTIONS(2499), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2501), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2485), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2487), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2489), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2509), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2507), 15,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [55746] = 21,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2481), 1,
      anon_sym_PIPE,
    ACTIONS(2483), 1,
      anon_sym_CARET,
    ACTIONS(2491), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2493), 1,
      anon_sym_AMP,
    ACTIONS(2499), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2501), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2485), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2487), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2489), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2513), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2511), 15,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [55834] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(635), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [55916] = 21,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2481), 1,
      anon_sym_PIPE,
    ACTIONS(2483), 1,
      anon_sym_CARET,
    ACTIONS(2491), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2493), 1,
      anon_sym_AMP,
    ACTIONS(2499), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2501), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2485), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2487), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2489), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2517), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2515), 15,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [56004] = 21,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2481), 1,
      anon_sym_PIPE,
    ACTIONS(2483), 1,
      anon_sym_CARET,
    ACTIONS(2491), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2493), 1,
      anon_sym_AMP,
    ACTIONS(2499), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2501), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2455), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2479), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2485), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2487), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2489), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2521), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2477), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2519), 15,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [56092] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1331), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [56174] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1344), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [56256] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1345), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [56338] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1346), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [56420] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1351), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [56502] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1286), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [56584] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1302), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [56666] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1157), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [56748] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1324), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [56830] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1350), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [56912] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1361), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [56994] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1178), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [57076] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1336), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [57158] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1192), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [57240] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1193), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [57322] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1328), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [57404] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1228), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [57486] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1198), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [57568] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1201), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [57650] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1205), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [57732] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1206), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [57814] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1208), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [57896] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1209), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [57978] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1210), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [58060] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1157), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [58142] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1212), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [58224] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1213), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [58306] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1178), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [58388] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1235), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [58470] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1192), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [58552] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1193), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [58634] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1250), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [58716] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1228), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [58798] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2257), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2259), 28,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym_symbol_token3,
  [58850] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(915), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [58932] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(888), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [59014] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(866), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [59096] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(869), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [59178] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(871), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [59260] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(873), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [59342] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(876), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [59424] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(879), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [59506] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(885), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [59588] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(886), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [59670] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2275), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2277), 28,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym_symbol_token3,
  [59722] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(883), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [59804] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(895), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [59886] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(892), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [59968] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(893), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [60050] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(907), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [60132] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(897), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [60214] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(934), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [60296] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1151), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [60378] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1117), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [60460] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(844), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [60542] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2523), 1,
      anon_sym_LPAREN,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1075), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [60624] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    ACTIONS(2525), 1,
      anon_sym_LPAREN,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1391), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [60706] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1446), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [60788] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1447), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [60870] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1372), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [60952] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1448), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [61034] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1449), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [61116] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1450), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [61198] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1451), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [61280] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1376), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [61362] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1452), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [61444] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1079), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [61526] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1081), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [61608] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1082), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [61690] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1083), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [61772] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1084), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [61854] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1086), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [61936] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1088), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [62018] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1091), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [62100] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1092), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [62182] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1093), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [62264] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1453), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [62346] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1018), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [62428] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1015), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [62510] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(799), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [62592] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    ACTIONS(2527), 1,
      anon_sym_LPAREN,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1061), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [62674] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1377), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [62756] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1454), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [62838] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1455), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [62920] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(954), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63002] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(432), 1,
      anon_sym_LPAREN,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(955), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63084] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(864), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63166] = 18,
    ACTIONS(430), 1,
      sym_identifier,
    ACTIONS(442), 1,
      anon_sym_BSLASH,
    ACTIONS(444), 1,
      anon_sym_sizeof,
    ACTIONS(450), 1,
      aux_sym_symbol_token1,
    ACTIONS(452), 1,
      aux_sym_symbol_token2,
    ACTIONS(454), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2529), 1,
      anon_sym_LPAREN,
    STATE(991), 1,
      sym_trace32_hll_variable,
    STATE(1691), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(434), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(436), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(438), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(446), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(448), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(977), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(957), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63248] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1098), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63330] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(848), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63412] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(380), 1,
      anon_sym_LPAREN,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(849), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63494] = 18,
    ACTIONS(378), 1,
      sym_identifier,
    ACTIONS(390), 1,
      anon_sym_BSLASH,
    ACTIONS(392), 1,
      anon_sym_sizeof,
    ACTIONS(398), 1,
      aux_sym_symbol_token1,
    ACTIONS(400), 1,
      aux_sym_symbol_token2,
    ACTIONS(402), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2531), 1,
      anon_sym_LPAREN,
    STATE(882), 1,
      sym_trace32_hll_variable,
    STATE(1701), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(382), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(384), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(386), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(394), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(396), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(881), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(851), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63576] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1018), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63658] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1015), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63740] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    ACTIONS(2533), 1,
      anon_sym_LPAREN,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1171), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63822] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(612), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63904] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(232), 1,
      anon_sym_LPAREN,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(613), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [63986] = 18,
    ACTIONS(228), 1,
      sym_identifier,
    ACTIONS(244), 1,
      anon_sym_BSLASH,
    ACTIONS(246), 1,
      anon_sym_sizeof,
    ACTIONS(252), 1,
      aux_sym_symbol_token1,
    ACTIONS(254), 1,
      aux_sym_symbol_token2,
    ACTIONS(256), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2535), 1,
      anon_sym_LPAREN,
    STATE(819), 1,
      sym_trace32_hll_variable,
    STATE(1708), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(236), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(238), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(240), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(248), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(250), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(818), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(614), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [64068] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1164), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [64150] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(733), 1,
      anon_sym_LPAREN,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1166), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [64232] = 18,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(731), 1,
      sym_identifier,
    ACTIONS(741), 1,
      anon_sym_sizeof,
    ACTIONS(2537), 1,
      anon_sym_LPAREN,
    STATE(944), 1,
      sym_trace32_hll_variable,
    STATE(1710), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(735), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(737), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(739), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(942), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1327), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [64314] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1102), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [64396] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1164), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [64478] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(667), 1,
      anon_sym_LPAREN,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1166), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [64560] = 18,
    ACTIONS(665), 1,
      sym_identifier,
    ACTIONS(677), 1,
      anon_sym_BSLASH,
    ACTIONS(679), 1,
      anon_sym_sizeof,
    ACTIONS(685), 1,
      aux_sym_symbol_token1,
    ACTIONS(687), 1,
      aux_sym_symbol_token2,
    ACTIONS(689), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2539), 1,
      anon_sym_LPAREN,
    STATE(939), 1,
      sym_trace32_hll_variable,
    STATE(1712), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(669), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(671), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(673), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(681), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(683), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(938), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1217), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [64642] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(867), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [64724] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_LPAREN,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(868), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [64806] = 18,
    ACTIONS(404), 1,
      sym_identifier,
    ACTIONS(416), 1,
      anon_sym_BSLASH,
    ACTIONS(418), 1,
      anon_sym_sizeof,
    ACTIONS(424), 1,
      aux_sym_symbol_token1,
    ACTIONS(426), 1,
      aux_sym_symbol_token2,
    ACTIONS(428), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(2541), 1,
      anon_sym_LPAREN,
    STATE(930), 1,
      sym_trace32_hll_variable,
    STATE(1716), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(408), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(410), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(412), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(675), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(420), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(422), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(928), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(870), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [64888] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1406), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [64970] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1106), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [65052] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1109), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [65134] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1107), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [65216] = 18,
    ACTIONS(627), 1,
      sym_identifier,
    ACTIONS(629), 1,
      anon_sym_LPAREN,
    ACTIONS(639), 1,
      anon_sym_BSLASH,
    ACTIONS(641), 1,
      anon_sym_sizeof,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    STATE(891), 1,
      sym_trace32_hll_variable,
    STATE(1724), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(631), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(633), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(635), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(637), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(890), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1110), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [65298] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1431), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [65380] = 18,
    ACTIONS(647), 1,
      aux_sym_symbol_token1,
    ACTIONS(649), 1,
      aux_sym_symbol_token2,
    ACTIONS(651), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(993), 1,
      sym_identifier,
    ACTIONS(995), 1,
      anon_sym_LPAREN,
    ACTIONS(1005), 1,
      anon_sym_BSLASH,
    ACTIONS(1007), 1,
      anon_sym_sizeof,
    STATE(990), 1,
      sym_trace32_hll_variable,
    STATE(1687), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(997), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(999), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(1001), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1003), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(643), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(645), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(974), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1385), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [65462] = 18,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(653), 1,
      sym_identifier,
    ACTIONS(655), 1,
      anon_sym_LPAREN,
    ACTIONS(663), 1,
      anon_sym_sizeof,
    STATE(914), 1,
      sym_trace32_hll_variable,
    STATE(1705), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(657), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(659), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(661), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(913), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1434), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [65544] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2449), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2545), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2543), 22,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [65608] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1066), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [65690] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1067), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [65772] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1068), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [65854] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1069), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [65936] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1070), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [66018] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1071), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [66100] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1065), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [66182] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1020), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [66264] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1053), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [66346] = 18,
    ACTIONS(474), 1,
      anon_sym_LPAREN,
    ACTIONS(484), 1,
      anon_sym_BSLASH,
    ACTIONS(486), 1,
      anon_sym_sizeof,
    ACTIONS(504), 1,
      aux_sym_symbol_token1,
    ACTIONS(506), 1,
      aux_sym_symbol_token2,
    ACTIONS(510), 1,
      sym__hll_number_literal_pre_hook,
    ACTIONS(512), 1,
      sym_identifier,
    STATE(857), 1,
      sym_trace32_hll_variable,
    STATE(1681), 1,
      sym__hll_assignment_left_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(476), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(478), 2,
      anon_sym_TILDE,
      anon_sym_BANG,
    ACTIONS(480), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(482), 2,
      anon_sym_STAR,
      anon_sym_AMP,
    ACTIONS(488), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(490), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
    STATE(856), 5,
      sym_hll_call_expression,
      sym_hll_field_expression,
      sym_hll_subscript_expression,
      sym_hll_pointer_expression,
      sym_hll_parenthesized_expression,
    STATE(1064), 11,
      sym__hll_expression,
      sym_hll_cast_expression,
      sym_hll_sizeof_expression,
      sym_hll_update_expression,
      sym_hll_unary_expression,
      sym_hll_binary_expression,
      sym_hll_assignment_expression,
      sym_hll_conditional_expression,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_symbol,
  [66428] = 4,
    ACTIONS(2547), 1,
      aux_sym_symbol_token3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2279), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2281), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [66482] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2551), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2549), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [66533] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2555), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2553), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [66584] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2559), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2557), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [66635] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2563), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2561), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [66686] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2311), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2313), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [66737] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2567), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2565), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [66788] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2571), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2569), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [66839] = 5,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2575), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2580), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
    ACTIONS(2573), 17,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [66894] = 5,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2575), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2580), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
    ACTIONS(2573), 17,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [66949] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2585), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2583), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67000] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2589), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2587), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67051] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2593), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2591), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67102] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2597), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2595), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67153] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2390), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2392), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67204] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2601), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2599), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67255] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2605), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2603), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67306] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2609), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2607), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67357] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2613), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2611), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67408] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2617), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2615), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67459] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2319), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2321), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67510] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2621), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2619), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67561] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2625), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2623), 27,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [67612] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2633), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2637), 1,
      anon_sym_PIPE,
    ACTIONS(2639), 1,
      anon_sym_CARET,
    ACTIONS(2649), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2651), 1,
      anon_sym_AMP,
    ACTIONS(2655), 1,
      anon_sym_QMARK,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2641), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2643), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2645), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2509), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2507), 13,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [67696] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2275), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2277), 27,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      aux_sym_symbol_token3,
      sym_comment,
      sym__blank_line,
  [67746] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2633), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2637), 1,
      anon_sym_PIPE,
    ACTIONS(2639), 1,
      anon_sym_CARET,
    ACTIONS(2649), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2651), 1,
      anon_sym_AMP,
    ACTIONS(2655), 1,
      anon_sym_QMARK,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2641), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2643), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2645), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2517), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2515), 13,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [67830] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2633), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2637), 1,
      anon_sym_PIPE,
    ACTIONS(2639), 1,
      anon_sym_CARET,
    ACTIONS(2649), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2651), 1,
      anon_sym_AMP,
    ACTIONS(2655), 1,
      anon_sym_QMARK,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2641), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2643), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2645), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2521), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2519), 13,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [67914] = 11,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 20,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [67980] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2459), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 20,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68040] = 16,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2639), 1,
      anon_sym_CARET,
    ACTIONS(2651), 1,
      anon_sym_AMP,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2641), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2643), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2645), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68116] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2463), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2461), 20,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68176] = 9,
    ACTIONS(2662), 1,
      anon_sym_EQ,
    ACTIONS(2664), 1,
      anon_sym_STAR,
    ACTIONS(2671), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2669), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(2659), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 11,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
    ACTIONS(2573), 12,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
  [68238] = 5,
    ACTIONS(2387), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2370), 17,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      anon_sym_const,
      anon_sym_volatile,
      anon_sym_struct,
      anon_sym_union,
      anon_sym_enum,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_hll_primitive_type,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(2368), 22,
      sym__hll_number_literal_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_RBRACK,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [68292] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2675), 1,
      aux_sym_symbol_token3,
    ACTIONS(2279), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2281), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68344] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2545), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2543), 20,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68404] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2467), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2465), 20,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68464] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2471), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2469), 20,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68524] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2475), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2473), 20,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68584] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2437), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2433), 22,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68642] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2447), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2445), 20,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68702] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2257), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2259), 27,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      aux_sym_symbol_token3,
      sym_comment,
      sym__blank_line,
  [68752] = 9,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2453), 13,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2451), 20,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68814] = 17,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2637), 1,
      anon_sym_PIPE,
    ACTIONS(2639), 1,
      anon_sym_CARET,
    ACTIONS(2651), 1,
      anon_sym_AMP,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2641), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2643), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2645), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2459), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2457), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68892] = 14,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2641), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2643), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2645), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [68964] = 15,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2651), 1,
      anon_sym_AMP,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2641), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2643), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2645), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 5,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [69038] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2633), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2637), 1,
      anon_sym_PIPE,
    ACTIONS(2639), 1,
      anon_sym_CARET,
    ACTIONS(2649), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2651), 1,
      anon_sym_AMP,
    ACTIONS(2655), 1,
      anon_sym_QMARK,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2641), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2643), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2645), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2497), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2495), 13,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [69122] = 6,
    ACTIONS(2575), 1,
      sym__blank,
    ACTIONS(2662), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
    ACTIONS(2573), 16,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [69178] = 6,
    ACTIONS(2575), 1,
      sym__blank,
    ACTIONS(2662), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_DOT,
    ACTIONS(2573), 16,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [69234] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2633), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2637), 1,
      anon_sym_PIPE,
    ACTIONS(2639), 1,
      anon_sym_CARET,
    ACTIONS(2649), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2651), 1,
      anon_sym_AMP,
    ACTIONS(2655), 1,
      anon_sym_QMARK,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2641), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2643), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2645), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2505), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2503), 13,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [69318] = 10,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 10,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 20,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [69382] = 18,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2637), 1,
      anon_sym_PIPE,
    ACTIONS(2639), 1,
      anon_sym_CARET,
    ACTIONS(2649), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2651), 1,
      anon_sym_AMP,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2641), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2643), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2645), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2459), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2457), 15,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [69462] = 9,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2459), 13,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 20,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [69524] = 13,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2643), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2645), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [69594] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2633), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2637), 1,
      anon_sym_PIPE,
    ACTIONS(2639), 1,
      anon_sym_CARET,
    ACTIONS(2649), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2651), 1,
      anon_sym_AMP,
    ACTIONS(2655), 1,
      anon_sym_QMARK,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2629), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2631), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2641), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2643), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2645), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2647), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2513), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2635), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2511), 13,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [69678] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2545), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2543), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [69737] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2390), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2392), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [69786] = 9,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 13,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [69847] = 7,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2437), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2433), 20,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [69904] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2447), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2445), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [69963] = 16,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2691), 1,
      anon_sym_CARET,
    ACTIONS(2699), 1,
      anon_sym_AMP,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2693), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2695), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2697), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [70038] = 9,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2453), 13,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2451), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [70099] = 15,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2699), 1,
      anon_sym_AMP,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2693), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2695), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2697), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 5,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [70172] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2621), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2619), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [70221] = 13,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2695), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2697), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 16,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [70290] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2319), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2321), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [70339] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2311), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2313), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [70388] = 11,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [70453] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2625), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2623), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [70502] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2551), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2549), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [70551] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2459), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [70610] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2585), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2583), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [70659] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2575), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2580), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2573), 16,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [70712] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2575), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2580), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2573), 16,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [70765] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2463), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2461), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [70824] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2563), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2561), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [70873] = 17,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2691), 1,
      anon_sym_CARET,
    ACTIONS(2699), 1,
      anon_sym_AMP,
    ACTIONS(2701), 1,
      anon_sym_PIPE,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2693), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2695), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2697), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2459), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2457), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [70950] = 14,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2693), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2695), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2697), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [71021] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2617), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2615), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [71070] = 18,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2691), 1,
      anon_sym_CARET,
    ACTIONS(2699), 1,
      anon_sym_AMP,
    ACTIONS(2701), 1,
      anon_sym_PIPE,
    ACTIONS(2703), 1,
      anon_sym_AMP_AMP,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2693), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2695), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2697), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2459), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2457), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [71149] = 20,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2691), 1,
      anon_sym_CARET,
    ACTIONS(2699), 1,
      anon_sym_AMP,
    ACTIONS(2701), 1,
      anon_sym_PIPE,
    ACTIONS(2703), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2705), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2707), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2693), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2695), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2697), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2497), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2495), 11,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [71232] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2575), 1,
      sym__blank,
    ACTIONS(2662), 1,
      anon_sym_EQ,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2573), 16,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [71287] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2575), 1,
      sym__blank,
    ACTIONS(2662), 1,
      anon_sym_EQ,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2573), 16,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [71342] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2467), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2465), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [71401] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2471), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2469), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [71460] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2589), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2587), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [71509] = 20,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2691), 1,
      anon_sym_CARET,
    ACTIONS(2699), 1,
      anon_sym_AMP,
    ACTIONS(2701), 1,
      anon_sym_PIPE,
    ACTIONS(2703), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2705), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2707), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2693), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2695), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2697), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2505), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2503), 11,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [71592] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2593), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2591), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [71641] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2475), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2473), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [71700] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2597), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2595), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [71749] = 20,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2691), 1,
      anon_sym_CARET,
    ACTIONS(2699), 1,
      anon_sym_AMP,
    ACTIONS(2701), 1,
      anon_sym_PIPE,
    ACTIONS(2703), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2705), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2707), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2693), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2695), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2697), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2509), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2507), 11,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [71832] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2601), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2599), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [71881] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2605), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2603), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [71930] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2609), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2607), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [71979] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2559), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2557), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [72028] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2567), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2565), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [72077] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2571), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2569), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [72126] = 20,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2691), 1,
      anon_sym_CARET,
    ACTIONS(2699), 1,
      anon_sym_AMP,
    ACTIONS(2701), 1,
      anon_sym_PIPE,
    ACTIONS(2703), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2705), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2707), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2693), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2695), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2697), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2513), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2511), 11,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [72209] = 20,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2691), 1,
      anon_sym_CARET,
    ACTIONS(2699), 1,
      anon_sym_AMP,
    ACTIONS(2701), 1,
      anon_sym_PIPE,
    ACTIONS(2703), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2705), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2707), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2693), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2695), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2697), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2517), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2515), 11,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [72292] = 20,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2691), 1,
      anon_sym_CARET,
    ACTIONS(2699), 1,
      anon_sym_AMP,
    ACTIONS(2701), 1,
      anon_sym_PIPE,
    ACTIONS(2703), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2705), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2707), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2687), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2693), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2695), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2697), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2521), 3,
      anon_sym_EQ,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2519), 11,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [72375] = 4,
    ACTIONS(2709), 1,
      aux_sym_symbol_token3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2279), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2281), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [72426] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2257), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2259), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      aux_sym_symbol_token3,
  [72475] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2275), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2277), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      aux_sym_symbol_token3,
  [72524] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2613), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2611), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [72573] = 5,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2662), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 12,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
    ACTIONS(2573), 16,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [72626] = 5,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2662), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 12,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
    ACTIONS(2573), 16,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [72679] = 10,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2679), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2685), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2689), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 10,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 18,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [72742] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2555), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2553), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [72791] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2617), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2615), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [72839] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2555), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2553), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [72887] = 13,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2719), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2721), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 5,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 17,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [72955] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2275), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2277), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym_symbol_token3,
      sym_comment,
      sym__blank_line,
  [73003] = 11,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 19,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [73067] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2559), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2557), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [73115] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(2729), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2731), 1,
      anon_sym_PIPE,
    ACTIONS(2733), 1,
      anon_sym_CARET,
    ACTIONS(2737), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2739), 1,
      anon_sym_AMP,
    ACTIONS(2741), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2517), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2719), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2721), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2735), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2515), 12,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [73197] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(2729), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2731), 1,
      anon_sym_PIPE,
    ACTIONS(2733), 1,
      anon_sym_CARET,
    ACTIONS(2737), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2739), 1,
      anon_sym_AMP,
    ACTIONS(2741), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2521), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2719), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2721), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2735), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2519), 12,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [73279] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2567), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2565), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [73327] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2571), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2569), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [73375] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2390), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2392), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [73423] = 5,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2575), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2580), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2573), 14,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [73475] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2319), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2321), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [73523] = 5,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2575), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2580), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2573), 14,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [73575] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2311), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2313), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [73623] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2551), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2549), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [73671] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(2729), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2731), 1,
      anon_sym_PIPE,
    ACTIONS(2733), 1,
      anon_sym_CARET,
    ACTIONS(2737), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2739), 1,
      anon_sym_AMP,
    ACTIONS(2741), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2497), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2719), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2721), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2735), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2495), 12,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [73753] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2545), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2543), 19,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [73811] = 18,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(2731), 1,
      anon_sym_PIPE,
    ACTIONS(2733), 1,
      anon_sym_CARET,
    ACTIONS(2737), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2739), 1,
      anon_sym_AMP,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2459), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2719), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2721), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2735), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2457), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [73889] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2597), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2595), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [73937] = 10,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 19,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [73999] = 6,
    ACTIONS(2575), 1,
      sym__blank,
    ACTIONS(2662), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2573), 14,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [74053] = 6,
    ACTIONS(2575), 1,
      sym__blank,
    ACTIONS(2662), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2573), 14,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [74107] = 17,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(2731), 1,
      anon_sym_PIPE,
    ACTIONS(2733), 1,
      anon_sym_CARET,
    ACTIONS(2739), 1,
      anon_sym_AMP,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2459), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2719), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2721), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2735), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2457), 15,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [74183] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2257), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2259), 26,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym_symbol_token3,
      sym_comment,
      sym__blank_line,
  [74231] = 5,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2662), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2573), 14,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [74283] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2585), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2583), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [74331] = 5,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2662), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 13,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2573), 14,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [74383] = 9,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2459), 12,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 19,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [74443] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2459), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 19,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [74501] = 14,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2719), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2721), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2735), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 5,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 15,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [74571] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2563), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2561), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [74619] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2601), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2599), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [74667] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2589), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2587), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [74715] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2593), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2591), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [74763] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(2729), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2731), 1,
      anon_sym_PIPE,
    ACTIONS(2733), 1,
      anon_sym_CARET,
    ACTIONS(2737), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2739), 1,
      anon_sym_AMP,
    ACTIONS(2741), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2505), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2719), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2721), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2735), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2503), 12,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [74845] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2463), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2461), 19,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [74903] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2437), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2433), 21,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [74959] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2447), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2445), 19,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75017] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2621), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2619), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [75065] = 9,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2453), 12,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2451), 19,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75125] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2467), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2465), 19,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75183] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2625), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2623), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [75231] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2471), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2469), 19,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75289] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2475), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2473), 19,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75347] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2605), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2603), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [75395] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(2729), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2731), 1,
      anon_sym_PIPE,
    ACTIONS(2733), 1,
      anon_sym_CARET,
    ACTIONS(2737), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2739), 1,
      anon_sym_AMP,
    ACTIONS(2741), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2509), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2719), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2721), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2735), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2507), 12,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75477] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(2729), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2731), 1,
      anon_sym_PIPE,
    ACTIONS(2733), 1,
      anon_sym_CARET,
    ACTIONS(2737), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2739), 1,
      anon_sym_AMP,
    ACTIONS(2741), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2513), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2719), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2721), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2735), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2511), 12,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75559] = 16,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(2733), 1,
      anon_sym_CARET,
    ACTIONS(2739), 1,
      anon_sym_AMP,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2719), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2721), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2735), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2459), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      sym__blank,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2457), 15,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75633] = 15,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(2739), 1,
      anon_sym_AMP,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2713), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2715), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2719), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2721), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2723), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2735), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2717), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2459), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      sym__blank,
    ACTIONS(2457), 15,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75705] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2613), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2611), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [75753] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2609), 15,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2607), 24,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      aux_sym__hll_expression_separator_token1,
  [75801] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2743), 1,
      aux_sym_symbol_token3,
    ACTIONS(2279), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2281), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75851] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2319), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2321), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75898] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2390), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2392), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75945] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2597), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2595), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [75992] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2621), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2619), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76039] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2662), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 12,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
    ACTIONS(2573), 15,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      sym_comment,
      sym__blank_line,
  [76090] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2609), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2607), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76137] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2605), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2603), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76184] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2575), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2580), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 12,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
    ACTIONS(2573), 15,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      sym_comment,
      sym__blank_line,
  [76235] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2559), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2557), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76282] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2563), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2561), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76329] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2585), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2583), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76376] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2589), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2587), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76423] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2567), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2565), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76470] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2593), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2591), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76517] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2625), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2623), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76564] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2555), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2553), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76611] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2551), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2549), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76658] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2311), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2313), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76705] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2571), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2569), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76752] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2613), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2611), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76799] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2662), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2667), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 12,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
    ACTIONS(2573), 15,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      sym_comment,
      sym__blank_line,
  [76850] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2575), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2580), 10,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
    ACTIONS(2578), 12,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
    ACTIONS(2573), 15,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      sym_comment,
      sym__blank_line,
  [76901] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2601), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2599), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76948] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2617), 14,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2615), 25,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
      sym_comment,
      sym__blank_line,
  [76995] = 8,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    ACTIONS(470), 1,
      sym__blank,
    STATE(1002), 1,
      aux_sym_block_repeat1,
    STATE(2098), 1,
      sym_option_expression,
    STATE(2376), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2745), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2747), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77051] = 9,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    ACTIONS(470), 1,
      sym__blank,
    ACTIONS(2749), 1,
      anon_sym_COMMA,
    STATE(1002), 1,
      aux_sym_block_repeat1,
    STATE(2196), 1,
      sym_option_expression,
    STATE(2436), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2745), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2747), 17,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77109] = 8,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    ACTIONS(470), 1,
      sym__blank,
    STATE(1002), 1,
      aux_sym_block_repeat1,
    STATE(2196), 1,
      sym_option_expression,
    STATE(2436), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2745), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2747), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77165] = 8,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    ACTIONS(470), 1,
      sym__blank,
    STATE(1002), 1,
      aux_sym_block_repeat1,
    STATE(2098), 1,
      sym_option_expression,
    STATE(2376), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2752), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2754), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77221] = 9,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    ACTIONS(470), 1,
      sym__blank,
    ACTIONS(2749), 1,
      anon_sym_COMMA,
    STATE(1002), 1,
      aux_sym_block_repeat1,
    STATE(2098), 1,
      sym_option_expression,
    STATE(2376), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2745), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2747), 17,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77279] = 8,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    ACTIONS(470), 1,
      sym__blank,
    STATE(1002), 1,
      aux_sym_block_repeat1,
    STATE(2201), 1,
      sym_option_expression,
    STATE(2444), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2752), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2754), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77335] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2756), 1,
      sym__blank,
    STATE(1000), 1,
      aux_sym_block_repeat1,
    ACTIONS(2370), 15,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2368), 20,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
      sym_comment,
      sym__blank_line,
  [77384] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2765), 1,
      sym__blank,
    STATE(996), 1,
      aux_sym_block_repeat1,
    ACTIONS(2763), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(2759), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2761), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77434] = 5,
    ACTIONS(2767), 1,
      sym__blank,
    STATE(1002), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2370), 15,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2368), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77482] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2772), 1,
      sym__blank,
    STATE(994), 1,
      aux_sym_block_repeat1,
    ACTIONS(2770), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(2759), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2761), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77532] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2774), 1,
      sym__blank,
    STATE(997), 1,
      aux_sym_block_repeat1,
    ACTIONS(2770), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(2745), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2747), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77582] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2778), 1,
      sym__blank,
    STATE(999), 1,
      aux_sym_block_repeat1,
    ACTIONS(2776), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(2745), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2747), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77632] = 5,
    ACTIONS(2780), 1,
      sym__blank,
    STATE(1008), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2752), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2754), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77679] = 5,
    ACTIONS(2782), 1,
      sym__blank,
    STATE(1009), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2759), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2761), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77726] = 5,
    ACTIONS(470), 1,
      sym__blank,
    STATE(1002), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2784), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2786), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77773] = 5,
    ACTIONS(470), 1,
      sym__blank,
    STATE(1002), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2745), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2747), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77820] = 5,
    ACTIONS(2788), 1,
      sym__blank,
    STATE(1012), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2745), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2747), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77867] = 6,
    ACTIONS(470), 1,
      sym__blank,
    ACTIONS(2749), 1,
      anon_sym_COMMA,
    STATE(1002), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2745), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2747), 17,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77916] = 5,
    ACTIONS(470), 1,
      sym__blank,
    STATE(1002), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2752), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2754), 18,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_COMMA,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [77963] = 5,
    ACTIONS(2790), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2370), 14,
      aux_sym_if_block_token1,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_const,
      anon_sym_volatile,
      anon_sym_struct,
      anon_sym_union,
      anon_sym_enum,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_hll_primitive_type,
      sym_identifier,
    ACTIONS(2368), 16,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [78008] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2257), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2259), 23,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_symbol_token3,
  [78049] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2447), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2445), 17,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [78102] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2275), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2277), 23,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      aux_sym_symbol_token3,
  [78143] = 4,
    ACTIONS(2795), 1,
      aux_sym_symbol_token3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2279), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2281), 22,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [78186] = 8,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2437), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2433), 19,
      anon_sym_RPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [78237] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2463), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2461), 17,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [78290] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2459), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 17,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [78343] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2467), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2465), 17,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [78396] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2471), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2469), 17,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [78449] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2475), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2473), 17,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [78502] = 7,
    ACTIONS(2804), 1,
      sym__blank,
    STATE(2013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2801), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1024), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(2797), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(2799), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_RBRACK,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [78551] = 24,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2807), 1,
      anon_sym_RPAREN,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2837), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(1989), 1,
      aux_sym_block_repeat1,
    STATE(2250), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [78633] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2589), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2587), 22,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [78673] = 24,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    ACTIONS(2867), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2869), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2873), 1,
      sym__blank,
    STATE(872), 1,
      sym_hll_argument_list,
    STATE(1135), 1,
      aux_sym_block_repeat1,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2871), 2,
      sym_comment,
      sym__blank_line,
  [78755] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2593), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2591), 22,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [78795] = 22,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2497), 1,
      sym__blank,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2495), 4,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [78873] = 8,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1925), 1,
      sym_path,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1905), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 13,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [78923] = 24,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2875), 1,
      anon_sym_RPAREN,
    ACTIONS(2877), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2020), 1,
      aux_sym_block_repeat1,
    STATE(2390), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [79005] = 8,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(2248), 1,
      sym_option_expression,
    STATE(2251), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2879), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(2881), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [79055] = 24,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    ACTIONS(2867), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2869), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2885), 1,
      sym__blank,
    STATE(872), 1,
      sym_hll_argument_list,
    STATE(1036), 1,
      aux_sym_block_repeat1,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2883), 2,
      sym_comment,
      sym__blank_line,
  [79137] = 24,
    ACTIONS(2067), 1,
      anon_sym_RPAREN,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2887), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(1993), 1,
      aux_sym_block_repeat1,
    STATE(2296), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [79219] = 22,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2513), 1,
      sym__blank,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2511), 4,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [79297] = 8,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(2463), 1,
      sym_option_expression,
    STATE(2464), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2889), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(2891), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [79347] = 24,
    ACTIONS(2198), 1,
      anon_sym_RPAREN,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2893), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(1928), 1,
      aux_sym_block_repeat1,
    STATE(2441), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [79429] = 24,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2895), 1,
      anon_sym_RPAREN,
    ACTIONS(2897), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(1977), 1,
      aux_sym_block_repeat1,
    STATE(2445), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [79511] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2613), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2611), 22,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [79551] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2617), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2615), 22,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [79591] = 24,
    ACTIONS(2140), 1,
      anon_sym_RPAREN,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2899), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(1988), 1,
      aux_sym_block_repeat1,
    STATE(2247), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [79673] = 8,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(2248), 1,
      sym_option_expression,
    STATE(2251), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2889), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(2891), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [79723] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2390), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2392), 22,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [79763] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1921), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1923), 20,
      sym__and_operator_pre_hook,
      sym_path,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [79803] = 22,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2505), 1,
      sym__blank,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2503), 4,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [79881] = 22,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2509), 1,
      sym__blank,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2507), 4,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [79959] = 24,
    ACTIONS(2105), 1,
      anon_sym_RPAREN,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2901), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(1948), 1,
      aux_sym_block_repeat1,
    STATE(2412), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [80041] = 24,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2903), 1,
      anon_sym_RPAREN,
    ACTIONS(2905), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(1955), 1,
      aux_sym_block_repeat1,
    STATE(2225), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [80123] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2585), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2583), 22,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [80163] = 24,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    ACTIONS(2867), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2869), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2909), 1,
      sym__blank,
    STATE(872), 1,
      sym_hll_argument_list,
    STATE(1120), 1,
      aux_sym_block_repeat1,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2907), 2,
      sym_comment,
      sym__blank_line,
  [80245] = 8,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(2457), 1,
      sym_option_expression,
    STATE(2458), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2879), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(2881), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [80295] = 24,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    ACTIONS(2867), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2869), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2913), 1,
      sym__blank,
    STATE(872), 1,
      sym_hll_argument_list,
    STATE(1042), 1,
      aux_sym_block_repeat1,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2911), 2,
      sym_comment,
      sym__blank_line,
  [80377] = 19,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2459), 1,
      sym__blank,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 7,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [80449] = 8,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1925), 1,
      sym_path,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1905), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 11,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [80499] = 24,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    ACTIONS(2867), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2869), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2915), 1,
      sym__blank,
    STATE(872), 1,
      sym_hll_argument_list,
    STATE(1032), 1,
      aux_sym_block_repeat1,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2911), 2,
      sym_comment,
      sym__blank_line,
  [80581] = 24,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    ACTIONS(2867), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2869), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2919), 1,
      sym__blank,
    STATE(872), 1,
      sym_hll_argument_list,
    STATE(1051), 1,
      aux_sym_block_repeat1,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2917), 2,
      sym_comment,
      sym__blank_line,
  [80663] = 22,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2517), 1,
      sym__blank,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2515), 4,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [80741] = 24,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    ACTIONS(2867), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2869), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2923), 1,
      sym__blank,
    STATE(872), 1,
      sym_hll_argument_list,
    STATE(1137), 1,
      aux_sym_block_repeat1,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2921), 2,
      sym_comment,
      sym__blank_line,
  [80823] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1901), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1903), 20,
      sym__and_operator_pre_hook,
      sym_path,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [80863] = 6,
    ACTIONS(1925), 1,
      sym_path,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1917), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1909), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1912), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [80909] = 10,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2453), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2451), 14,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [80963] = 22,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2521), 1,
      sym__blank,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2519), 4,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [81041] = 24,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    ACTIONS(2867), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2869), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2925), 1,
      sym__blank,
    STATE(872), 1,
      sym_hll_argument_list,
    STATE(1123), 1,
      aux_sym_block_repeat1,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2871), 2,
      sym_comment,
      sym__blank_line,
  [81123] = 16,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 3,
      anon_sym_PIPE,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 8,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [81189] = 13,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 5,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 12,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [81249] = 12,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 12,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [81307] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2459), 1,
      sym__blank,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 6,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [81381] = 10,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 14,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [81435] = 18,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2459), 2,
      anon_sym_PIPE,
      sym__blank,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 7,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [81505] = 17,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2459), 2,
      anon_sym_PIPE,
      sym__blank,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 8,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [81573] = 15,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 3,
      anon_sym_PIPE,
      anon_sym_AMP,
      sym__blank,
    ACTIONS(2457), 10,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [81637] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2319), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      anon_sym_DOT,
      sym__blank,
    ACTIONS(2321), 22,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [81677] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1961), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1963), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [81716] = 24,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2927), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2929), 1,
      anon_sym_RBRACK,
    ACTIONS(2931), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2374), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [81797] = 9,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2453), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2451), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [81848] = 24,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2927), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2931), 1,
      sym__blank,
    ACTIONS(2933), 1,
      anon_sym_RBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2374), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [81929] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2935), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2938), 16,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [81968] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2941), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2944), 16,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [82007] = 11,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 11,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [82062] = 23,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2947), 1,
      anon_sym_RPAREN,
    ACTIONS(2949), 1,
      anon_sym_COMMA,
    ACTIONS(2951), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2233), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [82141] = 19,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2459), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [82212] = 9,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [82263] = 17,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 3,
      anon_sym_PIPE,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 6,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [82330] = 16,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 3,
      anon_sym_PIPE,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 7,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [82395] = 23,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2949), 1,
      anon_sym_COMMA,
    ACTIONS(2953), 1,
      anon_sym_RPAREN,
    ACTIONS(2955), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2242), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [82474] = 14,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 9,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [82535] = 23,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2949), 1,
      anon_sym_COMMA,
    ACTIONS(2957), 1,
      anon_sym_RPAREN,
    ACTIONS(2959), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2394), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [82614] = 12,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 6,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 11,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [82671] = 24,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2927), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2931), 1,
      sym__blank,
    ACTIONS(2961), 1,
      anon_sym_RBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2374), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [82752] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1925), 1,
      sym_path,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1917), 4,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1909), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1912), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [82797] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2459), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 15,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [82846] = 18,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2459), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 6,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [82915] = 15,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 7,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [82978] = 24,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2439), 1,
      anon_sym_DOT,
    ACTIONS(2441), 1,
      anon_sym_DASH_GT,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2927), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2931), 1,
      sym__blank,
    ACTIONS(2963), 1,
      anon_sym_RBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2374), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [83059] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2497), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2495), 3,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [83134] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2033), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(2035), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [83173] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(2965), 1,
      sym_path,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1905), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 13,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [83222] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2463), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2461), 15,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [83271] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1949), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1951), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [83310] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2275), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2277), 22,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      aux_sym_symbol_token3,
      sym_comment,
      sym__blank_line,
  [83349] = 23,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2949), 1,
      anon_sym_COMMA,
    ACTIONS(2967), 1,
      anon_sym_RPAREN,
    ACTIONS(2969), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2232), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [83428] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2505), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2503), 3,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [83503] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2509), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2507), 3,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [83578] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2513), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2511), 3,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [83653] = 8,
    ACTIONS(1925), 1,
      sym_path,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1917), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1905), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 10,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [83702] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2467), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2465), 15,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [83751] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2517), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2515), 3,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [83826] = 21,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    ACTIONS(2843), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2847), 1,
      anon_sym_SLASH,
    ACTIONS(2849), 1,
      anon_sym_PIPE,
    ACTIONS(2851), 1,
      anon_sym_CARET,
    ACTIONS(2861), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2863), 1,
      anon_sym_AMP,
    ACTIONS(2865), 1,
      anon_sym_QMARK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2521), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2839), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2845), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2853), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2855), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2857), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2859), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2519), 3,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [83901] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2471), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2469), 15,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [83950] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2475), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2473), 15,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [83999] = 7,
    ACTIONS(2977), 1,
      anon_sym_SLASH,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2975), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2973), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1961), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1963), 14,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [84046] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1945), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1947), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [84085] = 8,
    ACTIONS(2977), 1,
      anon_sym_SLASH,
    ACTIONS(2979), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2975), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2973), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1961), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1963), 14,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [84134] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2981), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(2983), 16,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [84173] = 7,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1905), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 11,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [84220] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2047), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(2049), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [84259] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2447), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2445), 15,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [84308] = 7,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1935), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1939), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1942), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1937), 11,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [84355] = 6,
    ACTIONS(154), 1,
      anon_sym_AMP,
    STATE(1131), 1,
      aux_sym__path_expression_repeat2,
    STATE(1297), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2099), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2101), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [84400] = 7,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(2985), 1,
      anon_sym_SLASH,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(2643), 1,
      sym_option_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2889), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(2891), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [84447] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1927), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1929), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [84486] = 22,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2989), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2089), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2987), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [84563] = 7,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(2985), 1,
      anon_sym_SLASH,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(2652), 1,
      sym_option_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2889), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(2891), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [84610] = 12,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2977), 1,
      anon_sym_SLASH,
    ACTIONS(2979), 1,
      anon_sym_CARET,
    ACTIONS(2991), 1,
      anon_sym_PIPE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2975), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2973), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1975), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1977), 13,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [84667] = 23,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2867), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2869), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2997), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    ACTIONS(3003), 1,
      anon_sym_PIPE,
    ACTIONS(3005), 1,
      anon_sym_CARET,
    ACTIONS(3015), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3017), 1,
      anon_sym_AMP,
    ACTIONS(3019), 1,
      anon_sym_QMARK,
    ACTIONS(3021), 1,
      sym__blank,
    STATE(956), 1,
      sym_hll_argument_list,
    STATE(1314), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [84746] = 7,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1905), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 13,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [84793] = 7,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1935), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1939), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1942), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1937), 13,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [84840] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2055), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(2057), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [84879] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1931), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1933), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [84918] = 12,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2977), 1,
      anon_sym_SLASH,
    ACTIONS(2979), 1,
      anon_sym_CARET,
    ACTIONS(2991), 1,
      anon_sym_PIPE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2975), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2973), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(2059), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2061), 13,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [84975] = 6,
    ACTIONS(3023), 1,
      anon_sym_AMP,
    STATE(1131), 1,
      aux_sym__path_expression_repeat2,
    STATE(2708), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2088), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2090), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [85020] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2257), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2259), 22,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      aux_sym_symbol_token3,
      sym_comment,
      sym__blank_line,
  [85059] = 5,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1917), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1909), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1912), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [85102] = 5,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1917), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1939), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1942), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [85145] = 7,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(2985), 1,
      anon_sym_SLASH,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(2652), 1,
      sym_option_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2879), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(2881), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [85192] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1921), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1923), 20,
      sym__and_operator_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [85231] = 7,
    ACTIONS(1857), 1,
      sym__blank,
    ACTIONS(2985), 1,
      anon_sym_SLASH,
    STATE(842), 1,
      aux_sym_block_repeat1,
    STATE(2696), 1,
      sym_option_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2879), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(2881), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [85278] = 22,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(3028), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2073), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [85355] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3030), 1,
      aux_sym_symbol_token3,
    ACTIONS(2279), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2281), 21,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [85396] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3032), 14,
      anon_sym_PLUS,
      anon_sym_DASH,
      aux_sym__hll_non_default_format_var_command_token1,
      anon_sym_BSLASH,
      sym_access_class,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
      sym_float,
      aux_sym_bitmask_token1,
      aux_sym_bitmask_token2,
      aux_sym_symbol_token2,
      sym_time,
      sym_frequency,
      sym_identifier,
    ACTIONS(3036), 16,
      sym__decimal_number_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym__composed_name,
      sym_percentage,
      sym_file_handle,
  [85435] = 13,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2977), 1,
      anon_sym_SLASH,
    ACTIONS(2979), 1,
      anon_sym_CARET,
    ACTIONS(2991), 1,
      anon_sym_PIPE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2971), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2975), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2042), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    ACTIONS(2973), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(2044), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2039), 10,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [85494] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1953), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1955), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [85533] = 23,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2949), 1,
      anon_sym_COMMA,
    ACTIONS(3040), 1,
      anon_sym_RPAREN,
    ACTIONS(3042), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2422), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [85612] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2051), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(2053), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [85651] = 6,
    ACTIONS(2977), 1,
      anon_sym_SLASH,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2975), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2973), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1961), 10,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1963), 14,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [85696] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1987), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1989), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [85735] = 4,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2973), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1961), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1963), 16,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [85776] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1901), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1903), 20,
      sym__and_operator_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [85815] = 23,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2949), 1,
      anon_sym_COMMA,
    ACTIONS(3044), 1,
      anon_sym_RPAREN,
    ACTIONS(3046), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2273), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [85894] = 23,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2949), 1,
      anon_sym_COMMA,
    ACTIONS(3048), 1,
      anon_sym_RPAREN,
    ACTIONS(3050), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2286), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [85973] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2627), 1,
      anon_sym_LPAREN,
    ACTIONS(2657), 1,
      anon_sym_LBRACK,
    STATE(872), 1,
      sym_hll_argument_list,
    ACTIONS(2653), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2437), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2433), 17,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [86020] = 23,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2949), 1,
      anon_sym_COMMA,
    ACTIONS(3052), 1,
      anon_sym_RPAREN,
    ACTIONS(3054), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2432), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [86099] = 23,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2867), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2869), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2997), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    ACTIONS(3003), 1,
      anon_sym_PIPE,
    ACTIONS(3005), 1,
      anon_sym_CARET,
    ACTIONS(3015), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3017), 1,
      anon_sym_AMP,
    ACTIONS(3019), 1,
      anon_sym_QMARK,
    ACTIONS(3056), 1,
      sym__blank,
    STATE(956), 1,
      sym_hll_argument_list,
    STATE(1287), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [86178] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1957), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1959), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
  [86217] = 11,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2133), 8,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 10,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_COMMA,
  [86271] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3076), 1,
      anon_sym_SLASH,
    ACTIONS(3070), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3074), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3072), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1961), 8,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1963), 14,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [86317] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2459), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [86365] = 9,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1925), 1,
      sym_path,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(3078), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(1905), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 10,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [86415] = 17,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 6,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [86481] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3100), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      anon_sym_const,
      anon_sym_volatile,
      aux_sym_symbol_token2,
      sym_identifier,
      sym__blank,
    ACTIONS(3102), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_RBRACK,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [86519] = 22,
    ACTIONS(2178), 1,
      anon_sym_RBRACK,
    ACTIONS(2180), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2302), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [86595] = 22,
    ACTIONS(2063), 1,
      anon_sym_RBRACK,
    ACTIONS(2065), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2331), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [86671] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1917), 4,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1939), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1942), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [86713] = 7,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2437), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2433), 15,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [86759] = 7,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1917), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1905), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 10,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [86805] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2447), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2445), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [86853] = 7,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1917), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1935), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1939), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1942), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1937), 10,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [86899] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2390), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2392), 21,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [86937] = 15,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2459), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 7,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [86999] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3120), 1,
      sym__blank,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2168), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [87071] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2453), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2451), 13,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [87121] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3023), 1,
      anon_sym_AMP,
    STATE(1172), 1,
      aux_sym__path_expression_repeat2,
    STATE(2701), 1,
      sym_macro,
    ACTIONS(2088), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2090), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [87165] = 12,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3076), 1,
      anon_sym_SLASH,
    ACTIONS(3124), 1,
      anon_sym_PIPE,
    ACTIONS(3126), 1,
      anon_sym_CARET,
    ACTIONS(3070), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3074), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3072), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1975), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1977), 13,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [87221] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2585), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2583), 21,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [87259] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2495), 3,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
  [87331] = 22,
    ACTIONS(2084), 1,
      anon_sym_RBRACK,
    ACTIONS(2086), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2369), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [87407] = 12,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3076), 1,
      anon_sym_SLASH,
    ACTIONS(3124), 1,
      anon_sym_PIPE,
    ACTIONS(3126), 1,
      anon_sym_CARET,
    ACTIONS(3070), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3074), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3072), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(2059), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2061), 13,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [87463] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2463), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2461), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [87511] = 22,
    ACTIONS(2225), 1,
      anon_sym_RBRACK,
    ACTIONS(2227), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2340), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [87587] = 22,
    ACTIONS(2142), 1,
      anon_sym_RBRACK,
    ACTIONS(2144), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2288), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [87663] = 22,
    ACTIONS(2241), 1,
      anon_sym_RBRACK,
    ACTIONS(2243), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2372), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [87739] = 22,
    ACTIONS(2194), 1,
      anon_sym_RBRACK,
    ACTIONS(2196), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2382), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [87815] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2503), 3,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
  [87887] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2589), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2587), 21,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [87925] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2507), 3,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
  [87997] = 22,
    ACTIONS(2217), 1,
      anon_sym_RBRACK,
    ACTIONS(2219), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2357), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [88073] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1905), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 13,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [88119] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2047), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(2049), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [88157] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2033), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(2035), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [88195] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2511), 3,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
  [88267] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2613), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2611), 21,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [88305] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2467), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2465), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [88353] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2471), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2469), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [88401] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2797), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      anon_sym_const,
      anon_sym_volatile,
      aux_sym_symbol_token2,
      sym_identifier,
      sym__blank,
    ACTIONS(2799), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_RBRACK,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [88439] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1987), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1989), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [88477] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2617), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2615), 21,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [88515] = 22,
    ACTIONS(2123), 1,
      anon_sym_RBRACK,
    ACTIONS(2125), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2378), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [88591] = 11,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 9,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [88645] = 22,
    ACTIONS(2095), 1,
      anon_sym_RBRACK,
    ACTIONS(2097), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2375), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [88721] = 22,
    ACTIONS(2221), 1,
      anon_sym_RBRACK,
    ACTIONS(2223), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2470), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [88797] = 19,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    ACTIONS(3003), 1,
      anon_sym_PIPE,
    ACTIONS(3005), 1,
      anon_sym_CARET,
    ACTIONS(3015), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3017), 1,
      anon_sym_AMP,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2459), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [88867] = 22,
    ACTIONS(2237), 1,
      anon_sym_RBRACK,
    ACTIONS(2239), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2364), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [88943] = 22,
    ACTIONS(2107), 1,
      anon_sym_RBRACK,
    ACTIONS(2109), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2379), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [89019] = 22,
    ACTIONS(2111), 1,
      anon_sym_RBRACK,
    ACTIONS(2113), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2380), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [89095] = 9,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 11,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [89145] = 17,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    ACTIONS(3005), 1,
      anon_sym_CARET,
    ACTIONS(3017), 1,
      anon_sym_AMP,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 3,
      anon_sym_PIPE,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [89211] = 22,
    ACTIONS(2115), 1,
      anon_sym_RBRACK,
    ACTIONS(2117), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2381), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [89287] = 16,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    ACTIONS(3017), 1,
      anon_sym_AMP,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 3,
      anon_sym_PIPE,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [89351] = 14,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 7,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [89411] = 12,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 6,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 9,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [89467] = 22,
    ACTIONS(2119), 1,
      anon_sym_RBRACK,
    ACTIONS(2121), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2239), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [89543] = 18,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    ACTIONS(3003), 1,
      anon_sym_PIPE,
    ACTIONS(3005), 1,
      anon_sym_CARET,
    ACTIONS(3017), 1,
      anon_sym_AMP,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2459), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [89611] = 15,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2457), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [89673] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2515), 3,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
  [89745] = 21,
    ACTIONS(2495), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2997), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    ACTIONS(3003), 1,
      anon_sym_PIPE,
    ACTIONS(3005), 1,
      anon_sym_CARET,
    ACTIONS(3015), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3017), 1,
      anon_sym_AMP,
    ACTIONS(3019), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2497), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [89819] = 5,
    ACTIONS(154), 1,
      anon_sym_AMP,
    STATE(1335), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2099), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2101), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [89861] = 9,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2453), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2451), 11,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [89911] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2519), 3,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
  [89983] = 11,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2457), 11,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [90037] = 18,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 5,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [90105] = 6,
    ACTIONS(2200), 1,
      anon_sym_LPAREN,
    ACTIONS(2203), 1,
      anon_sym_DOT,
    STATE(2373), 1,
      aux_sym__function_identifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1905), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1907), 17,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [90149] = 9,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2457), 13,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [90199] = 17,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2459), 1,
      anon_sym_PIPE,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 6,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [90265] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1935), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1939), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1942), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1937), 13,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [90311] = 16,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2459), 1,
      anon_sym_PIPE,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 7,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [90375] = 14,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2459), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 9,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [90435] = 5,
    ACTIONS(3134), 1,
      sym_path,
    STATE(1243), 1,
      aux_sym__path_expression_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2099), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2101), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [90477] = 8,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2475), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2473), 13,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [90525] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2051), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(2053), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [90563] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2319), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2321), 21,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [90601] = 8,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2158), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2160), 14,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [90649] = 7,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2162), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2164), 15,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [90695] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3120), 1,
      sym__blank,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2152), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [90767] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2055), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(2057), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [90805] = 21,
    ACTIONS(2503), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2997), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    ACTIONS(3003), 1,
      anon_sym_PIPE,
    ACTIONS(3005), 1,
      anon_sym_CARET,
    ACTIONS(3015), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3017), 1,
      anon_sym_AMP,
    ACTIONS(3019), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2505), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [90879] = 13,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3076), 1,
      anon_sym_SLASH,
    ACTIONS(3124), 1,
      anon_sym_PIPE,
    ACTIONS(3126), 1,
      anon_sym_CARET,
    ACTIONS(3070), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3074), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2042), 3,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
    ACTIONS(3072), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(2044), 4,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2039), 10,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [90937] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1931), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1933), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [90975] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(187), 1,
      anon_sym_AMP,
    STATE(1172), 1,
      aux_sym__path_expression_repeat2,
    STATE(1384), 1,
      sym_macro,
    ACTIONS(2099), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2101), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [91019] = 22,
    ACTIONS(2213), 1,
      anon_sym_RBRACK,
    ACTIONS(2215), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2392), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [91095] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1953), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1955), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [91133] = 10,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1925), 1,
      sym_path,
    ACTIONS(3136), 1,
      sym__blank,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(3078), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(1905), 3,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 10,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [91185] = 5,
    ACTIONS(154), 1,
      anon_sym_AMP,
    STATE(1335), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2249), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2251), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [91227] = 5,
    ACTIONS(3138), 1,
      sym_path,
    STATE(1243), 1,
      aux_sym__path_expression_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2077), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2079), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [91269] = 22,
    ACTIONS(2205), 1,
      anon_sym_RBRACK,
    ACTIONS(2207), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2306), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [91345] = 21,
    ACTIONS(2507), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2997), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    ACTIONS(3003), 1,
      anon_sym_PIPE,
    ACTIONS(3005), 1,
      anon_sym_CARET,
    ACTIONS(3015), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3017), 1,
      anon_sym_AMP,
    ACTIONS(3019), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2509), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [91419] = 21,
    ACTIONS(2511), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2997), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    ACTIONS(3003), 1,
      anon_sym_PIPE,
    ACTIONS(3005), 1,
      anon_sym_CARET,
    ACTIONS(3015), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3017), 1,
      anon_sym_AMP,
    ACTIONS(3019), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2513), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [91493] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1917), 4,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
    ACTIONS(1909), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1912), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [91535] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2593), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2591), 21,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      sym_comment,
      sym__blank_line,
  [91573] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1949), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1951), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [91611] = 21,
    ACTIONS(2515), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2997), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    ACTIONS(3003), 1,
      anon_sym_PIPE,
    ACTIONS(3005), 1,
      anon_sym_CARET,
    ACTIONS(3015), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3017), 1,
      anon_sym_AMP,
    ACTIONS(3019), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2517), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [91685] = 21,
    ACTIONS(2519), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2997), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3001), 1,
      anon_sym_SLASH,
    ACTIONS(3003), 1,
      anon_sym_PIPE,
    ACTIONS(3005), 1,
      anon_sym_CARET,
    ACTIONS(3015), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3017), 1,
      anon_sym_AMP,
    ACTIONS(3019), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2521), 2,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2993), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2999), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3007), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3009), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3011), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3013), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [91759] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1957), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1959), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [91797] = 22,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3141), 1,
      anon_sym_RPAREN,
    ACTIONS(3143), 1,
      anon_sym_COMMA,
    ACTIONS(3145), 1,
      sym__blank,
    STATE(1128), 1,
      sym_argument_list,
    STATE(1974), 1,
      aux_sym_argument_list_repeat1,
    STATE(2171), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [91873] = 17,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2146), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2148), 5,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
  [91939] = 4,
    ACTIONS(3147), 1,
      aux_sym_symbol_token3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2279), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2281), 19,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [91979] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2257), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2259), 20,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      aux_sym_symbol_token3,
  [92017] = 14,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2146), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      sym__blank,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2148), 6,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
  [92077] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3120), 1,
      sym__blank,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2156), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [92149] = 22,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3143), 1,
      anon_sym_COMMA,
    ACTIONS(3145), 1,
      sym__blank,
    ACTIONS(3149), 1,
      anon_sym_RPAREN,
    STATE(1128), 1,
      sym_argument_list,
    STATE(1937), 1,
      aux_sym_argument_list_repeat1,
    STATE(2171), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [92225] = 22,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3143), 1,
      anon_sym_COMMA,
    ACTIONS(3145), 1,
      sym__blank,
    ACTIONS(3151), 1,
      anon_sym_RPAREN,
    STATE(1128), 1,
      sym_argument_list,
    STATE(1944), 1,
      aux_sym_argument_list_repeat1,
    STATE(2171), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [92301] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1945), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1947), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [92339] = 22,
    ACTIONS(2069), 1,
      anon_sym_RBRACK,
    ACTIONS(2071), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2472), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [92415] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3120), 1,
      sym__blank,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2129), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [92487] = 22,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    ACTIONS(2949), 1,
      anon_sym_COMMA,
    ACTIONS(3153), 1,
      anon_sym_RPAREN,
    ACTIONS(3155), 1,
      sym__blank,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [92563] = 5,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2133), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [92605] = 17,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2133), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2135), 5,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
  [92671] = 18,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2133), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2135), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [92739] = 9,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2133), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 12,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_COMMA,
  [92789] = 16,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3157), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2133), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      sym__blank,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2135), 5,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
  [92853] = 15,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3157), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2133), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      sym__blank,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2135), 5,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
  [92915] = 12,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2133), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 10,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_COMMA,
  [92971] = 8,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2133), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 14,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [93019] = 22,
    ACTIONS(2073), 1,
      anon_sym_RBRACK,
    ACTIONS(2075), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2221), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [93095] = 22,
    ACTIONS(2174), 1,
      anon_sym_RBRACK,
    ACTIONS(2176), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2223), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [93171] = 22,
    ACTIONS(2190), 1,
      anon_sym_RBRACK,
    ACTIONS(2192), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2230), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [93247] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2275), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2277), 20,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
      aux_sym_symbol_token3,
  [93285] = 22,
    ACTIONS(2209), 1,
      anon_sym_RBRACK,
    ACTIONS(2211), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2272), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [93361] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3076), 1,
      anon_sym_SLASH,
    ACTIONS(3074), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3072), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1961), 10,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1963), 14,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [93405] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1961), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1963), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [93443] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3072), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1961), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1963), 16,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [93483] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3076), 1,
      anon_sym_SLASH,
    ACTIONS(3126), 1,
      anon_sym_CARET,
    ACTIONS(3070), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3074), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3072), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1961), 7,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1963), 14,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [93531] = 12,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 4,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2457), 11,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_RBRACK,
  [93587] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1927), 11,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
      sym__blank,
    ACTIONS(1929), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_COLON_COLON_COLON,
      sym_comment,
      sym__blank_line,
  [93625] = 22,
    ACTIONS(2229), 1,
      anon_sym_RBRACK,
    ACTIONS(2231), 1,
      sym__blank,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(2811), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2815), 1,
      anon_sym_SLASH,
    ACTIONS(2817), 1,
      anon_sym_PIPE,
    ACTIONS(2819), 1,
      anon_sym_CARET,
    ACTIONS(2829), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2831), 1,
      anon_sym_AMP,
    ACTIONS(2835), 1,
      anon_sym_QMARK,
    STATE(831), 1,
      sym_hll_argument_list,
    STATE(2345), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2809), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2813), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(2821), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2823), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(2825), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2827), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [93701] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2257), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2259), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      aux_sym_symbol_token3,
  [93738] = 14,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 3,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2457), 7,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [93797] = 5,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2889), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(2891), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [93838] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3172), 1,
      sym_path,
    STATE(1288), 1,
      aux_sym__path_expression_repeat1,
    ACTIONS(2077), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2079), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [93879] = 10,
    ACTIONS(3179), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3181), 1,
      sym__blank_line,
    ACTIONS(3184), 1,
      sym__blank,
    STATE(1563), 1,
      sym_else_block,
    STATE(1567), 1,
      sym_elif_block,
    STATE(1878), 1,
      aux_sym_if_block_repeat2,
    STATE(1889), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3175), 6,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3177), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [93930] = 10,
    ACTIONS(3191), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3194), 1,
      sym__blank_line,
    ACTIONS(3197), 1,
      sym__blank,
    STATE(1526), 1,
      sym_elif_block,
    STATE(1556), 1,
      sym_else_block,
    STATE(1916), 1,
      aux_sym_if_block_repeat2,
    STATE(1917), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3189), 6,
      sym_label_identifier,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3187), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [93981] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2200), 1,
      anon_sym_LPAREN,
    ACTIONS(2203), 1,
      anon_sym_DOT,
    STATE(2373), 1,
      aux_sym__function_identifier_repeat1,
    ACTIONS(1905), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1907), 17,
      sym__and_operator_pre_hook,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [94024] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2088), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2090), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP,
      anon_sym_COMMA,
  [94061] = 10,
    ACTIONS(3204), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3207), 1,
      sym__blank_line,
    ACTIONS(3210), 1,
      sym__blank,
    STATE(1525), 1,
      sym_else_block,
    STATE(1539), 1,
      sym_elif_block,
    STATE(1819), 1,
      aux_sym_if_block_repeat2,
    STATE(1919), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3202), 6,
      sym_label_identifier,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3200), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [94112] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2158), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2160), 14,
      sym__and_operator_pre_hook,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [94159] = 10,
    ACTIONS(3181), 1,
      sym__blank_line,
    ACTIONS(3184), 1,
      sym__blank,
    ACTIONS(3219), 1,
      aux_sym_elif_block_token1,
    STATE(1563), 1,
      sym_else_block,
    STATE(1595), 1,
      sym_elif_block,
    STATE(1920), 1,
      aux_sym_if_block_repeat2,
    STATE(1921), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3175), 6,
      sym_label_identifier,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3177), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [94210] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2168), 3,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [94281] = 4,
    ACTIONS(1925), 1,
      sym_path,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2249), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2251), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [94320] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2156), 3,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [94391] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(187), 1,
      anon_sym_AMP,
    STATE(1369), 1,
      sym_macro,
    ACTIONS(2249), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2251), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [94432] = 10,
    ACTIONS(3179), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3207), 1,
      sym__blank_line,
    ACTIONS(3210), 1,
      sym__blank,
    STATE(1525), 1,
      sym_else_block,
    STATE(1528), 1,
      sym_elif_block,
    STATE(1832), 1,
      aux_sym_if_block_repeat2,
    STATE(1835), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3202), 6,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3200), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [94483] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2133), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 14,
      sym__and_operator_pre_hook,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [94530] = 12,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 5,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2457), 9,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [94585] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2613), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2611), 19,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [94622] = 17,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(2146), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2148), 5,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [94687] = 15,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3248), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2133), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      sym__blank,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2135), 5,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [94748] = 21,
    ACTIONS(2507), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2509), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    ACTIONS(3251), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3253), 1,
      anon_sym_PIPE,
    ACTIONS(3255), 1,
      anon_sym_CARET,
    ACTIONS(3259), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3261), 1,
      anon_sym_AMP,
    ACTIONS(3263), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3257), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [94821] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3267), 1,
      sym__blank,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3265), 3,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [94892] = 21,
    ACTIONS(2511), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2513), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    ACTIONS(3251), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3253), 1,
      anon_sym_PIPE,
    ACTIONS(3255), 1,
      anon_sym_CARET,
    ACTIONS(3259), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3261), 1,
      anon_sym_AMP,
    ACTIONS(3263), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3257), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [94965] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2390), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2392), 19,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [95002] = 14,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2146), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      sym__blank,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2148), 6,
      sym__and_operator_pre_hook,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [95061] = 4,
    ACTIONS(3270), 1,
      aux_sym_symbol_token3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2279), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2281), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [95100] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2617), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2615), 19,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [95137] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2589), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2587), 19,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [95174] = 5,
    ACTIONS(1857), 1,
      sym__blank,
    STATE(842), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2879), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(2881), 20,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_AMP,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [95215] = 16,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3248), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2133), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      sym__blank,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2135), 5,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [95278] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2152), 3,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [95349] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3274), 1,
      sym__blank,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3272), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [95420] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3279), 1,
      sym__blank,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3277), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [95491] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2129), 3,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [95562] = 10,
    ACTIONS(3207), 1,
      sym__blank_line,
    ACTIONS(3210), 1,
      sym__blank,
    ACTIONS(3282), 1,
      aux_sym_elif_block_token1,
    STATE(1525), 1,
      sym_else_block,
    STATE(1539), 1,
      sym_elif_block,
    STATE(1907), 1,
      aux_sym_if_block_repeat2,
    STATE(1908), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3202), 6,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3200), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [95613] = 10,
    ACTIONS(3194), 1,
      sym__blank_line,
    ACTIONS(3197), 1,
      sym__blank,
    ACTIONS(3285), 1,
      aux_sym_elif_block_token1,
    STATE(1544), 1,
      sym_elif_block,
    STATE(1556), 1,
      sym_else_block,
    STATE(1903), 1,
      aux_sym_if_block_repeat2,
    STATE(1912), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3189), 6,
      sym_label_identifier,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3187), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [95664] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2319), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2321), 19,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [95701] = 10,
    ACTIONS(3179), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3194), 1,
      sym__blank_line,
    ACTIONS(3197), 1,
      sym__blank,
    STATE(1544), 1,
      sym_elif_block,
    STATE(1556), 1,
      sym_else_block,
    STATE(1844), 1,
      aux_sym_if_block_repeat3,
    STATE(1925), 1,
      aux_sym_if_block_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3189), 6,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3187), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [95752] = 18,
    ACTIONS(2459), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    ACTIONS(3253), 1,
      anon_sym_PIPE,
    ACTIONS(3255), 1,
      anon_sym_CARET,
    ACTIONS(3261), 1,
      anon_sym_AMP,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3257), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2457), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [95819] = 9,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2133), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 12,
      sym__and_operator_pre_hook,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [95868] = 11,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2133), 8,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 10,
      sym__and_operator_pre_hook,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [95921] = 9,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2453), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2451), 11,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [95970] = 21,
    ACTIONS(2515), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2517), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    ACTIONS(3251), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3253), 1,
      anon_sym_PIPE,
    ACTIONS(3255), 1,
      anon_sym_CARET,
    ACTIONS(3259), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3261), 1,
      anon_sym_AMP,
    ACTIONS(3263), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3257), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [96043] = 9,
    ACTIONS(1917), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1925), 1,
      sym_path,
    ACTIONS(3136), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1915), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1905), 3,
      anon_sym_EQ,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(1909), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(1912), 5,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_DOT_DOT,
    ACTIONS(1907), 10,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [96092] = 21,
    ACTIONS(2519), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2521), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    ACTIONS(3251), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3253), 1,
      anon_sym_PIPE,
    ACTIONS(3255), 1,
      anon_sym_CARET,
    ACTIONS(3259), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3261), 1,
      anon_sym_AMP,
    ACTIONS(3263), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3257), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [96165] = 11,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2457), 9,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [96218] = 10,
    ACTIONS(3207), 1,
      sym__blank_line,
    ACTIONS(3210), 1,
      sym__blank,
    ACTIONS(3285), 1,
      aux_sym_elif_block_token1,
    STATE(1525), 1,
      sym_else_block,
    STATE(1528), 1,
      sym_elif_block,
    STATE(1821), 1,
      aux_sym_if_block_repeat2,
    STATE(1822), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3202), 6,
      sym_label_identifier,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3200), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [96269] = 10,
    ACTIONS(3181), 1,
      sym__blank_line,
    ACTIONS(3184), 1,
      sym__blank,
    ACTIONS(3287), 1,
      aux_sym_elif_block_token1,
    STATE(1563), 1,
      sym_else_block,
    STATE(1595), 1,
      sym_elif_block,
    STATE(1909), 1,
      aux_sym_if_block_repeat2,
    STATE(1910), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3175), 6,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3177), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [96320] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3292), 1,
      sym__blank,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3290), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [96391] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2077), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2079), 19,
      sym__and_operator_pre_hook,
      sym_path,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [96428] = 21,
    ACTIONS(2503), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2505), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    ACTIONS(3251), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3253), 1,
      anon_sym_PIPE,
    ACTIONS(3255), 1,
      anon_sym_CARET,
    ACTIONS(3259), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3261), 1,
      anon_sym_AMP,
    ACTIONS(3263), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3257), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [96501] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(2133), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [96542] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2162), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2164), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [96587] = 17,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(2133), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(2135), 5,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [96652] = 18,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(2133), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2135), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [96719] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(187), 1,
      anon_sym_AMP,
    STATE(1369), 1,
      sym_macro,
    ACTIONS(2099), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2101), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [96760] = 10,
    ACTIONS(3194), 1,
      sym__blank_line,
    ACTIONS(3197), 1,
      sym__blank,
    ACTIONS(3295), 1,
      aux_sym_elif_block_token1,
    STATE(1526), 1,
      sym_elif_block,
    STATE(1556), 1,
      sym_else_block,
    STATE(1905), 1,
      aux_sym_if_block_repeat2,
    STATE(1906), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3189), 6,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3187), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [96811] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2275), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2277), 19,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      aux_sym_symbol_token3,
  [96848] = 19,
    ACTIONS(2459), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    ACTIONS(3253), 1,
      anon_sym_PIPE,
    ACTIONS(3255), 1,
      anon_sym_CARET,
    ACTIONS(3259), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3261), 1,
      anon_sym_AMP,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3257), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2457), 3,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [96917] = 9,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 8,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2457), 11,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [96966] = 17,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    ACTIONS(3255), 1,
      anon_sym_CARET,
    ACTIONS(3261), 1,
      anon_sym_AMP,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2459), 2,
      anon_sym_PIPE,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3257), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2457), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [97031] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3298), 1,
      sym_path,
    STATE(1288), 1,
      aux_sym__path_expression_repeat1,
    ACTIONS(2099), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2101), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [97072] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2585), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2583), 19,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [97109] = 10,
    ACTIONS(3181), 1,
      sym__blank_line,
    ACTIONS(3184), 1,
      sym__blank,
    ACTIONS(3285), 1,
      aux_sym_elif_block_token1,
    STATE(1563), 1,
      sym_else_block,
    STATE(1567), 1,
      sym_elif_block,
    STATE(1826), 1,
      aux_sym_if_block_repeat2,
    STATE(1827), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3175), 6,
      sym_label_identifier,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3177), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [97160] = 15,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3257), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2459), 3,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2457), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [97221] = 16,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    ACTIONS(3261), 1,
      anon_sym_AMP,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2459), 2,
      anon_sym_PIPE,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3257), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2457), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      aux_sym__hll_expression_separator_token1,
  [97284] = 21,
    ACTIONS(2495), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2497), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    ACTIONS(3251), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3253), 1,
      anon_sym_PIPE,
    ACTIONS(3255), 1,
      anon_sym_CARET,
    ACTIONS(3259), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3261), 1,
      anon_sym_AMP,
    ACTIONS(3263), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3257), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [97357] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2593), 9,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
      aux_sym__hll_expression_separator_token2,
      sym__blank,
    ACTIONS(2591), 19,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_DOT,
      anon_sym_QMARK,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      aux_sym__hll_expression_separator_token1,
  [97394] = 12,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2133), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 10,
      sym__and_operator_pre_hook,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [97449] = 21,
    ACTIONS(2677), 1,
      anon_sym_LPAREN,
    ACTIONS(2683), 1,
      anon_sym_LBRACK,
    ACTIONS(2867), 1,
      aux_sym__hll_expression_separator_token1,
    ACTIONS(2869), 1,
      aux_sym__hll_expression_separator_token2,
    ACTIONS(3164), 1,
      anon_sym_SLASH,
    ACTIONS(3251), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3253), 1,
      anon_sym_PIPE,
    ACTIONS(3255), 1,
      anon_sym_CARET,
    ACTIONS(3259), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3261), 1,
      anon_sym_AMP,
    ACTIONS(3263), 1,
      anon_sym_QMARK,
    STATE(956), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2681), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2995), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3160), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3162), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3166), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3168), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3170), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3257), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [97522] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3300), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [97592] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3302), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [97662] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2327), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2329), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [97698] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2463), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2461), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [97744] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2200), 1,
      anon_sym_LPAREN,
    ACTIONS(2203), 1,
      anon_sym_DOT,
    STATE(2373), 1,
      aux_sym__function_identifier_repeat1,
    ACTIONS(3078), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(1905), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1907), 14,
      sym__and_operator_pre_hook,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [97788] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2503), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [97858] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2390), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2392), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [97894] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2507), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [97964] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2511), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [98034] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3330), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [98104] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2200), 1,
      anon_sym_LPAREN,
    ACTIONS(2203), 1,
      anon_sym_DOT,
    ACTIONS(3136), 1,
      sym__blank,
    STATE(2373), 1,
      aux_sym__function_identifier_repeat1,
    ACTIONS(3078), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(1905), 8,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(1907), 14,
      sym__and_operator_pre_hook,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [98150] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2319), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2321), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [98186] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3332), 2,
      sym_comment,
      sym__blank_line,
  [98256] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2077), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2079), 19,
      sym__and_operator_pre_hook,
      sym_path,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [98292] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3120), 1,
      sym__blank,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3334), 1,
      anon_sym_RPAREN,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [98362] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3120), 1,
      sym__blank,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3334), 1,
      anon_sym_RBRACE,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [98432] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2467), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2465), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [98478] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2311), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2313), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [98514] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3336), 1,
      aux_sym_symbol_token3,
    ACTIONS(2279), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2281), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [98552] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2257), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2259), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      aux_sym_symbol_token3,
      sym_comment,
      sym__blank_line,
  [98588] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2471), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2469), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [98634] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2515), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [98704] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3338), 1,
      anon_sym_RBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [98774] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2519), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [98844] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3340), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [98914] = 7,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(3346), 1,
      sym__blank,
    STATE(1497), 1,
      aux_sym_block_repeat1,
    STATE(1549), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3342), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3344), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [98958] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2088), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2090), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_AMP,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [98994] = 7,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(3348), 1,
      sym__blank,
    STATE(1509), 1,
      aux_sym_block_repeat1,
    STATE(1521), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3342), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3344), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [99038] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2965), 1,
      sym_path,
    ACTIONS(2249), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2251), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [99076] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2475), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2473), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [99122] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3350), 1,
      anon_sym_RBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [99192] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3352), 1,
      anon_sym_RBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [99262] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3354), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [99332] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2437), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2433), 16,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [99376] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2447), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2445), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [99422] = 9,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2453), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2451), 12,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [99470] = 6,
    ACTIONS(3360), 1,
      anon_sym_EQ,
    ACTIONS(3362), 1,
      sym__blank,
    STATE(2231), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3356), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3358), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [99512] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3364), 1,
      anon_sym_RBRACK,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [99582] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3366), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [99652] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3368), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [99722] = 6,
    ACTIONS(3360), 1,
      anon_sym_EQ,
    ACTIONS(3362), 1,
      sym__blank,
    STATE(2231), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3370), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3372), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [99764] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3265), 1,
      anon_sym_COMMA,
    ACTIONS(3374), 1,
      anon_sym_EQ,
    ACTIONS(3380), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3382), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3388), 1,
      anon_sym_PIPE,
    ACTIONS(3390), 1,
      anon_sym_CARET,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3400), 1,
      sym__blank,
    ACTIONS(3403), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [99834] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3405), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [99904] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3407), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [99974] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3409), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [100044] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3120), 1,
      sym__blank,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3411), 1,
      anon_sym_RPAREN,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [100114] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3120), 1,
      sym__blank,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3411), 1,
      anon_sym_RBRACE,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [100184] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3332), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [100254] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3413), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [100324] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3415), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [100394] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3417), 1,
      anon_sym_RPAREN,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [100464] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3419), 1,
      anon_sym_COLON,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [100534] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3421), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [100604] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3423), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [100674] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3425), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [100744] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2315), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2317), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [100780] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3427), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [100850] = 20,
    ACTIONS(2129), 1,
      anon_sym_COMMA,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3374), 1,
      anon_sym_EQ,
    ACTIONS(3380), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3382), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3388), 1,
      anon_sym_PIPE,
    ACTIONS(3390), 1,
      anon_sym_CARET,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3403), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3429), 1,
      sym__blank,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [100920] = 11,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2133), 8,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 8,
      sym__and_operator_pre_hook,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_COMMA,
  [100972] = 17,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3388), 1,
      anon_sym_PIPE,
    ACTIONS(3390), 1,
      anon_sym_CARET,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3403), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2133), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2135), 3,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [101036] = 18,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3380), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3388), 1,
      anon_sym_PIPE,
    ACTIONS(3390), 1,
      anon_sym_CARET,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3403), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2133), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(2135), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [101102] = 9,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2133), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 10,
      sym__and_operator_pre_hook,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_COMMA,
  [101150] = 16,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3390), 1,
      anon_sym_CARET,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3431), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2133), 3,
      anon_sym_EQ,
      anon_sym_PIPE,
      sym__blank,
    ACTIONS(2135), 3,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [101212] = 15,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3431), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2135), 3,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
    ACTIONS(2133), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      sym__blank,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [101272] = 12,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2133), 6,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 8,
      sym__and_operator_pre_hook,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_COMMA,
  [101326] = 8,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2133), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 12,
      sym__and_operator_pre_hook,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [101372] = 17,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3388), 1,
      anon_sym_PIPE,
    ACTIONS(3390), 1,
      anon_sym_CARET,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3403), 1,
      sym__and_operator_pre_hook,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2146), 2,
      anon_sym_EQ,
      sym__blank,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2148), 3,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [101436] = 14,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2146), 4,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_CARET,
      sym__blank,
    ACTIONS(2148), 4,
      sym__and_operator_pre_hook,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_COMMA,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [101494] = 20,
    ACTIONS(2152), 1,
      anon_sym_COMMA,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3374), 1,
      anon_sym_EQ,
    ACTIONS(3380), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3382), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3388), 1,
      anon_sym_PIPE,
    ACTIONS(3390), 1,
      anon_sym_CARET,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3403), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3429), 1,
      sym__blank,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [101564] = 20,
    ACTIONS(2156), 1,
      anon_sym_COMMA,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3374), 1,
      anon_sym_EQ,
    ACTIONS(3380), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3382), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3388), 1,
      anon_sym_PIPE,
    ACTIONS(3390), 1,
      anon_sym_CARET,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3403), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3429), 1,
      sym__blank,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [101634] = 20,
    ACTIONS(2168), 1,
      anon_sym_COMMA,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3374), 1,
      anon_sym_EQ,
    ACTIONS(3380), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3382), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3388), 1,
      anon_sym_PIPE,
    ACTIONS(3390), 1,
      anon_sym_CARET,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3403), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3429), 1,
      sym__blank,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [101704] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3120), 1,
      sym__blank,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3434), 1,
      anon_sym_RPAREN,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [101774] = 20,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3064), 1,
      anon_sym_SLASH,
    ACTIONS(3068), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3104), 1,
      anon_sym_EQ,
    ACTIONS(3108), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3110), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3112), 1,
      anon_sym_PIPE,
    ACTIONS(3114), 1,
      anon_sym_CARET,
    ACTIONS(3120), 1,
      sym__blank,
    ACTIONS(3122), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3434), 1,
      anon_sym_RBRACE,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2261), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3060), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3062), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3066), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3106), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3118), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3116), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [101844] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2275), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2277), 19,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      aux_sym_symbol_token3,
      sym_comment,
      sym__blank_line,
  [101880] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3436), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [101950] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3438), 1,
      anon_sym_RPAREN,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [102020] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3440), 1,
      anon_sym_COLON,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [102090] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3436), 2,
      sym_comment,
      sym__blank_line,
  [102160] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3442), 1,
      anon_sym_RPAREN,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [102230] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3444), 1,
      anon_sym_COLON,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [102300] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3446), 1,
      anon_sym_COLON,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [102370] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3448), 1,
      anon_sym_COLON,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [102440] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3450), 1,
      anon_sym_COLON,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [102510] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3452), 1,
      anon_sym_COLON,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [102580] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3454), 1,
      anon_sym_COLON,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [102650] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3456), 1,
      anon_sym_COLON,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [102720] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3458), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [102790] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3460), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [102860] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3460), 2,
      sym_comment,
      sym__blank_line,
  [102930] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3458), 2,
      sym_comment,
      sym__blank_line,
  [103000] = 11,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2457), 10,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [103052] = 18,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 4,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [103118] = 9,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2457), 12,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [103166] = 17,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2459), 1,
      anon_sym_PIPE,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [103230] = 16,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2459), 1,
      anon_sym_PIPE,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 6,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [103292] = 14,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2459), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 8,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [103350] = 12,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2459), 4,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2457), 10,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [103404] = 8,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2459), 7,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_GT,
      anon_sym_LT,
      anon_sym_AMP,
    ACTIONS(2457), 14,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_CARET,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [103450] = 17,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 5,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [103514] = 15,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2459), 2,
      anon_sym_PIPE,
      anon_sym_AMP,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2457), 6,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_AMP_AMP,
      anon_sym_QMARK,
      sym_comment,
      sym__blank_line,
  [103574] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3462), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [103644] = 20,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2443), 1,
      anon_sym_LBRACK,
    ACTIONS(3084), 1,
      anon_sym_SLASH,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    ACTIONS(3088), 1,
      anon_sym_CARET,
    ACTIONS(3098), 1,
      anon_sym_AMP,
    ACTIONS(3128), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3130), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3132), 1,
      anon_sym_QMARK,
    ACTIONS(3464), 1,
      anon_sym_COLON,
    STATE(831), 1,
      sym_hll_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2441), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2793), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3080), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3082), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3090), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3092), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3094), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3096), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [103714] = 8,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2158), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2160), 12,
      sym__and_operator_pre_hook,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [103760] = 7,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2162), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2164), 13,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_COMMA,
  [103804] = 5,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2133), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2135), 16,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [103844] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3466), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [103914] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3468), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [103984] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3468), 2,
      sym_comment,
      sym__blank_line,
  [104054] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3466), 2,
      sym_comment,
      sym__blank_line,
  [104124] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3470), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [104194] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2495), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [104264] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3472), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [104334] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3472), 2,
      sym_comment,
      sym__blank_line,
  [104404] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3474), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [104474] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3474), 2,
      sym_comment,
      sym__blank_line,
  [104544] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2323), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2325), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [104580] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2333), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2335), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [104616] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2333), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2335), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [104652] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2337), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2339), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [104688] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2337), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2339), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [104724] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2355), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2357), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [104760] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3476), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [104830] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3478), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [104900] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3478), 2,
      sym_comment,
      sym__blank_line,
  [104970] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3480), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [105040] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3480), 2,
      sym_comment,
      sym__blank_line,
  [105110] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2345), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2347), 18,
      sym__and_operator_pre_hook,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
  [105146] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3482), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [105216] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3484), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [105286] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3484), 2,
      sym_comment,
      sym__blank_line,
  [105356] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3486), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [105426] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3488), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [105496] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2711), 1,
      anon_sym_LPAREN,
    ACTIONS(2727), 1,
      anon_sym_LBRACK,
    ACTIONS(3306), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3310), 1,
      anon_sym_SLASH,
    ACTIONS(3312), 1,
      anon_sym_PIPE,
    ACTIONS(3314), 1,
      anon_sym_CARET,
    ACTIONS(3324), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3326), 1,
      anon_sym_AMP,
    ACTIONS(3328), 1,
      anon_sym_QMARK,
    STATE(973), 1,
      sym_hll_argument_list,
    ACTIONS(2725), 2,
      anon_sym_DOT,
      anon_sym_DASH_GT,
    ACTIONS(2841), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3304), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3308), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3316), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(3318), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3320), 2,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
    ACTIONS(3322), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3488), 2,
      sym_comment,
      sym__blank_line,
  [105566] = 20,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3217), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3222), 1,
      anon_sym_EQ,
    ACTIONS(3226), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3228), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3232), 1,
      anon_sym_SLASH,
    ACTIONS(3234), 1,
      anon_sym_PIPE,
    ACTIONS(3236), 1,
      anon_sym_CARET,
    ACTIONS(3244), 1,
      sym__blank,
    ACTIONS(3246), 1,
      sym__and_operator_pre_hook,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2469), 1,
      aux_sym_block_repeat1,
    ACTIONS(3215), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3224), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3230), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3240), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3242), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3490), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(3238), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [105636] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2355), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2357), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [105671] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2315), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2317), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [105706] = 17,
    ACTIONS(494), 1,
      anon_sym_struct,
    ACTIONS(496), 1,
      anon_sym_union,
    ACTIONS(498), 1,
      anon_sym_enum,
    ACTIONS(3492), 1,
      sym_identifier,
    ACTIONS(3494), 1,
      sym_hll_primitive_type,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(1608), 1,
      aux_sym__hll_declaration_specifiers_repeat1,
    STATE(1685), 1,
      sym__hll_declaration_specifiers,
    STATE(1812), 1,
      aux_sym_hll_sized_type_specifier_repeat1,
    STATE(2500), 1,
      sym_hll_declaration,
    STATE(2532), 1,
      sym__var_definition_command_argument_list_format,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(492), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(2489), 2,
      sym_hll_type_qualifier,
      sym__hll_declaration_modifiers,
    ACTIONS(500), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1810), 6,
      sym__hll_type_specifier,
      sym_hll_struct_specifier,
      sym_hll_union_specifier,
      sym_hll_enum_specifier,
      sym_hll_sized_type_specifier,
      sym__hll_type_identifier,
  [105769] = 19,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    ACTIONS(3374), 1,
      anon_sym_EQ,
    ACTIONS(3380), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3382), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3388), 1,
      anon_sym_PIPE,
    ACTIONS(3390), 1,
      anon_sym_CARET,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3403), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3429), 1,
      sym__blank,
    STATE(1128), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [105836] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2319), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2321), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [105871] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2327), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2329), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [105906] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2345), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2347), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [105941] = 6,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(3502), 1,
      sym__blank,
    STATE(1500), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3498), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3500), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [105982] = 6,
    ACTIONS(3508), 1,
      aux_sym_while_block_token1,
    ACTIONS(3511), 1,
      sym__blank_line,
    STATE(2383), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3504), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3506), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106023] = 4,
    ACTIONS(3518), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3514), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      aux_sym_while_block_token2,
      anon_sym_EQ,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3516), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106060] = 5,
    ACTIONS(3520), 1,
      sym__blank,
    STATE(1500), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2368), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(2370), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [106099] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3514), 9,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_EQ,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3516), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106134] = 4,
    ACTIONS(3527), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3523), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      aux_sym_while_block_token2,
      anon_sym_EQ,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3525), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106171] = 7,
    ACTIONS(2200), 1,
      anon_sym_LPAREN,
    ACTIONS(2203), 1,
      anon_sym_DOT,
    ACTIONS(3136), 1,
      sym__blank,
    STATE(2373), 1,
      aux_sym__function_identifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1905), 8,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(1907), 14,
      sym__and_operator_pre_hook,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [106214] = 6,
    ACTIONS(3533), 1,
      aux_sym_while_block_token1,
    ACTIONS(3536), 1,
      sym__blank_line,
    STATE(2414), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3529), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3531), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106255] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2323), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2325), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [106290] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3539), 9,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_EQ,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3541), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106325] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2390), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2392), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [106360] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2311), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2313), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [106395] = 6,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(3502), 1,
      sym__blank,
    STATE(1500), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3498), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3500), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [106436] = 19,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    ACTIONS(3374), 1,
      anon_sym_EQ,
    ACTIONS(3380), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3382), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3388), 1,
      anon_sym_PIPE,
    ACTIONS(3390), 1,
      anon_sym_CARET,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3403), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3429), 1,
      sym__blank,
    STATE(1234), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [106503] = 6,
    ACTIONS(3547), 1,
      aux_sym_while_block_token1,
    ACTIONS(3550), 1,
      sym__blank_line,
    STATE(2407), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3543), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3545), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106544] = 19,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    ACTIONS(3374), 1,
      anon_sym_EQ,
    ACTIONS(3380), 1,
      anon_sym_CARET_CARET,
    ACTIONS(3382), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3386), 1,
      anon_sym_SLASH,
    ACTIONS(3388), 1,
      anon_sym_PIPE,
    ACTIONS(3390), 1,
      anon_sym_CARET,
    ACTIONS(3398), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3403), 1,
      sym__and_operator_pre_hook,
    ACTIONS(3429), 1,
      sym__blank,
    STATE(432), 1,
      sym_argument_list,
    STATE(2397), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3376), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3378), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(3384), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3394), 2,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(3396), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3392), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
  [106611] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2333), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2335), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [106646] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2333), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2335), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [106681] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2337), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2339), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [106716] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2337), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(2339), 18,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [106751] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3553), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3555), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106785] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3557), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3559), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106819] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3561), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3563), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106853] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3565), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3567), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106887] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3498), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3500), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106921] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3569), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3571), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106955] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3565), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3567), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [106989] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3565), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3567), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107023] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3175), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3177), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107057] = 5,
    ACTIONS(3576), 1,
      sym__blank_line,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3573), 2,
      aux_sym_elif_block_token1,
      sym__blank,
    ACTIONS(3202), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3200), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [107095] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3579), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3581), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107129] = 6,
    ACTIONS(3583), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3585), 1,
      sym__blank_line,
    ACTIONS(3588), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3175), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3177), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [107169] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3591), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3593), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107203] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3595), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3597), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107237] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3599), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3601), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107271] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3599), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3601), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107305] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3078), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(1905), 9,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
      sym__blank,
    ACTIONS(1907), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [107341] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3603), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3605), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107375] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3607), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3609), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107409] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3136), 1,
      sym__blank,
    ACTIONS(3078), 2,
      sym_comment,
      sym__blank_line,
    ACTIONS(1905), 8,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(1907), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [107447] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3611), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3613), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107481] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3615), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3617), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107515] = 5,
    ACTIONS(3585), 1,
      sym__blank_line,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3588), 2,
      aux_sym_elif_block_token1,
      sym__blank,
    ACTIONS(3175), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3177), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [107553] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3619), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3621), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107587] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3523), 8,
      sym_label_identifier,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_EQ,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3525), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107621] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3579), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3581), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107655] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3623), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3625), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107689] = 6,
    ACTIONS(3573), 1,
      sym__blank,
    ACTIONS(3576), 1,
      sym__blank_line,
    ACTIONS(3583), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3202), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3200), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [107729] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3627), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3629), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107763] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3553), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3555), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107797] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3631), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3633), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107831] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3631), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3633), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107865] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3498), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3500), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107899] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3553), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3555), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107933] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3553), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3555), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [107967] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3635), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3637), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108001] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3631), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3633), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108035] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3635), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3637), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108069] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3639), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3641), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108103] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3202), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3200), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108137] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3643), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3645), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108171] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3647), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3649), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108205] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3514), 8,
      sym_label_identifier,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_EQ,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3516), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108239] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3651), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3653), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108273] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3631), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3633), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108307] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3342), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3344), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108341] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3655), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3657), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108375] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3659), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3661), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108409] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3663), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3665), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108443] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3659), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3661), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108477] = 6,
    ACTIONS(3583), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3667), 1,
      sym__blank_line,
    ACTIONS(3670), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3655), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3657), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [108517] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3673), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3675), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108551] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3677), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3679), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108585] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3681), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3683), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108619] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3685), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3687), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108653] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3607), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3609), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108687] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3689), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3691), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108721] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3693), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3695), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108755] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3697), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3699), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108789] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3697), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3699), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108823] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3701), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3703), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108857] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3565), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3567), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108891] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3705), 6,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BSLASH,
      anon_sym_sizeof,
      aux_sym_symbol_token2,
      sym_identifier,
    ACTIONS(3707), 19,
      sym__hll_number_literal_pre_hook,
      anon_sym_LPAREN,
      anon_sym_TILDE,
      anon_sym_BANG,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_AMP,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
  [108925] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3709), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3711), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108959] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3713), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3715), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [108993] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3717), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3719), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109027] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3721), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3723), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109061] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3725), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3727), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109095] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3729), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3731), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109129] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3733), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3735), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109163] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3693), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3695), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109197] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3693), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3695), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109231] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3693), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3695), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109265] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3717), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3719), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109299] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3737), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3739), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109333] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3356), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3358), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109367] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3741), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3743), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109401] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3745), 8,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym__blank_line,
    ACTIONS(3747), 17,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109435] = 5,
    ACTIONS(3667), 1,
      sym__blank_line,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3670), 2,
      aux_sym_elif_block_token1,
      sym__blank,
    ACTIONS(3655), 7,
      sym_label_identifier,
      ts_builtin_sym_end,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3657), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [109473] = 15,
    ACTIONS(494), 1,
      anon_sym_struct,
    ACTIONS(496), 1,
      anon_sym_union,
    ACTIONS(498), 1,
      anon_sym_enum,
    ACTIONS(3492), 1,
      sym_identifier,
    ACTIONS(3749), 1,
      anon_sym_RPAREN,
    ACTIONS(3751), 1,
      sym_hll_primitive_type,
    STATE(1607), 1,
      aux_sym__hll_declaration_specifiers_repeat1,
    STATE(1611), 1,
      sym__hll_declaration_specifiers,
    STATE(1812), 1,
      aux_sym_hll_sized_type_specifier_repeat1,
    STATE(2454), 1,
      sym_hll_parameter_declaration,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(492), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(2489), 2,
      sym_hll_type_qualifier,
      sym__hll_declaration_modifiers,
    ACTIONS(500), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1769), 6,
      sym__hll_type_specifier,
      sym_hll_struct_specifier,
      sym_hll_union_specifier,
      sym_hll_enum_specifier,
      sym_hll_sized_type_specifier,
      sym__hll_type_identifier,
  [109530] = 22,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(3755), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(3757), 1,
      aux_sym_subroutine_call_expression_token1,
    ACTIONS(3759), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2353), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2683), 1,
      sym__host_command,
    STATE(2684), 1,
      sym__other_command,
    STATE(2685), 1,
      sym__var_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3753), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
  [109601] = 4,
    ACTIONS(3136), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1905), 8,
      anon_sym_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_GT,
      anon_sym_LT,
    ACTIONS(1907), 15,
      sym__and_operator_pre_hook,
      anon_sym_LPAREN,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_CARET_CARET,
      anon_sym_PIPE_PIPE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [109636] = 22,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(35), 1,
      aux_sym__host_command_with_default_format_token1,
    ACTIONS(3759), 1,
      aux_sym__hll_non_default_format_var_command_token1,
    ACTIONS(3763), 1,
      aux_sym_parameter_declaration_token3,
    ACTIONS(3765), 1,
      aux_sym_subroutine_call_expression_token1,
    STATE(2241), 1,
      sym__command_identifier,
    STATE(2361), 1,
      sym_macro_definition_command,
    STATE(2425), 1,
      sym__hll_var_command_with_default_format,
    STATE(2429), 1,
      sym__var_command_with_non_hll_format,
    STATE(2493), 1,
      sym__host_command_with_default_format,
    STATE(2544), 1,
      sym__hll_var_definition_command_expression,
    STATE(2605), 1,
      sym__hll_var_definition_command,
    STATE(2650), 1,
      sym__hll_var_command,
    STATE(2667), 1,
      sym__other_command,
    STATE(2668), 1,
      sym__var_command,
    STATE(2670), 1,
      sym__hll_default_format_var_command,
    STATE(2695), 1,
      sym__host_command,
    STATE(2697), 1,
      sym__hll_non_default_format_var_command,
    STATE(2718), 1,
      sym__non_hll_var_command,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3761), 2,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
    ACTIONS(33), 3,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
  [109707] = 15,
    ACTIONS(494), 1,
      anon_sym_struct,
    ACTIONS(496), 1,
      anon_sym_union,
    ACTIONS(498), 1,
      anon_sym_enum,
    ACTIONS(3492), 1,
      sym_identifier,
    ACTIONS(3751), 1,
      sym_hll_primitive_type,
    ACTIONS(3767), 1,
      anon_sym_RPAREN,
    STATE(1607), 1,
      aux_sym__hll_declaration_specifiers_repeat1,
    STATE(1611), 1,
      sym__hll_declaration_specifiers,
    STATE(1812), 1,
      aux_sym_hll_sized_type_specifier_repeat1,
    STATE(2384), 1,
      sym_hll_parameter_declaration,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(492), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(2489), 2,
      sym_hll_type_qualifier,
      sym__hll_declaration_modifiers,
    ACTIONS(500), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1769), 6,
      sym__hll_type_specifier,
      sym_hll_struct_specifier,
      sym_hll_union_specifier,
      sym_hll_enum_specifier,
      sym_hll_sized_type_specifier,
      sym__hll_type_identifier,
  [109764] = 5,
    ACTIONS(3769), 1,
      sym__blank,
    STATE(1601), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2368), 7,
      sym_label_identifier,
      aux_sym_block_token1,
      anon_sym_RPAREN,
      aux_sym_while_block_token2,
      anon_sym_EQ,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(2370), 15,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
  [109801] = 5,
    ACTIONS(3776), 1,
      sym__blank_line,
    STATE(1602), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3774), 5,
      sym_label_identifier,
      aux_sym_block_token1,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
    ACTIONS(3772), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109837] = 14,
    ACTIONS(494), 1,
      anon_sym_struct,
    ACTIONS(496), 1,
      anon_sym_union,
    ACTIONS(498), 1,
      anon_sym_enum,
    ACTIONS(3492), 1,
      sym_identifier,
    ACTIONS(3751), 1,
      sym_hll_primitive_type,
    STATE(1607), 1,
      aux_sym__hll_declaration_specifiers_repeat1,
    STATE(1611), 1,
      sym__hll_declaration_specifiers,
    STATE(1812), 1,
      aux_sym_hll_sized_type_specifier_repeat1,
    STATE(2518), 1,
      sym_hll_parameter_declaration,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(492), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(2489), 2,
      sym_hll_type_qualifier,
      sym__hll_declaration_modifiers,
    ACTIONS(500), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1769), 6,
      sym__hll_type_specifier,
      sym_hll_struct_specifier,
      sym_hll_union_specifier,
      sym_hll_enum_specifier,
      sym_hll_sized_type_specifier,
      sym__hll_type_identifier,
  [109891] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3779), 1,
      sym__blank_line,
    STATE(1604), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3774), 6,
      sym_label_identifier,
      aux_sym_block_token1,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
    ACTIONS(3772), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [109927] = 13,
    ACTIONS(494), 1,
      anon_sym_struct,
    ACTIONS(496), 1,
      anon_sym_union,
    ACTIONS(498), 1,
      anon_sym_enum,
    ACTIONS(3492), 1,
      sym_identifier,
    ACTIONS(3782), 1,
      sym_hll_primitive_type,
    ACTIONS(3784), 1,
      sym__blank,
    STATE(1614), 1,
      aux_sym_block_repeat1,
    STATE(1812), 1,
      aux_sym_hll_sized_type_specifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(492), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1649), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat1,
    ACTIONS(500), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1620), 6,
      sym__hll_type_specifier,
      sym_hll_struct_specifier,
      sym_hll_union_specifier,
      sym_hll_enum_specifier,
      sym_hll_sized_type_specifier,
      sym__hll_type_identifier,
  [109978] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(3788), 6,
      sym_label_identifier,
      aux_sym_while_block_token2,
      anon_sym_AMP_AMP,
      aux_sym_macro_definition_token1,
      sym_comment,
      sym__blank_line,
    ACTIONS(3786), 16,
      aux_sym_subroutine_block_token1,
      aux_sym_if_block_token1,
      aux_sym_while_block_token1,
      aux_sym__repeat_block_pre_condition_format_token1,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token1,
      aux_sym_parameter_declaration_token2,
      aux_sym_parameter_declaration_token3,
      aux_sym_subroutine_call_expression_token1,
      aux_sym_macro_definition_command_token1,
      aux_sym_macro_definition_command_token2,
      aux_sym_macro_definition_command_token3,
      aux_sym__host_command_with_default_format_token1,
      aux_sym__hll_non_default_format_var_command_token1,
      sym_identifier,
      sym__blank,
  [110008] = 12,
    ACTIONS(494), 1,
      anon_sym_struct,
    ACTIONS(496), 1,
      anon_sym_union,
    ACTIONS(498), 1,
      anon_sym_enum,
    ACTIONS(3492), 1,
      sym_identifier,
    ACTIONS(3790), 1,
      sym_hll_primitive_type,
    STATE(1651), 1,
      aux_sym__hll_declaration_specifiers_repeat1,
    STATE(1812), 1,
      aux_sym_hll_sized_type_specifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(492), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(2489), 2,
      sym_hll_type_qualifier,
      sym__hll_declaration_modifiers,
    ACTIONS(500), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1764), 6,
      sym__hll_type_specifier,
      sym_hll_struct_specifier,
      sym_hll_union_specifier,
      sym_hll_enum_specifier,
      sym_hll_sized_type_specifier,
      sym__hll_type_identifier,
  [110056] = 12,
    ACTIONS(494), 1,
      anon_sym_struct,
    ACTIONS(496), 1,
      anon_sym_union,
    ACTIONS(498), 1,
      anon_sym_enum,
    ACTIONS(3492), 1,
      sym_identifier,
    ACTIONS(3792), 1,
      sym_hll_primitive_type,
    STATE(1651), 1,
      aux_sym__hll_declaration_specifiers_repeat1,
    STATE(1812), 1,
      aux_sym_hll_sized_type_specifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(492), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(2489), 2,
      sym_hll_type_qualifier,
      sym__hll_declaration_modifiers,
    ACTIONS(500), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1805), 6,
      sym__hll_type_specifier,
      sym_hll_struct_specifier,
      sym_hll_union_specifier,
      sym_hll_enum_specifier,
      sym_hll_sized_type_specifier,
      sym__hll_type_identifier,
  [110104] = 11,
    ACTIONS(3796), 1,
      anon_sym_LPAREN,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3802), 1,
      anon_sym_LBRACK,
    ACTIONS(3804), 1,
      sym__blank,
    STATE(1609), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1610), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3794), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1795), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
    STATE(1881), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [110149] = 11,
    ACTIONS(3796), 1,
      anon_sym_LPAREN,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3802), 1,
      anon_sym_LBRACK,
    ACTIONS(3808), 1,
      sym__blank,
    STATE(1609), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3806), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1815), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
    STATE(1895), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [110194] = 9,
    ACTIONS(3796), 1,
      anon_sym_LPAREN,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3802), 1,
      anon_sym_LBRACK,
    STATE(1609), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3810), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1780), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
    STATE(1922), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [110233] = 10,
    ACTIONS(3796), 1,
      anon_sym_LPAREN,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3802), 1,
      anon_sym_LBRACK,
    ACTIONS(3812), 1,
      sym__blank,
    STATE(1609), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1615), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1915), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
    STATE(1943), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [110274] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3102), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
    ACTIONS(3100), 13,
      anon_sym_BSLASH,
      anon_sym_const,
      anon_sym_volatile,
      anon_sym_struct,
      anon_sym_union,
      anon_sym_enum,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_hll_primitive_type,
      sym_identifier,
      sym__blank,
  [110301] = 11,
    ACTIONS(494), 1,
      anon_sym_struct,
    ACTIONS(496), 1,
      anon_sym_union,
    ACTIONS(498), 1,
      anon_sym_enum,
    ACTIONS(3492), 1,
      sym_identifier,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(3814), 1,
      sym_hll_primitive_type,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(1812), 1,
      aux_sym_hll_sized_type_specifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(500), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
    STATE(1624), 6,
      sym__hll_type_specifier,
      sym_hll_struct_specifier,
      sym_hll_union_specifier,
      sym_hll_enum_specifier,
      sym_hll_sized_type_specifier,
      sym__hll_type_identifier,
  [110344] = 10,
    ACTIONS(3796), 1,
      anon_sym_LPAREN,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3802), 1,
      anon_sym_LBRACK,
    ACTIONS(3808), 1,
      sym__blank,
    STATE(1609), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1852), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
    STATE(1953), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [110385] = 7,
    ACTIONS(3822), 1,
      sym__decimal_number_pre_hook,
    ACTIONS(3824), 1,
      sym_hll_number_literal,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3820), 2,
      aux_sym_integer_token1,
      aux_sym_integer_token2,
    STATE(2716), 3,
      sym_hll_char_literal,
      sym_hll_string_literal,
      sym_integer,
    ACTIONS(3816), 5,
      anon_sym_L_SQUOTE,
      anon_sym_u_SQUOTE,
      anon_sym_U_SQUOTE,
      anon_sym_u8_SQUOTE,
      anon_sym_SQUOTE,
    ACTIONS(3818), 5,
      anon_sym_L_DQUOTE,
      anon_sym_u_DQUOTE,
      anon_sym_U_DQUOTE,
      anon_sym_u8_DQUOTE,
      anon_sym_DQUOTE,
  [110419] = 11,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3826), 1,
      anon_sym_RPAREN,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    ACTIONS(3835), 1,
      sym__blank,
    STATE(1650), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1753), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    STATE(1926), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [110460] = 11,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    ACTIONS(3839), 1,
      anon_sym_RPAREN,
    ACTIONS(3841), 1,
      sym__blank,
    STATE(1660), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1753), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    STATE(1853), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [110501] = 11,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    ACTIONS(3844), 1,
      anon_sym_RPAREN,
    ACTIONS(3846), 1,
      sym__blank,
    STATE(1652), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1753), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    STATE(1829), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [110542] = 11,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    ACTIONS(3849), 1,
      anon_sym_RPAREN,
    ACTIONS(3851), 1,
      sym__blank,
    STATE(1653), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1619), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    STATE(1914), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [110583] = 11,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    ACTIONS(3854), 1,
      anon_sym_RPAREN,
    ACTIONS(3857), 1,
      sym__blank,
    STATE(1657), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1617), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    STATE(1923), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [110624] = 11,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    ACTIONS(3861), 1,
      anon_sym_RPAREN,
    ACTIONS(3863), 1,
      sym__blank,
    STATE(1647), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1618), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    STATE(1913), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [110665] = 11,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    ACTIONS(3866), 1,
      anon_sym_RPAREN,
    ACTIONS(3868), 1,
      sym__blank,
    STATE(1654), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1753), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    STATE(1841), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [110706] = 11,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    ACTIONS(3871), 1,
      anon_sym_RPAREN,
    ACTIONS(3873), 1,
      sym__blank,
    STATE(1643), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1623), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    STATE(1892), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [110747] = 6,
    ACTIONS(3880), 1,
      anon_sym_SLASH,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3878), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3876), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1961), 4,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1963), 5,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_COLON_COLON_COLON,
  [110777] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1961), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1963), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_DOT_DOT,
      anon_sym_COLON_COLON_COLON,
  [110801] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3884), 1,
      anon_sym_LPAREN,
    ACTIONS(3887), 1,
      anon_sym_LBRACK,
    ACTIONS(3890), 1,
      sym__blank,
    STATE(1662), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3882), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1837), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [110839] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2051), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(2053), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_DOT_DOT,
      anon_sym_COLON_COLON_COLON,
  [110863] = 4,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3876), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1961), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1963), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_COLON_COLON_COLON,
  [110889] = 8,
    ACTIONS(3880), 1,
      anon_sym_SLASH,
    ACTIONS(3895), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1961), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(3878), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3893), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3876), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1963), 4,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_COLON_COLON_COLON,
  [110923] = 7,
    ACTIONS(3880), 1,
      anon_sym_SLASH,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1961), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(3878), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3893), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3876), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1963), 5,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_COLON_COLON_COLON,
  [110955] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3899), 1,
      anon_sym_LPAREN,
    ACTIONS(3902), 1,
      anon_sym_LBRACK,
    ACTIONS(3905), 1,
      sym__blank,
    STATE(1675), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3897), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1862), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [110993] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3910), 1,
      anon_sym_LPAREN,
    ACTIONS(3913), 1,
      anon_sym_LBRACK,
    ACTIONS(3916), 1,
      sym__blank,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1742), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3908), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1836), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [111031] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3921), 1,
      anon_sym_LPAREN,
    ACTIONS(3924), 1,
      anon_sym_LBRACK,
    ACTIONS(3927), 1,
      sym__blank,
    STATE(1665), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3919), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1846), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [111069] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3932), 1,
      anon_sym_LPAREN,
    ACTIONS(3935), 1,
      anon_sym_LBRACK,
    ACTIONS(3938), 1,
      sym__blank,
    STATE(1666), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3930), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1847), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [111107] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1949), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1951), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_DOT_DOT,
      anon_sym_COLON_COLON_COLON,
  [111131] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3943), 1,
      anon_sym_LPAREN,
    ACTIONS(3946), 1,
      anon_sym_LBRACK,
    ACTIONS(3949), 1,
      sym__blank,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1739), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3941), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1894), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [111169] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3954), 1,
      anon_sym_LPAREN,
    ACTIONS(3957), 1,
      anon_sym_LBRACK,
    ACTIONS(3960), 1,
      sym__blank,
    STATE(1668), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3952), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1851), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [111207] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3965), 1,
      anon_sym_LPAREN,
    ACTIONS(3968), 1,
      anon_sym_LBRACK,
    ACTIONS(3971), 1,
      sym__blank,
    STATE(1673), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3963), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1858), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [111245] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3976), 1,
      anon_sym_LPAREN,
    ACTIONS(3979), 1,
      anon_sym_LBRACK,
    ACTIONS(3982), 1,
      sym__blank,
    STATE(1670), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3974), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1849), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [111283] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3987), 1,
      anon_sym_LPAREN,
    ACTIONS(3990), 1,
      anon_sym_LBRACK,
    ACTIONS(3993), 1,
      sym__blank,
    STATE(1671), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3985), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1856), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [111321] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3998), 1,
      anon_sym_LPAREN,
    ACTIONS(4001), 1,
      anon_sym_LBRACK,
    ACTIONS(4004), 1,
      sym__blank,
    STATE(1672), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3996), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    STATE(1857), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [111359] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1799), 1,
      sym_hll_type_qualifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1841), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [111396] = 11,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3880), 1,
      anon_sym_SLASH,
    ACTIONS(3895), 1,
      anon_sym_CARET,
    ACTIONS(4007), 1,
      anon_sym_RBRACE,
    ACTIONS(4009), 1,
      anon_sym_PIPE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3878), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3893), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3876), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [111435] = 11,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3880), 1,
      anon_sym_SLASH,
    ACTIONS(3895), 1,
      anon_sym_CARET,
    ACTIONS(4009), 1,
      anon_sym_PIPE,
    ACTIONS(4011), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3878), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3893), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3876), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [111474] = 11,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3880), 1,
      anon_sym_SLASH,
    ACTIONS(3895), 1,
      anon_sym_CARET,
    ACTIONS(4009), 1,
      anon_sym_PIPE,
    ACTIONS(4013), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3878), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3893), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3876), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [111513] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1799), 1,
      sym_hll_type_qualifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1853), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [111550] = 11,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3880), 1,
      anon_sym_SLASH,
    ACTIONS(3895), 1,
      anon_sym_CARET,
    ACTIONS(4009), 1,
      anon_sym_PIPE,
    ACTIONS(4015), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3878), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3893), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3876), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [111589] = 4,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4019), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1649), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat1,
    ACTIONS(4017), 10,
      anon_sym_struct,
      anon_sym_union,
      anon_sym_enum,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_hll_primitive_type,
      sym_identifier,
      sym__blank,
  [111614] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1799), 1,
      sym_hll_type_qualifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1839), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [111651] = 5,
    STATE(1651), 1,
      aux_sym__hll_declaration_specifiers_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4024), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(2489), 2,
      sym_hll_type_qualifier,
      sym__hll_declaration_modifiers,
    ACTIONS(4022), 9,
      anon_sym_struct,
      anon_sym_union,
      anon_sym_enum,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_hll_primitive_type,
      sym_identifier,
  [111678] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1799), 1,
      sym_hll_type_qualifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1840), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [111715] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1799), 1,
      sym_hll_type_qualifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1829), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [111752] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1799), 1,
      sym_hll_type_qualifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1880), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [111789] = 11,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3880), 1,
      anon_sym_SLASH,
    ACTIONS(3895), 1,
      anon_sym_CARET,
    ACTIONS(4009), 1,
      anon_sym_PIPE,
    ACTIONS(4015), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3878), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3893), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3876), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [111828] = 11,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3880), 1,
      anon_sym_SLASH,
    ACTIONS(3895), 1,
      anon_sym_CARET,
    ACTIONS(4009), 1,
      anon_sym_PIPE,
    ACTIONS(4013), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3878), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3893), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3876), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [111867] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1799), 1,
      sym_hll_type_qualifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1926), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [111904] = 11,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3880), 1,
      anon_sym_SLASH,
    ACTIONS(3895), 1,
      anon_sym_CARET,
    ACTIONS(4009), 1,
      anon_sym_PIPE,
    ACTIONS(4011), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3878), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3893), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3876), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [111943] = 11,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3880), 1,
      anon_sym_SLASH,
    ACTIONS(3895), 1,
      anon_sym_CARET,
    ACTIONS(4007), 1,
      anon_sym_RPAREN,
    ACTIONS(4009), 1,
      anon_sym_PIPE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3878), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(3893), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3876), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [111982] = 10,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1799), 1,
      sym_hll_type_qualifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1879), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [112019] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4031), 1,
      sym__blank,
    STATE(1673), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1858), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112053] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1849), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112087] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4033), 1,
      sym__blank,
    STATE(1668), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1851), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112121] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4035), 1,
      sym__blank,
    STATE(1670), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1849), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112155] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1855), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112189] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1856), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112223] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4037), 1,
      sym__blank,
    STATE(1671), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1856), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112257] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1857), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112291] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4039), 1,
      sym__blank,
    STATE(1672), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1857), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112325] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1859), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112359] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1860), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112393] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1861), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112427] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1862), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112461] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4041), 1,
      sym__blank,
    STATE(1675), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1862), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112495] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1863), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112529] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4049), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1721), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1924), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112563] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1854), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112597] = 10,
    ACTIONS(1981), 1,
      anon_sym_COLON,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4057), 1,
      anon_sym_SLASH,
    ACTIONS(4059), 1,
      anon_sym_PIPE,
    ACTIONS(4061), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4051), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4055), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(4053), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [112633] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4063), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1719), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1842), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112667] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1828), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112701] = 4,
    ACTIONS(4067), 1,
      sym__blank,
    STATE(1686), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4065), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [112725] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1890), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112759] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1843), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112793] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4069), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1688), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1843), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112827] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4071), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1677), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1848), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112861] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4073), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [112885] = 4,
    ACTIONS(4077), 1,
      sym__blank,
    STATE(1700), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4075), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [112909] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1866), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [112943] = 6,
    ACTIONS(4057), 1,
      anon_sym_SLASH,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4055), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(1963), 3,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(4053), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1961), 4,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
  [112971] = 4,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4053), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
    ACTIONS(1961), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1963), 5,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_COLON_COLON_COLON,
  [112995] = 4,
    ACTIONS(4081), 1,
      sym__blank,
    STATE(1699), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4079), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113019] = 8,
    ACTIONS(4057), 1,
      anon_sym_SLASH,
    ACTIONS(4061), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1961), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(1963), 2,
      anon_sym_PIPE,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(4051), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4055), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(4053), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [113051] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1867), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113085] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4083), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1728), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1867), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113119] = 7,
    ACTIONS(4057), 1,
      anon_sym_SLASH,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1961), 2,
      anon_sym_COLON,
      anon_sym_COLON_COLON,
    ACTIONS(4051), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4055), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(1963), 3,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(4053), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [113149] = 10,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(4057), 1,
      anon_sym_SLASH,
    ACTIONS(4059), 1,
      anon_sym_PIPE,
    ACTIONS(4061), 1,
      anon_sym_CARET,
    ACTIONS(4085), 1,
      anon_sym_COLON,
    ACTIONS(4087), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4051), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4055), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(4053), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [113185] = 10,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4057), 1,
      anon_sym_SLASH,
    ACTIONS(4059), 1,
      anon_sym_PIPE,
    ACTIONS(4061), 1,
      anon_sym_CARET,
    ACTIONS(4089), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4051), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4055), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(4053), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [113221] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4091), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1729), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1870), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113255] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4093), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113279] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4095), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113303] = 4,
    ACTIONS(4099), 1,
      sym__blank,
    STATE(1702), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4097), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113327] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4101), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113351] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1873), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113385] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4103), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1731), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1874), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113419] = 4,
    ACTIONS(4107), 1,
      sym__blank,
    STATE(1706), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4105), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113443] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4109), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113467] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4111), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1737), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1873), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113501] = 4,
    ACTIONS(4115), 1,
      sym__blank,
    STATE(1744), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4113), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113525] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4117), 1,
      sym__blank,
    STATE(1666), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1847), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113559] = 4,
    ACTIONS(4121), 1,
      sym__blank,
    STATE(1711), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4119), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113583] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4123), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113607] = 4,
    ACTIONS(4127), 1,
      sym__blank,
    STATE(1713), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4125), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113631] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4129), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113655] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4131), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1693), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1877), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113689] = 5,
    ACTIONS(4133), 1,
      sym__blank,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2370), 4,
      anon_sym_BSLASH,
      anon_sym_const,
      anon_sym_volatile,
      sym_identifier,
    ACTIONS(2368), 7,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_COMMA,
      anon_sym_QMARK,
      anon_sym_LBRACK,
  [113715] = 4,
    ACTIONS(4138), 1,
      sym__blank,
    STATE(1717), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4136), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113739] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4140), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113763] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4142), 1,
      sym__blank,
    STATE(1665), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1846), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113797] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1838), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113831] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4144), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1703), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1838), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113865] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1927), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113899] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4146), 1,
      sym__blank,
    STATE(1723), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1943), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113933] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1953), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [113967] = 4,
    ACTIONS(4150), 1,
      sym__blank,
    STATE(1726), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4148), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [113991] = 10,
    ACTIONS(1983), 1,
      anon_sym_COLON_COLON_COLON,
    ACTIONS(1985), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4057), 1,
      anon_sym_SLASH,
    ACTIONS(4059), 1,
      anon_sym_PIPE,
    ACTIONS(4061), 1,
      anon_sym_CARET,
    ACTIONS(4152), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4051), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(4055), 2,
      anon_sym_STAR,
      anon_sym_PERCENT,
    ACTIONS(4053), 3,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_DOT_DOT,
  [114027] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4154), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [114051] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4156), 1,
      sym__blank,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1736), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1881), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114085] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1830), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114119] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1833), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114153] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4158), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1680), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1833), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114187] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1869), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114221] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4160), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1682), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1869), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114255] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4022), 11,
      anon_sym_const,
      anon_sym_volatile,
      anon_sym_struct,
      anon_sym_union,
      anon_sym_enum,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_hll_primitive_type,
      sym_identifier,
  [114279] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4162), 1,
      sym__blank,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1739), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1894), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114313] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    ACTIONS(4164), 1,
      sym__blank,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1683), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1876), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114347] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1895), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114381] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4043), 1,
      anon_sym_LPAREN,
    ACTIONS(4045), 1,
      anon_sym_BSLASH,
    ACTIONS(4047), 1,
      anon_sym_LBRACK,
    STATE(1676), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1884), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114415] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4166), 1,
      sym__blank,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1742), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1836), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114449] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1837), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114483] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4168), 1,
      sym__blank,
    STATE(1662), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1837), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114517] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    ACTIONS(4170), 1,
      sym__blank,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1743), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1929), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114551] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1846), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114585] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3800), 1,
      anon_sym_BSLASH,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4027), 1,
      anon_sym_LPAREN,
    ACTIONS(4029), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1727), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1961), 6,
      sym_trace32_hll_variable,
      sym__hll_declarator,
      sym_hll_pointer_declarator,
      sym_hll_function_declarator,
      sym_hll_array_declarator,
      sym_hll_parenthesized_declarator,
  [114619] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4172), 11,
      anon_sym_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_CARET_EQ,
      anon_sym_PIPE_EQ,
  [114643] = 9,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3806), 1,
      anon_sym_RPAREN,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1815), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [114676] = 9,
    ACTIONS(3794), 1,
      anon_sym_RPAREN,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    ACTIONS(4174), 1,
      sym__blank,
    STATE(1745), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1795), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [114709] = 11,
    ACTIONS(4176), 1,
      anon_sym_BSLASH,
    ACTIONS(4178), 1,
      anon_sym_DQUOTE,
    ACTIONS(4180), 1,
      aux_sym_symbol_token1,
    ACTIONS(4182), 1,
      aux_sym_symbol_token2,
    ACTIONS(4186), 1,
      sym__blank,
    STATE(1772), 1,
      aux_sym_block_repeat1,
    STATE(2602), 1,
      sym_trace32_hll_variable,
    STATE(2612), 1,
      sym__hll_var_command_argument_list_non_default_format,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4184), 2,
      sym_path,
      sym_file_handle,
    STATE(2536), 2,
      sym_string,
      sym_symbol,
  [114746] = 12,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4188), 1,
      sym_identifier,
    ACTIONS(4192), 1,
      anon_sym_PERCENT,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1763), 1,
      aux_sym__hll_var_pattern_command_argument_list_format_repeat1,
    STATE(1886), 1,
      aux_sym__hll_var_pattern_command_argument_list_format_repeat2,
    STATE(1887), 1,
      aux_sym_hll_type_pattern_repeat1,
    STATE(2541), 1,
      sym_format_expression,
    STATE(2542), 1,
      sym_hll_type_pattern,
    STATE(2612), 1,
      sym__hll_var_pattern_command_argument_list_format,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4190), 2,
      anon_sym_STAR,
      anon_sym_QMARK,
  [114785] = 8,
    ACTIONS(4196), 1,
      anon_sym_BSLASH,
    ACTIONS(4198), 1,
      sym__blank,
    STATE(1754), 1,
      sym_hll_type_qualifier,
    STATE(1893), 1,
      aux_sym_block_repeat1,
    STATE(2097), 1,
      aux_sym__hll_pointer_declarator_specifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(4194), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [114816] = 8,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    ACTIONS(4201), 1,
      sym__blank,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    STATE(1752), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1915), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [114846] = 7,
    ACTIONS(4203), 1,
      sym_identifier,
    ACTIONS(4205), 1,
      aux_sym_if_block_token1,
    ACTIONS(4209), 1,
      aux_sym__hll_non_default_format_var_command_token5,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4213), 2,
      aux_sym__hll_var_definition_command_expression_token1,
      aux_sym__hll_var_definition_command_expression_token2,
    ACTIONS(4207), 3,
      aux_sym__hll_non_default_format_var_command_token2,
      aux_sym__hll_non_default_format_var_command_token3,
      aux_sym__hll_non_default_format_var_command_token4,
    ACTIONS(4211), 3,
      aux_sym__var_command_with_non_hll_format_token1,
      aux_sym__var_command_with_non_hll_format_token2,
      aux_sym__var_command_with_non_hll_format_token3,
  [114874] = 8,
    ACTIONS(3798), 1,
      anon_sym_STAR,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(3829), 1,
      anon_sym_LPAREN,
    ACTIONS(3831), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(1746), 1,
      sym__hll_pointer_declarator_specifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1852), 5,
      sym__hll_abstract_declarator,
      sym_hll_abstract_pointer_declarator,
      sym_hll_abstract_function_declarator,
      sym_hll_abstract_array_declarator,
      sym_hll_abstract_parenthesized_declarator,
  [114904] = 6,
    ACTIONS(4218), 1,
      sym__blank,
    STATE(2007), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4215), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1753), 2,
      sym_hll_type_qualifier,
      aux_sym_hll_type_descriptor_repeat2,
    ACTIONS(2799), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_LBRACK,
  [114929] = 6,
    ACTIONS(4223), 1,
      anon_sym_BSLASH,
    ACTIONS(4227), 1,
      sym__blank,
    STATE(2072), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4225), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(4221), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [114954] = 6,
    ACTIONS(4232), 1,
      anon_sym_BSLASH,
    ACTIONS(4234), 1,
      sym__blank,
    STATE(2072), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4225), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(4230), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [114979] = 6,
    ACTIONS(4203), 1,
      sym_identifier,
    ACTIONS(4209), 1,
      aux_sym__hll_non_default_format_var_command_token5,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4213), 2,
      aux_sym__hll_var_definition_command_expression_token1,
      aux_sym__hll_var_definition_command_expression_token2,
    ACTIONS(4207), 3,
      aux_sym__hll_non_default_format_var_command_token2,
      aux_sym__hll_non_default_format_var_command_token3,
      aux_sym__hll_non_default_format_var_command_token4,
    ACTIONS(4211), 3,
      aux_sym__var_command_with_non_hll_format_token1,
      aux_sym__var_command_with_non_hll_format_token2,
      aux_sym__var_command_with_non_hll_format_token3,
  [115004] = 5,
    ACTIONS(4239), 1,
      anon_sym_BSLASH,
    ACTIONS(4241), 1,
      sym__blank,
    STATE(1803), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4237), 7,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_const,
      anon_sym_volatile,
  [115027] = 6,
    ACTIONS(4246), 1,
      anon_sym_BSLASH,
    ACTIONS(4248), 1,
      sym__blank,
    STATE(2072), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4225), 2,
      anon_sym_const,
      anon_sym_volatile,
    ACTIONS(4244), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [115052] = 5,
    ACTIONS(4253), 1,
      anon_sym_BSLASH,
    ACTIONS(4255), 1,
      sym__blank,
    STATE(1803), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4251), 7,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_const,
      anon_sym_volatile,
  [115075] = 6,
    ACTIONS(4260), 1,
      anon_sym_BSLASH,
    ACTIONS(4262), 1,
      sym__blank,
    STATE(1765), 1,
      aux_sym__hll_declaration_specifiers_repeat2,
    STATE(1888), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4258), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [115099] = 6,
    ACTIONS(4262), 1,
      sym__blank,
    ACTIONS(4266), 1,
      anon_sym_BSLASH,
    STATE(1765), 1,
      aux_sym__hll_declaration_specifiers_repeat2,
    STATE(1888), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4264), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [115123] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4270), 2,
      anon_sym_BSLASH,
      sym__blank,
    ACTIONS(4268), 7,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_const,
      anon_sym_volatile,
  [115141] = 9,
    ACTIONS(4188), 1,
      sym_identifier,
    ACTIONS(4192), 1,
      anon_sym_PERCENT,
    STATE(1820), 1,
      aux_sym__hll_var_pattern_command_argument_list_format_repeat2,
    STATE(1887), 1,
      aux_sym_hll_type_pattern_repeat1,
    STATE(1918), 1,
      aux_sym__hll_var_pattern_command_argument_list_format_repeat1,
    STATE(2497), 1,
      sym_hll_type_pattern,
    STATE(2541), 1,
      sym_format_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4190), 2,
      anon_sym_STAR,
      anon_sym_QMARK,
  [115171] = 6,
    ACTIONS(4262), 1,
      sym__blank,
    ACTIONS(4274), 1,
      anon_sym_BSLASH,
    STATE(1760), 1,
      aux_sym__hll_declaration_specifiers_repeat2,
    STATE(1888), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4272), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [115195] = 6,
    ACTIONS(4278), 1,
      anon_sym_BSLASH,
    ACTIONS(4280), 1,
      sym__blank,
    STATE(1765), 1,
      aux_sym__hll_declaration_specifiers_repeat2,
    STATE(1888), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4276), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [115219] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4285), 2,
      anon_sym_BSLASH,
      sym__blank,
    ACTIONS(4283), 7,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_const,
      anon_sym_volatile,
  [115237] = 9,
    ACTIONS(4287), 1,
      anon_sym_AMP,
    ACTIONS(4289), 1,
      aux_sym__host_command_argument_list_default_format_token1,
    ACTIONS(4291), 1,
      anon_sym_DQUOTE,
    ACTIONS(4293), 1,
      sym__blank,
    STATE(1868), 1,
      aux_sym__host_command_argument_list_default_format_repeat1,
    STATE(2039), 1,
      aux_sym_block_repeat1,
    STATE(2609), 1,
      sym__host_command_argument_list_default_format,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(2168), 2,
      sym_macro,
      sym_string,
  [115267] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4297), 2,
      anon_sym_BSLASH,
      sym__blank,
    ACTIONS(4295), 7,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_const,
      anon_sym_volatile,
  [115285] = 6,
    ACTIONS(4262), 1,
      sym__blank,
    ACTIONS(4301), 1,
      anon_sym_BSLASH,
    STATE(1761), 1,
      aux_sym__hll_declaration_specifiers_repeat2,
    STATE(1888), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4299), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [115309] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4303), 2,
      anon_sym_BSLASH,
      sym__blank,
    ACTIONS(2669), 7,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_const,
      anon_sym_volatile,
  [115327] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4307), 2,
      anon_sym_BSLASH,
      sym__blank,
    ACTIONS(4305), 7,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
      anon_sym_const,
      anon_sym_volatile,
  [115345] = 5,
    ACTIONS(4309), 1,
      sym__blank,
    STATE(1772), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2370), 2,
      anon_sym_BSLASH,
      aux_sym_symbol_token2,
    ACTIONS(2368), 4,
      sym_path,
      anon_sym_DQUOTE,
      aux_sym_symbol_token1,
      sym_file_handle,
  [115366] = 9,
    ACTIONS(3202), 1,
      aux_sym_while_block_token1,
    ACTIONS(3207), 1,
      sym__blank_line,
    ACTIONS(3210), 1,
      sym__blank,
    ACTIONS(4312), 1,
      aux_sym_elif_block_token1,
    STATE(1899), 1,
      aux_sym_if_block_repeat2,
    STATE(1900), 1,
      aux_sym_if_block_repeat3,
    STATE(2120), 1,
      sym_else_block,
    STATE(2169), 1,
      sym_elif_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [115395] = 9,
    ACTIONS(3189), 1,
      aux_sym_while_block_token1,
    ACTIONS(3194), 1,
      sym__blank_line,
    ACTIONS(3197), 1,
      sym__blank,
    ACTIONS(4315), 1,
      aux_sym_elif_block_token1,
    STATE(1897), 1,
      aux_sym_if_block_repeat2,
    STATE(1898), 1,
      aux_sym_if_block_repeat3,
    STATE(2091), 1,
      sym_else_block,
    STATE(2207), 1,
      sym_elif_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [115424] = 9,
    ACTIONS(3175), 1,
      aux_sym_while_block_token1,
    ACTIONS(3181), 1,
      sym__blank_line,
    ACTIONS(3184), 1,
      sym__blank,
    ACTIONS(4318), 1,
      aux_sym_elif_block_token1,
    STATE(1901), 1,
      aux_sym_if_block_repeat2,
    STATE(1902), 1,
      aux_sym_if_block_repeat3,
    STATE(2085), 1,
      sym_elif_block,
    STATE(2143), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [115453] = 8,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    ACTIONS(458), 1,
      anon_sym_PERCENT,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(1994), 1,
      aux_sym__var_command_argument_list_non_hll_format_repeat1,
    STATE(2726), 1,
      sym__var_command_argument_list_non_hll_format,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(2275), 2,
      sym_option_expression,
      sym_format_expression,
  [115480] = 8,
    ACTIONS(4287), 1,
      anon_sym_AMP,
    ACTIONS(4291), 1,
      anon_sym_DQUOTE,
    ACTIONS(4293), 1,
      sym__blank,
    ACTIONS(4321), 1,
      aux_sym__host_command_argument_list_default_format_token1,
    STATE(2039), 1,
      aux_sym_block_repeat1,
    STATE(2192), 1,
      aux_sym__host_command_argument_list_default_format_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(2547), 2,
      sym_macro,
      sym_string,
  [115507] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4325), 1,
      anon_sym_DQUOTE,
    ACTIONS(4327), 1,
      aux_sym_string_token1,
    ACTIONS(4329), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1787), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115529] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4333), 1,
      anon_sym_DQUOTE,
    ACTIONS(4335), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4337), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1781), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115551] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4339), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [115575] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4347), 1,
      anon_sym_DQUOTE,
    ACTIONS(4349), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4351), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115597] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4353), 1,
      anon_sym_DQUOTE,
    ACTIONS(4355), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4357), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1800), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115619] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4359), 1,
      anon_sym_DQUOTE,
    ACTIONS(4361), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4363), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1784), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115641] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4349), 1,
      aux_sym_string_token1,
    ACTIONS(4365), 1,
      anon_sym_DQUOTE,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4351), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115663] = 6,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4367), 1,
      sym__blank,
    STATE(1824), 1,
      aux_sym_block_repeat1,
    STATE(2200), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3342), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [115685] = 6,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4369), 1,
      sym__blank,
    STATE(1825), 1,
      aux_sym_block_repeat1,
    STATE(2202), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3342), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [115707] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4349), 1,
      aux_sym_string_token1,
    ACTIONS(4371), 1,
      anon_sym_DQUOTE,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4351), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115729] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4373), 1,
      anon_sym_DQUOTE,
    ACTIONS(4375), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4377), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1789), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115751] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4349), 1,
      aux_sym_string_token1,
    ACTIONS(4379), 1,
      anon_sym_DQUOTE,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4351), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115773] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4381), 1,
      anon_sym_DQUOTE,
    ACTIONS(4383), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4385), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1793), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115795] = 8,
    ACTIONS(3194), 1,
      sym__blank_line,
    ACTIONS(3197), 1,
      sym__blank,
    ACTIONS(4387), 1,
      aux_sym_elif_block_token1,
    STATE(1882), 1,
      aux_sym_if_block_repeat2,
    STATE(1883), 1,
      aux_sym_if_block_repeat3,
    STATE(2091), 1,
      sym_else_block,
    STATE(2207), 1,
      sym_elif_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [115821] = 6,
    ACTIONS(4260), 1,
      anon_sym_BSLASH,
    ACTIONS(4390), 1,
      sym__blank,
    STATE(1765), 1,
      aux_sym__hll_declaration_specifiers_repeat2,
    STATE(1888), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4258), 3,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_LBRACK,
  [115843] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4349), 1,
      aux_sym_string_token1,
    ACTIONS(4393), 1,
      anon_sym_DQUOTE,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4351), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115865] = 8,
    ACTIONS(3207), 1,
      sym__blank_line,
    ACTIONS(3210), 1,
      sym__blank,
    ACTIONS(4395), 1,
      aux_sym_elif_block_token1,
    STATE(1904), 1,
      aux_sym_if_block_repeat2,
    STATE(1911), 1,
      aux_sym_if_block_repeat3,
    STATE(2120), 1,
      sym_else_block,
    STATE(2169), 1,
      sym_elif_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [115891] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4398), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [115915] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4400), 1,
      anon_sym_DQUOTE,
    ACTIONS(4402), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4404), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1818), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115937] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4406), 1,
      anon_sym_DQUOTE,
    ACTIONS(4408), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4410), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1808), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115959] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4349), 1,
      aux_sym_string_token1,
    ACTIONS(4412), 1,
      anon_sym_DQUOTE,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4351), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [115981] = 3,
    ACTIONS(2797), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2799), 6,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_const,
      anon_sym_volatile,
  [115997] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4349), 1,
      aux_sym_string_token1,
    ACTIONS(4414), 1,
      anon_sym_DQUOTE,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4351), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [116019] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4416), 1,
      anon_sym_DQUOTE,
    ACTIONS(4418), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4420), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1802), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [116041] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4349), 1,
      aux_sym_string_token1,
    ACTIONS(4422), 1,
      anon_sym_DQUOTE,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4351), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [116063] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4424), 5,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
      sym_identifier,
  [116081] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4426), 1,
      anon_sym_DQUOTE,
    ACTIONS(4428), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4430), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1806), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [116103] = 6,
    ACTIONS(4274), 1,
      anon_sym_BSLASH,
    ACTIONS(4432), 1,
      sym__blank,
    STATE(1792), 1,
      aux_sym__hll_declaration_specifiers_repeat2,
    STATE(1888), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4272), 3,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_LBRACK,
  [116125] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4349), 1,
      aux_sym_string_token1,
    ACTIONS(4435), 1,
      anon_sym_DQUOTE,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4351), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [116147] = 6,
    ACTIONS(4266), 1,
      anon_sym_BSLASH,
    ACTIONS(4437), 1,
      sym__blank,
    STATE(1765), 1,
      aux_sym__hll_declaration_specifiers_repeat2,
    STATE(1888), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4264), 3,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_LBRACK,
  [116169] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4349), 1,
      aux_sym_string_token1,
    ACTIONS(4440), 1,
      anon_sym_DQUOTE,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4351), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [116191] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4442), 1,
      anon_sym_DQUOTE,
    ACTIONS(4444), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4446), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1811), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [116213] = 6,
    ACTIONS(4301), 1,
      anon_sym_BSLASH,
    ACTIONS(4448), 1,
      sym__blank,
    STATE(1807), 1,
      aux_sym__hll_declaration_specifiers_repeat2,
    STATE(1888), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4299), 3,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_LBRACK,
  [116235] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4349), 1,
      aux_sym_string_token1,
    ACTIONS(4451), 1,
      anon_sym_DQUOTE,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4351), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [116257] = 5,
    ACTIONS(3492), 1,
      sym_identifier,
    STATE(1762), 1,
      sym__hll_type_identifier,
    STATE(1872), 1,
      aux_sym_hll_sized_type_specifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4453), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
  [116277] = 8,
    ACTIONS(3181), 1,
      sym__blank_line,
    ACTIONS(3184), 1,
      sym__blank,
    ACTIONS(4455), 1,
      aux_sym_elif_block_token1,
    STATE(1834), 1,
      aux_sym_if_block_repeat2,
    STATE(1845), 1,
      aux_sym_if_block_repeat3,
    STATE(2085), 1,
      sym_elif_block,
    STATE(2143), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [116303] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4458), 1,
      anon_sym_DQUOTE,
    ACTIONS(4460), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4462), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1798), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [116325] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4464), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [116349] = 6,
    ACTIONS(4466), 1,
      anon_sym_AMP,
    ACTIONS(4469), 1,
      anon_sym_DQUOTE,
    ACTIONS(4471), 1,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4474), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [116371] = 3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4278), 2,
      anon_sym_BSLASH,
      sym__blank,
    ACTIONS(4276), 5,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [116387] = 6,
    ACTIONS(4323), 1,
      anon_sym_AMP,
    ACTIONS(4349), 1,
      aux_sym_string_token1,
    ACTIONS(4477), 1,
      anon_sym_DQUOTE,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4351), 2,
      sym__ampersand_char,
      aux_sym_string_token2,
    STATE(1816), 2,
      sym_macro,
      aux_sym_string_repeat1,
  [116409] = 6,
    ACTIONS(4479), 1,
      aux_sym_elif_block_token1,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    STATE(1563), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [116430] = 6,
    ACTIONS(4188), 1,
      sym_identifier,
    STATE(1865), 1,
      aux_sym__hll_var_pattern_command_argument_list_format_repeat2,
    STATE(1887), 1,
      aux_sym_hll_type_pattern_repeat1,
    STATE(2535), 1,
      sym_hll_type_pattern,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4190), 2,
      anon_sym_STAR,
      anon_sym_QMARK,
  [116451] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4485), 1,
      aux_sym_elif_block_token1,
    STATE(1563), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [116472] = 6,
    ACTIONS(4487), 1,
      aux_sym_elif_block_token1,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1563), 2,
      sym_elif_block,
      sym_else_block,
  [116493] = 4,
    ACTIONS(4493), 1,
      sym__blank,
    STATE(1823), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2368), 4,
      aux_sym_block_token1,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [116510] = 5,
    ACTIONS(1785), 1,
      aux_sym_block_token1,
    ACTIONS(4496), 1,
      sym__blank,
    STATE(1823), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3498), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [116529] = 5,
    ACTIONS(1785), 1,
      aux_sym_block_token1,
    ACTIONS(4496), 1,
      sym__blank,
    STATE(1823), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3498), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [116548] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4485), 1,
      aux_sym_elif_block_token1,
    STATE(1522), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [116569] = 6,
    ACTIONS(4487), 1,
      aux_sym_elif_block_token1,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1522), 2,
      sym_elif_block,
      sym_else_block,
  [116590] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4500), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [116611] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    ACTIONS(4504), 1,
      anon_sym_RPAREN,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [116634] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4506), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [116655] = 5,
    ACTIONS(4508), 1,
      anon_sym_EQ,
    ACTIONS(4510), 1,
      sym__blank,
    STATE(2411), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3356), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [116674] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4512), 1,
      aux_sym_elif_block_token1,
    STATE(1563), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [116695] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4514), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [116716] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4516), 1,
      aux_sym_elif_block_token1,
    STATE(2156), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [116737] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4518), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1563), 2,
      sym_elif_block,
      sym_else_block,
  [116758] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4520), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [116779] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4524), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [116800] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4524), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [116821] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    ACTIONS(4526), 1,
      anon_sym_RPAREN,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [116844] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    ACTIONS(4529), 1,
      anon_sym_RPAREN,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [116867] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    ACTIONS(4531), 1,
      anon_sym_RPAREN,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [116890] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4533), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [116911] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4535), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [116932] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4518), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1525), 2,
      sym_elif_block,
      sym_else_block,
  [116953] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4516), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(2156), 2,
      sym_elif_block,
      sym_else_block,
  [116974] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4537), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [116995] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4539), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117016] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4541), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [117037] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4543), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117058] = 7,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4545), 1,
      anon_sym_RPAREN,
    ACTIONS(4547), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(2005), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [117081] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4549), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117102] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4545), 1,
      anon_sym_RPAREN,
    ACTIONS(4551), 1,
      sym__blank,
    STATE(1891), 1,
      aux_sym_block_repeat1,
    STATE(1945), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [117125] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    ACTIONS(4553), 1,
      anon_sym_RPAREN,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [117148] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4555), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [117169] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4506), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117190] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4514), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117211] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4557), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117232] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4559), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117253] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4561), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117274] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4500), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117295] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4563), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117316] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4535), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117337] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4565), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117358] = 4,
    ACTIONS(3525), 1,
      sym__blank,
    ACTIONS(4567), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3523), 4,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      anon_sym_EQ,
      sym__blank_line,
  [117375] = 6,
    ACTIONS(4569), 1,
      sym_identifier,
    STATE(1865), 1,
      aux_sym__hll_var_pattern_command_argument_list_format_repeat2,
    STATE(1990), 1,
      aux_sym_hll_type_pattern_repeat1,
    STATE(2607), 1,
      sym_hll_type_pattern,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4572), 2,
      anon_sym_STAR,
      anon_sym_QMARK,
  [117396] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4565), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [117417] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4537), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [117438] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4575), 1,
      aux_sym__host_command_argument_list_default_format_token1,
    ACTIONS(4579), 1,
      sym__blank,
    STATE(1777), 1,
      aux_sym_block_repeat1,
    STATE(2136), 1,
      aux_sym__host_command_argument_list_default_format_repeat1,
    STATE(2216), 1,
      aux_sym__host_command_argument_list_default_format_repeat2,
    ACTIONS(4577), 2,
      sym_comment,
      sym__blank_line,
  [117461] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4557), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [117482] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4539), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [117503] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4581), 1,
      sym_identifier,
    ACTIONS(4589), 1,
      sym__blank,
    STATE(1871), 1,
      aux_sym_hll_type_pattern_repeat1,
    ACTIONS(4584), 2,
      anon_sym_STAR,
      anon_sym_QMARK,
    ACTIONS(4587), 2,
      sym_comment,
      sym__blank_line,
  [117524] = 4,
    ACTIONS(4424), 1,
      sym_identifier,
    STATE(1872), 1,
      aux_sym_hll_sized_type_specifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4591), 4,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_long,
      anon_sym_short,
  [117541] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4543), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [117562] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4549), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [117583] = 5,
    ACTIONS(4508), 1,
      anon_sym_EQ,
    ACTIONS(4510), 1,
      sym__blank,
    STATE(2411), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3370), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [117602] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4559), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [117623] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4520), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [117644] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4512), 1,
      aux_sym_elif_block_token1,
    STATE(1522), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [117665] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    ACTIONS(4594), 1,
      anon_sym_RPAREN,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [117688] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    ACTIONS(4596), 1,
      anon_sym_RPAREN,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [117711] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4598), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [117732] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4516), 1,
      aux_sym_elif_block_token1,
    STATE(2120), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [117753] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4516), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(2120), 2,
      sym_elif_block,
      sym_else_block,
  [117774] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4561), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [117795] = 4,
    ACTIONS(3516), 1,
      sym__blank,
    ACTIONS(4600), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3514), 4,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      anon_sym_EQ,
      sym__blank_line,
  [117812] = 6,
    ACTIONS(4188), 1,
      sym_identifier,
    STATE(1865), 1,
      aux_sym__hll_var_pattern_command_argument_list_format_repeat2,
    STATE(1887), 1,
      aux_sym_hll_type_pattern_repeat1,
    STATE(2497), 1,
      sym_hll_type_pattern,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4190), 2,
      anon_sym_STAR,
      anon_sym_QMARK,
  [117833] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4188), 1,
      sym_identifier,
    ACTIONS(4606), 1,
      sym__blank,
    STATE(1871), 1,
      aux_sym_hll_type_pattern_repeat1,
    ACTIONS(4602), 2,
      anon_sym_STAR,
      anon_sym_QMARK,
    ACTIONS(4604), 2,
      sym_comment,
      sym__blank_line,
  [117854] = 5,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
    STATE(1817), 2,
      sym_hll_type_qualifier,
      sym__hll_declaration_modifiers,
  [117873] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4518), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1522), 2,
      sym_elif_block,
      sym_else_block,
  [117894] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4563), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [117915] = 7,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4547), 1,
      anon_sym_LBRACK,
    ACTIONS(4608), 1,
      anon_sym_RPAREN,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(2005), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [117938] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    ACTIONS(4610), 1,
      anon_sym_RPAREN,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [117961] = 6,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(1758), 1,
      sym_hll_type_qualifier,
    STATE(2081), 1,
      aux_sym__hll_pointer_declarator_specifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
  [117982] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4533), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [118003] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4612), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [118024] = 4,
    ACTIONS(3808), 1,
      sym__blank,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4614), 4,
      anon_sym_STAR,
      anon_sym_PERCENT,
      anon_sym_QMARK,
      sym_identifier,
  [118041] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4616), 1,
      aux_sym_elif_block_token1,
    STATE(2120), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [118062] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4618), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(2120), 2,
      sym_elif_block,
      sym_else_block,
  [118083] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4616), 1,
      aux_sym_elif_block_token1,
    STATE(2143), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [118104] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4618), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(2143), 2,
      sym_elif_block,
      sym_else_block,
  [118125] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4616), 1,
      aux_sym_elif_block_token1,
    STATE(2156), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [118146] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4618), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(2156), 2,
      sym_elif_block,
      sym_else_block,
  [118167] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4485), 1,
      aux_sym_elif_block_token1,
    STATE(1525), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [118188] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4516), 1,
      aux_sym_elif_block_token1,
    STATE(2143), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [118209] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4620), 1,
      aux_sym_elif_block_token1,
    STATE(1525), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [118230] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4622), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1525), 2,
      sym_elif_block,
      sym_else_block,
  [118251] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4620), 1,
      aux_sym_elif_block_token1,
    STATE(1563), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [118272] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4622), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1563), 2,
      sym_elif_block,
      sym_else_block,
  [118293] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4620), 1,
      aux_sym_elif_block_token1,
    STATE(1522), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [118314] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4622), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1522), 2,
      sym_elif_block,
      sym_else_block,
  [118335] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4516), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(2143), 2,
      sym_elif_block,
      sym_else_block,
  [118356] = 6,
    ACTIONS(4487), 1,
      aux_sym_elif_block_token1,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1525), 2,
      sym_elif_block,
      sym_else_block,
  [118377] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    ACTIONS(4624), 1,
      anon_sym_RPAREN,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [118400] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    ACTIONS(4626), 1,
      anon_sym_RPAREN,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [118423] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4628), 1,
      anon_sym_RPAREN,
    ACTIONS(4630), 1,
      sym__blank,
    STATE(1850), 1,
      aux_sym_block_repeat1,
    STATE(1945), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [118446] = 6,
    ACTIONS(4479), 1,
      aux_sym_elif_block_token1,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    STATE(1525), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [118467] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4632), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1525), 2,
      sym_elif_block,
      sym_else_block,
  [118488] = 5,
    ACTIONS(4634), 1,
      anon_sym_PERCENT,
    STATE(1918), 1,
      aux_sym__hll_var_pattern_command_argument_list_format_repeat1,
    STATE(2541), 1,
      sym_format_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4614), 3,
      anon_sym_STAR,
      anon_sym_QMARK,
      sym_identifier,
  [118507] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4632), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1563), 2,
      sym_elif_block,
      sym_else_block,
  [118528] = 6,
    ACTIONS(4479), 1,
      aux_sym_elif_block_token1,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    STATE(1522), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [118549] = 6,
    ACTIONS(4489), 1,
      sym__blank_line,
    ACTIONS(4491), 1,
      sym__blank,
    ACTIONS(4632), 1,
      aux_sym_elif_block_token1,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1522), 2,
      sym_elif_block,
      sym_else_block,
  [118570] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4522), 1,
      sym__blank,
    STATE(2075), 1,
      sym_hll_parameter_list,
    STATE(2101), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4339), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [118591] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    ACTIONS(4637), 1,
      anon_sym_RPAREN,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [118614] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4598), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [118635] = 6,
    ACTIONS(4481), 1,
      sym__blank_line,
    ACTIONS(4483), 1,
      sym__blank,
    ACTIONS(4512), 1,
      aux_sym_elif_block_token1,
    STATE(1525), 1,
      sym_else_block,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [118656] = 7,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4343), 1,
      anon_sym_LBRACK,
    ACTIONS(4345), 1,
      sym__blank,
    ACTIONS(4640), 1,
      anon_sym_RPAREN,
    STATE(1945), 1,
      sym_hll_parameter_list,
    STATE(1949), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [118679] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    ACTIONS(4502), 1,
      sym__blank,
    STATE(2163), 1,
      sym_hll_parameter_list,
    STATE(2172), 1,
      aux_sym_block_repeat1,
    ACTIONS(4612), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [118700] = 6,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2895), 1,
      anon_sym_RPAREN,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(2445), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [118720] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4643), 1,
      anon_sym_RPAREN,
    ACTIONS(4645), 1,
      sym__blank,
    STATE(1960), 1,
      aux_sym_block_repeat1,
    STATE(2075), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [118740] = 6,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4647), 1,
      anon_sym_RPAREN,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(2119), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [118760] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4649), 1,
      sym_comment,
    ACTIONS(4651), 1,
      sym__blank,
    STATE(1935), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2138), 1,
      sym_block,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [118782] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4653), 1,
      anon_sym_DOT,
    ACTIONS(4657), 1,
      sym__blank,
    STATE(2035), 1,
      aux_sym_format_expression_repeat1,
    ACTIONS(4655), 3,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [118800] = 6,
    ACTIONS(4287), 1,
      anon_sym_AMP,
    ACTIONS(4659), 1,
      anon_sym_SLASH,
    ACTIONS(4661), 1,
      sym__blank,
    STATE(1991), 1,
      aux_sym_block_repeat1,
    STATE(2558), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [118820] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(1986), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [118842] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(2079), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [118864] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1320), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [118886] = 6,
    ACTIONS(3143), 1,
      anon_sym_COMMA,
    ACTIONS(4667), 1,
      anon_sym_RPAREN,
    ACTIONS(4669), 1,
      sym__blank,
    STATE(1970), 1,
      aux_sym_argument_list_repeat1,
    STATE(2395), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [118906] = 3,
    ACTIONS(1921), 2,
      anon_sym_DQUOTE,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1923), 3,
      sym__ampersand_char,
      anon_sym_AMP,
      aux_sym_string_token2,
  [118920] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(2094), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [118942] = 5,
    ACTIONS(4671), 1,
      anon_sym_SLASH,
    ACTIONS(4674), 1,
      anon_sym_PERCENT,
    STATE(1940), 1,
      aux_sym__var_command_argument_list_non_hll_format_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(2585), 2,
      sym_option_expression,
      sym_format_expression,
  [118960] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4677), 1,
      anon_sym_DOT,
    ACTIONS(4682), 1,
      sym__blank,
    STATE(1941), 1,
      aux_sym_format_expression_repeat1,
    ACTIONS(4680), 3,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [118978] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(2107), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [119000] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4684), 1,
      anon_sym_RPAREN,
    ACTIONS(4686), 1,
      sym__blank,
    STATE(1952), 1,
      aux_sym_block_repeat1,
    STATE(2075), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119020] = 6,
    ACTIONS(3143), 1,
      anon_sym_COMMA,
    ACTIONS(4669), 1,
      sym__blank,
    ACTIONS(4688), 1,
      anon_sym_RPAREN,
    STATE(1970), 1,
      aux_sym_argument_list_repeat1,
    STATE(2395), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119040] = 3,
    ACTIONS(4692), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4690), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119054] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1333), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [119076] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(2122), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [119098] = 6,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2903), 1,
      anon_sym_RPAREN,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(2225), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119118] = 6,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4547), 1,
      anon_sym_LBRACK,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(2005), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119138] = 6,
    ACTIONS(3529), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3531), 1,
      sym__blank,
    ACTIONS(4694), 1,
      aux_sym_while_block_token1,
    ACTIONS(4697), 1,
      sym__blank_line,
    STATE(2439), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119158] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(2134), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [119180] = 6,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4700), 1,
      anon_sym_RPAREN,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(2119), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119200] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4700), 1,
      anon_sym_RPAREN,
    ACTIONS(4702), 1,
      sym__blank,
    STATE(1956), 1,
      aux_sym_block_repeat1,
    STATE(2075), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119220] = 3,
    ACTIONS(4706), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4704), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119234] = 6,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(4708), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(2257), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119254] = 6,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4710), 1,
      anon_sym_RPAREN,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(2119), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119274] = 3,
    ACTIONS(4714), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4712), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119288] = 3,
    ACTIONS(4718), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4716), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119302] = 3,
    ACTIONS(4722), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4720), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119316] = 6,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4724), 1,
      anon_sym_RPAREN,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(2119), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119336] = 6,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    ACTIONS(4724), 1,
      anon_sym_RPAREN,
    ACTIONS(4726), 1,
      sym__blank,
    STATE(1930), 1,
      aux_sym_block_repeat1,
    STATE(2075), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119356] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(2165), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [119378] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(2166), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [119400] = 6,
    ACTIONS(3529), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3531), 1,
      sym__blank,
    ACTIONS(4697), 1,
      sym__blank_line,
    ACTIONS(4728), 1,
      aux_sym_while_block_token1,
    STATE(2439), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119420] = 3,
    ACTIONS(3541), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3539), 4,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      anon_sym_EQ,
      sym__blank_line,
  [119434] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1569), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [119456] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4665), 1,
      sym__blank,
    ACTIONS(4730), 1,
      sym_comment,
    STATE(1545), 1,
      sym_block,
    STATE(1966), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2401), 1,
      aux_sym_block_repeat1,
  [119478] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(1950), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [119500] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1289), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [119522] = 6,
    ACTIONS(3290), 1,
      anon_sym_RPAREN,
    ACTIONS(4732), 1,
      anon_sym_COMMA,
    ACTIONS(4735), 1,
      sym__blank,
    STATE(1970), 1,
      aux_sym_argument_list_repeat1,
    STATE(2395), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119542] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(2030), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [119564] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(1794), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [119586] = 5,
    ACTIONS(4738), 1,
      aux_sym_elif_block_token1,
    ACTIONS(4741), 1,
      sym__blank_line,
    ACTIONS(4744), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(1973), 2,
      sym_elif_block,
      aux_sym_if_block_repeat2,
  [119604] = 6,
    ACTIONS(3143), 1,
      anon_sym_COMMA,
    ACTIONS(4669), 1,
      sym__blank,
    ACTIONS(4747), 1,
      anon_sym_RPAREN,
    STATE(1970), 1,
      aux_sym_argument_list_repeat1,
    STATE(2395), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119624] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(1813), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [119646] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1300), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [119668] = 6,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(4749), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(2448), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119688] = 3,
    ACTIONS(4753), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4751), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119702] = 3,
    ACTIONS(4757), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4755), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119716] = 3,
    ACTIONS(4761), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4759), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119730] = 3,
    ACTIONS(4765), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4763), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119744] = 3,
    ACTIONS(4769), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4767), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119758] = 3,
    ACTIONS(4773), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4771), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119772] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1577), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [119794] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1504), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [119816] = 6,
    ACTIONS(3543), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3545), 1,
      sym__blank,
    ACTIONS(4775), 1,
      aux_sym_while_block_token1,
    ACTIONS(4777), 1,
      sym__blank_line,
    STATE(2440), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119836] = 3,
    ACTIONS(4782), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4780), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119850] = 6,
    ACTIONS(2807), 1,
      anon_sym_RPAREN,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(2250), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119870] = 6,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(4784), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(2252), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119890] = 5,
    ACTIONS(4606), 1,
      sym__blank,
    ACTIONS(4786), 1,
      sym_identifier,
    STATE(2004), 1,
      aux_sym_hll_type_pattern_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4788), 2,
      anon_sym_STAR,
      anon_sym_QMARK,
  [119908] = 5,
    ACTIONS(2370), 1,
      anon_sym_SLASH,
    ACTIONS(4790), 1,
      sym__blank,
    STATE(1991), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2368), 2,
      anon_sym_AMP,
      aux_sym_parameter_declaration_token4,
  [119926] = 3,
    ACTIONS(4795), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4793), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119940] = 6,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2875), 1,
      anon_sym_RPAREN,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(2390), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [119960] = 5,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    ACTIONS(458), 1,
      anon_sym_PERCENT,
    STATE(1940), 1,
      aux_sym__var_command_argument_list_non_hll_format_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    STATE(2289), 2,
      sym_option_expression,
      sym_format_expression,
  [119978] = 3,
    ACTIONS(4799), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4797), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [119992] = 3,
    ACTIONS(4803), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4801), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120006] = 3,
    ACTIONS(4807), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4805), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120020] = 3,
    ACTIONS(4811), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4809), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120034] = 3,
    ACTIONS(4815), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4813), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120048] = 3,
    ACTIONS(4819), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4817), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120062] = 3,
    ACTIONS(4823), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4821), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120076] = 3,
    ACTIONS(4827), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4825), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120090] = 6,
    ACTIONS(4287), 1,
      anon_sym_AMP,
    ACTIONS(4661), 1,
      sym__blank,
    ACTIONS(4829), 1,
      aux_sym_parameter_declaration_token4,
    STATE(1991), 1,
      aux_sym_block_repeat1,
    STATE(2529), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [120110] = 5,
    ACTIONS(4589), 1,
      sym__blank,
    ACTIONS(4831), 1,
      sym_identifier,
    STATE(2004), 1,
      aux_sym_hll_type_pattern_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4834), 2,
      anon_sym_STAR,
      anon_sym_QMARK,
  [120128] = 3,
    ACTIONS(4839), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4837), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120142] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1530), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [120164] = 5,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(1799), 1,
      sym_hll_type_qualifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
  [120182] = 3,
    ACTIONS(4843), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4841), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120196] = 3,
    ACTIONS(4847), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4845), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120210] = 3,
    ACTIONS(4851), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4849), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120224] = 3,
    ACTIONS(4855), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4853), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120238] = 3,
    ACTIONS(4859), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4857), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120252] = 5,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(1194), 1,
      sym_hll_type_qualifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4861), 2,
      anon_sym_const,
      anon_sym_volatile,
  [120270] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1538), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [120292] = 3,
    ACTIONS(4865), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4863), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120306] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1293), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [120328] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1295), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [120350] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1540), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [120372] = 3,
    ACTIONS(4869), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4867), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120386] = 6,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(4871), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(2291), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [120406] = 3,
    ACTIONS(4875), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4873), 4,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACK,
  [120420] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1511), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [120442] = 6,
    ACTIONS(4287), 1,
      anon_sym_AMP,
    ACTIONS(4661), 1,
      sym__blank,
    ACTIONS(4877), 1,
      aux_sym_parameter_declaration_token4,
    STATE(1991), 1,
      aux_sym_block_repeat1,
    STATE(2529), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [120462] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1583), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [120484] = 6,
    ACTIONS(4287), 1,
      anon_sym_AMP,
    ACTIONS(4661), 1,
      sym__blank,
    ACTIONS(4879), 1,
      aux_sym_parameter_declaration_token4,
    STATE(1991), 1,
      aux_sym_block_repeat1,
    STATE(2529), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [120504] = 6,
    ACTIONS(4287), 1,
      anon_sym_AMP,
    ACTIONS(4661), 1,
      sym__blank,
    ACTIONS(4881), 1,
      aux_sym_parameter_declaration_token4,
    STATE(1991), 1,
      aux_sym_block_repeat1,
    STATE(2529), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [120524] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1332), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [120546] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1349), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [120568] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(9), 1,
      aux_sym_block_token1,
    ACTIONS(4663), 1,
      sym_comment,
    ACTIONS(4665), 1,
      sym__blank,
    STATE(1580), 1,
      sym_block,
    STATE(2401), 1,
      aux_sym_block_repeat1,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [120590] = 6,
    ACTIONS(3543), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3545), 1,
      sym__blank,
    ACTIONS(4777), 1,
      sym__blank_line,
    ACTIONS(4883), 1,
      aux_sym_while_block_token1,
    STATE(2440), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [120610] = 6,
    ACTIONS(3504), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3506), 1,
      sym__blank,
    ACTIONS(4886), 1,
      aux_sym_while_block_token1,
    ACTIONS(4888), 1,
      sym__blank_line,
    STATE(2435), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [120630] = 3,
    ACTIONS(3516), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3514), 4,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      anon_sym_EQ,
      sym__blank_line,
  [120644] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4893), 1,
      sym__blank,
    ACTIONS(4891), 5,
      anon_sym_STAR,
      anon_sym_QMARK,
      sym_identifier,
      sym_comment,
      sym__blank_line,
  [120658] = 6,
    ACTIONS(3504), 1,
      aux_sym_elif_block_token1,
    ACTIONS(3506), 1,
      sym__blank,
    ACTIONS(4888), 1,
      sym__blank_line,
    ACTIONS(4895), 1,
      aux_sym_while_block_token1,
    STATE(2435), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [120678] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4653), 1,
      anon_sym_DOT,
    ACTIONS(4900), 1,
      sym__blank,
    STATE(1941), 1,
      aux_sym_format_expression_repeat1,
    ACTIONS(4898), 3,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [120696] = 3,
    ACTIONS(1901), 2,
      anon_sym_DQUOTE,
      aux_sym_string_token1,
    ACTIONS(4331), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(1903), 3,
      sym__ampersand_char,
      anon_sym_AMP,
      aux_sym_string_token2,
  [120710] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(1773), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [120732] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(1775), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [120754] = 5,
    ACTIONS(2370), 1,
      aux_sym__host_command_argument_list_default_format_token1,
    ACTIONS(4902), 1,
      sym__blank,
    STATE(2039), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2368), 2,
      anon_sym_AMP,
      anon_sym_DQUOTE,
  [120772] = 7,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1558), 1,
      aux_sym_block_token1,
    ACTIONS(4651), 1,
      sym__blank,
    ACTIONS(4663), 1,
      sym_comment,
    STATE(1964), 1,
      sym_block,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
    STATE(2413), 1,
      aux_sym_block_repeat1,
  [120794] = 3,
    ACTIONS(4907), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4905), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
  [120807] = 3,
    ACTIONS(4909), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3985), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [120820] = 3,
    ACTIONS(4911), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3996), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [120833] = 3,
    ACTIONS(4913), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3963), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [120846] = 3,
    ACTIONS(4915), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3930), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [120859] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(4917), 1,
      anon_sym_DQUOTE,
    ACTIONS(4919), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(4921), 1,
      sym_hll_escape_sequence,
    STATE(2052), 1,
      aux_sym_hll_string_literal_repeat1,
  [120878] = 3,
    ACTIONS(4923), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3952), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [120891] = 5,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(4925), 1,
      anon_sym_SLASH,
    ACTIONS(4927), 1,
      anon_sym_PERCENT,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [120908] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4929), 1,
      sym__blank,
    STATE(2198), 1,
      aux_sym_block_repeat1,
    STATE(2199), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(2770), 2,
      sym_comment,
      sym__blank_line,
  [120925] = 3,
    ACTIONS(4931), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3974), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [120938] = 5,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(4933), 1,
      sym_identifier,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(1766), 1,
      sym__hll_type_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [120955] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(4935), 1,
      anon_sym_DQUOTE,
    ACTIONS(4937), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(4939), 1,
      sym_hll_escape_sequence,
    STATE(2104), 1,
      aux_sym_hll_string_literal_repeat1,
  [120974] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4941), 1,
      sym_comment,
    ACTIONS(4943), 1,
      sym__blank_line,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [120993] = 3,
    ACTIONS(2311), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(2313), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
  [121006] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(4937), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(4939), 1,
      sym_hll_escape_sequence,
    ACTIONS(4947), 1,
      anon_sym_DQUOTE,
    STATE(2104), 1,
      aux_sym_hll_string_literal_repeat1,
  [121025] = 5,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(4933), 1,
      sym_identifier,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(1771), 1,
      sym__hll_type_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [121042] = 5,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(4933), 1,
      sym_identifier,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    STATE(1768), 1,
      sym__hll_type_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [121059] = 3,
    ACTIONS(3675), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3673), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121072] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4951), 1,
      sym__blank,
    ACTIONS(4949), 4,
      anon_sym_LPAREN,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [121085] = 3,
    ACTIONS(3683), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3681), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121098] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4953), 1,
      anon_sym_COMMA,
    ACTIONS(4955), 1,
      sym__blank,
    STATE(998), 1,
      aux_sym_block_repeat1,
    ACTIONS(2770), 2,
      sym_comment,
      sym__blank_line,
  [121115] = 3,
    ACTIONS(3699), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3697), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121128] = 3,
    ACTIONS(3699), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3697), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121141] = 3,
    ACTIONS(4893), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4891), 3,
      anon_sym_STAR,
      anon_sym_QMARK,
      sym_identifier,
  [121154] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(4957), 1,
      anon_sym_DQUOTE,
    ACTIONS(4959), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(4961), 1,
      sym_hll_escape_sequence,
    STATE(2068), 1,
      aux_sym_hll_string_literal_repeat1,
  [121173] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4963), 1,
      anon_sym_DOT,
    ACTIONS(4967), 1,
      sym__blank,
    STATE(2209), 1,
      aux_sym__command_identifier_repeat1,
    ACTIONS(4965), 2,
      sym_comment,
      sym__blank_line,
  [121190] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4969), 1,
      sym__blank,
    STATE(1933), 1,
      aux_sym_block_repeat1,
    STATE(2212), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(2770), 2,
      sym_comment,
      sym__blank_line,
  [121207] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(4937), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(4939), 1,
      sym_hll_escape_sequence,
    ACTIONS(4971), 1,
      anon_sym_DQUOTE,
    STATE(2104), 1,
      aux_sym_hll_string_literal_repeat1,
  [121226] = 3,
    ACTIONS(3727), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3725), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121239] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4973), 1,
      anon_sym_DOT,
    ACTIONS(4977), 1,
      sym__blank,
    STATE(2100), 1,
      aux_sym__command_identifier_repeat1,
    ACTIONS(4975), 2,
      sym_comment,
      sym__blank_line,
  [121256] = 3,
    ACTIONS(3731), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3729), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121269] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4979), 2,
      anon_sym_const,
      anon_sym_volatile,
  [121284] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4981), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [121299] = 5,
    ACTIONS(4898), 1,
      anon_sym_COMMA,
    ACTIONS(4900), 1,
      sym__blank,
    ACTIONS(4983), 1,
      anon_sym_DOT,
    STATE(2116), 1,
      aux_sym_format_expression_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [121316] = 3,
    ACTIONS(4987), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4985), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
  [121329] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(4989), 1,
      anon_sym_DQUOTE,
    ACTIONS(4991), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(4993), 1,
      sym_hll_escape_sequence,
    STATE(2077), 1,
      aux_sym_hll_string_literal_repeat1,
  [121348] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(4937), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(4939), 1,
      sym_hll_escape_sequence,
    ACTIONS(4995), 1,
      anon_sym_DQUOTE,
    STATE(2104), 1,
      aux_sym_hll_string_literal_repeat1,
  [121367] = 3,
    ACTIONS(3665), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3663), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121380] = 3,
    ACTIONS(3679), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3677), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121393] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(4997), 1,
      anon_sym_DQUOTE,
    ACTIONS(4999), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(5001), 1,
      sym_hll_escape_sequence,
    STATE(2083), 1,
      aux_sym_hll_string_literal_repeat1,
  [121412] = 4,
    STATE(1755), 1,
      sym_hll_type_qualifier,
    STATE(2084), 1,
      aux_sym__hll_pointer_declarator_specifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
  [121427] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4929), 1,
      sym__blank,
    STATE(2198), 1,
      aux_sym_block_repeat1,
    STATE(2199), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(5003), 2,
      sym_comment,
      sym__blank_line,
  [121444] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(4937), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(4939), 1,
      sym_hll_escape_sequence,
    ACTIONS(5005), 1,
      anon_sym_DQUOTE,
    STATE(2104), 1,
      aux_sym_hll_string_literal_repeat1,
  [121463] = 4,
    STATE(2084), 1,
      aux_sym__hll_pointer_declarator_specifier_repeat1,
    STATE(2178), 1,
      sym_hll_type_qualifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(5007), 2,
      anon_sym_const,
      anon_sym_volatile,
  [121478] = 4,
    ACTIONS(3655), 1,
      aux_sym_while_block_token1,
    ACTIONS(3670), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3667), 2,
      aux_sym_elif_block_token1,
      sym__blank_line,
  [121493] = 3,
    ACTIONS(5010), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3908), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [121506] = 3,
    ACTIONS(5012), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3882), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [121519] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5014), 1,
      anon_sym_DQUOTE,
    ACTIONS(5016), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(5018), 1,
      sym_hll_escape_sequence,
    STATE(2090), 1,
      aux_sym_hll_string_literal_repeat1,
  [121538] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3026), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [121553] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(4937), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(4939), 1,
      sym_hll_escape_sequence,
    ACTIONS(5020), 1,
      anon_sym_DQUOTE,
    STATE(2104), 1,
      aux_sym_hll_string_literal_repeat1,
  [121572] = 3,
    ACTIONS(3200), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3202), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121585] = 3,
    ACTIONS(5022), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3897), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [121598] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2311), 1,
      sym__blank,
    ACTIONS(2313), 4,
      anon_sym_LPAREN,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [121611] = 3,
    ACTIONS(3703), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3701), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121624] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5024), 1,
      sym_comment,
    ACTIONS(5026), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [121643] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4929), 1,
      sym__blank,
    STATE(2198), 1,
      aux_sym_block_repeat1,
    STATE(2199), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(5028), 2,
      sym_comment,
      sym__blank_line,
  [121660] = 4,
    STATE(1758), 1,
      sym_hll_type_qualifier,
    STATE(2084), 1,
      aux_sym__hll_pointer_declarator_specifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3833), 2,
      anon_sym_const,
      anon_sym_volatile,
  [121675] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4969), 1,
      sym__blank,
    STATE(1933), 1,
      aux_sym_block_repeat1,
    STATE(2191), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(5028), 2,
      sym_comment,
      sym__blank_line,
  [121692] = 3,
    ACTIONS(3625), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3623), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121705] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4973), 1,
      anon_sym_DOT,
    ACTIONS(5032), 1,
      sym__blank,
    STATE(2133), 1,
      aux_sym__command_identifier_repeat1,
    ACTIONS(5030), 2,
      sym_comment,
      sym__blank_line,
  [121722] = 5,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4341), 1,
      anon_sym_LPAREN,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(2119), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [121739] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5034), 1,
      sym_comment,
    ACTIONS(5036), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [121758] = 3,
    ACTIONS(3555), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3553), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121771] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5038), 1,
      anon_sym_DQUOTE,
    ACTIONS(5040), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(5043), 1,
      sym_hll_escape_sequence,
    STATE(2104), 1,
      aux_sym_hll_string_literal_repeat1,
  [121790] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5046), 1,
      sym_comment,
    ACTIONS(5048), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [121809] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4967), 1,
      sym__blank,
    ACTIONS(5050), 1,
      anon_sym_DOT,
    STATE(2209), 1,
      aux_sym__command_identifier_repeat1,
    ACTIONS(4965), 2,
      sym_comment,
      sym__blank_line,
  [121826] = 3,
    ACTIONS(3711), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3709), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121839] = 3,
    ACTIONS(3719), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3717), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121852] = 3,
    ACTIONS(3719), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3717), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121865] = 3,
    ACTIONS(3743), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3741), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121878] = 3,
    ACTIONS(3555), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3553), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121891] = 3,
    ACTIONS(3747), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3745), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121904] = 3,
    ACTIONS(3555), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3553), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [121917] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4929), 1,
      sym__blank,
    STATE(2198), 1,
      aux_sym_block_repeat1,
    STATE(2199), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(5052), 2,
      sym_comment,
      sym__blank_line,
  [121934] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4969), 1,
      sym__blank,
    STATE(1933), 1,
      aux_sym_block_repeat1,
    STATE(2082), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(5052), 2,
      sym_comment,
      sym__blank_line,
  [121951] = 5,
    ACTIONS(4680), 1,
      anon_sym_COMMA,
    ACTIONS(4682), 1,
      sym__blank,
    ACTIONS(5054), 1,
      anon_sym_DOT,
    STATE(2116), 1,
      aux_sym_format_expression_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [121968] = 3,
    ACTIONS(4951), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4949), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
  [121981] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5059), 1,
      sym__blank,
    ACTIONS(5057), 4,
      anon_sym_LPAREN,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [121994] = 3,
    ACTIONS(5063), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(5061), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
  [122007] = 3,
    ACTIONS(3177), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3175), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122020] = 3,
    ACTIONS(3555), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3553), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122033] = 3,
    ACTIONS(3597), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3595), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122046] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(4937), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(4939), 1,
      sym_hll_escape_sequence,
    ACTIONS(5065), 1,
      anon_sym_DQUOTE,
    STATE(2104), 1,
      aux_sym_hll_string_literal_repeat1,
  [122065] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4843), 1,
      sym__blank,
    ACTIONS(4841), 4,
      anon_sym_LPAREN,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [122078] = 3,
    ACTIONS(3637), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3635), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122091] = 3,
    ACTIONS(3637), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3635), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122104] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5067), 1,
      sym_comment,
    ACTIONS(5069), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [122123] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5071), 1,
      anon_sym_COMMA,
    ACTIONS(5073), 1,
      sym__blank,
    STATE(995), 1,
      aux_sym_block_repeat1,
    ACTIONS(2763), 2,
      sym_comment,
      sym__blank_line,
  [122140] = 3,
    ACTIONS(3641), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3639), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122153] = 3,
    ACTIONS(3645), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3643), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122166] = 5,
    ACTIONS(4655), 1,
      anon_sym_COMMA,
    ACTIONS(4657), 1,
      sym__blank,
    ACTIONS(4983), 1,
      anon_sym_DOT,
    STATE(2074), 1,
      aux_sym_format_expression_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [122183] = 3,
    ACTIONS(3715), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3713), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122196] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5075), 1,
      anon_sym_DOT,
    ACTIONS(5080), 1,
      sym__blank,
    STATE(2133), 1,
      aux_sym__command_identifier_repeat1,
    ACTIONS(5078), 2,
      sym_comment,
      sym__blank_line,
  [122213] = 3,
    ACTIONS(3723), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3721), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122226] = 3,
    ACTIONS(3735), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3733), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122239] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5082), 1,
      aux_sym__host_command_argument_list_default_format_token1,
    ACTIONS(5087), 1,
      sym__blank,
    STATE(2136), 1,
      aux_sym__host_command_argument_list_default_format_repeat1,
    ACTIONS(5085), 2,
      sym_comment,
      sym__blank_line,
  [122256] = 3,
    ACTIONS(5089), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3919), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [122269] = 3,
    ACTIONS(3629), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3627), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122282] = 3,
    ACTIONS(3739), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3737), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122295] = 3,
    ACTIONS(3601), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3599), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122308] = 3,
    ACTIONS(3601), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3599), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122321] = 3,
    ACTIONS(3653), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3651), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122334] = 3,
    ACTIONS(3657), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3655), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122347] = 5,
    ACTIONS(4287), 1,
      anon_sym_AMP,
    ACTIONS(4661), 1,
      sym__blank,
    STATE(1991), 1,
      aux_sym_block_repeat1,
    STATE(2529), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [122364] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5091), 1,
      anon_sym_DQUOTE,
    ACTIONS(5093), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(5095), 1,
      sym_hll_escape_sequence,
    STATE(2174), 1,
      aux_sym_hll_string_literal_repeat1,
  [122383] = 3,
    ACTIONS(3687), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3685), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122396] = 3,
    ACTIONS(5059), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(5057), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_COMMA,
  [122409] = 3,
    ACTIONS(3344), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3342), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122422] = 3,
    ACTIONS(3567), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3565), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122435] = 3,
    ACTIONS(3567), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3565), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122448] = 3,
    ACTIONS(3567), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3565), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122461] = 3,
    ACTIONS(3567), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3565), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122474] = 3,
    ACTIONS(3581), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3579), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122487] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5063), 1,
      sym__blank,
    ACTIONS(5061), 4,
      anon_sym_LPAREN,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [122500] = 3,
    ACTIONS(3559), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3557), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122513] = 3,
    ACTIONS(3571), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3569), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122526] = 3,
    ACTIONS(3581), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3579), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122539] = 3,
    ACTIONS(3358), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3356), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122552] = 3,
    ACTIONS(3593), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3591), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122565] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5097), 1,
      sym_comment,
    ACTIONS(5099), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [122584] = 3,
    ACTIONS(3605), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3603), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122597] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5101), 1,
      sym_comment,
    ACTIONS(5103), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [122616] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4987), 1,
      sym__blank,
    ACTIONS(4985), 4,
      anon_sym_LPAREN,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [122629] = 3,
    ACTIONS(3613), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3611), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122642] = 3,
    ACTIONS(3617), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3615), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122655] = 3,
    ACTIONS(3621), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3619), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122668] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4969), 1,
      sym__blank,
    STATE(1933), 1,
      aux_sym_block_repeat1,
    STATE(2049), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(2763), 2,
      sym_comment,
      sym__blank_line,
  [122685] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4579), 1,
      sym__blank,
    STATE(1777), 1,
      aux_sym_block_repeat1,
    STATE(2216), 1,
      aux_sym__host_command_argument_list_default_format_repeat2,
    ACTIONS(4577), 2,
      sym_comment,
      sym__blank_line,
  [122702] = 4,
    ACTIONS(3175), 1,
      aux_sym_while_block_token1,
    ACTIONS(3588), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3585), 2,
      aux_sym_elif_block_token1,
      sym__blank_line,
  [122717] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5105), 1,
      sym_comment,
    ACTIONS(5107), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [122736] = 5,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5109), 1,
      anon_sym_EQ,
    ACTIONS(5111), 1,
      anon_sym_COMMA,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [122753] = 5,
    ACTIONS(3808), 1,
      sym__blank,
    ACTIONS(4498), 1,
      anon_sym_LPAREN,
    STATE(1715), 1,
      aux_sym_block_repeat1,
    STATE(2154), 1,
      sym_hll_parameter_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [122770] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5113), 1,
      sym_comment,
    ACTIONS(5115), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [122789] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(4937), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(4939), 1,
      sym_hll_escape_sequence,
    ACTIONS(5117), 1,
      anon_sym_DQUOTE,
    STATE(2104), 1,
      aux_sym_hll_string_literal_repeat1,
  [122808] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5119), 1,
      sym_comment,
    ACTIONS(5121), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [122827] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4875), 1,
      sym__blank,
    ACTIONS(4873), 4,
      anon_sym_LPAREN,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [122840] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5123), 1,
      sym_comment,
    ACTIONS(5125), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [122859] = 4,
    ACTIONS(5127), 1,
      sym__blank,
    STATE(2072), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(4225), 2,
      anon_sym_const,
      anon_sym_volatile,
  [122874] = 3,
    ACTIONS(5129), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3941), 3,
      anon_sym_RPAREN,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
  [122887] = 3,
    ACTIONS(3633), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3631), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122900] = 3,
    ACTIONS(3633), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3631), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122913] = 3,
    ACTIONS(3633), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3631), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122926] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5131), 1,
      sym_comment,
    ACTIONS(5133), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [122945] = 3,
    ACTIONS(3633), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3631), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122958] = 5,
    ACTIONS(5135), 1,
      aux_sym_elif_block_token1,
    ACTIONS(5137), 1,
      sym__blank_line,
    ACTIONS(5140), 1,
      sym__blank,
    STATE(2185), 1,
      aux_sym_if_block_repeat3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [122975] = 3,
    ACTIONS(3661), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3659), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [122988] = 3,
    ACTIONS(3661), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3659), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [123001] = 3,
    ACTIONS(3695), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3693), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [123014] = 3,
    ACTIONS(3695), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3693), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [123027] = 3,
    ACTIONS(3695), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3693), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [123040] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4929), 1,
      sym__blank,
    STATE(2198), 1,
      aux_sym_block_repeat1,
    STATE(2199), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(5143), 2,
      sym_comment,
      sym__blank_line,
  [123057] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4575), 1,
      aux_sym__host_command_argument_list_default_format_token1,
    ACTIONS(5147), 1,
      sym__blank,
    STATE(2136), 1,
      aux_sym__host_command_argument_list_default_format_repeat1,
    ACTIONS(5145), 2,
      sym_comment,
      sym__blank_line,
  [123074] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5149), 1,
      sym__blank,
    STATE(1777), 1,
      aux_sym_block_repeat1,
    STATE(2193), 1,
      aux_sym__host_command_argument_list_default_format_repeat2,
    ACTIONS(5145), 2,
      sym_comment,
      sym__blank_line,
  [123091] = 3,
    ACTIONS(3695), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3693), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [123104] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5152), 1,
      anon_sym_DQUOTE,
    ACTIONS(5154), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(5156), 1,
      sym_hll_escape_sequence,
    STATE(2055), 1,
      aux_sym_hll_string_literal_repeat1,
  [123123] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4969), 1,
      sym__blank,
    STATE(1933), 1,
      aux_sym_block_repeat1,
    STATE(2096), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(2776), 2,
      sym_comment,
      sym__blank_line,
  [123140] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5160), 1,
      sym__blank,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(5158), 2,
      sym_comment,
      sym__blank_line,
  [123157] = 5,
    ACTIONS(4287), 1,
      anon_sym_AMP,
    ACTIONS(4661), 1,
      sym__blank,
    STATE(1991), 1,
      aux_sym_block_repeat1,
    STATE(2558), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123174] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5165), 1,
      sym__blank,
    STATE(2198), 1,
      aux_sym_block_repeat1,
    STATE(2199), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(5163), 2,
      sym_comment,
      sym__blank_line,
  [123191] = 3,
    ACTIONS(3500), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3498), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [123204] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4969), 1,
      sym__blank,
    STATE(1933), 1,
      aux_sym_block_repeat1,
    STATE(2114), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(5143), 2,
      sym_comment,
      sym__blank_line,
  [123221] = 3,
    ACTIONS(3500), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3498), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [123234] = 3,
    ACTIONS(3609), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3607), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [123247] = 3,
    ACTIONS(3609), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3607), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [123260] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5170), 1,
      sym__blank,
    ACTIONS(5168), 4,
      anon_sym_DOT,
      anon_sym_COMMA,
      sym_comment,
      sym__blank_line,
  [123273] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4869), 1,
      sym__blank,
    ACTIONS(4867), 4,
      anon_sym_LPAREN,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [123286] = 4,
    ACTIONS(3202), 1,
      aux_sym_while_block_token1,
    ACTIONS(3573), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3576), 2,
      aux_sym_elif_block_token1,
      sym__blank_line,
  [123301] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5172), 1,
      sym_comment,
    ACTIONS(5174), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [123320] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4973), 1,
      anon_sym_DOT,
    ACTIONS(5178), 1,
      sym__blank,
    STATE(2133), 1,
      aux_sym__command_identifier_repeat1,
    ACTIONS(5176), 2,
      sym_comment,
      sym__blank_line,
  [123337] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5180), 1,
      sym_comment,
    ACTIONS(5182), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [123356] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5184), 1,
      sym_comment,
    ACTIONS(5186), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [123375] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4929), 1,
      sym__blank,
    STATE(2198), 1,
      aux_sym_block_repeat1,
    STATE(2199), 1,
      aux_sym__command_arguments_repeat3,
    ACTIONS(2776), 2,
      sym_comment,
      sym__blank_line,
  [123392] = 3,
    ACTIONS(3691), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3689), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [123405] = 6,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4945), 1,
      sym__blank,
    ACTIONS(5188), 1,
      sym_comment,
    ACTIONS(5190), 1,
      sym__blank_line,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2197), 1,
      aux_sym_macro_definition_repeat1,
  [123424] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4907), 1,
      sym__blank,
    ACTIONS(4905), 4,
      anon_sym_LPAREN,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [123437] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(4579), 1,
      sym__blank,
    STATE(1777), 1,
      aux_sym_block_repeat1,
    STATE(2193), 1,
      aux_sym__host_command_argument_list_default_format_repeat2,
    ACTIONS(5192), 2,
      sym_comment,
      sym__blank_line,
  [123454] = 3,
    ACTIONS(3563), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3561), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [123467] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5194), 1,
      anon_sym_DQUOTE,
    ACTIONS(5196), 1,
      aux_sym_hll_string_literal_token1,
    ACTIONS(5198), 1,
      sym_hll_escape_sequence,
    STATE(2123), 1,
      aux_sym_hll_string_literal_repeat1,
  [123486] = 3,
    ACTIONS(3649), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(3647), 3,
      aux_sym_elif_block_token1,
      aux_sym_while_block_token1,
      sym__blank_line,
  [123499] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5202), 1,
      sym__blank,
    STATE(2453), 1,
      aux_sym_block_repeat1,
    ACTIONS(5200), 2,
      sym_comment,
      sym__blank_line,
  [123513] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5204), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123527] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5206), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123541] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5208), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123555] = 4,
    ACTIONS(1787), 1,
      anon_sym_EQ,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123569] = 4,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(4708), 1,
      anon_sym_RPAREN,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123583] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5210), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123597] = 4,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2162), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123611] = 4,
    ACTIONS(2947), 1,
      anon_sym_RPAREN,
    ACTIONS(2951), 1,
      sym__blank,
    STATE(2233), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123625] = 4,
    ACTIONS(5212), 1,
      anon_sym_RPAREN,
    ACTIONS(5214), 1,
      sym__blank,
    STATE(2234), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123639] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5216), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123653] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5218), 1,
      anon_sym_EQ,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123667] = 4,
    ACTIONS(3044), 1,
      anon_sym_RPAREN,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123681] = 4,
    ACTIONS(2953), 1,
      anon_sym_RPAREN,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123695] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5220), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123709] = 4,
    ACTIONS(2953), 1,
      anon_sym_RPAREN,
    ACTIONS(2955), 1,
      sym__blank,
    STATE(2242), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123723] = 4,
    ACTIONS(5222), 1,
      anon_sym_RPAREN,
    ACTIONS(5224), 1,
      sym__blank,
    STATE(2243), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123737] = 4,
    ACTIONS(5226), 1,
      anon_sym_RPAREN,
    ACTIONS(5228), 1,
      sym__blank,
    STATE(2244), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123751] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5230), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123765] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5232), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123779] = 4,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    STATE(2067), 1,
      sym_option_expression,
    STATE(2368), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123793] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5236), 1,
      sym__blank,
    STATE(31), 1,
      aux_sym_block_repeat1,
    ACTIONS(5234), 2,
      sym_comment,
      sym__blank_line,
  [123807] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5238), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123821] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5240), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123835] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5242), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123849] = 4,
    ACTIONS(5244), 1,
      anon_sym_RPAREN,
    ACTIONS(5246), 1,
      sym__blank,
    STATE(2249), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123863] = 4,
    ACTIONS(5248), 1,
      anon_sym_RPAREN,
    ACTIONS(5250), 1,
      anon_sym_COMMA,
    STATE(2415), 1,
      aux_sym_hll_parameter_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123877] = 4,
    ACTIONS(2807), 1,
      anon_sym_RPAREN,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123891] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5254), 1,
      sym__blank,
    STATE(2446), 1,
      aux_sym_block_repeat1,
    ACTIONS(5252), 2,
      sym_comment,
      sym__blank_line,
  [123905] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5256), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123919] = 4,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(4784), 1,
      anon_sym_RPAREN,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123933] = 4,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    STATE(2368), 1,
      aux_sym__command_arguments_repeat2,
    STATE(2463), 1,
      sym_option_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123947] = 4,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(5258), 1,
      anon_sym_RPAREN,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123961] = 4,
    ACTIONS(5260), 1,
      sym_identifier,
    ACTIONS(5262), 1,
      anon_sym_LPAREN,
    ACTIONS(5264), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123975] = 4,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    ACTIONS(5266), 1,
      sym_identifier,
    STATE(1311), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [123989] = 4,
    ACTIONS(5268), 1,
      sym__blank,
    STATE(2003), 1,
      aux_sym_block_repeat1,
    STATE(2105), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124003] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5272), 1,
      sym__blank,
    STATE(2398), 1,
      aux_sym_block_repeat1,
    ACTIONS(5270), 2,
      sym_comment,
      sym__blank_line,
  [124017] = 4,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(5274), 1,
      anon_sym_RPAREN,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124031] = 4,
    ACTIONS(5276), 1,
      sym__blank,
    STATE(2023), 1,
      aux_sym_block_repeat1,
    STATE(2173), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124045] = 4,
    ACTIONS(5278), 1,
      anon_sym_RPAREN,
    ACTIONS(5280), 1,
      sym__blank,
    STATE(2262), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124059] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5284), 1,
      sym__blank,
    STATE(110), 1,
      aux_sym_block_repeat1,
    ACTIONS(5282), 2,
      sym_comment,
      sym__blank_line,
  [124073] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5109), 1,
      anon_sym_EQ,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124087] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5287), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124101] = 4,
    ACTIONS(5289), 1,
      anon_sym_RPAREN,
    ACTIONS(5291), 1,
      sym__blank,
    STATE(2265), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124115] = 4,
    ACTIONS(5293), 1,
      anon_sym_RPAREN,
    ACTIONS(5295), 1,
      sym__blank,
    STATE(2266), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124129] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5297), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124143] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5299), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124157] = 4,
    ACTIONS(5301), 1,
      anon_sym_RPAREN,
    ACTIONS(5303), 1,
      sym__blank,
    STATE(2268), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124171] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5305), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124185] = 4,
    ACTIONS(5307), 1,
      sym_identifier,
    ACTIONS(5309), 1,
      anon_sym_LPAREN,
    ACTIONS(5311), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124199] = 4,
    ACTIONS(5313), 1,
      sym_identifier,
    ACTIONS(5315), 1,
      anon_sym_DQUOTE,
    STATE(1017), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124213] = 3,
    ACTIONS(5320), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(5317), 2,
      aux_sym_elif_block_token1,
      sym__blank_line,
  [124225] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5323), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124239] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5325), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124253] = 4,
    ACTIONS(5327), 1,
      anon_sym_RPAREN,
    ACTIONS(5329), 1,
      sym__blank,
    STATE(2387), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124267] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5333), 1,
      sym__blank,
    STATE(2048), 1,
      aux_sym_block_repeat1,
    ACTIONS(5331), 2,
      sym_comment,
      sym__blank_line,
  [124281] = 4,
    ACTIONS(5335), 1,
      anon_sym_RPAREN,
    ACTIONS(5337), 1,
      sym__blank,
    STATE(2277), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124295] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5339), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124309] = 4,
    ACTIONS(5341), 1,
      anon_sym_RPAREN,
    ACTIONS(5343), 1,
      sym__blank,
    STATE(2280), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124323] = 4,
    ACTIONS(5345), 1,
      anon_sym_RPAREN,
    ACTIONS(5347), 1,
      sym__blank,
    STATE(2281), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124337] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5349), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124351] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5351), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124365] = 4,
    ACTIONS(5353), 1,
      anon_sym_RPAREN,
    ACTIONS(5355), 1,
      sym__blank,
    STATE(2283), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124379] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5357), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124393] = 4,
    ACTIONS(5359), 1,
      sym_identifier,
    ACTIONS(5361), 1,
      anon_sym_LPAREN,
    ACTIONS(5363), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124407] = 4,
    ACTIONS(5365), 1,
      sym_identifier,
    ACTIONS(5367), 1,
      anon_sym_DQUOTE,
    STATE(969), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124421] = 4,
    ACTIONS(2957), 1,
      anon_sym_RPAREN,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124435] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5369), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124449] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5371), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124463] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5333), 1,
      sym__blank,
    STATE(2048), 1,
      aux_sym_block_repeat1,
    ACTIONS(5373), 2,
      sym_comment,
      sym__blank_line,
  [124477] = 4,
    ACTIONS(5375), 1,
      anon_sym_RPAREN,
    ACTIONS(5377), 1,
      sym__blank,
    STATE(2293), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124491] = 4,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(5379), 1,
      anon_sym_RPAREN,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124505] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5381), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124519] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5383), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124533] = 4,
    ACTIONS(5385), 1,
      anon_sym_RPAREN,
    ACTIONS(5387), 1,
      sym__blank,
    STATE(2297), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124547] = 4,
    ACTIONS(5389), 1,
      anon_sym_RPAREN,
    ACTIONS(5391), 1,
      sym__blank,
    STATE(2298), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124561] = 4,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2875), 1,
      anon_sym_RPAREN,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124575] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5393), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124589] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5395), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124603] = 4,
    ACTIONS(5397), 1,
      anon_sym_RPAREN,
    ACTIONS(5399), 1,
      sym__blank,
    STATE(2300), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124617] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5401), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124631] = 4,
    ACTIONS(5403), 1,
      sym_identifier,
    ACTIONS(5405), 1,
      anon_sym_DQUOTE,
    STATE(843), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124645] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5407), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124659] = 4,
    ACTIONS(2957), 1,
      anon_sym_RPAREN,
    ACTIONS(2959), 1,
      sym__blank,
    STATE(2394), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124673] = 4,
    ACTIONS(5409), 1,
      anon_sym_RPAREN,
    ACTIONS(5411), 1,
      sym__blank,
    STATE(2307), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124687] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5415), 1,
      sym__blank,
    STATE(99), 1,
      aux_sym_block_repeat1,
    ACTIONS(5413), 2,
      sym_comment,
      sym__blank_line,
  [124701] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5417), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124715] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5419), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124729] = 4,
    ACTIONS(5421), 1,
      anon_sym_RPAREN,
    ACTIONS(5423), 1,
      sym__blank,
    STATE(2310), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124743] = 4,
    ACTIONS(5425), 1,
      anon_sym_RPAREN,
    ACTIONS(5427), 1,
      sym__blank,
    STATE(2311), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124757] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5429), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124771] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5431), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124785] = 4,
    ACTIONS(5433), 1,
      anon_sym_RPAREN,
    ACTIONS(5435), 1,
      sym__blank,
    STATE(2313), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124799] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5437), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124813] = 4,
    ACTIONS(5313), 1,
      sym_identifier,
    ACTIONS(5439), 1,
      anon_sym_DQUOTE,
    STATE(810), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124827] = 4,
    ACTIONS(3044), 1,
      anon_sym_RPAREN,
    ACTIONS(3046), 1,
      sym__blank,
    STATE(2273), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124841] = 4,
    ACTIONS(3360), 1,
      anon_sym_EQ,
    ACTIONS(3362), 1,
      sym__blank,
    STATE(2231), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124855] = 4,
    ACTIONS(5441), 1,
      anon_sym_RPAREN,
    ACTIONS(5443), 1,
      sym__blank,
    STATE(2319), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124869] = 4,
    ACTIONS(5445), 1,
      anon_sym_RPAREN,
    ACTIONS(5447), 1,
      sym__blank,
    STATE(2292), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124883] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5449), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124897] = 4,
    ACTIONS(5451), 1,
      anon_sym_RPAREN,
    ACTIONS(5453), 1,
      sym__blank,
    STATE(2322), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124911] = 4,
    ACTIONS(5455), 1,
      anon_sym_RPAREN,
    ACTIONS(5457), 1,
      sym__blank,
    STATE(2323), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124925] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5459), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124939] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5461), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124953] = 4,
    ACTIONS(5463), 1,
      anon_sym_RPAREN,
    ACTIONS(5465), 1,
      sym__blank,
    STATE(2325), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124967] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5467), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124981] = 4,
    ACTIONS(4178), 1,
      anon_sym_DQUOTE,
    ACTIONS(5469), 1,
      sym_identifier,
    STATE(2556), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [124995] = 4,
    ACTIONS(5471), 1,
      anon_sym_RPAREN,
    ACTIONS(5473), 1,
      sym__blank,
    STATE(2396), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125009] = 4,
    ACTIONS(3774), 1,
      aux_sym_while_block_token1,
    ACTIONS(5475), 1,
      sym__blank_line,
    STATE(2328), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125023] = 4,
    ACTIONS(5478), 1,
      anon_sym_RPAREN,
    ACTIONS(5480), 1,
      sym__blank,
    STATE(2403), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125037] = 4,
    ACTIONS(5482), 1,
      anon_sym_RPAREN,
    ACTIONS(5484), 1,
      sym__blank,
    STATE(2332), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125051] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5486), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125065] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5488), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125079] = 4,
    ACTIONS(5490), 1,
      anon_sym_RPAREN,
    ACTIONS(5492), 1,
      sym__blank,
    STATE(2335), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125093] = 4,
    ACTIONS(5494), 1,
      anon_sym_RPAREN,
    ACTIONS(5496), 1,
      sym__blank,
    STATE(2336), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125107] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5498), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125121] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5500), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125135] = 4,
    ACTIONS(5502), 1,
      anon_sym_RPAREN,
    ACTIONS(5504), 1,
      sym__blank,
    STATE(2338), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125149] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5506), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125163] = 4,
    ACTIONS(5508), 1,
      sym_identifier,
    ACTIONS(5510), 1,
      anon_sym_DQUOTE,
    STATE(1255), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125177] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5512), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125191] = 4,
    ACTIONS(2967), 1,
      anon_sym_RPAREN,
    ACTIONS(2969), 1,
      sym__blank,
    STATE(2232), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125205] = 4,
    ACTIONS(5508), 1,
      sym_identifier,
    ACTIONS(5514), 1,
      anon_sym_DQUOTE,
    STATE(909), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125219] = 4,
    ACTIONS(5250), 1,
      anon_sym_COMMA,
    ACTIONS(5516), 1,
      anon_sym_RPAREN,
    STATE(2415), 1,
      aux_sym_hll_parameter_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125233] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5520), 1,
      sym__blank,
    STATE(123), 1,
      aux_sym_block_repeat1,
    ACTIONS(5518), 2,
      sym_comment,
      sym__blank_line,
  [125247] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5522), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125261] = 4,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2210), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125275] = 4,
    ACTIONS(5524), 1,
      sym__blank,
    STATE(2025), 1,
      aux_sym_block_repeat1,
    STATE(2211), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125289] = 4,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2214), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125303] = 4,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    ACTIONS(5526), 1,
      sym_identifier,
    STATE(1374), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125317] = 4,
    ACTIONS(5528), 1,
      anon_sym_RPAREN,
    ACTIONS(5530), 1,
      sym__blank,
    STATE(2238), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125331] = 4,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2208), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125345] = 4,
    ACTIONS(5532), 1,
      sym__blank,
    STATE(2026), 1,
      aux_sym_block_repeat1,
    STATE(2160), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125359] = 4,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2183), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125373] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5534), 1,
      sym_identifier,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125387] = 4,
    ACTIONS(5536), 1,
      anon_sym_RPAREN,
    ACTIONS(5538), 1,
      sym__blank,
    STATE(2226), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125401] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2873), 1,
      sym__blank,
    STATE(1135), 1,
      aux_sym_block_repeat1,
    ACTIONS(2871), 2,
      sym_comment,
      sym__blank_line,
  [125415] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5540), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125429] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5542), 1,
      sym_identifier,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125443] = 4,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2170), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125457] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5544), 1,
      sym_identifier,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125471] = 4,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2127), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125485] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5254), 1,
      sym__blank,
    STATE(2446), 1,
      aux_sym_block_repeat1,
    ACTIONS(5546), 2,
      sym_comment,
      sym__blank_line,
  [125499] = 4,
    ACTIONS(3048), 1,
      anon_sym_RPAREN,
    ACTIONS(3050), 1,
      sym__blank,
    STATE(2286), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125513] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5548), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125527] = 4,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2095), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125541] = 4,
    ACTIONS(5550), 1,
      anon_sym_RPAREN,
    ACTIONS(5552), 1,
      sym__blank,
    STATE(2287), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125555] = 3,
    ACTIONS(5170), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(5168), 2,
      anon_sym_DOT,
      anon_sym_COMMA,
  [125567] = 4,
    ACTIONS(5554), 1,
      anon_sym_SLASH,
    STATE(2368), 1,
      aux_sym__command_arguments_repeat2,
    STATE(2580), 1,
      sym_option_expression,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125581] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5557), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125595] = 4,
    ACTIONS(5559), 1,
      sym_identifier,
    ACTIONS(5561), 1,
      anon_sym_LPAREN,
    ACTIONS(5563), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125609] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5567), 1,
      sym__blank,
    STATE(2466), 1,
      aux_sym_block_repeat1,
    ACTIONS(5565), 2,
      sym_comment,
      sym__blank_line,
  [125623] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5569), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125637] = 4,
    ACTIONS(5571), 1,
      anon_sym_LPAREN,
    ACTIONS(5573), 1,
      anon_sym_DOT,
    STATE(2400), 1,
      aux_sym__function_identifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125651] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5575), 1,
      anon_sym_DOT_DOT,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125665] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5577), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125679] = 4,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    STATE(2201), 1,
      sym_option_expression,
    STATE(2368), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125693] = 4,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    ACTIONS(5579), 1,
      sym_identifier,
    STATE(518), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125707] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5581), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125721] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5583), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125735] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5585), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125749] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5587), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125763] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5589), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125777] = 4,
    ACTIONS(5591), 1,
      aux_sym_while_block_token1,
    ACTIONS(5593), 1,
      sym__blank_line,
    STATE(2328), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125791] = 4,
    ACTIONS(5250), 1,
      anon_sym_COMMA,
    ACTIONS(5595), 1,
      anon_sym_RPAREN,
    STATE(2343), 1,
      aux_sym_hll_parameter_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125805] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2919), 1,
      sym__blank,
    STATE(1051), 1,
      aux_sym_block_repeat1,
    ACTIONS(2917), 2,
      sym_comment,
      sym__blank_line,
  [125819] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5599), 1,
      sym__blank,
    STATE(138), 1,
      aux_sym_block_repeat1,
    ACTIONS(5597), 2,
      sym_comment,
      sym__blank_line,
  [125833] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5601), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125847] = 4,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2102), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125861] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5202), 1,
      sym__blank,
    STATE(2453), 1,
      aux_sym_block_repeat1,
    ACTIONS(5603), 2,
      sym_comment,
      sym__blank_line,
  [125875] = 4,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(4871), 1,
      anon_sym_RPAREN,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125889] = 4,
    ACTIONS(2987), 1,
      anon_sym_RPAREN,
    ACTIONS(5605), 1,
      anon_sym_COMMA,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125903] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5608), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125917] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5612), 1,
      sym__blank,
    STATE(146), 1,
      aux_sym_block_repeat1,
    ACTIONS(5610), 2,
      sym_comment,
      sym__blank_line,
  [125931] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5614), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125945] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5111), 1,
      anon_sym_COMMA,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125959] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5616), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125973] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5618), 1,
      anon_sym_EQ,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [125987] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5620), 1,
      aux_sym_if_block_token1,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126001] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5622), 1,
      sym_identifier,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126015] = 4,
    ACTIONS(5624), 1,
      anon_sym_LPAREN,
    ACTIONS(5626), 1,
      anon_sym_DOT,
    STATE(2400), 1,
      aux_sym__function_identifier_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126029] = 4,
    ACTIONS(1757), 1,
      aux_sym_block_token1,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126043] = 5,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5629), 1,
      aux_sym_block_token1,
    ACTIONS(5631), 1,
      sym_comment,
    ACTIONS(5634), 1,
      sym__blank,
    STATE(2402), 1,
      aux_sym_subroutine_block_repeat1,
  [126059] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5636), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126073] = 4,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2053), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126087] = 4,
    ACTIONS(3040), 1,
      anon_sym_RPAREN,
    ACTIONS(3042), 1,
      sym__blank,
    STATE(2422), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126101] = 4,
    ACTIONS(5638), 1,
      anon_sym_RPAREN,
    ACTIONS(5640), 1,
      sym__blank,
    STATE(2424), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126115] = 4,
    ACTIONS(5593), 1,
      sym__blank_line,
    ACTIONS(5642), 1,
      aux_sym_while_block_token1,
    STATE(2328), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126129] = 4,
    ACTIONS(5644), 1,
      anon_sym_RPAREN,
    ACTIONS(5646), 1,
      sym__blank,
    STATE(2222), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126143] = 4,
    ACTIONS(4508), 1,
      anon_sym_EQ,
    ACTIONS(4510), 1,
      sym__blank,
    STATE(2411), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126157] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5648), 1,
      sym_identifier,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126171] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5650), 1,
      anon_sym_EQ,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126185] = 4,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2903), 1,
      anon_sym_RPAREN,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126199] = 4,
    ACTIONS(1785), 1,
      aux_sym_block_token1,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126213] = 4,
    ACTIONS(5593), 1,
      sym__blank_line,
    ACTIONS(5652), 1,
      aux_sym_while_block_token1,
    STATE(2328), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126227] = 4,
    ACTIONS(5654), 1,
      anon_sym_RPAREN,
    ACTIONS(5656), 1,
      anon_sym_COMMA,
    STATE(2415), 1,
      aux_sym_hll_parameter_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126241] = 4,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2175), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126255] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5661), 1,
      sym__blank,
    STATE(32), 1,
      aux_sym_block_repeat1,
    ACTIONS(5659), 2,
      sym_comment,
      sym__blank_line,
  [126269] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5663), 1,
      sym__blank,
    STATE(113), 1,
      aux_sym_block_repeat1,
    ACTIONS(5282), 2,
      sym_comment,
      sym__blank_line,
  [126283] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5667), 1,
      sym__blank,
    STATE(34), 1,
      aux_sym_block_repeat1,
    ACTIONS(5665), 2,
      sym_comment,
      sym__blank_line,
  [126297] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2923), 1,
      sym__blank,
    STATE(1137), 1,
      aux_sym_block_repeat1,
    ACTIONS(2921), 2,
      sym_comment,
      sym__blank_line,
  [126311] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5669), 1,
      sym_identifier,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126325] = 4,
    ACTIONS(3052), 1,
      anon_sym_RPAREN,
    ACTIONS(3496), 1,
      sym__blank,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126339] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5673), 1,
      sym__blank,
    STATE(2468), 1,
      aux_sym_block_repeat1,
    ACTIONS(5671), 2,
      sym_comment,
      sym__blank_line,
  [126353] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5675), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126367] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5679), 1,
      sym__blank,
    STATE(390), 1,
      aux_sym_block_repeat1,
    ACTIONS(5677), 2,
      sym_comment,
      sym__blank_line,
  [126381] = 4,
    ACTIONS(3052), 1,
      anon_sym_RPAREN,
    ACTIONS(3054), 1,
      sym__blank,
    STATE(2432), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126395] = 4,
    ACTIONS(5681), 1,
      anon_sym_RPAREN,
    ACTIONS(5683), 1,
      sym__blank,
    STATE(2433), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126409] = 4,
    ACTIONS(5685), 1,
      anon_sym_RPAREN,
    ACTIONS(5687), 1,
      sym__blank,
    STATE(2434), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126423] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5691), 1,
      sym__blank,
    STATE(1776), 1,
      aux_sym_block_repeat1,
    ACTIONS(5689), 2,
      sym_comment,
      sym__blank_line,
  [126437] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2915), 1,
      sym__blank,
    STATE(1032), 1,
      aux_sym_block_repeat1,
    ACTIONS(2911), 2,
      sym_comment,
      sym__blank_line,
  [126451] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5695), 1,
      sym__blank,
    STATE(35), 1,
      aux_sym_block_repeat1,
    ACTIONS(5693), 2,
      sym_comment,
      sym__blank_line,
  [126465] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5697), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126479] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5699), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126493] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5701), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126507] = 4,
    ACTIONS(4728), 1,
      aux_sym_while_block_token1,
    ACTIONS(5593), 1,
      sym__blank_line,
    STATE(2328), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126521] = 4,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    STATE(2098), 1,
      sym_option_expression,
    STATE(2368), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126535] = 4,
    ACTIONS(5703), 1,
      anon_sym_RPAREN,
    ACTIONS(5705), 1,
      sym__blank,
    STATE(2443), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126549] = 4,
    ACTIONS(4945), 1,
      sym__blank,
    STATE(2144), 1,
      aux_sym_block_repeat1,
    STATE(2177), 1,
      aux_sym_macro_definition_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126563] = 4,
    ACTIONS(4775), 1,
      aux_sym_while_block_token1,
    ACTIONS(5593), 1,
      sym__blank_line,
    STATE(2328), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126577] = 4,
    ACTIONS(5593), 1,
      sym__blank_line,
    ACTIONS(5707), 1,
      aux_sym_while_block_token1,
    STATE(2328), 1,
      aux_sym_if_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126591] = 4,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(2895), 1,
      anon_sym_RPAREN,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126605] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(5709), 3,
      anon_sym_STAR,
      anon_sym_QMARK,
      sym_identifier,
  [126615] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5711), 1,
      anon_sym_RPAREN,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126629] = 4,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    STATE(2115), 1,
      sym_option_expression,
    STATE(2368), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126643] = 4,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(4749), 1,
      anon_sym_RPAREN,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126657] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(4659), 1,
      anon_sym_SLASH,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126671] = 4,
    ACTIONS(5713), 1,
      sym_identifier,
    ACTIONS(5715), 1,
      anon_sym_LPAREN,
    ACTIONS(5717), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126685] = 4,
    ACTIONS(2833), 1,
      anon_sym_COMMA,
    ACTIONS(5719), 1,
      anon_sym_RPAREN,
    STATE(2391), 1,
      aux_sym_hll_argument_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126699] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5723), 1,
      sym__blank,
    STATE(155), 1,
      aux_sym_block_repeat1,
    ACTIONS(5721), 2,
      sym_comment,
      sym__blank_line,
  [126713] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5727), 1,
      sym__blank,
    STATE(33), 1,
      aux_sym_block_repeat1,
    ACTIONS(5725), 2,
      sym_comment,
      sym__blank_line,
  [126727] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5272), 1,
      sym__blank,
    STATE(2398), 1,
      aux_sym_block_repeat1,
    ACTIONS(5729), 2,
      sym_comment,
      sym__blank_line,
  [126741] = 4,
    ACTIONS(5731), 1,
      sym_identifier,
    ACTIONS(5733), 1,
      anon_sym_LPAREN,
    ACTIONS(5735), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126755] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5737), 1,
      aux_sym_if_block_token1,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126769] = 4,
    ACTIONS(5250), 1,
      anon_sym_COMMA,
    ACTIONS(5739), 1,
      anon_sym_RPAREN,
    STATE(2246), 1,
      aux_sym_hll_parameter_list_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126783] = 4,
    ACTIONS(5741), 1,
      anon_sym_COMMA,
    ACTIONS(5743), 1,
      sym__blank,
    STATE(1011), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126797] = 4,
    ACTIONS(5403), 1,
      sym_identifier,
    ACTIONS(5745), 1,
      anon_sym_DQUOTE,
    STATE(1139), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126811] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5254), 1,
      sym__blank,
    STATE(2446), 1,
      aux_sym_block_repeat1,
    ACTIONS(2883), 2,
      sym_comment,
      sym__blank_line,
  [126825] = 4,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    STATE(2248), 1,
      sym_option_expression,
    STATE(2368), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126839] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5749), 1,
      sym__blank,
    ACTIONS(5747), 3,
      anon_sym_DOT,
      sym_comment,
      sym__blank_line,
  [126851] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5753), 1,
      sym__blank,
    STATE(157), 1,
      aux_sym_block_repeat1,
    ACTIONS(5751), 2,
      sym_comment,
      sym__blank_line,
  [126865] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5567), 1,
      sym__blank,
    STATE(2466), 1,
      aux_sym_block_repeat1,
    ACTIONS(5200), 2,
      sym_comment,
      sym__blank_line,
  [126879] = 4,
    ACTIONS(5365), 1,
      sym_identifier,
    ACTIONS(5745), 1,
      anon_sym_DQUOTE,
    STATE(1139), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126893] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5254), 1,
      sym__blank,
    STATE(2446), 1,
      aux_sym_block_repeat1,
    ACTIONS(5755), 2,
      sym_comment,
      sym__blank_line,
  [126907] = 4,
    ACTIONS(456), 1,
      anon_sym_SLASH,
    STATE(2362), 1,
      sym_option_expression,
    STATE(2368), 1,
      aux_sym__command_arguments_repeat2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126921] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5202), 1,
      sym__blank,
    STATE(2453), 1,
      aux_sym_block_repeat1,
    ACTIONS(5729), 2,
      sym_comment,
      sym__blank_line,
  [126935] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5757), 1,
      aux_sym_if_block_token1,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126949] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5759), 1,
      aux_sym_if_block_token1,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126963] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5761), 1,
      aux_sym_if_block_token1,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126977] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5763), 1,
      anon_sym_EQ,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [126991] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5765), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127005] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5202), 1,
      sym__blank,
    STATE(2453), 1,
      aux_sym_block_repeat1,
    ACTIONS(5270), 2,
      sym_comment,
      sym__blank_line,
  [127019] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5767), 1,
      anon_sym_RBRACK,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127033] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5202), 1,
      sym__blank,
    STATE(2453), 1,
      aux_sym_block_repeat1,
    ACTIONS(5565), 2,
      sym_comment,
      sym__blank_line,
  [127047] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5202), 1,
      sym__blank,
    STATE(2453), 1,
      aux_sym_block_repeat1,
    ACTIONS(5671), 2,
      sym_comment,
      sym__blank_line,
  [127061] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5769), 1,
      sym__blank,
    STATE(2467), 1,
      aux_sym_block_repeat1,
    ACTIONS(5565), 2,
      sym_comment,
      sym__blank_line,
  [127075] = 4,
    ACTIONS(3496), 1,
      sym__blank,
    ACTIONS(5771), 1,
      aux_sym_if_block_token1,
    STATE(1013), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127089] = 4,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5773), 1,
      sym__blank,
    STATE(2476), 1,
      aux_sym_block_repeat1,
    ACTIONS(5270), 2,
      sym_comment,
      sym__blank_line,
  [127103] = 4,
    ACTIONS(5775), 1,
      sym_identifier,
    ACTIONS(5777), 1,
      anon_sym_LPAREN,
    ACTIONS(5779), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127117] = 3,
    ACTIONS(5781), 1,
      sym__blank,
    STATE(109), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127128] = 3,
    ACTIONS(3153), 1,
      anon_sym_RPAREN,
    ACTIONS(3155), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127139] = 3,
    ACTIONS(5783), 1,
      sym__blank,
    STATE(2057), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127150] = 3,
    ACTIONS(71), 1,
      anon_sym_DQUOTE,
    STATE(518), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127161] = 3,
    ACTIONS(3058), 1,
      anon_sym_LPAREN,
    STATE(1128), 1,
      sym_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127172] = 3,
    ACTIONS(5785), 1,
      sym_identifier,
    STATE(985), 1,
      sym__hll_field_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127183] = 3,
    ACTIONS(5787), 1,
      anon_sym_LPAREN,
    STATE(1129), 1,
      sym_hll_var_call_expression_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127194] = 3,
    ACTIONS(5789), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5791), 1,
      anon_sym_AMP,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127205] = 3,
    ACTIONS(5793), 1,
      sym__blank,
    STATE(129), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127216] = 3,
    ACTIONS(5202), 1,
      sym__blank,
    STATE(2453), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127227] = 3,
    ACTIONS(5795), 1,
      sym__blank,
    STATE(1733), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127238] = 3,
    ACTIONS(5797), 1,
      sym_identifier,
    ACTIONS(5799), 1,
      anon_sym_LPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127249] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1901), 1,
      sym__blank,
    ACTIONS(1903), 2,
      sym_comment,
      sym__blank_line,
  [127260] = 3,
    ACTIONS(5801), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5803), 1,
      anon_sym_AMP,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127271] = 3,
    ACTIONS(5805), 1,
      sym__blank,
    STATE(1767), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127282] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(5807), 2,
      aux_sym_character_token1,
      aux_sym_character_token2,
  [127291] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5809), 1,
      aux_sym_hll_char_literal_token1,
    ACTIONS(5811), 1,
      sym_hll_escape_sequence,
  [127304] = 3,
    ACTIONS(5813), 1,
      sym_identifier,
    STATE(916), 1,
      sym__hll_field_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127315] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5817), 1,
      sym__blank,
    ACTIONS(5815), 2,
      sym_comment,
      sym__blank_line,
  [127326] = 3,
    ACTIONS(2275), 1,
      sym__blank,
    ACTIONS(2277), 1,
      aux_sym_symbol_token3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127337] = 3,
    ACTIONS(5405), 1,
      anon_sym_DQUOTE,
    STATE(843), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127348] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5819), 3,
      aux_sym__hll_var_definition_command_token1,
      sym_comment,
      sym__blank_line,
  [127357] = 3,
    ACTIONS(5510), 1,
      anon_sym_DQUOTE,
    STATE(1255), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127368] = 3,
    ACTIONS(5821), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5823), 1,
      anon_sym_AMP,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127379] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5825), 1,
      aux_sym_hll_char_literal_token1,
    ACTIONS(5827), 1,
      sym_hll_escape_sequence,
  [127392] = 3,
    ACTIONS(5829), 1,
      sym__blank,
    STATE(108), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127403] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(5831), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
  [127412] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5833), 1,
      aux_sym_hll_char_literal_token1,
    ACTIONS(5835), 1,
      sym_hll_escape_sequence,
  [127425] = 3,
    ACTIONS(5837), 1,
      sym_identifier,
    STATE(918), 1,
      sym__hll_field_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127436] = 3,
    ACTIONS(5839), 1,
      sym__blank,
    STATE(2421), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127447] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2275), 1,
      sym__blank,
    ACTIONS(2277), 2,
      sym_comment,
      sym__blank_line,
  [127458] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5841), 1,
      aux_sym_hll_char_literal_token1,
    ACTIONS(5843), 1,
      sym_hll_escape_sequence,
  [127471] = 3,
    ACTIONS(5845), 1,
      sym__blank,
    STATE(122), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127482] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5847), 1,
      aux_sym_hll_char_literal_token1,
    ACTIONS(5849), 1,
      sym_hll_escape_sequence,
  [127495] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5851), 1,
      aux_sym_hll_char_literal_token1,
    ACTIONS(5853), 1,
      sym_hll_escape_sequence,
  [127508] = 3,
    ACTIONS(5855), 1,
      sym__blank,
    STATE(2399), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127519] = 3,
    ACTIONS(1921), 1,
      sym__blank,
    ACTIONS(1923), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127530] = 3,
    ACTIONS(5857), 1,
      sym__blank,
    STATE(2354), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127541] = 3,
    ACTIONS(5859), 1,
      sym__blank,
    STATE(501), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127552] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(5654), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [127561] = 3,
    ACTIONS(5861), 1,
      sym_identifier,
    STATE(812), 1,
      sym__hll_field_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127572] = 3,
    ACTIONS(5863), 1,
      sym__blank,
    STATE(2358), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127583] = 3,
    ACTIONS(5865), 1,
      sym__blank,
    STATE(115), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127594] = 3,
    ACTIONS(5514), 1,
      anon_sym_DQUOTE,
    STATE(909), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127605] = 3,
    ACTIONS(5867), 1,
      sym__blank,
    STATE(131), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127616] = 3,
    ACTIONS(187), 1,
      anon_sym_AMP,
    STATE(1369), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127627] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1921), 1,
      sym__blank,
    ACTIONS(1923), 2,
      sym_comment,
      sym__blank_line,
  [127638] = 3,
    ACTIONS(5869), 1,
      sym__blank,
    STATE(132), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127649] = 3,
    ACTIONS(5871), 1,
      sym__blank,
    STATE(133), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127660] = 3,
    ACTIONS(5873), 1,
      sym__blank,
    STATE(134), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127671] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5877), 1,
      sym__blank,
    ACTIONS(5875), 2,
      sym_comment,
      sym__blank_line,
  [127682] = 3,
    ACTIONS(5879), 1,
      sym__blank,
    STATE(128), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127693] = 3,
    ACTIONS(5881), 1,
      anon_sym_LPAREN,
    STATE(1237), 1,
      sym_hll_var_call_expression_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127704] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5883), 1,
      aux_sym__hll_var_definition_command_token1,
    ACTIONS(5885), 2,
      sym_comment,
      sym__blank_line,
  [127715] = 3,
    ACTIONS(154), 1,
      anon_sym_AMP,
    STATE(1335), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127726] = 3,
    ACTIONS(5887), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5889), 1,
      anon_sym_AMP,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127737] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5817), 1,
      sym__blank,
    ACTIONS(5891), 2,
      sym_comment,
      sym__blank_line,
  [127748] = 3,
    ACTIONS(5893), 1,
      sym__blank,
    STATE(402), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127759] = 3,
    ACTIONS(2257), 1,
      sym__blank,
    ACTIONS(2259), 1,
      aux_sym_symbol_token3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127770] = 3,
    ACTIONS(5367), 1,
      anon_sym_DQUOTE,
    STATE(969), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127781] = 3,
    ACTIONS(5895), 1,
      sym__blank,
    STATE(1803), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127792] = 3,
    ACTIONS(5439), 1,
      anon_sym_DQUOTE,
    STATE(810), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127803] = 3,
    ACTIONS(5897), 1,
      sym__blank,
    STATE(1896), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127814] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5817), 1,
      sym__blank,
    ACTIONS(5899), 2,
      sym_comment,
      sym__blank_line,
  [127825] = 3,
    ACTIONS(5901), 1,
      sym__blank,
    STATE(141), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127836] = 3,
    ACTIONS(5904), 1,
      sym__blank,
    STATE(1492), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127847] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5906), 1,
      aux_sym_hll_char_literal_token1,
    ACTIONS(5908), 1,
      sym_hll_escape_sequence,
  [127860] = 3,
    ACTIONS(5910), 1,
      sym__blank,
    STATE(547), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127871] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5147), 1,
      sym__blank,
    ACTIONS(5145), 2,
      sym_comment,
      sym__blank_line,
  [127882] = 3,
    ACTIONS(5912), 1,
      sym__blank,
    STATE(2410), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127893] = 3,
    ACTIONS(5914), 1,
      sym__blank,
    STATE(143), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127904] = 3,
    ACTIONS(5916), 1,
      sym__blank,
    STATE(94), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127915] = 3,
    ACTIONS(5918), 1,
      sym__blank,
    STATE(504), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127926] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2257), 1,
      sym__blank,
    ACTIONS(2259), 2,
      sym_comment,
      sym__blank_line,
  [127937] = 3,
    ACTIONS(5920), 1,
      sym__blank,
    STATE(505), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127948] = 3,
    ACTIONS(3213), 1,
      anon_sym_LPAREN,
    STATE(1234), 1,
      sym_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127959] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(5922), 2,
      aux_sym_character_token1,
      aux_sym_character_token2,
  [127968] = 3,
    ACTIONS(2279), 1,
      sym__blank,
    ACTIONS(5924), 1,
      aux_sym_symbol_token3,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [127979] = 2,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
    ACTIONS(5926), 2,
      aux_sym_character_token1,
      aux_sym_character_token2,
  [127988] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5928), 1,
      sym__blank,
    ACTIONS(5163), 2,
      sym_comment,
      sym__blank_line,
  [127999] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(4331), 1,
      sym__line_continuation,
    ACTIONS(5930), 1,
      aux_sym_hll_char_literal_token1,
    ACTIONS(5932), 1,
      sym_hll_escape_sequence,
  [128012] = 3,
    ACTIONS(5934), 1,
      sym__blank,
    STATE(2051), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128023] = 3,
    ACTIONS(5936), 1,
      sym__blank,
    STATE(145), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128034] = 3,
    ACTIONS(715), 1,
      anon_sym_AMP,
    STATE(503), 1,
      sym_macro,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128045] = 3,
    ACTIONS(5938), 1,
      sym_identifier,
    ACTIONS(5940), 1,
      anon_sym_LPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128056] = 3,
    ACTIONS(5942), 1,
      sym__blank,
    STATE(147), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128067] = 3,
    ACTIONS(5944), 1,
      sym__blank,
    STATE(148), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128078] = 3,
    ACTIONS(5946), 1,
      sym__blank,
    STATE(519), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128089] = 3,
    ACTIONS(5948), 1,
      sym__blank,
    STATE(520), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128100] = 3,
    ACTIONS(5950), 1,
      sym__blank,
    STATE(149), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128111] = 3,
    ACTIONS(5952), 1,
      sym__blank,
    STATE(150), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128122] = 3,
    ACTIONS(5954), 1,
      sym__blank,
    STATE(151), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128133] = 3,
    ACTIONS(5956), 1,
      sym__blank,
    STATE(522), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128144] = 3,
    ACTIONS(5958), 1,
      sym__blank,
    STATE(523), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128155] = 3,
    ACTIONS(5960), 1,
      sym__blank,
    STATE(152), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128166] = 3,
    ACTIONS(5962), 1,
      sym__blank,
    STATE(153), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128177] = 3,
    ACTIONS(5964), 1,
      sym__blank,
    STATE(526), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128188] = 3,
    ACTIONS(5966), 1,
      sym__blank,
    STATE(154), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128199] = 3,
    ACTIONS(5968), 1,
      sym__blank,
    STATE(528), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128210] = 3,
    ACTIONS(5315), 1,
      anon_sym_DQUOTE,
    STATE(1017), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128221] = 3,
    ACTIONS(1901), 1,
      sym__blank,
    ACTIONS(1903), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128232] = 3,
    ACTIONS(5254), 1,
      sym__blank,
    STATE(2446), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128243] = 3,
    ACTIONS(5970), 1,
      sym__blank,
    STATE(1747), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128254] = 3,
    ACTIONS(5972), 1,
      sym__blank,
    STATE(1748), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128265] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(5976), 1,
      sym__blank,
    ACTIONS(5974), 2,
      sym_comment,
      sym__blank_line,
  [128276] = 3,
    ACTIONS(3021), 1,
      sym__blank,
    STATE(1314), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128287] = 3,
    ACTIONS(5333), 1,
      sym__blank,
    STATE(2048), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128298] = 3,
    ACTIONS(124), 1,
      anon_sym_DQUOTE,
    STATE(1311), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128309] = 3,
    ACTIONS(5978), 1,
      sym_identifier,
    ACTIONS(5980), 1,
      anon_sym_LPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128320] = 3,
    ACTIONS(5745), 1,
      anon_sym_DQUOTE,
    STATE(1139), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128331] = 3,
    ACTIONS(270), 1,
      anon_sym_LPAREN,
    STATE(432), 1,
      sym_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128342] = 3,
    ACTIONS(5982), 1,
      anon_sym_LPAREN,
    STATE(410), 1,
      sym_hll_var_call_expression_argument_list,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128353] = 3,
    ACTIONS(5984), 1,
      sym__blank,
    STATE(2056), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128364] = 3,
    ACTIONS(193), 1,
      anon_sym_DQUOTE,
    STATE(1374), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128375] = 3,
    ACTIONS(5986), 1,
      sym__blank,
    STATE(158), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128386] = 3,
    ACTIONS(5988), 1,
      sym__blank,
    STATE(548), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128397] = 3,
    ACTIONS(5990), 1,
      sym__blank,
    STATE(159), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128408] = 3,
    ACTIONS(5992), 1,
      sym__blank,
    STATE(120), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128419] = 3,
    ACTIONS(4178), 1,
      anon_sym_DQUOTE,
    STATE(2556), 1,
      sym_string,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128430] = 3,
    ACTIONS(5994), 1,
      sym__blank,
    STATE(130), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128441] = 3,
    ACTIONS(5996), 1,
      sym__blank,
    STATE(550), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128452] = 3,
    ACTIONS(5998), 1,
      sym__blank,
    STATE(144), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128463] = 3,
    ACTIONS(6000), 1,
      sym__blank,
    STATE(2360), 1,
      aux_sym_block_repeat1,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128474] = 2,
    ACTIONS(2355), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128482] = 2,
    ACTIONS(6002), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128490] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6004), 2,
      sym_comment,
      sym__blank_line,
  [128498] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6006), 2,
      sym_comment,
      sym__blank_line,
  [128506] = 2,
    ACTIONS(6008), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128514] = 2,
    ACTIONS(5817), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128522] = 2,
    ACTIONS(6010), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128530] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6012), 2,
      sym_comment,
      sym__blank_line,
  [128538] = 2,
    ACTIONS(6014), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128546] = 2,
    ACTIONS(6016), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128554] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6018), 2,
      sym_comment,
      sym__blank_line,
  [128562] = 2,
    ACTIONS(6020), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128570] = 2,
    ACTIONS(6022), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128578] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6024), 2,
      sym_comment,
      sym__blank_line,
  [128586] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6026), 2,
      sym_comment,
      sym__blank_line,
  [128594] = 2,
    ACTIONS(6028), 1,
      sym_hll_number_literal,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128602] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6030), 1,
      sym_comment,
    ACTIONS(6032), 1,
      sym__blank_line,
  [128612] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6034), 2,
      sym_comment,
      sym__blank_line,
  [128620] = 2,
    ACTIONS(6036), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128628] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2587), 2,
      sym_comment,
      sym__blank_line,
  [128636] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6038), 2,
      sym_comment,
      sym__blank_line,
  [128644] = 2,
    ACTIONS(2929), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128652] = 2,
    ACTIONS(6040), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128660] = 2,
    ACTIONS(6042), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128668] = 2,
    ACTIONS(6044), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128676] = 2,
    ACTIONS(6046), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128684] = 2,
    ACTIONS(6048), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128692] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6050), 1,
      sym_comment,
    ACTIONS(6052), 1,
      sym__blank_line,
  [128702] = 2,
    ACTIONS(1903), 1,
      sym_path,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128710] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1929), 2,
      sym_comment,
      sym__blank_line,
  [128718] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6054), 2,
      sym_comment,
      sym__blank_line,
  [128726] = 2,
    ACTIONS(6008), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128734] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6056), 2,
      sym_comment,
      sym__blank_line,
  [128742] = 2,
    ACTIONS(6058), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128750] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6060), 2,
      sym_comment,
      sym__blank_line,
  [128758] = 2,
    ACTIONS(6062), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128766] = 2,
    ACTIONS(6064), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128774] = 2,
    ACTIONS(6066), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128782] = 2,
    ACTIONS(6068), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128790] = 2,
    ACTIONS(6070), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128798] = 2,
    ACTIONS(6072), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128806] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6074), 2,
      sym_comment,
      sym__blank_line,
  [128814] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6076), 2,
      sym_comment,
      sym__blank_line,
  [128822] = 2,
    ACTIONS(6078), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128830] = 2,
    ACTIONS(6070), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128838] = 2,
    ACTIONS(6080), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128846] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6082), 2,
      sym_comment,
      sym__blank_line,
  [128854] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6084), 2,
      sym_comment,
      sym__blank_line,
  [128862] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6086), 2,
      sym_comment,
      sym__blank_line,
  [128870] = 2,
    ACTIONS(6088), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128878] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6090), 2,
      sym_comment,
      sym__blank_line,
  [128886] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(1947), 2,
      sym_comment,
      sym__blank_line,
  [128894] = 2,
    ACTIONS(6080), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128902] = 2,
    ACTIONS(2963), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128910] = 2,
    ACTIONS(1919), 1,
      sym_path,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128918] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6092), 1,
      sym_comment,
    ACTIONS(6094), 1,
      sym__blank_line,
  [128928] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6096), 1,
      sym_comment,
    ACTIONS(6098), 1,
      sym__blank_line,
  [128938] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6100), 2,
      sym_comment,
      sym__blank_line,
  [128946] = 2,
    ACTIONS(6102), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128954] = 2,
    ACTIONS(6104), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128962] = 2,
    ACTIONS(6106), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128970] = 2,
    ACTIONS(6108), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128978] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6110), 2,
      sym_comment,
      sym__blank_line,
  [128986] = 2,
    ACTIONS(6112), 1,
      sym__decimal_number,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [128994] = 2,
    ACTIONS(6114), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129002] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6116), 1,
      sym_comment,
    ACTIONS(6118), 1,
      sym__blank_line,
  [129012] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6120), 1,
      sym_comment,
    ACTIONS(6122), 1,
      sym__blank_line,
  [129022] = 2,
    ACTIONS(2961), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129030] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6124), 2,
      sym_comment,
      sym__blank_line,
  [129038] = 2,
    ACTIONS(6126), 1,
      sym_hll_number_literal,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129046] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6128), 1,
      sym_comment,
    ACTIONS(6130), 1,
      sym__blank_line,
  [129056] = 2,
    ACTIONS(6132), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129064] = 2,
    ACTIONS(2933), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129072] = 2,
    ACTIONS(6134), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129080] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6136), 2,
      sym_comment,
      sym__blank_line,
  [129088] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6138), 1,
      sym_comment,
    ACTIONS(6140), 1,
      sym__blank_line,
  [129098] = 2,
    ACTIONS(6142), 1,
      sym__decimal_number,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129106] = 2,
    ACTIONS(6144), 1,
      sym_hll_number_literal,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129114] = 2,
    ACTIONS(1923), 1,
      sym_path,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129122] = 2,
    ACTIONS(6146), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129130] = 2,
    ACTIONS(6148), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129138] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6150), 1,
      sym_comment,
    ACTIONS(6152), 1,
      sym__blank_line,
  [129148] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6154), 1,
      sym_comment,
    ACTIONS(6156), 1,
      sym__blank_line,
  [129158] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6158), 1,
      sym_comment,
    ACTIONS(6160), 1,
      sym__blank_line,
  [129168] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6162), 2,
      sym_comment,
      sym__blank_line,
  [129176] = 2,
    ACTIONS(6164), 1,
      sym__decimal_number,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129184] = 2,
    ACTIONS(6166), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129192] = 2,
    ACTIONS(6148), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129200] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6168), 2,
      sym_comment,
      sym__blank_line,
  [129208] = 2,
    ACTIONS(6170), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129216] = 2,
    ACTIONS(6172), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129224] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2611), 2,
      sym_comment,
      sym__blank_line,
  [129232] = 2,
    ACTIONS(6174), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129240] = 3,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6176), 1,
      sym_comment,
    ACTIONS(6178), 1,
      sym__blank_line,
  [129250] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2907), 2,
      sym_comment,
      sym__blank_line,
  [129258] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6124), 2,
      sym_comment,
      sym__blank_line,
  [129266] = 2,
    ACTIONS(6180), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129274] = 2,
    ACTIONS(6182), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129282] = 2,
    ACTIONS(6184), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129290] = 2,
    ACTIONS(2965), 1,
      sym_path,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129298] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6186), 2,
      sym_comment,
      sym__blank_line,
  [129306] = 2,
    ACTIONS(6188), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129314] = 2,
    ACTIONS(6190), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129322] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6192), 2,
      sym_comment,
      sym__blank_line,
  [129330] = 2,
    ACTIONS(6194), 1,
      sym__decimal_number,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129338] = 2,
    ACTIONS(6196), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129346] = 2,
    ACTIONS(1925), 1,
      sym_path,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129354] = 2,
    ACTIONS(6180), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129362] = 2,
    ACTIONS(2390), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129370] = 2,
    ACTIONS(6198), 1,
      sym_hll_number_literal,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129378] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6200), 2,
      sym_comment,
      sym__blank_line,
  [129386] = 2,
    ACTIONS(6202), 1,
      sym_hll_number_literal,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129394] = 2,
    ACTIONS(6204), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129402] = 2,
    ACTIONS(6206), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129410] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6208), 2,
      sym_comment,
      sym__blank_line,
  [129418] = 2,
    ACTIONS(6210), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129426] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6086), 2,
      sym_comment,
      sym__blank_line,
  [129434] = 2,
    ACTIONS(6212), 1,
      sym_hll_number_literal,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129442] = 2,
    ACTIONS(6214), 1,
      ts_builtin_sym_end,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129450] = 2,
    ACTIONS(6216), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129458] = 2,
    ACTIONS(6218), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129466] = 2,
    ACTIONS(6220), 1,
      sym_hll_number_literal,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129474] = 2,
    ACTIONS(2319), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129482] = 2,
    ACTIONS(6222), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129490] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6224), 2,
      sym_comment,
      sym__blank_line,
  [129498] = 2,
    ACTIONS(6134), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129506] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(2615), 2,
      sym_comment,
      sym__blank_line,
  [129514] = 2,
    ACTIONS(6172), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129522] = 2,
    ACTIONS(6226), 1,
      sym__blank,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129530] = 2,
    ACTIONS(6228), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129538] = 2,
    ACTIONS(5469), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129546] = 2,
    ACTIONS(6230), 1,
      anon_sym_LPAREN,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129554] = 2,
    ACTIONS(6232), 1,
      anon_sym_SQUOTE,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129562] = 2,
    ACTIONS(6234), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129570] = 2,
    ACTIONS(3), 1,
      sym__line_continuation,
    ACTIONS(6236), 2,
      sym_comment,
      sym__blank_line,
  [129578] = 2,
    ACTIONS(6238), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
  [129586] = 2,
    ACTIONS(6240), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_comment,
      sym__line_continuation,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(161)] = 0,
  [SMALL_STATE(162)] = 147,
  [SMALL_STATE(163)] = 254,
  [SMALL_STATE(164)] = 361,
  [SMALL_STATE(165)] = 468,
  [SMALL_STATE(166)] = 575,
  [SMALL_STATE(167)] = 722,
  [SMALL_STATE(168)] = 869,
  [SMALL_STATE(169)] = 1013,
  [SMALL_STATE(170)] = 1157,
  [SMALL_STATE(171)] = 1301,
  [SMALL_STATE(172)] = 1445,
  [SMALL_STATE(173)] = 1589,
  [SMALL_STATE(174)] = 1733,
  [SMALL_STATE(175)] = 1877,
  [SMALL_STATE(176)] = 2021,
  [SMALL_STATE(177)] = 2165,
  [SMALL_STATE(178)] = 2309,
  [SMALL_STATE(179)] = 2453,
  [SMALL_STATE(180)] = 2597,
  [SMALL_STATE(181)] = 2741,
  [SMALL_STATE(182)] = 2885,
  [SMALL_STATE(183)] = 3029,
  [SMALL_STATE(184)] = 3173,
  [SMALL_STATE(185)] = 3317,
  [SMALL_STATE(186)] = 3461,
  [SMALL_STATE(187)] = 3605,
  [SMALL_STATE(188)] = 3749,
  [SMALL_STATE(189)] = 3893,
  [SMALL_STATE(190)] = 4037,
  [SMALL_STATE(191)] = 4181,
  [SMALL_STATE(192)] = 4325,
  [SMALL_STATE(193)] = 4469,
  [SMALL_STATE(194)] = 4613,
  [SMALL_STATE(195)] = 4757,
  [SMALL_STATE(196)] = 4901,
  [SMALL_STATE(197)] = 5045,
  [SMALL_STATE(198)] = 5189,
  [SMALL_STATE(199)] = 5333,
  [SMALL_STATE(200)] = 5477,
  [SMALL_STATE(201)] = 5621,
  [SMALL_STATE(202)] = 5765,
  [SMALL_STATE(203)] = 5909,
  [SMALL_STATE(204)] = 6053,
  [SMALL_STATE(205)] = 6193,
  [SMALL_STATE(206)] = 6337,
  [SMALL_STATE(207)] = 6477,
  [SMALL_STATE(208)] = 6621,
  [SMALL_STATE(209)] = 6765,
  [SMALL_STATE(210)] = 6909,
  [SMALL_STATE(211)] = 7053,
  [SMALL_STATE(212)] = 7197,
  [SMALL_STATE(213)] = 7341,
  [SMALL_STATE(214)] = 7485,
  [SMALL_STATE(215)] = 7625,
  [SMALL_STATE(216)] = 7769,
  [SMALL_STATE(217)] = 7913,
  [SMALL_STATE(218)] = 8057,
  [SMALL_STATE(219)] = 8201,
  [SMALL_STATE(220)] = 8341,
  [SMALL_STATE(221)] = 8485,
  [SMALL_STATE(222)] = 8625,
  [SMALL_STATE(223)] = 8765,
  [SMALL_STATE(224)] = 8905,
  [SMALL_STATE(225)] = 9045,
  [SMALL_STATE(226)] = 9189,
  [SMALL_STATE(227)] = 9333,
  [SMALL_STATE(228)] = 9473,
  [SMALL_STATE(229)] = 9613,
  [SMALL_STATE(230)] = 9757,
  [SMALL_STATE(231)] = 9901,
  [SMALL_STATE(232)] = 10045,
  [SMALL_STATE(233)] = 10189,
  [SMALL_STATE(234)] = 10333,
  [SMALL_STATE(235)] = 10477,
  [SMALL_STATE(236)] = 10621,
  [SMALL_STATE(237)] = 10765,
  [SMALL_STATE(238)] = 10909,
  [SMALL_STATE(239)] = 11049,
  [SMALL_STATE(240)] = 11189,
  [SMALL_STATE(241)] = 11333,
  [SMALL_STATE(242)] = 11477,
  [SMALL_STATE(243)] = 11621,
  [SMALL_STATE(244)] = 11765,
  [SMALL_STATE(245)] = 11909,
  [SMALL_STATE(246)] = 12053,
  [SMALL_STATE(247)] = 12197,
  [SMALL_STATE(248)] = 12341,
  [SMALL_STATE(249)] = 12485,
  [SMALL_STATE(250)] = 12629,
  [SMALL_STATE(251)] = 12773,
  [SMALL_STATE(252)] = 12917,
  [SMALL_STATE(253)] = 13061,
  [SMALL_STATE(254)] = 13205,
  [SMALL_STATE(255)] = 13349,
  [SMALL_STATE(256)] = 13493,
  [SMALL_STATE(257)] = 13637,
  [SMALL_STATE(258)] = 13781,
  [SMALL_STATE(259)] = 13925,
  [SMALL_STATE(260)] = 14069,
  [SMALL_STATE(261)] = 14213,
  [SMALL_STATE(262)] = 14357,
  [SMALL_STATE(263)] = 14501,
  [SMALL_STATE(264)] = 14645,
  [SMALL_STATE(265)] = 14789,
  [SMALL_STATE(266)] = 14929,
  [SMALL_STATE(267)] = 15079,
  [SMALL_STATE(268)] = 15229,
  [SMALL_STATE(269)] = 15375,
  [SMALL_STATE(270)] = 15525,
  [SMALL_STATE(271)] = 15675,
  [SMALL_STATE(272)] = 15825,
  [SMALL_STATE(273)] = 15975,
  [SMALL_STATE(274)] = 16125,
  [SMALL_STATE(275)] = 16275,
  [SMALL_STATE(276)] = 16425,
  [SMALL_STATE(277)] = 16575,
  [SMALL_STATE(278)] = 16725,
  [SMALL_STATE(279)] = 16875,
  [SMALL_STATE(280)] = 17021,
  [SMALL_STATE(281)] = 17171,
  [SMALL_STATE(282)] = 17321,
  [SMALL_STATE(283)] = 17471,
  [SMALL_STATE(284)] = 17621,
  [SMALL_STATE(285)] = 17771,
  [SMALL_STATE(286)] = 17917,
  [SMALL_STATE(287)] = 18067,
  [SMALL_STATE(288)] = 18217,
  [SMALL_STATE(289)] = 18367,
  [SMALL_STATE(290)] = 18517,
  [SMALL_STATE(291)] = 18667,
  [SMALL_STATE(292)] = 18817,
  [SMALL_STATE(293)] = 18967,
  [SMALL_STATE(294)] = 19117,
  [SMALL_STATE(295)] = 19267,
  [SMALL_STATE(296)] = 19417,
  [SMALL_STATE(297)] = 19567,
  [SMALL_STATE(298)] = 19717,
  [SMALL_STATE(299)] = 19867,
  [SMALL_STATE(300)] = 20017,
  [SMALL_STATE(301)] = 20167,
  [SMALL_STATE(302)] = 20317,
  [SMALL_STATE(303)] = 20463,
  [SMALL_STATE(304)] = 20609,
  [SMALL_STATE(305)] = 20755,
  [SMALL_STATE(306)] = 20905,
  [SMALL_STATE(307)] = 21051,
  [SMALL_STATE(308)] = 21197,
  [SMALL_STATE(309)] = 21343,
  [SMALL_STATE(310)] = 21493,
  [SMALL_STATE(311)] = 21639,
  [SMALL_STATE(312)] = 21785,
  [SMALL_STATE(313)] = 21935,
  [SMALL_STATE(314)] = 22085,
  [SMALL_STATE(315)] = 22235,
  [SMALL_STATE(316)] = 22385,
  [SMALL_STATE(317)] = 22535,
  [SMALL_STATE(318)] = 22685,
  [SMALL_STATE(319)] = 22835,
  [SMALL_STATE(320)] = 22985,
  [SMALL_STATE(321)] = 23135,
  [SMALL_STATE(322)] = 23285,
  [SMALL_STATE(323)] = 23435,
  [SMALL_STATE(324)] = 23585,
  [SMALL_STATE(325)] = 23735,
  [SMALL_STATE(326)] = 23885,
  [SMALL_STATE(327)] = 24035,
  [SMALL_STATE(328)] = 24185,
  [SMALL_STATE(329)] = 24335,
  [SMALL_STATE(330)] = 24485,
  [SMALL_STATE(331)] = 24635,
  [SMALL_STATE(332)] = 24785,
  [SMALL_STATE(333)] = 24935,
  [SMALL_STATE(334)] = 25085,
  [SMALL_STATE(335)] = 25235,
  [SMALL_STATE(336)] = 25385,
  [SMALL_STATE(337)] = 25535,
  [SMALL_STATE(338)] = 25680,
  [SMALL_STATE(339)] = 25825,
  [SMALL_STATE(340)] = 25970,
  [SMALL_STATE(341)] = 26115,
  [SMALL_STATE(342)] = 26260,
  [SMALL_STATE(343)] = 26405,
  [SMALL_STATE(344)] = 26550,
  [SMALL_STATE(345)] = 26695,
  [SMALL_STATE(346)] = 26840,
  [SMALL_STATE(347)] = 26985,
  [SMALL_STATE(348)] = 27130,
  [SMALL_STATE(349)] = 27275,
  [SMALL_STATE(350)] = 27415,
  [SMALL_STATE(351)] = 27555,
  [SMALL_STATE(352)] = 27695,
  [SMALL_STATE(353)] = 27835,
  [SMALL_STATE(354)] = 27975,
  [SMALL_STATE(355)] = 28115,
  [SMALL_STATE(356)] = 28255,
  [SMALL_STATE(357)] = 28395,
  [SMALL_STATE(358)] = 28535,
  [SMALL_STATE(359)] = 28675,
  [SMALL_STATE(360)] = 28815,
  [SMALL_STATE(361)] = 28955,
  [SMALL_STATE(362)] = 29092,
  [SMALL_STATE(363)] = 29229,
  [SMALL_STATE(364)] = 29366,
  [SMALL_STATE(365)] = 29503,
  [SMALL_STATE(366)] = 29640,
  [SMALL_STATE(367)] = 29777,
  [SMALL_STATE(368)] = 29914,
  [SMALL_STATE(369)] = 30051,
  [SMALL_STATE(370)] = 30188,
  [SMALL_STATE(371)] = 30325,
  [SMALL_STATE(372)] = 30462,
  [SMALL_STATE(373)] = 30599,
  [SMALL_STATE(374)] = 30733,
  [SMALL_STATE(375)] = 30867,
  [SMALL_STATE(376)] = 31001,
  [SMALL_STATE(377)] = 31135,
  [SMALL_STATE(378)] = 31269,
  [SMALL_STATE(379)] = 31403,
  [SMALL_STATE(380)] = 31504,
  [SMALL_STATE(381)] = 31605,
  [SMALL_STATE(382)] = 31706,
  [SMALL_STATE(383)] = 31807,
  [SMALL_STATE(384)] = 31908,
  [SMALL_STATE(385)] = 32009,
  [SMALL_STATE(386)] = 32110,
  [SMALL_STATE(387)] = 32211,
  [SMALL_STATE(388)] = 32312,
  [SMALL_STATE(389)] = 32413,
  [SMALL_STATE(390)] = 32514,
  [SMALL_STATE(391)] = 32617,
  [SMALL_STATE(392)] = 32718,
  [SMALL_STATE(393)] = 32818,
  [SMALL_STATE(394)] = 32918,
  [SMALL_STATE(395)] = 33018,
  [SMALL_STATE(396)] = 33118,
  [SMALL_STATE(397)] = 33218,
  [SMALL_STATE(398)] = 33318,
  [SMALL_STATE(399)] = 33418,
  [SMALL_STATE(400)] = 33518,
  [SMALL_STATE(401)] = 33618,
  [SMALL_STATE(402)] = 33718,
  [SMALL_STATE(403)] = 33818,
  [SMALL_STATE(404)] = 33918,
  [SMALL_STATE(405)] = 34018,
  [SMALL_STATE(406)] = 34075,
  [SMALL_STATE(407)] = 34142,
  [SMALL_STATE(408)] = 34199,
  [SMALL_STATE(409)] = 34262,
  [SMALL_STATE(410)] = 34318,
  [SMALL_STATE(411)] = 34374,
  [SMALL_STATE(412)] = 34438,
  [SMALL_STATE(413)] = 34494,
  [SMALL_STATE(414)] = 34588,
  [SMALL_STATE(415)] = 34644,
  [SMALL_STATE(416)] = 34700,
  [SMALL_STATE(417)] = 34756,
  [SMALL_STATE(418)] = 34812,
  [SMALL_STATE(419)] = 34870,
  [SMALL_STATE(420)] = 34936,
  [SMALL_STATE(421)] = 35010,
  [SMALL_STATE(422)] = 35066,
  [SMALL_STATE(423)] = 35160,
  [SMALL_STATE(424)] = 35216,
  [SMALL_STATE(425)] = 35280,
  [SMALL_STATE(426)] = 35358,
  [SMALL_STATE(427)] = 35422,
  [SMALL_STATE(428)] = 35484,
  [SMALL_STATE(429)] = 35540,
  [SMALL_STATE(430)] = 35596,
  [SMALL_STATE(431)] = 35656,
  [SMALL_STATE(432)] = 35716,
  [SMALL_STATE(433)] = 35772,
  [SMALL_STATE(434)] = 35866,
  [SMALL_STATE(435)] = 35940,
  [SMALL_STATE(436)] = 36031,
  [SMALL_STATE(437)] = 36122,
  [SMALL_STATE(438)] = 36213,
  [SMALL_STATE(439)] = 36304,
  [SMALL_STATE(440)] = 36363,
  [SMALL_STATE(441)] = 36454,
  [SMALL_STATE(442)] = 36515,
  [SMALL_STATE(443)] = 36606,
  [SMALL_STATE(444)] = 36665,
  [SMALL_STATE(445)] = 36756,
  [SMALL_STATE(446)] = 36847,
  [SMALL_STATE(447)] = 36938,
  [SMALL_STATE(448)] = 37029,
  [SMALL_STATE(449)] = 37120,
  [SMALL_STATE(450)] = 37211,
  [SMALL_STATE(451)] = 37302,
  [SMALL_STATE(452)] = 37373,
  [SMALL_STATE(453)] = 37432,
  [SMALL_STATE(454)] = 37515,
  [SMALL_STATE(455)] = 37600,
  [SMALL_STATE(456)] = 37667,
  [SMALL_STATE(457)] = 37748,
  [SMALL_STATE(458)] = 37827,
  [SMALL_STATE(459)] = 37900,
  [SMALL_STATE(460)] = 37991,
  [SMALL_STATE(461)] = 38082,
  [SMALL_STATE(462)] = 38165,
  [SMALL_STATE(463)] = 38242,
  [SMALL_STATE(464)] = 38333,
  [SMALL_STATE(465)] = 38424,
  [SMALL_STATE(466)] = 38489,
  [SMALL_STATE(467)] = 38552,
  [SMALL_STATE(468)] = 38643,
  [SMALL_STATE(469)] = 38734,
  [SMALL_STATE(470)] = 38825,
  [SMALL_STATE(471)] = 38916,
  [SMALL_STATE(472)] = 39007,
  [SMALL_STATE(473)] = 39098,
  [SMALL_STATE(474)] = 39189,
  [SMALL_STATE(475)] = 39280,
  [SMALL_STATE(476)] = 39371,
  [SMALL_STATE(477)] = 39432,
  [SMALL_STATE(478)] = 39523,
  [SMALL_STATE(479)] = 39614,
  [SMALL_STATE(480)] = 39705,
  [SMALL_STATE(481)] = 39796,
  [SMALL_STATE(482)] = 39887,
  [SMALL_STATE(483)] = 39978,
  [SMALL_STATE(484)] = 40069,
  [SMALL_STATE(485)] = 40160,
  [SMALL_STATE(486)] = 40251,
  [SMALL_STATE(487)] = 40342,
  [SMALL_STATE(488)] = 40407,
  [SMALL_STATE(489)] = 40495,
  [SMALL_STATE(490)] = 40583,
  [SMALL_STATE(491)] = 40671,
  [SMALL_STATE(492)] = 40759,
  [SMALL_STATE(493)] = 40815,
  [SMALL_STATE(494)] = 40903,
  [SMALL_STATE(495)] = 40991,
  [SMALL_STATE(496)] = 41079,
  [SMALL_STATE(497)] = 41167,
  [SMALL_STATE(498)] = 41255,
  [SMALL_STATE(499)] = 41343,
  [SMALL_STATE(500)] = 41397,
  [SMALL_STATE(501)] = 41485,
  [SMALL_STATE(502)] = 41573,
  [SMALL_STATE(503)] = 41661,
  [SMALL_STATE(504)] = 41715,
  [SMALL_STATE(505)] = 41803,
  [SMALL_STATE(506)] = 41891,
  [SMALL_STATE(507)] = 41979,
  [SMALL_STATE(508)] = 42067,
  [SMALL_STATE(509)] = 42155,
  [SMALL_STATE(510)] = 42243,
  [SMALL_STATE(511)] = 42331,
  [SMALL_STATE(512)] = 42419,
  [SMALL_STATE(513)] = 42507,
  [SMALL_STATE(514)] = 42595,
  [SMALL_STATE(515)] = 42683,
  [SMALL_STATE(516)] = 42771,
  [SMALL_STATE(517)] = 42827,
  [SMALL_STATE(518)] = 42881,
  [SMALL_STATE(519)] = 42937,
  [SMALL_STATE(520)] = 43025,
  [SMALL_STATE(521)] = 43113,
  [SMALL_STATE(522)] = 43201,
  [SMALL_STATE(523)] = 43289,
  [SMALL_STATE(524)] = 43377,
  [SMALL_STATE(525)] = 43465,
  [SMALL_STATE(526)] = 43553,
  [SMALL_STATE(527)] = 43641,
  [SMALL_STATE(528)] = 43729,
  [SMALL_STATE(529)] = 43817,
  [SMALL_STATE(530)] = 43905,
  [SMALL_STATE(531)] = 43993,
  [SMALL_STATE(532)] = 44081,
  [SMALL_STATE(533)] = 44169,
  [SMALL_STATE(534)] = 44257,
  [SMALL_STATE(535)] = 44345,
  [SMALL_STATE(536)] = 44433,
  [SMALL_STATE(537)] = 44521,
  [SMALL_STATE(538)] = 44609,
  [SMALL_STATE(539)] = 44697,
  [SMALL_STATE(540)] = 44785,
  [SMALL_STATE(541)] = 44873,
  [SMALL_STATE(542)] = 44961,
  [SMALL_STATE(543)] = 45049,
  [SMALL_STATE(544)] = 45137,
  [SMALL_STATE(545)] = 45225,
  [SMALL_STATE(546)] = 45313,
  [SMALL_STATE(547)] = 45401,
  [SMALL_STATE(548)] = 45489,
  [SMALL_STATE(549)] = 45577,
  [SMALL_STATE(550)] = 45665,
  [SMALL_STATE(551)] = 45753,
  [SMALL_STATE(552)] = 45841,
  [SMALL_STATE(553)] = 45926,
  [SMALL_STATE(554)] = 45979,
  [SMALL_STATE(555)] = 46064,
  [SMALL_STATE(556)] = 46117,
  [SMALL_STATE(557)] = 46202,
  [SMALL_STATE(558)] = 46287,
  [SMALL_STATE(559)] = 46372,
  [SMALL_STATE(560)] = 46425,
  [SMALL_STATE(561)] = 46478,
  [SMALL_STATE(562)] = 46563,
  [SMALL_STATE(563)] = 46616,
  [SMALL_STATE(564)] = 46701,
  [SMALL_STATE(565)] = 46754,
  [SMALL_STATE(566)] = 46807,
  [SMALL_STATE(567)] = 46892,
  [SMALL_STATE(568)] = 46977,
  [SMALL_STATE(569)] = 47030,
  [SMALL_STATE(570)] = 47083,
  [SMALL_STATE(571)] = 47168,
  [SMALL_STATE(572)] = 47253,
  [SMALL_STATE(573)] = 47338,
  [SMALL_STATE(574)] = 47391,
  [SMALL_STATE(575)] = 47476,
  [SMALL_STATE(576)] = 47529,
  [SMALL_STATE(577)] = 47614,
  [SMALL_STATE(578)] = 47699,
  [SMALL_STATE(579)] = 47784,
  [SMALL_STATE(580)] = 47869,
  [SMALL_STATE(581)] = 47954,
  [SMALL_STATE(582)] = 48007,
  [SMALL_STATE(583)] = 48080,
  [SMALL_STATE(584)] = 48133,
  [SMALL_STATE(585)] = 48218,
  [SMALL_STATE(586)] = 48303,
  [SMALL_STATE(587)] = 48388,
  [SMALL_STATE(588)] = 48470,
  [SMALL_STATE(589)] = 48552,
  [SMALL_STATE(590)] = 48634,
  [SMALL_STATE(591)] = 48716,
  [SMALL_STATE(592)] = 48798,
  [SMALL_STATE(593)] = 48880,
  [SMALL_STATE(594)] = 48962,
  [SMALL_STATE(595)] = 49044,
  [SMALL_STATE(596)] = 49126,
  [SMALL_STATE(597)] = 49208,
  [SMALL_STATE(598)] = 49290,
  [SMALL_STATE(599)] = 49372,
  [SMALL_STATE(600)] = 49454,
  [SMALL_STATE(601)] = 49536,
  [SMALL_STATE(602)] = 49618,
  [SMALL_STATE(603)] = 49700,
  [SMALL_STATE(604)] = 49782,
  [SMALL_STATE(605)] = 49864,
  [SMALL_STATE(606)] = 49946,
  [SMALL_STATE(607)] = 50028,
  [SMALL_STATE(608)] = 50110,
  [SMALL_STATE(609)] = 50192,
  [SMALL_STATE(610)] = 50274,
  [SMALL_STATE(611)] = 50356,
  [SMALL_STATE(612)] = 50438,
  [SMALL_STATE(613)] = 50500,
  [SMALL_STATE(614)] = 50564,
  [SMALL_STATE(615)] = 50630,
  [SMALL_STATE(616)] = 50712,
  [SMALL_STATE(617)] = 50794,
  [SMALL_STATE(618)] = 50876,
  [SMALL_STATE(619)] = 50958,
  [SMALL_STATE(620)] = 51040,
  [SMALL_STATE(621)] = 51122,
  [SMALL_STATE(622)] = 51204,
  [SMALL_STATE(623)] = 51286,
  [SMALL_STATE(624)] = 51368,
  [SMALL_STATE(625)] = 51450,
  [SMALL_STATE(626)] = 51532,
  [SMALL_STATE(627)] = 51596,
  [SMALL_STATE(628)] = 51678,
  [SMALL_STATE(629)] = 51742,
  [SMALL_STATE(630)] = 51824,
  [SMALL_STATE(631)] = 51906,
  [SMALL_STATE(632)] = 51988,
  [SMALL_STATE(633)] = 52052,
  [SMALL_STATE(634)] = 52116,
  [SMALL_STATE(635)] = 52198,
  [SMALL_STATE(636)] = 52262,
  [SMALL_STATE(637)] = 52344,
  [SMALL_STATE(638)] = 52426,
  [SMALL_STATE(639)] = 52508,
  [SMALL_STATE(640)] = 52590,
  [SMALL_STATE(641)] = 52672,
  [SMALL_STATE(642)] = 52754,
  [SMALL_STATE(643)] = 52836,
  [SMALL_STATE(644)] = 52918,
  [SMALL_STATE(645)] = 53000,
  [SMALL_STATE(646)] = 53082,
  [SMALL_STATE(647)] = 53164,
  [SMALL_STATE(648)] = 53246,
  [SMALL_STATE(649)] = 53328,
  [SMALL_STATE(650)] = 53410,
  [SMALL_STATE(651)] = 53492,
  [SMALL_STATE(652)] = 53574,
  [SMALL_STATE(653)] = 53656,
  [SMALL_STATE(654)] = 53738,
  [SMALL_STATE(655)] = 53820,
  [SMALL_STATE(656)] = 53902,
  [SMALL_STATE(657)] = 53984,
  [SMALL_STATE(658)] = 54066,
  [SMALL_STATE(659)] = 54148,
  [SMALL_STATE(660)] = 54230,
  [SMALL_STATE(661)] = 54312,
  [SMALL_STATE(662)] = 54394,
  [SMALL_STATE(663)] = 54476,
  [SMALL_STATE(664)] = 54544,
  [SMALL_STATE(665)] = 54628,
  [SMALL_STATE(666)] = 54694,
  [SMALL_STATE(667)] = 54774,
  [SMALL_STATE(668)] = 54852,
  [SMALL_STATE(669)] = 54926,
  [SMALL_STATE(670)] = 54996,
  [SMALL_STATE(671)] = 55078,
  [SMALL_STATE(672)] = 55154,
  [SMALL_STATE(673)] = 55236,
  [SMALL_STATE(674)] = 55324,
  [SMALL_STATE(675)] = 55406,
  [SMALL_STATE(676)] = 55488,
  [SMALL_STATE(677)] = 55576,
  [SMALL_STATE(678)] = 55658,
  [SMALL_STATE(679)] = 55746,
  [SMALL_STATE(680)] = 55834,
  [SMALL_STATE(681)] = 55916,
  [SMALL_STATE(682)] = 56004,
  [SMALL_STATE(683)] = 56092,
  [SMALL_STATE(684)] = 56174,
  [SMALL_STATE(685)] = 56256,
  [SMALL_STATE(686)] = 56338,
  [SMALL_STATE(687)] = 56420,
  [SMALL_STATE(688)] = 56502,
  [SMALL_STATE(689)] = 56584,
  [SMALL_STATE(690)] = 56666,
  [SMALL_STATE(691)] = 56748,
  [SMALL_STATE(692)] = 56830,
  [SMALL_STATE(693)] = 56912,
  [SMALL_STATE(694)] = 56994,
  [SMALL_STATE(695)] = 57076,
  [SMALL_STATE(696)] = 57158,
  [SMALL_STATE(697)] = 57240,
  [SMALL_STATE(698)] = 57322,
  [SMALL_STATE(699)] = 57404,
  [SMALL_STATE(700)] = 57486,
  [SMALL_STATE(701)] = 57568,
  [SMALL_STATE(702)] = 57650,
  [SMALL_STATE(703)] = 57732,
  [SMALL_STATE(704)] = 57814,
  [SMALL_STATE(705)] = 57896,
  [SMALL_STATE(706)] = 57978,
  [SMALL_STATE(707)] = 58060,
  [SMALL_STATE(708)] = 58142,
  [SMALL_STATE(709)] = 58224,
  [SMALL_STATE(710)] = 58306,
  [SMALL_STATE(711)] = 58388,
  [SMALL_STATE(712)] = 58470,
  [SMALL_STATE(713)] = 58552,
  [SMALL_STATE(714)] = 58634,
  [SMALL_STATE(715)] = 58716,
  [SMALL_STATE(716)] = 58798,
  [SMALL_STATE(717)] = 58850,
  [SMALL_STATE(718)] = 58932,
  [SMALL_STATE(719)] = 59014,
  [SMALL_STATE(720)] = 59096,
  [SMALL_STATE(721)] = 59178,
  [SMALL_STATE(722)] = 59260,
  [SMALL_STATE(723)] = 59342,
  [SMALL_STATE(724)] = 59424,
  [SMALL_STATE(725)] = 59506,
  [SMALL_STATE(726)] = 59588,
  [SMALL_STATE(727)] = 59670,
  [SMALL_STATE(728)] = 59722,
  [SMALL_STATE(729)] = 59804,
  [SMALL_STATE(730)] = 59886,
  [SMALL_STATE(731)] = 59968,
  [SMALL_STATE(732)] = 60050,
  [SMALL_STATE(733)] = 60132,
  [SMALL_STATE(734)] = 60214,
  [SMALL_STATE(735)] = 60296,
  [SMALL_STATE(736)] = 60378,
  [SMALL_STATE(737)] = 60460,
  [SMALL_STATE(738)] = 60542,
  [SMALL_STATE(739)] = 60624,
  [SMALL_STATE(740)] = 60706,
  [SMALL_STATE(741)] = 60788,
  [SMALL_STATE(742)] = 60870,
  [SMALL_STATE(743)] = 60952,
  [SMALL_STATE(744)] = 61034,
  [SMALL_STATE(745)] = 61116,
  [SMALL_STATE(746)] = 61198,
  [SMALL_STATE(747)] = 61280,
  [SMALL_STATE(748)] = 61362,
  [SMALL_STATE(749)] = 61444,
  [SMALL_STATE(750)] = 61526,
  [SMALL_STATE(751)] = 61608,
  [SMALL_STATE(752)] = 61690,
  [SMALL_STATE(753)] = 61772,
  [SMALL_STATE(754)] = 61854,
  [SMALL_STATE(755)] = 61936,
  [SMALL_STATE(756)] = 62018,
  [SMALL_STATE(757)] = 62100,
  [SMALL_STATE(758)] = 62182,
  [SMALL_STATE(759)] = 62264,
  [SMALL_STATE(760)] = 62346,
  [SMALL_STATE(761)] = 62428,
  [SMALL_STATE(762)] = 62510,
  [SMALL_STATE(763)] = 62592,
  [SMALL_STATE(764)] = 62674,
  [SMALL_STATE(765)] = 62756,
  [SMALL_STATE(766)] = 62838,
  [SMALL_STATE(767)] = 62920,
  [SMALL_STATE(768)] = 63002,
  [SMALL_STATE(769)] = 63084,
  [SMALL_STATE(770)] = 63166,
  [SMALL_STATE(771)] = 63248,
  [SMALL_STATE(772)] = 63330,
  [SMALL_STATE(773)] = 63412,
  [SMALL_STATE(774)] = 63494,
  [SMALL_STATE(775)] = 63576,
  [SMALL_STATE(776)] = 63658,
  [SMALL_STATE(777)] = 63740,
  [SMALL_STATE(778)] = 63822,
  [SMALL_STATE(779)] = 63904,
  [SMALL_STATE(780)] = 63986,
  [SMALL_STATE(781)] = 64068,
  [SMALL_STATE(782)] = 64150,
  [SMALL_STATE(783)] = 64232,
  [SMALL_STATE(784)] = 64314,
  [SMALL_STATE(785)] = 64396,
  [SMALL_STATE(786)] = 64478,
  [SMALL_STATE(787)] = 64560,
  [SMALL_STATE(788)] = 64642,
  [SMALL_STATE(789)] = 64724,
  [SMALL_STATE(790)] = 64806,
  [SMALL_STATE(791)] = 64888,
  [SMALL_STATE(792)] = 64970,
  [SMALL_STATE(793)] = 65052,
  [SMALL_STATE(794)] = 65134,
  [SMALL_STATE(795)] = 65216,
  [SMALL_STATE(796)] = 65298,
  [SMALL_STATE(797)] = 65380,
  [SMALL_STATE(798)] = 65462,
  [SMALL_STATE(799)] = 65544,
  [SMALL_STATE(800)] = 65608,
  [SMALL_STATE(801)] = 65690,
  [SMALL_STATE(802)] = 65772,
  [SMALL_STATE(803)] = 65854,
  [SMALL_STATE(804)] = 65936,
  [SMALL_STATE(805)] = 66018,
  [SMALL_STATE(806)] = 66100,
  [SMALL_STATE(807)] = 66182,
  [SMALL_STATE(808)] = 66264,
  [SMALL_STATE(809)] = 66346,
  [SMALL_STATE(810)] = 66428,
  [SMALL_STATE(811)] = 66482,
  [SMALL_STATE(812)] = 66533,
  [SMALL_STATE(813)] = 66584,
  [SMALL_STATE(814)] = 66635,
  [SMALL_STATE(815)] = 66686,
  [SMALL_STATE(816)] = 66737,
  [SMALL_STATE(817)] = 66788,
  [SMALL_STATE(818)] = 66839,
  [SMALL_STATE(819)] = 66894,
  [SMALL_STATE(820)] = 66949,
  [SMALL_STATE(821)] = 67000,
  [SMALL_STATE(822)] = 67051,
  [SMALL_STATE(823)] = 67102,
  [SMALL_STATE(824)] = 67153,
  [SMALL_STATE(825)] = 67204,
  [SMALL_STATE(826)] = 67255,
  [SMALL_STATE(827)] = 67306,
  [SMALL_STATE(828)] = 67357,
  [SMALL_STATE(829)] = 67408,
  [SMALL_STATE(830)] = 67459,
  [SMALL_STATE(831)] = 67510,
  [SMALL_STATE(832)] = 67561,
  [SMALL_STATE(833)] = 67612,
  [SMALL_STATE(834)] = 67696,
  [SMALL_STATE(835)] = 67746,
  [SMALL_STATE(836)] = 67830,
  [SMALL_STATE(837)] = 67914,
  [SMALL_STATE(838)] = 67980,
  [SMALL_STATE(839)] = 68040,
  [SMALL_STATE(840)] = 68116,
  [SMALL_STATE(841)] = 68176,
  [SMALL_STATE(842)] = 68238,
  [SMALL_STATE(843)] = 68292,
  [SMALL_STATE(844)] = 68344,
  [SMALL_STATE(845)] = 68404,
  [SMALL_STATE(846)] = 68464,
  [SMALL_STATE(847)] = 68524,
  [SMALL_STATE(848)] = 68584,
  [SMALL_STATE(849)] = 68642,
  [SMALL_STATE(850)] = 68702,
  [SMALL_STATE(851)] = 68752,
  [SMALL_STATE(852)] = 68814,
  [SMALL_STATE(853)] = 68892,
  [SMALL_STATE(854)] = 68964,
  [SMALL_STATE(855)] = 69038,
  [SMALL_STATE(856)] = 69122,
  [SMALL_STATE(857)] = 69178,
  [SMALL_STATE(858)] = 69234,
  [SMALL_STATE(859)] = 69318,
  [SMALL_STATE(860)] = 69382,
  [SMALL_STATE(861)] = 69462,
  [SMALL_STATE(862)] = 69524,
  [SMALL_STATE(863)] = 69594,
  [SMALL_STATE(864)] = 69678,
  [SMALL_STATE(865)] = 69737,
  [SMALL_STATE(866)] = 69786,
  [SMALL_STATE(867)] = 69847,
  [SMALL_STATE(868)] = 69904,
  [SMALL_STATE(869)] = 69963,
  [SMALL_STATE(870)] = 70038,
  [SMALL_STATE(871)] = 70099,
  [SMALL_STATE(872)] = 70172,
  [SMALL_STATE(873)] = 70221,
  [SMALL_STATE(874)] = 70290,
  [SMALL_STATE(875)] = 70339,
  [SMALL_STATE(876)] = 70388,
  [SMALL_STATE(877)] = 70453,
  [SMALL_STATE(878)] = 70502,
  [SMALL_STATE(879)] = 70551,
  [SMALL_STATE(880)] = 70610,
  [SMALL_STATE(881)] = 70659,
  [SMALL_STATE(882)] = 70712,
  [SMALL_STATE(883)] = 70765,
  [SMALL_STATE(884)] = 70824,
  [SMALL_STATE(885)] = 70873,
  [SMALL_STATE(886)] = 70950,
  [SMALL_STATE(887)] = 71021,
  [SMALL_STATE(888)] = 71070,
  [SMALL_STATE(889)] = 71149,
  [SMALL_STATE(890)] = 71232,
  [SMALL_STATE(891)] = 71287,
  [SMALL_STATE(892)] = 71342,
  [SMALL_STATE(893)] = 71401,
  [SMALL_STATE(894)] = 71460,
  [SMALL_STATE(895)] = 71509,
  [SMALL_STATE(896)] = 71592,
  [SMALL_STATE(897)] = 71641,
  [SMALL_STATE(898)] = 71700,
  [SMALL_STATE(899)] = 71749,
  [SMALL_STATE(900)] = 71832,
  [SMALL_STATE(901)] = 71881,
  [SMALL_STATE(902)] = 71930,
  [SMALL_STATE(903)] = 71979,
  [SMALL_STATE(904)] = 72028,
  [SMALL_STATE(905)] = 72077,
  [SMALL_STATE(906)] = 72126,
  [SMALL_STATE(907)] = 72209,
  [SMALL_STATE(908)] = 72292,
  [SMALL_STATE(909)] = 72375,
  [SMALL_STATE(910)] = 72426,
  [SMALL_STATE(911)] = 72475,
  [SMALL_STATE(912)] = 72524,
  [SMALL_STATE(913)] = 72573,
  [SMALL_STATE(914)] = 72626,
  [SMALL_STATE(915)] = 72679,
  [SMALL_STATE(916)] = 72742,
  [SMALL_STATE(917)] = 72791,
  [SMALL_STATE(918)] = 72839,
  [SMALL_STATE(919)] = 72887,
  [SMALL_STATE(920)] = 72955,
  [SMALL_STATE(921)] = 73003,
  [SMALL_STATE(922)] = 73067,
  [SMALL_STATE(923)] = 73115,
  [SMALL_STATE(924)] = 73197,
  [SMALL_STATE(925)] = 73279,
  [SMALL_STATE(926)] = 73327,
  [SMALL_STATE(927)] = 73375,
  [SMALL_STATE(928)] = 73423,
  [SMALL_STATE(929)] = 73475,
  [SMALL_STATE(930)] = 73523,
  [SMALL_STATE(931)] = 73575,
  [SMALL_STATE(932)] = 73623,
  [SMALL_STATE(933)] = 73671,
  [SMALL_STATE(934)] = 73753,
  [SMALL_STATE(935)] = 73811,
  [SMALL_STATE(936)] = 73889,
  [SMALL_STATE(937)] = 73937,
  [SMALL_STATE(938)] = 73999,
  [SMALL_STATE(939)] = 74053,
  [SMALL_STATE(940)] = 74107,
  [SMALL_STATE(941)] = 74183,
  [SMALL_STATE(942)] = 74231,
  [SMALL_STATE(943)] = 74283,
  [SMALL_STATE(944)] = 74331,
  [SMALL_STATE(945)] = 74383,
  [SMALL_STATE(946)] = 74443,
  [SMALL_STATE(947)] = 74501,
  [SMALL_STATE(948)] = 74571,
  [SMALL_STATE(949)] = 74619,
  [SMALL_STATE(950)] = 74667,
  [SMALL_STATE(951)] = 74715,
  [SMALL_STATE(952)] = 74763,
  [SMALL_STATE(953)] = 74845,
  [SMALL_STATE(954)] = 74903,
  [SMALL_STATE(955)] = 74959,
  [SMALL_STATE(956)] = 75017,
  [SMALL_STATE(957)] = 75065,
  [SMALL_STATE(958)] = 75125,
  [SMALL_STATE(959)] = 75183,
  [SMALL_STATE(960)] = 75231,
  [SMALL_STATE(961)] = 75289,
  [SMALL_STATE(962)] = 75347,
  [SMALL_STATE(963)] = 75395,
  [SMALL_STATE(964)] = 75477,
  [SMALL_STATE(965)] = 75559,
  [SMALL_STATE(966)] = 75633,
  [SMALL_STATE(967)] = 75705,
  [SMALL_STATE(968)] = 75753,
  [SMALL_STATE(969)] = 75801,
  [SMALL_STATE(970)] = 75851,
  [SMALL_STATE(971)] = 75898,
  [SMALL_STATE(972)] = 75945,
  [SMALL_STATE(973)] = 75992,
  [SMALL_STATE(974)] = 76039,
  [SMALL_STATE(975)] = 76090,
  [SMALL_STATE(976)] = 76137,
  [SMALL_STATE(977)] = 76184,
  [SMALL_STATE(978)] = 76235,
  [SMALL_STATE(979)] = 76282,
  [SMALL_STATE(980)] = 76329,
  [SMALL_STATE(981)] = 76376,
  [SMALL_STATE(982)] = 76423,
  [SMALL_STATE(983)] = 76470,
  [SMALL_STATE(984)] = 76517,
  [SMALL_STATE(985)] = 76564,
  [SMALL_STATE(986)] = 76611,
  [SMALL_STATE(987)] = 76658,
  [SMALL_STATE(988)] = 76705,
  [SMALL_STATE(989)] = 76752,
  [SMALL_STATE(990)] = 76799,
  [SMALL_STATE(991)] = 76850,
  [SMALL_STATE(992)] = 76901,
  [SMALL_STATE(993)] = 76948,
  [SMALL_STATE(994)] = 76995,
  [SMALL_STATE(995)] = 77051,
  [SMALL_STATE(996)] = 77109,
  [SMALL_STATE(997)] = 77165,
  [SMALL_STATE(998)] = 77221,
  [SMALL_STATE(999)] = 77279,
  [SMALL_STATE(1000)] = 77335,
  [SMALL_STATE(1001)] = 77384,
  [SMALL_STATE(1002)] = 77434,
  [SMALL_STATE(1003)] = 77482,
  [SMALL_STATE(1004)] = 77532,
  [SMALL_STATE(1005)] = 77582,
  [SMALL_STATE(1006)] = 77632,
  [SMALL_STATE(1007)] = 77679,
  [SMALL_STATE(1008)] = 77726,
  [SMALL_STATE(1009)] = 77773,
  [SMALL_STATE(1010)] = 77820,
  [SMALL_STATE(1011)] = 77867,
  [SMALL_STATE(1012)] = 77916,
  [SMALL_STATE(1013)] = 77963,
  [SMALL_STATE(1014)] = 78008,
  [SMALL_STATE(1015)] = 78049,
  [SMALL_STATE(1016)] = 78102,
  [SMALL_STATE(1017)] = 78143,
  [SMALL_STATE(1018)] = 78186,
  [SMALL_STATE(1019)] = 78237,
  [SMALL_STATE(1020)] = 78290,
  [SMALL_STATE(1021)] = 78343,
  [SMALL_STATE(1022)] = 78396,
  [SMALL_STATE(1023)] = 78449,
  [SMALL_STATE(1024)] = 78502,
  [SMALL_STATE(1025)] = 78551,
  [SMALL_STATE(1026)] = 78633,
  [SMALL_STATE(1027)] = 78673,
  [SMALL_STATE(1028)] = 78755,
  [SMALL_STATE(1029)] = 78795,
  [SMALL_STATE(1030)] = 78873,
  [SMALL_STATE(1031)] = 78923,
  [SMALL_STATE(1032)] = 79005,
  [SMALL_STATE(1033)] = 79055,
  [SMALL_STATE(1034)] = 79137,
  [SMALL_STATE(1035)] = 79219,
  [SMALL_STATE(1036)] = 79297,
  [SMALL_STATE(1037)] = 79347,
  [SMALL_STATE(1038)] = 79429,
  [SMALL_STATE(1039)] = 79511,
  [SMALL_STATE(1040)] = 79551,
  [SMALL_STATE(1041)] = 79591,
  [SMALL_STATE(1042)] = 79673,
  [SMALL_STATE(1043)] = 79723,
  [SMALL_STATE(1044)] = 79763,
  [SMALL_STATE(1045)] = 79803,
  [SMALL_STATE(1046)] = 79881,
  [SMALL_STATE(1047)] = 79959,
  [SMALL_STATE(1048)] = 80041,
  [SMALL_STATE(1049)] = 80123,
  [SMALL_STATE(1050)] = 80163,
  [SMALL_STATE(1051)] = 80245,
  [SMALL_STATE(1052)] = 80295,
  [SMALL_STATE(1053)] = 80377,
  [SMALL_STATE(1054)] = 80449,
  [SMALL_STATE(1055)] = 80499,
  [SMALL_STATE(1056)] = 80581,
  [SMALL_STATE(1057)] = 80663,
  [SMALL_STATE(1058)] = 80741,
  [SMALL_STATE(1059)] = 80823,
  [SMALL_STATE(1060)] = 80863,
  [SMALL_STATE(1061)] = 80909,
  [SMALL_STATE(1062)] = 80963,
  [SMALL_STATE(1063)] = 81041,
  [SMALL_STATE(1064)] = 81123,
  [SMALL_STATE(1065)] = 81189,
  [SMALL_STATE(1066)] = 81249,
  [SMALL_STATE(1067)] = 81307,
  [SMALL_STATE(1068)] = 81381,
  [SMALL_STATE(1069)] = 81435,
  [SMALL_STATE(1070)] = 81505,
  [SMALL_STATE(1071)] = 81573,
  [SMALL_STATE(1072)] = 81637,
  [SMALL_STATE(1073)] = 81677,
  [SMALL_STATE(1074)] = 81716,
  [SMALL_STATE(1075)] = 81797,
  [SMALL_STATE(1076)] = 81848,
  [SMALL_STATE(1077)] = 81929,
  [SMALL_STATE(1078)] = 81968,
  [SMALL_STATE(1079)] = 82007,
  [SMALL_STATE(1080)] = 82062,
  [SMALL_STATE(1081)] = 82141,
  [SMALL_STATE(1082)] = 82212,
  [SMALL_STATE(1083)] = 82263,
  [SMALL_STATE(1084)] = 82330,
  [SMALL_STATE(1085)] = 82395,
  [SMALL_STATE(1086)] = 82474,
  [SMALL_STATE(1087)] = 82535,
  [SMALL_STATE(1088)] = 82614,
  [SMALL_STATE(1089)] = 82671,
  [SMALL_STATE(1090)] = 82752,
  [SMALL_STATE(1091)] = 82797,
  [SMALL_STATE(1092)] = 82846,
  [SMALL_STATE(1093)] = 82915,
  [SMALL_STATE(1094)] = 82978,
  [SMALL_STATE(1095)] = 83059,
  [SMALL_STATE(1096)] = 83134,
  [SMALL_STATE(1097)] = 83173,
  [SMALL_STATE(1098)] = 83222,
  [SMALL_STATE(1099)] = 83271,
  [SMALL_STATE(1100)] = 83310,
  [SMALL_STATE(1101)] = 83349,
  [SMALL_STATE(1102)] = 83428,
  [SMALL_STATE(1103)] = 83503,
  [SMALL_STATE(1104)] = 83578,
  [SMALL_STATE(1105)] = 83653,
  [SMALL_STATE(1106)] = 83702,
  [SMALL_STATE(1107)] = 83751,
  [SMALL_STATE(1108)] = 83826,
  [SMALL_STATE(1109)] = 83901,
  [SMALL_STATE(1110)] = 83950,
  [SMALL_STATE(1111)] = 83999,
  [SMALL_STATE(1112)] = 84046,
  [SMALL_STATE(1113)] = 84085,
  [SMALL_STATE(1114)] = 84134,
  [SMALL_STATE(1115)] = 84173,
  [SMALL_STATE(1116)] = 84220,
  [SMALL_STATE(1117)] = 84259,
  [SMALL_STATE(1118)] = 84308,
  [SMALL_STATE(1119)] = 84355,
  [SMALL_STATE(1120)] = 84400,
  [SMALL_STATE(1121)] = 84447,
  [SMALL_STATE(1122)] = 84486,
  [SMALL_STATE(1123)] = 84563,
  [SMALL_STATE(1124)] = 84610,
  [SMALL_STATE(1125)] = 84667,
  [SMALL_STATE(1126)] = 84746,
  [SMALL_STATE(1127)] = 84793,
  [SMALL_STATE(1128)] = 84840,
  [SMALL_STATE(1129)] = 84879,
  [SMALL_STATE(1130)] = 84918,
  [SMALL_STATE(1131)] = 84975,
  [SMALL_STATE(1132)] = 85020,
  [SMALL_STATE(1133)] = 85059,
  [SMALL_STATE(1134)] = 85102,
  [SMALL_STATE(1135)] = 85145,
  [SMALL_STATE(1136)] = 85192,
  [SMALL_STATE(1137)] = 85231,
  [SMALL_STATE(1138)] = 85278,
  [SMALL_STATE(1139)] = 85355,
  [SMALL_STATE(1140)] = 85396,
  [SMALL_STATE(1141)] = 85435,
  [SMALL_STATE(1142)] = 85494,
  [SMALL_STATE(1143)] = 85533,
  [SMALL_STATE(1144)] = 85612,
  [SMALL_STATE(1145)] = 85651,
  [SMALL_STATE(1146)] = 85696,
  [SMALL_STATE(1147)] = 85735,
  [SMALL_STATE(1148)] = 85776,
  [SMALL_STATE(1149)] = 85815,
  [SMALL_STATE(1150)] = 85894,
  [SMALL_STATE(1151)] = 85973,
  [SMALL_STATE(1152)] = 86020,
  [SMALL_STATE(1153)] = 86099,
  [SMALL_STATE(1154)] = 86178,
  [SMALL_STATE(1155)] = 86217,
  [SMALL_STATE(1156)] = 86271,
  [SMALL_STATE(1157)] = 86317,
  [SMALL_STATE(1158)] = 86365,
  [SMALL_STATE(1159)] = 86415,
  [SMALL_STATE(1160)] = 86481,
  [SMALL_STATE(1161)] = 86519,
  [SMALL_STATE(1162)] = 86595,
  [SMALL_STATE(1163)] = 86671,
  [SMALL_STATE(1164)] = 86713,
  [SMALL_STATE(1165)] = 86759,
  [SMALL_STATE(1166)] = 86805,
  [SMALL_STATE(1167)] = 86853,
  [SMALL_STATE(1168)] = 86899,
  [SMALL_STATE(1169)] = 86937,
  [SMALL_STATE(1170)] = 86999,
  [SMALL_STATE(1171)] = 87071,
  [SMALL_STATE(1172)] = 87121,
  [SMALL_STATE(1173)] = 87165,
  [SMALL_STATE(1174)] = 87221,
  [SMALL_STATE(1175)] = 87259,
  [SMALL_STATE(1176)] = 87331,
  [SMALL_STATE(1177)] = 87407,
  [SMALL_STATE(1178)] = 87463,
  [SMALL_STATE(1179)] = 87511,
  [SMALL_STATE(1180)] = 87587,
  [SMALL_STATE(1181)] = 87663,
  [SMALL_STATE(1182)] = 87739,
  [SMALL_STATE(1183)] = 87815,
  [SMALL_STATE(1184)] = 87887,
  [SMALL_STATE(1185)] = 87925,
  [SMALL_STATE(1186)] = 87997,
  [SMALL_STATE(1187)] = 88073,
  [SMALL_STATE(1188)] = 88119,
  [SMALL_STATE(1189)] = 88157,
  [SMALL_STATE(1190)] = 88195,
  [SMALL_STATE(1191)] = 88267,
  [SMALL_STATE(1192)] = 88305,
  [SMALL_STATE(1193)] = 88353,
  [SMALL_STATE(1194)] = 88401,
  [SMALL_STATE(1195)] = 88439,
  [SMALL_STATE(1196)] = 88477,
  [SMALL_STATE(1197)] = 88515,
  [SMALL_STATE(1198)] = 88591,
  [SMALL_STATE(1199)] = 88645,
  [SMALL_STATE(1200)] = 88721,
  [SMALL_STATE(1201)] = 88797,
  [SMALL_STATE(1202)] = 88867,
  [SMALL_STATE(1203)] = 88943,
  [SMALL_STATE(1204)] = 89019,
  [SMALL_STATE(1205)] = 89095,
  [SMALL_STATE(1206)] = 89145,
  [SMALL_STATE(1207)] = 89211,
  [SMALL_STATE(1208)] = 89287,
  [SMALL_STATE(1209)] = 89351,
  [SMALL_STATE(1210)] = 89411,
  [SMALL_STATE(1211)] = 89467,
  [SMALL_STATE(1212)] = 89543,
  [SMALL_STATE(1213)] = 89611,
  [SMALL_STATE(1214)] = 89673,
  [SMALL_STATE(1215)] = 89745,
  [SMALL_STATE(1216)] = 89819,
  [SMALL_STATE(1217)] = 89861,
  [SMALL_STATE(1218)] = 89911,
  [SMALL_STATE(1219)] = 89983,
  [SMALL_STATE(1220)] = 90037,
  [SMALL_STATE(1221)] = 90105,
  [SMALL_STATE(1222)] = 90149,
  [SMALL_STATE(1223)] = 90199,
  [SMALL_STATE(1224)] = 90265,
  [SMALL_STATE(1225)] = 90311,
  [SMALL_STATE(1226)] = 90375,
  [SMALL_STATE(1227)] = 90435,
  [SMALL_STATE(1228)] = 90477,
  [SMALL_STATE(1229)] = 90525,
  [SMALL_STATE(1230)] = 90563,
  [SMALL_STATE(1231)] = 90601,
  [SMALL_STATE(1232)] = 90649,
  [SMALL_STATE(1233)] = 90695,
  [SMALL_STATE(1234)] = 90767,
  [SMALL_STATE(1235)] = 90805,
  [SMALL_STATE(1236)] = 90879,
  [SMALL_STATE(1237)] = 90937,
  [SMALL_STATE(1238)] = 90975,
  [SMALL_STATE(1239)] = 91019,
  [SMALL_STATE(1240)] = 91095,
  [SMALL_STATE(1241)] = 91133,
  [SMALL_STATE(1242)] = 91185,
  [SMALL_STATE(1243)] = 91227,
  [SMALL_STATE(1244)] = 91269,
  [SMALL_STATE(1245)] = 91345,
  [SMALL_STATE(1246)] = 91419,
  [SMALL_STATE(1247)] = 91493,
  [SMALL_STATE(1248)] = 91535,
  [SMALL_STATE(1249)] = 91573,
  [SMALL_STATE(1250)] = 91611,
  [SMALL_STATE(1251)] = 91685,
  [SMALL_STATE(1252)] = 91759,
  [SMALL_STATE(1253)] = 91797,
  [SMALL_STATE(1254)] = 91873,
  [SMALL_STATE(1255)] = 91939,
  [SMALL_STATE(1256)] = 91979,
  [SMALL_STATE(1257)] = 92017,
  [SMALL_STATE(1258)] = 92077,
  [SMALL_STATE(1259)] = 92149,
  [SMALL_STATE(1260)] = 92225,
  [SMALL_STATE(1261)] = 92301,
  [SMALL_STATE(1262)] = 92339,
  [SMALL_STATE(1263)] = 92415,
  [SMALL_STATE(1264)] = 92487,
  [SMALL_STATE(1265)] = 92563,
  [SMALL_STATE(1266)] = 92605,
  [SMALL_STATE(1267)] = 92671,
  [SMALL_STATE(1268)] = 92739,
  [SMALL_STATE(1269)] = 92789,
  [SMALL_STATE(1270)] = 92853,
  [SMALL_STATE(1271)] = 92915,
  [SMALL_STATE(1272)] = 92971,
  [SMALL_STATE(1273)] = 93019,
  [SMALL_STATE(1274)] = 93095,
  [SMALL_STATE(1275)] = 93171,
  [SMALL_STATE(1276)] = 93247,
  [SMALL_STATE(1277)] = 93285,
  [SMALL_STATE(1278)] = 93361,
  [SMALL_STATE(1279)] = 93405,
  [SMALL_STATE(1280)] = 93443,
  [SMALL_STATE(1281)] = 93483,
  [SMALL_STATE(1282)] = 93531,
  [SMALL_STATE(1283)] = 93587,
  [SMALL_STATE(1284)] = 93625,
  [SMALL_STATE(1285)] = 93701,
  [SMALL_STATE(1286)] = 93738,
  [SMALL_STATE(1287)] = 93797,
  [SMALL_STATE(1288)] = 93838,
  [SMALL_STATE(1289)] = 93879,
  [SMALL_STATE(1290)] = 93930,
  [SMALL_STATE(1291)] = 93981,
  [SMALL_STATE(1292)] = 94024,
  [SMALL_STATE(1293)] = 94061,
  [SMALL_STATE(1294)] = 94112,
  [SMALL_STATE(1295)] = 94159,
  [SMALL_STATE(1296)] = 94210,
  [SMALL_STATE(1297)] = 94281,
  [SMALL_STATE(1298)] = 94320,
  [SMALL_STATE(1299)] = 94391,
  [SMALL_STATE(1300)] = 94432,
  [SMALL_STATE(1301)] = 94483,
  [SMALL_STATE(1302)] = 94530,
  [SMALL_STATE(1303)] = 94585,
  [SMALL_STATE(1304)] = 94622,
  [SMALL_STATE(1305)] = 94687,
  [SMALL_STATE(1306)] = 94748,
  [SMALL_STATE(1307)] = 94821,
  [SMALL_STATE(1308)] = 94892,
  [SMALL_STATE(1309)] = 94965,
  [SMALL_STATE(1310)] = 95002,
  [SMALL_STATE(1311)] = 95061,
  [SMALL_STATE(1312)] = 95100,
  [SMALL_STATE(1313)] = 95137,
  [SMALL_STATE(1314)] = 95174,
  [SMALL_STATE(1315)] = 95215,
  [SMALL_STATE(1316)] = 95278,
  [SMALL_STATE(1317)] = 95349,
  [SMALL_STATE(1318)] = 95420,
  [SMALL_STATE(1319)] = 95491,
  [SMALL_STATE(1320)] = 95562,
  [SMALL_STATE(1321)] = 95613,
  [SMALL_STATE(1322)] = 95664,
  [SMALL_STATE(1323)] = 95701,
  [SMALL_STATE(1324)] = 95752,
  [SMALL_STATE(1325)] = 95819,
  [SMALL_STATE(1326)] = 95868,
  [SMALL_STATE(1327)] = 95921,
  [SMALL_STATE(1328)] = 95970,
  [SMALL_STATE(1329)] = 96043,
  [SMALL_STATE(1330)] = 96092,
  [SMALL_STATE(1331)] = 96165,
  [SMALL_STATE(1332)] = 96218,
  [SMALL_STATE(1333)] = 96269,
  [SMALL_STATE(1334)] = 96320,
  [SMALL_STATE(1335)] = 96391,
  [SMALL_STATE(1336)] = 96428,
  [SMALL_STATE(1337)] = 96501,
  [SMALL_STATE(1338)] = 96542,
  [SMALL_STATE(1339)] = 96587,
  [SMALL_STATE(1340)] = 96652,
  [SMALL_STATE(1341)] = 96719,
  [SMALL_STATE(1342)] = 96760,
  [SMALL_STATE(1343)] = 96811,
  [SMALL_STATE(1344)] = 96848,
  [SMALL_STATE(1345)] = 96917,
  [SMALL_STATE(1346)] = 96966,
  [SMALL_STATE(1347)] = 97031,
  [SMALL_STATE(1348)] = 97072,
  [SMALL_STATE(1349)] = 97109,
  [SMALL_STATE(1350)] = 97160,
  [SMALL_STATE(1351)] = 97221,
  [SMALL_STATE(1352)] = 97284,
  [SMALL_STATE(1353)] = 97357,
  [SMALL_STATE(1354)] = 97394,
  [SMALL_STATE(1355)] = 97449,
  [SMALL_STATE(1356)] = 97522,
  [SMALL_STATE(1357)] = 97592,
  [SMALL_STATE(1358)] = 97662,
  [SMALL_STATE(1359)] = 97698,
  [SMALL_STATE(1360)] = 97744,
  [SMALL_STATE(1361)] = 97788,
  [SMALL_STATE(1362)] = 97858,
  [SMALL_STATE(1363)] = 97894,
  [SMALL_STATE(1364)] = 97964,
  [SMALL_STATE(1365)] = 98034,
  [SMALL_STATE(1366)] = 98104,
  [SMALL_STATE(1367)] = 98150,
  [SMALL_STATE(1368)] = 98186,
  [SMALL_STATE(1369)] = 98256,
  [SMALL_STATE(1370)] = 98292,
  [SMALL_STATE(1371)] = 98362,
  [SMALL_STATE(1372)] = 98432,
  [SMALL_STATE(1373)] = 98478,
  [SMALL_STATE(1374)] = 98514,
  [SMALL_STATE(1375)] = 98552,
  [SMALL_STATE(1376)] = 98588,
  [SMALL_STATE(1377)] = 98634,
  [SMALL_STATE(1378)] = 98704,
  [SMALL_STATE(1379)] = 98774,
  [SMALL_STATE(1380)] = 98844,
  [SMALL_STATE(1381)] = 98914,
  [SMALL_STATE(1382)] = 98958,
  [SMALL_STATE(1383)] = 98994,
  [SMALL_STATE(1384)] = 99038,
  [SMALL_STATE(1385)] = 99076,
  [SMALL_STATE(1386)] = 99122,
  [SMALL_STATE(1387)] = 99192,
  [SMALL_STATE(1388)] = 99262,
  [SMALL_STATE(1389)] = 99332,
  [SMALL_STATE(1390)] = 99376,
  [SMALL_STATE(1391)] = 99422,
  [SMALL_STATE(1392)] = 99470,
  [SMALL_STATE(1393)] = 99512,
  [SMALL_STATE(1394)] = 99582,
  [SMALL_STATE(1395)] = 99652,
  [SMALL_STATE(1396)] = 99722,
  [SMALL_STATE(1397)] = 99764,
  [SMALL_STATE(1398)] = 99834,
  [SMALL_STATE(1399)] = 99904,
  [SMALL_STATE(1400)] = 99974,
  [SMALL_STATE(1401)] = 100044,
  [SMALL_STATE(1402)] = 100114,
  [SMALL_STATE(1403)] = 100184,
  [SMALL_STATE(1404)] = 100254,
  [SMALL_STATE(1405)] = 100324,
  [SMALL_STATE(1406)] = 100394,
  [SMALL_STATE(1407)] = 100464,
  [SMALL_STATE(1408)] = 100534,
  [SMALL_STATE(1409)] = 100604,
  [SMALL_STATE(1410)] = 100674,
  [SMALL_STATE(1411)] = 100744,
  [SMALL_STATE(1412)] = 100780,
  [SMALL_STATE(1413)] = 100850,
  [SMALL_STATE(1414)] = 100920,
  [SMALL_STATE(1415)] = 100972,
  [SMALL_STATE(1416)] = 101036,
  [SMALL_STATE(1417)] = 101102,
  [SMALL_STATE(1418)] = 101150,
  [SMALL_STATE(1419)] = 101212,
  [SMALL_STATE(1420)] = 101272,
  [SMALL_STATE(1421)] = 101326,
  [SMALL_STATE(1422)] = 101372,
  [SMALL_STATE(1423)] = 101436,
  [SMALL_STATE(1424)] = 101494,
  [SMALL_STATE(1425)] = 101564,
  [SMALL_STATE(1426)] = 101634,
  [SMALL_STATE(1427)] = 101704,
  [SMALL_STATE(1428)] = 101774,
  [SMALL_STATE(1429)] = 101844,
  [SMALL_STATE(1430)] = 101880,
  [SMALL_STATE(1431)] = 101950,
  [SMALL_STATE(1432)] = 102020,
  [SMALL_STATE(1433)] = 102090,
  [SMALL_STATE(1434)] = 102160,
  [SMALL_STATE(1435)] = 102230,
  [SMALL_STATE(1436)] = 102300,
  [SMALL_STATE(1437)] = 102370,
  [SMALL_STATE(1438)] = 102440,
  [SMALL_STATE(1439)] = 102510,
  [SMALL_STATE(1440)] = 102580,
  [SMALL_STATE(1441)] = 102650,
  [SMALL_STATE(1442)] = 102720,
  [SMALL_STATE(1443)] = 102790,
  [SMALL_STATE(1444)] = 102860,
  [SMALL_STATE(1445)] = 102930,
  [SMALL_STATE(1446)] = 103000,
  [SMALL_STATE(1447)] = 103052,
  [SMALL_STATE(1448)] = 103118,
  [SMALL_STATE(1449)] = 103166,
  [SMALL_STATE(1450)] = 103230,
  [SMALL_STATE(1451)] = 103292,
  [SMALL_STATE(1452)] = 103350,
  [SMALL_STATE(1453)] = 103404,
  [SMALL_STATE(1454)] = 103450,
  [SMALL_STATE(1455)] = 103514,
  [SMALL_STATE(1456)] = 103574,
  [SMALL_STATE(1457)] = 103644,
  [SMALL_STATE(1458)] = 103714,
  [SMALL_STATE(1459)] = 103760,
  [SMALL_STATE(1460)] = 103804,
  [SMALL_STATE(1461)] = 103844,
  [SMALL_STATE(1462)] = 103914,
  [SMALL_STATE(1463)] = 103984,
  [SMALL_STATE(1464)] = 104054,
  [SMALL_STATE(1465)] = 104124,
  [SMALL_STATE(1466)] = 104194,
  [SMALL_STATE(1467)] = 104264,
  [SMALL_STATE(1468)] = 104334,
  [SMALL_STATE(1469)] = 104404,
  [SMALL_STATE(1470)] = 104474,
  [SMALL_STATE(1471)] = 104544,
  [SMALL_STATE(1472)] = 104580,
  [SMALL_STATE(1473)] = 104616,
  [SMALL_STATE(1474)] = 104652,
  [SMALL_STATE(1475)] = 104688,
  [SMALL_STATE(1476)] = 104724,
  [SMALL_STATE(1477)] = 104760,
  [SMALL_STATE(1478)] = 104830,
  [SMALL_STATE(1479)] = 104900,
  [SMALL_STATE(1480)] = 104970,
  [SMALL_STATE(1481)] = 105040,
  [SMALL_STATE(1482)] = 105110,
  [SMALL_STATE(1483)] = 105146,
  [SMALL_STATE(1484)] = 105216,
  [SMALL_STATE(1485)] = 105286,
  [SMALL_STATE(1486)] = 105356,
  [SMALL_STATE(1487)] = 105426,
  [SMALL_STATE(1488)] = 105496,
  [SMALL_STATE(1489)] = 105566,
  [SMALL_STATE(1490)] = 105636,
  [SMALL_STATE(1491)] = 105671,
  [SMALL_STATE(1492)] = 105706,
  [SMALL_STATE(1493)] = 105769,
  [SMALL_STATE(1494)] = 105836,
  [SMALL_STATE(1495)] = 105871,
  [SMALL_STATE(1496)] = 105906,
  [SMALL_STATE(1497)] = 105941,
  [SMALL_STATE(1498)] = 105982,
  [SMALL_STATE(1499)] = 106023,
  [SMALL_STATE(1500)] = 106060,
  [SMALL_STATE(1501)] = 106099,
  [SMALL_STATE(1502)] = 106134,
  [SMALL_STATE(1503)] = 106171,
  [SMALL_STATE(1504)] = 106214,
  [SMALL_STATE(1505)] = 106255,
  [SMALL_STATE(1506)] = 106290,
  [SMALL_STATE(1507)] = 106325,
  [SMALL_STATE(1508)] = 106360,
  [SMALL_STATE(1509)] = 106395,
  [SMALL_STATE(1510)] = 106436,
  [SMALL_STATE(1511)] = 106503,
  [SMALL_STATE(1512)] = 106544,
  [SMALL_STATE(1513)] = 106611,
  [SMALL_STATE(1514)] = 106646,
  [SMALL_STATE(1515)] = 106681,
  [SMALL_STATE(1516)] = 106716,
  [SMALL_STATE(1517)] = 106751,
  [SMALL_STATE(1518)] = 106785,
  [SMALL_STATE(1519)] = 106819,
  [SMALL_STATE(1520)] = 106853,
  [SMALL_STATE(1521)] = 106887,
  [SMALL_STATE(1522)] = 106921,
  [SMALL_STATE(1523)] = 106955,
  [SMALL_STATE(1524)] = 106989,
  [SMALL_STATE(1525)] = 107023,
  [SMALL_STATE(1526)] = 107057,
  [SMALL_STATE(1527)] = 107095,
  [SMALL_STATE(1528)] = 107129,
  [SMALL_STATE(1529)] = 107169,
  [SMALL_STATE(1530)] = 107203,
  [SMALL_STATE(1531)] = 107237,
  [SMALL_STATE(1532)] = 107271,
  [SMALL_STATE(1533)] = 107305,
  [SMALL_STATE(1534)] = 107341,
  [SMALL_STATE(1535)] = 107375,
  [SMALL_STATE(1536)] = 107409,
  [SMALL_STATE(1537)] = 107447,
  [SMALL_STATE(1538)] = 107481,
  [SMALL_STATE(1539)] = 107515,
  [SMALL_STATE(1540)] = 107553,
  [SMALL_STATE(1541)] = 107587,
  [SMALL_STATE(1542)] = 107621,
  [SMALL_STATE(1543)] = 107655,
  [SMALL_STATE(1544)] = 107689,
  [SMALL_STATE(1545)] = 107729,
  [SMALL_STATE(1546)] = 107763,
  [SMALL_STATE(1547)] = 107797,
  [SMALL_STATE(1548)] = 107831,
  [SMALL_STATE(1549)] = 107865,
  [SMALL_STATE(1550)] = 107899,
  [SMALL_STATE(1551)] = 107933,
  [SMALL_STATE(1552)] = 107967,
  [SMALL_STATE(1553)] = 108001,
  [SMALL_STATE(1554)] = 108035,
  [SMALL_STATE(1555)] = 108069,
  [SMALL_STATE(1556)] = 108103,
  [SMALL_STATE(1557)] = 108137,
  [SMALL_STATE(1558)] = 108171,
  [SMALL_STATE(1559)] = 108205,
  [SMALL_STATE(1560)] = 108239,
  [SMALL_STATE(1561)] = 108273,
  [SMALL_STATE(1562)] = 108307,
  [SMALL_STATE(1563)] = 108341,
  [SMALL_STATE(1564)] = 108375,
  [SMALL_STATE(1565)] = 108409,
  [SMALL_STATE(1566)] = 108443,
  [SMALL_STATE(1567)] = 108477,
  [SMALL_STATE(1568)] = 108517,
  [SMALL_STATE(1569)] = 108551,
  [SMALL_STATE(1570)] = 108585,
  [SMALL_STATE(1571)] = 108619,
  [SMALL_STATE(1572)] = 108653,
  [SMALL_STATE(1573)] = 108687,
  [SMALL_STATE(1574)] = 108721,
  [SMALL_STATE(1575)] = 108755,
  [SMALL_STATE(1576)] = 108789,
  [SMALL_STATE(1577)] = 108823,
  [SMALL_STATE(1578)] = 108857,
  [SMALL_STATE(1579)] = 108891,
  [SMALL_STATE(1580)] = 108925,
  [SMALL_STATE(1581)] = 108959,
  [SMALL_STATE(1582)] = 108993,
  [SMALL_STATE(1583)] = 109027,
  [SMALL_STATE(1584)] = 109061,
  [SMALL_STATE(1585)] = 109095,
  [SMALL_STATE(1586)] = 109129,
  [SMALL_STATE(1587)] = 109163,
  [SMALL_STATE(1588)] = 109197,
  [SMALL_STATE(1589)] = 109231,
  [SMALL_STATE(1590)] = 109265,
  [SMALL_STATE(1591)] = 109299,
  [SMALL_STATE(1592)] = 109333,
  [SMALL_STATE(1593)] = 109367,
  [SMALL_STATE(1594)] = 109401,
  [SMALL_STATE(1595)] = 109435,
  [SMALL_STATE(1596)] = 109473,
  [SMALL_STATE(1597)] = 109530,
  [SMALL_STATE(1598)] = 109601,
  [SMALL_STATE(1599)] = 109636,
  [SMALL_STATE(1600)] = 109707,
  [SMALL_STATE(1601)] = 109764,
  [SMALL_STATE(1602)] = 109801,
  [SMALL_STATE(1603)] = 109837,
  [SMALL_STATE(1604)] = 109891,
  [SMALL_STATE(1605)] = 109927,
  [SMALL_STATE(1606)] = 109978,
  [SMALL_STATE(1607)] = 110008,
  [SMALL_STATE(1608)] = 110056,
  [SMALL_STATE(1609)] = 110104,
  [SMALL_STATE(1610)] = 110149,
  [SMALL_STATE(1611)] = 110194,
  [SMALL_STATE(1612)] = 110233,
  [SMALL_STATE(1613)] = 110274,
  [SMALL_STATE(1614)] = 110301,
  [SMALL_STATE(1615)] = 110344,
  [SMALL_STATE(1616)] = 110385,
  [SMALL_STATE(1617)] = 110419,
  [SMALL_STATE(1618)] = 110460,
  [SMALL_STATE(1619)] = 110501,
  [SMALL_STATE(1620)] = 110542,
  [SMALL_STATE(1621)] = 110583,
  [SMALL_STATE(1622)] = 110624,
  [SMALL_STATE(1623)] = 110665,
  [SMALL_STATE(1624)] = 110706,
  [SMALL_STATE(1625)] = 110747,
  [SMALL_STATE(1626)] = 110777,
  [SMALL_STATE(1627)] = 110801,
  [SMALL_STATE(1628)] = 110839,
  [SMALL_STATE(1629)] = 110863,
  [SMALL_STATE(1630)] = 110889,
  [SMALL_STATE(1631)] = 110923,
  [SMALL_STATE(1632)] = 110955,
  [SMALL_STATE(1633)] = 110993,
  [SMALL_STATE(1634)] = 111031,
  [SMALL_STATE(1635)] = 111069,
  [SMALL_STATE(1636)] = 111107,
  [SMALL_STATE(1637)] = 111131,
  [SMALL_STATE(1638)] = 111169,
  [SMALL_STATE(1639)] = 111207,
  [SMALL_STATE(1640)] = 111245,
  [SMALL_STATE(1641)] = 111283,
  [SMALL_STATE(1642)] = 111321,
  [SMALL_STATE(1643)] = 111359,
  [SMALL_STATE(1644)] = 111396,
  [SMALL_STATE(1645)] = 111435,
  [SMALL_STATE(1646)] = 111474,
  [SMALL_STATE(1647)] = 111513,
  [SMALL_STATE(1648)] = 111550,
  [SMALL_STATE(1649)] = 111589,
  [SMALL_STATE(1650)] = 111614,
  [SMALL_STATE(1651)] = 111651,
  [SMALL_STATE(1652)] = 111678,
  [SMALL_STATE(1653)] = 111715,
  [SMALL_STATE(1654)] = 111752,
  [SMALL_STATE(1655)] = 111789,
  [SMALL_STATE(1656)] = 111828,
  [SMALL_STATE(1657)] = 111867,
  [SMALL_STATE(1658)] = 111904,
  [SMALL_STATE(1659)] = 111943,
  [SMALL_STATE(1660)] = 111982,
  [SMALL_STATE(1661)] = 112019,
  [SMALL_STATE(1662)] = 112053,
  [SMALL_STATE(1663)] = 112087,
  [SMALL_STATE(1664)] = 112121,
  [SMALL_STATE(1665)] = 112155,
  [SMALL_STATE(1666)] = 112189,
  [SMALL_STATE(1667)] = 112223,
  [SMALL_STATE(1668)] = 112257,
  [SMALL_STATE(1669)] = 112291,
  [SMALL_STATE(1670)] = 112325,
  [SMALL_STATE(1671)] = 112359,
  [SMALL_STATE(1672)] = 112393,
  [SMALL_STATE(1673)] = 112427,
  [SMALL_STATE(1674)] = 112461,
  [SMALL_STATE(1675)] = 112495,
  [SMALL_STATE(1676)] = 112529,
  [SMALL_STATE(1677)] = 112563,
  [SMALL_STATE(1678)] = 112597,
  [SMALL_STATE(1679)] = 112633,
  [SMALL_STATE(1680)] = 112667,
  [SMALL_STATE(1681)] = 112701,
  [SMALL_STATE(1682)] = 112725,
  [SMALL_STATE(1683)] = 112759,
  [SMALL_STATE(1684)] = 112793,
  [SMALL_STATE(1685)] = 112827,
  [SMALL_STATE(1686)] = 112861,
  [SMALL_STATE(1687)] = 112885,
  [SMALL_STATE(1688)] = 112909,
  [SMALL_STATE(1689)] = 112943,
  [SMALL_STATE(1690)] = 112971,
  [SMALL_STATE(1691)] = 112995,
  [SMALL_STATE(1692)] = 113019,
  [SMALL_STATE(1693)] = 113051,
  [SMALL_STATE(1694)] = 113085,
  [SMALL_STATE(1695)] = 113119,
  [SMALL_STATE(1696)] = 113149,
  [SMALL_STATE(1697)] = 113185,
  [SMALL_STATE(1698)] = 113221,
  [SMALL_STATE(1699)] = 113255,
  [SMALL_STATE(1700)] = 113279,
  [SMALL_STATE(1701)] = 113303,
  [SMALL_STATE(1702)] = 113327,
  [SMALL_STATE(1703)] = 113351,
  [SMALL_STATE(1704)] = 113385,
  [SMALL_STATE(1705)] = 113419,
  [SMALL_STATE(1706)] = 113443,
  [SMALL_STATE(1707)] = 113467,
  [SMALL_STATE(1708)] = 113501,
  [SMALL_STATE(1709)] = 113525,
  [SMALL_STATE(1710)] = 113559,
  [SMALL_STATE(1711)] = 113583,
  [SMALL_STATE(1712)] = 113607,
  [SMALL_STATE(1713)] = 113631,
  [SMALL_STATE(1714)] = 113655,
  [SMALL_STATE(1715)] = 113689,
  [SMALL_STATE(1716)] = 113715,
  [SMALL_STATE(1717)] = 113739,
  [SMALL_STATE(1718)] = 113763,
  [SMALL_STATE(1719)] = 113797,
  [SMALL_STATE(1720)] = 113831,
  [SMALL_STATE(1721)] = 113865,
  [SMALL_STATE(1722)] = 113899,
  [SMALL_STATE(1723)] = 113933,
  [SMALL_STATE(1724)] = 113967,
  [SMALL_STATE(1725)] = 113991,
  [SMALL_STATE(1726)] = 114027,
  [SMALL_STATE(1727)] = 114051,
  [SMALL_STATE(1728)] = 114085,
  [SMALL_STATE(1729)] = 114119,
  [SMALL_STATE(1730)] = 114153,
  [SMALL_STATE(1731)] = 114187,
  [SMALL_STATE(1732)] = 114221,
  [SMALL_STATE(1733)] = 114255,
  [SMALL_STATE(1734)] = 114279,
  [SMALL_STATE(1735)] = 114313,
  [SMALL_STATE(1736)] = 114347,
  [SMALL_STATE(1737)] = 114381,
  [SMALL_STATE(1738)] = 114415,
  [SMALL_STATE(1739)] = 114449,
  [SMALL_STATE(1740)] = 114483,
  [SMALL_STATE(1741)] = 114517,
  [SMALL_STATE(1742)] = 114551,
  [SMALL_STATE(1743)] = 114585,
  [SMALL_STATE(1744)] = 114619,
  [SMALL_STATE(1745)] = 114643,
  [SMALL_STATE(1746)] = 114676,
  [SMALL_STATE(1747)] = 114709,
  [SMALL_STATE(1748)] = 114746,
  [SMALL_STATE(1749)] = 114785,
  [SMALL_STATE(1750)] = 114816,
  [SMALL_STATE(1751)] = 114846,
  [SMALL_STATE(1752)] = 114874,
  [SMALL_STATE(1753)] = 114904,
  [SMALL_STATE(1754)] = 114929,
  [SMALL_STATE(1755)] = 114954,
  [SMALL_STATE(1756)] = 114979,
  [SMALL_STATE(1757)] = 115004,
  [SMALL_STATE(1758)] = 115027,
  [SMALL_STATE(1759)] = 115052,
  [SMALL_STATE(1760)] = 115075,
  [SMALL_STATE(1761)] = 115099,
  [SMALL_STATE(1762)] = 115123,
  [SMALL_STATE(1763)] = 115141,
  [SMALL_STATE(1764)] = 115171,
  [SMALL_STATE(1765)] = 115195,
  [SMALL_STATE(1766)] = 115219,
  [SMALL_STATE(1767)] = 115237,
  [SMALL_STATE(1768)] = 115267,
  [SMALL_STATE(1769)] = 115285,
  [SMALL_STATE(1770)] = 115309,
  [SMALL_STATE(1771)] = 115327,
  [SMALL_STATE(1772)] = 115345,
  [SMALL_STATE(1773)] = 115366,
  [SMALL_STATE(1774)] = 115395,
  [SMALL_STATE(1775)] = 115424,
  [SMALL_STATE(1776)] = 115453,
  [SMALL_STATE(1777)] = 115480,
  [SMALL_STATE(1778)] = 115507,
  [SMALL_STATE(1779)] = 115529,
  [SMALL_STATE(1780)] = 115551,
  [SMALL_STATE(1781)] = 115575,
  [SMALL_STATE(1782)] = 115597,
  [SMALL_STATE(1783)] = 115619,
  [SMALL_STATE(1784)] = 115641,
  [SMALL_STATE(1785)] = 115663,
  [SMALL_STATE(1786)] = 115685,
  [SMALL_STATE(1787)] = 115707,
  [SMALL_STATE(1788)] = 115729,
  [SMALL_STATE(1789)] = 115751,
  [SMALL_STATE(1790)] = 115773,
  [SMALL_STATE(1791)] = 115795,
  [SMALL_STATE(1792)] = 115821,
  [SMALL_STATE(1793)] = 115843,
  [SMALL_STATE(1794)] = 115865,
  [SMALL_STATE(1795)] = 115891,
  [SMALL_STATE(1796)] = 115915,
  [SMALL_STATE(1797)] = 115937,
  [SMALL_STATE(1798)] = 115959,
  [SMALL_STATE(1799)] = 115981,
  [SMALL_STATE(1800)] = 115997,
  [SMALL_STATE(1801)] = 116019,
  [SMALL_STATE(1802)] = 116041,
  [SMALL_STATE(1803)] = 116063,
  [SMALL_STATE(1804)] = 116081,
  [SMALL_STATE(1805)] = 116103,
  [SMALL_STATE(1806)] = 116125,
  [SMALL_STATE(1807)] = 116147,
  [SMALL_STATE(1808)] = 116169,
  [SMALL_STATE(1809)] = 116191,
  [SMALL_STATE(1810)] = 116213,
  [SMALL_STATE(1811)] = 116235,
  [SMALL_STATE(1812)] = 116257,
  [SMALL_STATE(1813)] = 116277,
  [SMALL_STATE(1814)] = 116303,
  [SMALL_STATE(1815)] = 116325,
  [SMALL_STATE(1816)] = 116349,
  [SMALL_STATE(1817)] = 116371,
  [SMALL_STATE(1818)] = 116387,
  [SMALL_STATE(1819)] = 116409,
  [SMALL_STATE(1820)] = 116430,
  [SMALL_STATE(1821)] = 116451,
  [SMALL_STATE(1822)] = 116472,
  [SMALL_STATE(1823)] = 116493,
  [SMALL_STATE(1824)] = 116510,
  [SMALL_STATE(1825)] = 116529,
  [SMALL_STATE(1826)] = 116548,
  [SMALL_STATE(1827)] = 116569,
  [SMALL_STATE(1828)] = 116590,
  [SMALL_STATE(1829)] = 116611,
  [SMALL_STATE(1830)] = 116634,
  [SMALL_STATE(1831)] = 116655,
  [SMALL_STATE(1832)] = 116674,
  [SMALL_STATE(1833)] = 116695,
  [SMALL_STATE(1834)] = 116716,
  [SMALL_STATE(1835)] = 116737,
  [SMALL_STATE(1836)] = 116758,
  [SMALL_STATE(1837)] = 116779,
  [SMALL_STATE(1838)] = 116800,
  [SMALL_STATE(1839)] = 116821,
  [SMALL_STATE(1840)] = 116844,
  [SMALL_STATE(1841)] = 116867,
  [SMALL_STATE(1842)] = 116890,
  [SMALL_STATE(1843)] = 116911,
  [SMALL_STATE(1844)] = 116932,
  [SMALL_STATE(1845)] = 116953,
  [SMALL_STATE(1846)] = 116974,
  [SMALL_STATE(1847)] = 116995,
  [SMALL_STATE(1848)] = 117016,
  [SMALL_STATE(1849)] = 117037,
  [SMALL_STATE(1850)] = 117058,
  [SMALL_STATE(1851)] = 117081,
  [SMALL_STATE(1852)] = 117102,
  [SMALL_STATE(1853)] = 117125,
  [SMALL_STATE(1854)] = 117148,
  [SMALL_STATE(1855)] = 117169,
  [SMALL_STATE(1856)] = 117190,
  [SMALL_STATE(1857)] = 117211,
  [SMALL_STATE(1858)] = 117232,
  [SMALL_STATE(1859)] = 117253,
  [SMALL_STATE(1860)] = 117274,
  [SMALL_STATE(1861)] = 117295,
  [SMALL_STATE(1862)] = 117316,
  [SMALL_STATE(1863)] = 117337,
  [SMALL_STATE(1864)] = 117358,
  [SMALL_STATE(1865)] = 117375,
  [SMALL_STATE(1866)] = 117396,
  [SMALL_STATE(1867)] = 117417,
  [SMALL_STATE(1868)] = 117438,
  [SMALL_STATE(1869)] = 117461,
  [SMALL_STATE(1870)] = 117482,
  [SMALL_STATE(1871)] = 117503,
  [SMALL_STATE(1872)] = 117524,
  [SMALL_STATE(1873)] = 117541,
  [SMALL_STATE(1874)] = 117562,
  [SMALL_STATE(1875)] = 117583,
  [SMALL_STATE(1876)] = 117602,
  [SMALL_STATE(1877)] = 117623,
  [SMALL_STATE(1878)] = 117644,
  [SMALL_STATE(1879)] = 117665,
  [SMALL_STATE(1880)] = 117688,
  [SMALL_STATE(1881)] = 117711,
  [SMALL_STATE(1882)] = 117732,
  [SMALL_STATE(1883)] = 117753,
  [SMALL_STATE(1884)] = 117774,
  [SMALL_STATE(1885)] = 117795,
  [SMALL_STATE(1886)] = 117812,
  [SMALL_STATE(1887)] = 117833,
  [SMALL_STATE(1888)] = 117854,
  [SMALL_STATE(1889)] = 117873,
  [SMALL_STATE(1890)] = 117894,
  [SMALL_STATE(1891)] = 117915,
  [SMALL_STATE(1892)] = 117938,
  [SMALL_STATE(1893)] = 117961,
  [SMALL_STATE(1894)] = 117982,
  [SMALL_STATE(1895)] = 118003,
  [SMALL_STATE(1896)] = 118024,
  [SMALL_STATE(1897)] = 118041,
  [SMALL_STATE(1898)] = 118062,
  [SMALL_STATE(1899)] = 118083,
  [SMALL_STATE(1900)] = 118104,
  [SMALL_STATE(1901)] = 118125,
  [SMALL_STATE(1902)] = 118146,
  [SMALL_STATE(1903)] = 118167,
  [SMALL_STATE(1904)] = 118188,
  [SMALL_STATE(1905)] = 118209,
  [SMALL_STATE(1906)] = 118230,
  [SMALL_STATE(1907)] = 118251,
  [SMALL_STATE(1908)] = 118272,
  [SMALL_STATE(1909)] = 118293,
  [SMALL_STATE(1910)] = 118314,
  [SMALL_STATE(1911)] = 118335,
  [SMALL_STATE(1912)] = 118356,
  [SMALL_STATE(1913)] = 118377,
  [SMALL_STATE(1914)] = 118400,
  [SMALL_STATE(1915)] = 118423,
  [SMALL_STATE(1916)] = 118446,
  [SMALL_STATE(1917)] = 118467,
  [SMALL_STATE(1918)] = 118488,
  [SMALL_STATE(1919)] = 118507,
  [SMALL_STATE(1920)] = 118528,
  [SMALL_STATE(1921)] = 118549,
  [SMALL_STATE(1922)] = 118570,
  [SMALL_STATE(1923)] = 118591,
  [SMALL_STATE(1924)] = 118614,
  [SMALL_STATE(1925)] = 118635,
  [SMALL_STATE(1926)] = 118656,
  [SMALL_STATE(1927)] = 118679,
  [SMALL_STATE(1928)] = 118700,
  [SMALL_STATE(1929)] = 118720,
  [SMALL_STATE(1930)] = 118740,
  [SMALL_STATE(1931)] = 118760,
  [SMALL_STATE(1932)] = 118782,
  [SMALL_STATE(1933)] = 118800,
  [SMALL_STATE(1934)] = 118820,
  [SMALL_STATE(1935)] = 118842,
  [SMALL_STATE(1936)] = 118864,
  [SMALL_STATE(1937)] = 118886,
  [SMALL_STATE(1938)] = 118906,
  [SMALL_STATE(1939)] = 118920,
  [SMALL_STATE(1940)] = 118942,
  [SMALL_STATE(1941)] = 118960,
  [SMALL_STATE(1942)] = 118978,
  [SMALL_STATE(1943)] = 119000,
  [SMALL_STATE(1944)] = 119020,
  [SMALL_STATE(1945)] = 119040,
  [SMALL_STATE(1946)] = 119054,
  [SMALL_STATE(1947)] = 119076,
  [SMALL_STATE(1948)] = 119098,
  [SMALL_STATE(1949)] = 119118,
  [SMALL_STATE(1950)] = 119138,
  [SMALL_STATE(1951)] = 119158,
  [SMALL_STATE(1952)] = 119180,
  [SMALL_STATE(1953)] = 119200,
  [SMALL_STATE(1954)] = 119220,
  [SMALL_STATE(1955)] = 119234,
  [SMALL_STATE(1956)] = 119254,
  [SMALL_STATE(1957)] = 119274,
  [SMALL_STATE(1958)] = 119288,
  [SMALL_STATE(1959)] = 119302,
  [SMALL_STATE(1960)] = 119316,
  [SMALL_STATE(1961)] = 119336,
  [SMALL_STATE(1962)] = 119356,
  [SMALL_STATE(1963)] = 119378,
  [SMALL_STATE(1964)] = 119400,
  [SMALL_STATE(1965)] = 119420,
  [SMALL_STATE(1966)] = 119434,
  [SMALL_STATE(1967)] = 119456,
  [SMALL_STATE(1968)] = 119478,
  [SMALL_STATE(1969)] = 119500,
  [SMALL_STATE(1970)] = 119522,
  [SMALL_STATE(1971)] = 119542,
  [SMALL_STATE(1972)] = 119564,
  [SMALL_STATE(1973)] = 119586,
  [SMALL_STATE(1974)] = 119604,
  [SMALL_STATE(1975)] = 119624,
  [SMALL_STATE(1976)] = 119646,
  [SMALL_STATE(1977)] = 119668,
  [SMALL_STATE(1978)] = 119688,
  [SMALL_STATE(1979)] = 119702,
  [SMALL_STATE(1980)] = 119716,
  [SMALL_STATE(1981)] = 119730,
  [SMALL_STATE(1982)] = 119744,
  [SMALL_STATE(1983)] = 119758,
  [SMALL_STATE(1984)] = 119772,
  [SMALL_STATE(1985)] = 119794,
  [SMALL_STATE(1986)] = 119816,
  [SMALL_STATE(1987)] = 119836,
  [SMALL_STATE(1988)] = 119850,
  [SMALL_STATE(1989)] = 119870,
  [SMALL_STATE(1990)] = 119890,
  [SMALL_STATE(1991)] = 119908,
  [SMALL_STATE(1992)] = 119926,
  [SMALL_STATE(1993)] = 119940,
  [SMALL_STATE(1994)] = 119960,
  [SMALL_STATE(1995)] = 119978,
  [SMALL_STATE(1996)] = 119992,
  [SMALL_STATE(1997)] = 120006,
  [SMALL_STATE(1998)] = 120020,
  [SMALL_STATE(1999)] = 120034,
  [SMALL_STATE(2000)] = 120048,
  [SMALL_STATE(2001)] = 120062,
  [SMALL_STATE(2002)] = 120076,
  [SMALL_STATE(2003)] = 120090,
  [SMALL_STATE(2004)] = 120110,
  [SMALL_STATE(2005)] = 120128,
  [SMALL_STATE(2006)] = 120142,
  [SMALL_STATE(2007)] = 120164,
  [SMALL_STATE(2008)] = 120182,
  [SMALL_STATE(2009)] = 120196,
  [SMALL_STATE(2010)] = 120210,
  [SMALL_STATE(2011)] = 120224,
  [SMALL_STATE(2012)] = 120238,
  [SMALL_STATE(2013)] = 120252,
  [SMALL_STATE(2014)] = 120270,
  [SMALL_STATE(2015)] = 120292,
  [SMALL_STATE(2016)] = 120306,
  [SMALL_STATE(2017)] = 120328,
  [SMALL_STATE(2018)] = 120350,
  [SMALL_STATE(2019)] = 120372,
  [SMALL_STATE(2020)] = 120386,
  [SMALL_STATE(2021)] = 120406,
  [SMALL_STATE(2022)] = 120420,
  [SMALL_STATE(2023)] = 120442,
  [SMALL_STATE(2024)] = 120462,
  [SMALL_STATE(2025)] = 120484,
  [SMALL_STATE(2026)] = 120504,
  [SMALL_STATE(2027)] = 120524,
  [SMALL_STATE(2028)] = 120546,
  [SMALL_STATE(2029)] = 120568,
  [SMALL_STATE(2030)] = 120590,
  [SMALL_STATE(2031)] = 120610,
  [SMALL_STATE(2032)] = 120630,
  [SMALL_STATE(2033)] = 120644,
  [SMALL_STATE(2034)] = 120658,
  [SMALL_STATE(2035)] = 120678,
  [SMALL_STATE(2036)] = 120696,
  [SMALL_STATE(2037)] = 120710,
  [SMALL_STATE(2038)] = 120732,
  [SMALL_STATE(2039)] = 120754,
  [SMALL_STATE(2040)] = 120772,
  [SMALL_STATE(2041)] = 120794,
  [SMALL_STATE(2042)] = 120807,
  [SMALL_STATE(2043)] = 120820,
  [SMALL_STATE(2044)] = 120833,
  [SMALL_STATE(2045)] = 120846,
  [SMALL_STATE(2046)] = 120859,
  [SMALL_STATE(2047)] = 120878,
  [SMALL_STATE(2048)] = 120891,
  [SMALL_STATE(2049)] = 120908,
  [SMALL_STATE(2050)] = 120925,
  [SMALL_STATE(2051)] = 120938,
  [SMALL_STATE(2052)] = 120955,
  [SMALL_STATE(2053)] = 120974,
  [SMALL_STATE(2054)] = 120993,
  [SMALL_STATE(2055)] = 121006,
  [SMALL_STATE(2056)] = 121025,
  [SMALL_STATE(2057)] = 121042,
  [SMALL_STATE(2058)] = 121059,
  [SMALL_STATE(2059)] = 121072,
  [SMALL_STATE(2060)] = 121085,
  [SMALL_STATE(2061)] = 121098,
  [SMALL_STATE(2062)] = 121115,
  [SMALL_STATE(2063)] = 121128,
  [SMALL_STATE(2064)] = 121141,
  [SMALL_STATE(2065)] = 121154,
  [SMALL_STATE(2066)] = 121173,
  [SMALL_STATE(2067)] = 121190,
  [SMALL_STATE(2068)] = 121207,
  [SMALL_STATE(2069)] = 121226,
  [SMALL_STATE(2070)] = 121239,
  [SMALL_STATE(2071)] = 121256,
  [SMALL_STATE(2072)] = 121269,
  [SMALL_STATE(2073)] = 121284,
  [SMALL_STATE(2074)] = 121299,
  [SMALL_STATE(2075)] = 121316,
  [SMALL_STATE(2076)] = 121329,
  [SMALL_STATE(2077)] = 121348,
  [SMALL_STATE(2078)] = 121367,
  [SMALL_STATE(2079)] = 121380,
  [SMALL_STATE(2080)] = 121393,
  [SMALL_STATE(2081)] = 121412,
  [SMALL_STATE(2082)] = 121427,
  [SMALL_STATE(2083)] = 121444,
  [SMALL_STATE(2084)] = 121463,
  [SMALL_STATE(2085)] = 121478,
  [SMALL_STATE(2086)] = 121493,
  [SMALL_STATE(2087)] = 121506,
  [SMALL_STATE(2088)] = 121519,
  [SMALL_STATE(2089)] = 121538,
  [SMALL_STATE(2090)] = 121553,
  [SMALL_STATE(2091)] = 121572,
  [SMALL_STATE(2092)] = 121585,
  [SMALL_STATE(2093)] = 121598,
  [SMALL_STATE(2094)] = 121611,
  [SMALL_STATE(2095)] = 121624,
  [SMALL_STATE(2096)] = 121643,
  [SMALL_STATE(2097)] = 121660,
  [SMALL_STATE(2098)] = 121675,
  [SMALL_STATE(2099)] = 121692,
  [SMALL_STATE(2100)] = 121705,
  [SMALL_STATE(2101)] = 121722,
  [SMALL_STATE(2102)] = 121739,
  [SMALL_STATE(2103)] = 121758,
  [SMALL_STATE(2104)] = 121771,
  [SMALL_STATE(2105)] = 121790,
  [SMALL_STATE(2106)] = 121809,
  [SMALL_STATE(2107)] = 121826,
  [SMALL_STATE(2108)] = 121839,
  [SMALL_STATE(2109)] = 121852,
  [SMALL_STATE(2110)] = 121865,
  [SMALL_STATE(2111)] = 121878,
  [SMALL_STATE(2112)] = 121891,
  [SMALL_STATE(2113)] = 121904,
  [SMALL_STATE(2114)] = 121917,
  [SMALL_STATE(2115)] = 121934,
  [SMALL_STATE(2116)] = 121951,
  [SMALL_STATE(2117)] = 121968,
  [SMALL_STATE(2118)] = 121981,
  [SMALL_STATE(2119)] = 121994,
  [SMALL_STATE(2120)] = 122007,
  [SMALL_STATE(2121)] = 122020,
  [SMALL_STATE(2122)] = 122033,
  [SMALL_STATE(2123)] = 122046,
  [SMALL_STATE(2124)] = 122065,
  [SMALL_STATE(2125)] = 122078,
  [SMALL_STATE(2126)] = 122091,
  [SMALL_STATE(2127)] = 122104,
  [SMALL_STATE(2128)] = 122123,
  [SMALL_STATE(2129)] = 122140,
  [SMALL_STATE(2130)] = 122153,
  [SMALL_STATE(2131)] = 122166,
  [SMALL_STATE(2132)] = 122183,
  [SMALL_STATE(2133)] = 122196,
  [SMALL_STATE(2134)] = 122213,
  [SMALL_STATE(2135)] = 122226,
  [SMALL_STATE(2136)] = 122239,
  [SMALL_STATE(2137)] = 122256,
  [SMALL_STATE(2138)] = 122269,
  [SMALL_STATE(2139)] = 122282,
  [SMALL_STATE(2140)] = 122295,
  [SMALL_STATE(2141)] = 122308,
  [SMALL_STATE(2142)] = 122321,
  [SMALL_STATE(2143)] = 122334,
  [SMALL_STATE(2144)] = 122347,
  [SMALL_STATE(2145)] = 122364,
  [SMALL_STATE(2146)] = 122383,
  [SMALL_STATE(2147)] = 122396,
  [SMALL_STATE(2148)] = 122409,
  [SMALL_STATE(2149)] = 122422,
  [SMALL_STATE(2150)] = 122435,
  [SMALL_STATE(2151)] = 122448,
  [SMALL_STATE(2152)] = 122461,
  [SMALL_STATE(2153)] = 122474,
  [SMALL_STATE(2154)] = 122487,
  [SMALL_STATE(2155)] = 122500,
  [SMALL_STATE(2156)] = 122513,
  [SMALL_STATE(2157)] = 122526,
  [SMALL_STATE(2158)] = 122539,
  [SMALL_STATE(2159)] = 122552,
  [SMALL_STATE(2160)] = 122565,
  [SMALL_STATE(2161)] = 122584,
  [SMALL_STATE(2162)] = 122597,
  [SMALL_STATE(2163)] = 122616,
  [SMALL_STATE(2164)] = 122629,
  [SMALL_STATE(2165)] = 122642,
  [SMALL_STATE(2166)] = 122655,
  [SMALL_STATE(2167)] = 122668,
  [SMALL_STATE(2168)] = 122685,
  [SMALL_STATE(2169)] = 122702,
  [SMALL_STATE(2170)] = 122717,
  [SMALL_STATE(2171)] = 122736,
  [SMALL_STATE(2172)] = 122753,
  [SMALL_STATE(2173)] = 122770,
  [SMALL_STATE(2174)] = 122789,
  [SMALL_STATE(2175)] = 122808,
  [SMALL_STATE(2176)] = 122827,
  [SMALL_STATE(2177)] = 122840,
  [SMALL_STATE(2178)] = 122859,
  [SMALL_STATE(2179)] = 122874,
  [SMALL_STATE(2180)] = 122887,
  [SMALL_STATE(2181)] = 122900,
  [SMALL_STATE(2182)] = 122913,
  [SMALL_STATE(2183)] = 122926,
  [SMALL_STATE(2184)] = 122945,
  [SMALL_STATE(2185)] = 122958,
  [SMALL_STATE(2186)] = 122975,
  [SMALL_STATE(2187)] = 122988,
  [SMALL_STATE(2188)] = 123001,
  [SMALL_STATE(2189)] = 123014,
  [SMALL_STATE(2190)] = 123027,
  [SMALL_STATE(2191)] = 123040,
  [SMALL_STATE(2192)] = 123057,
  [SMALL_STATE(2193)] = 123074,
  [SMALL_STATE(2194)] = 123091,
  [SMALL_STATE(2195)] = 123104,
  [SMALL_STATE(2196)] = 123123,
  [SMALL_STATE(2197)] = 123140,
  [SMALL_STATE(2198)] = 123157,
  [SMALL_STATE(2199)] = 123174,
  [SMALL_STATE(2200)] = 123191,
  [SMALL_STATE(2201)] = 123204,
  [SMALL_STATE(2202)] = 123221,
  [SMALL_STATE(2203)] = 123234,
  [SMALL_STATE(2204)] = 123247,
  [SMALL_STATE(2205)] = 123260,
  [SMALL_STATE(2206)] = 123273,
  [SMALL_STATE(2207)] = 123286,
  [SMALL_STATE(2208)] = 123301,
  [SMALL_STATE(2209)] = 123320,
  [SMALL_STATE(2210)] = 123337,
  [SMALL_STATE(2211)] = 123356,
  [SMALL_STATE(2212)] = 123375,
  [SMALL_STATE(2213)] = 123392,
  [SMALL_STATE(2214)] = 123405,
  [SMALL_STATE(2215)] = 123424,
  [SMALL_STATE(2216)] = 123437,
  [SMALL_STATE(2217)] = 123454,
  [SMALL_STATE(2218)] = 123467,
  [SMALL_STATE(2219)] = 123486,
  [SMALL_STATE(2220)] = 123499,
  [SMALL_STATE(2221)] = 123513,
  [SMALL_STATE(2222)] = 123527,
  [SMALL_STATE(2223)] = 123541,
  [SMALL_STATE(2224)] = 123555,
  [SMALL_STATE(2225)] = 123569,
  [SMALL_STATE(2226)] = 123583,
  [SMALL_STATE(2227)] = 123597,
  [SMALL_STATE(2228)] = 123611,
  [SMALL_STATE(2229)] = 123625,
  [SMALL_STATE(2230)] = 123639,
  [SMALL_STATE(2231)] = 123653,
  [SMALL_STATE(2232)] = 123667,
  [SMALL_STATE(2233)] = 123681,
  [SMALL_STATE(2234)] = 123695,
  [SMALL_STATE(2235)] = 123709,
  [SMALL_STATE(2236)] = 123723,
  [SMALL_STATE(2237)] = 123737,
  [SMALL_STATE(2238)] = 123751,
  [SMALL_STATE(2239)] = 123765,
  [SMALL_STATE(2240)] = 123779,
  [SMALL_STATE(2241)] = 123793,
  [SMALL_STATE(2242)] = 123807,
  [SMALL_STATE(2243)] = 123821,
  [SMALL_STATE(2244)] = 123835,
  [SMALL_STATE(2245)] = 123849,
  [SMALL_STATE(2246)] = 123863,
  [SMALL_STATE(2247)] = 123877,
  [SMALL_STATE(2248)] = 123891,
  [SMALL_STATE(2249)] = 123905,
  [SMALL_STATE(2250)] = 123919,
  [SMALL_STATE(2251)] = 123933,
  [SMALL_STATE(2252)] = 123947,
  [SMALL_STATE(2253)] = 123961,
  [SMALL_STATE(2254)] = 123975,
  [SMALL_STATE(2255)] = 123989,
  [SMALL_STATE(2256)] = 124003,
  [SMALL_STATE(2257)] = 124017,
  [SMALL_STATE(2258)] = 124031,
  [SMALL_STATE(2259)] = 124045,
  [SMALL_STATE(2260)] = 124059,
  [SMALL_STATE(2261)] = 124073,
  [SMALL_STATE(2262)] = 124087,
  [SMALL_STATE(2263)] = 124101,
  [SMALL_STATE(2264)] = 124115,
  [SMALL_STATE(2265)] = 124129,
  [SMALL_STATE(2266)] = 124143,
  [SMALL_STATE(2267)] = 124157,
  [SMALL_STATE(2268)] = 124171,
  [SMALL_STATE(2269)] = 124185,
  [SMALL_STATE(2270)] = 124199,
  [SMALL_STATE(2271)] = 124213,
  [SMALL_STATE(2272)] = 124225,
  [SMALL_STATE(2273)] = 124239,
  [SMALL_STATE(2274)] = 124253,
  [SMALL_STATE(2275)] = 124267,
  [SMALL_STATE(2276)] = 124281,
  [SMALL_STATE(2277)] = 124295,
  [SMALL_STATE(2278)] = 124309,
  [SMALL_STATE(2279)] = 124323,
  [SMALL_STATE(2280)] = 124337,
  [SMALL_STATE(2281)] = 124351,
  [SMALL_STATE(2282)] = 124365,
  [SMALL_STATE(2283)] = 124379,
  [SMALL_STATE(2284)] = 124393,
  [SMALL_STATE(2285)] = 124407,
  [SMALL_STATE(2286)] = 124421,
  [SMALL_STATE(2287)] = 124435,
  [SMALL_STATE(2288)] = 124449,
  [SMALL_STATE(2289)] = 124463,
  [SMALL_STATE(2290)] = 124477,
  [SMALL_STATE(2291)] = 124491,
  [SMALL_STATE(2292)] = 124505,
  [SMALL_STATE(2293)] = 124519,
  [SMALL_STATE(2294)] = 124533,
  [SMALL_STATE(2295)] = 124547,
  [SMALL_STATE(2296)] = 124561,
  [SMALL_STATE(2297)] = 124575,
  [SMALL_STATE(2298)] = 124589,
  [SMALL_STATE(2299)] = 124603,
  [SMALL_STATE(2300)] = 124617,
  [SMALL_STATE(2301)] = 124631,
  [SMALL_STATE(2302)] = 124645,
  [SMALL_STATE(2303)] = 124659,
  [SMALL_STATE(2304)] = 124673,
  [SMALL_STATE(2305)] = 124687,
  [SMALL_STATE(2306)] = 124701,
  [SMALL_STATE(2307)] = 124715,
  [SMALL_STATE(2308)] = 124729,
  [SMALL_STATE(2309)] = 124743,
  [SMALL_STATE(2310)] = 124757,
  [SMALL_STATE(2311)] = 124771,
  [SMALL_STATE(2312)] = 124785,
  [SMALL_STATE(2313)] = 124799,
  [SMALL_STATE(2314)] = 124813,
  [SMALL_STATE(2315)] = 124827,
  [SMALL_STATE(2316)] = 124841,
  [SMALL_STATE(2317)] = 124855,
  [SMALL_STATE(2318)] = 124869,
  [SMALL_STATE(2319)] = 124883,
  [SMALL_STATE(2320)] = 124897,
  [SMALL_STATE(2321)] = 124911,
  [SMALL_STATE(2322)] = 124925,
  [SMALL_STATE(2323)] = 124939,
  [SMALL_STATE(2324)] = 124953,
  [SMALL_STATE(2325)] = 124967,
  [SMALL_STATE(2326)] = 124981,
  [SMALL_STATE(2327)] = 124995,
  [SMALL_STATE(2328)] = 125009,
  [SMALL_STATE(2329)] = 125023,
  [SMALL_STATE(2330)] = 125037,
  [SMALL_STATE(2331)] = 125051,
  [SMALL_STATE(2332)] = 125065,
  [SMALL_STATE(2333)] = 125079,
  [SMALL_STATE(2334)] = 125093,
  [SMALL_STATE(2335)] = 125107,
  [SMALL_STATE(2336)] = 125121,
  [SMALL_STATE(2337)] = 125135,
  [SMALL_STATE(2338)] = 125149,
  [SMALL_STATE(2339)] = 125163,
  [SMALL_STATE(2340)] = 125177,
  [SMALL_STATE(2341)] = 125191,
  [SMALL_STATE(2342)] = 125205,
  [SMALL_STATE(2343)] = 125219,
  [SMALL_STATE(2344)] = 125233,
  [SMALL_STATE(2345)] = 125247,
  [SMALL_STATE(2346)] = 125261,
  [SMALL_STATE(2347)] = 125275,
  [SMALL_STATE(2348)] = 125289,
  [SMALL_STATE(2349)] = 125303,
  [SMALL_STATE(2350)] = 125317,
  [SMALL_STATE(2351)] = 125331,
  [SMALL_STATE(2352)] = 125345,
  [SMALL_STATE(2353)] = 125359,
  [SMALL_STATE(2354)] = 125373,
  [SMALL_STATE(2355)] = 125387,
  [SMALL_STATE(2356)] = 125401,
  [SMALL_STATE(2357)] = 125415,
  [SMALL_STATE(2358)] = 125429,
  [SMALL_STATE(2359)] = 125443,
  [SMALL_STATE(2360)] = 125457,
  [SMALL_STATE(2361)] = 125471,
  [SMALL_STATE(2362)] = 125485,
  [SMALL_STATE(2363)] = 125499,
  [SMALL_STATE(2364)] = 125513,
  [SMALL_STATE(2365)] = 125527,
  [SMALL_STATE(2366)] = 125541,
  [SMALL_STATE(2367)] = 125555,
  [SMALL_STATE(2368)] = 125567,
  [SMALL_STATE(2369)] = 125581,
  [SMALL_STATE(2370)] = 125595,
  [SMALL_STATE(2371)] = 125609,
  [SMALL_STATE(2372)] = 125623,
  [SMALL_STATE(2373)] = 125637,
  [SMALL_STATE(2374)] = 125651,
  [SMALL_STATE(2375)] = 125665,
  [SMALL_STATE(2376)] = 125679,
  [SMALL_STATE(2377)] = 125693,
  [SMALL_STATE(2378)] = 125707,
  [SMALL_STATE(2379)] = 125721,
  [SMALL_STATE(2380)] = 125735,
  [SMALL_STATE(2381)] = 125749,
  [SMALL_STATE(2382)] = 125763,
  [SMALL_STATE(2383)] = 125777,
  [SMALL_STATE(2384)] = 125791,
  [SMALL_STATE(2385)] = 125805,
  [SMALL_STATE(2386)] = 125819,
  [SMALL_STATE(2387)] = 125833,
  [SMALL_STATE(2388)] = 125847,
  [SMALL_STATE(2389)] = 125861,
  [SMALL_STATE(2390)] = 125875,
  [SMALL_STATE(2391)] = 125889,
  [SMALL_STATE(2392)] = 125903,
  [SMALL_STATE(2393)] = 125917,
  [SMALL_STATE(2394)] = 125931,
  [SMALL_STATE(2395)] = 125945,
  [SMALL_STATE(2396)] = 125959,
  [SMALL_STATE(2397)] = 125973,
  [SMALL_STATE(2398)] = 125987,
  [SMALL_STATE(2399)] = 126001,
  [SMALL_STATE(2400)] = 126015,
  [SMALL_STATE(2401)] = 126029,
  [SMALL_STATE(2402)] = 126043,
  [SMALL_STATE(2403)] = 126059,
  [SMALL_STATE(2404)] = 126073,
  [SMALL_STATE(2405)] = 126087,
  [SMALL_STATE(2406)] = 126101,
  [SMALL_STATE(2407)] = 126115,
  [SMALL_STATE(2408)] = 126129,
  [SMALL_STATE(2409)] = 126143,
  [SMALL_STATE(2410)] = 126157,
  [SMALL_STATE(2411)] = 126171,
  [SMALL_STATE(2412)] = 126185,
  [SMALL_STATE(2413)] = 126199,
  [SMALL_STATE(2414)] = 126213,
  [SMALL_STATE(2415)] = 126227,
  [SMALL_STATE(2416)] = 126241,
  [SMALL_STATE(2417)] = 126255,
  [SMALL_STATE(2418)] = 126269,
  [SMALL_STATE(2419)] = 126283,
  [SMALL_STATE(2420)] = 126297,
  [SMALL_STATE(2421)] = 126311,
  [SMALL_STATE(2422)] = 126325,
  [SMALL_STATE(2423)] = 126339,
  [SMALL_STATE(2424)] = 126353,
  [SMALL_STATE(2425)] = 126367,
  [SMALL_STATE(2426)] = 126381,
  [SMALL_STATE(2427)] = 126395,
  [SMALL_STATE(2428)] = 126409,
  [SMALL_STATE(2429)] = 126423,
  [SMALL_STATE(2430)] = 126437,
  [SMALL_STATE(2431)] = 126451,
  [SMALL_STATE(2432)] = 126465,
  [SMALL_STATE(2433)] = 126479,
  [SMALL_STATE(2434)] = 126493,
  [SMALL_STATE(2435)] = 126507,
  [SMALL_STATE(2436)] = 126521,
  [SMALL_STATE(2437)] = 126535,
  [SMALL_STATE(2438)] = 126549,
  [SMALL_STATE(2439)] = 126563,
  [SMALL_STATE(2440)] = 126577,
  [SMALL_STATE(2441)] = 126591,
  [SMALL_STATE(2442)] = 126605,
  [SMALL_STATE(2443)] = 126615,
  [SMALL_STATE(2444)] = 126629,
  [SMALL_STATE(2445)] = 126643,
  [SMALL_STATE(2446)] = 126657,
  [SMALL_STATE(2447)] = 126671,
  [SMALL_STATE(2448)] = 126685,
  [SMALL_STATE(2449)] = 126699,
  [SMALL_STATE(2450)] = 126713,
  [SMALL_STATE(2451)] = 126727,
  [SMALL_STATE(2452)] = 126741,
  [SMALL_STATE(2453)] = 126755,
  [SMALL_STATE(2454)] = 126769,
  [SMALL_STATE(2455)] = 126783,
  [SMALL_STATE(2456)] = 126797,
  [SMALL_STATE(2457)] = 126811,
  [SMALL_STATE(2458)] = 126825,
  [SMALL_STATE(2459)] = 126839,
  [SMALL_STATE(2460)] = 126851,
  [SMALL_STATE(2461)] = 126865,
  [SMALL_STATE(2462)] = 126879,
  [SMALL_STATE(2463)] = 126893,
  [SMALL_STATE(2464)] = 126907,
  [SMALL_STATE(2465)] = 126921,
  [SMALL_STATE(2466)] = 126935,
  [SMALL_STATE(2467)] = 126949,
  [SMALL_STATE(2468)] = 126963,
  [SMALL_STATE(2469)] = 126977,
  [SMALL_STATE(2470)] = 126991,
  [SMALL_STATE(2471)] = 127005,
  [SMALL_STATE(2472)] = 127019,
  [SMALL_STATE(2473)] = 127033,
  [SMALL_STATE(2474)] = 127047,
  [SMALL_STATE(2475)] = 127061,
  [SMALL_STATE(2476)] = 127075,
  [SMALL_STATE(2477)] = 127089,
  [SMALL_STATE(2478)] = 127103,
  [SMALL_STATE(2479)] = 127117,
  [SMALL_STATE(2480)] = 127128,
  [SMALL_STATE(2481)] = 127139,
  [SMALL_STATE(2482)] = 127150,
  [SMALL_STATE(2483)] = 127161,
  [SMALL_STATE(2484)] = 127172,
  [SMALL_STATE(2485)] = 127183,
  [SMALL_STATE(2486)] = 127194,
  [SMALL_STATE(2487)] = 127205,
  [SMALL_STATE(2488)] = 127216,
  [SMALL_STATE(2489)] = 127227,
  [SMALL_STATE(2490)] = 127238,
  [SMALL_STATE(2491)] = 127249,
  [SMALL_STATE(2492)] = 127260,
  [SMALL_STATE(2493)] = 127271,
  [SMALL_STATE(2494)] = 127282,
  [SMALL_STATE(2495)] = 127291,
  [SMALL_STATE(2496)] = 127304,
  [SMALL_STATE(2497)] = 127315,
  [SMALL_STATE(2498)] = 127326,
  [SMALL_STATE(2499)] = 127337,
  [SMALL_STATE(2500)] = 127348,
  [SMALL_STATE(2501)] = 127357,
  [SMALL_STATE(2502)] = 127368,
  [SMALL_STATE(2503)] = 127379,
  [SMALL_STATE(2504)] = 127392,
  [SMALL_STATE(2505)] = 127403,
  [SMALL_STATE(2506)] = 127412,
  [SMALL_STATE(2507)] = 127425,
  [SMALL_STATE(2508)] = 127436,
  [SMALL_STATE(2509)] = 127447,
  [SMALL_STATE(2510)] = 127458,
  [SMALL_STATE(2511)] = 127471,
  [SMALL_STATE(2512)] = 127482,
  [SMALL_STATE(2513)] = 127495,
  [SMALL_STATE(2514)] = 127508,
  [SMALL_STATE(2515)] = 127519,
  [SMALL_STATE(2516)] = 127530,
  [SMALL_STATE(2517)] = 127541,
  [SMALL_STATE(2518)] = 127552,
  [SMALL_STATE(2519)] = 127561,
  [SMALL_STATE(2520)] = 127572,
  [SMALL_STATE(2521)] = 127583,
  [SMALL_STATE(2522)] = 127594,
  [SMALL_STATE(2523)] = 127605,
  [SMALL_STATE(2524)] = 127616,
  [SMALL_STATE(2525)] = 127627,
  [SMALL_STATE(2526)] = 127638,
  [SMALL_STATE(2527)] = 127649,
  [SMALL_STATE(2528)] = 127660,
  [SMALL_STATE(2529)] = 127671,
  [SMALL_STATE(2530)] = 127682,
  [SMALL_STATE(2531)] = 127693,
  [SMALL_STATE(2532)] = 127704,
  [SMALL_STATE(2533)] = 127715,
  [SMALL_STATE(2534)] = 127726,
  [SMALL_STATE(2535)] = 127737,
  [SMALL_STATE(2536)] = 127748,
  [SMALL_STATE(2537)] = 127759,
  [SMALL_STATE(2538)] = 127770,
  [SMALL_STATE(2539)] = 127781,
  [SMALL_STATE(2540)] = 127792,
  [SMALL_STATE(2541)] = 127803,
  [SMALL_STATE(2542)] = 127814,
  [SMALL_STATE(2543)] = 127825,
  [SMALL_STATE(2544)] = 127836,
  [SMALL_STATE(2545)] = 127847,
  [SMALL_STATE(2546)] = 127860,
  [SMALL_STATE(2547)] = 127871,
  [SMALL_STATE(2548)] = 127882,
  [SMALL_STATE(2549)] = 127893,
  [SMALL_STATE(2550)] = 127904,
  [SMALL_STATE(2551)] = 127915,
  [SMALL_STATE(2552)] = 127926,
  [SMALL_STATE(2553)] = 127937,
  [SMALL_STATE(2554)] = 127948,
  [SMALL_STATE(2555)] = 127959,
  [SMALL_STATE(2556)] = 127968,
  [SMALL_STATE(2557)] = 127979,
  [SMALL_STATE(2558)] = 127988,
  [SMALL_STATE(2559)] = 127999,
  [SMALL_STATE(2560)] = 128012,
  [SMALL_STATE(2561)] = 128023,
  [SMALL_STATE(2562)] = 128034,
  [SMALL_STATE(2563)] = 128045,
  [SMALL_STATE(2564)] = 128056,
  [SMALL_STATE(2565)] = 128067,
  [SMALL_STATE(2566)] = 128078,
  [SMALL_STATE(2567)] = 128089,
  [SMALL_STATE(2568)] = 128100,
  [SMALL_STATE(2569)] = 128111,
  [SMALL_STATE(2570)] = 128122,
  [SMALL_STATE(2571)] = 128133,
  [SMALL_STATE(2572)] = 128144,
  [SMALL_STATE(2573)] = 128155,
  [SMALL_STATE(2574)] = 128166,
  [SMALL_STATE(2575)] = 128177,
  [SMALL_STATE(2576)] = 128188,
  [SMALL_STATE(2577)] = 128199,
  [SMALL_STATE(2578)] = 128210,
  [SMALL_STATE(2579)] = 128221,
  [SMALL_STATE(2580)] = 128232,
  [SMALL_STATE(2581)] = 128243,
  [SMALL_STATE(2582)] = 128254,
  [SMALL_STATE(2583)] = 128265,
  [SMALL_STATE(2584)] = 128276,
  [SMALL_STATE(2585)] = 128287,
  [SMALL_STATE(2586)] = 128298,
  [SMALL_STATE(2587)] = 128309,
  [SMALL_STATE(2588)] = 128320,
  [SMALL_STATE(2589)] = 128331,
  [SMALL_STATE(2590)] = 128342,
  [SMALL_STATE(2591)] = 128353,
  [SMALL_STATE(2592)] = 128364,
  [SMALL_STATE(2593)] = 128375,
  [SMALL_STATE(2594)] = 128386,
  [SMALL_STATE(2595)] = 128397,
  [SMALL_STATE(2596)] = 128408,
  [SMALL_STATE(2597)] = 128419,
  [SMALL_STATE(2598)] = 128430,
  [SMALL_STATE(2599)] = 128441,
  [SMALL_STATE(2600)] = 128452,
  [SMALL_STATE(2601)] = 128463,
  [SMALL_STATE(2602)] = 128474,
  [SMALL_STATE(2603)] = 128482,
  [SMALL_STATE(2604)] = 128490,
  [SMALL_STATE(2605)] = 128498,
  [SMALL_STATE(2606)] = 128506,
  [SMALL_STATE(2607)] = 128514,
  [SMALL_STATE(2608)] = 128522,
  [SMALL_STATE(2609)] = 128530,
  [SMALL_STATE(2610)] = 128538,
  [SMALL_STATE(2611)] = 128546,
  [SMALL_STATE(2612)] = 128554,
  [SMALL_STATE(2613)] = 128562,
  [SMALL_STATE(2614)] = 128570,
  [SMALL_STATE(2615)] = 128578,
  [SMALL_STATE(2616)] = 128586,
  [SMALL_STATE(2617)] = 128594,
  [SMALL_STATE(2618)] = 128602,
  [SMALL_STATE(2619)] = 128612,
  [SMALL_STATE(2620)] = 128620,
  [SMALL_STATE(2621)] = 128628,
  [SMALL_STATE(2622)] = 128636,
  [SMALL_STATE(2623)] = 128644,
  [SMALL_STATE(2624)] = 128652,
  [SMALL_STATE(2625)] = 128660,
  [SMALL_STATE(2626)] = 128668,
  [SMALL_STATE(2627)] = 128676,
  [SMALL_STATE(2628)] = 128684,
  [SMALL_STATE(2629)] = 128692,
  [SMALL_STATE(2630)] = 128702,
  [SMALL_STATE(2631)] = 128710,
  [SMALL_STATE(2632)] = 128718,
  [SMALL_STATE(2633)] = 128726,
  [SMALL_STATE(2634)] = 128734,
  [SMALL_STATE(2635)] = 128742,
  [SMALL_STATE(2636)] = 128750,
  [SMALL_STATE(2637)] = 128758,
  [SMALL_STATE(2638)] = 128766,
  [SMALL_STATE(2639)] = 128774,
  [SMALL_STATE(2640)] = 128782,
  [SMALL_STATE(2641)] = 128790,
  [SMALL_STATE(2642)] = 128798,
  [SMALL_STATE(2643)] = 128806,
  [SMALL_STATE(2644)] = 128814,
  [SMALL_STATE(2645)] = 128822,
  [SMALL_STATE(2646)] = 128830,
  [SMALL_STATE(2647)] = 128838,
  [SMALL_STATE(2648)] = 128846,
  [SMALL_STATE(2649)] = 128854,
  [SMALL_STATE(2650)] = 128862,
  [SMALL_STATE(2651)] = 128870,
  [SMALL_STATE(2652)] = 128878,
  [SMALL_STATE(2653)] = 128886,
  [SMALL_STATE(2654)] = 128894,
  [SMALL_STATE(2655)] = 128902,
  [SMALL_STATE(2656)] = 128910,
  [SMALL_STATE(2657)] = 128918,
  [SMALL_STATE(2658)] = 128928,
  [SMALL_STATE(2659)] = 128938,
  [SMALL_STATE(2660)] = 128946,
  [SMALL_STATE(2661)] = 128954,
  [SMALL_STATE(2662)] = 128962,
  [SMALL_STATE(2663)] = 128970,
  [SMALL_STATE(2664)] = 128978,
  [SMALL_STATE(2665)] = 128986,
  [SMALL_STATE(2666)] = 128994,
  [SMALL_STATE(2667)] = 129002,
  [SMALL_STATE(2668)] = 129012,
  [SMALL_STATE(2669)] = 129022,
  [SMALL_STATE(2670)] = 129030,
  [SMALL_STATE(2671)] = 129038,
  [SMALL_STATE(2672)] = 129046,
  [SMALL_STATE(2673)] = 129056,
  [SMALL_STATE(2674)] = 129064,
  [SMALL_STATE(2675)] = 129072,
  [SMALL_STATE(2676)] = 129080,
  [SMALL_STATE(2677)] = 129088,
  [SMALL_STATE(2678)] = 129098,
  [SMALL_STATE(2679)] = 129106,
  [SMALL_STATE(2680)] = 129114,
  [SMALL_STATE(2681)] = 129122,
  [SMALL_STATE(2682)] = 129130,
  [SMALL_STATE(2683)] = 129138,
  [SMALL_STATE(2684)] = 129148,
  [SMALL_STATE(2685)] = 129158,
  [SMALL_STATE(2686)] = 129168,
  [SMALL_STATE(2687)] = 129176,
  [SMALL_STATE(2688)] = 129184,
  [SMALL_STATE(2689)] = 129192,
  [SMALL_STATE(2690)] = 129200,
  [SMALL_STATE(2691)] = 129208,
  [SMALL_STATE(2692)] = 129216,
  [SMALL_STATE(2693)] = 129224,
  [SMALL_STATE(2694)] = 129232,
  [SMALL_STATE(2695)] = 129240,
  [SMALL_STATE(2696)] = 129250,
  [SMALL_STATE(2697)] = 129258,
  [SMALL_STATE(2698)] = 129266,
  [SMALL_STATE(2699)] = 129274,
  [SMALL_STATE(2700)] = 129282,
  [SMALL_STATE(2701)] = 129290,
  [SMALL_STATE(2702)] = 129298,
  [SMALL_STATE(2703)] = 129306,
  [SMALL_STATE(2704)] = 129314,
  [SMALL_STATE(2705)] = 129322,
  [SMALL_STATE(2706)] = 129330,
  [SMALL_STATE(2707)] = 129338,
  [SMALL_STATE(2708)] = 129346,
  [SMALL_STATE(2709)] = 129354,
  [SMALL_STATE(2710)] = 129362,
  [SMALL_STATE(2711)] = 129370,
  [SMALL_STATE(2712)] = 129378,
  [SMALL_STATE(2713)] = 129386,
  [SMALL_STATE(2714)] = 129394,
  [SMALL_STATE(2715)] = 129402,
  [SMALL_STATE(2716)] = 129410,
  [SMALL_STATE(2717)] = 129418,
  [SMALL_STATE(2718)] = 129426,
  [SMALL_STATE(2719)] = 129434,
  [SMALL_STATE(2720)] = 129442,
  [SMALL_STATE(2721)] = 129450,
  [SMALL_STATE(2722)] = 129458,
  [SMALL_STATE(2723)] = 129466,
  [SMALL_STATE(2724)] = 129474,
  [SMALL_STATE(2725)] = 129482,
  [SMALL_STATE(2726)] = 129490,
  [SMALL_STATE(2727)] = 129498,
  [SMALL_STATE(2728)] = 129506,
  [SMALL_STATE(2729)] = 129514,
  [SMALL_STATE(2730)] = 129522,
  [SMALL_STATE(2731)] = 129530,
  [SMALL_STATE(2732)] = 129538,
  [SMALL_STATE(2733)] = 129546,
  [SMALL_STATE(2734)] = 129554,
  [SMALL_STATE(2735)] = 129562,
  [SMALL_STATE(2736)] = 129570,
  [SMALL_STATE(2737)] = 129578,
  [SMALL_STATE(2738)] = 129586,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script, 0, 0, 0),
  [7] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2070),
  [9] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2648),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2514),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2596),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2504),
  [17] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2517),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2305),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2490),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2478),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1599),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2227),
  [29] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2258),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2508),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2730),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2638),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2066),
  [39] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [41] = {.entry = {.count = 1, .reusable = false}}, SHIFT(361),
  [43] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2731),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(476),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [49] = {.entry = {.count = 1, .reusable = true}}, SHIFT(196),
  [51] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__prefix_postfix_expression, 2, 0, 27),
  [53] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__prefix_postfix_expression, 2, 0, 27),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(204),
  [57] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [59] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [61] = {.entry = {.count = 1, .reusable = true}}, SHIFT(569),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2370),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2620),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2377),
  [69] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2555),
  [71] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1796),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3),
  [75] = {.entry = {.count = 1, .reusable = false}}, SHIFT(412),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(452),
  [79] = {.entry = {.count = 1, .reusable = false}}, SHIFT(560),
  [81] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [83] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2482),
  [85] = {.entry = {.count = 1, .reusable = true}}, SHIFT(452),
  [87] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2706),
  [89] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_address, 1, 0, 0), SHIFT(1221),
  [92] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_address, 1, 0, 0), SHIFT(229),
  [95] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [97] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_address, 1, 0, 0),
  [99] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_address, 1, 0, 0),
  [101] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_address, 1, 0, 0), SHIFT(262),
  [104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [106] = {.entry = {.count = 1, .reusable = false}}, SHIFT(262),
  [108] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_address, 1, 0, 0), SHIFT(263),
  [111] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_address, 1, 0, 0), SHIFT(1482),
  [114] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_address, 1, 0, 0), SHIFT(2370),
  [117] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_address, 1, 0, 0), SHIFT(2620),
  [120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2254),
  [122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2494),
  [124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1790),
  [126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5),
  [128] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1512),
  [130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1471),
  [132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1362),
  [134] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2586),
  [136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1512),
  [138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1216),
  [140] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1221),
  [142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(217),
  [148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1482),
  [154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2253),
  [156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1112),
  [158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1265),
  [160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1265),
  [162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2678),
  [164] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_address, 1, 0, 0), SHIFT(233),
  [167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1493),
  [171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1493),
  [173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1291),
  [175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(209),
  [181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1496),
  [187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2447),
  [189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2349),
  [191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2557),
  [193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1782),
  [195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7),
  [197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1261),
  [199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1337),
  [201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1505),
  [203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1507),
  [205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2592),
  [207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1337),
  [209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2665),
  [211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1341),
  [213] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_address, 1, 0, 0), SHIFT(225),
  [216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1510),
  [220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1510),
  [222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1460),
  [226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1460),
  [228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(818),
  [230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_sizeof_expression, 4, 0, 67),
  [232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [234] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_sizeof_expression, 4, 0, 67),
  [236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(778),
  [238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(778),
  [240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(779),
  [242] = {.entry = {.count = 1, .reusable = false}}, SHIFT(762),
  [244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2314),
  [246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(780),
  [248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2506),
  [250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2065),
  [252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(824),
  [254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2540),
  [256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2679),
  [258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_sizeof_expression, 5, 0, 67),
  [260] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_sizeof_expression, 5, 0, 67),
  [262] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_sizeof_expression, 5, 0, 85),
  [264] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_sizeof_expression, 5, 0, 85),
  [266] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_sizeof_expression, 6, 0, 85),
  [268] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_sizeof_expression, 6, 0, 85),
  [270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(136),
  [274] = {.entry = {.count = 1, .reusable = false}}, SHIFT(214),
  [276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2),
  [278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [284] = {.entry = {.count = 1, .reusable = false}}, SHIFT(221),
  [286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(222),
  [288] = {.entry = {.count = 1, .reusable = false}}, SHIFT(223),
  [290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(224),
  [294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(368),
  [302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2492),
  [304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2564),
  [306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2598),
  [308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2599),
  [310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2449),
  [312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [314] = {.entry = {.count = 1, .reusable = false}}, SHIFT(370),
  [316] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2569),
  [318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2593),
  [320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2594),
  [322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2460),
  [324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2563),
  [326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(371),
  [330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2549),
  [332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2550),
  [334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2551),
  [336] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2344),
  [338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(364),
  [342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2548),
  [344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2574),
  [346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2565),
  [348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2566),
  [350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2386),
  [352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2587),
  [354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1597),
  [356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2346),
  [358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2347),
  [360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2516),
  [362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(363),
  [366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2576),
  [368] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2570),
  [370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2571),
  [372] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2393),
  [374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(369),
  [378] = {.entry = {.count = 1, .reusable = false}}, SHIFT(881),
  [380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(772),
  [384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(773),
  [388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(737),
  [390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2301),
  [392] = {.entry = {.count = 1, .reusable = false}}, SHIFT(774),
  [394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2503),
  [396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2046),
  [398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(865),
  [400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2499),
  [402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2723),
  [404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(928),
  [406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(788),
  [410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(789),
  [414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(769),
  [416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2342),
  [418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(790),
  [420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2513),
  [422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2088),
  [424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(927),
  [426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2522),
  [428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2719),
  [430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(977),
  [432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(767),
  [436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(767),
  [438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(768),
  [440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(734),
  [442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2285),
  [444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(770),
  [446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2495),
  [448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2145),
  [450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(971),
  [452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2538),
  [454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2671),
  [456] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2645),
  [458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2738),
  [460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1001),
  [462] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1307),
  [464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1307),
  [466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__other_command, 2, 0, 3),
  [468] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1000),
  [470] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1002),
  [472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(841),
  [474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(760),
  [478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(760),
  [480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(761),
  [482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [484] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2270),
  [486] = {.entry = {.count = 1, .reusable = false}}, SHIFT(763),
  [488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2559),
  [490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2218),
  [492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1613),
  [494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2560),
  [496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2591),
  [498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2481),
  [500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1759),
  [502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1622),
  [504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1043),
  [506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2578),
  [508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(73),
  [510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2713),
  [512] = {.entry = {.count = 1, .reusable = false}}, SHIFT(856),
  [514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(42),
  [516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1621),
  [518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(582),
  [520] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(1221),
  [523] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(168),
  [526] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(218),
  [529] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(262),
  [532] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(262),
  [535] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(263),
  [538] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(1482),
  [541] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(2608),
  [544] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(2253),
  [547] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(1007),
  [550] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(2620),
  [553] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(2254),
  [556] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(2494),
  [559] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(1790),
  [562] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(5),
  [565] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(1112),
  [568] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(1397),
  [571] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(1471),
  [574] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(1362),
  [577] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(2586),
  [580] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(1397),
  [583] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(2678),
  [586] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(1216),
  [589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1003),
  [591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(44),
  [593] = {.entry = {.count = 1, .reusable = false}}, SHIFT(39),
  [595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(48),
  [597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(50),
  [599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(52),
  [601] = {.entry = {.count = 1, .reusable = false}}, SHIFT(54),
  [603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(56),
  [605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(58),
  [607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(60),
  [609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(62),
  [611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(64),
  [613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(66),
  [615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(68),
  [617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(70),
  [619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(72),
  [621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(75),
  [623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(77),
  [625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(79),
  [627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(890),
  [629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(735),
  [633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(735),
  [635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(736),
  [637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(737),
  [639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2456),
  [641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(738),
  [643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2545),
  [645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2195),
  [647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1168),
  [649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2588),
  [651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2711),
  [653] = {.entry = {.count = 1, .reusable = false}}, SHIFT(913),
  [655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(71),
  [657] = {.entry = {.count = 1, .reusable = false}}, SHIFT(775),
  [659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(775),
  [661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(777),
  [665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(938),
  [667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(785),
  [671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(785),
  [673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(786),
  [675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(769),
  [677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2339),
  [679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(787),
  [681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2510),
  [683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2076),
  [685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1309),
  [687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2501),
  [689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2617),
  [691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1443),
  [693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1443),
  [695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1263),
  [697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1263),
  [699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(96),
  [701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1233),
  [703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1233),
  [705] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1258),
  [707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1258),
  [709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(98),
  [711] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1170),
  [713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1170),
  [715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2370),
  [717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(13),
  [719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(13),
  [721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1483),
  [723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1483),
  [725] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1356),
  [727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1356),
  [729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(135),
  [731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(942),
  [733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(781),
  [737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(781),
  [739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [741] = {.entry = {.count = 1, .reusable = false}}, SHIFT(783),
  [743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1316),
  [745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1316),
  [747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1298),
  [749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1298),
  [751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(112),
  [753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1430),
  [755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1430),
  [757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1365),
  [759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1365),
  [761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1366),
  [763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1536),
  [765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1536),
  [767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1334),
  [769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1334),
  [771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(119),
  [773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1296),
  [775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1296),
  [777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1360),
  [779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1533),
  [781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1533),
  [783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1413),
  [785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1413),
  [787] = {.entry = {.count = 1, .reusable = false}}, SHIFT(116),
  [789] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1380),
  [791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1380),
  [793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1424),
  [795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1424),
  [797] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1425),
  [799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1425),
  [801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(118),
  [803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1426),
  [805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1426),
  [807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1317),
  [809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1317),
  [811] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1403),
  [813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1403),
  [815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(125),
  [817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1388),
  [819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1388),
  [821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(16),
  [823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1400),
  [827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1400),
  [829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(126),
  [831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1318),
  [833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1318),
  [835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1404),
  [837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1404),
  [839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1405),
  [841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1405),
  [843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(160),
  [845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1394),
  [847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1394),
  [849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1395),
  [851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1395),
  [853] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1484),
  [855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1484),
  [857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1409),
  [859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1409),
  [861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1410),
  [863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1410),
  [865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1489),
  [867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1489),
  [869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1412),
  [871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1412),
  [873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1398),
  [875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1398),
  [877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(450),
  [879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(450),
  [881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(139),
  [883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1399),
  [885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1399),
  [887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(100),
  [889] = {.entry = {.count = 1, .reusable = false}}, SHIFT(17),
  [891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(463),
  [895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(464),
  [899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(142),
  [903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1503),
  [905] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1598),
  [907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1598),
  [909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(467),
  [911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [913] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1442),
  [915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1442),
  [917] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1486),
  [919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1486),
  [921] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1456),
  [923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1456),
  [925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(18),
  [927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(18),
  [929] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1461),
  [931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1461),
  [933] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1462),
  [935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1462),
  [937] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1465),
  [939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1465),
  [941] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1467),
  [943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1467),
  [945] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1478),
  [947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1478),
  [949] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1477),
  [951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1477),
  [953] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1469),
  [955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1469),
  [957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1480),
  [959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1480),
  [961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(14),
  [963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(14),
  [965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1319),
  [967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1319),
  [969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(106),
  [971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(15),
  [973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1487),
  [977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1487),
  [979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1357),
  [981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1357),
  [983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1408),
  [985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1408),
  [987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1142),
  [989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1253),
  [991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1253),
  [993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(974),
  [995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [997] = {.entry = {.count = 1, .reusable = false}}, SHIFT(605),
  [999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(605),
  [1001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [1003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(734),
  [1005] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2462),
  [1007] = {.entry = {.count = 1, .reusable = false}}, SHIFT(739),
  [1009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [1011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1260),
  [1013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1260),
  [1015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1240),
  [1017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1259),
  [1019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1259),
  [1021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1427),
  [1023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1427),
  [1025] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1420),
  [1027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1420),
  [1029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1421),
  [1031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1421),
  [1033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1422),
  [1035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1422),
  [1037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1423),
  [1039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1423),
  [1041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1340),
  [1043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1340),
  [1045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1325),
  [1047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1325),
  [1049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1315),
  [1051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1315),
  [1053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1305),
  [1055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1305),
  [1057] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1354),
  [1059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1354),
  [1061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [1063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1301),
  [1065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1301),
  [1067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1339),
  [1069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1339),
  [1071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [1073] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1415),
  [1075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1415),
  [1077] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1401),
  [1079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1401),
  [1081] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1402),
  [1083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1402),
  [1085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1416),
  [1087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1416),
  [1089] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(1221),
  [1092] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(168),
  [1095] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(218),
  [1098] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(262),
  [1101] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(262),
  [1104] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(263),
  [1107] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(1482),
  [1110] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(2253),
  [1113] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(2620),
  [1116] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(2254),
  [1119] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(2494),
  [1122] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(1790),
  [1125] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(5),
  [1128] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(1112),
  [1131] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(1493),
  [1134] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(1471),
  [1137] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(1362),
  [1140] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(2586),
  [1143] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(1493),
  [1146] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(2678),
  [1149] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 30), SHIFT(1216),
  [1152] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(1221),
  [1155] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(168),
  [1158] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(218),
  [1161] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(262),
  [1164] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(262),
  [1167] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(263),
  [1170] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(1482),
  [1173] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(2253),
  [1176] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(2620),
  [1179] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(2254),
  [1182] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(2494),
  [1185] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(1790),
  [1188] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(5),
  [1191] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(1112),
  [1194] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(1493),
  [1197] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(1471),
  [1200] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(1362),
  [1203] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(2586),
  [1206] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(1493),
  [1209] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(2678),
  [1212] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), SHIFT(1216),
  [1215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1338),
  [1217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1338),
  [1219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(465),
  [1221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [1223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1232),
  [1225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1232),
  [1227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(466),
  [1229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [1231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [1233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1155),
  [1235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1155),
  [1237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1294),
  [1239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1294),
  [1241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1304),
  [1243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1304),
  [1245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1266),
  [1247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1266),
  [1249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(451),
  [1251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(451),
  [1253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1267),
  [1255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1267),
  [1257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1268),
  [1259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1268),
  [1261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1231),
  [1263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1231),
  [1265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1428),
  [1267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1428),
  [1269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(453),
  [1271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(453),
  [1273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1414),
  [1275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1414),
  [1277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(455),
  [1279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(455),
  [1281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(456),
  [1283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(456),
  [1285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(457),
  [1287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(457),
  [1289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(458),
  [1291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(458),
  [1293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(487),
  [1295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [1297] = {.entry = {.count = 1, .reusable = false}}, SHIFT(461),
  [1299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [1301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(462),
  [1303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(462),
  [1305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1269),
  [1307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1269),
  [1309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1310),
  [1311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1310),
  [1313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1270),
  [1315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1270),
  [1317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1271),
  [1319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1271),
  [1321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1371),
  [1323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1371),
  [1325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1272),
  [1327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1272),
  [1329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1370),
  [1331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1370),
  [1333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1254),
  [1335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1254),
  [1337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1257),
  [1339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1257),
  [1341] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1326),
  [1343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1326),
  [1345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1417),
  [1347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1417),
  [1349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1418),
  [1351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1418),
  [1353] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(1221),
  [1357] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(168),
  [1361] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(218),
  [1365] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(262),
  [1369] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(262),
  [1373] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(263),
  [1377] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(1482),
  [1381] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(2253),
  [1385] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(2620),
  [1389] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(2254),
  [1393] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(2494),
  [1397] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(1790),
  [1401] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(5),
  [1405] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(1112),
  [1409] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(1493),
  [1413] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(1471),
  [1417] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(1362),
  [1421] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(2586),
  [1425] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(1493),
  [1429] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(2678),
  [1433] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 31), REDUCE(sym_memory_space, 4, 0, 61), SHIFT(1216),
  [1437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1419),
  [1439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1419),
  [1441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1458),
  [1443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1458),
  [1445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1459),
  [1447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1459),
  [1449] = {.entry = {.count = 1, .reusable = false}}, SHIFT(454),
  [1451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(454),
  [1453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2029),
  [1455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [1457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2018),
  [1459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1604),
  [1461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(372),
  [1463] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0),
  [1465] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2070),
  [1468] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2648),
  [1471] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2514),
  [1474] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2596),
  [1477] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2504),
  [1480] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2517),
  [1483] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2305),
  [1486] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2490),
  [1489] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2478),
  [1492] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(1599),
  [1495] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2227),
  [1498] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2258),
  [1501] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2508),
  [1504] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2730),
  [1507] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2638),
  [1510] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2066),
  [1513] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(268),
  [1516] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(361),
  [1519] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_script_repeat1, 2, 0, 0), SHIFT_REPEAT(2731),
  [1522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1985),
  [1524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [1526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2022),
  [1528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2027),
  [1530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [1532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(365),
  [1534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2014),
  [1536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [1538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(367),
  [1540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1984),
  [1542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [1544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [1546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2028),
  [1548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2006),
  [1550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2024),
  [1552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2649),
  [1554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [1556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(351),
  [1558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2664),
  [1560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1962),
  [1562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [1564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(362),
  [1566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1963),
  [1568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2040),
  [1570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [1572] = {.entry = {.count = 1, .reusable = false}}, SHIFT(366),
  [1574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1934),
  [1576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1936),
  [1578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [1580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2659),
  [1582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(350),
  [1584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1939),
  [1586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [1588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [1590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1942),
  [1592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [1594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1946),
  [1596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1947),
  [1598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1951),
  [1600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [1602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1968),
  [1604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [1606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1971),
  [1608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2016),
  [1610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [1612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1969),
  [1614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2017),
  [1616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1972),
  [1618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [1620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1975),
  [1622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2705),
  [1624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [1626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(349),
  [1628] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_script, 1, 0, 0),
  [1630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [1632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2644),
  [1634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(354),
  [1636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [1638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [1640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2615),
  [1642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(355),
  [1644] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2070),
  [1647] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2648),
  [1650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0),
  [1652] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2514),
  [1655] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2549),
  [1658] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2550),
  [1661] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2551),
  [1664] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2344),
  [1667] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2563),
  [1670] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2478),
  [1673] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(1599),
  [1676] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2227),
  [1679] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2258),
  [1682] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2508),
  [1685] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2730),
  [1688] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2638),
  [1691] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2066),
  [1694] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(310),
  [1697] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(372),
  [1700] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2731),
  [1703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2619),
  [1705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [1707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(352),
  [1709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [1711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [1713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [1715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2037),
  [1717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [1719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2038),
  [1721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [1723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [1725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [1727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [1729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [1731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [1733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [1735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [1737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [1739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1976),
  [1741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [1743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [1745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1602),
  [1747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [1749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [1751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [1753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [1755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [1757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2616),
  [1759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1601),
  [1761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2632),
  [1763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1785),
  [1765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1786),
  [1767] = {.entry = {.count = 1, .reusable = false}}, SHIFT(374),
  [1769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2702),
  [1771] = {.entry = {.count = 1, .reusable = false}}, SHIFT(377),
  [1773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1381),
  [1775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1383),
  [1777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(376),
  [1779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(378),
  [1781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(375),
  [1783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(373),
  [1785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2676),
  [1787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [1789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(446),
  [1791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1740),
  [1793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1160),
  [1795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(403),
  [1797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [1799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1679),
  [1801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(394),
  [1803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [1805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2179),
  [1807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(397),
  [1809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(477),
  [1811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1637),
  [1813] = {.entry = {.count = 1, .reusable = false}}, SHIFT(393),
  [1815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [1817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1734),
  [1819] = {.entry = {.count = 1, .reusable = false}}, SHIFT(399),
  [1821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(469),
  [1823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1958),
  [1825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(398),
  [1827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(480),
  [1829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1720),
  [1831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(404),
  [1833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(473),
  [1835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1959),
  [1837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(396),
  [1839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [1841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2087),
  [1843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(400),
  [1845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(483),
  [1847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1987),
  [1849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(392),
  [1851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(448),
  [1853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2001),
  [1855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(401),
  [1857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(842),
  [1859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [1861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1627),
  [1863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(395),
  [1865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(449),
  [1867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2002),
  [1869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [1871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(478),
  [1873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [1875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1640),
  [1877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(482),
  [1879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(481),
  [1881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [1883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1982),
  [1885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(445),
  [1887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [1889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2050),
  [1891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [1893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(447),
  [1895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1664),
  [1897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [1899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1707),
  [1901] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro, 2, 0, 7),
  [1903] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro, 2, 0, 7),
  [1905] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0),
  [1907] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0),
  [1909] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0), REDUCE(sym__address_expression, 1, 0, 0),
  [1912] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0), REDUCE(sym__address_expression, 1, 0, 0),
  [1915] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__address_expression, 1, 0, 0),
  [1917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__address_expression, 1, 0, 0),
  [1919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(568),
  [1921] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro, 4, 0, 35),
  [1923] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro, 4, 0, 35),
  [1925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1292),
  [1927] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_integer, 2, 0, 0),
  [1929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_integer, 2, 0, 0),
  [1931] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_var_call_expression, 2, 0, 28),
  [1933] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_call_expression, 2, 0, 28),
  [1935] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__literal, 1, 0, 0),
  [1937] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__literal, 1, 0, 0),
  [1939] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__literal, 1, 0, 0), REDUCE(sym__address_expression, 1, 0, 0),
  [1942] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__literal, 1, 0, 0), REDUCE(sym__address_expression, 1, 0, 0),
  [1945] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_integer, 1, 0, 0),
  [1947] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_integer, 1, 0, 0),
  [1949] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__address_expression, 1, 0, 11),
  [1951] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__address_expression, 1, 0, 11),
  [1953] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 2, 0, 0),
  [1955] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 2, 0, 0),
  [1957] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_var_call_expression_argument_list, 3, 0, 0),
  [1959] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_var_call_expression_argument_list, 3, 0, 0),
  [1961] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__address_binary_expression, 3, 0, 38),
  [1963] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__address_binary_expression, 3, 0, 38),
  [1965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [1967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(179),
  [1969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [1971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(181),
  [1973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(183),
  [1975] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_address, 2, 0, 26),
  [1977] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_address, 2, 0, 26),
  [1979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(182),
  [1981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1114),
  [1983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [1985] = {.entry = {.count = 1, .reusable = false}}, SHIFT(202),
  [1987] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 3, 0, 0),
  [1989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 3, 0, 0),
  [1991] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(938),
  [1994] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(76),
  [1997] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(785),
  [2000] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(785),
  [2003] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(786),
  [2006] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(769),
  [2009] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(2608),
  [2012] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(2339),
  [2015] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(787),
  [2018] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(2510),
  [2021] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(2076),
  [2024] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(1309),
  [2027] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(2501),
  [2030] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(2617),
  [2033] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_argument_list, 4, 0, 0),
  [2035] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument_list, 4, 0, 0),
  [2037] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_address, 3, 0, 43),
  [2039] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_address, 2, 0, 32), REDUCE(sym_address, 3, 0, 43),
  [2042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_address, 3, 0, 43),
  [2044] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_address, 2, 0, 32), REDUCE(sym_address, 3, 0, 43),
  [2047] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_call_expression, 1, 0, 12),
  [2049] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_call_expression, 1, 0, 12),
  [2051] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__parenthesized_address_expression, 3, 0, 0),
  [2053] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__parenthesized_address_expression, 3, 0, 0),
  [2055] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_call_expression, 2, 0, 28),
  [2057] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_call_expression, 2, 0, 28),
  [2059] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_address, 2, 0, 32),
  [2061] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_address, 2, 0, 32),
  [2063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1997),
  [2065] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2331),
  [2067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(992),
  [2069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2044),
  [2071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2472),
  [2073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1980),
  [2075] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2221),
  [2077] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__path_expression_repeat1, 2, 0, 0),
  [2079] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__path_expression_repeat1, 2, 0, 0),
  [2081] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__path_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(2562),
  [2084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1635),
  [2086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2369),
  [2088] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__path_expression_repeat2, 2, 0, 0),
  [2090] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__path_expression_repeat2, 2, 0, 0),
  [2092] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__path_expression_repeat2, 2, 0, 0), SHIFT_REPEAT(2284),
  [2095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1738),
  [2097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2375),
  [2099] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__path_expression, 1, 0, 0),
  [2101] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__path_expression, 1, 0, 0),
  [2103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(516),
  [2105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(900),
  [2107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1709),
  [2109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2379),
  [2111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1663),
  [2113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2380),
  [2115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1661),
  [2117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2381),
  [2119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1957),
  [2121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2239),
  [2123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2012),
  [2125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2378),
  [2127] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment_expression, 3, 0, 38),
  [2129] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 3, 0, 38),
  [2131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2224),
  [2133] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binary_expression, 3, 0, 38),
  [2135] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binary_expression, 3, 0, 38),
  [2137] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__binary_expression, 3, 0, 38), SHIFT(2492),
  [2140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(949),
  [2142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1714),
  [2144] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2288),
  [2146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__and_expression, 4, 0, 53),
  [2148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__and_expression, 4, 0, 53),
  [2150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment_expression, 4, 0, 52),
  [2152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 4, 0, 52),
  [2154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment_expression, 4, 0, 53),
  [2156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 4, 0, 53),
  [2158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__unary_expression, 2, 0, 25),
  [2160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__unary_expression, 2, 0, 25),
  [2162] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__prefix_postfix_expression, 2, 0, 25),
  [2164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__prefix_postfix_expression, 2, 0, 25),
  [2166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment_expression, 5, 0, 72),
  [2168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 5, 0, 72),
  [2170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(878),
  [2172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(444),
  [2174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1981),
  [2176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2223),
  [2178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2086),
  [2180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2302),
  [2182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [2184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(436),
  [2186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(811),
  [2188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(475),
  [2190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1983),
  [2192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2230),
  [2194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1639),
  [2196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2382),
  [2198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(825),
  [2200] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0), REDUCE(sym__function_identifier, 1, 0, 9),
  [2203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2700),
  [2205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1633),
  [2207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2306),
  [2209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1698),
  [2211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2272),
  [2213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1735),
  [2215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2392),
  [2217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1704),
  [2219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2357),
  [2221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2045),
  [2223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2470),
  [2225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1999),
  [2227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2340),
  [2229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2000),
  [2231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2345),
  [2233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(932),
  [2235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(459),
  [2237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2047),
  [2239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2364),
  [2241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1638),
  [2243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2372),
  [2245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(497),
  [2247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(529),
  [2249] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__path_expression, 2, 0, 0),
  [2251] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__path_expression, 2, 0, 0),
  [2253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(489),
  [2255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(502),
  [2257] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 2, 0, 0),
  [2259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 2, 0, 0),
  [2261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(542),
  [2263] = {.entry = {.count = 1, .reusable = false}}, SHIFT(488),
  [2265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(537),
  [2267] = {.entry = {.count = 1, .reusable = false}}, SHIFT(493),
  [2269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(513),
  [2271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(551),
  [2273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(527),
  [2275] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 3, 0, 0),
  [2277] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 3, 0, 0),
  [2279] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_symbol, 2, 0, 0),
  [2281] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_symbol, 2, 0, 0),
  [2283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(559),
  [2285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(508),
  [2287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(511),
  [2289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(531),
  [2291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(534),
  [2293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(538),
  [2295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(536),
  [2297] = {.entry = {.count = 1, .reusable = false}}, SHIFT(541),
  [2299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(498),
  [2301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(546),
  [2303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(506),
  [2305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(525),
  [2307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(496),
  [2309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [2311] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trace32_hll_variable, 2, 0, 7),
  [2313] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trace32_hll_variable, 2, 0, 7),
  [2315] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__parenthesized_expression, 3, 0, 0),
  [2317] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__parenthesized_expression, 3, 0, 0),
  [2319] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_symbol, 3, 0, 0),
  [2321] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_symbol, 3, 0, 0),
  [2323] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bitmask, 1, 0, 0),
  [2325] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bitmask, 1, 0, 0),
  [2327] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_character, 3, 0, 0),
  [2329] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_character, 3, 0, 0),
  [2331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(589),
  [2333] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_expression, 1, 0, 10),
  [2335] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_expression, 1, 0, 10),
  [2337] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 1, 0, 11),
  [2339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 1, 0, 11),
  [2341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(784),
  [2343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [2345] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__special_command_operator, 1, 0, 0),
  [2347] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__special_command_operator, 1, 0, 0),
  [2349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(695),
  [2351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(729),
  [2353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(647),
  [2355] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_symbol, 1, 0, 13),
  [2357] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_symbol, 1, 0, 13),
  [2359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(627),
  [2361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(587),
  [2363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [2365] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT(1770),
  [2368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0),
  [2370] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0),
  [2372] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT(2560),
  [2375] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT(2591),
  [2378] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT(2481),
  [2381] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT(1759),
  [2384] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT(1620),
  [2387] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(842),
  [2390] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_symbol, 1, 0, 0),
  [2392] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_symbol, 1, 0, 0),
  [2394] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(942),
  [2397] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(74),
  [2400] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(781),
  [2403] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(781),
  [2406] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(782),
  [2409] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(769),
  [2412] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(2339),
  [2415] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(783),
  [2418] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(2510),
  [2421] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(2076),
  [2424] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(1309),
  [2427] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(2501),
  [2430] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(2617),
  [2433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_unary_expression, 2, 0, 25),
  [2435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [2437] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_unary_expression, 2, 0, 25),
  [2439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2519),
  [2441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2519),
  [2443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(579),
  [2445] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_update_expression, 2, 0, 25),
  [2447] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_update_expression, 2, 0, 25),
  [2449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(822),
  [2451] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_sizeof_expression, 2, 0, 33),
  [2453] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_sizeof_expression, 2, 0, 33),
  [2455] = {.entry = {.count = 1, .reusable = false}}, SHIFT(659),
  [2457] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_binary_expression, 3, 0, 38),
  [2459] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_binary_expression, 3, 0, 38),
  [2461] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_cast_expression, 4, 0, 64),
  [2463] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_cast_expression, 4, 0, 64),
  [2465] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_cast_expression, 5, 0, 78),
  [2467] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_cast_expression, 5, 0, 78),
  [2469] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_cast_expression, 5, 0, 82),
  [2471] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_cast_expression, 5, 0, 82),
  [2473] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_cast_expression, 6, 0, 96),
  [2475] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_cast_expression, 6, 0, 96),
  [2477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(654),
  [2479] = {.entry = {.count = 1, .reusable = false}}, SHIFT(652),
  [2481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(655),
  [2483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(656),
  [2485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(657),
  [2487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(658),
  [2489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(658),
  [2491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [2493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(661),
  [2495] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_assignment_expression, 3, 0, 38),
  [2497] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_assignment_expression, 3, 0, 38),
  [2499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [2501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(552),
  [2503] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_conditional_expression, 4, 0, 68),
  [2505] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_conditional_expression, 4, 0, 68),
  [2507] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_assignment_expression, 4, 0, 52),
  [2509] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_assignment_expression, 4, 0, 52),
  [2511] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_assignment_expression, 4, 0, 53),
  [2513] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_assignment_expression, 4, 0, 53),
  [2515] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_conditional_expression, 5, 0, 86),
  [2517] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_conditional_expression, 5, 0, 86),
  [2519] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_assignment_expression, 5, 0, 72),
  [2521] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_assignment_expression, 5, 0, 72),
  [2523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [2525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [2527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [2529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [2531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [2533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [2535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [2537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [2539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [2541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [2543] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_pointer_expression, 2, 0, 25),
  [2545] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_pointer_expression, 2, 0, 25),
  [2547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(830),
  [2549] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_argument_list, 2, 0, 0),
  [2551] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_argument_list, 2, 0, 0),
  [2553] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_field_expression, 3, 0, 47),
  [2555] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_field_expression, 3, 0, 47),
  [2557] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_argument_list, 4, 0, 0),
  [2559] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_argument_list, 4, 0, 0),
  [2561] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_field_identifier, 1, 0, 46),
  [2563] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_field_identifier, 1, 0, 46),
  [2565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_argument_list, 5, 0, 0),
  [2567] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_argument_list, 5, 0, 0),
  [2569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_argument_list, 6, 0, 0),
  [2571] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_argument_list, 6, 0, 0),
  [2573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_expression, 1, 0, 0),
  [2575] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__hll_expression, 1, 0, 0), REDUCE(sym__hll_assignment_left_expression, 1, 0, 0),
  [2578] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_expression, 1, 0, 0),
  [2580] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__hll_expression, 1, 0, 0), REDUCE(sym__hll_assignment_left_expression, 1, 0, 0),
  [2583] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_expression, 2, 0, 0),
  [2585] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_expression, 2, 0, 0),
  [2587] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_string_literal, 2, 0, 0),
  [2589] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_string_literal, 2, 0, 0),
  [2591] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_update_expression, 2, 0, 27),
  [2593] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_update_expression, 2, 0, 27),
  [2595] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_parenthesized_expression, 4, 0, 0),
  [2597] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_parenthesized_expression, 4, 0, 0),
  [2599] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_argument_list, 3, 0, 0),
  [2601] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_argument_list, 3, 0, 0),
  [2603] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_subscript_expression, 4, 0, 69),
  [2605] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_subscript_expression, 4, 0, 69),
  [2607] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_parenthesized_expression, 5, 0, 0),
  [2609] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_parenthesized_expression, 5, 0, 0),
  [2611] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_char_literal, 3, 0, 0),
  [2613] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_char_literal, 3, 0, 0),
  [2615] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_string_literal, 3, 0, 0),
  [2617] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_string_literal, 3, 0, 0),
  [2619] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_call_expression, 2, 0, 28),
  [2621] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_call_expression, 2, 0, 28),
  [2623] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_parenthesized_expression, 3, 0, 0),
  [2625] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_parenthesized_expression, 3, 0, 0),
  [2627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(468),
  [2629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(615),
  [2631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(896),
  [2633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [2635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(617),
  [2637] = {.entry = {.count = 1, .reusable = false}}, SHIFT(618),
  [2639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(619),
  [2641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(620),
  [2643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(621),
  [2645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(621),
  [2647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(622),
  [2649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(623),
  [2651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(624),
  [2653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2496),
  [2655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(576),
  [2657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(585),
  [2659] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__hll_expression, 1, 0, 0), REDUCE(sym__hll_type_identifier, 1, 0, 23),
  [2662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_assignment_left_expression, 1, 0, 0),
  [2664] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__hll_expression, 1, 0, 0), REDUCE(sym__hll_type_identifier, 1, 0, 23),
  [2667] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_assignment_left_expression, 1, 0, 0),
  [2669] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_type_identifier, 1, 0, 23),
  [2671] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym__hll_expression, 1, 0, 0), REDUCE(sym__hll_assignment_left_expression, 1, 0, 0), REDUCE(sym__hll_type_identifier, 1, 0, 23),
  [2675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [2677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [2679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [2681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2507),
  [2683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(558),
  [2685] = {.entry = {.count = 1, .reusable = false}}, SHIFT(724),
  [2687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(717),
  [2689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(719),
  [2691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(721),
  [2693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(722),
  [2695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(723),
  [2697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(723),
  [2699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(726),
  [2701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(720),
  [2703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(725),
  [2705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(718),
  [2707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [2709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(929),
  [2711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(471),
  [2713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(594),
  [2715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(983),
  [2717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(596),
  [2719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(600),
  [2721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(600),
  [2723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(601),
  [2725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2484),
  [2727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(584),
  [2729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(595),
  [2731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(597),
  [2733] = {.entry = {.count = 1, .reusable = false}}, SHIFT(598),
  [2735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(599),
  [2737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(602),
  [2739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(603),
  [2741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(577),
  [2743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(970),
  [2745] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0),
  [2747] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0),
  [2749] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 2, 0, 0), SHIFT_REPEAT(1006),
  [2752] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 3, 0, 0),
  [2754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 3, 0, 0),
  [2756] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1000),
  [2759] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 1, 0, 0),
  [2761] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 1, 0, 0),
  [2763] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_arguments, 1, 0, 0),
  [2765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(996),
  [2767] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1002),
  [2770] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_arguments, 2, 0, 0),
  [2772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(994),
  [2774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(997),
  [2776] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_arguments, 3, 0, 0),
  [2778] = {.entry = {.count = 1, .reusable = false}}, SHIFT(999),
  [2780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1008),
  [2782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1009),
  [2784] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat1, 4, 0, 0),
  [2786] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat1, 4, 0, 0),
  [2788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1012),
  [2790] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1013),
  [2793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1028),
  [2795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1072),
  [2797] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_hll_type_descriptor_repeat2, 2, 0, 0),
  [2799] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_hll_type_descriptor_repeat2, 2, 0, 0),
  [2801] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_hll_type_descriptor_repeat2, 2, 0, 0), SHIFT_REPEAT(1160),
  [2804] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_hll_type_descriptor_repeat2, 2, 0, 0), SHIFT_REPEAT(2013),
  [2807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(922),
  [2809] = {.entry = {.count = 1, .reusable = false}}, SHIFT(800),
  [2811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(801),
  [2813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(802),
  [2815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(802),
  [2817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(803),
  [2819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [2821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(805),
  [2823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(806),
  [2825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(806),
  [2827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(807),
  [2829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(808),
  [2831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(809),
  [2833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(521),
  [2835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(563),
  [2837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1989),
  [2839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(749),
  [2841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1248),
  [2843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(750),
  [2845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(751),
  [2847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(751),
  [2849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(752),
  [2851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(753),
  [2853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(754),
  [2855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(755),
  [2857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(755),
  [2859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(756),
  [2861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(757),
  [2863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(758),
  [2865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [2867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1579),
  [2869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1579),
  [2871] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_argument_list_non_default_format, 4, 0, 0),
  [2873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1135),
  [2875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(978),
  [2877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2020),
  [2879] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0),
  [2881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 2, 0, 0),
  [2883] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_argument_list_default_format, 3, 0, 0),
  [2885] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1036),
  [2887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1993),
  [2889] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 3, 0, 0),
  [2891] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat2, 3, 0, 0),
  [2893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1928),
  [2895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(813),
  [2897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1977),
  [2899] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1988),
  [2901] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1948),
  [2903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(903),
  [2905] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1955),
  [2907] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_argument_list_non_default_format, 5, 0, 0),
  [2909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1120),
  [2911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_argument_list_default_format, 2, 0, 0),
  [2913] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1042),
  [2915] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1032),
  [2917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_argument_list_default_format, 1, 0, 0),
  [2919] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1051),
  [2921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_argument_list_non_default_format, 3, 0, 0),
  [2923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1137),
  [2925] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1123),
  [2927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [2929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [2931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2374),
  [2933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(976),
  [2935] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 29), REDUCE(sym_memory_space, 4, 0, 60),
  [2938] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 29), REDUCE(sym_memory_space, 4, 0, 60),
  [2941] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 29), REDUCE(sym_memory_space, 4, 0, 62),
  [2944] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 29), REDUCE(sym_memory_space, 4, 0, 62),
  [2947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(959),
  [2949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(572),
  [2951] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2233),
  [2953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(936),
  [2955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2242),
  [2957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(898),
  [2959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2394),
  [2961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(962),
  [2963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(901),
  [2965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1382),
  [2967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(984),
  [2969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2232),
  [2971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(186),
  [2973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [2975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [2977] = {.entry = {.count = 1, .reusable = false}}, SHIFT(188),
  [2979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(190),
  [2981] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 29),
  [2983] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 29),
  [2985] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2661),
  [2987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_hll_argument_list_repeat1, 2, 0, 0),
  [2989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2089),
  [2991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(189),
  [2993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(700),
  [2995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1353),
  [2997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [2999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(702),
  [3001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(702),
  [3003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(703),
  [3005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(704),
  [3007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(705),
  [3009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(706),
  [3011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(706),
  [3013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(707),
  [3015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [3017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(709),
  [3019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(580),
  [3021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1314),
  [3023] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__path_expression_repeat2, 2, 0, 0), SHIFT_REPEAT(2284),
  [3026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_hll_argument_list_repeat1, 3, 0, 0),
  [3028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2073),
  [3030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1230),
  [3032] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_memory_space, 2, 0, 29), REDUCE(sym_memory_space, 4, 0, 62), REDUCE(sym_memory_space, 6, 0, 91),
  [3036] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_memory_space, 2, 0, 29), REDUCE(sym_memory_space, 4, 0, 62), REDUCE(sym_memory_space, 6, 0, 91),
  [3040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [3042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2422),
  [3044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(972),
  [3046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2273),
  [3048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(877),
  [3050] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2286),
  [3052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(823),
  [3054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2432),
  [3056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1287),
  [3058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [3060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4),
  [3062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [3064] = {.entry = {.count = 1, .reusable = false}}, SHIFT(216),
  [3066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [3068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [3070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(231),
  [3072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [3074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [3076] = {.entry = {.count = 1, .reusable = false}}, SHIFT(235),
  [3078] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_option_expression, 4, 0, 73),
  [3080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(636),
  [3082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(638),
  [3084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(638),
  [3086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(639),
  [3088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(640),
  [3090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(641),
  [3092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(642),
  [3094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(642),
  [3096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(643),
  [3098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(645),
  [3100] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_type_qualifier, 1, 0, 0),
  [3102] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_qualifier, 1, 0, 0),
  [3104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(95),
  [3106] = {.entry = {.count = 1, .reusable = false}}, SHIFT(208),
  [3108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [3110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [3112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(240),
  [3114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(242),
  [3116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [3118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(243),
  [3120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2261),
  [3122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2502),
  [3124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(236),
  [3126] = {.entry = {.count = 1, .reusable = false}}, SHIFT(237),
  [3128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(637),
  [3130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(644),
  [3132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [3134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1242),
  [3136] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_option_expression, 4, 0, 73),
  [3138] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__path_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(2533),
  [3141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1146),
  [3143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [3145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2171),
  [3147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1322),
  [3149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1195),
  [3151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [3153] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_comma_expression, 3, 0, 63),
  [3155] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_comma_expression, 3, 0, 63),
  [3157] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__binary_expression, 3, 0, 38), SHIFT(2502),
  [3160] = {.entry = {.count = 1, .reusable = false}}, SHIFT(683),
  [3162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(685),
  [3164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(685),
  [3166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(689),
  [3168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [3170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(690),
  [3172] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__path_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(2524),
  [3175] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_block, 7, 0, 34),
  [3177] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_block, 7, 0, 34),
  [3179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2256),
  [3181] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_block, 7, 0, 34), SHIFT(2271),
  [3184] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_block, 7, 0, 34), SHIFT(2271),
  [3187] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_block, 5, 0, 34),
  [3189] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_block, 5, 0, 34),
  [3191] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_block, 5, 0, 34), SHIFT(2461),
  [3194] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_block, 5, 0, 34), SHIFT(2271),
  [3197] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_block, 5, 0, 34), SHIFT(2271),
  [3200] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_block, 6, 0, 34),
  [3202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_block, 6, 0, 34),
  [3204] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_block, 6, 0, 34), SHIFT(2461),
  [3207] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_block, 6, 0, 34), SHIFT(2271),
  [3210] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_block, 6, 0, 34), SHIFT(2271),
  [3213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [3215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [3217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [3219] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_block, 7, 0, 34), SHIFT(2461),
  [3222] = {.entry = {.count = 1, .reusable = false}}, SHIFT(156),
  [3224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(251),
  [3226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [3228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [3230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [3232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(175),
  [3234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(176),
  [3236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(177),
  [3238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [3240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(178),
  [3242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [3244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2469),
  [3246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2486),
  [3248] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__binary_expression, 3, 0, 38), SHIFT(2486),
  [3251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(684),
  [3253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(686),
  [3255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(687),
  [3257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(688),
  [3259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(691),
  [3261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(692),
  [3263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [3265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_argument, 1, 0, 0),
  [3267] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__command_argument, 1, 0, 0), SHIFT(2469),
  [3270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1367),
  [3272] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 3, 0, 0),
  [3274] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 3, 0, 0), SHIFT(2261),
  [3277] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 4, 0, 0),
  [3279] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 4, 0, 0), SHIFT(2261),
  [3282] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_block, 6, 0, 34), SHIFT(2451),
  [3285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2371),
  [3287] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_block, 7, 0, 34), SHIFT(2451),
  [3290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0),
  [3292] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT(2261),
  [3295] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_if_block, 5, 0, 34), SHIFT(2451),
  [3298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1299),
  [3300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1557),
  [3302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [3304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(740),
  [3306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(741),
  [3308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(743),
  [3310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(743),
  [3312] = {.entry = {.count = 1, .reusable = false}}, SHIFT(744),
  [3314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(745),
  [3316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(746),
  [3318] = {.entry = {.count = 1, .reusable = false}}, SHIFT(748),
  [3320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [3322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(759),
  [3324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(765),
  [3326] = {.entry = {.count = 1, .reusable = false}}, SHIFT(766),
  [3328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(567),
  [3330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1581),
  [3332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [3334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1491),
  [3336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1494),
  [3338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_range_expression, 3, 0, 38),
  [3340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1571),
  [3342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_expression, 2, 0, 8),
  [3344] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_expression, 2, 0, 8),
  [3346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1497),
  [3348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1509),
  [3350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_range_expression, 4, 0, 52),
  [3352] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_range_expression, 4, 0, 53),
  [3354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1529),
  [3356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statement, 2, 0, 0),
  [3358] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statement, 2, 0, 0),
  [3360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [3362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2231),
  [3364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_range_expression, 5, 0, 72),
  [3366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1534),
  [3368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [3370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statement, 1, 0, 0),
  [3372] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statement, 1, 0, 0),
  [3374] = {.entry = {.count = 1, .reusable = false}}, SHIFT(114),
  [3376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(220),
  [3378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8),
  [3380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [3382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [3384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [3386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(252),
  [3388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(258),
  [3390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(260),
  [3392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [3394] = {.entry = {.count = 1, .reusable = false}}, SHIFT(169),
  [3396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [3398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [3400] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__command_argument, 1, 0, 0), SHIFT(2397),
  [3403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2534),
  [3405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1584),
  [3407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1585),
  [3409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2130),
  [3411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(555),
  [3413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2069),
  [3415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2071),
  [3417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [3419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(794),
  [3421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2112),
  [3423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2132),
  [3425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2146),
  [3427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2161),
  [3429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2397),
  [3431] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__binary_expression, 3, 0, 38), SHIFT(2534),
  [3434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1411),
  [3436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [3438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1154),
  [3440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(592),
  [3442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1252),
  [3444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [3446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(631),
  [3448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(650),
  [3450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(677),
  [3452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [3454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(714),
  [3456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(732),
  [3458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [3460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [3462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [3464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(764),
  [3466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [3468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [3470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [3472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [3474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [3476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [3478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [3480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [3482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1594),
  [3484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [3486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [3488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [3490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2159),
  [3492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1770),
  [3494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1810),
  [3496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1013),
  [3498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_expression, 3, 0, 8),
  [3500] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_expression, 3, 0, 8),
  [3502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1500),
  [3504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 3, 0, 3),
  [3506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__repeat_block_pre_condition_format, 3, 0, 3),
  [3508] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__repeat_block_pre_condition_format, 3, 0, 3), SHIFT(2479),
  [3511] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 3, 0, 3), SHIFT(2383),
  [3514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recursive_macro_expansion, 3, 0, 0),
  [3516] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_recursive_macro_expansion, 3, 0, 0),
  [3518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1506),
  [3520] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1500),
  [3523] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recursive_macro_expansion, 2, 0, 0),
  [3525] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_recursive_macro_expansion, 2, 0, 0),
  [3527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1501),
  [3529] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 4, 0, 3),
  [3531] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__repeat_block_pre_condition_format, 4, 0, 3),
  [3533] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__repeat_block_pre_condition_format, 4, 0, 3), SHIFT(2521),
  [3536] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 4, 0, 3), SHIFT(2414),
  [3539] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recursive_macro_expansion, 4, 0, 0),
  [3541] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_recursive_macro_expansion, 4, 0, 0),
  [3543] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 5, 0, 3),
  [3545] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__repeat_block_pre_condition_format, 5, 0, 3),
  [3547] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__repeat_block_pre_condition_format, 5, 0, 3), SHIFT(2511),
  [3550] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 5, 0, 3), SHIFT(2407),
  [3553] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 4, 0, 36),
  [3555] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_declaration, 4, 0, 36),
  [3557] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_block, 4, 0, 3),
  [3559] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_block, 4, 0, 3),
  [3561] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_block, 5, 0, 34),
  [3563] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_block, 5, 0, 34),
  [3565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_expression, 2, 0, 4),
  [3567] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_expression, 2, 0, 4),
  [3569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_block, 9, 0, 34),
  [3571] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_block, 9, 0, 34),
  [3573] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_if_block_repeat2, 1, 0, 0), REDUCE(sym_if_block, 6, 0, 34),
  [3576] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_block_repeat2, 1, 0, 0), REDUCE(sym_if_block, 6, 0, 34),
  [3579] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_expression, 2, 0, 5),
  [3581] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_expression, 2, 0, 5),
  [3583] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_block_repeat2, 1, 0, 0),
  [3585] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_block_repeat2, 1, 0, 0), REDUCE(sym_if_block, 7, 0, 34),
  [3588] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_if_block_repeat2, 1, 0, 0), REDUCE(sym_if_block, 7, 0, 34),
  [3591] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__repeat_block_post_condition_format, 9, 1, 109),
  [3593] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__repeat_block_post_condition_format, 9, 1, 109),
  [3595] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_block, 7, 0, 34),
  [3597] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_block, 7, 0, 34),
  [3599] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_repeat_block, 1, 0, 2),
  [3601] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_repeat_block, 1, 0, 2),
  [3603] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__repeat_block_post_condition_format, 10, 1, 118),
  [3605] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__repeat_block_post_condition_format, 10, 1, 118),
  [3607] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_definition, 3, 0, 17),
  [3609] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_definition, 3, 0, 17),
  [3611] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elif_block, 7, 0, 124),
  [3613] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_elif_block, 7, 0, 124),
  [3615] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elif_block, 8, 0, 124),
  [3617] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_elif_block, 8, 0, 124),
  [3619] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elif_block, 9, 0, 124),
  [3621] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_elif_block, 9, 0, 124),
  [3623] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 4, 0, 34),
  [3625] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__repeat_block_pre_condition_format, 4, 0, 34),
  [3627] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_block, 5, 0, 37),
  [3629] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_block, 5, 0, 37),
  [3631] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_expression, 3, 0, 14),
  [3633] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_expression, 3, 0, 14),
  [3635] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_definition, 4, 0, 36),
  [3637] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_definition, 4, 0, 36),
  [3639] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_call_expression, 4, 0, 37),
  [3641] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_call_expression, 4, 0, 37),
  [3643] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_assignment_expression, 4, 0, 38),
  [3645] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_assignment_expression, 4, 0, 38),
  [3647] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 5, 0, 0),
  [3649] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 5, 0, 0),
  [3651] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_block, 3, 0, 3),
  [3653] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_block, 3, 0, 3),
  [3655] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_block, 8, 0, 34),
  [3657] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_block, 8, 0, 34),
  [3659] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_command_expression, 3, 0, 15),
  [3661] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_expression, 3, 0, 15),
  [3663] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 6, 0, 0),
  [3665] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 6, 0, 0),
  [3667] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_block_repeat2, 1, 0, 0), REDUCE(sym_if_block, 8, 0, 34),
  [3670] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_if_block_repeat2, 1, 0, 0), REDUCE(sym_if_block, 8, 0, 34),
  [3673] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 5, 0, 34),
  [3675] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__repeat_block_pre_condition_format, 5, 0, 34),
  [3677] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_block, 6, 0, 37),
  [3679] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_block, 6, 0, 37),
  [3681] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_call_expression, 5, 0, 48),
  [3683] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_call_expression, 5, 0, 48),
  [3685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__repeat_block_post_condition_format, 8, 1, 99),
  [3687] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__repeat_block_post_condition_format, 8, 1, 99),
  [3689] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 4, 0, 0),
  [3691] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 4, 0, 0),
  [3693] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 3, 0, 17),
  [3695] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_declaration, 3, 0, 17),
  [3697] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 5, 0, 49),
  [3699] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_declaration, 5, 0, 49),
  [3701] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_block, 6, 0, 34),
  [3703] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_block, 6, 0, 34),
  [3705] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0),
  [3707] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__hll_var_command_argument_list_default_format_repeat1, 2, 0, 0),
  [3709] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 6, 0, 34),
  [3711] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__repeat_block_pre_condition_format, 6, 0, 34),
  [3713] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__repeat_block_post_condition_format, 7, 1, 87),
  [3715] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__repeat_block_post_condition_format, 7, 1, 87),
  [3717] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_declaration, 6, 0, 70),
  [3719] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter_declaration, 6, 0, 70),
  [3721] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 7, 0, 34),
  [3723] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__repeat_block_pre_condition_format, 7, 0, 34),
  [3725] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_assignment_expression, 5, 0, 52),
  [3727] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_assignment_expression, 5, 0, 52),
  [3729] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_assignment_expression, 5, 0, 53),
  [3731] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_assignment_expression, 5, 0, 53),
  [3733] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_call_expression, 7, 0, 88),
  [3735] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_call_expression, 7, 0, 88),
  [3737] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 7, 0, 0),
  [3739] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 7, 0, 0),
  [3741] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subroutine_call_expression, 6, 0, 71),
  [3743] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subroutine_call_expression, 6, 0, 71),
  [3745] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_assignment_expression, 6, 0, 72),
  [3747] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_assignment_expression, 6, 0, 72),
  [3749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2008),
  [3751] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1769),
  [3753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2351),
  [3755] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2352),
  [3757] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2520),
  [3759] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2106),
  [3761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2388),
  [3763] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2255),
  [3765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2601),
  [3767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2124),
  [3769] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1601),
  [3772] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_block_repeat1, 2, 0, 0),
  [3774] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_block_repeat1, 2, 0, 0),
  [3776] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1602),
  [3779] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1604),
  [3782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1620),
  [3784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1614),
  [3786] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__label, 2, 0, 6),
  [3788] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__label, 2, 0, 6),
  [3790] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1764),
  [3792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1805),
  [3794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_pointer_declarator, 1, 0, 0),
  [3796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1612),
  [3798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1749),
  [3800] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2732),
  [3802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [3804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1610),
  [3806] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_pointer_declarator, 2, 0, 0),
  [3808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1715),
  [3810] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_parameter_declaration, 1, 0, 76),
  [3812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1615),
  [3814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1624),
  [3816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2512),
  [3818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2080),
  [3820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2653),
  [3822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2687),
  [3824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2716),
  [3826] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 2, 0, 24), REDUCE(sym_hll_type_descriptor, 3, 0, 41),
  [3829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1750),
  [3831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [3833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1613),
  [3835] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_hll_type_descriptor, 2, 0, 24), REDUCE(sym_hll_type_descriptor, 3, 0, 41), SHIFT(1650),
  [3839] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 2, 0, 24),
  [3841] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_type_descriptor, 2, 0, 24), SHIFT(1660),
  [3844] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 3, 0, 41),
  [3846] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_type_descriptor, 3, 0, 41), SHIFT(1652),
  [3849] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 2, 0, 41),
  [3851] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_type_descriptor, 2, 0, 41), SHIFT(1653),
  [3854] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 1, 0, 24), REDUCE(sym_hll_type_descriptor, 2, 0, 41),
  [3857] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym_hll_type_descriptor, 1, 0, 24), REDUCE(sym_hll_type_descriptor, 2, 0, 41), SHIFT(1657),
  [3861] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 1, 0, 24),
  [3863] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_type_descriptor, 1, 0, 24), SHIFT(1647),
  [3866] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 4, 0, 67),
  [3868] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_type_descriptor, 4, 0, 67), SHIFT(1654),
  [3871] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 3, 0, 67),
  [3873] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_type_descriptor, 3, 0, 67), SHIFT(1643),
  [3876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [3878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [3880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(255),
  [3882] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 3, 0, 0),
  [3884] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 3, 0, 0), SHIFT(1722),
  [3887] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 3, 0, 0), SHIFT(383),
  [3890] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 3, 0, 0), SHIFT(1662),
  [3893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(253),
  [3895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [3897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 104),
  [3899] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 104), SHIFT(1722),
  [3902] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 104), SHIFT(383),
  [3905] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 104), SHIFT(1675),
  [3908] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 3, 0, 79),
  [3910] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 3, 0, 79), SHIFT(1722),
  [3913] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 3, 0, 79), SHIFT(383),
  [3916] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 3, 0, 79), SHIFT(1742),
  [3919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 79),
  [3921] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 79), SHIFT(1722),
  [3924] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 79), SHIFT(383),
  [3927] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 79), SHIFT(1665),
  [3930] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 92),
  [3932] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 92), SHIFT(1722),
  [3935] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 92), SHIFT(383),
  [3938] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 92), SHIFT(1666),
  [3941] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 2, 0, 0),
  [3943] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 2, 0, 0), SHIFT(1722),
  [3946] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 2, 0, 0), SHIFT(383),
  [3949] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 2, 0, 0), SHIFT(1739),
  [3952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 93),
  [3954] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 93), SHIFT(1722),
  [3957] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 93), SHIFT(383),
  [3960] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 93), SHIFT(1668),
  [3963] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 104),
  [3965] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 104), SHIFT(1722),
  [3968] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 104), SHIFT(383),
  [3971] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 104), SHIFT(1673),
  [3974] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 0),
  [3976] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 0), SHIFT(1722),
  [3979] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 0), SHIFT(383),
  [3982] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 0), SHIFT(1670),
  [3985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 92),
  [3987] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 92), SHIFT(1722),
  [3990] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 92), SHIFT(383),
  [3993] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 92), SHIFT(1671),
  [3996] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 93),
  [3998] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 93), SHIFT(1722),
  [4001] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 93), SHIFT(383),
  [4004] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 93), SHIFT(1672),
  [4007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1144),
  [4009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [4011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1628),
  [4013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1229),
  [4015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [4017] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_hll_type_descriptor_repeat1, 2, 0, 0),
  [4019] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_hll_type_descriptor_repeat1, 2, 0, 0), SHIFT_REPEAT(1613),
  [4022] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__hll_declaration_specifiers_repeat1, 2, 0, 0),
  [4024] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__hll_declaration_specifiers_repeat1, 2, 0, 0), SHIFT_REPEAT(1613),
  [4027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1722),
  [4029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [4031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1673),
  [4033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1668),
  [4035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1670),
  [4037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1671),
  [4039] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1672),
  [4041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1675),
  [4043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1741),
  [4045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2663),
  [4047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [4049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1721),
  [4051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(193),
  [4053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [4055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [4057] = {.entry = {.count = 1, .reusable = false}}, SHIFT(197),
  [4059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [4061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [4063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1719),
  [4065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(549),
  [4067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1686),
  [4069] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1688),
  [4071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1677),
  [4073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(510),
  [4075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(500),
  [4077] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1700),
  [4079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(540),
  [4081] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1699),
  [4083] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1728),
  [4085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1077),
  [4087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(259),
  [4089] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1078),
  [4091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1729),
  [4093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(507),
  [4095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(545),
  [4097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [4099] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1702),
  [4101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(533),
  [4103] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1731),
  [4105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(509),
  [4107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1706),
  [4109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [4111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1737),
  [4113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(491),
  [4115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1744),
  [4117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1666),
  [4119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(512),
  [4121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1711),
  [4123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [4125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(532),
  [4127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1713),
  [4129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(535),
  [4131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1693),
  [4133] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1715),
  [4136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [4138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1717),
  [4140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [4142] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1665),
  [4144] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1703),
  [4146] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1723),
  [4148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [4150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1726),
  [4152] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1140),
  [4154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(490),
  [4156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1736),
  [4158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1680),
  [4160] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1682),
  [4162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1739),
  [4164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1683),
  [4166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1742),
  [4168] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1662),
  [4170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1743),
  [4172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(544),
  [4174] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1745),
  [4176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2326),
  [4178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1814),
  [4180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2710),
  [4182] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2597),
  [4184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2536),
  [4186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1772),
  [4188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2033),
  [4190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1887),
  [4192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2608),
  [4194] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_pointer_declarator_specifier, 1, 0, 0),
  [4196] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_pointer_declarator_specifier, 1, 0, 0),
  [4198] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__hll_pointer_declarator_specifier, 1, 0, 0), SHIFT(1893),
  [4201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1752),
  [4203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2459),
  [4205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2694),
  [4207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2581),
  [4209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2582),
  [4211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2583),
  [4213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2699),
  [4215] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hll_type_descriptor_repeat2, 2, 0, 0), SHIFT_REPEAT(1613),
  [4218] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_hll_type_descriptor_repeat2, 2, 0, 0), SHIFT_REPEAT(2007),
  [4221] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_pointer_declarator_specifier, 2, 0, 0),
  [4223] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_pointer_declarator_specifier, 2, 0, 0),
  [4225] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__hll_pointer_declarator_specifier_repeat1, 1, 0, 0),
  [4227] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__hll_pointer_declarator_specifier, 2, 0, 0), SHIFT(2072),
  [4230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_pointer_declarator_specifier, 4, 0, 0),
  [4232] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_pointer_declarator_specifier, 4, 0, 0),
  [4234] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__hll_pointer_declarator_specifier, 4, 0, 0), SHIFT(2072),
  [4237] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_sized_type_specifier, 2, 0, 0),
  [4239] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_sized_type_specifier, 2, 0, 0),
  [4241] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_sized_type_specifier, 2, 0, 0), SHIFT(1803),
  [4244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_pointer_declarator_specifier, 3, 0, 0),
  [4246] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_pointer_declarator_specifier, 3, 0, 0),
  [4248] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__hll_pointer_declarator_specifier, 3, 0, 0), SHIFT(2072),
  [4251] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_sized_type_specifier, 1, 0, 0),
  [4253] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_sized_type_specifier, 1, 0, 0),
  [4255] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_hll_sized_type_specifier, 1, 0, 0), SHIFT(1803),
  [4258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_declaration_specifiers, 3, 0, 41),
  [4260] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_declaration_specifiers, 3, 0, 41),
  [4262] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1888),
  [4264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_declaration_specifiers, 2, 0, 24),
  [4266] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_declaration_specifiers, 2, 0, 24),
  [4268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_sized_type_specifier, 2, 0, 41),
  [4270] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_sized_type_specifier, 2, 0, 41),
  [4272] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_declaration_specifiers, 2, 0, 41),
  [4274] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_declaration_specifiers, 2, 0, 41),
  [4276] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__hll_declaration_specifiers_repeat2, 2, 0, 0),
  [4278] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__hll_declaration_specifiers_repeat2, 2, 0, 0),
  [4280] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__hll_declaration_specifiers_repeat2, 2, 0, 0), SHIFT_REPEAT(1888),
  [4283] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_struct_specifier, 3, 0, 55),
  [4285] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_struct_specifier, 3, 0, 55),
  [4287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2452),
  [4289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1868),
  [4291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1797),
  [4293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2039),
  [4295] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_enum_specifier, 3, 0, 55),
  [4297] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_enum_specifier, 3, 0, 55),
  [4299] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_declaration_specifiers, 1, 0, 24),
  [4301] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_declaration_specifiers, 1, 0, 24),
  [4303] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_type_identifier, 1, 0, 23),
  [4305] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_union_specifier, 3, 0, 55),
  [4307] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_union_specifier, 3, 0, 55),
  [4309] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1772),
  [4312] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_block, 6, 0, 34), SHIFT(2423),
  [4315] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_block, 5, 0, 34), SHIFT(2423),
  [4318] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_block, 7, 0, 34), SHIFT(2423),
  [4321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2192),
  [4323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2269),
  [4325] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1256),
  [4327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1787),
  [4329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1787),
  [4331] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [4333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(850),
  [4335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1781),
  [4337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1781),
  [4339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_parameter_declaration, 2, 0, 40),
  [4341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1596),
  [4343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [4345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1949),
  [4347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(834),
  [4349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1816),
  [4351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1816),
  [4353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1375),
  [4355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1800),
  [4357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1800),
  [4359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(716),
  [4361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1784),
  [4363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1784),
  [4365] = {.entry = {.count = 1, .reusable = false}}, SHIFT(727),
  [4367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1824),
  [4369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1825),
  [4371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1276),
  [4373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(910),
  [4375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1789),
  [4377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1789),
  [4379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(911),
  [4381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1285),
  [4383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1793),
  [4385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1793),
  [4387] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_block, 5, 0, 34), SHIFT(2389),
  [4390] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__hll_declaration_specifiers, 3, 0, 41), SHIFT(1888),
  [4393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1343),
  [4395] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_block, 6, 0, 34), SHIFT(2389),
  [4398] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_pointer_declarator, 2, 0, 58),
  [4400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(499),
  [4402] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1818),
  [4404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1818),
  [4406] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2552),
  [4408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1808),
  [4410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1808),
  [4412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2498),
  [4414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1429),
  [4416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1132),
  [4418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1802),
  [4420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1802),
  [4422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1100),
  [4424] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_hll_sized_type_specifier_repeat1, 2, 0, 0),
  [4426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1014),
  [4428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1806),
  [4430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1806),
  [4432] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__hll_declaration_specifiers, 2, 0, 41), SHIFT(1888),
  [4435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1016),
  [4437] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__hll_declaration_specifiers, 2, 0, 24), SHIFT(1888),
  [4440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2509),
  [4442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(941),
  [4444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1811),
  [4446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1811),
  [4448] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__hll_declaration_specifiers, 1, 0, 24), SHIFT(1888),
  [4451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(920),
  [4453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1757),
  [4455] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_if_block, 7, 0, 34), SHIFT(2389),
  [4458] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2537),
  [4460] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1798),
  [4462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1798),
  [4464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_pointer_declarator, 3, 0, 75),
  [4466] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_repeat1, 2, 0, 0), SHIFT_REPEAT(2269),
  [4469] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2, 0, 0),
  [4471] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1816),
  [4474] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1816),
  [4477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(517),
  [4479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2220),
  [4481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1973),
  [4483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1973),
  [4485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2473),
  [4487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2475),
  [4489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2185),
  [4491] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2185),
  [4493] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1823),
  [4496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1823),
  [4498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1600),
  [4500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 7, 0, 119),
  [4502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2172),
  [4504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 4, 0, 83),
  [4506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 6, 0, 110),
  [4508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [4510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2411),
  [4512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2471),
  [4514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 6, 0, 111),
  [4516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2389),
  [4518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2477),
  [4520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 4, 0, 89),
  [4522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2101),
  [4524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 4, 0, 90),
  [4526] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 4, 0, 81), REDUCE(sym_hll_type_descriptor, 5, 0, 97),
  [4529] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 5, 0, 97),
  [4531] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 5, 0, 98),
  [4533] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 3, 0, 75),
  [4535] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 7, 0, 121),
  [4537] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 5, 0, 100),
  [4539] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 5, 0, 101),
  [4541] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_declaration, 2, 0, 40),
  [4543] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 5, 0, 102),
  [4545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1954),
  [4547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [4549] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 5, 0, 103),
  [4551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1891),
  [4553] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 3, 0, 65),
  [4555] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_declaration, 3, 0, 59),
  [4557] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 6, 0, 112),
  [4559] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 6, 0, 113),
  [4561] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 6, 0, 114),
  [4563] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 7, 0, 120),
  [4565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_array_declarator, 8, 0, 123),
  [4567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2032),
  [4569] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_pattern_command_argument_list_format_repeat2, 2, 0, 0), SHIFT_REPEAT(2064),
  [4572] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_pattern_command_argument_list_format_repeat2, 2, 0, 0), SHIFT_REPEAT(1990),
  [4575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2136),
  [4577] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__host_command_argument_list_default_format, 1, 0, 0),
  [4579] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1777),
  [4581] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hll_type_pattern_repeat1, 2, 0, 0), SHIFT_REPEAT(2033),
  [4584] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hll_type_pattern_repeat1, 2, 0, 0), SHIFT_REPEAT(1871),
  [4587] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_hll_type_pattern_repeat1, 2, 0, 0),
  [4589] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_hll_type_pattern_repeat1, 2, 0, 0),
  [4591] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_hll_sized_type_specifier_repeat1, 2, 0, 0), SHIFT_REPEAT(2539),
  [4594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 4, 0, 81),
  [4596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 6, 0, 108),
  [4598] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_pointer_declarator, 2, 0, 58),
  [4600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1965),
  [4602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1871),
  [4604] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_pattern, 1, 0, 0),
  [4606] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_type_pattern, 1, 0, 0),
  [4608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1978),
  [4610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 4, 0, 84),
  [4612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_pointer_declarator, 3, 0, 75),
  [4614] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__hll_var_pattern_command_argument_list_format_repeat1, 2, 0, 0),
  [4616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2474),
  [4618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2423),
  [4620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2465),
  [4622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2451),
  [4624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 2, 0, 45),
  [4626] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 3, 0, 66),
  [4628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1992),
  [4630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1850),
  [4632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2461),
  [4634] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_var_pattern_command_argument_list_format_repeat1, 2, 0, 0), SHIFT_REPEAT(2608),
  [4637] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 2, 0, 45), REDUCE(sym_hll_type_descriptor, 3, 0, 66),
  [4640] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_hll_type_descriptor, 3, 0, 65), REDUCE(sym_hll_type_descriptor, 4, 0, 83),
  [4643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2059),
  [4645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1960),
  [4647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2118),
  [4649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1935),
  [4651] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2413),
  [4653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2635),
  [4655] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_format_expression, 2, 0, 33),
  [4657] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_format_expression, 2, 0, 33),
  [4659] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat2, 2, 0, 0),
  [4661] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1991),
  [4663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2402),
  [4665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2401),
  [4667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1189),
  [4669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2395),
  [4671] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__var_command_argument_list_non_hll_format_repeat1, 2, 0, 0), SHIFT_REPEAT(2725),
  [4674] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__var_command_argument_list_non_hll_format_repeat1, 2, 0, 0), SHIFT_REPEAT(2608),
  [4677] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_format_expression_repeat1, 2, 0, 74), SHIFT_REPEAT(2635),
  [4680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_format_expression_repeat1, 2, 0, 74),
  [4682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_format_expression_repeat1, 2, 0, 74),
  [4684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2117),
  [4686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1952),
  [4688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [4690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_function_declarator, 2, 0, 57),
  [4692] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_function_declarator, 2, 0, 57),
  [4694] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 4, 0, 3), SHIFT(2526),
  [4697] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 4, 0, 3), SHIFT(2439),
  [4700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2041),
  [4702] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1956),
  [4704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_parenthesized_declarator, 4, 0, 0),
  [4706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_parenthesized_declarator, 4, 0, 0),
  [4708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(904),
  [4710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2147),
  [4712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 94),
  [4714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 94),
  [4716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 80),
  [4718] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 80),
  [4720] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 95),
  [4722] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 95),
  [4724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2215),
  [4726] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1930),
  [4728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2526),
  [4730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1966),
  [4732] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(111),
  [4735] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2395),
  [4738] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_block_repeat2, 2, 0, 0), SHIFT_REPEAT(2488),
  [4741] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_block_repeat2, 2, 0, 0), SHIFT_REPEAT(1973),
  [4744] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_if_block_repeat2, 2, 0, 0), SHIFT_REPEAT(1973),
  [4747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1096),
  [4749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(816),
  [4751] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_parenthesized_declarator, 5, 0, 0),
  [4753] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_parenthesized_declarator, 5, 0, 0),
  [4755] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 94),
  [4757] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 94),
  [4759] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 105),
  [4761] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 105),
  [4763] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 106),
  [4765] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 106),
  [4767] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 80),
  [4769] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 80),
  [4771] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 107),
  [4773] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 107),
  [4775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2527),
  [4777] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 5, 0, 3), SHIFT(2440),
  [4780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 95),
  [4782] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 95),
  [4784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [4786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2064),
  [4788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2004),
  [4790] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1991),
  [4793] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_parenthesized_declarator, 3, 0, 0),
  [4795] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_parenthesized_declarator, 3, 0, 0),
  [4797] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 105),
  [4799] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 105),
  [4801] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 106),
  [4803] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 106),
  [4805] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 115),
  [4807] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 115),
  [4809] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 107),
  [4811] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 107),
  [4813] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 116),
  [4815] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 116),
  [4817] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 117),
  [4819] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 117),
  [4821] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 3, 0, 80),
  [4823] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 3, 0, 80),
  [4825] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 95),
  [4827] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 95),
  [4829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2416),
  [4831] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hll_type_pattern_repeat1, 2, 0, 0), SHIFT_REPEAT(2064),
  [4834] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hll_type_pattern_repeat1, 2, 0, 0), SHIFT_REPEAT(2004),
  [4837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_function_declarator, 3, 0, 77),
  [4839] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_function_declarator, 3, 0, 77),
  [4841] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_parameter_list, 2, 0, 0),
  [4843] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_parameter_list, 2, 0, 0),
  [4845] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 7, 0, 115),
  [4847] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 7, 0, 115),
  [4849] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 7, 0, 116),
  [4851] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 7, 0, 116),
  [4853] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 7, 0, 117),
  [4855] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 7, 0, 117),
  [4857] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 7, 0, 122),
  [4859] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 7, 0, 122),
  [4861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [4863] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_abstract_array_declarator, 8, 0, 122),
  [4865] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 8, 0, 122),
  [4867] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_parameter_list, 3, 0, 0),
  [4869] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_parameter_list, 3, 0, 0),
  [4871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(982),
  [4873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_parameter_list, 4, 0, 0),
  [4875] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_parameter_list, 4, 0, 0),
  [4877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2438),
  [4879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2359),
  [4881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2365),
  [4883] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 5, 0, 3), SHIFT(2527),
  [4886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2523),
  [4888] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 3, 0, 3), SHIFT(2435),
  [4891] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_hll_type_pattern_repeat1, 1, 0, 51),
  [4893] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_hll_type_pattern_repeat1, 1, 0, 51),
  [4895] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__repeat_block_pre_condition_format, 3, 0, 3), SHIFT(2523),
  [4898] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_format_expression, 3, 0, 54),
  [4900] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_format_expression, 3, 0, 54),
  [4902] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_block_repeat1, 2, 0, 0), SHIFT_REPEAT(2039),
  [4905] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_parenthesized_declarator, 4, 0, 0),
  [4907] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_parenthesized_declarator, 4, 0, 0),
  [4909] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 92),
  [4911] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 93),
  [4913] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 5, 0, 104),
  [4915] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 92),
  [4917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(894),
  [4919] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2052),
  [4921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2052),
  [4923] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 93),
  [4925] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__var_command_argument_list_non_hll_format_repeat1, 2, 0, 0),
  [4927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__var_command_argument_list_non_hll_format_repeat1, 2, 0, 0),
  [4929] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2198),
  [4931] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 0),
  [4933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1770),
  [4935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(887),
  [4937] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2104),
  [4939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2104),
  [4941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1572),
  [4943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1535),
  [4945] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2144),
  [4947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1196),
  [4949] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_parenthesized_declarator, 3, 0, 0),
  [4951] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_parenthesized_declarator, 3, 0, 0),
  [4953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1005),
  [4955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(998),
  [4957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(821),
  [4959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2068),
  [4961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2068),
  [4963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1751),
  [4965] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_with_default_format, 1, 0, 0),
  [4967] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_var_command_with_default_format, 1, 0, 0),
  [4969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1933),
  [4971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(829),
  [4973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2717),
  [4975] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_identifier, 1, 0, 1),
  [4977] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__command_identifier, 1, 0, 1),
  [4979] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__hll_pointer_declarator_specifier_repeat1, 2, 0, 0),
  [4981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_hll_argument_list_repeat1, 4, 0, 0),
  [4983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2722),
  [4985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_function_declarator, 2, 0, 57),
  [4987] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_function_declarator, 2, 0, 57),
  [4989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1313),
  [4991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2077),
  [4993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2077),
  [4995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1312),
  [4997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2621),
  [4999] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2083),
  [5001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2083),
  [5003] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_arguments, 7, 0, 0),
  [5005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2728),
  [5007] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__hll_pointer_declarator_specifier_repeat1, 2, 0, 0), SHIFT_REPEAT(1613),
  [5010] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 3, 0, 79),
  [5012] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 3, 0, 0),
  [5014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(950),
  [5016] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2090),
  [5018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2090),
  [5020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(917),
  [5022] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 6, 0, 104),
  [5024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2108),
  [5026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2109),
  [5028] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_arguments, 4, 0, 0),
  [5030] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_identifier, 2, 0, 1),
  [5032] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__command_identifier, 2, 0, 1),
  [5034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1546),
  [5036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1517),
  [5038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_hll_string_literal_repeat1, 2, 0, 0),
  [5040] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_hll_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(2104),
  [5043] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hll_string_literal_repeat1, 2, 0, 0), SHIFT_REPEAT(2104),
  [5046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1550),
  [5048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1551),
  [5050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1756),
  [5052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_arguments, 6, 0, 0),
  [5054] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_format_expression_repeat1, 2, 0, 74), SHIFT_REPEAT(2722),
  [5057] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_parenthesized_declarator, 5, 0, 0),
  [5059] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_parenthesized_declarator, 5, 0, 0),
  [5061] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hll_function_declarator, 3, 0, 77),
  [5063] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_function_declarator, 3, 0, 77),
  [5065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1040),
  [5067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1552),
  [5069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1554),
  [5071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1004),
  [5073] = {.entry = {.count = 1, .reusable = false}}, SHIFT(995),
  [5075] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__command_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(2717),
  [5078] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__command_identifier_repeat1, 2, 0, 0),
  [5080] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__command_identifier_repeat1, 2, 0, 0),
  [5082] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__host_command_argument_list_default_format_repeat1, 2, 0, 0), SHIFT_REPEAT(2136),
  [5085] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__host_command_argument_list_default_format_repeat1, 2, 0, 0),
  [5087] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__host_command_argument_list_default_format_repeat1, 2, 0, 0),
  [5089] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 4, 0, 79),
  [5091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(981),
  [5093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2174),
  [5095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2174),
  [5097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2113),
  [5099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2121),
  [5101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1574),
  [5103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1587),
  [5105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2062),
  [5107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2063),
  [5109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [5111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [5113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1588),
  [5115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1589),
  [5117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(993),
  [5119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1582),
  [5121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1590),
  [5123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1575),
  [5125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1576),
  [5127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2072),
  [5129] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hll_abstract_array_declarator, 2, 0, 0),
  [5131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2125),
  [5133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2126),
  [5135] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_block_repeat3, 2, 0, 0),
  [5137] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_block_repeat3, 2, 0, 0), SHIFT_REPEAT(2185),
  [5140] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_if_block_repeat3, 2, 0, 0), SHIFT_REPEAT(2185),
  [5143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__command_arguments, 5, 0, 0),
  [5145] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__host_command_argument_list_default_format_repeat2, 2, 0, 0),
  [5147] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__host_command_argument_list_default_format_repeat2, 2, 0, 0),
  [5149] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__host_command_argument_list_default_format_repeat2, 2, 0, 0), SHIFT_REPEAT(1777),
  [5152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1184),
  [5154] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2055),
  [5156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2055),
  [5158] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_macro_definition_repeat1, 2, 0, 18),
  [5160] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_macro_definition_repeat1, 2, 0, 18), SHIFT_REPEAT(2144),
  [5163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__command_arguments_repeat3, 2, 0, 0),
  [5165] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat3, 2, 0, 0), SHIFT_REPEAT(2198),
  [5168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_format_expression_repeat1, 2, 0, 33),
  [5170] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_format_expression_repeat1, 2, 0, 33),
  [5172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2103),
  [5174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2111),
  [5176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_with_default_format, 2, 0, 0),
  [5178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_var_command_with_default_format, 2, 0, 0),
  [5180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2188),
  [5182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2189),
  [5184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2190),
  [5186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2194),
  [5188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2203),
  [5190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2204),
  [5192] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__host_command_argument_list_default_format, 2, 0, 0),
  [5194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1026),
  [5196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2123),
  [5198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2123),
  [5200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [5202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2453),
  [5204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1995),
  [5206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [5208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1996),
  [5210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [5212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [5214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2234),
  [5216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1998),
  [5218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [5220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(606),
  [5222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(607),
  [5224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2243),
  [5226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(29),
  [5228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2244),
  [5230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(742),
  [5232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1979),
  [5234] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__other_command, 1, 0, 3),
  [5236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(31),
  [5238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(968),
  [5240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [5242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [5244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(28),
  [5246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2249),
  [5248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2021),
  [5250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1603),
  [5252] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_argument_list_default_format, 4, 0, 0),
  [5254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2446),
  [5256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [5258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(926),
  [5260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1059),
  [5262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2613),
  [5264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2614),
  [5266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1373),
  [5268] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2003),
  [5270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [5272] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2398),
  [5274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [5276] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2023),
  [5278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(625),
  [5280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2262),
  [5282] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_option_expression, 2, 0, 39),
  [5284] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_option_expression, 2, 0, 39), SHIFT(110),
  [5287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [5289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(630),
  [5291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2265),
  [5293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [5295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2266),
  [5297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [5299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [5301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(21),
  [5303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2268),
  [5305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [5307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2036),
  [5309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2626),
  [5311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2627),
  [5313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(815),
  [5315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1804),
  [5317] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_block_repeat2, 1, 0, 0), REDUCE(aux_sym_if_block_repeat3, 1, 0, 0),
  [5320] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_if_block_repeat2, 1, 0, 0), REDUCE(aux_sym_if_block_repeat3, 1, 0, 0),
  [5323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1730),
  [5325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(975),
  [5327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [5329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2387),
  [5331] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__var_command_argument_list_non_hll_format, 1, 0, 0),
  [5333] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2048),
  [5335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(646),
  [5337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2277),
  [5339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(648),
  [5341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(649),
  [5343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2280),
  [5345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [5347] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2281),
  [5349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(651),
  [5351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [5353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [5355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2283),
  [5357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [5359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2630),
  [5361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2639),
  [5363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2640),
  [5365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [5367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1809),
  [5369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [5371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1694),
  [5373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__var_command_argument_list_non_hll_format, 2, 0, 0),
  [5375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(662),
  [5377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2293),
  [5379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(988),
  [5381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(797),
  [5383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(674),
  [5385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(675),
  [5387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2297),
  [5389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(9),
  [5391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2298),
  [5393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(680),
  [5395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [5397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(11),
  [5399] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2300),
  [5401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(12),
  [5403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(875),
  [5405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1779),
  [5407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2137),
  [5409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [5411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2307),
  [5413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [5415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(99),
  [5417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1634),
  [5419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(696),
  [5421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [5423] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2310),
  [5425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [5427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2311),
  [5429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(699),
  [5431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [5433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [5435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2313),
  [5437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(105),
  [5439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1783),
  [5441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(710),
  [5443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2319),
  [5445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(747),
  [5447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2292),
  [5449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(712),
  [5451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(713),
  [5453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2322),
  [5455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [5457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2323),
  [5459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(715),
  [5461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [5463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [5465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2325),
  [5467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [5469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2054),
  [5471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(793),
  [5473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2396),
  [5475] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_block_repeat1, 2, 0, 0), SHIFT_REPEAT(2328),
  [5478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [5480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2403),
  [5482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(728),
  [5484] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2332),
  [5486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2009),
  [5488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(730),
  [5490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(731),
  [5492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2335),
  [5494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [5496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2336),
  [5498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(733),
  [5500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [5502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [5504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2338),
  [5506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(26),
  [5508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(931),
  [5510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1778),
  [5512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2010),
  [5514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1788),
  [5516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2176),
  [5518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [5520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(123),
  [5522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2011),
  [5524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2025),
  [5526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1508),
  [5528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(611),
  [5530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2238),
  [5532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2026),
  [5534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2419),
  [5536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [5538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2226),
  [5540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1732),
  [5542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2431),
  [5544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2417),
  [5546] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_argument_list_default_format, 6, 0, 0),
  [5548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2043),
  [5550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(771),
  [5552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2287),
  [5554] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat2, 2, 0, 0), SHIFT_REPEAT(2725),
  [5557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1641),
  [5559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [5561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2625),
  [5563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2628),
  [5565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [5567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2466),
  [5569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1642),
  [5571] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__function_identifier, 2, 0, 9),
  [5573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2700),
  [5575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(515),
  [5577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1718),
  [5579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(553),
  [5581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2015),
  [5583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1667),
  [5585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1669),
  [5587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1674),
  [5589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1632),
  [5591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2521),
  [5593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2328),
  [5595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2206),
  [5597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [5599] = {.entry = {.count = 1, .reusable = false}}, SHIFT(138),
  [5601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [5603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [5605] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hll_argument_list_repeat1, 2, 0, 0), SHIFT_REPEAT(521),
  [5608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1684),
  [5610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [5612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(146),
  [5614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(902),
  [5616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(795),
  [5618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(117),
  [5620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2487),
  [5622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2634),
  [5624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__function_identifier_repeat1, 2, 0, 0),
  [5626] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__function_identifier_repeat1, 2, 0, 0), SHIFT_REPEAT(2700),
  [5629] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_subroutine_block_repeat1, 2, 0, 0),
  [5631] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subroutine_block_repeat1, 2, 0, 0), SHIFT_REPEAT(2402),
  [5634] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_subroutine_block_repeat1, 2, 0, 0),
  [5636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [5638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(588),
  [5640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2424),
  [5642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2530),
  [5644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [5646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2222),
  [5648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2636),
  [5650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [5652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2511),
  [5654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_hll_parameter_list_repeat1, 2, 0, 0),
  [5656] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_hll_parameter_list_repeat1, 2, 0, 0), SHIFT_REPEAT(1603),
  [5659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1570),
  [5661] = {.entry = {.count = 1, .reusable = false}}, SHIFT(32),
  [5663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(113),
  [5665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2129),
  [5667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(34),
  [5669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2450),
  [5671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [5673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2468),
  [5675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(590),
  [5677] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_default_format_var_command, 1, 0, 3),
  [5679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(390),
  [5681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [5683] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2433),
  [5685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [5687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2434),
  [5689] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__non_hll_var_command, 1, 0, 3),
  [5691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1776),
  [5693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2060),
  [5695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(35),
  [5697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [5699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(593),
  [5701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [5703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(37),
  [5705] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2443),
  [5707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2528),
  [5709] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__hll_var_pattern_command_argument_list_format_repeat2, 2, 0, 0),
  [5711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [5713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1148),
  [5715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2610),
  [5717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2611),
  [5719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [5721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [5723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(155),
  [5725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1555),
  [5727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(33),
  [5729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [5731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2491),
  [5733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2714),
  [5735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2715),
  [5737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2561),
  [5739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2019),
  [5741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1010),
  [5743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1011),
  [5745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1801),
  [5747] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__command_identifier_repeat1, 2, 0, 21),
  [5749] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__command_identifier_repeat1, 2, 0, 21),
  [5751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [5753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(157),
  [5755] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_argument_list_default_format, 5, 0, 0),
  [5757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2568),
  [5759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2595),
  [5761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2573),
  [5763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(107),
  [5765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2042),
  [5767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2092),
  [5769] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2467),
  [5771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2600),
  [5773] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2476),
  [5775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2579),
  [5777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2624),
  [5779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2637),
  [5781] = {.entry = {.count = 1, .reusable = false}}, SHIFT(109),
  [5783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2057),
  [5785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(979),
  [5787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(796),
  [5789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [5791] = {.entry = {.count = 1, .reusable = false}}, SHIFT(241),
  [5793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(129),
  [5795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1733),
  [5797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1502),
  [5799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2673),
  [5801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [5803] = {.entry = {.count = 1, .reusable = false}}, SHIFT(239),
  [5805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1767),
  [5807] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2704),
  [5809] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2662),
  [5811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2662),
  [5813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(884),
  [5815] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_pattern_command_argument_list_format, 2, 0, 0),
  [5817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2442),
  [5819] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__var_definition_command_argument_list_format, 1, 0, 0),
  [5821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [5823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(250),
  [5825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2707),
  [5827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2707),
  [5829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(108),
  [5831] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__function_identifier_repeat1, 2, 0, 44),
  [5833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2691),
  [5835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2691),
  [5837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(948),
  [5839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2421),
  [5841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2721),
  [5843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2721),
  [5845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(122),
  [5847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2666),
  [5849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2666),
  [5851] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2651),
  [5853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2651),
  [5855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2399),
  [5857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2354),
  [5859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(501),
  [5861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(814),
  [5863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2358),
  [5865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(115),
  [5867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(131),
  [5869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(132),
  [5871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(133),
  [5873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(134),
  [5875] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_macro_definition_repeat1, 2, 0, 16),
  [5877] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_macro_definition_repeat1, 2, 0, 16),
  [5879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(128),
  [5881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(798),
  [5883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1616),
  [5885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_definition_command, 3, 0, 22),
  [5887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [5889] = {.entry = {.count = 1, .reusable = false}}, SHIFT(173),
  [5891] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_pattern_command_argument_list_format, 3, 0, 0),
  [5893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(402),
  [5895] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1803),
  [5897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1896),
  [5899] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_pattern_command_argument_list_format, 1, 0, 0),
  [5901] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_option_expression, 2, 0, 39), SHIFT(141),
  [5904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1492),
  [5906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2734),
  [5908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2734),
  [5910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(547),
  [5912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2410),
  [5914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(143),
  [5916] = {.entry = {.count = 1, .reusable = false}}, SHIFT(94),
  [5918] = {.entry = {.count = 1, .reusable = false}}, SHIFT(504),
  [5920] = {.entry = {.count = 1, .reusable = false}}, SHIFT(505),
  [5922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2660),
  [5924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2724),
  [5926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2703),
  [5928] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__command_arguments_repeat3, 2, 0, 0),
  [5930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2603),
  [5932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2603),
  [5934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2051),
  [5936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(145),
  [5938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1541),
  [5940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2681),
  [5942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(147),
  [5944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(148),
  [5946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(519),
  [5948] = {.entry = {.count = 1, .reusable = false}}, SHIFT(520),
  [5950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(149),
  [5952] = {.entry = {.count = 1, .reusable = false}}, SHIFT(150),
  [5954] = {.entry = {.count = 1, .reusable = false}}, SHIFT(151),
  [5956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(522),
  [5958] = {.entry = {.count = 1, .reusable = false}}, SHIFT(523),
  [5960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(152),
  [5962] = {.entry = {.count = 1, .reusable = false}}, SHIFT(153),
  [5964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(526),
  [5966] = {.entry = {.count = 1, .reusable = false}}, SHIFT(154),
  [5968] = {.entry = {.count = 1, .reusable = false}}, SHIFT(528),
  [5970] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1747),
  [5972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1748),
  [5974] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__var_command_with_non_hll_format, 3, 0, 0),
  [5976] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__var_command_with_non_hll_format, 3, 0, 0),
  [5978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1864),
  [5980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2737),
  [5982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(791),
  [5984] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2056),
  [5986] = {.entry = {.count = 1, .reusable = false}}, SHIFT(158),
  [5988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(548),
  [5990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(159),
  [5992] = {.entry = {.count = 1, .reusable = false}}, SHIFT(120),
  [5994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(130),
  [5996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(550),
  [5998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(144),
  [6000] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2360),
  [6002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1039),
  [6004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1593),
  [6006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__var_command, 1, 0, 5),
  [6008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2515),
  [6010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2131),
  [6012] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__host_command, 3, 0, 22),
  [6014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2692),
  [6016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2729),
  [6018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_non_default_format_var_command, 5, 0, 50),
  [6020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2698),
  [6022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2709),
  [6024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2078),
  [6026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [6028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1348),
  [6030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1527),
  [6032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1542),
  [6034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1573),
  [6036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2642),
  [6038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2135),
  [6040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2606),
  [6042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2647),
  [6044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2682),
  [6046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2689),
  [6048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2654),
  [6050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1578),
  [6052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1520),
  [6054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1591),
  [6056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1967),
  [6058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2205),
  [6060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1931),
  [6062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2633),
  [6064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2735),
  [6066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2641),
  [6068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2646),
  [6070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2680),
  [6072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2733),
  [6074] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_argument_list_non_default_format, 7, 0, 0),
  [6076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2219),
  [6078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2260),
  [6080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(407),
  [6082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [6084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1558),
  [6086] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__var_command, 1, 0, 4),
  [6088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(967),
  [6090] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command_argument_list_non_default_format, 6, 0, 0),
  [6092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2149),
  [6094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2150),
  [6096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2151),
  [6098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2152),
  [6100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1565),
  [6102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(562),
  [6104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2418),
  [6106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [6108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2093),
  [6110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [6112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1283),
  [6114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2693),
  [6116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1553),
  [6118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1561),
  [6120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1564),
  [6122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1566),
  [6124] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_command, 1, 0, 4),
  [6126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(980),
  [6128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1523),
  [6130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1524),
  [6132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1499),
  [6134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2525),
  [6136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [6138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2153),
  [6140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2157),
  [6142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1121),
  [6144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [6146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1559),
  [6148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1938),
  [6150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2180),
  [6152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2181),
  [6154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2182),
  [6156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2184),
  [6158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2186),
  [6160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2187),
  [6162] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__other_command, 3, 0, 22),
  [6164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2631),
  [6166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__host_command_with_default_format, 3, 0, 19),
  [6168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2110),
  [6170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(828),
  [6172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1136),
  [6174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_var_if_command, 3, 0, 20),
  [6176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1547),
  [6178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1548),
  [6180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1044),
  [6182] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__hll_var_definition_command_expression, 3, 0, 0),
  [6184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2505),
  [6186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2139),
  [6188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1495),
  [6190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1358),
  [6192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2213),
  [6194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [6196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(912),
  [6198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1174),
  [6200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_default_format_var_command, 3, 0, 22),
  [6202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1049),
  [6204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2727),
  [6206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2675),
  [6208] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_definition_command, 5, 0, 56),
  [6210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2459),
  [6212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(943),
  [6214] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [6216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1303),
  [6218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2367),
  [6220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [6222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2543),
  [6224] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__non_hll_var_command, 3, 0, 22),
  [6226] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_definition_command, 1, 0, 0),
  [6228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1606),
  [6230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__hll_var_call_expression_function, 3, 0, 42),
  [6232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1191),
  [6234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2688),
  [6236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1586),
  [6238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1885),
  [6240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1932),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token_label_identifier = 0,
  ts_external_token__and_operator_pre_hook = 1,
  ts_external_token__decimal_number = 2,
  ts_external_token__decimal_number_pre_hook = 3,
  ts_external_token_hll_number_literal = 4,
  ts_external_token__hll_number_literal_pre_hook = 5,
  ts_external_token_path = 6,
  ts_external_token_AMP_AMP = 7,
  ts_external_token_AMP = 8,
  ts_external_token__ampersand_char = 9,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token_label_identifier] = sym_label_identifier,
  [ts_external_token__and_operator_pre_hook] = sym__and_operator_pre_hook,
  [ts_external_token__decimal_number] = sym__decimal_number,
  [ts_external_token__decimal_number_pre_hook] = sym__decimal_number_pre_hook,
  [ts_external_token_hll_number_literal] = sym_hll_number_literal,
  [ts_external_token__hll_number_literal_pre_hook] = sym__hll_number_literal_pre_hook,
  [ts_external_token_path] = sym_path,
  [ts_external_token_AMP_AMP] = anon_sym_AMP_AMP,
  [ts_external_token_AMP] = anon_sym_AMP,
  [ts_external_token__ampersand_char] = sym__ampersand_char,
};

static const bool ts_external_scanner_states[20][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token_label_identifier] = true,
    [ts_external_token__and_operator_pre_hook] = true,
    [ts_external_token__decimal_number] = true,
    [ts_external_token__decimal_number_pre_hook] = true,
    [ts_external_token_hll_number_literal] = true,
    [ts_external_token__hll_number_literal_pre_hook] = true,
    [ts_external_token_path] = true,
    [ts_external_token_AMP] = true,
    [ts_external_token__ampersand_char] = true,
  },
  [2] = {
    [ts_external_token_label_identifier] = true,
    [ts_external_token_AMP_AMP] = true,
    [ts_external_token_AMP] = true,
  },
  [3] = {
    [ts_external_token_label_identifier] = true,
    [ts_external_token__and_operator_pre_hook] = true,
    [ts_external_token__decimal_number_pre_hook] = true,
    [ts_external_token_path] = true,
    [ts_external_token_AMP_AMP] = true,
    [ts_external_token_AMP] = true,
  },
  [4] = {
    [ts_external_token__and_operator_pre_hook] = true,
    [ts_external_token__decimal_number_pre_hook] = true,
    [ts_external_token_path] = true,
    [ts_external_token_AMP] = true,
  },
  [5] = {
    [ts_external_token__hll_number_literal_pre_hook] = true,
    [ts_external_token_AMP_AMP] = true,
    [ts_external_token_AMP] = true,
  },
  [6] = {
    [ts_external_token_label_identifier] = true,
    [ts_external_token__and_operator_pre_hook] = true,
    [ts_external_token_AMP_AMP] = true,
    [ts_external_token_AMP] = true,
  },
  [7] = {
    [ts_external_token__decimal_number_pre_hook] = true,
    [ts_external_token_path] = true,
    [ts_external_token_AMP] = true,
  },
  [8] = {
    [ts_external_token__hll_number_literal_pre_hook] = true,
    [ts_external_token_AMP] = true,
  },
  [9] = {
    [ts_external_token_label_identifier] = true,
    [ts_external_token__and_operator_pre_hook] = true,
    [ts_external_token_path] = true,
    [ts_external_token_AMP_AMP] = true,
    [ts_external_token_AMP] = true,
  },
  [10] = {
    [ts_external_token_AMP_AMP] = true,
    [ts_external_token_AMP] = true,
  },
  [11] = {
    [ts_external_token__and_operator_pre_hook] = true,
    [ts_external_token_path] = true,
  },
  [12] = {
    [ts_external_token__and_operator_pre_hook] = true,
  },
  [13] = {
    [ts_external_token__and_operator_pre_hook] = true,
    [ts_external_token_AMP] = true,
  },
  [14] = {
    [ts_external_token__decimal_number_pre_hook] = true,
    [ts_external_token_hll_number_literal] = true,
  },
  [15] = {
    [ts_external_token_path] = true,
  },
  [16] = {
    [ts_external_token_AMP] = true,
  },
  [17] = {
    [ts_external_token_AMP] = true,
    [ts_external_token__ampersand_char] = true,
  },
  [18] = {
    [ts_external_token_hll_number_literal] = true,
  },
  [19] = {
    [ts_external_token__decimal_number] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_t32_external_scanner_create(void);
void tree_sitter_t32_external_scanner_destroy(void *);
bool tree_sitter_t32_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_t32_external_scanner_serialize(void *, char *);
void tree_sitter_t32_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_t32(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym_identifier,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_t32_external_scanner_create,
      tree_sitter_t32_external_scanner_destroy,
      tree_sitter_t32_external_scanner_scan,
      tree_sitter_t32_external_scanner_serialize,
      tree_sitter_t32_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "t32",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 5,
      .minor_version = 5,
      .patch_version = 3,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
